/*
	$Id: F5.m,v 1.3 2015/04/02 05:30:02 kako Exp $
*/

proc poly(f);
begin
	return third(f);
end;


proc poly_list(R);
begin local l, r;
	l:={};
	foreach r in R do
		l:=append(l, {third(r)});
	od;
	return l;
end;

debug:=false;

proc check_dipoly_list(F);
begin
        local o, f, l;
	global debug;

        if not is_list(F) then
                if debug then error("not list", F);fi;
                return false;
        fi;
        f:=car(F);
        l:=cdr(F);
        if not is_dipoly(f) then
                if debug then error("not dipoly", f);fi;
                return false;
        fi;
        o:=dipoly_order(f);
        if o <> getdipolyorder() then
                if debug then error("order mismatch", o);fi;
                return false;
        fi;
        foreach f in l do
                if not is_dipoly(f) then
                        if debug then error("not dipoly", f);fi;
                        return false;
                fi;
                if o <> dipoly_order(f) then
                        if debug then error("order mismatch", f);fi;
                        return false;
                fi;
        od;
        return true;
end;


proc F5(F);
begin local G, rm, m;
	local i, fm;
	if not check_dipoly_list(F) then
		write "not dipoly list ", F;
		return {};
	fi;
	m:=length(F);
	ResetRules(m);

	F:=reverse(F);
	fm:=first(F);
	fm:=fm/HC(fm);
	F:=rest(F);
	rm:={1, m, fm};
	G:={rm};
write "f5: G", m, "=  ", G;
	for i:=m-1 downto 1 do
		fm:=first(F);
		fm:=fm/HC(fm);
		F:=rest(F);
		G:=F5_inc(i, fm, G);
write "f5: G", i, "=  ", G;
	od;
	return poly_list(G);
end;

proc F5_inc(i, f, Gp);
begin local ri;
	local r, g, s;
	local F, P, P2, Pd, G, L;
	local p, Rd, d;

	ri:={dipoly(1), i, f};
	G:=append(Gp, {ri});
	L:={};
	foreach r in Gp do
		g:=CritPair(ri, r, i, Gp);
		if g <> {} then L:=append(L, {g});
		fi;
	od;
	P:=sort_by_degree(L);
	while P <> {} do
		r:=first(P);
	  	d:=tdeg(first(r));

write "d=", d;
write "critical pairs = ", P;
		Pd:={};
		foreach g in P do
			if tdeg(first(g)) = d then
				Pd:=append(Pd, {g});
			fi;
		od;
		write "P",d,"=", Pd;
		P:=setdiff(P, Pd);
		F:=Spol(Pd);
		write "call reduction: F=", F;
		Rd:=Reduction(F, G, i, Gp);
		write "R",d,"=", Rd;
		L:={};
		foreach r in Rd do
			foreach g in G do
				s:=CritPair(r, g, i, Gp);
				if s<> {} then
					L:=append(L, {s});
				else
					write "pair ", r, "  and ", g, " is rejected";
				fi;
			od;
			/*P:=append(P, L);*/
			G:=append(G, {r});
		od;
		L:=sort_by_degree(L);
		write "L = ", L;
		write "G = ", G;
		P:=merge_by_degree(P, L);
		/*P:=sort_by_degree(P);*/
	od;
	return G;
end;
				
proc CritPair(r1, r2, k, G);
begin local r;
	r:=CritPair2(r1, r2, k, G);
	write "Pair(", r1," , ", r2, ") --> ",r;
	return r;
end;
				
proc CritPair2(r1, r2, k, G);
begin local p1, p2, t;
	local u1, u2, s1, s2;
	local u, v, k2, t2, k1, t1;
	local rtmp;

	p1:=third(r1);
	p2:=third(r2);
	t:=lcm(HT(p1), HT(p2));
	u1:=dipoly(t/HT(p1));
	u2:=dipoly(t/HT(p2));
	if second(r1) > second(r2) then 
		rtmp := r1;
		r1:=r2;
		r2:=rtmp;
		rtmp:=u1;
		u1:=u2;
		u2:=rtmp;
	elif second(r1) = second(r2) 
		and compare(u1*first(r1), u2*first(r2)) < 0 then
		rtmp:=r1;
		r1:=r2;
		r2:=rtmp;
		rtmp:=u1;
		u1:=u2;
		u2:=rtmp;
	fi;
	t1:=first(r1); k1:=second(r1);
	t2:=first(r2); k2:=second(r2);
	if k1 > k then return {};
	fi;
	u:=t1*u1;
	v:=NF(u, G);
write "NF(",u, ",", G, ") ==> ", v;
	if u <> v then return {};
	fi;
	if k2 = k then
		u:=t2*u2;
		v:=NF(u, G);
write "NF(",u, ",", G, ") ==> ", v;
		if u <> v then return {};
		fi;
	fi;
	return {dipoly(t), u1, r1, u2, r2};
end;
	
proc Spol(P);
begin local F;
	local ri, rj, r;
	local i, j, ul, vl, tl;
	local p;
	global N;

	F:={};
	foreach p in P do
		tl:=first(p);
		ul:=second(p);
		ri:=third(p);
		vl:=nth(p, 4);
		rj:=nth(p, 5);
		if IsRewritten(ul, ri) = false  and IsRewritten(vl, rj) = false then
			N:=N+1;
			vl:=vl*HC(poly(ri))/HC(poly(rj));
			vl:=ul*poly(ri) - vl*poly(rj);
			vl:=vl/HC(vl);
			r:={ul*first(ri), second(ri), vl};
write "Spol: addrules ", N;
			AddRules(r, N);
			F:=append(F, {r});
		fi;
	od;
	F:=sort_by_degree(F);
	return F;
end;

proc NF(f, F);
begin local g, r;

	g:=foreach r in F collect third(r);
	return reduce(f, g);
end;

proc NF2(f,F);
begin local r, g, n;
	local i, j;
	local c;

	n:=length(F);
	array g[n];

	for i:=0 to n-1 do
		g[i]:=third(first(F));
		F:=rest(F);
	od;
	r:=0;
	while f <> 0 do
		for j:=0 to n-1 do
			c:=ht_div(f, g[j]);
			if c <> false then break;
			fi;
		od;
		if c = false then
			c:=mlterm(f);
			r:=r+c;
			f:=f-c;
		else
			c:=dipoly(c);
			f:=f-g[j]*c;
		fi;
	od;
	return r;
end;

proc ht_div(f, g);
begin 
	local r;
	r = mlterm(f)/mlterm(g);
	if is_number(r) or is_poly(r) then
		return dipoly(r);
	fi;
	return false;
end;

proc HM(f);
begin
	return mlterm(f);
end;
	
proc HT(f);
begin  
	return mpower(f);
end;

proc HC(f);
begin
	return mlcoeff(f);
end;
proc ResetRules(m);
begin local i;
	global N, Rules;
	vector Rules[m];
	N:=m;

	for i:=1 to m do
		Rules[i]:={};
	od;
	
end;

proc AddRules(r, n); /* r = {t, i, p} */
		/* t: term */
		/* i: index Fi */
		/* p: polynomial */
begin local i;
	global Rules;
	i:=second(r);
write "add rules[",i,"]=", {r,n};
 	i:= second(r);
	Rules[i]:=cons({first(r), r, n}, Rules[i]);
end;

proc Rewritten(u, rk);
begin local L, i, t, ti, d, l;
	local ut;
	global Rules;

 	i:=second(rk);
 	t:=first(rk);
	ut:=t*u;
	L:=Rules[i];
write "Rewritten: rule = ", L;
	foreach l in L do
    		ti:=first(l);
		d:=term_div(ut, ti);
		if d then
			return {d, second(l)};
    		fi;
	od;
	return {u, rk};
end;

proc term_div(f, g);
begin local a;
  	a:=f/g;
  	if is_number(a) or is_poly(a) then return dipoly(a);
  	fi;
  	return false;
end;

proc IsRewritten(u, rk);
begin local l, rl;
	local r;
	l:=Rewritten(u,rk);
	rl:=second(l);
	if rl <> rk then r:=true;
	else r:=false;
	fi;
write "isRewritten :", {u, rk}, " --> ", l, r;
	return r;
end;
proc merge_by_degree(u, v);
begin
	local P;
	movd('merge_sort_compare_function', 'compare_by_degree');
	P:=merge_sort(u, v);
	return P;
end;

proc merge_sort(u, v);
begin local a, b, w;
	w := {};
	while u <> {} and v <> {} do
		a := first(u);
		b := first(v);
		if merge_sort_compare_function(a, b) then
			w:=cons(a, w);
			u:=rest(u);
			continue;
		else
			w:=cons(b, w);
			v:=rest(v);
			continue;
		fi;
	od;
	w:=reversip(w);
	w:=append(w, u);
	w:=append(w, v);
	return w;
end;
		
proc sort_by_signature(u);
begin
	local a, i, n, s;
	movd('quick_sort_compare_function', 'compare_by_signature');
	n := length(u);
	array a[n];
	list2array(a, u);
	quick_sort(a, 0, n-1);
	s:=nil;
	for i:=n-1 downto 0 do
		s:=cons(a[i], s);
	od;
	return s;
end;

proc sort_by_degree(u);
begin
	local a, i, n, s;
	movd('quick_sort_compare_function', 'compare_by_degree');
	n := length(u);
	if n <= 0 then return u;
	fi;
	array a[n];
	list2array(a, u);
	quick_sort(a, 0, n-1);
	s:=nil;
	for i:=n-1 downto 0 do
		s:=cons(a[i], s);
	od;
	return s;
end;

/*
proc quick_sort(a, l, r);
begin local ls, rs; 
	local pivot;
	local ll;


	rs:=r;
	ls:=l;
	pivot := a[l];

	while l < r do
		while quick_sort_compare_function(pivot, a[r]) and l < r do
			r := r - 1;
		od;
		if r <>  l then
			a[l] := a[r];
			l := l + 1;
		fi;
		while quick_sort_compare_function(a[l], pivot) and l < r do
			l := l + 1;
		od;
		if l <> r then
			a[r] := a[l];
			r := r - 1;
		fi;
	od;
	a[l] := pivot;
	ll:=l;
	l := ls;
	r:=rs;
	if l < ll then
		quick_sort(a, l, ll-1);
	fi;
	if r > ll then
		quick_sort(a, ll+1, r);
	fi;
	return a;
end;
*/

proc quick_sort(a, l, r);
begin local i, j, pivot, t;
	local m;
	i := l;
	j := r;
	m:=quotient(i+j, 2);
	pivot:=a[m];
	while i <= j do
		while quick_sort_compare_function(a[i], pivot) do
			i:=i+1;
		od;
		while quick_sort_compare_function(pivot, a[j]) do
			j:=j-1;
		od;
		if i <= j then
			t:=a[i];
			a[i]:=a[j];
			a[j]:=t;
			i:=i+1;
			j:=j-1;
		fi;
	od;
	if l < j then
		quick_sort(a, l, j);
	fi;
	if i < r then
		quick_sort(a, i, r);
	fi;
	return a;
end;


proc compare_by_degree(u, v);
begin 
/*write "compare ", u, " and ", v;*/
	if tdeg(car(u)) < tdeg(car(v)) then
		return true;
	fi;
	return nil;
end;

proc compare_by_signature(u, v);
begin 
	local ui, vi;
	ui:=second(u); vi:=second(v);
	if ui > vi then return true;
	elif ui < vi then return false;
	fi;
	if compare(first(u), first(v)) < 0 then
		return true;
	fi;
	return false;
end;
proc compare_signature(u, v);
	/* 1: u > v 
          -1: u < v
           0: u = v
         */
begin
	local ut, vt, ui, vi;
	ut:=first(u); ui:=second(u);
	vt:=first(v); vi := second(v);
	if ui < vi then return 1;
	elif ui > vi then return -1;
	else return compare(ut, vt);
	fi;
end;

proc minimal_by_signature(u);
begin local v, w;
/*write "minimal_by_signature: u=", u;*/
	if u = {} then return u;
	fi;
	v:=first(u);
	u:=rest(u);
	foreach w in u do
		if compare_signature(v, w) = 1 then v:=w;
		fi;
	od;
/*write "minimal_by_signature: return ", v;*/
	return v;
end;


proc Reduction(Todo, G, k, Gp);
begin
	local Done, h, hh;
	local To;
	To := Todo;
	Done:={};
	while Todo <> {} do
		h:=minimal_by_signature(Todo);
		Todo:=delete(h, Todo);
		hh:=TopReduction(NF_with_signature(h, Gp), append(G, Done), k, Gp);
		/*write "reduction:", hh;*/
		Done:=append(Done, first(hh));
		Todo:=append(Todo, second(hh));
	od;
write "Reduction: ", To, " --> ", Done;
	return Done;
end;

proc IsReducible(r0, G, k, Gp);
begin
	local rj;
	local tj, kj, u1, u2, k0, t0;
	local u;

	t0 := first(r0);
	k0 := second(r0);

	foreach rj in G do
		tj:=first(rj);
		kj:=second(rj);
		u:=HM(third(r0))/HM(third(rj));
		if not (is_poly(u) or is_number(u)) then continue;
		fi;

		u:=dipoly(u);
		u1:=u*tj;
		u2:=NF(u1, Gp);
		if u1 <> u2 or IsRewritten(u, rj) 
			or (kj = k0 and u1 = t0) then continue;
		fi;

		return rj;
	od;
	return 0;
end;

proc TopReduction(r0, G, k, Gp);
begin
	local rp, rk, r;
	local pk, k0, k1;
	local u;
	global N;

	if poly(r0) = 0 then
		error("the system is not a regular sequence",nil);
		return {{}, {}};
	fi;
	rp:=IsReducible(r0, G, k, Gp);
	if rp = 0 then
		return {{{first(r0), second(r0), third(r0)/HC(third(r0))}}, {}};
	fi;
	rk:=rp;
	u:=HM(third(r0))/HM(third(rk));
		u:=dipoly(u);
	k0:=second(r0);
	k1:=second(rk);
	if k1 > k0 or (k1 = k0 and compare(u*first(rk), first(r0)) < 0) then
		pk:=poly(r0) - u*poly(rk);
		r0:={first(r0), second(r0), pk};
		return {{}, {r0}};
	else
		N:=N+1;
		r:={u*first(rk), second(rk), u*poly(rk)-poly(r0)};
write "TopReduction: addrules ", N;
		AddRules(r, N);
		return {{}, {r, r0}};
	fi;
end;


proc NF_with_signature(f,F);
begin local r, g, n;
        local i, j;
        local c;
	local rj, rt, gt, gj;

/*write "nf_with_signature: f=", f, ", F=", F;*/

        n:=length(F);
        array g[n];

        for i:=0 to n-1 do
                g[i]:=first(F);
                F:=rest(F);
        od;

	rt:=first(f);
	rj:=second(f);
	f:=third(f);
	r:=0;

        while f <> 0 do
                for j:=0 to n-1 do
                        c:=ht_div(f, third(g[j]));
                        if c <> false then break;
                        fi;
                od;
                if c = false then
                        c:=mlterm(f);
                        r:=r+c;
                        f:=f-c;
                else
			c:=dipoly(c);
			gt:=first(g[j]) *c;
			gj:=second(g[j]);
                        f:=f-third(g[j])*c;
			if rj > gj then
				rj:=gj;
				rt:=gt;
			elif rj=gj and compare(rt, gt) < 0 then rt:=gt;
			fi;
                fi;
        od;
/*write "nf_with_signature: r=", r;*/
        return {rt, rj, r};
end;

proc bench_cyclic(n);
begin
	local g, v, i, tim0, tim1;
	local h;
	global c;
	g:=cyclic(n);
	v:=for i:=0 to n-1 collect makeid(c, i);
	setorder_tlex(v);
	g:=dipoly(g);
	tim0:=time();
	h:=F5(g);
	tim1:=time();
	if isgb(h, g) = false then
		write "calculation failed";
	else
		write "computation time = ", tim1 - tim0; 
	fi;
	return h;
end;

proc bench_katsura(n);
begin
	local g, v, i, tim0, tim1;
	local h;
	global u;
	g:=katsura(n);
	v:=for i:=0 to n collect makeid(u, i);
	setorder_tlex(v);
	g:=dipoly(g);
	tim0:=time();
	h:=F5(g);
	tim1:=time();
	if isgb(h, g) = false then
		write "calculation failed";
	else
		write "computation time = ", tim1 - tim0; 
	fi;
	return h;
end;


proc cyclic(n);
begin
	local i, j, l, r;
	local a, b;
	r:={};
	for l:=1 to n do
		a:=0;
		for i:=0 to n-1 do
			b:=1;
			for j:=i to l+i-1 do
				b:=b*mkc(j, n);
			od;
			a:=a+b;
		od;
		if l = n then
			a:=a-1;
		fi;
		r := cons(a, r);
	od;
	return r;
end;


proc katsura(n);
begin
	local r, l, a;
	local i, k;
	r:={};
	for l:=0 to n-1 do
		a:=-1*mku(l, n);
		for i:=-n to n do
			a:= a + mku(l-i, n)*mku(i, n);
		od;
		r:=cons(a, r);
	od;
	a:=0;
	for l:=-n to n do
		a:=a+mku(l, n);
	od;
	r := cons(a-1, r);
	return r;
end;

proc mkc(l, n);
begin
	global c;
	if l >= n then l := l - n; fi;
	if l >= 0 then return makeid(c, l);
	else	return 0;
	fi;
end;

proc mku(l, n);
begin
	global u;
	if l < 0 then l:=-l; fi;
	if l <= n then return makeid(u, l);
	fi;
	return 0;
end;

f:={y*z^3-x^2*t^2, x*z^2-y^2*t, x^2*y-z^2*t};
setorder revlex x,y,z,t;
g:=dipoly(f);

proc test(g);
begin
	return F5(g);
end;

proc test2(g);
begin local g1, g2, g3;
	local r1, r2, r3, r4, r5;
	local G;
global N;


	ResetRules(3);
	g1:=first(g);
	g2:=second(g);
	g3:=third(g);
	r1:={dipoly(1), 1, g1};
	r3:={dipoly(1), 3, g3};
	r2:={dipoly(1), 2, g2};
	r4:={dipoly(x*y), 2, dipoly(-x*y^3*t+z^4*t)};
	r5:={dipoly(x*y*z^2), 2, dipoly(z^6*t-y^5*t^2)};;
	G:={r3, r2, r4, r5};	
	AddRules(r4, 4);
	AddRules(r5, 5);
	N:=5; 
  G:=F5(1, g1, G);
	return G;
end;
