/*
 * algnum.c                (C) Copyright 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/algnum.c,v 2.7 2015/04/23 04:04:47 kako Exp $
 */

static char const rcsid[] =
    "$Id: algnum.c,v 2.7 2015/04/23 04:04:47 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "entries.h"
#include "polynomial.h"
#include "fns.h"

#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)


Sex             negatep(Sex);
Sex             modulusp(Sex, Sex);
Sex             smodulusp(Sex, Sex);

#define ONEP(x)                onep(x)
#define ZEROP(x)        zerop(x)


/***************************************************************/

void            print_algnum(Sex a, putchar_fn * fn);
Sex             Lis_algnum(Sex v);
Sex             mkalgnum(Sex v, Sex d);
Sex             make_algnum(Sex v, Sex p, Sex d);
static Sex      mk_algebraic(Sex args, Sex env);
Sex             gplus2an(Sex a, Sex b);
Sex             gplus2na(Sex a, Sex b);
Sex             gplus2aa(Sex a, Sex b);
Sex             gdiff2an(Sex a, Sex b);
Sex             gdiff2na(Sex a, Sex b);
Sex             gdiff2aa(Sex a, Sex b);
Sex             gnegatea(Sex a);
Sex             gmodulusa(Sex a, Sex b);
Sex             gsmodulusa(Sex a, Sex b);
Sex             adivide(Sex a, Sex b);
Sex             aremainder(Sex a, Sex b);
Sex             gtimes2an(Sex p, Sex t);
Sex             gtimes2na(Sex t, Sex p);
Sex             gtimes2aa(Sex a, Sex b);
Sex             exptia(Sex a, long n, Sex d);
Sex             gexpt2ai(Sex a, Sex b);
Sex             quot2ax(Sex a, Sex b);
Sex             gquot2an(Sex a, Sex b);
Sex             gquot2na(Sex a, Sex b);
Sex             gquot2aa(Sex a, Sex b);
bool            zeropa(Sex a);
bool            onepa(Sex a);
Sex             recipa(Sex a);

/***************************************************************/
/***************************************************************/

void
print_algnum(Sex a, putchar_fn * fn)
{
    stackcheck();
    push(a);
    fn('(');
    pop(a);
    a = make_poly(algnum_var(a), algnum_pol(a));
    print_sex(a, fn);
    fn(')');
}

/***************************************************************/
/***************************************************************/
Sex
Lis_algnum(Sex a)
{
    return Bool_to_Sex(is_algnum(a));
}
/***************************************************************/
/***************************************************************/

Sex
mkalgnum(Sex v, Sex d)
{
    Sex             c,
                    x;

    stackcheck();
    x = v;
    if (is_poly(v)) {
	x = poly_var(v);
    } else {
	push2(x, d);
	v = make_polynomial(x, mkfxat(1), mkfxat(1));
	pop2(d, x);
    }
    push2(x, v);
    kernel_order = cons(x, kernel_order);
    d = Lreorder(d);
    push(d);
    kernel_order = cdr(kernel_order);
    c = getvector(TAG_VECTOR, T_ALGNUM, sizeof(Algebraic_Number));
    x = stack[-2];
    d = stack[-0];
    if (!is_poly(d) || poly_var(d) != x) {
	popv(3);
	return error1("mkalgnum: invalid defining polynomial", d);
    }
    x = stack[-2];
    x = Lgen_sharp_symbol(x);
    d = stack[0];
    v = stack[-1];
    // pop3(d, v, x);
    popv(3);
    algnum_var(c) = x;
    algnum_pol(c) = poly_pol(v);
    algnum_def(c) = poly_pol(d);
    return c;
}

Sex
make_algnum(Sex v, Sex p, Sex d)
{
    Sex             c;

    if (p == nil || zerop(p))
	return mkfxat(0);
    c = lterm(p);
    if (reduct(p) == nil && power(c) == mkfxat(0))
	return coeff(c);
    if (is_poly(v))
	v = poly_var(v);
    stackcheck();
    push3(v, p, d);
    c = getvector(TAG_VECTOR, T_ALGNUM, sizeof(Algebraic_Number));
    pop3(d, p, v);
    algnum_var(c) = v;
    algnum_pol(c) = p;
    algnum_def(c) = d;
    return c;
}


static          Sex
mk_algebraic(Sex args, Sex env)
{
    Sex             var = nil,
	val = nil;
    stackcheck();
    if (is_cons(args)) {
	var = car(args);
	if (!is_symbol(var) || var == nil || var == true)
	    return aerror("setq (bad variable)");
	args = cdr(args);
	push(var);
	if (is_cons(args)) {
	    val = car(args);

	    val = eval(val, env);

	    var = stack[0];
	}
	val = mkalgnum(var, val);
	pop(var);
	value(var) = val;

    }
    return var;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/




/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
gplus2an(Sex a, Sex b)
{
    Sex             p;
    Sex             d;

    if (ZEROP(b))
	return a;
    p = algnum_pol(a);
    stackcheck();
    push(a);
    p = plus2xp(b, p);
    pop(a);
    return make_algnum(algnum_var(a), p, algnum_def(a));
}



Sex
gplus2na(Sex a, Sex b)
{
    Sex             p;
    Sex             d;

    if (ZEROP(a))
	return b;
    p = algnum_pol(b);
    stackcheck();
    push(b);
    p = plus2xp(a, p);
    pop(b);
    return make_algnum(algnum_var(b), p, algnum_def(b));
}




Sex
gplus2aa(Sex a, Sex b)
{
    long            o;

    o = compare_var(algnum_var(a), algnum_var(b));
    if (o > 0)
	return gplus2na(a, b);
    else if (o < 0)
	return gplus2na(b, a);
    stackcheck();
    push(a);
    b = plus2pp(poly_pol(a), poly_pol(b));
    pop(a);
    return make_algnum(algnum_var(a), b, algnum_def(a));
}



Sex
gdiff2an(Sex a, Sex b)
{
    Sex             p;
    Sex             d;

    if (ZEROP(b))
	return a;
    p = algnum_pol(a);
    stackcheck();
    push(a);
    p = diff2px(p, b);
    pop(a);
    return make_algnum(algnum_var(a), p, algnum_def(a));
}



Sex
gdiff2na(Sex a, Sex b)
{
    Sex             p;
    Sex             d;

    if (ZEROP(a))
	return gnegatea(b);
    p = algnum_pol(b);
    stackcheck();
    push(b);
    p = diff2xp(a, p);
    pop(b);
    return make_algnum(algnum_var(b), p, algnum_def(b));
}




Sex
gdiff2aa(Sex a, Sex b)
{
    long            o;

    o = compare_var(algnum_var(a), algnum_var(b));
    if (o > 0)
	return gdiff2na(a, b);
    else if (o < 0)
	return gdiff2an(a, b);
    stackcheck();
    push(a);
    b = diff2pp(algnum_pol(a), algnum_pol(b));
    pop(a);
    return make_algnum(algnum_var(a), b, algnum_def(a));
}





Sex
gnegatea(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = negatep(algnum_pol(a));
    pop(a);
    return make_algnum(algnum_var(a), c, algnum_def(a));
}

Sex
gmodulusa(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push(a);
    c = modulusp(algnum_pol(a), b);
    pop(a);
    return make_algnum(algnum_var(a), c, algnum_def(a));
}

Sex
gsmodulusa(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push(a);
    c = smodulusp(algnum_pol(a), b);
    pop(a);
    return make_algnum(algnum_var(a), c, algnum_def(a));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

#define putv(v,i,a) vector_value(v,i) = a
#define getv(v,i)   vector_value(v,i)

Sex
adivide(Sex a, Sex b)
{
    long            i,
                    j,
                    k;
    long            n,
                    m;

    if (a == nil)
	return a;
    m = mkint(power(lterm(a)));
    n = mkint(power(lterm(b)));
    k = m - n;
    if (k < 0) {
	b = a;
	a = nil;
	return cons(a, b);
    }

    push5(nil, nil, nil, a, b);
    push4(nil, nil, nil, nil);

#define q0        stack[-8]
#define cu         stack[-7]
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define u stack[-2]
#define w stack[-1]
#define cv stack[0]

    u = Lmkvect(mkfxat(m));
    for (i = m; i >= 0; i--) {
	if (a != nil && power(lterm(a)) == mkfxat(i)) {
	    putv(u, i, coeff(lterm(a)));
	    a = reduct(a);
	} else
	    putv(u, i, mkfxat(0));
    }
    w = Lmkvect(mkfxat(n));
    for (i = n; i >= 0; i--) {
	if (b != nil && power(lterm(b)) == mkfxat(i)) {
	    putv(w, i, coeff(lterm(b)));
	    b = reduct(b);
	} else
	    putv(w, i, mkfxat(0));
    }
    cv = getv(w, n);
    for (; k >= 0; k--) {
	cu = getv(u, n + k);
	if (!zerop(cu)) {
	    cu = gratio2(cu, cv);
	    b = cons(cu, mkfxat(k));
	    q0 = cons(b, q0);
	    for (j = n + k - 1; j >= k; j--) {
		b = gtimes2(cu, getv(w, j - k));
		b = gdiff2(getv(u, j), b);
		putv(u, j, b);
	    }
	}
    }
    a = Lnreverse(q0);
    b = nil;
    for (k = 0; k < n; k++) {
	if (!zerop(getv(u, k))) {
	    w = cons(getv(u, k), mkfxat(k));
	    b = cons(w, b);
	}
    }
#undef q0
#undef cu
#undef x
#undef u
#undef w
#undef cv
#undef a
#undef b
    a = stack[-5];
    b = stack[-4];
    popv(9);
    return cons(a, b);
}


Sex
aremainder(Sex a, Sex b)
{
    long            i,
                    j,
                    k;
    long            n,
                    m;
    Sex             c;

    m = mkint(power(lterm(a)));
    n = mkint(power(lterm(b)));
    k = m - n;
    if (k < 0)
	return a;

    push5(nil, nil, nil, a, b);
    push(nil);

#define cu         stack[-5]
#define u stack[-4]
#define w stack[-3]
#define a stack[-2]
#define b stack[-1]
#define cv stack[0]

    u = Lmkvect(mkfxat(m));
    c = u;
    for (i = m; i >= 0; i--) {
	if (a != nil && power(lterm(a)) == mkfxat(i)) {
	    putv(c, i, coeff(lterm(a)));
	    a = reduct(a);
	} else
	    putv(c, i, mkfxat(0));
    }
    w = Lmkvect(mkfxat(n));
    c = w;
    for (i = n; i >= 0; i--) {
	if (b != nil && power(lterm(b)) == mkfxat(i)) {
	    putv(c, i, coeff(lterm(b)));
	    b = reduct(b);
	} else
	    putv(c, i, mkfxat(0));
    }
    cv = getv(w, n);
    for (; k >= 0; k--) {
	cu = getv(u, n + k);
	if (!zerop(cu)) {
	    cu = gratio2(cu, cv);
	    // lisp_printf("cu = %p\n", cu);
	    for (j = n + k - 1; j >= k; j--) {

		a = gtimes2(cu, getv(w, j - k));
		a = gdiff2(getv(u, j), a);
		putv(u, j, a);
	    }
	}
    }

    b = nil;
    for (k = 0; k < n; k++) {
	if (!zerop(getv(u, k))) {
	    a = cons(getv(u, k), mkfxat(k));
	    b = cons(a, b);
	}
    }
#undef cu
#undef u
#undef w
#undef cv
#undef a
#undef b
    b = stack[-1];
    popv(6);
    return b;
}

Sex
gtimes2an(Sex p, Sex t)
{
    if (ZEROP(t))
	return t;
    if (ONEP(t))
	return p;
    stackcheck();
    push(p);
    t = times2xp(t, algnum_pol(p));
    pop(p);
    return make_algnum(algnum_var(p), t, algnum_def(p));
}

Sex
gtimes2na(Sex t, Sex p)
{
    // lisp_printf("gtimes2na: t = %p, p = %p\n", t, p);
    if (ZEROP(t))
	return t;
    if (ONEP(t))
	return p;
    stackcheck();
    push(p);
    t = times2xp(t, algnum_pol(p));
    pop(p);
    return make_algnum(algnum_var(p), t, algnum_def(p));
}


Sex
gtimes2aa(Sex a, Sex b)
{
    long            o;

    o = compare_var(algnum_var(a), algnum_var(b));
    if (o > 0)
	return gtimes2na(a, b);
    else if (o < 0)
	return gtimes2na(b, a);
    stackcheck();
    push(a);
    b = times2pp(algnum_pol(a), algnum_pol(b));
    a = stack[0];
    // lisp_printf("b = %p\n", b);
    // lisp_printf("a = %p\n", algnum_def(a));
    if (!lessp2(power(lterm(b)), power(lterm(algnum_def(a))))) {
	b = aremainder(b, algnum_def(a));
    }
    // lisp_printf("b = %p\n", b);
    pop(a);

    return make_algnum(algnum_var(a), b, algnum_def(a));
}



Sex
exptia(Sex a, long n, Sex d)
{
    Sex             b;

    if (n == 1)
	return a;

    stackcheck();

    if ((n & 1) == 0) {
	push(d);
	a = exptia(a, n / 2, d);
	a = times2pp(a, a);
    } else {
	push2(d, a);
	b = exptia(a, n / 2, d);
	b = times2pp(b, b);
	d = stack[-1];
	b = aremainder(b, d);
	pop(a);
	a = times2pp(a, b);
    }
    pop(d);
    a = aremainder(a, d);
    return a;
}


Sex
gexpt2ai(Sex a, Sex b)
{
    long            n = mkint(b);

    if (n == 0)
	return mkfxat(1);
    else if (n == 1)
	return a;
    stackcheck();
    push(a);
    b = exptia(algnum_pol(a), n, algnum_def(a));
    pop(a);
    return make_algnum(algnum_var(a), b, algnum_def(a));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
quot2ax(Sex a, Sex b)
{
    if (b == nil)
	aerror2("quotient", a, b);
    if (a == nil)
	return nil;
    stackcheck();
    push5(a, b, nil, nil, nil);
    push(nil);
#define a stack[-5]
#define b stack[-4]
#define c stack[-3]
#define d stack[-2]
#define e stack[-1]
#define f stack[0]
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = gratio2(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);

#undef a
#undef b
#undef c
#undef d
#undef e
#undef f
    a = cdr(stack[0]);
    popv(6);
    return a;
}


Sex
gquot2an(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push(a);
    c = quot2ax(algnum_pol(a), b);
    pop(a);
    return make_algnum(algnum_var(a), c, algnum_def(a));
}



Sex
gquot2na(Sex a, Sex b)
{
    push(a);
    b = recipa(b);
    pop(a);
    return gtimes2na(a, b);
}

Sex
gquot2aa(Sex a, Sex b)
{
    push(a);
    b = recipa(b);
    pop(a);
    return gtimes2aa(a, b);
}


/***************************************************************/
/***************************************************************/
Sex
ggcd2aa(Sex a, Sex b)
{
    return mkfxat(1);
}

Sex
ggcd2an(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    return mkfxat(1);
}

Sex
ggcd2na(Sex a, Sex b)
{
    if (zerop(a))
	return b;
    return mkfxat(1);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
zeropa(Sex a)
{
    return a == nil;
}

bool
onepa(Sex a)
{
    if (reduct(a) != nil)
	return NO;
    a = lterm(a);
    return power(a) == mkfxat(0);
}


Sex
recipa(Sex a)
{
    Sex             m,
                    n;

    m = algnum_pol(a);
    n = algnum_def(a);

    push5(a, m, n, nil, nil);
    push2(nil, nil);
#define c stack[-6]
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define y stack[-2]
#define w stack[-1]
#define t stack[0]
    x = Lncons(cons(mkfxat(1), mkfxat(0)));
    y = nil;
    while (!zeropa(b)) {
	// lisp_printf("recipa: a = %p, b=%p\n", a, b);
	w = adivide(a, b);
	// lisp_printf("recipa: w = %p\n", w);
	a = b;
	b = cdr(w);		/* 
				 * remainder 
				 */
	t = y;
	w = times2pp(y, car(w));
	y = diff2pp(x, w);
	x = t;
    }
    // lisp_printf("x = %p, a = %p\n", x, a);
    if (power(lterm(a)) != mkfxat(0)) {
	error1("recipa: divide error", a);
    }
    a = coeff(lterm(a));
    n = x;
    if (!onep(a))
	n = quot2ax(n, a);
    m = c;
#undef c
#undef a
#undef b
#undef x
#undef y
#undef w
#undef t
    popv(7);
    return make_algnum(algnum_var(m), n, algnum_def(m));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


setup_type const algnum_setup[] = {
    {"mkalgnum", (lisp_proc *) mkalgnum, 2},
    {"algebraic", (lisp_proc *) mk_algebraic, -1},
    {"is_algnum", (lisp_proc *) Lis_algnum, 1},
    {NULL, 0, 0}
};
