/*
 * apply.c      (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/apply.c,v 2.4 2015/06/02 08:22:16 kako Exp $
 */

static char const rcsid[] =
    "$Id: apply.c,v 2.4 2015/06/02 08:22:16 kako Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"


extern long     need_trace;
extern long     trace_count;
extern long     trace_flag;
extern long     trace_level;

/***************************************************************/
/***************************************************************/
/***************************************************************/

void            trace_arg_print(long n);
void            trace_val_print(void);

Sex             apply0(Sex fn);
Sex             apply1(Sex fn, Sex arg1);
Sex             apply2(Sex fn, Sex arg1, Sex arg2);
Sex             apply3(Sex fn, Sex arg1, Sex arg2, Sex arg3);
Sex             apply4(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4);
Sex             apply5(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4,
		       Sex arg5);
Sex             apply6(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4,
		       Sex arg5, Sex arg6);
Sex             apply7(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4,
		       Sex arg5, Sex arg6, Sex arg7);
Sex             apply8(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4,
		       Sex arg5, Sex arg6, Sex arg7, Sex arg8);
Sex             apply9(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4,
		       Sex arg5, Sex arg6, Sex arg7, Sex arg8, Sex arg9);
Sex             applyn(Sex fn, long n, ...);

/***************************************************************/
/***************************************************************/
/***************************************************************/


void
trace_arg_print(long n)
{
    Sex            *sp = stack;
    long            i;

    lisp_printf("\n*     %04d+:%P\n", trace_count++, sp[-n]);
    for (i = 1; i <= n; i++)
	lisp_printf("         Arg %d:%P\n", i, sp[i - n]);
}

void
trace_val_print()
{
    Sex            *sp = stack;

    lisp_printf("\n*     %04d-:%P\n", --trace_count, sp[-1]);
    lisp_printf("         Value : %P\n", *sp);
}

Sex
apply0(Sex fn)
{
    Sex             def;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push(fn);
	trace_arg_print(0);
    }
    if (typ == PROC_EXPR) {
	def = (Sex) proc(fn);
	if (!is_cons(def) || (car(def) != nil))
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
    } else if (typ & PROC_CCODE) {
	def = (*proc0(fn)) ();
    } else if (typ & PROC_BCODE) {
	def = interpreter(bcode(fn), 0, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply1(Sex fn, Sex arg1)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push2(fn, arg1);
	trace_arg_print(1);
	pop2(arg1, fn);
	push(fn);
    }
    switch (typ) {
    case PROC_EXPR:
	{
	    Sex             bvl;
	    def = (Sex) proc(fn);
	    if (!is_cons(def)) {
		error1("bad proc for", fn);
	    }
	    bvl = car(def);
	    if (bvl == nil)
		error1("too many args", fn);
	    v = car(bvl);
	    if (is_cons(v))
		error1("bad arg for ", fn);
	    pushb(v);
	    value(v) = arg1;
	    if (cdr(bvl) != nil)
		error1("too few args", fn);
	    def = progn_fn(cdr(def), nil);
	    popb();
	}
	break;
    case PROC_CSMACRO:
    case PROC_SUBR:
	def = (*proc1(fn)) (arg1);
	break;
    case PROC_CODE:
	bytecode_regs[1] = arg1;
	def = interpreter(bcode(fn), 1, nil);
	break;
    case PROC_FCODE:
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = env;
	def = interpreter(bcode(fn), 2, nil);
	break;
    case PROC_FSUBR:
	def = ((Special *) proc(fn)) (arg1, env);
	break;
    case PROC_CMACRO:
	push(env);
	push(arg1);
	def = apply_lambda(fn, Lget(fn, macro_symbol), 1, env);
	pop(env);
	def = eval(def, env);
	break;
    case PROC_MACRO:
	push(env);
	push(arg1);
	def = apply_lambda(fn, (Sex) proc(fn), 1, env);
	pop(env);
	def = eval(def, env);
	break;
    case PROC_SMACRO:
	push(env);
	fn = (Sex) proc(fn);
	def = sublis(car(fn), arg1, cdr(cdr(fn)));
	pop(env);
	def = eval(def, env);
	break;
    case PROC_FEXPR:
    default:
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}


Sex
apply2(Sex fn, Sex arg1, Sex arg2)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push3(fn, arg1, arg2);
	trace_arg_print(2);
	pop3(arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
	popb();
    } else if (typ & PROC_CCODE) {
	def = (*proc2(fn)) (arg1, arg2);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	def = interpreter(bcode(fn), 2, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply3(Sex fn, Sex arg1, Sex arg2, Sex arg3)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push4(fn, arg1, arg2, arg3);
	trace_arg_print(3);
	pop4(arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def = (*proc3(fn)) (arg1, arg2, arg3);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	def = interpreter(bcode(fn), 3, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply4(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push5(fn, arg1, arg2, arg3, arg4);
	trace_arg_print(4);
	pop5(arg4, arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg4;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def = (*proc4(fn)) (arg1, arg2, arg3, arg4);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	bytecode_regs[4] = arg4;
	def = interpreter(bcode(fn), 4, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply5(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4, Sex arg5)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push5(fn, arg1, arg2, arg3, arg4);
	push(arg5);
	trace_arg_print(5);
	pop(arg5);
	pop5(arg4, arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg4;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg5;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def = (*proc5(fn)) (arg1, arg2, arg3, arg4, arg5);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	bytecode_regs[4] = arg4;
	bytecode_regs[5] = arg5;
	def = interpreter(bcode(fn), 5, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply6(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4, Sex arg5, Sex arg6)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push5(fn, arg1, arg2, arg3, arg4);
	push2(arg5, arg6);
	trace_arg_print(6);
	pop2(arg6, arg5);
	pop5(arg4, arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg4;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg5;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg6;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def = (*proc6(fn)) (arg1, arg2, arg3, arg4, arg5, arg6);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	bytecode_regs[4] = arg4;
	bytecode_regs[5] = arg5;
	bytecode_regs[6] = arg6;
	def = interpreter(bcode(fn), 6, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply7(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4, Sex arg5,
       Sex arg6, Sex arg7)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push5(fn, arg1, arg2, arg3, arg4);
	push3(arg5, arg6, arg7);
	trace_arg_print(7);
	pop3(arg7, arg6, arg5);
	pop5(arg4, arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg4;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg5;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg6;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg7;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def = (*proc7(fn)) (arg1, arg2, arg3, arg4, arg5, arg6, arg7);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	bytecode_regs[4] = arg4;
	bytecode_regs[5] = arg5;
	bytecode_regs[6] = arg6;
	bytecode_regs[7] = arg7;
	def = interpreter(bcode(fn), 7, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply8(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4, Sex arg5,
       Sex arg6, Sex arg7, Sex arg8)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push5(fn, arg1, arg2, arg3, arg4);
	push4(arg5, arg6, arg7, arg8);
	trace_arg_print(8);
	pop4(arg8, arg7, arg6, arg5);
	pop5(arg4, arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg4;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg5;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg6;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg7;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg8;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def =
	    (*proc8(fn)) (arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	bytecode_regs[4] = arg4;
	bytecode_regs[5] = arg5;
	bytecode_regs[6] = arg6;
	bytecode_regs[7] = arg7;
	bytecode_regs[8] = arg8;
	def = interpreter(bcode(fn), 8, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}

Sex
apply9(Sex fn, Sex arg1, Sex arg2, Sex arg3, Sex arg4, Sex arg5,
       Sex arg6, Sex arg7, Sex arg8, Sex arg9)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    if (traced) {
	push5(fn, arg1, arg2, arg3, arg4);
	push5(arg5, arg6, arg7, arg8, arg9);
	trace_arg_print(9);
	pop5(arg9, arg8, arg7, arg6, arg5);
	pop5(arg4, arg3, arg2, arg1, fn);
	push(fn);
    }
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg1;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg2;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg3;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg4;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg5;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg6;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg7;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg8;
	bvl = cdr(bvl);
	if (bvl == nil)
	    error1("too many args", fn);
	v = car(bvl);
	pushb(v);
	value(v) = arg9;

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
	def =
	    (*proc9(fn)) (arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8,
			  arg9);
    } else if (typ & PROC_BCODE) {
	bytecode_regs[1] = arg1;
	bytecode_regs[2] = arg2;
	bytecode_regs[3] = arg3;
	bytecode_regs[4] = arg4;
	bytecode_regs[5] = arg5;
	bytecode_regs[6] = arg6;
	bytecode_regs[7] = arg7;
	bytecode_regs[8] = arg8;
	bytecode_regs[9] = arg9;
	def = interpreter(bcode(fn), 9, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}


Sex
applyn(Sex fn, long n, ...)
{
    Sex             def,
                    v;
    long            typ;
    long            traced;
    long            i;
    va_list         a;
    Sex             args[20];

    if (!is_symbol(fn))
	error1("bad function", fn);

    typ = qtag(fn) & PROC_MASK;
    traced = trace_flag & qtag(fn);
    stackcheck();
    va_start(a, n);
    if (traced) {
	push(fn);
	for (i = 0; i < n; i++)
	    push(va_arg(a, Sex));
	trace_arg_print(n);
	for (i = 0; i < n; i++)
	    pop(args[n - i - 1]);
	fn = stack[0];
    } else {
	for (i = 0; i < n; i++)
	    args[i] = va_arg(a, Sex);
    }
    va_end(a);
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	for (i = 0; i < n; i++) {
	    if (bvl == nil)
		error1("too many args", fn);
	    v = car(bvl);
	    pushb(v);
	    value(v) = args[i];
	}

	if (cdr(bvl) != nil)
	    error1("too few args", fn);
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
#if MAX_NARGS < 15
	if (n >= MAX_NARGS)
	    error1("argument exceed MAX", fn);
#endif
	switch (n) {
	case 0:
	    def = (*proc0(fn)) ();
	    break;
	case 1:
	    def = (*proc1(fn)) (args[0]);
	    break;
	case 2:
	    def = (*proc2(fn)) (args[0], args[1]);
	    break;
	case 3:
	    def = (*proc3(fn)) (args[0], args[1], args[2]);
	    break;
	case 4:
	    def = (*proc4(fn)) (args[0], args[1], args[2], args[3]);
	    break;
	case 5:
	    def =
		(*proc5(fn)) (args[0], args[1], args[2], args[3], args[4]);
	    break;
	case 6:
	    def =
		(*proc6(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5]);
	    break;
	case 7:
	    def =
		(*proc7(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5], args[6]);
	    break;
	case 8:
	    def =
		(*proc8(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5], args[6], args[7]);
	    break;
	case 9:
	    def =
		(*proc9(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5], args[6], args[7], args[8]);
	    break;
	case 10:
	    def =
		(*proc10(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8],
			       args[9]);
	    break;
	case 11:
	    def =
		(*proc11(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10]);
	    break;
	case 12:
	    def =
		(*proc12(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11]);
	    break;
	case 13:
	    def =
		(*proc13(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11], args[12]);
	    break;
	case 14:
	    def =
		(*proc14(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11], args[12], args[13]);
	    break;
	case 15:
	    def =
		(*proc15(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11], args[12], args[13],
			       args[14]);
	    break;
	default:
	    return error1("too many argument", fn);
	}
    } else if (typ & PROC_BCODE) {
	if (n < 0 || n > MAX_NARGS)
	    return error1("too many argument", fn);

	for (i = 0; i < n; i++)
	    bytecode_regs[i + 1] = args[i];
	def = interpreter(bcode(fn), n, nil);
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}
