/*
 * arith.h      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/arith.h,v 2.12 2016/03/19 07:34:05 kako Exp $
 */

#ifndef _ARITH_H_
#define _ARITH_H_

#define new_binteger(tag, size)	getvector(TAG_NUMBERS, tag, size)
#define new_bignum(size)	getvector(TAG_NUMBERS, T_BIGNUM, size)
#define new_bigneg(size)	getvector(TAG_NUMBERS, T_BIGNEG, size)
#define new_logical(size)	getvector(TAG_NUMBERS, T_LOGICAL, size)
#define new_modular()		\
	getvector(TAG_NUMBERS, T_MODULAR, sizeof(struct Modular_Number))
#define new_rational()		\
	getvector(TAG_NUMBERS, T_RATIONAL, sizeof(struct Rational_Number))
#define new_float()		\
	getvector(TAG_NUMBERS, T_FLOAT, sizeof(struct Float_Number))
#define new_bfloat()		\
	getvector(TAG_NUMBERS, T_BFLOAT, sizeof(struct Big_Float_Number))
#define new_efloat()		\
	getvector(TAG_NUMBERS, T_EFLOAT, sizeof(struct EFloat_Number))
#define new_befloat()		\
	getvector(TAG_NUMBERS, T_BEFLOAT, sizeof(struct Big_EFloat_Number))
#define new_interval()		\
	getvector(TAG_NUMBERS, T_INTERVAL, sizeof(struct Interval_Number))
#define new_binterval()		\
	getvector(TAG_NUMBERS, T_BINTERVAL, sizeof(struct Big_Interval_Number))
#define new_complex()		\
	getvector(TAG_NUMBERS, T_COMPLEX, sizeof(struct Complex_Number))
#define new_ecomplex()		\
	getvector(TAG_NUMBERS, T_ECOMPLEX, sizeof(struct EComplex_Number))
#define new_becomplex()		\
	getvector(TAG_NUMBERS, T_BECOMPLEX, sizeof(struct Big_EComplex_Number))
#define new_bmodular()		\
	getvector(TAG_NUMBERS, T_BMODULAR, sizeof(struct Big_Modular_Number))
#define new_qmodular()		\
	getvector(TAG_NUMBERS, T_QMODULAR, sizeof(struct q_Modular_Number))
#define new_fmodular()		\
	getvector(TAG_NUMBERS, T_FMODULAR, sizeof(struct f_Modular_Number))


#ifdef _LP64
#define fix_mask      (long)(-0x0800000000000000L)
#define fix_min       (long)( 0xf800000000000000L)
#define fix_max       (long)( 0x07ffffffffffffffL)
#define int_max       (long)( 0x7fffffffffffffffL)
#define fxat_min      (long)( 0x8000000000000000L)
#define fxat_max      (long)( 0x7ffffffffffffff0L)
#define big_min       (long)( 0x0800000000000000L)
#define flt_min  (2.22045e-16)

#define M10_19 10000000000000000000UL
#define M10_17 100000000000000000L
#define M10_PRINT_EXP	M10_19
#define M10_READ_EXP	M10_17
#define M10_PRINT_DIG	19
#define M10_READ_DIG	17
#else
#define fix_mask      (long)(-0x08000000)
#define fix_min       (long)( 0xf8000000)
#define fix_max       (long)( 0x07ffffff)
#define int_max       (long)( 0x7fffffff)
#define fxat_min      (long)( 0x80000000)
#define fxat_max      (long)( 0x7ffffff0)
#define big_min       (long)( 0x08000000)
#define flt_min  (2.22045e-16)

#define M10_9 1000000000
#define M10_8 100000000
#define M10_PRINT_EXP	M10_9
#define M10_READ_EXP	M10_8
#define M10_PRINT_DIG	9
#define M10_READ_DIG	8
#endif



long            bbadd(unsigned long a[], long lena, unsigned long b[],
		      long lenb, unsigned long c[]);
		/*
		 * big number and big number addition a + b ==> c
		 * 
		 * len(a) should be greater of equal to len(b) len(c)
		 * should be greater of equal to len(a) returns overflow
		 * digit (if zero, no overflow occurs) 
		 */

long            bbsub(unsigned long a[], long lena, unsigned long b[],
		      long lenb, unsigned long c[]);
		/*
		 * big number and big number subtraction a - b ==> c
		 * 
		 * len(a) should be greater of equal to len(b) len(c)
		 * should be greater of equal to len(a) returns resulting
		 * length of c if it returns negative value, it means the
		 * result is negative. 
		 */

long            bfmul(unsigned long a[], long lena, unsigned long b,
		      unsigned long c[]);
long            bbmul(unsigned long a[], long lena, unsigned long b[],
		      long lenb, unsigned long c[]);
long            bfdiv(unsigned long a[], long lena, unsigned long b,
		      unsigned long c[]);
long            bbsqr(unsigned long a[], long lena, unsigned long c[]);
#ifdef SAVE_SPACE
#else
long            bbquo(unsigned long a[], long lena, unsigned long b[],
		      long lenb, unsigned long c[], unsigned long d[],
		      unsigned long e[]);
long            bbrem(unsigned long a[], long lena, unsigned long b[],
		      long lenb, unsigned long c[], unsigned long d[],
		      unsigned long e[]);
#endif
long            bbdiv(unsigned long a[], long lena, unsigned long b[],
		      long lenb, unsigned long c[], unsigned long d[],
		      unsigned long e[], long *r);
long            bfrem(unsigned long a[], long lena, unsigned long b);

extern Sex      negateb(Sex);
extern Sex      negatel(Sex);
extern Sex      negated(Sex);
extern Sex      negateE(Sex);
extern Sex      negateZ(Sex);
extern Sex      negateV(Sex);

extern Sex      Ladd1(Sex a);
extern Sex      Lsub1(Sex a);

extern Sex      copyb(Sex);
extern Sex      negate(Sex);
extern Sex      plus2(Sex a, Sex b);
extern Sex      diff2(Sex a, Sex b);
extern Sex      times2(Sex a, Sex b);
extern Sex      quot2(Sex a, Sex b);
extern Sex      remaind2(Sex a, Sex b);
extern Sex      ratio2(Sex a, Sex b);
extern Sex      shrink_bignum(Sex a, long lena);
extern Sex      modulus(Sex a, Sex b);
extern Sex      smodulus(Sex a, Sex b);
extern Sex      rational(Sex a);
extern Sex      rationalize(Sex a);
extern bool     numeq2(Sex a, Sex b);
extern bool     appeq2(Sex a, Sex b);
extern bool     zerop(Sex a);
extern bool     zeropd(Sex a);
extern bool     onep(Sex a);
extern bool     monep(Sex a);
extern bool     minusp(Sex a);
extern bool     plusp(Sex a);
extern Sex      mklogical(long n);
extern Sex      mkbigat1(long n);
extern Sex      mkbigat1s(long s, long n);
extern Sex      mkbigat(long a, unsigned long b);
extern double   float_of_integer(Sex a);
extern double   floatit(Sex a);
extern Sex      make_float(double a);
extern Sex      make_efloat(double v, double e);
extern Sex      make_interval(double a, double b);
extern Sex      make_ratio(Sex p, Sex q);
extern Sex      make_bfloat(Sex m, long e);
extern Sex      make_polynomial(Sex a, Sex b, Sex c);
extern Sex      make_dipoly(Sex a, Sex b);
extern Sex      make_vector(long a);
extern Sex      make_ecomplex(double r, double i, double e);
extern Sex      make_befloat(Sex a, bf_error_type e);
extern Sex      make_becomplex(Sex r, Sex i, bf_error_type e);
extern Sex      make_binterval(Sex u, Sex l);
extern Sex      make_polynomial(Sex a, Sex b, Sex c);
extern Sex      make_poly(Sex x, Sex p);
extern Sex      make_rpoly(Sex a, Sex b);
extern Sex      symbol_to_polynomial(Sex p);
extern Sex      ash(Sex a, Sex b);
extern Sex      lognot(Sex a);
extern Sex      logior2(Sex a, Sex b);
extern Sex      logxor2(Sex a, Sex b);
extern Sex      logand2(Sex a, Sex b);
extern Sex      logeqv2(Sex a, Sex b);
extern Sex      logor2(Sex a, Sex b);
extern Sex      ilognot(Sex a);
extern Sex      ilogior2(Sex a, Sex b);
extern Sex      ilogxor2(Sex a, Sex b);
extern Sex      ilogand2(Sex a, Sex b);
extern Sex      ilogeqv2(Sex a, Sex b);
extern Sex      ilogor2(Sex a, Sex b);
extern Sex      rationalf(double d);
extern bool     lessp2(Sex a, Sex b);
extern bool     clessp2(Sex a, Sex b);

extern Sex      Lminus(Sex a);
extern Sex      Lzerop(Sex a);
extern Sex      Lonep(Sex a);
extern Sex      Levenp(Sex a);
extern Sex      Loddp(Sex a);
extern Sex      Labs(Sex a);
extern Sex      Lminusp(Sex a);
extern Sex      Lplusp(Sex a);
extern Sex      Leqn(Sex a, Sex b);
extern Sex      Lappeq(Sex a, Sex b);
extern Sex      Llessp(Sex a, Sex b);
extern Sex      Lgreaterp(Sex a, Sex b);
extern Sex      Lgeq(Sex a, Sex b);
extern Sex      Lleq(Sex a, Sex b);
extern Sex      Lclessp(Sex a, Sex b);
extern Sex      Lcgreaterp(Sex a, Sex b);
extern Sex      Lcgeq(Sex a, Sex b);
extern Sex      Lcleq(Sex a, Sex b);
extern Sex      Lmax2(Sex a, Sex b);
extern Sex      Lmin2(Sex a, Sex b);
extern Sex      Lnumber_log2(Sex a);
extern Sex      Lleftshift(Sex a, Sex b);
extern Sex      Lprecision(Sex a);

extern long     _reduced_exp(double, double *);


typedef struct Complex {
    double          real;
    double          imag;
} Complex;

typedef union DoubleFloat {
    double          d;
    long            l;
    int             i[2];
} DoubleFloat;

/*
 * extern Complex  cln(Complex a); extern Complex  ccos(Complex a); extern 
 * Complex  cexp(Complex a); extern Complex  cpow(Complex a, Complex b);
 * extern double   fcabs(Complex a); 
 */

extern bool     is_shortfloat(Sex a);
extern bool     is_longfloat(Sex a);

extern Sex      Lfrexp(Sex a);
extern Sex      Lceiling(Sex a);
extern Sex      Lfloor(Sex a);
extern Sex      Lround(Sex a);
extern Sex      Ltruncate(Sex a);

extern Sex      Ldenominator(Sex a);
extern Sex      Lnumerator(Sex a);

extern Sex      Linterval_lower(Sex a);
extern Sex      Linterval_upper(Sex a);
extern Sex      Linterval_mid(Sex a);
extern Sex      Linterval_width(Sex a);

extern Sex      Lefloat_error(Sex a);
extern Sex      Lefloat_value(Sex a);
extern Sex      Lecomplex_value(Sex a);
extern Sex      Lecomplex_error(Sex a);
extern Sex      Lecomplex_real(Sex a);
extern Sex      Lecomplex_imag(Sex a);
extern Sex      Lcomplex_real(Sex a);
extern Sex      Lcomplex_imag(Sex a);
extern Sex      Lbefloat_error(Sex a);
extern Sex      Lbefloat_value(Sex a);

extern Sex      Lfix(Sex a);
extern Sex      Llogical(Sex a);
extern Sex      Lrational(Sex a);
extern Sex      Lmkrational(Sex a, Sex b);
extern Sex      Lfloat(Sex a);
extern Sex      Lmkefloat(Sex a, Sex b);
extern Sex      Lmkefloat_by_ratio(Sex a, Sex b);
extern Sex      Lmkefloat1(Sex a);
extern Sex      Lefloat(Sex a);
extern Sex      Lmkcomplex(Sex a, Sex b);
extern Sex      Lcomplex(Sex a);
extern Sex      Lmkecomplex(Sex a, Sex b, Sex c);
extern Sex      Lecomplex(Sex a);
extern Sex      Lmkinterval(Sex a, Sex b);
extern Sex      Linterval(Sex a);
extern Sex      Linterval_by_ratio(Sex a, Sex f);
extern Sex      Lbfloat(Sex a);
extern Sex      Lbefloat(Sex a);
extern Sex      Lmkbefloat(Sex a, Sex b);
extern Sex      Lmkbefloat_by_ratio(Sex a, Sex b);
extern Sex      Lbefloatit(Sex a);
extern Sex      Lmkbinterval(Sex a, Sex b);
extern Sex      Lbinterval(Sex a);
extern Sex      Lmkbecomplex(Sex a, Sex b, Sex c);
extern Sex      Lbecomplex(Sex a);
extern Sex      Lcomplex_conjugate(Sex a);
extern Sex      Lmkmodular(Sex a);
extern Sex      Lunmodular(Sex a);

extern Sex      expti(Sex, long);
extern Sex      Lcbrt(Sex);
extern Sex      Lpi(void);
extern Sex      Le(void);



/*
 * conversions between numbers 
 */


extern Sex      becomplexitE(Sex);
extern Sex      becomplexitV(Sex);
extern Sex      becomplexitb(Sex);
extern Sex      becomplexitc(Sex);
extern Sex      becomplexitd(Sex);
extern Sex      becomplexite(Sex);
extern Sex      becomplexiti(Sex);
extern Sex      becomplexitv(Sex);
extern Sex      becomplexitz(Sex);

extern Sex      befloatitV(Sex);
extern Sex      befloatitb(Sex);
extern Sex      befloatitd(Sex);
extern Sex      befloatite(Sex);
extern Sex      befloatitf(double);
extern Sex      befloatiti(Sex);
extern Sex      befloatitr(Sex);
extern Sex      befloatitv(Sex);

extern Sex      bfloatitb(Sex);
extern Sex      bfloatitf(double);
extern Sex      bfloatiti(Sex);
extern Sex      bfloatitr(Sex);

extern Sex      ecomplexitE(Sex);
extern Sex      ecomplexitV(Sex);
extern Sex      ecomplexitZ(Sex);
extern Sex      ecomplexitc(Sex);

extern Sex      efloatitv(Sex);
extern Sex      efloatitE(Sex);
extern Sex      efloatitV(Sex);


extern double   floatitE(Sex);
extern double   floatitV(Sex);
extern double   floatitb(Sex);
extern double   floatitd(Sex);
extern double   floatitl(Sex);

extern Sex      bfloatit(Sex);

extern Sex      bintervalitE(Sex);
extern Sex      bintervalitb(Sex);
extern Sex      bintervalitd(Sex);
extern Sex      bintervalite(Sex);
extern Sex      bintervaliti(Sex);
extern Sex      bintervalitv(Sex);

extern Sex      intervalitV(Sex);

extern Sex      fixitd(Sex);
extern Sex      fixitZ(Sex);
extern Sex      fixitV(Sex);
extern Sex      fixitl(Sex);
extern Sex      fixitc(Sex);
extern Sex      fixitz(Sex);

extern Sex      exptip(Sex, long);
extern long     is_decimal_part(Sex);

extern long     number_log2(Sex a);
extern long     bignum_log2(Sex a);
extern long     bfloat_log2(Sex a);
extern long     float_log2(double);
extern long     integer_log2(long);
extern long     msd(unsigned long);
extern long     lsd(unsigned long);


#ifdef BEFLOAT_ERROR_IS_DOUBLE
#else
extern long     efloat_edigit(double, double);
extern long     ecomplex_edigit(double, double, double);
extern long     binterval_edigit(Sex);
#endif

extern Sex      leftshift(Sex, long);
extern Sex      rightshift(Sex, long);
extern Sex      leftshifti(Sex, long);
extern Sex      leftshiftd(Sex, long);

extern Sex      fixitf(double);

extern Sex      Lsafe_fp_plus(Sex, Sex);
extern Sex      Lsafe_fp_times(Sex, Sex);
extern Sex      Lsafe_fp_quot(Sex, Sex);


#ifdef ASUKA
/*
 * for gal structures 
 */
extern Sex      gplus2(Sex, Sex);
extern Sex      gdiff2(Sex, Sex);
extern Sex      gtimes2(Sex, Sex);
extern Sex      gquot2(Sex, Sex);
extern Sex      gratio2(Sex, Sex);
extern Sex      gremaind2(Sex, Sex);
extern Sex      gexpt(Sex, Sex);
extern Sex      gpremaind2(Sex, Sex, Sex);
extern Sex      gpquot2(Sex, Sex, Sex);
extern Sex      gpdivide2(Sex, Sex, Sex);
extern Sex      ggcd2(Sex, Sex);

extern Sex      gnegate(Sex);
extern Sex      gmodulus(Sex, Sex);
extern Sex      gsmodulus(Sex, Sex);
#endif


extern Sex      divide2ii(Sex, Sex);
extern Sex      divide2bi(Sex, Sex);
extern Sex      divide2ib(Sex, Sex);
extern Sex      divide2bb(Sex, Sex);

extern Sex      recipm(Sex a);
extern Sex      Lsqrt(Sex a);

extern Sex      gcdbb(Sex a, Sex b);
extern Sex      gcdbi(Sex a, Sex b);

extern Sex      gcdrr(Sex a, Sex b);

#endif
