%{ 

/*
 *    asuka.y           (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 *      read parser for maple-like syntax
 *      $Header: /Users/kako/asuka/src/RCS/asuka.y,v 2.24 2016/04/08 07:19:55 kako Exp $
 */

static char const rcsid[] =
    "$Id: asuka.y,v 2.24 2016/04/08 07:19:55 kako Exp $";


#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include "config.h"
#include "tags.h"
#include "externs.h"
#include "entries.h"
#include "macros.h"
#include "read.h"
#include "fns.h"

/*
 * #define YYSTYPE Sex
 */

void            yyerror(char *);
static int      stmt_count = 0;

extern Sex     *parse_value_ptr;
extern Sex      proc_frame;
extern Sex      loop_frame;

extern Sex      onbeepre_symbol;
extern Sex      twbeepre_symbol;
extern Sex      thbeepre_symbol;

Sex             symtab;
Sex             methnames,
                methdefs;

static int      label_count = 0;
static int      in_proc = 0;
static int      in_loop = 0;

static void     begin_loop(void);
static void     end_loop(void);
static void     begin_proc(void);
static void     end_proc(void);
void            clear_loop(void);

static Sex	make_try(Sex, Sex, Sex);

static Sex      pickup_globals(Sex);

static          Sex make_statements(Sex l, Sex a);
static          Sex make_statements1(Sex a);

static Sex      make_name1(Sex a, Sex b);
static Sex      make_proc(Sex, Sex, Sex);
static Sex      make_proc_body(Sex);
static Sex      make_func1(Sex, Sex, Sex);
static Sex      make_func2(Sex, Sex, Sex);
static Sex      make_funcall(Sex, Sex);
static Sex      make_if(Sex, Sex, Sex);

static Sex      make_switch(Sex, Sex);
static Sex      make_case(Sex, Sex, Sex);

static Sex      expand_loop(Sex);
static Sex      expand_for(Sex, Sex, Sex);
static Sex      expand_while(Sex, Sex, Sex);
static Sex      expand_repeat(Sex, Sex, Sex);
static Sex      expand_foreach(Sex, Sex, Sex);
static Sex      expand_foreach_expression(Sex, Sex, Sex);
static Sex      expand_for_expression(Sex, Sex, Sex);

static Sex      make_break(Sex);
static Sex      make_continue(Sex);
static Sex      make_return0(void);
static Sex      make_return1(Sex);

static Sex      expand2(Sex);

extern Sex      gdiff2(Sex, Sex);

extern Sex      error(char *s);
extern Sex      Lstrcat(Sex, Sex);
%}

/*
 * Inappropriate start symbol for an interpreter
 *
 */
%union {
	long itype;
	int lineno;
	Sex sex;
}


%left QUOTE

%left ',' '.'

%token <sex> LITERAL STOP SYMBOL STRING EOF_TOKEN UNKNOWN

%token <sex> NUOPE TWBEEPRE THBEEXPRE NUMBER SHOSSU ONEBEEPRE
%token <sex> OD OPTION OPTIONS PROC QUIT THEN TO DOWNTO WHILE
%token <sex> REPEAT UNTIL BEGIN RETURN
%token <sex> BREAK CONTINUE ARRAY VECTOR MATRIX
%token <sex> BY DO DONE ELIF ELSE FI FOR FROM IF IN LOCAL END
%token <sex> GLOBAL
%token <sex> SETORDER ORDER_LEX ORDER_TLEX ORDER_RLEX ORDER_GRADINGS
%token <sex> FOREACH ON
%token <sex> ENABLE DISABLE
%token <sex> OR AND NOT MOD SMOD
%token <sex> UNION NONCOMU INTERSECT
%token <sex> LESSER_EQUAL GREATER_EQUAL NOT_EQUAL COLON_EQUAL
%token <sex> READ IMPORT SAVE WRITE WRITEN PRINTF LOAD CLEAR
%token <sex> XDEBUG XDEBUGN
%token <sex> ALGEBRAIC WITH
%token <sex> TRY THROW CATCH
%token <sex> SWITCH HCTIWS CASE ESAC OF DEFAULT
%token <sex> QUOTE_SYMBOL
/*
%token <sex> ERROR
*/
%token <sex> SUM PRODUCT COLLECT CONC JOIN

%token <sex> '$' '%'
%token <sex> '[' ']' '{' '}'
%token <sex> '<' '>' '='
%token <sex> '+' '-' '*' '/' '^' REMAINDER
%token <sex> ITRSYTH ARROWOPE
%token <sex> SEMICOLON_SEMICOLON
%token <sex> '(' ')'
%token <sex> '!' 
%token <sex> '&'


%right '@'

%left UNION SET_DIFF INTERSECT 



%right ITRSYTH


//%left ARROWOPE

%nonassoc MOD SMOD '$' 

%nonassoc OMIT

%right COLON_EQUAL
%left OR
%left AND
%right NOT
%left '=' NOT_EQUAL
%left '<' '>' LESSER_EQUAL GREATER_EQUAL
%left '+' '-'
%left '*' '/' REMAINDER
%left '&' NONCOMU
%right UNARY
%left '^' 
%left '!' 




%type <sex> program   
%type <sex> statement
%type <sex> statements
%type <sex> array_declaration
%type <sex> matrix_declaration
%type <sex> vector_declaration
%type <sex> global_declaration
%type <sex> setorder_declaration
%type <sex> func_body
%type <sex> proc_declaration
%type <sex> proc_body
%type <sex> proc_start
%type <sex> return_statement
%type <sex> break_statement
%type <sex> continue_statement
%type <sex> if_statement
%type <sex> while_statement
%type <sex> repeat_statement
%type <sex> for_statement
%type <sex> for_head
%type <sex> foreach_statement
%type <sex> foreach_head
%type <sex> for_list
%type <sex> foreach_list
%type <sex> switch_statement
%type <sex> switch_clause
%type <sex> switch_clauses
%type <sex> case_statement
%type <sex> case_clause
%type <sex> case_clauses
%type <sex> clear_statement
%type <sex> try_statement
%type <sex> throw_statement
%type <sex> catch_clause
/*
%type <sex> error_statement
*/

%type <sex> enable_statement
%type <sex> disable_statement

%type <sex> write_statement
%type <sex> debug_statement
%type <sex> read_statement
%type <sex> import_statement
%type <sex> save_statement
%type <sex> load_statement
%type <sex> simple_statement
%type <sex> assign_statement
%type <sex> empty_statement
%type <sex> expression_statement
%type <sex> algebraic_statement
%type <sex> expression
%type <sex> expression1
//%type <sex> string_or_expressions
%type <sex> else_expression
%type <sex> else_parts
%type <sex> for_expression
%type <sex> foreach_expression
%type <sex> localopt
%type <sex> localopts 
%type <sex> nameseq  
%type <sex> name
%type <sex> filename
%type <sex> identifier
%type <sex> string
%type <sex> try_tag

%start program


%%
program   : statement ';' 
		{*parse_value_ptr = $1; 
		  value_print_flag = 1;
//		  lisp_printf("stat = %p\n", $1); 
		  YYACCEPT;}
	  | statement SEMICOLON_SEMICOLON
		{*parse_value_ptr = $1; 
		  value_print_flag = 0;
//		  lisp_printf("stat = %p\n", $1); 
		  YYACCEPT;}
/*
	  | statement error 
		{*parse_value_ptr = $1;
		  yyerror("missing ;");
		  YYACCEPT;}
*/
	  | EOF_TOKEN
		{*parse_value_ptr = $1;
	//	  yyerror("end of file;");
		  YYACCEPT;}
          | UNKNOWN
		{*parse_value_ptr = $1;
			yyerror("illegal character;");
		YYACCEPT;}
          ;

statement : if_statement
	  | switch_statement
	  | case_statement
	  | clear_statement
/*
	  | error_statement
*/
	  | enable_statement
	  | disable_statement
	  | write_statement
	  | debug_statement
	  | read_statement
	  | import_statement
	  | save_statement
	  | load_statement
	  | for_statement
	  | foreach_statement
	  | while_statement
	  | repeat_statement
	  | assign_statement
	  | simple_statement
	  | global_declaration
	  | setorder_declaration
	  | array_declaration
	  | matrix_declaration
	  | vector_declaration
	  | proc_declaration
	  | return_statement
	  | break_statement
	  | continue_statement
	  | algebraic_statement
	  | try_statement
	  | throw_statement
	  | QUIT		
		{ $$ = Lncons($1);}
	  | error
		{  yyerror("Malformed statement");
		   YYERROR;}
	  ;

statements
	  : statement
			{$$ = make_statements1($1);}
	  | statements ';' statement
			{ 
/*
		if(in_loop || in_proc) {
			if(car($1) == prog_symbol)
		} else 
			$$ = cons($3, $1);}
*/
			$$ = make_statements($1, $3);
		}
	  | statements SEMICOLON_SEMICOLON statement
			{ $$ = make_statements($1, $3);}
	  ;

if_statement
	  : IF expression THEN statements else_parts FI
		{ $$ = make_if($2, expand2($4), $5);}
	  | IF expression THEN statements else_parts 
		{ yyerror("missing fi");
		$$ = make_if($2, expand2($4), $5);}
	  ;

else_parts
	  :
		{ $$ = nil;}
	  | ELSE statements
		{ $$ = expand2($2);}
	  | ELIF expression THEN statements else_parts
		{ $$ = make_if($2, expand2($4), $5);}
	  ;

switch_statement
	  : SWITCH expression  OF switch_clauses HCTIWS
		{ $$ = make_switch($2, $4);}
	  ;

switch_clause
	  : expression1 ':' statements 
		{ $$ = Llist2(cons(list_symbol, $1), $3);}
	  ;

switch_clauses
	  : switch_clause 
		{$$ = cons($1, nil);}
	  | switch_clause ';'
		{$$ = cons($1, nil);}
	  | switch_clauses switch_clause
		{$$ = cons($2, $1);}
	  | switch_clauses ';' switch_clause
		{$$ = cons($3, $1);}
	  | switch_clauses ';' 
		{$$ = $1;}
	  ;


case_statement
	  : CASE expression OF case_clauses ESAC
		{ $$ = make_case($2, $4, nil);}
	  | CASE expression OF case_clauses ELSE statements ESAC
		{ $$ = make_case($2, $4, $6);}
	  ;

case_clause
	  : expression1 ':' statements ';'
		{ $$ = Llist2(cons(list_symbol, $1), $3);}
	  ;

case_clauses
	  : case_clause 
		{$$ = cons($1, nil);}
	  | case_clauses case_clause
		{$$ = cons($2, $1);}
	  ;

/*
while_statement
	  : WHILE expression DO statements OD
		{ $$ = Llist3($1, $2, $4);}
	  ;

repeat_statement
	  : REPEAT statements UNTIL expression
		{ $$ = Llist3($1, $4, $2);}
	  ;


for_statement
	  : FOR for_list DO statements OD
		{ $$ = expand_for($1, $2, $4);}
	  ;
*/
while_statement
	  : WHILE { begin_loop();} expression DO statements OD
		{ $$ = expand_while($1, $3, $5); end_loop();}
	  ;
/*
	  | WHILE {begin_loop();} error
		{  yyerror("error in while statement");
		   end_loop();
		   YYERROR;}
	  ;
*/

repeat_statement
	  : REPEAT {begin_loop();} statements UNTIL expression
		{ $$ = expand_repeat($1, $5, $3); end_loop();}
	  ;
/*
	  | REPEAT {begin_loop();} error
		{  yyerror("error in repeat statement");
		   end_loop();
		   YYERROR;}
	  ;
*/

/*
for_statement
	  : FOR {begin_loop();} for_list DO statements OD
		{ $$ = expand_for($1, $3, $5);
			end_loop();}
	  ;
*/
for_statement
	  : for_head DO statements OD
		{ $$ = expand_for(nil, $1, $3);
		  end_loop();}
	  ;

for_expression
	  : for_head SUM expression
		{$$ = expand_for_expression(mkfxat(1), $1, $3);
		 end_loop();}
	  | for_head PRODUCT expression
		{$$ = expand_for_expression(mkfxat(2), $1, $3);
		 end_loop();}
	  | for_head COLLECT expression
		{$$ = expand_for_expression(mkfxat(3), $1, $3);
		 end_loop();}
	  | for_head JOIN expression
		{$$ = expand_for_expression(mkfxat(4), $1, $3);
		 end_loop();}
	  | for_head CONC expression
		{$$ = expand_for_expression(mkfxat(5), $1, $3);
		 end_loop();}
	  ;
for_head
	  : FOR {begin_loop();} for_list
		{ $$ = $3;}
	  ;
/*
	  | FOR {begin_loop();} error
		{  yyerror("error in for statement");
		   end_loop();
		   YYERROR;}
	  ;
*/

for_list
	  : name COLON_EQUAL expression TO expression
		{$$ = Llist5(mkfxat(0), $1, $3, $5, mkfxat(1));}
	  | name COLON_EQUAL expression TO expression BY expression
		{$$ = Llist5(mkfxat(0), $1, $3, $5, $7);}
	  | name COLON_EQUAL expression BY expression
		{$$ = Llist5(mkfxat(0), $1, $3, nil, $5);}
	  | name COLON_EQUAL expression BY expression TO expression
		{$$ = Llist5(mkfxat(0), $1, $3, $7, $5);}
	  | name COLON_EQUAL expression DOWNTO expression
		{$$ = Llist5(mkfxat(1), $1, $3, $5, mkfxat(1));}
	  | name COLON_EQUAL expression DOWNTO expression BY expression
		{$$ = Llist5(mkfxat(1), $1, $3, $5, $7);}
	  | name COLON_EQUAL expression BY expression DOWNTO expression
		{$$ = Llist5(mkfxat(1), $1, $3, $7, $5);}
	  | name COLON_EQUAL expression WHILE expression
		{$$ = Llist5(mkfxat(2),$1, $3, $5, mkfxat(1));}
	  | name COLON_EQUAL expression WHILE expression BY expression
		{$$ = Llist5(mkfxat(2),$1, $3, $5, $7);}
	  | name COLON_EQUAL expression BY expression WHILE expression
		{$$ = Llist5(mkfxat(2),$1, $3, $7, $5);}
	  ;

/*
foreach_statement
	  : FOREACH {begin_loop();} foreach_body
		{ $$ = expand_foreach($1, $3); end_loop();}
	  | FOREACH {begin_loop();} foreach_body
		{ $$ = expand_foreach($1, $3); end_loop();}
	  ;

foreach_head
	  : FOREACH {begin_loop();} 
	  ;
foreach_body
	  : name IN expression DO statements OD
		{ $$ = Llist4($1, $3, $5, nil);}
	  | name ON expression DO statements OD
		{ $$ = Llist4($1, $3, $5, true);}
*/

/*
foreach_statement
	  : foreach_head name IN expression DO statements OD
		{ $$ = Llist4($2, $4, $6, nil);
		  $$ = expand_foreach($1, $$);
		  end_loop();}
	  | foreach_head name ON expression DO statements OD
		{ $$ = Llist4($2, $4, $6, true);
		  $$ = expand_foreach($1, $$);
		  end_loop();}
	  ;

foreach_expression
	  : foreach_head name IN expression SUM expression
		{ $$ = Llist4($2, $4, $6, nil);
		  $$ = expand_foreach_expression(mkfxat(1), $$);
		  end_loop();}
	  | foreach_head name ON expression SUM expression
		{ $$ = Llist4($2, $4, $6, true);
		  $$ = expand_foreach_expression(mkfxat(1), $$);
		  end_loop();}
	  | foreach_head name IN expression PRODUCT expression
		{ $$ = Llist4($2, $4, $6, nil);
		  $$ = expand_foreach_expression(mkfxat(2), $$);
		  end_loop();}
	  | foreach_head name ON expression PRODUCT expression
		{ $$ = Llist4($2, $4, $6, true);
		  $$ = expand_foreach_expression(mkfxat(2), $$);
		  end_loop();}
	  | foreach_head name IN expression COLLECT expression
		{ $$ = Llist4($2, $4, $6, nil);
		  $$ = expand_foreach_expression(mkfxat(3), $$);
		  end_loop();}
	  | foreach_head name ON expression COLLECT expression
		{ $$ = Llist4($2, $4, $6, true);
		  $$ = expand_foreach_expression(mkfxat(3), $$);
		  end_loop();}
	  | foreach_head name IN expression JOIN expression
		{ $$ = Llist4($2, $4, $6, nil);
		  $$ = expand_foreach_expression(mkfxat(4), $$);
		  end_loop();}
	  | foreach_head name ON expression JOIN expression
		{ $$ = Llist4($2, $4, $6, true);
		  $$ = expand_foreach_expression(mkfxat(4), $$);
		  end_loop();}
	  | foreach_head name IN expression CONC expression
		{ $$ = Llist4($2, $4, $6, nil);
		  $$ = expand_foreach_expression(mkfxat(5), $$);
		  end_loop();}
	  | foreach_head name ON expression CONC expression
		{ $$ = Llist4($2, $4, $6, true);
		  $$ = expand_foreach_expression(mkfxat(5), $$);
		  end_loop();}
	  ;
*/

foreach_statement
	  : foreach_head DO statements OD
		{ $$ = expand_foreach(nil, $1, $3);
		  end_loop();}
	  ;

foreach_expression
	  : foreach_head SUM expression
		{$$ = expand_foreach_expression(mkfxat(1), $1, $3);
		 end_loop();}
	  | foreach_head PRODUCT expression
		{$$ = expand_foreach_expression(mkfxat(2), $1, $3);
		 end_loop();}
	  | foreach_head COLLECT expression
		{$$ = expand_foreach_expression(mkfxat(3), $1, $3);
		 end_loop();}
	  | foreach_head JOIN expression
		{$$ = expand_foreach_expression(mkfxat(4), $1, $3);
		 end_loop();}
	  | foreach_head CONC expression
		{$$ = expand_foreach_expression(mkfxat(5), $1, $3);
		 end_loop();}
	  ;

foreach_head
	  : FOREACH {begin_loop();} foreach_list
		{$$ = $3;}
	  ;
foreach_list
	  : name IN expression
		{$$ = Llist3($1, $3, nil);}
	  | name ON expression
		{$$ = Llist3($1, $3, true);}
	  ;


simple_statement
	  : empty_statement
	  | expression_statement
	  ;

empty_statement
	  :
		{$$ = nil;}
	  ;

clear_statement
	  : CLEAR nameseq
		{$$ = cons(clear_symbol, $2);}
	  ;

/*
error_statement
	  : ERROR expression1
		{$$ = Llist3(error_symbol, mkfxat(99), cons(list_symbol, $2));}
	  ;
*/
enable_statement
	  : ENABLE SYMBOL
		{$$ = Llist2($1, $2);}
	  | ENABLE XDEBUG
		{$$ = Llist2($1, $2);}
	  | enable_statement ',' SYMBOL
		{$$ = Laconc($1, $3);}
	  | enable_statement ',' XDEBUG
		{$$ = Laconc($1, $3);}
	  ;
disable_statement
	  : DISABLE SYMBOL
		{$$ = Llist2($1, $2);}
	  | DISABLE XDEBUG
		{$$ = Llist2($1, $2);}
	  | disable_statement ',' SYMBOL
		{$$ = Laconc($1, $3);}
	  | disable_statement ',' XDEBUG
		{$$ = Laconc($1, $3);}
	  ;
/*
write_statement
	  : WRITE string_or_expressions
		{$$ = Llist2(write_symbol, cons(list_symbol, $2));}
	  ;
*/
write_statement
	  : WRITE expression1
		{$$ = Llist2(write_symbol, cons(list_symbol, $2));}
	  | WRITEN expression1
		{$$ = Llist2(writen_symbol, cons(list_symbol, $2));}
	  | PRINTF '(' string  ')'
		{$$ = Llist3($1, $3, nil);}
	  | PRINTF '(' identifier ')'
		{$$ = Llist3($1, $3, nil);}
	  | PRINTF '(' string ',' expression1')'
		{$$ = Llist3($1, $3, cons(list_symbol, $5));}
	  | PRINTF '(' identifier ',' expression1')'
		{$$ = Llist3($1, $3, cons(list_symbol, $5));}
	  ;

debug_statement
	  : XDEBUG expression1
		{$$ = Llist2(debug_symbol, cons(list_symbol, $2));}
	  | XDEBUGN expression1
		{$$ = Llist2(debugn_symbol, cons(list_symbol, $2));}
	  ;


read_statement
	  : READ string
		{$$ = Llist2(read_symbol, $2);}
	  ;
import_statement
	  : IMPORT string
		{$$ = Llist2(import_symbol, $2);}
	  | IMPORT '<' SYMBOL '>'
		{$$ = Llist2(import_symbol, Lsymbol_name($3));}
	  | IMPORT '<' filename '>'
		{$$ = Llist2(import_symbol, $3);}
	  ;

filename
	  : SYMBOL  '.' NUMBER
		{ $$ = Lstrcat($1, $3);}
	  | SYMBOL  '.' SYMBOL
		{ $$ = Lstrcat($1, $3);}
	  | filename  '.' NUMBER
		{ $$ = Lstrcat($1, $3);}
	  | filename  '.' SYMBOL
		{ $$ = Lstrcat($1, $3);}
	  ;

save_statement
	  : SAVE string
		{$$ = Llist2(save_symbol, $2);}
	  ;
load_statement
	  : LOAD string
		{$$ = Llist2(load_symbol, $2);}
	  ;

expression_statement
	  : expression
	  ;

assign_statement
	  : identifier COLON_EQUAL expression
			{ $$ = Llist3(setq_symbol, $1, $3);}
	  | identifier COLON_EQUAL assign_statement
			{ $$ = Llist3(setq_symbol, $1, $3);}
	  | identifier COLON_EQUAL func_body
			{ $$ = Llist3(setq_symbol, $1, $3);}
	  | identifier '[' expression ']' COLON_EQUAL expression
			{ $$ = Llist4(setarray1_symbol, $1, $6, $3);}
	  | identifier '[' expression ',' expression ']' COLON_EQUAL expression
			{ $$ = Llist5(setarray2_symbol, $1, $8, $3, $5);}
	  | identifier '[' expression ',' expression ',' expression ']' 
		COLON_EQUAL expression
			{ $$ = Llist6(setarray3_symbol, $1, $10, $3, $5, $7);}
	  | identifier '[' expression ']' COLON_EQUAL assign_statement
			{ $$ = Llist4(setarray1_symbol, $1, $6, $3);}
	  | identifier '[' expression ',' expression ']' COLON_EQUAL assign_statement
			{ $$ = Llist5(setarray2_symbol, $1, $8, $3, $5);}
	  | identifier '[' expression ',' expression ',' expression ']' 
		COLON_EQUAL assign_statement
			{ $$ = Llist6(setarray3_symbol, $1, $10, $3, $5, $7);}
	  ;

algebraic_statement
	  : ALGEBRAIC name WITH expression '=' expression
		{$$ = Llist3(algebraic_symbol, $2, gdiff2($4, $6));}		
	  | ALGEBRAIC name WITH expression 
		{$$ = Llist3(algebraic_symbol, $2, $4);}		
	  ;

try_statement
	  : TRY try_tag DO statements OD
		{$$ = make_try($2, $4, nil);}
	  | TRY try_tag DO statements catch_clause OD
		{$$ = make_try($2, $4, $5);}
	  ;

throw_statement
	  : THROW try_tag  ',' expression
		{$$ = Llist3(throw_symbol, $2, $4);}
	  | THROW try_tag  ',' string
		{$$ = Llist3(throw_symbol, $2, $4);}
	  | THROW try_tag 
		{$$ = Llist3(throw_symbol, $2, nil);}
	  ;

catch_clause
	  : ON CATCH DO statements OD ';'
		{$$ = expand2($4);}
	  | ON CATCH DO statements OD 
		{$$ = expand2($4);}
	  ;


try_tag
	  : SYMBOL
		{$$ = $1;}
	  ;

/*
string_or_expressions
	  : expression
		{$$ = Lncons($1);}
	  | string
		{$$ = Lncons($1);}
	  | string_or_expressions ',' expression
		{$$ = cons($3, $1);}
	  | string_or_expressions ',' string
		{$$ = cons($3, $1);}
	  ;
*/

expression1
	  : expression
		{ $$ = Lncons($1);}
	  | expression1 ',' expression
		{ $$ = Laconc($1, $3);}
	  |
		{ $$ = nil;}
	  ;

expression
	  : '-' expression	%prec UNARY
		{ $$ = Llist2(minus_symbol, $2);}
	  | '+' expression	%prec UNARY
		{ $$ = $2;}
	  | '(' expression ')'
		{ $$ = $2;}
	  | NOT expression
		{ $$ = Llist2($1, $2);}
	  | expression '!'
		{ $$ = Llist2($2, $1);}
          | '{' expression1 '}' 
		{$$ = cons(list_symbol, $2);}
          | '[' expression1 ']' 
		{$$ = cons(set_symbol, $2);}
	  | '%'
		{$$ = onbeepre_symbol;}
	  | '%' '%'
		{$$ = twbeepre_symbol;}
	  | '%' '%' '%'
		{$$ = thbeepre_symbol;}
	  ;

expression
	  : expression '*' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression '/' expression
		{ $$ = Llist3($2, $1, $3);}
          | expression MOD expression 
		{$$ = Llist3($2, $1, $3);}
          | expression SMOD expression 
		{$$ = Llist3($2, $1, $3);}
	  | expression REMAINDER expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression '^' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression '+' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression '-' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression '=' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression NOT_EQUAL expression
            	{$$ = Llist2(not_symbol, Llist3(equal_symbol, $1, $3));}
          | expression UNION expression 
		{$$ = Llist3($2, $1, $3);}
          | expression SET_DIFF expression 
		{$$ = Llist3(set_diff_symbol, $1, $3);}
          | expression INTERSECT expression 
		{$$ = Llist3($2, $1, $3);}
	  | expression '<' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression '>' expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression LESSER_EQUAL expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression GREATER_EQUAL expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression AND expression
		{ $$ = Llist3($2, $1, $3);}
	  | expression OR expression
		{ $$ = Llist3($2, $1, $3);}
	  | name '(' expression1 ')'
		{ $$ = make_funcall($1, $3);}
	  | UNION '(' expression ',' expression ')'
		{ $$ = Llist3($1, $3, $5);}
	  | INTERSECT '(' expression ',' expression ')'
		{ $$ = Llist3($1, $3, $5);}
	  | NUMBER '(' expression ')'
		{ $$ = Llist3(times_symbol, $1, $3);}
	  | NUMBER name
		{ $$ = Llist3(times_symbol, $1, $2);}
	  | name
	  | string
          | QUOTE SYMBOL QUOTE 
		{$$ = Llist2(quote_symbol, $2);}
          | QUOTE_SYMBOL
		{$$ = Llist2(quote_symbol, $1);}
	  | NUMBER
	  | identifier '[' expression ']'
		{$$ = Llist3(getarray1_symbol, $1, $3);}
	  | identifier '[' expression ',' expression ']'
		{$$ = Llist4(getarray2_symbol, $1, $3, $5);}
	  | identifier '[' expression ',' expression ',' expression ']'
		{$$ = Llist5(getarray3_symbol, $1, $3, $5, $7);}
	  | IF expression THEN expression else_expression
		{$$ = make_if($2, $4, $5);}
	  | for_expression
		{$$ = $1;}
	  | foreach_expression
		{$$ = $1;}
	  | error
		{ yyerror("Missing/malformed expression");
		  YYERROR;}
	  ;


else_expression
	  : 
		{$$ = nil;}
	  | ELSE expression
		{ $$ = $2;}
	  | ELIF expression THEN expression else_expression
		{ $$ = make_if($2, $4, $5);}
	  ;



array_declaration
	  : ARRAY identifier '[' expression ']'
		{$$ = Llist3(setq_symbol, $2, 
				Llist2(mkarray1_symbol, $4));}
	  | ARRAY identifier '[' expression ',' expression ']'
		{$$ = Llist3(setq_symbol, $2, 
				Llist3(mkarray2_symbol, $4, $6));}
	  | ARRAY identifier '[' expression ',' expression ',' expression ']'
		{$$ = Llist3(setq_symbol, $2, 
				Llist4(mkarray3_symbol, $4, $6, $8));}
	  | array_declaration ','  identifier '[' expression ']'
		{$$ = Llist3(setq_symbol, $3, 
				Llist2(mkarray1_symbol, $5));
		 $$ = Llist3(progn_symbol, $1, $$);}
	  | array_declaration ','  identifier '[' expression ',' expression ']'
		{$$ = Llist3(setq_symbol, $3, 
				Llist3(mkarray2_symbol, $5, $7));
		 $$ = Llist3(progn_symbol, $1, $$);}
	  | array_declaration ','  identifier '[' expression ',' expression ',' expression ']'
		{$$ = Llist3(setq_symbol, $3, 
				Llist4(mkarray3_symbol, $5, $7, $9));
		 $$ = Llist3(progn_symbol, $1, $$);}
	  | ARRAY error
		{ yyerror("syntax error at array declaration");
		  YYERROR;}

matrix_declaration
	  : MATRIX identifier '[' expression ',' expression ']'
		{$$ = Llist3(setq_symbol, $2, 
				Llist3(mkmatrix_symbol, $4, $6));}
	  | matrix_declaration ','  identifier '[' expression ',' expression ']'
		{$$ = Llist3(setq_symbol, $3, 
				Llist3(mkmatrix_symbol, $5, $7));
		 $$ = Llist3(progn_symbol, $1, $$);}
	  | MATRIX error
		{ yyerror("syntax error at matrix declaration");
		  YYERROR;}

vector_declaration
	  : VECTOR identifier '[' expression ']'
		{$$ = Llist3(setq_symbol, $2, 
				Llist2(mkvector_symbol, $4));}
	  | vector_declaration ',' identifier '[' expression ']'
		{$$ = Llist3(setq_symbol, $3, 
				Llist2(mkvector_symbol, $5));
		 $$ = Llist3(progn_symbol, $1, $$);}
	  | VECTOR error
		{ yyerror("syntax error at vector declaration");
		  YYERROR;}

global_declaration
	  : GLOBAL nameseq
		{if(in_proc)
		     proc_gvars = Lnconc(proc_gvars, $2);$$ = nil;}
	  | GLOBAL array_declaration
		{if(in_proc)
		     proc_gvars = Lnconc(proc_gvars, pickup_globals($2));
		 $$ = $2;}
	  | GLOBAL matrix_declaration
		{if(in_proc)
		     proc_gvars = Lnconc(proc_gvars, pickup_globals($2));
		 $$ = $2;}
	  | GLOBAL vector_declaration
		{if(in_proc)
		     proc_gvars = Lnconc(proc_gvars, pickup_globals($2));
		 $$ = $2;}
	  | GLOBAL assign_statement
		{if(in_proc)
		     proc_gvars = Lnconc(proc_gvars, pickup_globals($2));
		 $$ = $2;}
	  ;

setorder_declaration 
	  : SETORDER ORDER_LEX nameseq 
		{if(in_proc) proc_gvars = Lnconc(proc_gvars, $3);
		 $$ = Llist2($1, cons(list_symbol, 
			cons(mkfxat(TERM_ORDER_LEX), $3)));}
	  | SETORDER ORDER_TLEX nameseq 
		{if(in_proc) proc_gvars = Lnconc(proc_gvars, $3);
		 $$ = Llist2($1, cons(list_symbol, 
			cons(mkfxat(TERM_ORDER_TLEX), $3)));}
	  | SETORDER ORDER_RLEX nameseq 
		{if(in_proc) proc_gvars = Lnconc(proc_gvars, $3);
		 $$ = Llist2($1, cons(list_symbol, 
			cons(mkfxat(TERM_ORDER_TREVLEX), $3)));}
	  | SETORDER ORDER_GRADINGS expression ',' nameseq
		{if(in_proc) proc_gvars = Lnconc(proc_gvars, $3);
		 $$ = Llist2($1, cons(list_symbol,
			cons(mkfxat(TERM_ORDER_GRADINGS), cons($3, $5))));}
	  | SETORDER nameseq 
		{if(in_proc) proc_gvars = Lnconc(proc_gvars, $2);
		 $$ = Llist2($1, cons(list_symbol, $2));}
	  ;
	 


func_body
	  : name ARROWOPE localopt expression1
			{$$ = make_func1($1, $3, $4);}
	  | '(' name ')' ARROWOPE localopt expression1
			{$$ = make_func1($2, $5, $6);}
          | '(' nameseq ')' ARROWOPE localopt expression1
	  	  	{ $$ = make_func2($2, $5, $6);}
          | '(' ')' ARROWOPE localopt expression1
	    		{ $$ = make_func2(nil, $4, $5);}
	  ;

proc_declaration
	  : proc_start name '(' nameseq ')' ';' proc_body
		{
		 $$ = make_proc($2, $4, $7);
		 end_proc();
		 in_proc = 0;}
	  | proc_start name '(' ')' ';' proc_body
		{
		 $$ = make_proc($2, nil, $6);
		 end_proc();
		 in_proc = 0;}
	  | proc_start error
		{ end_proc();
		  in_proc = 0; YYERROR; }
	  ;
proc_start: PROC
		{ begin_proc();
		  in_loop = 0;
		  in_proc = 1;}
	  ;

proc_body : BEGIN localopts statements END
		{ $$ = make_proc_body($3);}
	  ;

return_statement
	  : RETURN expression
		{$$ = make_return1($2);}
	  | RETURN 
		{$$ = make_return0();}
	  ;
break_statement
	  : BREAK
		{$$ = make_break($1);}
	  ;
continue_statement
	  : CONTINUE
		{$$ = make_continue($1);}
	  ;


localopt  : LOCAL nameseq ';' 
		{proc_frame = Lnconc(proc_frame, $2);
		 $$ = $2;}
          | {$$ = nil;}
          ;
localopts : localopt
		{ $$ = $1;}
	  | localopts localopt
		{ $$ = $1;}
          ;



nameseq   
          : name {$$ = Lncons($1);}
	  | nameseq ',' name  {$$ = Laconc($1, $3);}
          ;

name      : identifier {$$ = $1;}
          | name '.' NUMBER { $$ = make_name1($1, $3);}
          | name '.' identifier { $$ = make_name1($1, $3);}
          ;

identifier    
	  : SYMBOL {if(in_proc) proc_vars = Lsetaddq($1, proc_vars);$$ = $1;}
          | '`' SYMBOL '`' {$$ = Lsymbol_name($2);}
          ;

string    
	  : STRING {$$ = $1;}
          ;

%%

void
yyerror(char *s)
{
    yacc_warnings(s, nil, nil);
}

Sex
make_name1(Sex a, Sex b)
{
    if (is_cons(a)) {
	return Laconc(a, b);
    }
    return Llist3(name_symbol, a, b);
}

Sex
pickup_globals(Sex a)
{
    Sex             c;
#if 0
    lisp_printf("pickup: %p\n", a);
#endif
    push3(a, nil, nil);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    if (is_cons(a)) {
	if (car(a) == progn_symbol) {
	    while (car(a) == progn_symbol) {
		b = caddr(a);
		if (car(b) == setq_symbol)
		    c = cons(cadr(b), c);
		a = cadr(a);
	    }
	}
	if (car(a) == setq_symbol) {
	    c = cons(cadr(a), c);
	}
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}




static          Sex
make_funcall(Sex a, Sex b)
{
    proc_vars = Ldeleq(a, proc_vars);
    return cons(a, b);
}

static          Sex
make_func1(b, c, d)
     Sex             b,
                     c,
                     d;
{
    Sex             p;
    stackcheck();
    push3(b, c, d);
    stack[0] = expand2(d);
    b = stack[-2];
    stack[-2] = Lncons(b);
    c = stack[-1];
    if (c != nil) {
	d = stack[0];
	d = Llist3(prog_symbol, c, d);
	stack[0] = d;
    } else
	d = stack[0];
    b = stack[-2];
    p = Llist3(lambda, b, d);
    popv(3);
    return p;
}


static          Sex
make_func2(b, c, d)
     Sex             b,
                     c,
                     d;
{
    Sex             p;
    stackcheck();
    push3(b, c, d);
    stack[0] = expand2(d);
    c = stack[-1];
    if (c != nil) {
	d = stack[0];
	d = Llist3(prog_symbol, c, d);
    } else
	d = stack[0];
    b = stack[-2];
    p = Llist3(lambda, b, d);
    popv(3);
    return p;
}

static          Sex
make_proc(a, b, c)
     Sex             a,
                     b,
                     c;
{
    Sex             d;
    push3(a, b, c);
#if 0
    lisp_printf("proc_vars = %p\n", proc_vars);
    lisp_printf("proc_gvar = %p\n", proc_gvars);
    lisp_printf("proc_args = %p\n", b);
    lisp_printf("proc_frame = %p\n", proc_frame);
#endif
    d = Lsetdiff(proc_vars, proc_frame);
    d = Lsetdiff(d, stack[-1]);
    d = Lsetdiff(d, proc_gvars);
    d = Ldeleq(stack[-2], d);
    d = Ldeleq(asuka_true_symbol, d);
    d = Ldeleq(asuka_false_symbol, d);
    d = Ldeleq(asuka_nil_symbol, d);
    if (d != nil) {
	lisp_printf(" *** Message: global variabls %p used in", d);
	lisp_printf(" procedure %p\n", stack[-2]);
    }
    d = Lintersectionq(proc_gvars, stack[-1]);
    if (d != nil) {
	lisp_printf
	    (" *** Message: variables %p is used as both global and arguments\n",
	     d);
    }
    d = Lintersectionq(proc_gvars, proc_frame);
    if (d != nil) {
	lisp_printf
	    (" *** Message: variables %p is used as both global and local\n",
	     d);
    }

    pop3(c, b, a);
    return Llist4(defproc_symbol, a, b, c);
}

static          Sex
make_proc_body(Sex a)
{
    Sex             b;
    if (!is_cons(a))
	a = cons(prog_symbol, cons(proc_frame, a));
    else if (caar(a) == prog_symbol) {
	proc_frame = Lnconc(proc_frame, cadar(a));
	a = cons(prog_symbol, cons(proc_frame, cddar(a)));
    } else if (caar(a) == progn_symbol) {
	a = cons(prog_symbol, cons(proc_frame, cdar(a)));
    } else {
	a = cons(prog_symbol, cons(proc_frame, a));
    }
    return a;
}


Sex
expand2(Sex a)
{
    if (a == nil)
	return a;
    if (cdr(a) == nil)
	return car(a);
    if (car(a) == prog_symbol)
	return a;
    return cons(progn_symbol, a);
}

/*
 *procedure definition
 */

void
begin_loop(void)
{
    char            buf[128];
    Sex             a;

    in_loop++;

    // lisp_printf("entering loop\n");
    stackcheck();
    push3(nil, nil, nil);	/* label symbol */
    sprintf(buf, "label%d", label_count++);	// start label
    stack[-2] = gensym2(buf);
    sprintf(buf, "label%d", label_count++);	// continue label
    stack[-1] = gensym2(buf);
    sprintf(buf, "label%d", label_count++);	// exit label
    stack[0] = gensym2(buf);
    a = Llist3(stack[0], stack[-1], stack[-2]);
    loop_frame = cons(a, loop_frame);
    popv(3);
}

void
end_loop(void)
{
    if (in_loop <= 0 || !is_cons(loop_frame)) {
	in_loop = 0;
	lisp_printf("over exitting loop\n");
    } else {
	loop_frame = cdr(loop_frame);
	in_loop--;
	// lisp_printf("exitting loop\n");
    }
}

void
clear_loop(void)
{
    while (in_loop > 0) {
	popv(3);
	in_loop--;
	// lisp_printf("cleaning loop\n");
    }
}


Sex
expand_loop(Sex body)
{
    if (!is_cons(body))
	return body;
    if (car(body) == progn_symbol)
	return cdr(body);
    if (car(body) == prog_symbol)
	return cddr(body);
    return body;
}


Sex
expand_while(a, exp, body)
     Sex             a,
                     exp,
                     body;
{
    Sex             q;
    Sex             l;

    if (!is_cons(loop_frame))
	error("invalid while");
    l = car(loop_frame);
    stackcheck();
    push3(a, exp, body);
    push4(nil, nil, nil, l);
#define a stack[-6]
#define exp stack[-5]
#define body stack[-4]
#define gost stack[-3]
#define cond stack[-2]
#define p stack[-1]
#define l stack[-0]
    gost = Llist2(go_symbol, car(l));
    cond = Llist4(if_symbol, exp, nil, gost);
    gost = Llist2(go_symbol, caddr(l));
    // lisp_printf("\n while body=%p\n\n", body);
    if (in_proc) {
	if (!is_cons(body)) {
	    p = Llist3(cadr(l), gost, car(l));
	    p = Lnconc(body, p);
	} else if (caar(body) == prog_symbol) {
	    proc_frame = Lnconc(proc_frame, cadar(body));
	    p = Llist3(cadr(l), gost, car(l));
	    p = Lnconc(cddar(body), p);
	} else if (caar(body) == progn_symbol) {
	    p = Llist3(cadr(l), gost, car(l));
	    p = Lnconc(cdar(body), p);
	} else {
	    p = Llist3(cadr(l), gost, car(l));
	    p = Lnconc(body, p);
	}
    } else {
	p = Llist3(cadr(l), gost, car(l));
	p = Lnconc(body, p);
    }

    p = Lnconc(Llist2(caddr(l), cond), p);
    q = Lnconc(Llist2(prog_symbol, nil), p);
#undef a
#undef exp
#undef body
#undef gost
#undef cond
#undef p
#undef l
    popv(7);
    return q;
}


static          Sex
expand_repeat(a, exp, body)
     Sex             a,
                     exp,
                     body;
{
    Sex             q;
    Sex             l;

    if (!is_cons(loop_frame))
	error("invalid repeat");
    l = car(loop_frame);
    stackcheck();
    push3(a, exp, body);
    push4(nil, nil, nil, l);
#define a stack[-6]
#define exp stack[-5]
#define body stack[-4]
#define gost stack[-3]
#define cond stack[-2]
#define p stack[-1]
#define l stack[-0]
    gost = Llist2(go_symbol, caddr(l));
    cond = Llist4(if_symbol, exp, nil, gost);
    if (in_proc) {
	if (car(body) == prog_symbol) {
	    proc_frame = Lnconc(proc_frame, cadr(body));
	    p = Llist3(cadr(l), cond, car(l));
	    p = Lnconc(cddr(body), p);
	} else if (car(body) == progn_symbol) {
	    p = Llist3(cadr(l), cond, car(l));
	    p = Lnconc(cdr(body), p);
	} else {
	    p = Llist3(cadr(l), cond, car(l));
	    p = Lnconc(body, p);
	}
    } else {
	p = Llist3(cadr(l), cond, car(l));
	p = Lnconc(body, p);
    }
    l = Llist3(prog_symbol, nil, caddr(l));
    q = Lnconc(l, p);
#undef a
#undef exp
#undef body
#undef gost
#undef cond
#undef p
#undef l
    popv(7);
    return q;
}

static          Sex
make_if(cond, t, f)
     Sex             cond,
                     t,
                     f;
{
    Sex             q;
    int             flg = 0;

    if (!in_proc)
	return Llist4(if_symbol, cond, t, f);

    stackcheck();
    push3(cond, t, f);
    push3(nil, nil, nil);

#define cond stack[-5]
#define t stack[-4]
#define f stack[-3]
#define p stack[-2]
#define l stack[-1]
#define gost stack[-0]

    if (is_cons(t) && car(t) == prog_symbol) {
	flg = 1;
	proc_frame = Lnconc(proc_frame, cadr(t));
	t = cons(progn_symbol, cddr(t));
    }
    if (is_cons(f) && car(f) == prog_symbol) {
	flg = 1;
	proc_frame = Lnconc(proc_frame, cadr(f));
	f = cons(progn_symbol, cddr(f));
    }
    if (flg == 0) {
	q = Llist4(if_symbol, cond, t, f);
	popv(6);
	return q;
    }
    begin_loop();
    if (!is_cons(loop_frame)) {
	popv(6);
	error("invalid if");
    }
    l = car(loop_frame);
    if (t == nil) {
	gost = Llist2(go_symbol, car(l));
	cond = Llist4(if_symbol, cond, gost, nil);
	if (car(f) == progn_symbol)
	    f = cdr(f);
	else
	    f = cons(f, nil);
	p = cons(cond, f);
	p = Laconc(p, car(l));
    } else {
	gost = Llist2(go_symbol, car(l));
	cond = Llist4(if_symbol, cond, nil, gost);
	if (car(t) == progn_symbol)
	    t = cdr(t);
	else
	    t = cons(t, nil);
	p = cons(cond, t);
	if (f == nil) {
	    p = Laconc(p, car(l));
	} else {
	    gost = Llist2(go_symbol, cadr(l));
	    p = Laconc(p, gost);
	    p = Laconc(p, car(l));
	    if (car(f) == progn_symbol)
		f = cdr(f);
	    else
		f = cons(f, nil);
	    p = Lnconc(p, f);
	    p = Laconc(p, cadr(l));
	}
    }
    end_loop();
    p = cons(nil, p);
    q = cons(prog_symbol, p);
#undef cond
#undef t
#undef f
#undef p
#undef l
#undef gost
    popv(6);
    return q;
}

static          Sex
make_break(Sex a)
{
    if (!is_cons(loop_frame))
	error("break outside loop");
    return Llist2(go_symbol, car(car(loop_frame)));
}

static          Sex
make_continue(Sex a)
{
    if (!is_cons(loop_frame))
	error("continue outside loop");
    return Llist2(go_symbol, cadr(car(loop_frame)));
}

static          Sex
make_return0(void)
{
    if (!in_proc)
	error("invalid return");
    return cons(return_symbol, nil);
}

static          Sex
make_return1(Sex a)
{
    if (!in_proc)
	error("invalid return");
    a = cons(a, nil);
    return cons(return_symbol, a);
}


static          Sex
expand_for(a, var, body)
     Sex             a,
                     var,
                     body;
{
    Sex             is,
                    ie,
                    ib;
    Sex             q;
    Sex             l;
    long            b;

    b = mkint(car(var));
    var = cdr(var);

    if (!is_cons(loop_frame))
	error("invalid for");

    is = cadr(var);
    ib = cdr(cdr(var));
    ie = car(ib);
    ib = cadr(ib);
    var = car(var);

    l = car(loop_frame);

    stackcheck();
    push3(a, var, body);
    push2(nil, nil);
    push3(is, ie, ib);
    push2(nil, l);
#define a stack[-9]
#define var stack[-8]
#define body stack[-7]
#define gost stack[-6]
#define cond stack[-5]
#define is stack[-4]
#define ie stack[-3]
#define ib stack[-2]
#define p stack[-1]
#define l stack[-0]
    is = Llist3(setq_symbol, var, is);
    gost = Llist2(go_symbol, car(l));
    switch (b) {
    case 0:
	cond = Llist3(gt_symbol, var, ie);
	ib = Llist3(add_symbol, var, ib);
	break;
    case 1:
	cond = Llist3(lt_symbol, var, ie);
	ib = Llist3(sub_symbol, var, ib);
	break;
    default:
	cond = Llist2(not_symbol, ie);
	ib = Llist3(add_symbol, var, ib);
	break;
    }
    cond = Llist4(if_symbol, cond, gost, nil);
    ib = Llist3(setq_symbol, var, ib);
    gost = Llist2(go_symbol, caddr(l));
    // lisp_printf("\n for body=%p\n\n", body);
    if (in_proc) {
	if (caar(body) == prog_symbol) {
	    proc_frame = Lnconc(proc_frame, cadar(body));
	    p = Llist4(cadr(l), ib, gost, car(l));
	    p = Lnconc(cddar(body), p);
	} else if (caar(body) == progn_symbol) {
	    p = Llist4(cadr(l), ib, gost, car(l));
	    p = Lnconc(cdar(body), p);
	} else {
	    p = Llist4(cadr(l), ib, gost, car(l));
	    p = Lnconc(body, p);
	}
    } else {
	p = Llist4(cadr(l), ib, gost, car(l));
	p = Lnconc(body, p);
    }
    // p = Llist7(stack[-2], cond, body, stack[-1], ib, gost, stack[0]);

    if (ie != nil) {
	l = Llist2(caddr(l), cond);
	p = Lnconc(l, p);
    } else {
	p = cons(caddr(l), p);
    }
    p = cons(is, p);
    p = cons(nil, p);
    q = cons(prog_symbol, p);
#undef a
#undef var
#undef body
#undef gost
#undef cond
#undef is
#undef ie
#undef ib
#undef p
#undef l
    popv(10);
    return q;
}

static          Sex
expand_foreach(a, b, body)		/* v, c, body, f */
     Sex             a,
                     b, body;
{
    Sex             q;

    if (!is_cons(loop_frame))
	error("invalid foreach");

    stackcheck();
    push2(a, b);
    push4(nil, body, nil, nil);
    push2(nil, nil);
#define a stack[-7]
#define b stack[-6]
#define p stack[-5]
#define body stack[-4]
#define v stack[-3]
#define gost stack[-2]
#define f stack[-1]
#define l stack[-0]
    l = car(loop_frame);
    v = Lgensym();
    p = Llist3(setq_symbol, v, car(cdr(b)));
    f = Llist3(equal_symbol, v, nil);
    gost = Llist2(go_symbol, car(l));
    gost = Llist4(if_symbol, f, gost, nil);
    f = caddr(b);
    if (f == nil) {
	f = Llist2(car_symbol, v);
	f = Llist3(setq_symbol, car(b), f);
	p = Llist4(p, caddr(l), gost, f);
    } else {
	f = Llist3(setq_symbol, car(b), v);
	p = Llist4(p, caddr(l), gost, f);
    }

    //body = caddr(b);
    // lisp_printf("\n foreach body=%p\n\n", body);
    if (in_proc) {
	if (caar(body) == prog_symbol) {
	    proc_frame = Lnconc(proc_frame, cadar(body));
	    p = Lnconc(p, cddar(body));
	} else if (caar(body) == progn_symbol) {
	    p = Lnconc(p, cdar(body));
	} else {
	    p = Lnconc(p, body);
	}
    } else {
	p = Lnconc(p, body);
    }
    gost = Llist2(go_symbol, caddr(l));
    f = Llist2(cdr_symbol, v);
    body = Llist3(setq_symbol, v, f);
    f = Llist4(cadr(l), body, gost, car(l));
    p = Lnconc(p, f);
    f = Lncons(v);
    f = Llist2(prog_symbol, f);
    q = Lnconc(f, p);
#undef a
#undef b
#undef p
#undef body
#undef v
#undef gost
#undef f
#undef l
    popv(8);
    return q;
}

/**********************************************************************/
/**********************************************************************/


static          Sex
expand_for_expression(a, var, body)
     Sex             a,
                     var,
                     body;
{
    Sex             is,
                    ie,
                    ib;
    Sex             q;
    Sex             l;
    long            b;
    long            c;

    c = mkint(a);
    b = mkint(car(var));
    var = cdr(var);

    if (!is_cons(loop_frame))
	error("invalid for");

    is = cadr(var);
    ib = cdr(cdr(var));
    ie = car(ib);
    ib = cadr(ib);
    var = car(var);

    l = car(loop_frame);

    stackcheck();
    push3(a, var, body);
    push2(nil, nil);
    push3(is, ie, ib);
    push3(nil, l, nil);
#define a stack[-10]
#define var stack[-9]
#define body stack[-8]
#define gost stack[-7]
#define cond stack[-6]
#define is stack[-5]
#define ie stack[-4]
#define ib stack[-3]
#define p stack[-2]
#define l stack[-1]
#define r stack[-0]
    r = Lgensym();
    switch (c) {
    case 1:
	p = Llist3(setq_symbol, r, mkfxat(0));
	break;
    case 2:
	p = Llist3(setq_symbol, r, mkfxat(1));
	break;
    default:
	p = Llist3(setq_symbol, r, nil);
	break;
    }
    is = Llist3(setq_symbol, var, is);
    is = Llist2(p, is);
    gost = Llist2(go_symbol, car(l));
    switch (b) {
    case 0:
	cond = Llist3(gt_symbol, var, ie);
	ib = Llist3(add_symbol, var, ib);
	break;
    case 1:
	cond = Llist3(lt_symbol, var, ie);
	ib = Llist3(sub_symbol, var, ib);
	break;
    default:
	cond = Llist2(not_symbol, ie);
	ib = Llist3(add_symbol, var, ib);
	break;
    }
    cond = Llist4(if_symbol, cond, gost, nil);
    ib = Llist3(setq_symbol, var, ib);
    gost = Llist2(go_symbol, caddr(l));

    p = Llist4(cadr(l), ib, gost, car(l));
    switch (c) {
    case 1:
	q = Llist3(add_symbol, r, body);
	break;
    case 2:
	q = Llist3(times_symbol, r, body);
	break;
    case 3:
	q = Llist3(cons_symbol, body, r);
	break;
    case 4:
	q = Llist3(append_symbol, r, body);
	break;
    default:
	q = Llist3(nconc_symbol, r, body);
	break;
    }
    q = Llist3(setq_symbol, r, q);
    p = cons(q, p);

    if (ie != nil) {
	l = Llist2(caddr(l), cond);
	p = Lnconc(l, p);
    } else {
	p = cons(caddr(l), p);
    }

    if (c == 3) {
	q = Llist2(reversip_symbol, r);
	q = Llist2(return_symbol, q);
    } else {
	q = Llist2(return_symbol, r);
    }
    q = cons(q, nil);
    p = Lnconc(p, q);
    p = Lnconc(is, p);
    q = cons(r, nil);
    p = cons(q, p);
    q = cons(prog_symbol, p);
#undef a
#undef var
#undef body
#undef gost
#undef cond
#undef is
#undef ie
#undef ib
#undef p
#undef l
#undef r
    popv(11);
    return q;
}


static          Sex
expand_foreach_expression(a, b, body)	/* v, c, body, f */
     Sex             a,
                     b, body;
{
    Sex             q;
    long            c;

    c = mkint(a);

    if (!is_cons(loop_frame))
	error("invalid foreach");

    stackcheck();
    push2(a, b);
    push4(nil, body, nil, nil);
    push3(nil, nil, nil);
#define a stack[-8]
#define b stack[-7]
#define p stack[-6]
#define body stack[-5]
#define v stack[-4]
#define gost stack[-3]
#define f stack[-2]
#define l stack[-1]
#define r stack[-0]
    l = car(loop_frame);
    v = Lgensym();
    r = Lgensym();
    switch (c) {
    case 1:
	p = Llist3(setq_symbol, r, mkfxat(0));
	break;
    case 2:
	p = Llist3(setq_symbol, r, mkfxat(1));
	break;
    default:
	p = Llist3(setq_symbol, r, nil);
	break;
    }
    q = Llist3(setq_symbol, v, car(cdr(b)));
    p = Llist2(p, q);
    f = Llist3(equal_symbol, v, nil);
    gost = Llist2(go_symbol, car(l));
    gost = Llist4(if_symbol, f, gost, nil);
    f = caddr(b);
    if (f == nil) {
	f = Llist2(car_symbol, v);
	f = Llist3(setq_symbol, car(b), f);
	f = Llist3(caddr(l), gost, f);
	p = Lnconc(p, f);
    } else {
	f = Llist3(setq_symbol, car(b), v);
	f = Llist3(caddr(l), gost, f);
	p = Lnconc(p, f);
    }

    //body = caddr(b);
    // lisp_printf("\n foreach body=%p\n\n", body);
    switch (c) {
    case 1:
	q = Llist3(add_symbol, r, body);
	break;
    case 2:
	q = Llist3(times_symbol, r, body);
	break;
    case 3:
	q = Llist3(cons_symbol, body, r);
	break;
    case 4:
	q = Llist3(append_symbol, r, body);
	break;
    default:
	q = Llist3(nconc_symbol, r, body);
	break;
    }
    q = Llist3(setq_symbol, r, q);
    q = cons(q, nil);
    p = Lnconc(p, q);
    gost = Llist2(go_symbol, caddr(l));
    f = Llist2(cdr_symbol, v);
    body = Llist3(setq_symbol, v, f);
    f = Llist4(cadr(l), body, gost, car(l));
    p = Lnconc(p, f);
    if (c == 3) {
	q = Llist2(reversip_symbol, r);
	q = Llist2(return_symbol, q);
    } else {
	q = Llist2(return_symbol, r);
    }
    q = cons(q, nil);
    p = Lnconc(p, q);
    f = Llist2(v, r);
    f = Llist2(prog_symbol, f);
    q = Lnconc(f, p);
#undef a
#undef b
#undef p
#undef body
#undef v
#undef gost
#undef f
#undef l
#undef r
    popv(9);
    return q;
}

/**********************************************************************/
/**********************************************************************/

static          Sex
make_statements1(Sex a)
{
    if (a == nil)
	return a;
    if (!in_proc)
	return Lncons(a);
    stackcheck();
    push(a);
#define a stack[0]
    if (is_cons(a) && car(a) == prog_symbol) {
	proc_frame = Lnconc(proc_frame, car(cdr(a)));
	a = cons(nil, cdr(cdr(a)));
	a = cons(prog_symbol, a);
    }
#undef a
    a = stack[0];
    popv(1);
    return Lncons(a);
}


static          Sex
make_statements(Sex l, Sex a)
{
    // lisp_printf("make_statements enters: %p, %p\n", l, a);
    if (a == nil)
	return l;
    if (l == nil)
	return make_statements1(a);
    if (!in_proc)
	return Laconc(l, a);
    stackcheck();
    push4(nil, nil, l, a);
#define c stack[-3]
#define la stack[-2]
#define l stack[-1]
#define a stack[0]
    la = car(l);
    if (car(la) != prog_symbol) {
	if (is_cons(a) && car(a) == prog_symbol) {
	    a = cdr(a);
	    proc_frame = Lnconc(proc_frame, car(a));
	    c = Llist2(prog_symbol, nil);
	    c = Lnconc(c, l);
	    l = Lnconc(c, cdr(a));
	    l = Lncons(l);
	} else {
	    l = Laconc(l, a);
	}
    } else if (is_cons(a) && car(a) == prog_symbol) {
	a = cdr(a);
	proc_frame = Lnconc(proc_frame, car(a));
	l = Lnconc(la, cdr(a));
	l = Lncons(l);
    } else {
	l = Laconc(la, a);
	l = Lncons(l);
    }
#undef c
#undef a
    a = l;
#undef l
#undef la
    popv(4);
    // lisp_printf("make_statements return with : %p\n", a);
    return a;
}

static void
begin_proc(void)
{
    Sex             a;
    in_proc = 1;
    proc_frame = nil;
    proc_vars = nil;
    proc_gvars = nil;
}

static void
end_proc(void)
{
    proc_frame = nil;
    proc_vars = nil;
    proc_gvars = nil;
    in_proc = 0;
}

Sex
make_try(Sex tag, Sex body, Sex catch_body)
{
    Sex             a;
    stackcheck();
    if (body == nil)
	return nil;
    if (catch_body == nil) {
	push(tag);
	body = expand2(body);
	pop(tag);
	a = Llist3(catch_symbol, tag, body);
    } else {
	/*
	 * (if (catch ...) nil catch_body) 
	 */
	push2(catch_body, tag);
	body = expand2(body);
	pop(tag);
	a = Llist3(catch_symbol, tag, body);
	pop(catch_body);
	a = Llist4(if_symbol, a, nil, catch_body);
    }
    return a;
}

Sex
make_switch(Sex tag, Sex cases)
{
    Sex             p;
    /*
     * case = (value, sts) . ... 
     */
    /*
     * tag:=tag, if tag = car 
     */


    // lisp_printf("tag = %p\n", tag);
    // lisp_printf("cases = %p\n", cases);

    p = cons(tag, cases);
    p = cons(switch_symbol, p);
    return p;
}

Sex
make_case(Sex tag, Sex cases, Sex elses)
{
    Sex             p;
    /*
     * case = (value, sts) . ... 
     */
    /*
     * tag:=tag, if tag = car 
     */


    // lisp_printf("tag = %p\n", tag);
    // lisp_printf("cases = %p\n", cases);

    p = Llist3(tag, cases, elses);
    p = cons(case_symbol, p);
    return p;
}
