#include <ctype.h>
#include <math.h>
#include <limits.h>

#include "config.h"

#ifdef Android
double my_atof(const char *p);
char * stpcpy(char *s1, const char *s2);

extern double   ldexp();

#define POW1_25LEN	9
static double   pow1_25[POW1_25LEN] = { 0.0 };
#define DBL_EXP_MAX	308


double
my_atof(p)
     const char     *p;
{
    int             c;
    int             exp = 0,
	neg_val = 0;
    long            high = 0,
	low = 0,
	scale = 1;
    int             decpt = 0,
	nzeroes = 0;
    double          fl_val;
    int             e_exp = 0,
	neg_exp = 0;
    double         *powptr = pow1_25,
	fl_exp;

    while (isspace(c = *p))	/* eat leading white space */
	p++;

    switch (c) {		/* process sign */
    case '-':
	neg_val++;
    case '+':
	p++;
    }

    while ((isdigit(c = *p++) || c == '.') && !decpt++) {
	if (c == '.')
	    continue;
	if (decpt) {		/* handle trailing zeroes specially */
	    if (c == '0') {	/* ignore zero for now */
		nzeroes++;
		continue;
	    }
	    while (nzeroes > 0) {	/* put zeroes back in */
		exp--;
		if (high < LONG_MAX / 10) {
		    high *= 10;
		} else if (scale < LONG_MAX / 10) {
		    scale *= 10;
		    low *= 10;
		} else
		    exp++;
		nzeroes--;
	    }
	    exp--;		/* decr exponent if decimal pt. seen */
	}
	if (high < LONG_MAX / 10) {
	    high *= 10;
	    high += c - '0';
	} else if (scale < LONG_MAX / 10) {
	    scale *= 10;
	    low *= 10;
	    low += c - '0';
	} else
	    exp++;
    }
    if (!high)
	return (0.0);
    fl_val = (double) high;
    if (scale > 1)
	fl_val = (double) scale *fl_val + (double) low;

    if (c == 'E' || c == 'e') {	/* accumulate exponent */

	switch (*p) {		/* process sign */
	case '-':
	    neg_exp++;
	case '+':
	case ' ':
	    p++;
	}
	if (isdigit(c = *p)) {	/* found a legitimate exponent */
	    do {
		/*
		 * limit outrageously large exponents 
		 */
		if (e_exp < DBL_EXP_MAX)
		    e_exp = 10 * e_exp + c - '0';
	    } while (isdigit(c = *++p));
	    if (neg_exp)
		exp -= e_exp;
	    else
		exp += e_exp;
	}
    }
    if (exp != 0) {		/* apply exponent */
	fl_exp = fl_val;

	if (*powptr == 0.0) {	/* need to initialize table */
	    *powptr = 1.25;
	    for (; powptr < &pow1_25[POW1_25LEN - 1]; powptr++)
		powptr[1] = *powptr * *powptr;
	    powptr = pow1_25;
	}
	if ((c = exp) < 0) {
	    c = -c;
	    fl_exp = 1.0;
	}
	if (c >= DBL_EXP_MAX / 2)	/* outrageously large exponents */
	    c = DBL_EXP_MAX / 2 - 1;	/* will be handled by ldexp */
	for (;; powptr++) {
	    if (c & 01)
		fl_exp *= *powptr;
	    if ((c >>= 1) == 0)
		break;
	}
	fl_val = ldexp(exp < 0 ? fl_val / fl_exp : fl_exp, 3 * exp);
    }
    return (neg_val ? -fl_val : fl_val);	/* apply sign */
}


char *
stpcpy(char *s1, const char *s2)
{
        char *p = s1;

        while(*s1++ = *s2++)
                ;
        return(p);
}

#endif
