/*
 * befloat.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * big effective float routine
 *      $Header: /Users/kako/asuka/src/RCS/befloat.c,v 2.2 2014/12/22 07:22:30 kako Exp $
 *
 * data structure of befloat is
 *    value   bfloat
 *    error   integer     relative error == log_2 (error_value / true_value)
 */
static char const rcsid[] =
    "$Id: befloat.c,v 2.2 2014/12/22 07:22:30 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex             befloatit(Sex a);
Sex             befloatiti(Sex a);
void            print_error(double d, putchar_fn * fn);
Sex             times2Ec(Sex a, Sex b);
Sex             times2Ez(Sex a, Sex b);


/***************************************************************/
/****** error is double ****************************************/
/***************************************************************/

Sex
Lbefloat_error(Sex a)
{
    return make_float(befloat_error(a));
}

Sex
Lbefloat_value(Sex a)
{
    return befloat_value(a);
}

Sex
Lmkbefloat(Sex a, Sex b)
{
    Sex             c;
    double          e;
    double          f;

    e = floatit(b);
    /*
     * f = floatit(a); 
     */
    /*
     * e = f*1.0e-5; 
     */
    e = fabs(e);
    stackcheck();
    if (is_befloat(a)) {
	push(a);
	c = new_befloat();
	pop(a);
	befloat_value(c) = befloat_value(a);
	befloat_error(c) = e;
	return c;
    } else if (is_becomplex(a)) {
	push(a);
	c = new_becomplex();
	pop(a);
	becomplex_real(c) = becomplex_real(a);
	becomplex_imag(c) = becomplex_imag(a);
	becomplex_error(c) = e;
	return c;
    } else {
	a = Lbefloat(a);
	if (is_befloat(a))
	    befloat_error(a) = e;
	else if (is_becomplex(a))
	    becomplex_error(a) = e;
	return a;
    }
}

Sex
Lmkbefloat_by_ratio(Sex a, Sex b)
{
    Sex             c;
    double          e;
    double          r;
    double          i;

    e = floatit(b);
    e = fabs(e);
    stackcheck();
    if (is_befloat(a)) {
	push(a);
	c = new_befloat();
	pop(a);
	befloat_value(c) = befloat_value(a);
	r = floatit(befloat_value(a));
	r = fabs(r);
	befloat_error(c) = r * e;
	return c;
    } else if (is_becomplex(a)) {
	push(a);
	c = new_becomplex();
	pop(a);
	becomplex_real(c) = becomplex_real(a);
	r = floatit(becomplex_real(a));
	r = fabs(r);
	becomplex_imag(c) = becomplex_imag(a);
	i = floatit(becomplex_imag(a));
	i = fabs(i);
	if (r < i)
	    r = i;
	becomplex_error(c) = r * e;
	return c;
    } else {
	a = Lbefloat(a);
	if (is_befloat(a)) {
	    r = floatit(befloat_value(a));
	    r = fabs(r);
	    befloat_error(a) = r * e;
	} else if (is_becomplex(a)) {
	    r = floatit(becomplex_real(a));
	    r = fabs(r);
	    i = floatit(becomplex_imag(a));
	    i = fabs(i);
	    if (r < i)
		r = i;
	    becomplex_error(a) = r * e;
	}
	return a;
    }
}




Sex
make_befloat(Sex v, double n)
{
    Sex             r;

    stackcheck();
    push(v);
    r = new_befloat();
    pop(v);
    befloat_value(r) = v;
    befloat_error(r) = n;
    return r;
}
/*
 * conversion routines 
 */


Sex
befloatiti(Sex a)
{
    a = bfloatiti(a);
    return make_befloat(a, 0.0);
}

Sex
befloatitb(Sex a)
{
    Sex             v;
    long            n;
    double          x;

    x = floatitb(a);
    x = fabs(x) * current_precision;
    v = bfloatitb(a);
    stackcheck();
    push(v);
    a = new_befloat();
    pop(v);
    befloat_value(a) = v;
    befloat_error(a) = x;
    return a;
}

Sex
befloatitr(Sex a)
{
    Sex             v;
    double          x;
    v = bfloatitr(a);
    x = floatitd(v);
    x = fabs(x) * current_precision;
    stackcheck();
    push(v);
    a = new_befloat();
    pop(v);
    befloat_value(a) = v;
    befloat_error(a) = x;
    return a;
}

Sex
befloatitd(Sex a)
{
    Sex             c;
    double          x;
    x = floatitd(a);
    x = fabs(x) * current_precision;
    stackcheck();
    push(a);
    c = new_befloat();
    pop(a);
    befloat_value(c) = a;
    befloat_error(c) = x;
    return c;
}

Sex
befloatitf(double x)
{
    long            e;
    double          y;
    /*
     * y = fabs(x) * flt_min;
     */
    y = fabs(x) * current_precision;
    return make_befloat(bfloatitf(x), y);
}

Sex
befloatite(Sex a)
{
    double          n;
    double          d;

    d = efloat_value(a);
    n = efloat_error(a);
    a = bfloatitf(d);
    return make_befloat(bfloatitf(d), n);
}

Sex
befloatitv(Sex a)
{
    double          d,
                    e;

    d = interval_upper(a) + interval_lower(a);
    e = interval_upper(a) - interval_lower(a);
    a = bfloatitf(d * 0.5);
    return make_befloat(a, e * 0.5);
}

Sex
befloatitV(Sex a)
{
    Sex             d;
    double          e;

    stackcheck();
    push(a);
    d = plus2dd(binterval_upper(a), binterval_lower(a));
    pop(a);
    push(d);
    e = floatitd(binterval_upper(a)) - floatitd(binterval_lower(a));
    pop(d);
    bfloat_exp(d) = bfloat_exp(d) - 1;
    return make_befloat(d, e * 0.5);
}

Sex
befloatit(Sex a)
{
    DoubleFloat     u;
    if (is_fixnum(a)) {
	return befloatiti(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return bfloatitb(a);
	case N_FLOAT:
	    return befloatitf(float_val(a));
	case N_RATIONAL:
	    return befloatitr(a);
	case N_BFLOAT:
	    return befloatitd(a);
	case N_EFLOAT:
	    return befloatite(a);
	case N_INTERVAL:
	    return befloatitv(a);
	case N_BEFLOAT:
	    return a;
	case N_BINTERVAL:
	    return befloatitV(a);
	}
    }
    return error1("bad argument to befloat", a);
}


Sex
Lbefloat(Sex a)
{
    DoubleFloat     u;

    if (is_fixnum(a)) {
	return befloatiti(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return bfloatitb(a);
	case N_FLOAT:
	    return befloatitf(float_val(a));
	case N_RATIONAL:
	    return befloatitr(a);
	case N_BFLOAT:
	    return befloatitd(a);
	case N_EFLOAT:
	    return befloatite(a);
	case N_INTERVAL:
	    return befloatitv(a);
	case N_BEFLOAT:
	    return a;
	case N_BINTERVAL:
	    return befloatitV(a);
	case N_COMPLEX:
	    return becomplexitc(a);
	case N_ECOMPLEX:
	    return becomplexitz(a);
	case N_BECOMPLEX:
	    return a;
	default:
	    break;
	}
    }
    return error1("bad argument to befloat", a);
}


/*
 * print routine 
 */

void
print_error(double d, putchar_fn * fn)
{
    char           *p;
    char            my_buff[64];
    sprintf(my_buff, "%.3g", d);
    for (p = my_buff; *p; p++)
	fn(*p);
}



void
print_befloat(Sex u, putchar_fn * fn)
{
    long            n;
    Sex             a;
    double          d;

    d = befloat_error(u);
    stackcheck();
    push(befloat_value(u));
    (*fn) ('#');
    (*fn) ('E');
    (*fn) ('(');
    u = stack[0];
    print_bfloat(u, fn);
    (*fn) (',');
    pop(u);
    print_error(d, fn);
    (*fn) (')');
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
plus2bE(Sex a, Sex b)
{
    return plus2xE(a, b);
}

Sex
plus2dE(Sex a, Sex b)
{
    double          n;
    n = befloat_error(b);
    a = plus2dd(a, befloat_value(b));
    return make_befloat(a, n);
}

Sex
plus2fE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    double          x,
                    y;
    double          n;
    n = befloat_error(b);
    x = float_val(a);
    b = befloat_value(b);
    y = floatitd(b);
    x = x + y;
    return make_efloat(x, n);

#else
    return plus2xE(a, b);
#endif
}

Sex
plus2iE(Sex a, Sex b)
{
    return plus2xE(a, b);
}

Sex
plus2rE(Sex a, Sex b)
{
    return plus2xE(a, b);
}

Sex
plus2xE(Sex a, Sex b)
{
    double          c;
    c = befloat_error(b);
    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    a = plus2dd(a, befloat_value(b));
    return make_befloat(a, c);
}



Sex
plus2eE(Sex a, Sex b)
{
    Sex             ev;
    double          ee;
    double          e;
#if TOWARD_LOWER_PRECISION
    double          y;

    e = efloat_value(a);
    ee = efloat_error(a);
    if (ee < befloat_error(b))
	ee = befloat_error(b);
    b = befloat_value(b);
    y = floatitd(b);
    e = e + y;
    return make_efloat(e, ee);
#else

    stackcheck();
    e = efloat_value(a);
    ee = efloat_error(a);
    if (ee < befloat_error(b))
	ee = befloat_error(b);
    push(b);
    ev = bfloatitf(e);
    pop(b);
    ev = plus2dd(ev, befloat_value(b));
    return make_befloat(ev, ee);
#endif
}


Sex
plus2vE(Sex a, Sex b)
{
    Sex             ev;
    double          vv,
                    ve;
#if TOWARD_LOWER_PRECISION
    double          y;
    vv = (interval_upper(a) + interval_lower(a)) / 2;
    ve = (interval_upper(a) - interval_lower(a)) / 2;
    if (ve < befloat_error(b))
	ve = befloat_error(b);
    b = befloat_value(b);
    y = floatitd(b);
    vv = vv + y;
    return make_efloat(vv, ve);
#else

    vv = (interval_upper(a) + interval_lower(a)) / 2;
    ve = (interval_upper(a) - interval_lower(a)) / 2;
    if (ve < befloat_error(b))
	ve = befloat_error(b);
    stackcheck();
    push(b);
    ev = bfloatitf(vv);
    pop(b);
    ev = plus2dd(ev, befloat_value(b));
    return make_befloat(ev, ve);
#endif
}



Sex
plus2cE(Sex a, Sex b)
{
    Sex             er,
                    ei;
    double          ee;
#if TOWARD_LOWER_PRECISION
    double          xr,
                    xi,
                    y;
    if (is_shortfloat(a)) {
	er = complex_real(a);
	ei = complex_imag(a);
	xr = floatit(er);
	xi = floatit(ei);
	y = floatitd(befloat_value(b));
	ee = befloat_error(b);
	return make_ecomplex(xr + y, xi, ee);
    }
#endif
    ee = befloat_error(b);
    stackcheck();
    push2(a, b);
    er = bfloatit(complex_real(a));
    pop(b);
    er = plus2dd(er, befloat_value(b));
    pop(a);
    push(er);
    ei = bfloatit(complex_imag(a));
    pop(er);
    return make_becomplex(er, ei, ee);
}



Sex
plus2zE(Sex a, Sex b)
{
    double          er,
                    ei;
    double          ee;
#if TOWARD_LOWER_PRECISION
    double          r,
                    i;
    er = ecomplex_real(a);
    ei = ecomplex_imag(a);
    ee = ecomplex_error(a);
    if (ee < befloat_error(b))
	ee = befloat_error(b);
    r = floatitd(befloat_value(b));
    er = er + r;
    return make_ecomplex(er, ei, ee);
#else
    Sex             r,
                    i;

    stackcheck();
    er = ecomplex_real(a);
    ei = ecomplex_imag(a);
    ee = ecomplex_error(a);
    if (ee < befloat_error(b))
	ee = befloat_error(b);
    push(b);
    r = bfloatitf(er);
    pop(b);
    r = plus2dd(r, befloat_value(b));
    push(r);
    i = bfloatitf(ei);
    pop(r);
    return make_becomplex(r, i, ee);
#endif
}

Sex
plus2EE(Sex a, Sex b)
{
    Sex             c;
    double          n;

    n = befloat_error(a);
    if (n < befloat_error(b))
	n = befloat_error(b);
    c = plus2dd(befloat_value(a), befloat_value(b));
    /*
     * lisp_printf("plus2EE: %p + %p = (%p, %g)\n", a, b, c, n);
     */
    return make_befloat(c, n);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
negateE(Sex a)
{
    double          e;
    Sex             v;
    e = befloat_error(a);
    v = befloat_value(a);
    v = negated(v);
    return make_befloat(v, e);
}

Sex
diff2fE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitE(b);
    pop(a);
    return diff2fe(a, b);
#else
    Sex             c;
    double          e;
    e = befloat_error(b);
    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    c = diff2dd(a, befloat_value(b));
    /*
     * lisp_printf("diff2xE: %p + %p = (%p, %g)\n", a, b, c, e);
     */
    return make_befloat(c, e);
#endif
}

Sex
diff2xE(Sex a, Sex b)
{
    Sex             c;
    double          e;
    e = befloat_error(b);
    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    c = diff2dd(a, befloat_value(b));
    /*
     * lisp_printf("diff2xE: %p + %p = (%p, %g)\n", a, b, c, e);
     */
    return make_befloat(c, e);
}

Sex
diff2Ef(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitE(a);
    pop(b);
    return diff2ef(a, b);
#else
    Sex             c;
    double          e;
    e = befloat_error(a);
    stackcheck();
    push(a);
    b = bfloatit(b);
    pop(a);
    c = diff2dd(befloat_value(a), b);
    /*
     * lisp_printf("diff2Ex: %p + %p = (%p, %g)\n", a, b, c, e);
     */
    return make_befloat(c, e);
#endif
}

Sex
diff2Ex(Sex a, Sex b)
{
    Sex             c;
    double          e;
    e = befloat_error(a);
    stackcheck();
    push(a);
    b = bfloatit(b);
    pop(a);
    c = diff2dd(befloat_value(a), b);
    /*
     * lisp_printf("diff2Ex: %p + %p = (%p, %g)\n", a, b, c, e);
     */
    return make_befloat(c, e);
}



Sex
diff2eE(Sex a, Sex b)
{
    Sex             ev;
    double          ee;
    double          e;

#if TOWARD_LOWER_PRECISION
    double          x,
                    y;
    x = efloat_value(a);
    ee = efloat_error(a);
    if (ee < befloat_error(b))
	ee = befloat_error(b);
    y = floatitd(befloat_value(b));
    x = x - y;
    return make_efloat(x, ee);
#else
    stackcheck();
    push(b);
    e = efloat_value(a);
    ee = efloat_error(a);
    if (ee < befloat_error(b))
	ee = befloat_error(b);
    ev = bfloatitf(e);
    pop(b);
    ev = diff2dd(ev, befloat_value(b));
    return make_befloat(ev, ee);
#endif
}



Sex
diff2Ee(Sex a, Sex b)
{
    Sex             ev;
    double          ee;
#if TOWARD_LOWER_PRECISION
    double          x,
                    y;
    y = efloat_value(b);
    ee = efloat_error(b);
    if (ee < befloat_error(a))
	ee = befloat_error(a);
    x = floatitd(befloat_value(a));
    x = x - y;
    return make_efloat(x, ee);
#else

    stackcheck();
    ee = efloat_error(b);
    if (ee < befloat_error(a))
	ee = befloat_error(a);
    push(a);
    ev = bfloatitf(efloat_value(b));
    pop(a);
    ev = diff2dd(befloat_value(a), ev);
    return make_befloat(ev, ee);
#endif
}



Sex
diff2vE(Sex a, Sex b)
{
    Sex             ev;
    double          vv,
                    ve;
    double          x;

    vv = (interval_upper(a) + interval_lower(a)) / 2;
    ve = (interval_upper(a) - interval_lower(a)) / 2;
    if (ve < befloat_error(b))
	ve = befloat_error(b);
#if TOWARD_LOWER_PRECISION
    x = floatitd(befloat_value(b));
    vv = vv - x;
    return make_efloat(vv, ve);
#else
    stackcheck();
    push(b);
    ev = bfloatitf(vv);
    pop(b);
    ev = diff2dd(ev, befloat_value(b));
    return make_befloat(ev, ve);
#endif
}


Sex
diff2Ev(Sex a, Sex b)
{
    Sex             ev;
    double          vv,
                    ve;
    double          x;

    vv = (interval_upper(b) + interval_lower(b)) / 2;
    ve = (interval_upper(b) - interval_lower(b)) / 2;
    if (ve < befloat_error(a))
	ve = befloat_error(a);
#if TOWARD_LOWER_PRECISION
    x = floatitd(befloat_value(a));
    x = x - vv;
    return make_efloat(x, ve);
#else
    stackcheck();
    push(a);
    ev = bfloatitf(vv);
    pop(a);
    ev = diff2dd(befloat_value(a), ev);
    return make_befloat(ev, ve);
#endif
}


Sex
diff2cE(Sex a, Sex b)
{
    Sex             er,
                    ei;
    double          ee;

#if TOWARD_LOWER_PRECISION
    double          xr,
                    xi,
                    y;
    if (is_shortfloat(a)) {
	er = complex_real(a);
	ei = complex_imag(a);
	/*
	 * in case of er or ei is interval number --> need fix 
	 */
	xr = floatit(er);
	xi = floatit(ei);
	y = floatitd(befloat_value(b));
	ee = befloat_error(b);
	return make_ecomplex(xr - y, xi, ee);
    }
#endif


    stackcheck();
    ee = befloat_error(b);
    push2(a, b);
    er = bfloatit(complex_real(a));
    pop(b);
    er = diff2dd(er, befloat_value(b));
    pop(a);
    push(er);
    ei = bfloatit(complex_imag(a));
    pop(er);
    return make_becomplex(er, ei, ee);

}



Sex
diff2Ec(Sex a, Sex b)
{
    Sex             er,
                    ei;
    double          ee;

#if TOWARD_LOWER_PRECISION
    double          xr,
                    xi,
                    y;
    if (is_shortfloat(b)) {
	er = complex_real(b);
	ei = complex_imag(b);
	/*
	 * in case of er or ei is interval number --> need fix 
	 */
	xr = floatit(er);
	xi = floatit(ei);
	y = floatitd(befloat_value(a));
	ee = befloat_error(a);
	return make_ecomplex(y - xr, -xi, ee);
    }
#endif


    ee = befloat_error(a);
    stackcheck();
    push2(a, b);
    er = bfloatit(complex_real(b));
    a = stack[-1];
    er = diff2dd(befloat_value(a), er);
    b = stack[0];
    push(er);
    ei = bfloatit(complex_imag(b));
    ei = negated(ei);
    pop(er);
    popv(2);
    return make_becomplex(er, ei, ee);

}



Sex
diff2zE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    double          ar,
                    ai,
                    y;
#else
    Sex             er,
                    ei;
#endif
    double          ee;
    double          n;

    ee = befloat_error(b);
    n = ecomplex_error(a);
    if (ee < n)
	ee = n;
#if TOWARD_LOWER_PRECISION
    ar = ecomplex_real(a);
    ai = ecomplex_imag(a);
    y = floatitd(befloat_value(b));
    return make_ecomplex(ar - y, ai, ee);
#else
    stackcheck();
    push2(a, b);
    er = bfloatitf(ecomplex_real(a));
    b = stack[0];
    er = diff2dd(er, befloat_value(b));
    a = stack[-1];
    push(er);
    ei = bfloatitf(ecomplex_imag(a));
    a = stack[-2];
    pop(er);
    popv(2);
    return make_becomplex(er, ei, ee);
#endif
}


Sex
diff2Ez(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    double          x,
                    yr,
                    yi;
#else
    Sex             er,
                    ei;
#endif
    double          ee;
    double          n;

    ee = befloat_error(a);
    n = ecomplex_error(a);
    if (ee < n)
	ee = n;
#if TOWARD_LOWER_PRECISION
    yr = ecomplex_real(b);
    yi = ecomplex_imag(b);
    x = floatitd(befloat_value(a));
    return make_ecomplex(x - yr, -yi, ee);
#else
    stackcheck();
    push2(a, b);
    er = bfloatitf(ecomplex_real(b));
    a = stack[-1];
    er = diff2dd(befloat_value(a), er);
    b = stack[0];
    push(er);
    ei = bfloatitf(-ecomplex_imag(b));
    pop(er);
    popv(2);
    return make_becomplex(er, ei, ee);
#endif
}


Sex
diff2EE(Sex a, Sex b)
{
    double          ea,
                    eb;
    Sex             c;
    ea = befloat_error(a);
    eb = befloat_error(b);
    if (ea < eb)
	ea = eb;
    c = diff2dd(befloat_value(a), befloat_value(b));
    /*
     * lisp_printf("diff2EE: %p - %p = (%p, %g)\n", a, b, c, ea);
     */
    return make_befloat(c, ea);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
times2fE(Sex a, Sex b)
{
    double          c;
    double          d;
#if TOWARD_LOWER_PRECISION
    double          x;
    c = befloat_error(b);
    d = float_val(a);
    c = fabs(c * d);
    x = floatitd(befloat_value(b));

    x = d * x;
    return make_efloat(x, c);
#else
    c = befloat_error(b);
    d = floatit(a);
    c = fabs(c * d);
    push(b);
    a = bfloatit(a);
    pop(b);

    a = times2dd(a, befloat_value(b));
    return make_befloat(a, c);
#endif
}

Sex
times2xE(Sex a, Sex b)
{
    double          c;
    double          d;
    c = befloat_error(b);
    d = floatit(a);
    c = fabs(c * d);
    push(b);
    a = bfloatit(a);
    pop(b);

    a = times2dd(a, befloat_value(b));
    return make_befloat(a, c);
}


Sex
times2eE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitE(b);
    pop(a);
    return times2ee(a, b);
#else
    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    return times2EE(a, b);
#endif
}


Sex
times2vE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitv(a);
    pop(b);
    push(a);
    b = efloatitE(b);
    pop(a);
    return times2ee(a, b);
#else
    stackcheck();
    push(b);
    a = befloatitv(a);
    pop(b);
    return times2EE(a, b);
#endif
}

#if 0
Sex
times2Ec(Sex a, Sex b)
{
    return times2cE(b, a);
}
#endif


Sex
times2cE(Sex a, Sex b)
{
    stackcheck();
#if TOWARD_LOWER_PRECISION
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	pop(b);
	return times2zE(a, b);
    }
#endif
    push(b);
    a = becomplexitc(a);
    pop(b);
    return times2EZ(b, a);
}



Sex
times2Ez(Sex a, Sex b)
{
    return times2zE(b, a);
}

Sex
times2zE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitE(b);
    pop(a);
    return times2ez(b, a);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return times2EZ(b, a);
#endif
}


Sex
times2EE(Sex a, Sex b)
{
    Sex             c;
    double          ea,
                    eb;
    double          da,
                    db;
    ea = befloat_error(a);
    eb = befloat_error(b);
    da = floatitd(befloat_value(a));
    db = floatitd(befloat_value(b));
    ea = fabs(ea * db);
    eb = fabs(eb * da);
    if (ea < eb)
	ea = eb;
    c = times2dd(befloat_value(a), befloat_value(b));
    return make_befloat(c, ea);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
quot2fE(Sex a, Sex b)
{
    Sex             c;
    double          e;
    double          d;

#if TOWARD_LOWER_PRECISION
    double          x;

    e = befloat_error(b);
    d = floatitd(befloat_value(b));
    e = e / d;
    x = float_val(a);
    x = x / d;
    e = fabs(e * d);
    return make_efloat(x, e);
#else
    e = befloat_error(b);
    d = floatit(befloat_value(b));
    e = e / d;
    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    c = quot2dd(a, befloat_value(b));
    d = floatit(c);
    e = fabs(e * d);
    return make_befloat(c, e);
#endif
}


Sex
quot2Ef(Sex a, Sex b)
{
    Sex             c;
    double          e;
    double          d;

#if TOWARD_LOWER_PRECISION
    double          x;
    e = befloat_error(a);
    d = float_val(b);
    e = fabs(e / d);
    x = floatitd(befloat_value(a));
    x = x / d;
    return make_efloat(x, e);
#else
    e = befloat_error(a);
    d = floatit(b);
    e = fabs(e / d);
    stackcheck();
    push(a);
    b = bfloatit(b);
    pop(a);
    c = quot2dd(befloat_value(a), b);
    return make_befloat(c, e);
#endif
}




Sex
quot2xE(Sex a, Sex b)
{
    Sex             c;
    double          e;
    double          d;

    e = befloat_error(b);
    d = floatit(befloat_value(b));
    e = e / d;
    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    c = quot2dd(a, befloat_value(b));
    d = floatit(c);
    e = fabs(e * d);
    return make_befloat(c, e);
}


Sex
quot2Ex(Sex a, Sex b)
{
    Sex             c;
    double          e;
    double          d;

    e = befloat_error(a);
    d = floatit(b);
    e = fabs(e / d);
    stackcheck();
    push(a);
    b = bfloatit(b);
    pop(a);
    c = quot2dd(befloat_value(a), b);
    return make_befloat(c, e);
}


Sex
quot2eE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitE(b);
    pop(a);
    return quot2ee(a, b);
#else
    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    return quot2EE(a, b);
#endif
}


Sex
quot2Ee(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitE(a);
    pop(b);
    return quot2ee(a, b);
#else
    stackcheck();
    push(a);
    b = befloatite(b);
    pop(a);
    return quot2EE(a, b);
#endif
}


Sex
quot2vE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitv(a);
    pop(b);
    push(a);
    b = efloatitE(b);
    pop(a);
    return quot2ee(a, b);
#else
    stackcheck();
    push(b);
    a = befloatitv(a);
    pop(b);
    return quot2EE(a, b);
#endif
}


Sex
quot2Ev(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitv(b);
    pop(a);
    push(b);
    a = efloatitE(a);
    pop(b);
    return quot2ee(a, b);
#else
    stackcheck();
    push(a);
    b = befloatitv(b);
    pop(a);
    return quot2EE(a, b);
#endif
}



Sex
quot2cE(Sex a, Sex b)
{
    stackcheck();
#if TOWARD_LOWER_PRECISION
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	pop(b);
	return quot2zE(a, b);
    }
#endif
    push(b);
    a = becomplexitc(a);
    pop(b);
    return quot2ZE(a, b);
}




Sex
quot2Ec(Sex a, Sex b)
{
    stackcheck();
#if TOWARD_LOWER_PRECISION
    if (is_shortfloat(b)) {
	push(a);
	b = ecomplexitc(b);
	pop(a);
	return quot2Ez(a, b);
    }
#endif
    push(a);
    b = becomplexitc(b);
    pop(a);
    return quot2EZ(a, b);
}




Sex
quot2zE(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitE(b);
    pop(a);
    return quot2ze(a, b);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return quot2ZE(a, b);
#endif
}


Sex
quot2Ez(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitE(a);
    pop(b);
    return quot2ez(a, b);
#else
    stackcheck();
    push(a);
    b = becomplexitz(b);
    pop(a);
    return quot2EZ(a, b);
#endif
}



Sex
quot2EE(Sex a, Sex b)
{
    Sex             v;
    double          ea,
                    eb;
    double          da,
                    db;

    ea = befloat_error(a);
    eb = befloat_error(b);
    da = floatit(befloat_value(a));
    db = floatit(befloat_value(b));
    ea = fabs(ea / db);
    eb = fabs(eb / db * da / db);
    v = quot2dd(befloat_value(a), befloat_value(b));
    if (eb > ea)
	ea = eb;
    return make_befloat(v, ea);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
numeq2EE(Sex a, Sex b)
{

    if (befloat_error(a) != befloat_error(b))
	return NO;
    return numeq2dd(befloat_value(a), befloat_value(b));
}

bool
numeq2EV(Sex a, Sex b)
{
    return zerop(diff2EV(a, b));
}

bool
numeq2bE(Sex a, Sex b)
{
    return numeq2bd(a, befloat_value(b));
}

bool
numeq2dE(Sex a, Sex b)
{
    return numeq2dd(a, befloat_value(b));
}

bool
numeq2eE(Sex a, Sex b)
{
    return zerop(diff2eE(a, b));
}

bool
numeq2fE(Sex a, Sex b)
{
    return numeq2fd(a, befloat_value(b));
}

bool
numeq2iE(Sex a, Sex b)
{
    return numeq2id(a, befloat_value(b));
}

bool
numeq2rE(Sex a, Sex b)
{
    return numeq2rd(a, befloat_value(b));
}

bool
numeq2vE(Sex a, Sex b)
{
    return zerop(diff2vE(a, b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


	/*
	 * lessp 
	 */

bool
lessp2bE(Sex a, Sex b)
{
    a = diff2bE(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2rE(Sex a, Sex b)
{
    a = diff2rE(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2EE(Sex a, Sex b)
{
    a = diff2EE(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2Eb(Sex a, Sex b)
{
    a = diff2Eb(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2Ed(Sex a, Sex b)
{
    a = diff2Ed(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2Ee(Sex a, Sex b)
{
    a = diff2Ee(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2Ef(Sex a, Sex b)
{
    a = diff2Ef(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2Ei(Sex a, Sex b)
{
    a = diff2Ei(a, b);
    return minusp(a) ? YES : NO;
}


bool
lessp2Er(Sex a, Sex b)
{
    a = diff2Er(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2Ev(Sex a, Sex b)
{
    a = diff2Ev(a, b);
    return minusp(a) ? YES : NO;
}



bool
lessp2dE(Sex a, Sex b)
{
    a = diff2dE(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2eE(Sex a, Sex b)
{
    a = diff2eE(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2fE(Sex a, Sex b)
{
    a = diff2fE(a, b);
    return minusp(a) ? YES : NO;
}

bool
lessp2iE(Sex a, Sex b)
{
    a = diff2iE(a, b);
    return minusp(a) ? YES : NO;
}



bool
lessp2vE(Sex a, Sex b)
{
    a = diff2vE(a, b);
    return minusp(a) ? YES : NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
appeq2EE(Sex a, Sex b)
{
    return zerop(diff2EE(a, b));
}
bool
appeq2EV(Sex a, Sex b)
{
    return zerop(diff2EV(a, b));
}
bool
appeq2bE(Sex a, Sex b)
{
    return zerop(diff2bE(a, b));
}
bool
appeq2dE(Sex a, Sex b)
{
    return zerop(diff2dE(a, b));
}
bool
appeq2eE(Sex a, Sex b)
{
    return zerop(diff2eE(a, b));
}
bool
appeq2fE(Sex a, Sex b)
{
    return zerop(diff2fE(a, b));
}
bool
appeq2iE(Sex a, Sex b)
{
    return zerop(diff2iE(a, b));
}
bool
appeq2rE(Sex a, Sex b)
{
    return zerop(diff2rE(a, b));
}
bool
appeq2vE(Sex a, Sex b)
{
    return zerop(diff2vE(a, b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
