/*
 * bfloat.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/bfloat.c,v 2.18 2016/06/06 08:21:21 kako Exp $
 * 
 */

static char const rcsid[] =
    "$Id: bfloat.c,v 2.18 2016/06/06 08:21:21 kako Exp $";


#include<stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>


#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"

#include "bfloat.h"

#define MPFR_DEBUG	0

/***************************************************************/
/***************************************************************/
/***************************************************************/

double
power_of_tenth(long n)
{
    double          x = 0.1;
    double          y = 1;

    while (n > 0) {
	if (n & 1)
	    y = y * x;
	x = x * x;
	n = n >> 1;
    }
    return y;
}

Sex
Lprecision(Sex a)
{
    long            p,
                    n;

    p = current_d_precision;
    if (is_fixnum(a)) {
	n = mkint(a);
	if (n <= 0)
	    n = 16;
	// printf("n=%ld\n", n);
	current_d_precision = n;
	current_w_precision = (n * 7 + 127) / (LPSIZE * 2);	/* 
								 * == * 3.32129 
								 */
	current_b_precision = current_w_precision * LPSIZE;
	current_precision = power_of_tenth(n);
	// printf("current_precision = %lg\n", current_precision);
    }
    Lprint_precision(a);
    return mkfxat(p);
}


Sex
fixitd(Sex a)
{
    Sex             c;
    long            i,
                    n,
                    m;
    long            l;
    unsigned long  *da,
                   *dc;

    l = bfloat_exp(a);
    a = bfloat_man(a);
    if (l == 0) {
	l = bnum_n(numlen(a)) - 1;
	if (bignum_value(a)[l] == 0)
	    return mkfxat(0);
	return a;
    }
    stackcheck();
    if (l > 0) {
	n = l >> LP_LOG2;
	m = l & (LPSIZE - 1);
	l = bnum_n(numlen(a)) - 1;

	if (m == 0) {
	    push(a);
	    c = new_binteger(T_BIGNUM | (numhdr(a) & T_NEG),
			     bnum_len(l + n + 1));
	    pop(a);
	    da = &bignum_value(a)[0];
	    dc = &bignum_value(c)[0];
	    for (i = 0; i < n; i++)
		dc[i] = 0;
	    for (i = 0; i <= l; i++) {
		dc[i + n] = da[i];
	    }
	} else {
	    push(a);
	    c = new_binteger(T_BIGNUM | (numhdr(a) & T_NEG),
			     bnum_len(l + n + 2));
	    pop(a);
	    da = &bignum_value(a)[0];
	    dc = &bignum_value(c)[0];
	    for (i = 0; i < n; i++)
		dc[i] = 0;
	    dc[n] = da[0] << m;
	    for (i = 1; i <= l; i++) {
		dc[i + n] = (da[i] << m) | (da[i - 1] >> (LPSIZE - m));
	    }
	    dc[l + n + 1] = da[l] >> (LPSIZE - m);
	}
	return c;

    } else {
	l = -l;
	n = l >> LP_LOG2;
	m = l & (LPSIZE - 1);
	l = bnum_n(numlen(a)) - 1;
	if (n > l)
	    return mkfxat(0);
	if (n == l) {
	    if (m < 4) {
		unsigned long   a0;
		a0 = bignum_value(a)[l];
		if (m)
		    a0 = a0 >> m;
		a = new_binteger(T_BIGNUM | (numhdr(a) & T_NEG),
				 bnum_len(1));
		bignum_value(a)[0] = a0;
		return a;
	    } else {
		unsigned long   a0;
		long            t;
		a0 = bignum_value(a)[l];
		n = a0 >> m;
		if (numhdr(a) & T_NEG) {
		    n = -n;
		    t = n & fix_mask;
		    if (t == 0 || t == fix_mask)
			return mkfxat(n);
		    return mkbigat1s(-1, -n);
		} else {
		    t = n & fix_mask;
		    if (t == 0 || t == fix_mask)
			return mkfxat(n);
		    return mkbigat1s(1, n);
		}
	    }
	}
	push(a);
	c = new_binteger(T_BIGNUM | (numhdr(a) & T_NEG),
			 bnum_len(l - n + 1));
	pop(a);
	da = &bignum_value(a)[0];
	dc = &bignum_value(c)[0];
	if (m == 0) {
	    for (i = 0; i <= l - n; i++) {
		dc[i] = da[i + l];
	    }
	} else {
	    for (i = n; i <= l; i++) {
		dc[i - n] = (da[i] >> m) | (da[i + 1] << (LPSIZE - m));
	    }
	    dc[l - n] = da[l] >> m;
	}
	return c;

    }
}

Sex
bfloatiti(Sex a)
{
    Sex             c;
    long            i,
                    n;
    long            l;
    unsigned long  *da;

    ASSERT(is_fixnum(a));

    stackcheck();
    n = mkint(a);
    if (n == 0) {
	a = new_bignum(bnum_len(1));
	bignum_value(a)[0] = 0;
	push(a);
	c = new_bfloat();
	pop(a);
	bfloat_man(c) = a;
	bfloat_exp(c) = 0;
	return c;
    } else if (n < 0) {
	a = new_bigneg(bnum_len(current_w_precision));
	n = -n;
    } else {
	a = new_bignum(bnum_len(current_w_precision));
    }
    da = &bignum_value(a)[0];

    for (i = 0; i < current_w_precision; i++)
	da[i] = 0;
    l = msd(n);
    da[current_w_precision - 1] = n << (LPSIZE - l);
    push(a);
    c = new_bfloat();
    pop(a);
    bfloat_man(c) = a;
    bfloat_exp(c) = l - current_b_precision;
    return c;
}

Sex
bfloatitb(Sex a)
{
    Sex             b;
    long            i,
                    j;
    long            exp;
    long            l;
    unsigned long   last,
                    top;

    unsigned long  *da;
    unsigned long  *db;

    ASSERT(is_bigint(a));
    stackcheck();
    j = bnum_n(numlen(a));
    l = msd(bignum_value(a)[j - 1]);
    j = j - current_w_precision;
    exp = (j * LPSIZE) - (LPSIZE - l);
    push(a);
    b = new_binteger(T_BIGNUM | (numhdr(a) & T_NEG),
		     bnum_len(current_w_precision));
    pop(a);

    da = &bignum_value(a)[0];
    db = &bignum_value(b)[0];

    if (j > 0) {
	last = da[j - 1];
	i = 0;
    } else if (j == 0) {
	last = 0;
	i = 0;
    } else {
	i = -j;
	last = 0;
	for (j = 0; j < i; j++)
	    db[j] = 0;
	j = 0;
    }
    if (l == 0) {
	while (i < current_w_precision) {
	    top = da[j++];
	    db[i++] = last;
	    last = top;
	}
    } else if (l == LPSIZE) {
	while (i < current_w_precision) {
	    db[i++] = da[j++];
	}
    } else {
	while (i < current_w_precision) {
	    top = da[j++];
	    db[i++] = (top << (LPSIZE - l)) | (last >> l);
	    last = top;
	}
    }

    push(b);
    a = new_bfloat();
    pop(b);
    bfloat_man(a) = b;
    bfloat_exp(a) = exp;
    return a;
}

Sex
bfloatitr(Sex a)
{
    Sex             n,
                    d;
    ASSERT(is_rational(a));

    stackcheck();
    push(rational_den(a));
    n = bfloatit(rational_num(a));
    pop(d);
    push(n);
    d = bfloatit(d);
    pop(n);
    return quot2dd(n, d);
}

Sex
bfloatitf(double x)		// ??? 
{
    Sex             a,
                    c;
    DoubleFloat     u;
    long            exp;
    long            i;
    unsigned long   u1,
                    u2;

    u.d = x;
    if (x == 0.0) {
	return bfloatiti(mkfxat(0));
    } else {
	exp = ieee_exponent(u);
#ifdef _LP64
	u2 = ieee_word(u);
#else
	u1 = ieee_loword(u);
	u2 = ieee_hiword(u);
#endif
	exp = exp - 52 + 64 - current_b_precision - (32 - 21);
    }

    stackcheck();
    a = new_binteger(ieee_signum(u) < 0 ? T_BIGNEG : T_BIGNUM,
		     bnum_len(current_w_precision));
    for (i = 0; i < current_w_precision; i++)
	bignum_value(a)[i] = 0;
    /*
     * l = msd(u2);
     */
#ifdef _LP64
    bignum_value(a)[current_w_precision - 1] = u2 << (32 - 21);
#else
    bignum_value(a)[current_w_precision - 2] = u1 << (32 - 21);
    bignum_value(a)[current_w_precision - 1] =
	(u2 << (32 - 21)) | (u1 >> 21);
#endif
    push(a);
    c = new_bfloat();
    pop(a);
    bfloat_man(c) = a;
    bfloat_exp(c) = ieee_signum(u) ? exp : 0;
    return c;
}


Sex
bfloatitc(Sex a)
{
    Sex             c;

    ASSERT(is_complex(a));
    c = complex_imag(a);
    a = complex_real(a);
#if !CONVERT_FROM_COMPLEX
    if (zerop(c))
	return bfloatit(a);
    return aerror1("convert complex to bfloat", a);
#else
    stackcheck();
    push2(a, c);
    if (zerop(c)) {
	pop2(c, a);
	return bfloatit(a);
    }
    stack[-1] = bfloatit(stack[-1]);
    stack[0] = bfloatit(stack[0]);
    pop2(c, a);
    return Lmkcomplex(a, c);
#endif
}

Sex
bfloatitz(Sex a)
{
    Sex             c;
    double          fr,
                    fi;
    ASSERT(is_ecomplex(a));

    fi = ecomplex_imag(a);
    fr = ecomplex_real(a);
    c = bfloatitf(fr);
    if (fi == 0)
	return c;
#if !CONVERT_FROM_COMPLEX
    return aerror1("convert ecomplex to bfloat", a);
#else
    stackcheck();
    push(c);
    a = bfloatitf(fi);
    pop(c);
    return Lmkcomplex(c, a);
#endif
}


Sex
bfloatitZ(Sex a)
{
    Sex             c;
    ASSERT(is_becomplex(a));

    c = becomplex_imag(a);
    a = becomplex_real(a);
#if !CONVERT_FROM_COMPLEX
    if (zerop(c))
	return a;
    return aerror1("convert becomplex to bfloat", a);
#else
    stackcheck();
    push2(a, c);
    if (zerop(c)) {
	pop2(c, a);
	return a;
    }
    pop2(c, a);
    return Lmkcomplex(a, c);
#endif
}


Sex
bfloatit(Sex a)
{

    if (is_fixnum(a)) {
	return bfloatiti(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return bfloatitb(a);
	case N_FLOAT:
	    return bfloatitf(float_val(a));
	case N_EFLOAT:
	    return bfloatitf(efloat_value(a));
	case N_INTERVAL:
	    return bfloatitf((interval_lower(a) + interval_upper(a)) *
			     0.5);
	case N_RATIONAL:
	    return bfloatitr(a);
	case N_BFLOAT:
	    return a;
	case N_BEFLOAT:
	    return befloat_value(a);
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    return leftshiftd(a, -1);
	case N_COMPLEX:
	    return bfloatitc(a);
	case N_ECOMPLEX:
	    return bfloatitz(a);
	case N_BECOMPLEX:
	    return bfloatitZ(a);
	}
    }
    return error1("bad argument to bfloat ", a);
}



Sex
Lbfloat(Sex a)
{

    if (is_fixnum(a)) {
	return bfloatiti(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return bfloatitb(a);
	case N_FLOAT:
	    return bfloatitf(float_val(a));
	case N_EFLOAT:
	    return bfloatitf(efloat_value(a));
	case N_INTERVAL:
	    return bfloatitf((interval_lower(a) + interval_upper(a)) *
			     0.5);
	case N_RATIONAL:
	    return bfloatitr(a);
	case N_ECOMPLEX:
	    return bfloatitz(a);
	case N_COMPLEX:
	    return bfloatitc(a);
	case N_BFLOAT:
	    return a;
	case N_BEFLOAT:
	    return befloat_value(a);
	case N_BECOMPLEX:
	    return bfloatitZ(a);
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    return leftshiftd(a, -1);
	}
    }
    return error1("bad argument to bfloat ", a);
}


Sex
make_bfloat(Sex a, long b)
{
    if (a == mkfxat(0))
	return bfloatiti(a);
    if (is_fixnum(a))
	a = bfloatiti(a);
    else {
	ASSERT(is_bigint(a));
	a = bfloatitb(a);
    }
    bfloat_exp(a) = bfloat_exp(a) + b;
    return a;
}

Sex
leftshiftd(Sex a, long n)
{
    Sex             c;
    if (zeropd(a))
	return a;
    push(a);
    c = new_bfloat();
    pop(a);
    bfloat_man(c) = bfloat_man(a);
    bfloat_exp(c) = bfloat_exp(a) + n;
    return c;
}

#define _log_2    0.693147180559945309417
#define _log_10   2.30258509299404568402
#define _log_2_10 (_log_2/_log_10)
/*
 * #define app_log10(x) (1.003+x*2.31)
 */
#define app_log10(x) (1.00+x*(2.3064+2.654*x))

void
print_expt10(long n, putchar_fn * fn)
{
    double          d;
    char           *p;
    char            my_buff[64];

    if (n < -current_b_precision - 64) {
	fn('0');
	return;
    }
    d = n * _log_2_10;
    n = d;
    d = d - n;
    if (d < 0) {
	n--;
	d += 1.0;
    }
    if (d > 0.5) {
	d = d - 0.5;
	d = app_log10(d) * 3.16228;
    } else {
	d = app_log10(d);
    }

#if 1
    if (n > 0)
	sprintf(my_buff, "%.3fe+%ld", d, n);
    else if (n < 0)
	sprintf(my_buff, "%.3fe-%ld", d, -n);
    else
	sprintf(my_buff, "%.3f", d);
#else
    if (n > 0)
	sprintf(my_buff, "%fe+%ld", d, n);
    else if (n < 0)
	sprintf(my_buff, "%fe-%ld", d, -n);
    else
	sprintf(my_buff, "%f", d);
#endif

    for (p = my_buff; *p; p++)
	fn(*p);
}

void
print_float(double d, putchar_fn * fn)
{
    char           *p;
    char            my_buff[64];

    sprintf(my_buff, "%f", d);
    for (p = my_buff; *p; p++)
	fn(*p);
}


void
print_int(long dig, putchar_fn * fn)
{
    long            i;
    char            my_buff[20];


    if (dig == 0) {
	fn('0');
	return;
    }
    if (dig < 0) {
	dig = -dig;
	fn('-');
    }
    for (i = 0; dig != 0; i++) {
	my_buff[i] = (long) (dig % 10) + '0';
	dig = dig / 10;
    }
    while (i > 0)
	fn(my_buff[--i]);
}

void
print_uint(unsigned long dig, putchar_fn * fn)
{
    long            i;
    char            my_buff[20];


    if (dig == 0) {
	fn('0');
	return;
    }
    for (i = 0; dig != 0; i++) {
	my_buff[i] = (long) (dig % 10) + '0';
	dig = dig / 10;
    }
    while (i > 0)
	fn(my_buff[--i]);
}


void
print_int_with_exp(unsigned long u, long dexp, putchar_fn * fn)
{

    unsigned long   v;

    long            i,
                    nxt;
    char            my_buff[24];
    long            ndig;
    long            mdig;


    i = 0;
    if (u >= M10_PRINT_EXP) {
	u = udiv(&v, 0, u, M10_PRINT_EXP);
	while (i < M10_PRINT_DIG) {
	    nxt = v % 10;
	    v = v / 10;
	    my_buff[i++] = nxt + '0';
	}
    }
    while (u != 0) {
	nxt = u % 10;
	u = u / 10;
	my_buff[i++] = nxt + '0';
    }
    ndig = i;

    if (dexp < 0 && (dexp + ndig) > 0) {
	dexp = dexp + ndig;
	mdig = -ndig;
	while (i > 0) {
	    fn(my_buff[--i]);
	    mdig++;
	    if (--dexp == 0) {
		fn('.');
		mdig = 0;
	    }
	    if (mdig >= print_bfloat_precision)
		break;
	}
	dexp = 0;
    } else {
	fn(my_buff[--i]);
	fn('.');
	mdig = 0;
	dexp = dexp + ndig - 1;
	while (i > 0) {
	    fn(my_buff[--i]);
	    mdig++;
	    if (mdig >= print_bfloat_precision)
		break;
	}
    }
    if (dexp != 0) {
	fn('e');
	print_int(dexp, fn);
    }
}


void
print_bignum_with_exp(Sex u, long dexp, putchar_fn * fn)
{
    long            len;
    long            i,
                    len1,
                    len2;
    Sex             w;
    bool            sign = NO;
    char            my_buff[20];
    unsigned long   dig;
    long            ndig;
    long            mdig;

    if (is_fixnum(u)) {
	i = mkint(u);
	if (i >= 0) {
	    dig = i;
	    sign = NO;
	} else if (i < 0) {
	    sign = YES;
	    dig = -i;
	} else {
	    dexp = 0;
	    dig = 0;
	    sign = NO;
	}
	len = bnum_len(1);
    } else {
	if (type_of_header(numhdr(u)) == T_BIGNEG)
	    sign = YES;
	len = numlen(u);
	dig = bignum_value(u)[0];
    }
    if (len == bnum_len(1)) {
	if (sign)
	    fn('-');
	print_int_with_exp(dig, dexp, fn);
	return;
    }
    stackcheck();
    push(u);
    if (sign)
	fn('-');
    len1 = CELL + (11 * len) / 10;
    len1 = (long) alignit(len1);
    w = new_bignum(len1);
    pop(u);

    len = bnum_n(len);
    len1 = bnum_n(len1);
    for (i = 0; i < len; i++)
	bignum_value(w)[i] = bignum_value(u)[i];
    len2 = len1;

    while (len > 0) {
	unsigned long   carry;
	long            k;
	carry = 0;
	for (k = len - 1; k >= 0; k--) {
	    bignum_value(w)[k] = udiv(&carry,
				      carry, bignum_value(w)[k],
				      M10_PRINT_EXP);
	}
	if (bignum_value(w)[len - 1] == 0)
	    len--;
	bignum_value(w)[--len2] = carry;
    }
    push(w);
    if (bignum_value(w)[len2] == 0)
	len2++;
    dig = bignum_value(w)[len2++];
    i = 0;
    do {
	long            nxt = dig % 10;
	dig = dig / 10;
	my_buff[i++] = (long) nxt + '0';
    } while (dig != 0);
    ndig = i + (len1 - len2) * M10_PRINT_DIG;
    if (dexp < 0 && (dexp + ndig) > 0) {
	dexp = dexp + ndig;
	mdig = -ndig;
	while (i > 0) {
	    fn(my_buff[--i]);
	    mdig++;
	    if (--dexp == 0) {
		fn('.');
		mdig = 0;
	    }
	    if (mdig >= print_bfloat_precision)
		break;
	}
	pop(w);
	while (len2 < len1) {
	    dig = bignum_value(w)[len2++];
	    push(w);
	    for (i = M10_PRINT_DIG - 1; i >= 0; i--) {
		long            nxt = dig % 10;
		dig = dig / 10;
		my_buff[i] = (long) nxt + '0';
	    }
	    for (i = 0; i < M10_PRINT_DIG; i++) {
		fn(my_buff[i]);
		mdig++;
		if (--dexp == 0) {
		    fn('.');
		    mdig = 0;
		}
		if (mdig >= print_bfloat_precision)
		    break;
	    }
	    pop(w);
	    if (mdig >= print_bfloat_precision)
		break;
	}
	dexp = 0;
    } else {
	fn(my_buff[--i]);
	fn('.');
	mdig = 0;
	dexp = dexp + ndig - 1;
	while (i > 0) {
	    fn(my_buff[--i]);
	    mdig++;
	    if (mdig >= print_bfloat_precision)
		break;
	}
	pop(w);
	while (len2 < len1) {
	    dig = bignum_value(w)[len2++];
	    push(w);
	    for (i = M10_PRINT_DIG - 1; i >= 0; i--) {
		long            nxt = dig % 10;
		dig = dig / 10;
		my_buff[i] = (long) nxt + '0';
	    }
	    for (i = 0; i < M10_PRINT_DIG; i++) {
		fn(my_buff[i]);
		mdig++;
		if (mdig >= print_bfloat_precision)
		    break;
	    }
	    pop(w);
	    if (mdig >= print_bfloat_precision)
		break;
	}
    }
    if (dexp != 0) {
	fn('e');
	print_int(dexp, fn);
    }
}

#ifdef _LP64
#define LOG_10_2 0x4d104d427de7fbcc
#else
#define LOG_10_2 0x4d104d42
#endif

void
print_bfloat(Sex u, putchar_fn * fn)
{
    long            exp,
                    dexp,
                    len;

    unsigned long   h,
                    l;
    Sex             w;

    w = bfloat_man(u);
    dexp = bfloat_exp(u);
    len = bnum_n(numlen(w));
    if (bignum_value(w)[len - 1] == 0) {
	fn('0');
	fn('.');
	fn('0');
	return;
    }
#ifdef _LP64
    if (dexp == 9223372036854775745) {
	fn('0');
	fn('.');
	fn('0');
	return;
    }
#else
    if (dexp == 2147483585) {
	fn('0');
	fn('.');
	fn('0');
	return;
    }
#endif
    stackcheck();
    push(u);
    if (dexp > 0) {
	// printf("print: dexp = %ld\n", dexp);
	h = umul(&l, dexp, LOG_10_2);
	dexp = h;
	// printf("print: h = %ld\n", h);
#if 0
	w = expti(mkfxat(10), (long) h);
	w = is_fixnum(w) ? bfloatiti(w) : bfloatitb(w);
#else
	w = power_of_big_tenth((long) h);
#endif
	pop(u);
	u = quot2dd(u, w);
    } else if (dexp < 0) {
	h = umul(&l, -dexp, LOG_10_2);
	dexp = -(long) h;
	// printf("print: dexp = - %ld\n", h);
#if 0
	w = expti(mkfxat(10), (long) h);
	w = is_fixnum(w) ? bfloatiti(w) : bfloatitb(w);
#else
	w = power_of_big_tenth((long) h);
#endif
	pop(u);
	u = times2dd(u, w);
    } else {
	popv(1);
    }
    exp = bfloat_exp(u);
    u = bfloat_man(u);
    u = leftshift(u, exp);
    // printf("print2: dexp = %ld\n", dexp);
    print_bignum_with_exp(u, dexp, fn);
}

Sex
power_of_big_tenth(long n)
{
    long            m;
    Sex             w;
    m = current_d_precision;
    if (n > m) {
#ifdef USE_MPFR
	__mpfr_struct   r,
	                s;
	Sex             d;
	mpfr_prec_t     p;
	p = current_b_precision;
	mpfr_init2(&r, p);
	mpfr_init2(&s, p);
	mpfr_set_ui(&r, 10, GMP_RNDD);
	mpfr_pow_ui(&s, &r, n, GMP_RNDD);
	w = mpfr2bfloat(&s);
	mpfr_clear(&r);
	mpfr_clear(&s);
#else
#if 1
	stackcheck();
	push2(nil, nil);
#define a stack[-1]
#define b stack[0]
	a = bfloatiti(mkfxat(10));
	b = bfloatiti(mkfxat(1));

	while (n > 1) {
	    if (n & 1)
		b = times2dd(a, b);
	    a = times2dd(a, a);
	    n = n >> 1;
	}
	a = times2dd(a, b);
#undef a
#undef b
	w = stack[-1];
	popv(2);
#else
	extern Sex      Llog(Sex w);
	extern Sex      Lexp(Sex w);
	w = bfloatiti(mkfxat(10));
	w = Llog(w);
	w = times2(w, mkfxat(n));
	w = Lexp(w);
#endif
#endif
    } else {
	w = expti(mkfxat(10), n);
	w = is_fixnum(w) ? bfloatiti(w) : bfloatitb(w);
    }
    return w;
}

/***************************************************************/
/***************************************************************/
#ifdef USE_MPFR
void
print_mpfr(mpfr_t a)
{
    long            n,
                    i;
    long           *r;

    n = 4;
    r = (long *) a;
    for (i = 0; i < n; i++)
	printf("%016lx ", *r++);
    printf("\n");

    r = (long *) a;
    n = r[0] / 64;
    r = (long *) r[3];
    printf("value = ");
    for (i = 0; i < n; i++)
	printf("%016lx ", r[i]);
    printf("\n");
}
#endif
/***************************************************************/
/***************************************************************/
/***************************************************************/

/*
 * arithmetic routines 
 */


Sex
plus2rd(Sex a, Sex b)
{
    ASSERT(is_rational(a) && is_bfloat(b));
    stackcheck();
    push(b);
    a = bfloatitr(a);
    pop(b);
    return plus2dd(a, b);
}

Sex
plus2bd(Sex a, Sex b)
{
    ASSERT(is_bigint(a) && is_bfloat(b));
    stackcheck();
    push(b);
    a = bfloatitb(a);
    pop(b);
    return plus2dd(a, b);
}


Sex
plus2fd(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    ASSERT(is_float(a) && is_bfloat(b));
    x = float_val(a);
    y = floatitd(b);
    x = x + y;
    return make_float(x);
#else
    ASSERT(is_float(a) && is_bfloat(b));
    stackcheck();
    push(b);
    a = bfloatitf(float_val(a));
    pop(b);
    return plus2dd(a, b);
#endif
}

Sex
plus2id(Sex a, Sex b)
{
    ASSERT(is_fixnum(a) && is_bfloat(b));
    if (a == mkfxat(0))
	return b;
    stackcheck();
    push(b);
    a = bfloatiti(a);
    pop(b);
    return plus2dd(a, b);
}




Sex
leftshifti(Sex a, long s)
{
    long            m;
    long            sign;
    long            i;
    long            k;



    if (a == mkfxat(0))
	return a;
    sign = T_BIGNUM;
    i = mkint(a);
    if (i < 0) {
	sign = T_BIGNEG;
	i = -i;
    }

    m = CELL + SEX + ((s + LPSIZE - 1) & -LPSIZE) / 8;
    a = new_binteger(sign, m);

    m = bnum_n(m);
    for (k = 0; k < m; k++)
	bignum_value(a)[k] = 0;

    s = s & (LPSIZE - 1);
    if (s == 0) {
	bignum_value(a)[m - 1] = i;
	if (m != 1)
	    return a;

	if ((i & fix_mask) == 0)
	    return mkfxat(sign == T_BIGNEG ? -i : i);
	return a;
    } else {
	k = i >> (LPSIZE - s);
	bignum_value(a)[m - 1] = k;
	bignum_value(a)[m - 2] = i << s;

	if (k == 0) {
	    if (m == 2) {
		i = i << s;
		if ((i & fix_mask) == 0)
		    return mkfxat(sign == T_BIGNEG ? -i : i);
	    }
	    if ((m & 1) != 0) {
		bignum_value(a)[m - 1] = TAG_HDRS + T_STRING;
		bignum_value(a)[m] = bnum_len(0);
	    }
	    bignum_length(a) = bignum_length(a) - SEX;
	}
	return a;
    }
}

Sex
rightshifti(Sex a, long s)
{
    long            i;

    i = mkint(a);
    if (s >= LPSIZE)
	return mkfxat(0);
    if (i < 0) {
	i = -i;
	i = i >> s;
	i = -i;
    } else {
	i = i >> s;
    }
    return mkfxat(i);
}


Sex
leftshiftb(Sex a, long s)
{
    Sex             c;
    long            sign;
    long            n,
                    m;
    long            k;


    n = numlen(a);
    sign = numhdr(a) & 0xffff0;

    m = ((s + (LPSIZE - 1)) & -LPSIZE) >> 3;

    stackcheck();
    push(a);
    c = new_binteger(sign, n + m);
    pop(a);

    m = bnum_n(m) + 2;
    n = bnum_n(n);
    s = s & (LPSIZE - 1);
    for (k = 0; k < m; k++)
	bignum_value(c)[k] = 0;
    if (s == 0) {
	for (k = n - 1; k >= 0; k--) {
	    bignum_value(c)[k + m] = bignum_value(a)[k];
	}
    } else {
	bignum_value(c)[m - 1] = bignum_value(a)[0] << s;
	for (k = 0; k < n - 1; k++) {
	    bignum_value(c)[k + m] = (bignum_value(a)[k] >> (LPSIZE - s)) |
		(bignum_value(a)[k + 1] << s);
	}
	m = n + m - 1;
	bignum_value(c)[m] = bignum_value(a)[n - 1] >> (LPSIZE - s);
	if (bignum_value(c)[m] == 0) {
	    if ((m & 1) == 0) {
		bignum_value(c)[m] = TAG_HDRS + T_STRING;
		bignum_value(c)[m + 1] = bnum_len(0);
	    }
	    bignum_length(c) = bnum_len(m);
	}
    }
    return c;
}

Sex
rightshiftb(Sex a, long s)
{
    Sex             c;
    long            sign;
    long            n,
                    m;
    long            k;

    n = numlen(a);
    sign = numhdr(a) & 0xffff0;

    /*
     * if(bignum_log2(a) < s) return mkfxat(0); 
     */
    m = (s & -LPSIZE) >> 3;
    if (n - m <= CELL)
	return mkfxat(0);

    stackcheck();
    push(a);
    c = new_binteger(sign, n - m);
    pop(a);

    m = bnum_n(m) + 2;
    n = bnum_n(n);
    s = s & (LPSIZE - 1);
    if (s == 0) {
	k = n - m - 1;
	if (k == 0) {
	    k = bignum_value(a)[m];
	    if ((k & fix_mask) == 0)
		return mkfxat(sign == T_BIGNEG ? -k : k);
	    bignum_value(c)[0] = k;
	    return c;
	}
	for (; k >= 0; k--) {
	    bignum_value(c)[k] = bignum_value(a)[k + m];
	}
    } else {
	bignum_value(c)[n - m - 1] = bignum_value(a)[n - 1] >> s;
	for (k = n - m - 2; k >= 0; k--) {
	    bignum_value(c)[k] = (bignum_value(a)[k + m] >> s) |
		(bignum_value(a)[k + m + 1] << (LPSIZE - s));
	}
	m = n - m - 1;
	if (bignum_value(c)[m] == 0) {
	    if (m == 0)
		return mkfxat(0);
	    if (m == 1) {
		k = bignum_value(c)[0];
		if ((k & fix_mask) == 0)
		    return mkfxat(sign == T_BIGNEG ? -k : k);
	    }
	    if ((m & 1) == 0) {
		bignum_value(c)[m] = TAG_HDRS + T_STRING;
		bignum_value(c)[m + 1] = bnum_len(0);
	    }
	    bignum_length(c) = bnum_len(m);
	} else if (m == 0) {
	    k = bignum_value(c)[0];
	    if ((k & fix_mask) == 0)
		return mkfxat(sign == T_BIGNEG ? -k : k);
	}
    }
    return c;
}


Sex
leftshift(Sex a, long s)
{
    if (s == 0)
	return a;
    else if (s > 0) {
	if (is_fixnum(a))
	    return leftshifti(a, s);
	else
	    return leftshiftb(a, s);
    } else {
	if (is_fixnum(a))
	    return rightshifti(a, -s);
	else
	    return rightshiftb(a, -s);
    }

}

Sex
rightshift(Sex a, long s)
{
    if (s == 0)
	return a;
    else if (s > 0) {
	if (is_fixnum(a))
	    return rightshifti(a, s);
	else
	    return rightshiftb(a, s);
    } else {
	if (is_fixnum(a))
	    return leftshifti(a, -s);
	else
	    return leftshiftb(a, -s);
    }

}

Sex
Lleftshift(Sex a, Sex b)
{
    long            s;
    if (!is_fixnum(b))
	error("leftshift");
    s = mkint(b);
    if (s == 0)
	return a;
    else if (s > 0) {
	if (is_fixnum(a))
	    return leftshifti(a, s);
	else
	    return leftshiftb(a, s);
    } else {
	if (is_fixnum(a))
	    return rightshifti(a, -s);
	else
	    return rightshiftb(a, -s);
    }
}

Sex
extend_bignum(Sex a, long n)
{
    Sex             c;
    long            sign;
    long            k;
    long            m;
    unsigned long  *p,
                   *q;


    m = numlen(a);
    sign = numhdr(a) & 0xffff0;

    stackcheck();
    push(a);
    c = new_binteger(sign, m + n * sizeof(unsigned long));
    pop(a);

    m = bnum_n(m);
    p = &bignum_value(c)[0];
    q = &bignum_value(a)[0];
    m = n + m;
    for (k = 0; k < n; k++)
	*p++ = 0;
    for (; k < m; k++)
	*p++ = *q++;
    return c;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


#ifdef USE_MPFR


Sex
plus2dd(Sex a, Sex b)
{
    __mpfr_struct   op1,
                    op2,
                    rop;
    Sex             c;
    Sex             d,
                    e;
    long            n,
                    m,
                    s;

    mpfr_prec_t     p;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    d = bfloat_man(a);
    e = bfloat_man(b);

    // precision
    p = bignum_precision(d);

    n = bnum_n(numlen(d)) - 1;
    if (bignum_value(d)[n] == 0)
	return b;
    m = bnum_n(numlen(e)) - 1;
    if (bignum_value(e)[m] == 0)
	return a;

    mpfr_init2(&rop, p);

    op1 = bfloat2mpfr(a);
    op2 = bfloat2mpfr(b);

    mpfr_add(&rop, &op1, &op2, MPFR_RNDN);

    c = mpfr2bfloat(&rop);

    mpfr_clear(&rop);
    ASSERT(is_bfloat(c));
    return c;

}

Sex
plus2ddR(Sex a, Sex b, mpfr_rnd_t R)
{
    __mpfr_struct   op1,
                    op2,
                    rop;
    Sex             c;
    Sex             d,
                    e;
    long            n,
                    m,
                    s;

    mpfr_prec_t     p;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    d = bfloat_man(a);
    e = bfloat_man(b);

    // precision
    p = bignum_precision(d);

    n = bnum_n(numlen(d)) - 1;
    if (bignum_value(d)[n] == 0)
	return b;
    m = bnum_n(numlen(e)) - 1;
    if (bignum_value(e)[m] == 0)
	return a;

    mpfr_init2(&rop, p);

    op1 = bfloat2mpfr(a);
    op2 = bfloat2mpfr(b);

    mpfr_add(&rop, &op1, &op2, R);

    c = mpfr2bfloat(&rop);

    mpfr_clear(&rop);
    ASSERT(is_bfloat(c));
    return c;
}

#else
Sex
plus2dd(Sex a, Sex b)
{
    Sex             a_m,
                    b_m;
    long            a_e,
                    b_e;
    long            s;
    long            n,
                    m;


    ASSERT(is_bfloat(a) && is_bfloat(b));
    a_m = bfloat_man(a);
    a_e = bfloat_exp(a);
    b_m = bfloat_man(b);
    b_e = bfloat_exp(b);
    n = bnum_n(numlen(a_m)) - 1;
    if (bignum_value(a_m)[n] == 0)
	return b;
    m = bnum_n(numlen(b_m)) - 1;
    if (bignum_value(b_m)[m] == 0)
	return a;
    s = a_e - b_e;
    if (n == m) {
	if (s > current_b_precision + LPSIZE)
	    return a;
	else if (s < -current_b_precision - LPSIZE)
	    return b;
    }
    stackcheck();
    if (s == 0) {
	a = plus2bb(a_m, b_m);
	a = make_bfloat(a, a_e);
	ASSERT(is_bfloat(a));
	return a;
    } else if (s > 0) {
	push(b_m);
	a_m = leftshiftb(a_m, s);
	pop(b_m);
	a = plus2bb(a_m, b_m);
	a = make_bfloat(a, b_e);
	ASSERT(is_bfloat(a));
	return a;
    } else {
	push(a_m);
	b_m = leftshiftb(b_m, -s);
	pop(a_m);
	a = plus2bb(a_m, b_m);
	a = make_bfloat(a, a_e);
	ASSERT(is_bfloat(a));
	return a;
    }
}
#endif

/****  bfloat to mpfr ******************************************/

#ifdef USE_MPFR
#ifdef _LP64
#define MPFR_ZERO_EXP	0x8000000000000001
#else
#define MPFR_ZERO_EXP	0x80000001
#endif


__mpfr_struct
bfloat2mpfr(Sex a)
{
    Sex             b = bfloat_man(a);
    __mpfr_struct   m;
    long            n;

    m._mpfr_prec = bignum_precision(b);
    m._mpfr_sign = is_bigneg(b) ? -1 : 1;

    // printf("bfloat_exp: %d\n", bfloat_exp(a));
    n = bnum_n(numlen(b)) - 1;
    if (bignum_value(b)[n] == 0) {
	m._mpfr_exp = MPFR_ZERO_EXP;
    } else {
	m._mpfr_exp = bfloat_exp(a) + m._mpfr_prec;
    }
    m._mpfr_d = (mp_limb_t *) & bignum_value(b);

    // mpfr_printf("bfloat2mpfr: %.100Rf\n", &m);
    return m;
}


/****  mpfr to bfloat ******************************************/

Sex
mpfr2bfloat(__mpfr_struct * m)
{
    Sex             a;
    Sex             b;
    long            i;
    // int n = (m->_mpfr_prec*4)/32+8;
    long            n = (m->_mpfr_prec + (LPSIZE - 1)) / LPSIZE;

#if MPFR_DEBUG
    // mpfr_printf("mpfr2bfloat: %.100Rf\n", m);
    printf("n=%ld\n", n);

    {
	long            j;
	int            *p = (int *) m;
	j = sizeof(__mpfr_struct) / sizeof(int);
	for (i = 0; i < j; i++)
	    printf("%08x, ", *p++);
	printf("\n");
    }
    for (i = 0; i < n; i++)
	printf("%016lx, ", m->_mpfr_d[i]);
    printf("\n");
    printf("exp:%ld\n", m->_mpfr_exp);
    printf("prec:%ld\n", m->_mpfr_prec);
#endif

    if (m->_mpfr_exp == MPFR_ZERO_EXP || m->_mpfr_d[n - 1] == 0) {
	a = new_bignum(bnum_len(1));
	bignum_value(a)[0] = 0;
	push(a);
	b = new_bfloat();
	pop(a);
	bfloat_man(b) = a;
	bfloat_exp(b) = 0;
#if MPFR_DEBUG
	printf("return bigfloat zero\n");
#endif
	return b;
    }

    if (m->_mpfr_sign == -1) {
	b = new_bigneg(bnum_len(n));
    } else {
	b = new_bignum(bnum_len(n));
    }
    for (i = 0; i < n; i++) {
	bignum_value(b)[i] = m->_mpfr_d[i];
    }
    push(b);
    a = new_bfloat();
    pop(b);
    bfloat_man(a) = b;
    bfloat_exp(a) = m->_mpfr_exp - m->_mpfr_prec;
#if MPFR_DEBUG
    printf("(mpfr2bfloat)Sex a = %ld\n", a);
#endif
    return a;
}

#endif

/***************************************************************/
/***************************************************************/
/***************************************************************/

// #ifdef USE_MPFR
#if 0
Sex
negated(Sex a)
{
    __mpfr_struct   ma;
    __mpfr_struct   mb;

    ASSERT(is_bfloat(a));
    ma = bfloat2mpfr(a);
    mpfr_neg(&mb, &ma, MPFR_RNDN);
    a = mpfr2bfloat(&mb);
    return a;
}
#else

Sex
negated(Sex a)
{
    long            n;
    ASSERT(is_bfloat(a));
    n = bfloat_exp(a);
    a = bfloat_man(a);
    a = negateb(a);
    return make_bfloat(a, n);
}
#endif

Sex
diff2db(Sex a, Sex b)
{
    ASSERT(is_bfloat(a) && is_bigint(b));

    stackcheck();
    push(a);
    b = bfloatitb(b);
    pop(a);
    return diff2dd(a, b);
}

#ifdef USE_MPFR
Sex
diff2dd(Sex a, Sex b)
{
    __mpfr_struct   op1,
                    op2,
                    rop;
    Sex             c;
    Sex             d,
                    e;
    long            n,
                    m,
                    s;
    mpfr_prec_t     p;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    // 10ʿbitѴ
    d = bfloat_man(a);
    e = bfloat_man(b);
    p = bignum_precision(d);

    n = bnum_n(numlen(d)) - 1;
    if (bignum_value(d)[n] == 0)
	return negate(b);
    m = bnum_n(numlen(e)) - 1;
    if (bignum_value(e)[m] == 0)
	return a;


    mpfr_init2(&rop, p);

    op1 = bfloat2mpfr(a);
    op2 = bfloat2mpfr(b);
#if MPFR_DEBUG
    mpfr_printf("diff2dd: %.100Rf, ", &op1);
    mpfr_printf(" %.100Rf", &op2);
    mpfr_sub(&rop, &op1, &op2, GMP_RNDN);
    mpfr_printf(" -->  %.100Rf\n", &rop);
#endif
    mpfr_sub(&rop, &op1, &op2, GMP_RNDN);
    c = mpfr2bfloat(&rop);
    // printf("Sex c = %d\n",c);

#if MPFR_DEBUG
    {
	long            exp;
	push(c);
	exp = bfloat_exp(c);
	printf("bfloat_exp: %ld\n", exp);
	lisp_printf("bfloat_man: %p\n", bfloat_man(c));
	pop(c);
    }
#endif
    mpfr_clear(&rop);
    return c;

}
#else
Sex
diff2dd(Sex a, Sex b)
{
    Sex             a_m,
                    b_m;
    long            a_e,
                    b_e;
    long            s;
    long            n,
                    m;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    a_m = bfloat_man(a);
    a_e = bfloat_exp(a);
    n = bnum_n(numlen(a_m)) - 1;
    if (bignum_value(a_m)[n] == 0)
	return negate(b);
    b_m = bfloat_man(b);
    b_e = bfloat_exp(b);
    m = bnum_n(numlen(b_m)) - 1;
    if (bignum_value(b_m)[m] == 0)
	return a;
    s = a_e - b_e;
    if (n == m) {
	if (s > current_b_precision + LPSIZE)
	    return a;
	else if (s < -current_b_precision - LPSIZE)
	    return negate(b);
    }
    stackcheck();
    if (s == 0) {
	a = diff2bb(a_m, b_m);
	return make_bfloat(a, a_e);
    } else if (s > 0) {
	push(b_m);
	a_m = leftshiftb(a_m, s);
	pop(b_m);
	a = diff2bb(a_m, b_m);
	return make_bfloat(a, b_e);
    } else {
	push(a_m);
	b_m = leftshiftb(b_m, -s);
	pop(a_m);
	a = diff2bb(a_m, b_m);
	return make_bfloat(a, a_e);
    }
}
#endif

Sex
diff2di(Sex a, Sex b)
{
    if (b == mkfxat(0))
	return a;
    stackcheck();
    push(a);
    b = bfloatiti(b);
    pop(a);
    return diff2dd(a, b);
}


Sex
diff2dr(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = bfloatitr(b);
    pop(a);
    return diff2dd(a, b);
}

Sex
diff2df(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    y = float_val(b);
    x = floatitd(a);
    x = x - y;
    return make_float(x);
#else
    stackcheck();
    push(a);
    b = bfloatitf(float_val(b));
    pop(a);
    return diff2dd(a, b);
#endif
}


Sex
diff2rd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitr(a);
    pop(b);
    return diff2dd(a, b);
}

Sex
diff2id(Sex a, Sex b)
{
    if (a == mkfxat(0))
	return negated(b);
    stackcheck();
    push(b);
    a = bfloatiti(a);
    pop(b);
    return diff2dd(a, b);
}


Sex
diff2fd(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    x = float_val(a);
    y = floatitd(b);
    x = x - y;
    return make_float(x);

#else
    stackcheck();
    push(b);
    a = bfloatitf(float_val(a));
    pop(b);
    return diff2dd(a, b);
#endif
}


Sex
diff2bd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitb(a);
    pop(b);
    return diff2dd(a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


#ifdef USE_MPFR
Sex
times2dd(Sex a, Sex b)
{
    __mpfr_struct   op1,
                    op2,
                    rop;
    Sex             c;
    Sex             d,
                    e;
    long            n,
                    m,
                    s;
    mpfr_prec_t     p;

    // 10ʿbitѴ
    d = bfloat_man(a);
    e = bfloat_man(b);
    n = bnum_n(numlen(d)) - 1;
    if (bignum_value(d)[n] == 0)
	return a;
    n = bnum_n(numlen(e)) - 1;
    if (bignum_value(e)[n] == 0)
	return b;

    p = bignum_precision(d);

    mpfr_init2(&rop, p);

    op1 = bfloat2mpfr(a);
    op2 = bfloat2mpfr(b);

    mpfr_mul(&rop, &op1, &op2, GMP_RNDN);

    c = mpfr2bfloat(&rop);

    mpfr_clear(&rop);
    ASSERT(is_bfloat(c));
    return c;

}
#else
Sex
times2dd(Sex a, Sex b)
{
    Sex             a_m,
                    b_m;
    long            a_e,
                    b_e;
    long            n;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    a_m = bfloat_man(a);
    a_e = bfloat_exp(a);
    n = bnum_n(numlen(a_m)) - 1;
    if (bignum_value(a_m)[n] == 0)
	return a;		/* 
				 * return 0.0 
				 */
    b_m = bfloat_man(b);
    b_e = bfloat_exp(b);
    n = bnum_n(numlen(b_m)) - 1;
    if (bignum_value(b_m)[n] == 0)
	return b;		/* 
				 * return 0.0 
				 */
    a_m = times2bb(a_m, b_m);
    a_e = a_e + b_e;
    a = make_bfloat(a_m, a_e);
    ASSERT(is_bfloat(a));
    return a;
}
#endif

Sex
times2rd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitr(a);
    pop(b);
    return times2dd(a, b);
}

Sex
times2id(Sex a, Sex b)
{
    long            n;
    n = bfloat_exp(b);
    b = bfloat_man(b);
    a = times2ib(a, b);
    return make_bfloat(a, n);
}

Sex
times2bd(Sex a, Sex b)
{
    long            n;
    n = bfloat_exp(b);
    b = bfloat_man(b);
    a = times2bb(a, b);
    return make_bfloat(a, n);
}


Sex
times2fd(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    ASSERT(is_float(a) && is_bfloat(b));
    // printf("times2fd called\n");
    x = float_val(a);
    y = floatitd(b);
    x = x * y;
    return make_float(x);
#else
    ASSERT(is_float(a) && is_bfloat(b));
    // printf("times2fd called\n");
    stackcheck();
    push(b);
    a = bfloatitf(float_val(a));
    pop(b);
    return times2dd(a, b);
#endif
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
quot2db(Sex a, Sex b)
{
    long            l;
    long            n;

    l = bnum_n(numlen(b));
    n = bfloat_exp(a);
    a = bfloat_man(a);
    n = n - l * LPSIZE;
    stackcheck();
    push(b);
    a = extend_bignum(a, l);
    pop(b);
    a = quot2bb(a, b);
    return make_bfloat(a, n);
}

#ifdef USE_MPFR
Sex
quot2dd(Sex a, Sex b)
{
    __mpfr_struct   op1,
                    op2,
                    rop;
    Sex             c;
    Sex             d,
                    e;
    long            n,
                    m,
                    s;
    mpfr_prec_t     p;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    // 10ʿbitѴ
    d = bfloat_man(a);
    e = bfloat_man(b);

    p = bignum_precision(d);

    n = bnum_n(numlen(d)) - 1;
    if (bignum_value(d)[n] == 0)
	return a;
    m = bnum_n(numlen(e)) - 1;
    if (bignum_value(e)[m] == 0)
	return aerror("quot2dd: zero divide");

    mpfr_init2(&rop, p);

    op1 = bfloat2mpfr(a);
    op2 = bfloat2mpfr(b);

    mpfr_div(&rop, &op1, &op2, GMP_RNDN);
    // mpfr_printf("*%.10Rf/%.10Rf answer = %.100Rf\n",&op1,&op2,&rop);
    c = mpfr2bfloat(&rop);

    mpfr_clear(&rop);
    ASSERT(is_bfloat(c));
    return c;



}
#else
Sex
quot2dd(Sex a, Sex b)
{
    Sex             a_m,
                    b_m;
    long            a_e,
                    b_e;
    long            s;
    long            n;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    a_m = bfloat_man(a);
    a_e = bfloat_exp(a);
    n = bnum_n(numlen(a_m)) - 1;
    if (bignum_value(a_m)[n] == 0)
	return a;		/* 
				 * return 0.0 
				 */
    b_m = bfloat_man(b);
    b_e = bfloat_exp(b);
    n = bnum_n(numlen(b_m)) - 1;
    if (bignum_value(b_m)[n] == 0)
	return aerror("quot2dd: zero divide");	/* 
						 * return 0.0 
						 */
    s = current_b_precision - number_log2(b_m) + number_log2(a_m);
    if (s > 0) {
	stackcheck();
	push(b_m);
	a_m = leftshift(a_m, s);
	pop(b_m);
    } else
	s = 0;
    a_m = quot2bb(a_m, b_m);

    a_e = a_e - b_e - s;
    a = make_bfloat(a_m, a_e);
    ASSERT(is_bfloat(a));
    return a;
}
#endif

Sex
quot2di(Sex a, Sex b)
{
    long            sign = NO;
    long            n;
    long            m = mkint(b);
    if (m < 0) {
	m = -m;
	sign = YES;
    }
    n = bfloat_exp(a);
    a = bfloat_man(a);
    while (m != 0 && (m & 1) == 0) {
	m = m >> 1;
	n--;
    }
    if (m == 1) {
	if (sign)
	    a = negateb(a);
	return make_bfloat(a, n);
    }
    a = extend_bignum(a, 1);
    n = n - LPSIZE;
    a = quot2bi(a, mkfxat(m));
    if (sign)
	a = negateb(a);
    return make_bfloat(a, n);

    /*
     * stackcheck(); push(a); b = bfloatiti(b); pop(a); return quot2dd(a, b); 
     */
}


Sex
quot2dr(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = bfloatitr(b);
    pop(a);
    return quot2dd(a, b);
}

Sex
quot2df(Sex a, Sex b)
{
    // printf("quot2df called\n");
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    y = float_val(b);
    x = floatitd(a);
    x = x / y;
    return make_float(x);
#else
    stackcheck();
    push(a);
    b = bfloatitf(float_val(b));
    pop(a);
    return quot2dd(a, b);
#endif
}


Sex
quot2rd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitr(a);
    pop(b);
    return quot2dd(a, b);
}

Sex
quot2id(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatiti(a);
    pop(b);
    return quot2dd(a, b);
}

Sex
quot2bd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitb(a);
    pop(b);
    return quot2dd(a, b);
}


Sex
quot2fd(Sex a, Sex b)
{
    // printf("quot2fd called\n");
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    x = float_val(a);
    y = floatitd(b);
    x = x / y;
    return make_float(x);
#else
    stackcheck();
    push(b);
    a = bfloatitf(float_val(a));
    pop(b);
    return quot2dd(a, b);
#endif
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
numeq2dd_bad(Sex a, Sex b)
{
    Sex             a_m,
                    b_m;
    long            a_e,
                    b_e;
    long            s;

    a_m = bfloat_man(a);
    a_e = bfloat_exp(a);
    b_m = bfloat_man(b);
    b_e = bfloat_exp(b);
    s = a_e - b_e;
    if (bignum_length(a_m) == bignum_length(b_m)) {
	if (s > current_b_precision + LPSIZE) {
	    s = bnum_n(numlen(a));
	    if (bignum_value(a)[s - 1] == 0)
		return YES;
	    return NO;
	} else if (s < -current_b_precision - LPSIZE) {
	    s = bnum_n(numlen(b));
	    if (bignum_value(b)[s - 1] == 0)
		return YES;
	    return NO;
	}
    }
    stackcheck();
    if (s == 0) {
	return numeq2bb(a_m, b_m);
    } else if (s > 0) {
	push(b_m);
	a_m = leftshiftb(a_m, s);
	pop(b_m);
	return numeq2bb(a_m, b_m);
    } else {
	push(a_m);
	b_m = leftshiftb(b_m, -s);
	pop(a_m);
	return numeq2bb(a_m, b_m);
    }
}

bool
numeq2dd(Sex a, Sex b)
{
    return zerop(diff2dd(a, b));
}

bool
numeq2bd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitb(a);
    pop(b);
    return numeq2dd(a, b);
}

bool
numeq2fd(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    x = float_val(a);
    y = floatitd(b);
    return Bool_to_Sex(x == y);
#else
    stackcheck();
    push(b);
    a = bfloatitf(float_val(a));
    pop(b);
    return numeq2dd(a, b);
#endif
}

bool
numeq2rd(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitr(a);
    pop(b);
    return numeq2dd(a, b);
}

bool
numeq2id(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatiti(a);
    pop(b);
    return numeq2dd(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
lessp2bd(Sex a, Sex b)
{
    return minusp(diff2bd(a, b));
}



bool
lessp2db(Sex a, Sex b)
{
    return minusp(diff2db(a, b));
}

bool
lessp2dd(Sex a, Sex b)
{
    return minusp(diff2dd(a, b));
}

bool
lessp2df(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    y = float_val(b);
    x = floatitd(a);
    return (x < y);
#else
    return minusp(diff2df(a, b));
#endif
}

bool
lessp2di(Sex a, Sex b)
{
    return minusp(diff2di(a, b));
}


bool
lessp2dr(Sex a, Sex b)
{
    return minusp(diff2dr(a, b));
}


bool
lessp2rd(Sex a, Sex b)
{
    return minusp(diff2rd(a, b));
}


bool
lessp2id(Sex a, Sex b)
{
    return minusp(diff2id(a, b));
}




bool
lessp2fd(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION == 1
    double          x,
                    y;
    x = float_val(a);
    y = floatitd(b);
    return (x < y);
#else
    return minusp(diff2fd(a, b));
#endif
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
