/*
 * binterval.c          (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * big interval number
 * $Header: /Users/kako/asuka/src/RCS/binterval.c,v 2.9 2013/04/25 06:25:33 kako Exp $
 */
static char const rcsid[] =
    "$Id: binterval.c,v 2.9 2013/04/25 06:25:33 kako Exp $";



#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"

#include "bfloat.h"


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
fixitV(Sex a)
{
    a = plus2dd(binterval_lower(a), binterval_upper(a));
    bfloat_exp(a) = bfloat_exp(a) - 1;
    return fixitd(a);
}


Sex
bintervaliti(Sex a)
{
    Sex             c;
    a = bfloatiti(a);
    stackcheck();
    push(a);
    c = new_binterval();
    pop(a);
    binterval_upper(c) = a;
    binterval_lower(c) = a;
    return c;
}

Sex
bintervalitb(Sex a)
{
    Sex             c;
    a = bfloatitb(a);
    stackcheck();
    push(a);
    c = new_binterval();
    pop(a);
    binterval_upper(c) = a;
    binterval_lower(c) = a;
    return c;
}

Sex
bintervalitd(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = new_binterval();
    pop(a);
    binterval_upper(c) = a;
    binterval_lower(c) = a;
    return c;
}




Sex
bintervalitv(Sex a)
{
    Sex             c;
    double          u,
                    l;

    u = interval_upper(a);
    l = interval_lower(a);
    stackcheck();
    push2(nil, nil);
#define cu stack[-1]
#define cl stack[0]
    cu = bfloatitf(u);
    cl = bfloatitf(l);
    c = new_binterval();
    binterval_upper(c) = cu;
    binterval_lower(c) = cl;
    popv(2);
#undef cu
#undef cl
    return c;
}

Sex
bintervalite(Sex a)
{
    Sex             c;
    double          u,
                    l;

    u = efloat_value(a) + efloat_error(a);
    l = efloat_value(a) - efloat_error(a);
    stackcheck();
    push2(nil, nil);
#define cu stack[-1]
#define cl stack[0]
    cu = bfloatitf(u);
    cl = bfloatitf(l);
    c = new_binterval();
    binterval_upper(c) = cu;
    binterval_lower(c) = cl;
    popv(2);
#undef cu
#undef cl
    return c;
}

Sex
bintervalitE(Sex a)
{
    Sex             c;
    double          e;

#ifdef BEFLOAT_ERROR_IS_DOUBLE
    c = befloat_value(a);
    e = befloat_error(a);
#else
    long            n;

    n = befloat_error(a);
    c = befloat_value(a);
#endif

    stackcheck();
    push4(c, nil, nil, nil);
#define c  stack[-3]
#define d  stack[-2]
#define cu stack[-1]
#define cl stack[0]
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    d = bfloatitf(e);
#else
    d = leftshiftd(c, n);
    if (minusp(d))
	d = negated(d);
#endif
    cu = plus2dd(c, d);
    cl = diff2dd(c, d);
    c = new_binterval();
    binterval_upper(c) = cu;
    binterval_lower(c) = cl;
#undef c
    c = stack[-3];
    popv(4);
#undef d
#undef cu
#undef cl
    return c;
}




Sex
Lbinterval(Sex a)
{
    Sex             b;

    if (is_fixnum(a)) {
	return bintervaliti(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return bintervalitb(a);
	case N_FLOAT:
	case N_RATIONAL:
	    a = bfloatit(a);
	case N_COMPLEX:
	    b = Lbinterval(complex_imag(a));
	    stackcheck();
	    push(b);
	    a = Lbinterval(complex_real(a));
	    pop(b);
	    return Lmkcomplex(a, b);
	case N_BFLOAT:
	    return bintervalitd(a);
	case N_INTERVAL:
	    return bintervalitv(a);
	case N_BINTERVAL:
	    return a;
	case N_EFLOAT:
	    return bintervalite(a);
	case N_BEFLOAT:
	    return bintervalitE(a);
	}
    }
    return error1("bad argument to binterval", a);
}


Sex
Lmkbinterval(Sex l, Sex u)
{
    Sex             a;
    stackcheck();
    push2(l, u);
    stack[-1] = bfloatit(l);
    stack[0] = bfloatit(stack[0]);
    a = new_binterval();
    pop2(u, l);
    binterval_lower(a) = l;
    binterval_upper(a) = u;
    return a;
}




Sex
make_binterval(Sex l, Sex u)
{
    Sex             a;
    stackcheck();
    push2(l, u);
    a = new_binterval();
    pop2(u, l);
    binterval_lower(a) = l;
    binterval_upper(a) = u;
    return a;
}

void
print_binterval(Sex u, putchar_fn * fn)
{
    stackcheck();
    push2(binterval_lower(u), binterval_upper(u));
    (*fn) ('#');
    (*fn) ('V');
    (*fn) ('[');
    print_bfloat(stack[-1], fn);
    (*fn) (',');
    print_bfloat(stack[0], fn);
    (*fn) (']');
    popv(2);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


#ifdef USE_MPFR

Sex
plus2fV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return plus2fv(a, b);
#else
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = bfloatitf(float_val(a));
    stack[-1] = wl;
    b = stack[0];
    wl = plus2ddR(wl, binterval_lower(b), MPFR_RNDD);
    wu = stack[-1];
    push(wl);
    b = stack[-1];
    wu = plus2ddR(wu, binterval_upper(b), MPFR_RNDU);
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
#endif
}

Sex
plus2xV(Sex a, Sex b)
{
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = bfloatit(a);
    stack[-1] = wl;
    b = stack[0];
    wl = plus2ddR(wl, binterval_lower(b), MPFR_RNDD);
    wu = stack[-1];
    b = stack[0];
    push(wl);
    wu = plus2ddR(wu, binterval_upper(b), MPFR_RNDU);
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
}
#else
Sex
plus2fV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return plus2fv(a, b);
#else
    Sex             c,
                    d;

    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    push2(a, b);
    c = plus2dd(a, binterval_lower(b));
    pop2(b, a);
    push(c);
    d = plus2dd(a, binterval_upper(b));
    pop(c);
    return make_binterval(c, d);
#endif
}

Sex
plus2xV(Sex a, Sex b)
{
    Sex             c,
                    d;

    stackcheck();
    push(b);
    a = bfloatit(a);
    pop(b);
    push2(a, b);
    c = plus2dd(a, binterval_lower(b));
    pop2(b, a);
    push(c);
    d = plus2dd(a, binterval_upper(b));
    pop(c);
    return make_binterval(c, d);
}
#endif


Sex
plus2eV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return plus2ee(a, b);
#else
    long            ee,
                    ve;

    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    push(a);
    b = befloatitV(b);
    pop(a);
    return plus2EE(a, b);
#endif
}


Sex
plus2vV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return plus2vv(a, b);
#else
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = bfloatitf(interval_lower(a));
    b = stack[0];
    wl = plus2dd(wl, binterval_lower(b));
    a = stack[-1];
    push(wl);
    wu = bfloatitf(interval_upper(a));
    b = stack[-1];
    wu = plus2dd(wu, binterval_upper(b));
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
#endif
}


Sex
plus2EV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return plus2EE(a, b);
}


#ifdef USE_MPFR


Sex
plus2VV(Sex a, Sex b)
{
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = plus2ddR(binterval_lower(a), binterval_lower(b), MPFR_RNDD);
    // wl = plus2dd(binterval_lower(a), binterval_lower(b));
    push(wl);
    wu = plus2ddR(binterval_upper(a), binterval_upper(b), MPFR_RNDU);
    // wu = plus2dd(binterval_upper(a), binterval_upper(b));
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);

    /*
     * Sex wl, wu;
     * 
     * stackcheck(); push2(a, b); wl = plus2dd(binterval_lower(a),
     * binterval_lower(b)); a = stack[-1]; b = stack[0]; push(wl); wu =
     * plus2dd(binterval_upper(a), binterval_upper(b)); pop(wl); popv(2);
     * return make_binterval(wl, wu); 
     */
}
#else
Sex
plus2VV(Sex a, Sex b)
{
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = plus2dd(binterval_lower(a), binterval_lower(b));
    a = stack[-1];
    b = stack[0];
    push(wl);
    wu = plus2dd(binterval_upper(a), binterval_upper(b));
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
}
#endif

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
negateV(Sex a)
{
    Sex             u,
                    l;
    u = binterval_upper(a);
    l = binterval_lower(a);
    stackcheck();
    push(u);
    l = negated(l);
    pop(u);
    push(l);
    u = negated(u);
    pop(l);
    return make_binterval(u, l);
}

Sex
diff2fV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return diff2fv(a, b);
#else
    Sex             c,
                    d;

    stackcheck();
    push2(a, b);
    c = diff2(a, binterval_lower(b));
    pop2(b, a);
    push(c);
    d = diff2(a, binterval_upper(b));
    pop(c);
    return make_binterval(d, c);
#endif
}

Sex
diff2Vf(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = intervalitV(a);
    pop(b);
    return diff2vf(a, b);
#else
    Sex             c,
                    d;

    stackcheck();
    push2(a, b);
    c = diff2(binterval_upper(a), b);
    pop2(b, a);
    push(c);
    d = diff2(binterval_lower(a), b);
    pop(c);
    return make_binterval(d, c);
#endif
}

Sex
diff2xV(Sex a, Sex b)
{
    Sex             c,
                    d;

    stackcheck();
    push2(a, b);
    c = diff2(a, binterval_lower(b));
    pop2(b, a);
    push(c);
    d = diff2(a, binterval_upper(b));
    pop(c);
    return make_binterval(d, c);
}

Sex
diff2Vx(Sex a, Sex b)
{
    Sex             c,
                    d;

    stackcheck();
    push2(a, b);
    c = diff2(binterval_upper(a), b);
    pop2(b, a);
    push(c);
    d = diff2(binterval_lower(a), b);
    pop(c);
    return make_binterval(d, c);
}



Sex
diff2eV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return diff2ee(a, b);
#else
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    push(b);
    a = befloatite(a);
    pop(b);
    return diff2EE(a, b);
#endif
}


Sex
diff2Ve(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitV(a);
    pop(b);
    return diff2ee(a, b);
#else
    stackcheck();
    push(a);
    b = befloatite(b);
    pop(a);
    push(b);
    a = befloatitV(a);
    pop(b);
    return diff2EE(a, b);
#endif
}


#if 0
Sex
diff2cV(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = becomplexitc(a);
    pop(b);
    push(a);
    b = befloatitV(b);
    pop(a);
    return diff2ZE(a, b);
}
#endif


#if 0
Sex
diff2Vc(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = befloatitV(a);
    pop(b);
    push(a);
    b = becomplexitc(b);
    pop(a);
    return diff2EZ(a, b);
}
#endif

Sex
diff2zV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return diff2ze(a, b);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    push(a);
    b = befloatitV(b);
    pop(a);
    return diff2ZE(a, b);
#endif
}



Sex
diff2Vz(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitV(a);
    pop(b);
    return diff2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatitV(a);
    pop(b);
    push(a);
    b = becomplexitv(b);
    pop(a);
    return diff2EZ(a, b);
#endif
}


Sex
diff2vV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return diff2vv(a, b);
#else
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = bfloatitf(interval_lower(a));
    b = stack[0];
    wl = diff2dd(wl, binterval_upper(b));
    a = stack[-1];
    push(wl);
    wu = bfloatitf(interval_upper(a));
    b = stack[-1];
    wu = diff2dd(wu, binterval_lower(b));
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
#endif
}


Sex
diff2Vv(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = intervalitV(a);
    pop(b);
    return diff2vv(a, b);
#else
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = bfloatitf(interval_upper(b));
    a = stack[-1];
    wl = diff2dd(binterval_lower(a), wl);
    b = stack[0];
    push(wl);
    wu = bfloatitf(interval_lower(b));
    a = stack[-2];
    wu = diff2dd(binterval_upper(a), wu);
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
#endif
}

Sex
diff2EV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return diff2EE(a, b);
}

Sex
diff2VE(Sex a, Sex b)
{
    push(b);
    a = befloatitV(a);
    pop(b);
    return diff2EE(a, b);
}

Sex
diff2VV(Sex a, Sex b)
{
    Sex             wl,
                    wu;

    stackcheck();
    push2(a, b);
    wl = diff2dd(binterval_lower(a), binterval_upper(b));
    a = stack[-1];
    b = stack[0];
    push(wl);
    wu = diff2dd(binterval_upper(a), binterval_lower(b));
    pop(wl);
    popv(2);
    return make_binterval(wl, wu);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
times2fV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return times2fv(a, b);
#else
    Sex             c,
                    d;


    stackcheck();
    push3(a, b, nil);
    c = times2(a, binterval_lower(b));
    a = stack[-2];
    b = stack[-1];
    stack[0] = c;
    d = times2(a, binterval_upper(b));
    c = stack[0];
    b = stack[-1];
    popv(3);
    if (plusp(b))
	return make_binterval(c, d);
    else
	return make_binterval(d, c);
#endif
}

Sex
times2xV(Sex a, Sex b)
{
    Sex             c,
                    d;


    stackcheck();
    push3(a, b, nil);
    c = times2(a, binterval_lower(b));
    a = stack[-2];
    b = stack[-1];
    stack[0] = c;
    d = times2(a, binterval_upper(b));
    c = stack[0];
    b = stack[-1];
    popv(3);
    if (plusp(b))
	return make_binterval(c, d);
    else
	return make_binterval(d, c);
}


Sex
times2eV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return times2ee(a, b);
#else
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return times2eE(a, b);
#endif
}


#if 0
Sex
times2cV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return times2cE(a, b);
}
#endif


Sex
times2zV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return times2ez(b, a);
#else
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return times2zE(a, b);
#endif
}




Sex
times2vV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return times2vv(a, b);
#else
    stackcheck();
    push(b);
    a = bintervalitv(a);
    pop(b);
    return times2VV(a, b);
#endif
}

Sex
times2EV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return times2EE(a, b);
}

#ifdef USE_MPFR


Sex
times2VV(Sex a, Sex b)
{
    __mpfr_struct   a_u,
                    a_l,
                    b_u,
                    b_l;
    __mpfr_struct   l1,
                    l2,
                    u1,
                    u2;
    __mpfr_struct   l,
                    u;

    Sex             c;
    Sex             d;
    Sex             upper,
                    lower;

    mpfr_prec_t     p;
    d = bfloat_man(a);

    // precision
    p = bignum_precision(d);

    mpfr_init2(&l, p);
    mpfr_init2(&u, p);
    mpfr_init2(&l1, p);
    mpfr_init2(&l2, p);
    mpfr_init2(&u1, p);
    mpfr_init2(&u2, p);

    a_u = bfloat2mpfr(binterval_upper(a));
    a_l = bfloat2mpfr(binterval_lower(a));
    b_u = bfloat2mpfr(binterval_upper(b));
    b_l = bfloat2mpfr(binterval_lower(b));


    if (mpfr_sgn(&a_l) > 0) {
	if (mpfr_sgn(&b_l) > 0) {
	    mpfr_mul(&l, &a_l, &b_l, MPFR_RNDD);
	    mpfr_mul(&u, &a_u, &b_u, MPFR_RNDU);
	} else if (mpfr_sgn(&b_u) < 0) {
	    mpfr_mul(&l, &a_u, &b_l, MPFR_RNDD);
	    mpfr_mul(&u, &a_l, &b_u, MPFR_RNDU);
	} else {
	    mpfr_mul(&l, &a_u, &b_l, MPFR_RNDD);
	    mpfr_mul(&u, &a_u, &b_u, MPFR_RNDU);
	}
    } else if (mpfr_sgn(&a_u) < 0) {
	if (mpfr_sgn(&b_l) > 0) {
	    mpfr_mul(&l, &a_l, &b_u, MPFR_RNDD);
	    mpfr_mul(&u, &a_u, &b_l, MPFR_RNDU);
	} else if (mpfr_sgn(&b_u) < 0) {
	    mpfr_mul(&l, &a_u, &b_u, MPFR_RNDD);
	    mpfr_mul(&u, &a_l, &b_l, MPFR_RNDU);
	} else {
	    mpfr_mul(&l, &a_l, &b_u, MPFR_RNDD);
	    mpfr_mul(&u, &a_l, &b_l, MPFR_RNDU);
	}
    } else {
	if (mpfr_sgn(&b_l) > 0) {
	    mpfr_mul(&l, &a_l, &b_u, MPFR_RNDD);
	    mpfr_mul(&u, &a_u, &b_u, MPFR_RNDU);
	} else if (mpfr_sgn(&b_u) < 0) {
	    mpfr_mul(&l, &a_u, &b_l, MPFR_RNDD);
	    mpfr_mul(&u, &a_l, &b_l, MPFR_RNDU);
	} else {
	    mpfr_mul(&l1, &a_l, &b_u, MPFR_RNDD);
	    mpfr_mul(&l2, &a_u, &b_l, MPFR_RNDD);
	    mpfr_min(&l, &l1, &l2, MPFR_RNDD);
	    mpfr_mul(&u1, &a_l, &b_l, MPFR_RNDU);
	    mpfr_mul(&u2, &a_u, &b_u, MPFR_RNDU);
	    mpfr_max(&u, &u1, &u2, MPFR_RNDU);
	}
    }


    lower = mpfr2bfloat(&l);
    push(lower);
    upper = mpfr2bfloat(&u);
    pop(lower);

    mpfr_clear(&u);
    mpfr_clear(&l);
    mpfr_clear(&l1);
    mpfr_clear(&l2);
    mpfr_clear(&u1);
    mpfr_clear(&u2);


    return make_binterval(lower, upper);

    /*
     * Sex l, u; Sex l1, l2, u1, u2;
     * 
     * stackcheck(); push4(a, b, nil, nil); #define a stack[-3] #define b
     * stack[-2] #define l stack[-1] #define u stack[0]
     * 
     * //__mpfr_struct op1,op2,rop; //op1 = bfloat2mpfr(a); //op2 =
     * bfloat2mpfr(b);
     * 
     * 
     * if (plusp(binterval_lower(a))) { if (plusp(binterval_lower(b))) { l 
     * = times2(binterval_lower(a), binterval_lower(b)); u =
     * times2(binterval_upper(a), binterval_upper(b)); } else if
     * (minusp(binterval_upper(b))) { u = times2(binterval_lower(a),
     * binterval_upper(b)); l = times2(binterval_upper(a),
     * binterval_lower(b)); } else { u = times2(binterval_upper(a),
     * binterval_upper(b)); l = times2(binterval_upper(a),
     * binterval_lower(b)); } } else if (minusp(binterval_upper(a))) { if
     * (plusp(binterval_lower(b))) { l = times2(binterval_lower(a),
     * binterval_upper(b)); u = times2(binterval_upper(a),
     * binterval_lower(b)); } else if (minusp(binterval_upper(b))) { u =
     * times2(binterval_lower(a), binterval_lower(b)); l =
     * times2(binterval_upper(a), binterval_upper(b)); } else { u =
     * times2(binterval_lower(a), binterval_lower(b)); l =
     * times2(binterval_lower(a), binterval_upper(b)); } } else { if
     * (plusp(binterval_lower(b))) { l = times2(binterval_lower(a),
     * binterval_upper(b)); u = times2(binterval_upper(a),
     * binterval_upper(b)); } else if (minusp(binterval_upper(b))) { u =
     * times2(binterval_lower(a), binterval_lower(b)); l =
     * times2(binterval_upper(a), binterval_lower(b)); } else { u1 =
     * times2(binterval_lower(a), binterval_lower(b)); u2 =
     * times2(binterval_upper(a), binterval_upper(b)); u = Lmax2(u1, u2);
     * l1 = times2(binterval_lower(a), binterval_upper(b)); l2 =
     * times2(binterval_upper(a), binterval_lower(b)); l = Lmax2(l1, l2);
     * } } #undef a #undef b #undef l #undef u l = stack[-1]; u =
     * stack[0]; popv(4); return make_binterval(l, u); 
     */
}
#else
Sex
times2VV(Sex a, Sex b)
{
    Sex             l,
                    u;
    Sex             l2,
                    u2;

    stackcheck();
    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define l stack[-1]
#define u stack[0]

    if (plusp(binterval_lower(a))) {
	if (plusp(binterval_lower(b))) {
	    l = times2(binterval_lower(a), binterval_lower(b));
	    u = times2(binterval_upper(a), binterval_upper(b));
	} else if (minusp(binterval_upper(b))) {
	    u = times2(binterval_lower(a), binterval_upper(b));
	    l = times2(binterval_upper(a), binterval_lower(b));
	} else {
	    u = times2(binterval_upper(a), binterval_upper(b));
	    l = times2(binterval_upper(a), binterval_lower(b));
	}
    } else if (minusp(binterval_upper(a))) {
	if (plusp(binterval_lower(b))) {
	    l = times2(binterval_lower(a), binterval_upper(b));
	    u = times2(binterval_upper(a), binterval_lower(b));
	} else if (minusp(binterval_upper(b))) {
	    u = times2(binterval_lower(a), binterval_lower(b));
	    l = times2(binterval_upper(a), binterval_upper(b));
	} else {
	    u = times2(binterval_lower(a), binterval_lower(b));
	    l = times2(binterval_lower(a), binterval_upper(b));
	}
    } else {
	if (plusp(binterval_lower(b))) {
	    l = times2(binterval_lower(a), binterval_upper(b));
	    u = times2(binterval_upper(a), binterval_upper(b));
	} else if (minusp(binterval_upper(b))) {
	    u = times2(binterval_lower(a), binterval_lower(b));
	    l = times2(binterval_upper(a), binterval_lower(b));
	} else {
	    u = times2(binterval_lower(a), binterval_lower(b));
	    u2 = times2(binterval_upper(a), binterval_upper(b));
	    u = Lmax2(u, u2);
	    l = times2(binterval_lower(a), binterval_upper(b));
	    l2 = times2(binterval_upper(a), binterval_lower(b));
	    l = Lmax2(l, l2);
	}
    }
#undef a
#undef b
#undef l
#undef u
    l = stack[-1];
    u = stack[0];
    popv(4);
    return make_binterval(l, u);
}
#endif


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
quot2fV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return quot2fv(a, b);
#else
    Sex             wu,
                    wl;

    stackcheck();
    push2(a, b);
    wl = quot2(a, binterval_lower(b));
    a = stack[-1];
    b = stack[0];
    push(wl);
    wu = quot2(a, binterval_upper(b));
    pop(wl);
    popv(2);
    return make_binterval(wu, wl);
#endif
}


Sex
quot2Vf(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = intervalitV(a);
    pop(b);
    return quot2vf(a, b);
#else
    Sex             wu,
                    wl;

    stackcheck();
    push3(a, b, nil);
    wl = quot2(binterval_lower(a), b);
    a = stack[-2];
    b = stack[-1];
    stack[0] = wl;
    wu = quot2(binterval_upper(a), b);
    wl = stack[0];
    b = stack[-1];
    popv(3);
    if (plusp(b))
	return make_binterval(wl, wu);
    else
	return make_binterval(wu, wl);
#endif
}

Sex
quot2xV(Sex a, Sex b)
{
    Sex             wu,
                    wl;

    stackcheck();
    push2(a, b);
    wl = quot2(a, binterval_lower(b));
    a = stack[-1];
    b = stack[0];
    push(wl);
    wu = quot2(a, binterval_upper(b));
    pop(wl);
    popv(2);
    return make_binterval(wu, wl);
}


Sex
quot2Vx(Sex a, Sex b)
{
    Sex             wu,
                    wl;

    stackcheck();
    push3(a, b, nil);
    wl = quot2(binterval_lower(a), b);
    a = stack[-2];
    b = stack[-1];
    stack[0] = wl;
    wu = quot2(binterval_upper(a), b);
    wl = stack[0];
    b = stack[-1];
    popv(3);
    if (plusp(b))
	return make_binterval(wl, wu);
    else
	return make_binterval(wu, wl);
}

#if 0
Sex
quot2cV(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = becomplexitc(a);
    pop(b);
    push(a);
    b = befloatitV(b);
    pop(a);
    return quot2ZE(a, b);
}
#endif

Sex
quot2eV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return quot2ee(a, b);
#else
    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    push(a);
    b = befloatitV(b);
    pop(a);
    return quot2EE(a, b);
#endif
}

Sex
quot2EV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return quot2EE(a, b);
}

Sex
quot2vV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = intervalitV(b);
    pop(a);
    return quot2vv(a, b);
#else
    stackcheck();
    push(b);
    a = bintervalitv(a);
    pop(b);
    return quot2VV(a, b);
#endif
}

Sex
quot2Vv(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = intervalitV(a);
    pop(b);
    return quot2vv(a, b);
#else
    stackcheck();
    push(a);
    b = bintervalitv(b);
    pop(a);
    return quot2VV(a, b);
#endif
}


Sex
quot2Ve(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitV(a);
    pop(b);
    return quot2ee(a, b);
#else
    stackcheck();
    push(a);
    b = befloatite(b);
    pop(a);
    push(b);
    a = befloatitV(a);
    pop(b);
    return quot2EE(a, b);
#endif
}


#if 0
Sex
quot2Vc(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = becomplexitc(b);
    pop(a);
    push(b);
    a = befloatitV(a);
    pop(b);
    return quot2EZ(a, b);
}
#endif

Sex
quot2Vz(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = intervalitV(a);
    pop(b);
    return quot2ez(a, b);
#else
    stackcheck();
    push(a);
    b = becomplexitz(b);
    pop(a);
    push(b);
    a = befloatitV(a);
    pop(b);
    return quot2EZ(a, b);
#endif
}


#ifdef USE_MPFR



Sex
quot2VV(Sex a, Sex b)
{
    __mpfr_struct   a_u,
                    a_l,
                    b_u,
                    b_l;
    __mpfr_struct  *al,
                   *au;
    __mpfr_struct   u,
                    l;
    Sex             upper,
                    lower;
    Sex             d;

    mpfr_prec_t     p;
    d = bfloat_man(a);

    // precision
    p = bignum_precision(d);

    mpfr_init2(&l, p);
    mpfr_init2(&u, p);

    a_u = bfloat2mpfr(binterval_upper(a));
    a_l = bfloat2mpfr(binterval_lower(a));
    b_u = bfloat2mpfr(binterval_upper(b));
    b_l = bfloat2mpfr(binterval_lower(b));

    if (mpfr_sgn(&a_l) > 0) {
	if (mpfr_sgn(&b_l) > 0) {
	    al = &a_l;
	    au = &a_u;
	} else if (mpfr_sgn(&b_u) < 0) {
	    al = &a_u;
	    au = &a_l;
	} else {
	    mpfr_clear(&u);
	    mpfr_clear(&l);
	    return aerror("quot2VV: zero divide");
	}
	mpfr_div(&l, al, &b_u, MPFR_RNDD);
	mpfr_div(&u, au, &b_l, MPFR_RNDU);
    } else if (mpfr_sgn(&a_u) < 0) {
	if (mpfr_sgn(&b_l) > 0) {
	    al = &a_l;
	    au = &a_u;
	} else if (mpfr_sgn(&b_u) < 0) {
	    al = &a_u;
	    au = &a_l;
	} else {
	    mpfr_clear(&u);
	    mpfr_clear(&l);
	    return aerror("quot2VV: zero divide");
	}
	mpfr_div(&l, al, &b_l, MPFR_RNDD);
	mpfr_div(&u, au, &b_u, MPFR_RNDU);
    } else {
	if (mpfr_sgn(&b_l) > 0) {
	    mpfr_div(&l, &a_l, &b_l, MPFR_RNDD);
	    mpfr_div(&u, &a_u, &b_l, MPFR_RNDU);
	} else if (mpfr_sgn(&b_u) < 0) {
	    mpfr_div(&l, &a_u, &b_u, MPFR_RNDD);
	    mpfr_div(&u, &a_l, &b_u, MPFR_RNDU);
	} else {
	    mpfr_clear(&u);
	    mpfr_clear(&l);
	    return aerror("quot2VV: zero divide");
	}
    }
    lower = mpfr2bfloat(&l);
    push(lower);
    upper = mpfr2bfloat(&u);
    pop(lower);
    mpfr_clear(&u);
    mpfr_clear(&l);
    return make_binterval(lower, upper);

}
#else
Sex
quot2VV(Sex a, Sex b)
{
    Sex             wl,
                    wu;

    stackcheck();
    push4(nil, nil, a, b);
#define wu stack[-3]
#define wl stack[-2]
#define a stack[-1]
#define b stack[-0]
    if (plusp(binterval_lower(a))) {
	if (plusp(binterval_lower(b))) {
	    wl = binterval_lower(a);
	    wu = binterval_upper(a);
	} else if (minusp(binterval_upper(b))) {
	    wl = binterval_upper(a);
	    wu = binterval_lower(a);
	} else {
	    popv(4);
	    return aerror("quot2VV: zero divide");
	}
	wu = quot2dd(wu, binterval_lower(b));
	wl = quot2dd(wl, binterval_upper(b));
    } else if (minusp(binterval_upper(a))) {
	if (plusp(binterval_lower(b))) {
	    wu = binterval_upper(a);
	    wl = binterval_lower(a);
	} else if (minusp(binterval_upper(b))) {
	    wl = binterval_upper(a);
	    wu = binterval_lower(a);
	} else {
	    popv(4);
	    return aerror("quot2VV: zero divide");
	}
	wl = quot2dd(wl, binterval_lower(b));
	wu = quot2dd(wu, binterval_upper(b));
    } else {
	if (plusp(binterval_lower(b))) {
	    wl = quot2dd(binterval_lower(a), binterval_lower(b));
	    wu = quot2dd(binterval_upper(a), binterval_lower(b));
	} else if (minusp(binterval_upper(a))) {
	    wl = quot2dd(binterval_upper(a), binterval_upper(b));
	    wu = quot2dd(binterval_lower(a), binterval_upper(b));
	} else {
	    popv(4);
	    return aerror("quot2VV: zero divide");
	}
    }
    b = new_binterval();
    binterval_lower(b) = wl;
    binterval_upper(b) = wu;
#undef wu
#undef wl
#undef a
#undef b
    b = stack[0];
    popv(4);
    return b;
}
#endif


Sex
quot2VE(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = befloatitV(a);
    pop(b);
    return quot2EE(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
lessp2Vx(Sex a, Sex b)
{
    return lessp2(binterval_upper(a), b);
}

bool
lessp2xV(Sex a, Sex b)
{
    return lessp2(a, binterval_lower(b));
}

bool
lessp2VV(Sex a, Sex b)
{
    return lessp2dd(binterval_upper(a), binterval_lower(b));
}

bool
lessp2Vv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = bfloatitf(interval_lower(b));
    pop(a);
    return lessp2dd(binterval_upper(a), b);
}

bool
lessp2vV(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitf(interval_upper(a));
    pop(b);
    return lessp2dd(a, binterval_lower(b));
}

bool
lessp2Ve(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = bfloatitf(efloat_value(b));
    pop(a);
    return lessp2dd(binterval_upper(a), b);
}

bool
lessp2eV(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = bfloatitf(efloat_value(a));
    pop(b);
    return lessp2dd(a, binterval_lower(b));
}

/*
 * bool lessp2Vz(Sex a, Sex b) { stackcheck(); push(a); a =
 * bfloatitf(ecomplex_real(b)); pop(a); return
 * lessp2dd(binterval_upper(a), b); } 
 */

/*
 * bool lessp2cV(Sex a, Sex b) { return lessp2(complex_real(a),
 * binterval_lower(b)); }
 * 
 * bool lessp2Vc(Sex a, Sex b) { return lessp2(binterval_upper(a),
 * complex_real(b)); } 
 */

/*
 * bool lessp2zV(Sex a, Sex b) { stackcheck(); push(b); a =
 * bfloatitf(ecomplex_real(a)); pop(b); return lessp2dd(a,
 * binterval_lower(b)); } 
 */

bool
lessp2VE(Sex a, Sex b)
{
    return lessp2dd(binterval_upper(a), befloat_value(b));
}

bool
lessp2EV(Sex a, Sex b)
{
    return lessp2dd(befloat_value(a), binterval_upper(b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
numeq2VV(Sex a, Sex b)
{
    if (!numeq2dd(binterval_lower(a), binterval_lower(b)))
	return NO;
    return numeq2dd(binterval_upper(a), binterval_upper(b));
}

bool
numeq2bV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
numeq2dV(Sex a, Sex b)
{
    if (!numeq2dd(a, binterval_lower(b)))
	return NO;
    return numeq2dd(a, binterval_upper(b));
}

bool
numeq2eV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
numeq2fV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
numeq2iV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
numeq2rV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
numeq2vV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
appeq2bV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2dV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2eV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2fV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2iV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2rV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2vV(Sex a, Sex b)
{
    return zerop(diff2(a, b));
}

bool
appeq2VV(Sex a, Sex b)
{
    if (!appeq2dd(binterval_lower(a), binterval_lower(b)))
	return NO;
    return appeq2dd(binterval_upper(a), binterval_upper(b));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/
