/*
 * char.c               (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/char.c,v 2.7 2015/06/01 08:34:54 kako Exp $
 */

static char const rcsid[] =
    "$Id: char.c,v 2.7 2015/06/01 08:34:54 kako Exp $";


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "kctype.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "read.h"

/***************************************************************/

Sex             Lis_char(Sex a);
Sex             Lis_wchar(Sex a);
Sex             Lxdigitp(Sex a);
Sex             Lcode_char(Sex a);

Sex             Lallocate_string(Sex a);
Sex             Lstring_store(Sex v, Sex n, Sex x);
Sex             Lstring_length(Sex a);
extern Sex      Lcompress(Sex a);
extern Sex      Lexplode(Sex a);
extern Sex      Lstrlen(Sex a);
Sex             Lchar_upcase(Sex a);
Sex             Lchar_downcase(Sex a);
Sex             Lchar_to_id(Sex a);

/***************************************************************/
/***************************************************************/



Sex
Lseprp(Sex a)
{
    int             cc;
#ifdef Kanji
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    if (pname(a)[1]) {
	cc = ((unsigned short *) pname(a))[0];
	return Bool_to_Sex(iskspace(cc));
    }
    return Bool_to_Sex(isspace(pname(a)[0]));
#else
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    cc = pname(a)[0];
    return Bool_to_Sex(isspace(cc));
#endif
}

Sex
Lis_char(Sex a)
{
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    return pname(a)[1] == 0 ? true : nil;
}

Sex
Lis_wchar(Sex a)
{
#ifdef Kanji
    int             c0,
                    c1;
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    c0 = pname(a)[0];
    c1 = pname(a)[1];
    if (c1 == 0)
	return true;
    if (is2byte(c0) && pname(a)[2] == 0)
	return true;
    return nil;
#else
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    return pname(a)[1] == 0 ? true : nil;
#endif
}

Sex
Lliter(Sex a)
{
    int             cc;
#ifdef Kanji
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    if (pname(a)[1]) {
	cc = ((unsigned short *) pname(a))[0];
	return Bool_to_Sex(iskalpha(cc));
    }
    return Bool_to_Sex(isalpha(pname(a)[0]));
#else
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    cc = pname(a)[0];
    return Bool_to_Sex((cc & 0x80) || isalpha(cc));
#endif
}


Sex
Ldigitp(Sex a)
{
    int             cc;
#ifdef Kanji
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    if (pname(a)[1]) {
	cc = ((unsigned short *) pname(a))[0];
	return Bool_to_Sex(isedigit(cc));
    }
    return Bool_to_Sex(isdigit(pname(a)[0]));
#else
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    cc = pname(a)[0];
    return Bool_to_Sex(isdigit(cc));
#endif
}

Sex
Lxdigitp(Sex a)
{
    int             cc;
#ifdef Kanji
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    if (pname(a)[1]) {
	cc = ((unsigned short *) pname(a))[0];
	return Bool_to_Sex(isexdigit(cc));
    }
    return Bool_to_Sex(isxdigit(pname(a)[0]));
#else
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    cc = pname(a)[0];
    return Bool_to_Sex(isxdigit(cc));
#endif
}



Sex
Lspecial_char(Sex a)
{
    int             ch;

    if (!is_fixnum(a))
	return aerror("special-char");
    switch (mkint(a)) {
    case 0:			/* 
				 * space 
				 */
	ch = ' ';
	break;
    case 1:			/* 
				 * newline 
				 */
	ch = '\n';
	break;
    case 2:			/* 
				 * backspace 
				 */
	ch = '\b';
	break;
    case 3:			/* 
				 * tab 
				 */
	ch = '\t';
	break;
    case 4:			/* 
				 * linefeed 
				 */
	ch = '\v';
	break;
    case 5:			/* 
				 * page 
				 */
	ch = '\f';
	break;
    case 6:			/* 
				 * return 
				 */
	ch = '\r';
	break;
    case 7:			/* 
				 * rubout 
				 */
	ch = 0x7f;
	break;
    case 8:			/* 
				 * End Of File character 
				 */
	ch = 0x1a;
	break;
    case 9:			/* 
				 * attention 
				 */
	ch = '\a';
	break;
    default:
	return aerror("special-char");
    }
    return CHAR_OF(ch & 0xff);
}


Sex
Lchar_code(Sex a)
{
    int             cc;
#ifdef Kanji
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    cc = pname(a)[1] ? ((unsigned short *) pname(a))[0] : pname(a)[0];
#else
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
    cc = pname(a)[0];
#endif
    return mkfxat(cc);
}



Sex
Lcode_char(Sex a)
{
    int             cc;
    char            buf[4];
    cc = mkint(a);
    if (cc >= 0 && cc < 0x80)
	return CHAR_OF(cc);
    buf[2] = 0;
    buf[1] = cc >> 8;
    buf[0] = cc;
    return intern(buf, 2);
}

/***************************************************************/
/***************************************************************/
Sex
Lallocate_string(Sex a)
{
    Sex             r;
    char           *s;
    long            k;
    long            n;

    if (!is_fixnum(a))
	return aerror1("allocate_string not fixnum", a);
    n = mkint(a);
    if (n < 0)
	return aerror1("allocate_string negative size", a);

    r = getvector(TAG_BINARY, T_STRING, HDR + n);
    k = (n + (SEX - 1)) & -(2 * SEX);
    s = (char *) r - TAG_BINARY;

    *(long *) (s + k + 2 * SEX) = 0;
    if (k != 0)
	*(long *) (s + k + SEX) = 0;
    memset(s + HDR, 0, (size_t) n);
    return r;

}

Sex
Lstring_store(Sex v, Sex n, Sex x)
{
    long            n1,
                    hl;

    if (!is_fixnum(n))
	return aerror1("sputv offset not fixnum", n);
    if (x == CHAR_EOF || !is_symbol(x)) {
	return aerror1("sputv not character", x);
    }
    n1 = mkint(n);
    if (is_string(v)) {
	hl = string_length(v);
	if (n1 < 0 || n1 >= hl) {
	    return aerror1("sputv index range", n);
	}
	string_value(v, n1) = pname(x)[0];
	return v;
    }
    return aerror1("sputv not string", v);
}

Sex
Lstring_length(Sex a)
{
#if 0
    long            n;
    if (is_symbol(a))
	n = symbol_length(a);
    else if (is_string(a))
	n = string_length(a);
    else
	aerror1("string_length", a);
    return mkfxat(n);
#else
    return Lstrlen(a);
#endif
}


Sex
Lchar_upcase(Sex a)
{
    int             cc;
    char            buf[4];
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;

#ifdef Kanji
    cc = pname(a)[1] ? ((unsigned short *) pname(a))[0] : pname(a)[0];
    if (_isklower(cc))
	cc = _tokupper(cc);
#else
    cc = pname(a)[0];
    if (islower(cc))
	cc = toupper(cc);
#endif
    if (cc >= 0 && cc < 0x80)
	return CHAR_OF(cc);
    buf[2] = 0;
    buf[1] = cc >> 8;
    buf[0] = cc;
    return intern(buf, 2);
}

Sex
Lchar_downcase(Sex a)
{
    int             cc;
    char            buf[4];
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;
#ifdef Kanji
    cc = pname(a)[1] ? ((unsigned short *) pname(a))[0] : pname(a)[0];
    if (_iskupper(cc))
	cc = _toklower(cc);
#else
    cc = pname(a)[0];
    if (isupper(cc))
	cc = tolower(cc);
#endif
    if (cc >= 0 && cc < 0x80)
	return CHAR_OF(cc);
    buf[2] = 0;
    buf[1] = cc >> 8;
    buf[0] = cc;
    return intern(buf, 2);
}

Sex
Lchar_to_id(Sex a)
{
    int             cc;
    char            buf[4];
    if (a == CHAR_EOF || !is_symbol(a))
	return nil;

#ifdef Kanji
    cc = pname(a)[1] ? ((unsigned short *) pname(a))[0] : pname(a)[0];
#else
    cc = pname(a)[0];
#endif
    if (cc >= 0 && cc < 0x80)
	return CHAR_OF(cc);
    buf[2] = 0;
    buf[1] = cc >> 8;
    buf[0] = cc;
    return intern(buf, 2);
}


/***************************************************************/
/***************************************************************/


setup_type const char_setup[] = {
    {"char-code", (lisp_proc *) Lchar_code, 1},
    {"char_code", (lisp_proc *) Lchar_code, 1},
    {"code-char", (lisp_proc *) Lcode_char, 1},
    {"code_char", (lisp_proc *) Lcode_char, 1},
    {"special-char", (lisp_proc *) Lspecial_char, 1},
    {"special_char", (lisp_proc *) Lspecial_char, 1},
    {"digit", (lisp_proc *) Ldigitp, 1},
    {"liter", (lisp_proc *) Lliter, 1},
    {"is-char", (lisp_proc *) Lis_char, 1},
    {"is-wchar", (lisp_proc *) Lis_wchar, 1},
    {"seprp", (lisp_proc *) Lseprp, 1},
    {"isdigit", (lisp_proc *) Ldigitp, 1},
    {"isxdigit", (lisp_proc *) Lxdigitp, 1},
    {"isalpha", (lisp_proc *) Lliter, 1},
    {"isspace", (lisp_proc *) Lseprp, 1},

    {"allocate_string", (lisp_proc *) Lallocate_string, 1},
    {"string_store", (lisp_proc *) Lstring_store, 3},
    {"string_length", (lisp_proc *) Lstrlen, 1},
    {"char_upcase", (lisp_proc *) Lchar_upcase, 1},
    {"char_downcase", (lisp_proc *) Lchar_downcase, 1},
    {"char_to_id", (lisp_proc *) Lchar_to_id, 1},
    {NULL, 0, 0}
};
