/*
 * code.h       (C) Copyright 2011 by Fujio Kako, Nara Women's Univeristy
 *      $Header: /Users/kako/asuka/src/RCS/code.h,v 2.1 2012/10/21 13:32:48 kako Exp $
 */

/*
 * 0x00 nn nop nn 
 * 0x01 nnnn alloc nnnn 
 * 0x02 nnnn dealloc nnnn 
 * 0x03 entry
 * 0x04 exit
 * 
 * 0x05 00 jump 
 * 0x05 02 jumpc 
 * 0x05 03 junpnc 
 * 0x05 04 jumpt 
 * 0x05 05 jumpnil
 * 0x05 08 jumpe 
 * 0x05 09 jumpn
 * 
 * 0x06 link 
 * 0x07 linke 
 * 0x08 call
 * 
 * 0x09 lambind 
 * 0x0a progbind 
 * 0x0b freerstr 
 * 0x0c load 
 * 0x0d store 
 * 0x0e move
 * 
 * 0x0f builtin 
 */

#define	BPS	1000000
#define	NATOM	10000

typedef unsigned short int Code;

extern Code    *bytecode;
extern long     bpsorg;
extern long     bpsend;
extern Sex     *atoms;
extern long	atomorg;
extern long	atomend;


/*
 * builtin functions
 */


#ifdef T_INTERPRETER
zero_args *f0[] = {
	NULL,
/* 1 */		Lgc,		/* reclaim, 0 */
/* 2 */		Lgc,		/* gc, 0 */
/* 3 */		Lfslrclose,	/* fslrclose, 0 */
/* 4 */		Lfslwclose,	/* fslwclose, 0 */
/* 5 */		NULL,	/* fslclose, 0 */
/* 6 */		Lbpsleft,	/* bpsleft, 0 */
/* 7 */		Lanpbporg,	/* &bporg, 0 */
/* 8 */		Lanpbpend,	/* &bpend, 0 */
/* 9 */		Lchr1,		/* chr1, 0 */
/* 10 */	Lchr2,		/* chr2, 0 */
/* 11 */	Lchr4,		/* chr4, 0 */
/* 12 */	Lchrt,		/* chrt, 0 */
/* 13 */	Ldate,		/* date, 0 */
/* 14 */	Lerror1,	/* error1, 0 */
/* 15 */	Lgctime,	/* gctime, 0 */
/* 16 */	Ltime,		/* time, 0 */
/* 17 */	Lfloat_time,	/* float!-time, 0 */
/* 18 */	Lquit,		/* quit, 0 */
/* 19 */	Lbye,		/* bye, 0 */
/* 20 */	Leject,		/* eject, 0 */
/* 21 */	Lflush,		/* flush, 0 */
/* 22 */	Llposn,		/* lposn, 0 */
/* 23 */	Lposn,		/* posn, 0 */
/* 24 */	Lterpri,	/* terpri, 0 */
/* 25 */	Ltmpnam,	/* tmpnam, 0 */
/* 26 */	Lterpri,	/* terpri!*, 0 */
/* 27 */	Lbatchp,	/* batchp, 0 */
/* 28 */	Lrtell,		/* rtell, 0 */
/* 29 */	Lgensym,	/* gensym, 0 */
/* 30 */	Lread,		/* read, 0 */
/* 31 */	Lreadch,	/* readch, 0 */
};


one_args *f1[] = {
	NULL, 
/* 1 */		Lfrexp,		/* frexp, 1 */
/* 2 */		Lceiling,	/* ceiling, 1 */
/* 3 */		Lfloor,		/* floor, 1 */
/* 4 */		Lround,		/* round, 1 */
/* 5 */		Ltruncate,	/* truncate, 1 */
/* 6 */		Ldenominator,	/* denominator, 1 */
/* 7 */		Lnumerator,	/* numerator, 1 */
/* 8 */		Levenp,		/* evenp, 1 */
/* 9 */		lognot,		/* lognot, 1 */
/* 10 */	Llsd,		/* lsd, 1 */
/* 11 */	Lsrandom,	/* srandom, 1 */
/* 12 */	Lrandom,	/* random, 1 */
/* 13 */	Lminusp,	/* minusp, 1 */
/* 14 */	Lmsd,		/* msd, 1 */
/* 15 */	Loddp,		/* oddp, 1 */
/* 16 */	Lonep,		/* onep, 1 */
/* 17 */	Lplusp,		/* plusp, 1 */
/* 18 */	Lzerop,		/* zerop, 1 */
/* 19 */	Ladd1,		/* add1, 1 */
/* 20 */	reciprocal,	/* reciprocal, 1 */
/* 21 */	Lsub1,		/* sub1, 1 */
/* 22 */	Lprecision,	/* precision, 1 */
/* 23 */	Lfix,		/* fix, 1 */
/* 24 */	Llogical,	/* mklogical, 1 */
/* 25 */	Llogical,	/* logical, 1 */
/* 26 */	Lrational,	/* rational, 1 */
/* 27 */	Lfloat,		/* float, 1 */
/* 28 */	Lfloat,		/* mkfloat, 1 */
/* 29 */	Lefloat,	/* efloat, 1 */
/* 30 */	Lmkefloat1,	/* mkefloat1, 1 */
/* 31 */	Linterval,	/* interval, 1 */
/* 32 */	Lcomplex,	/* complex, 1 */
/* 33 */	Lecomplex,	/* ecomplex, 1 */
/* 34 */	Lbfloat,	/* bfloat, 1 */
/* 35 */	Lbfloat,	/* mkbfloat, 1 */
/* 36 */	Lbefloat,	/* befloat, 1 */
/* 37 */	Lbinterval,	/* binterval, 1 */
/* 38 */	Lbecomplex,	/* becomplex, 1 */
/* 39 */	Lcomplex_conjugate,	/* !#cc, 1 */
/* 40 */	Lmkmodular,	/* mkmodular, 1 */
/* 41 */	Linterval_lower,	/* !#itvlow, 1 */
/* 42 */	Linterval_upper,	/* !#itvupp, 1 */
/* 43 */	Linterval_mid,	/* !#itvmid, 1 */
/* 44 */	Linterval_width,	/* !#itvwid, 1 */
/* 45 */	Lefloat_error,	/* !#epart, 1 */
/* 46 */	Lefloat_value,	/* !#vpart, 1 */
/* 47 */	Lecomplex_value,	/* !#cvpart, 1 */
/* 48 */	Lecomplex_error,	/* !#cepart, 1 */
/* 49 */	Lecomplex_real,	/* !#rvpart, 1 */
/* 50 */	Lecomplex_imag,	/* !#ivpart, 1 */
/* 51 */	Lcomplex_real,	/* !#rpart, 1 */
/* 52 */	Lcomplex_imag,	/* !#ipart, 1 */
/* 53 */	Lbefloat_error,	/* !#bepart, 1 */
/* 54 */	Lbefloat_value,	/* !#bvpart, 1 */
/* 55 */	Lnext_prime,	/* next!-prime, 1 */
/* 56 */	Lprimep,	/* primep, 1 */
/* 57 */	Lbmodular_minus,	/* bmodular!-minus, 1 */
/* 58 */	Lbmodular_number,	/* bmodular!-number, 1 */
/* 59 */	Lbmodular_reciprocal,	/* bmodular!-reciprocal, 1 */
/* 60 */	Lset_modulus,	/* set!-modulus, 1 */
/* 61 */	Lchar_code,	/* char!-code, 1 */
/* 62 */	Lcode_char,	/* code!-char, 1 */
/* 63 */	Lspecial_char,	/* special!-char, 1 */
/* 64 */	Ldigitp,	/* digit, 1 */
/* 65 */	Lliter,		/* liter, 1 */
/* 66 */	Lis_char,	/* is!-char, 1 */
/* 67 */	Lis_wchar,	/* is!-wchar, 1 */
/* 68 */	Lseprp,		/* seprp, 1 */
/* 69 */	Ldigitp,	/* isdigit, 1 */
/* 70 */	Lxdigitp,	/* isxdigit, 1 */
/* 71 */	Lliter,		/* isalpha, 1 */
/* 72 */	Lseprp,		/* isspace, 1 */
/* 73 */	Lapply0,	/* apply0, 1 */
/* 74 */	Levlis,		/* evlis, 1 */
/* 75 */	Lfuncall0,	/* funcall0, 1 */
/* 76 */	Lverbos,	/* verbos, 1 */
/* 77 */	Lerrorset1,	/* errorset1, 1 */
/* 78 */	Lfslropen,	/* fslropen, 1 */
/* 79 */	Lfslwopen,	/* fslwopen, 1 */
/* 80 */	Lbpsmove,	/* bpsmove, 1 */
/* 81 */	Lchrbg,		/* chrbg, 1 */
/* 82 */	Latom,		/* atom, 1 */
/* 83 */	Lboundp,	/* boundp, 1 */
/* 84 */	Lcar,		/* car, 1 */
/* 85 */	Lcdr,		/* cdr, 1 */
/* 86 */	Lcaar,		/* caar, 1 */
/* 87 */	Lcadr,		/* cadr, 1 */
/* 88 */	Lcdar,		/* cdar, 1 */
/* 89 */	Lcddr,		/* cddr, 1 */
/* 90 */	Lcaaar,		/* caaar, 1 */
/* 91 */	Lcaadr,		/* caadr, 1 */
/* 92 */	Lcadar,		/* cadar, 1 */
/* 93 */	Lcaddr,		/* caddr, 1 */
/* 94 */	Lcdaar,		/* cdaar, 1 */
/* 95 */	Lcdadr,		/* cdadr, 1 */
/* 96 */	Lcddar,		/* cddar, 1 */
/* 97 */	Lcdddr,		/* cdddr, 1 */
/* 98 */	Lcaaaar,	/* caaaar, 1 */
/* 99 */	Lcaaadr,	/* caaadr, 1 */
/* 100 */	Lcaadar,	/* caadar, 1 */
/* 101 */	Lcaaddr,	/* caaddr, 1 */
/* 102 */	Lcadaar,	/* cadaar, 1 */
/* 103 */	Lcadadr,	/* cadadr, 1 */
/* 104 */	Lcaddar,	/* caddar, 1 */
/* 105 */	Lcadddr,	/* cadddr, 1 */
/* 106 */	Lcdaaar,	/* cdaaar, 1 */
/* 107 */	Lcdaadr,	/* cdaadr, 1 */
/* 108 */	Lcdadar,	/* cdadar, 1 */
/* 109 */	Lcdaddr,	/* cdaddr, 1 */
/* 110 */	Lcddaar,	/* cddaar, 1 */
/* 111 */	Lcddadr,	/* cddadr, 1 */
/* 112 */	Lcdddar,	/* cdddar, 1 */
/* 113 */	Lcddddr,	/* cddddr, 1 */
/* 114 */	Lcodep,		/* codep, 1 */
/* 115 */	Lbytecodep,	/* bytecodep, 1 */
/* 116 */	Lconstantp,	/* constantp, 1 */
/* 117 */	Lerror2,	/* error2, 1 */
/* 118 */	Lfixp,		/* fixp, 1 */
/* 119 */	Lbigp,		/* bigp, 1 */
/* 120 */	Llogicalp,	/* logicalp, 1 */
/* 121 */	Lfloatp,	/* floatp, 1 */
/* 122 */	Lbfloatp,	/* bfloatp, 1 */
/* 123 */	Lrationalp,	/* rationalp, 1 */
/* 124 */	Lcomplexp,	/* complexp, 1 */
/* 125 */	Lcomplexp,	/* complp, 1 */
/* 126 */	Lecomplexp,	/* ecomplexp, 1 */
/* 127 */	Lbecomplexp,	/* becomplexp, 1 */
/* 128 */	Lefloatp,	/* efloatp, 1 */
/* 129 */	Lbefloatp,	/* befloatp, 1 */
/* 130 */	Lintervalp,	/* intervalp, 1 */
/* 131 */	Lbintervalp,	/* bintervalp, 1 */
/* 132 */	Lfluidp,	/* fluidp, 1 */
/* 133 */	Lglobalp,	/* globalp, 1 */
/* 134 */	Lglobal,	/* global, 1 */
/* 135 */	Lfluid,		/* fluid, 1 */
/* 136 */	Lmkquote,	/* mkquote, 1 */
/* 137 */	Lncons,		/* ncons, 1 */
/* 138 */	Lnumberp,	/* numberp, 1 */
/* 139 */	Lstop,		/* stop, 1 */
/* 140 */	Lunglobal,	/* unglobal, 1 */
/* 141 */	Lunfluid,	/* unfluid, 1 */
/* 142 */	Lidp,		/* idp, 1 */
/* 143 */	Lpairp,		/* pairp, 1 */
/* 144 */	Llistp,		/* listp, 1 */
/* 145 */	Lstringp,	/* stringp, 1 */
/* 146 */	Lvectorp,	/* vectorp, 1 */
/* 147 */	Lendp,		/* endp, 1 */
/* 148 */	Lgetd,		/* getd, 1 */
/* 149 */	Llength,	/* length, 1 */
/* 150 */	Lremd,		/* remd, 1 */
/* 151 */	Ltrace,		/* trace, 1 */
/* 152 */	Luntrace,	/* untrace, 1 */
/* 153 */	Lupbv,		/* upbv, 1 */
/* 154 */	Lnull,		/* not, 1 */
/* 155 */	Lnull,		/* null, 1 */
/* 156 */	Lreverse,	/* reverse, 1 */
/* 157 */	Lmkvect,	/* mkvect, 1 */
/* 158 */	Larraydim,	/* arraydim, 1 */
/* 159 */	Lmkarray1,	/* mkarray1, 1 */
/* 160 */	Lmkiarray1,	/* mkiarray, 1 */
/* 161 */	Lmkiarray1,	/* mkivect, 1 */
/* 162 */	Lmkfarray1,	/* mkfvect, 1 */
/* 163 */	Lmkdarray1,	/* mkdvect, 1 */
/* 164 */	Lnreverse,	/* reversip, 1 */
/* 165 */	Lmaphcons,	/* maphcons, 1 */
/* 166 */	Lmodular_minus,	/* modular!-minus, 1 */
/* 167 */	Lmodular_number,	/* modular!-number, 1 */
/* 168 */	Lmodular_reciprocal,	/* modular!-reciprocal, 1 */
/* 169 */	Lset_small_modulus,	/* set!-small-modulus, 1 */
/* 170 */	Liadd1,	/* iadd1, 1 */
/* 171 */	Liminus,	/* iminus, 1 */
/* 172 */	Liminusp,	/* iminusp, 1 */
/* 173 */	Lisub1,		/* isub1, 1 */
/* 174 */	Lionep,		/* ionep, 1 */
/* 175 */	Lizerop,	/* izerop, 1 */
/* 176 */	Lfilep,		/* filep, 1 */
/* 177 */	Llengthc,	/* lengthc, 1 */
/* 178 */	Llinelength,	/* linelength, 1 */
/* 179 */	Lpagelength,	/* pagelength, 1 */
/* 180 */	Lxtab,		/* spaces, 1 */
/* 181 */	Lttab,		/* ttab, 1 */
/* 182 */	Lwrs,		/* wrs, 1 */
/* 183 */	Lxtab,		/* xtab, 1 */
/* 184 */	Lclose,		/* close, 1 */
/* 185 */	Lstrlen,	/* strlen, 1 */
/* 186 */	Lexplode,	/* explode, 1 */
/* 187 */	Lexplodew,	/* explodew, 1 */
/* 188 */	Lexplodec,	/* explodec, 1 */
/* 189 */	Lexplodec,	/* explode2, 1 */
/* 190 */	Lexploden,	/* exploden, 1 */
/* 191 */	Lexplodecn,	/* explodecn, 1 */
/* 192 */	Lexplodecn,	/* explode2n, 1 */
/* 193 */	Lexplodelc,	/* explode2lc, 1 */
/* 194 */	Lexplodeuc,	/* explode2uc, 1 */
/* 195 */	Lprin,		/* prin, 1 */
/* 196 */	Lprin,		/* prin1, 1 */
/* 197 */	Lprinc,		/* princ, 1 */
/* 198 */	Lprinc,		/* prin2, 1 */
/* 199 */	Lprin2t,	/* prin2t, 1 */
/* 200 */	Lprin2a,	/* prin2a, 1 */
/* 201 */	Lprint,		/* print, 1 */
/* 202 */	Lprintc,	/* printc, 1 */
/* 203 */	Lfflush,	/* fflush, 1 */
/* 204 */	Lfeject,	/* feject, 1 */
/* 205 */	Lfterpri,	/* fterpri, 1 */
/* 206 */	Lputch,		/* putch, 1 */
/* 207 */	Llprim,		/* lprim, 1 */
/* 208 */	Llprie,		/* lprie, 1 */
/* 209 */	Lprin,		/* maprin, 1 */
/* 210 */	Lprintf0,	/* printf0, 1 */
/* 211 */	Lsprintf0,	/* sprintf0, 1 */
/* 212 */	Lcompress,	/* compress, 1 */
/* 213 */	Lmake_id,	/* mk!.id, 1 */
/* 214 */	Lrseek,		/* rseek, 1 */
/* 215 */	Lgensym1,	/* gensym1, 1 */
/* 216 */	Lgetenv,	/* getenv, 1 */
/* 217 */	Lintern,	/* intern, 1 */
/* 218 */	Lrds,		/* rds, 1 */
/* 219 */	Lsetpchar,	/* setpchar, 1 */
/* 220 */	Lspool,		/* spool, 1 */
/* 221 */	Lremob,		/* remob, 1 */
/* 222 */	Lshowtime,	/* showtime, 1 */
/* 223 */	Labs,		/* abs, 1 */
/* 224 */	Lacos,		/* acos, 1 */
/* 225 */	Lacosh,		/* acosh, 1 */
/* 226 */	Lasec,		/* asec, 1 */
/* 227 */	Lasech,		/* asech, 1 */
/* 228 */	Lasin,		/* asin, 1 */
/* 229 */	Lasinh,		/* asinh, 1 */
/* 230 */	Latanh,		/* atanh, 1 */
/* 231 */	Lcbrt,		/* cbrt, 1 */
/* 232 */	Lcos,		/* cos, 1 */
/* 233 */	Lcosh,		/* cosh, 1 */
/* 234 */	Lexp,		/* exp, 1 */
/* 235 */	Llog,		/* ln, 1 */
/* 236 */	Llog,		/* log, 1 */
/* 237 */	Llog10,		/* log10, 1 */
/* 238 */	Lsec,		/* sec, 1 */
/* 239 */	Lsech,		/* sech, 1 */
/* 240 */	Lsin,		/* sin, 1 */
/* 241 */	Lsinh,		/* sinh, 1 */
/* 242 */	Lsqrt,		/* sqrt, 1 */
/* 243 */	Ltan,		/* tan, 1 */
/* 244 */	Ltanh,		/* tanh, 1 */
/* 245 */	Latan,		/* atan, 1 */
/* 246 */	Lcsc,		/* csc, 1 */
/* 247 */	Lcsch,		/* csch, 1 */
/* 248 */	Lcot,		/* cot, 1 */
/* 249 */	Lcoth,		/* coth, 1 */
/* 250 */	Lacsc,		/* acsc, 1 */
/* 251 */	Lacsch,		/* acsch, 1 */
/* 252 */	Lacot,		/* acot, 1 */
/* 253 */	Lacoth,		/* acoth, 1 */
/* 254 */	Lfactorial,	/* factorial, 1 */
	};


two_args *f2[] = {
	NULL,
/* 1 */		Llcm,		/* lcm, 2 */
/* 2 */		gcd,		/* gcdn, 2 */
/* 3 */		gcd,		/* gcd, 2 */
/* 4 */		Lash,		/* ash, 2 */
/* 5 */		Lash1,		/* ash1, 2 */
/* 6 */		Ldivide,	/* divide, 2 */
/* 7 */		logand2,	/* logand2, 2 */
/* 8 */		logeqv2,	/* logeqv2, 2 */
/* 9 */		logxor2,	/* logxor2, 2 */
/* 10 */	logor2,		/* logor2, 2 */
/* 11 */	Lmax2,		/* max2, 2 */
/* 12 */	Lmin2,		/* min2, 2 */
/* 13 */	modulus,	/* mod, 2 */
/* 14 */	plus2,		/* plus2, 2 */
/* 15 */	times2,		/* times2, 2 */
/* 16 */	diff2,		/* difference, 2 */
/* 17 */	Leqn,		/* eqn, 2 */
/* 18 */	Lappeq,		/* appeqn, 2 */
/* 19 */	Lgeq,		/* geq, 2 */
/* 20 */	Lgreaterp,	/* greaterp, 2 */
/* 21 */	Lleq,		/* leq, 2 */
/* 22 */	Lleq,		/* lep, 2 */
/* 23 */	Lgeq,		/* gep, 2 */
/* 24 */	Llessp,		/* lessp, 2 */
/* 25 */	Lcgeq,		/* cgeq, 2 */
/* 26 */	Lcgreaterp,	/* cgreaterp, 2 */
/* 27 */	Lcleq,		/* cleq, 2 */
/* 28 */	Lcleq,		/* clep, 2 */
/* 29 */	Lcgeq,		/* cgep, 2 */
/* 30 */	Lclessp,	/* clessp, 2 */
/* 31 */	ratio2,		/* ratio, 2 */
/* 32 */	quot2,		/* quotient, 2 */
/* 33 */	remaind2,	/* remainder, 2 */
/* 34 */	Lmkrational,	/* mkrational, 2 */
/* 35 */	Lmkefloat,	/* mkefloat, 2 */
/* 36 */	Lmkinterval,	/* mkinterval, 2 */
/* 37 */	Lmkcomplex,	/* mkcomplex, 2 */
/* 38 */	Lmkbefloat,	/* mkbefloat, 2 */
/* 39 */	Lmkbinterval,	/* mkbinterval, 2 */
/* 40 */	Lmkcomplex,	/* !#c, 2 */
/* 41 */	Lmkefloat,	/* !#ef, 2 */
/* 42 */	Lmkbefloat,	/* !#bef, 2 */
/* 43 */	Lmkinterval,	/* !#itv, 2 */
/* 44 */	Lleftshift,	/* leftshift, 2 */
/* 45 */	Lgen_prime,	/* gen!-prime, 2 */
/* 46 */	Lsafe_fp_plus,	/* safe!-fp!-plus, 2 */
/* 47 */	Lsafe_fp_times,	/* safe!-fp!-times, 2 */
/* 48 */	Lsafe_fp_quot,	/* safe!-fp!-quot, 2 */
/* 49 */	Lbmodular_difference,	/* bmodular!-difference, 2 */
/* 50 */	Lbmodular_plus,	/* bmodular!-plus, 2 */
/* 51 */	Lbmodular_quotient,	/* bmodular!-quotient, 2 */
/* 52 */	Lbmodular_times,	/* bmodular!-times, 2 */
/* 53 */	Lbmodular_expt,	/* bmodular!-expt, 2 */
/* 54 */	Lapply,	/* apply, 2 */
/* 55 */	Lapply1,	/* apply1, 2 */
/* 56 */	Lfuncall,	/* funcall, 2 */
/* 57 */	Lfuncall1,	/* funcall1, 2 */
/* 58 */	NULL,		/* acons, 2 */
/* 59 */	Lerror,		/* error, 2 */
/* 60 */	Lget,		/* get, 2 */
/* 61 */	Llist2,		/* list2, 2 */
/* 62 */	cons,		/* cons, 2 */
/* 63 */	Lxcons,		/* xcons, 2 */
/* 64 */	Lpair,		/* pair, 2 */
/* 65 */	Lremprop,	/* remprop, 2 */
/* 66 */	Lrplaca,	/* rplaca, 2 */
/* 67 */	Lrplacd,	/* rplacd, 2 */
/* 68 */	Lunion,		/* union, 2 */
/* 69 */	Lunionq,	/* unionq, 2 */
/* 70 */	Lintersection,	/* intersection, 2 */
/* 71 */	Lintersectionq,	/* intersectionq, 2 */
/* 72 */	Lflag,		/* flag, 2 */
/* 73 */	Lflagp,		/* flagp, 2 */
/* 74 */	Lflagpcar,	/* flagpcar, 2 */
/* 75 */	Lflagp,		/* flagp!*!*, 2 */
/* 76 */	Lget,		/* get!*, 2 */
/* 77 */	Lremflag,	/* remflag, 2 */
/* 78 */	Lset,		/* set, 2 */
/* 79 */	Lappend,	/* append, 2 */
/* 80 */	Leq,		/* eq, 2 */
/* 81 */	Leqcar,		/* eqcar, 2 */
/* 82 */	Leql,		/* eql, 2 */
/* 83 */	Lequal,		/* equal, 2 */
/* 84 */	Lmemq,		/* memq, 2 */
/* 85 */	Lnconc,		/* nconc, 2 */
/* 86 */	Laconc,		/* aconc, 2 */
/* 87 */	Lmap,		/* map, 2 */
/* 88 */	Lmapc,		/* mapc, 2 */
/* 89 */	Lmapcar,	/* mapcar, 2 */
/* 90 */	Lmaplist,	/* maplist, 2 */
/* 91 */	Lmapcan,	/* mapcan, 2 */
/* 92 */	Lmapcon,	/* mapcon, 2 */
/* 93 */	Lmapv,		/* mapv, 2 */
/* 94 */	Lneq,		/* neq, 2 */
/* 95 */	Lsmemq,		/* smemq, 2 */
/* 96 */	Lsubla,		/* subla, 2 */
/* 97 */	Lsublis,	/* sublis, 2 */
/* 98 */	Ldelete,	/* delete, 2 */
/* 99 */	Ldeleq,		/* deleq, 2 */
/* 100 */	Lrassoc,	/* rassoc, 2 */
/* 101 */	Lassoc,		/* assoc, 2 */
/* 102 */	Lassoc,		/* assoc!*!*, 2 */
/* 103 */	Latsoc,		/* atsoc, 2 */
/* 104 */	Latsoc,		/* assq, 2 */
/* 105 */	Lgetv,		/* getv, 2 */
/* 106 */	Laref,		/* aref, 2 */
/* 107 */	Lmember,	/* member, 2 */
/* 108 */	Lmember,	/* member!*!*, 2 */
/* 109 */	Lmkarray2,	/* mkarray2, 2 */
/* 110 */	Lmkiarray2,	/* mkiarray2, 2 */
/* 111 */	Lmkfarray2,	/* mkfarray2, 2 */
/* 112 */	Lmkdarray2,	/* mkdarray2, 2 */
/* 113 */	Lgetarray1,	/* getarray1, 2 */
/* 114 */	Lnth,		/* nth, 2 */
/* 115 */	Lpnth,		/* pnth, 2 */
/* 116 */	Lgethash,	/* gethash, 2 */
/* 117 */	Lremhash,	/* remhash, 2 */
/* 118 */	Lmaphash,	/* maphash, 2 */
/* 119 */	Lhcons,		/* hcons, 2 */
/* 120 */	Lhmapcar,	/* hmapcar, 2 */
/* 121 */	Lhappend,	/* happend, 2 */
/* 122 */	Lmodular_difference,	/* modular!-difference, 2 */
/* 123 */	Lmodular_plus,	/* modular!-plus, 2 */
/* 124 */	Lmodular_quotient,	/* modular!-quotient, 2 */
/* 125 */	Lmodular_times,	/* modular!-times, 2 */
/* 126 */	Lmodular_expt,	/* modular!-expt, 2 */
/* 127 */	Lidifference,	/* idifference, 2 */
/* 128 */	Ligeq,		/* igeq, 2 */
/* 129 */	Ligreaterp,	/* igreaterp, 2 */
/* 130 */	Lileq,		/* ileq, 2 */
/* 131 */	Lilessp,	/* ilessp, 2 */
/* 132 */	Lilogand2,	/* ilogand2, 2 */
/* 133 */	Lilogor2,	/* ilogor2, 2 */
/* 134 */	Lilogxor2,	/* ilogxor2, 2 */
/* 135 */	Limax2,		/* imax2, 2 */
/* 136 */	Limin2,		/* imin2, 2 */
/* 137 */	Liplus2,	/* iplus2, 2 */
/* 138 */	Liquotient,	/* iquotient, 2 */
/* 139 */	Liremainder,	/* iremainder, 2 */
/* 140 */	Lirightshift,	/* irightshift, 2 */
/* 141 */	Litimes2,	/* itimes, 2 */
/* 142 */	Litimes2,	/* itimes2, 2 */
/* 143 */	Lopen,		/* open, 2 */
/* 144 */	Lfprin,		/* fprin, 2 */
/* 145 */	Lfprin,		/* fprin1, 2 */
/* 146 */	Lfprinc,	/* fprinc, 2 */
/* 147 */	Lfprinc,	/* fprin2, 2 */
/* 148 */	Lfprin2a,	/* fprin2a, 2 */
/* 149 */	Lfprint,	/* fprint, 2 */
/* 150 */	Lfprintc,	/* fprintc, 2 */
/* 151 */	Lvprintf,	/* vprintf, 2 */
/* 152 */	Lvsprintf,	/* vsprintf, 2 */
/* 153 */	Lfprintf0,	/* fprintf0, 2 */
/* 154 */	Lprintf1,	/* printf1, 2 */
/* 155 */	Lsprintf1,	/* sprintf1, 2 */
/* 156 */	Lbldmsg1,	/* bldmsg1, 2 */
/* 157 */	Lstrcat,	/* strcat, 2 */
/* 158 */	Lorderp,	/* orderp, 2 */
/* 159 */	Lorderp,	/* ordp, 2 */
/* 160 */	Latan2,		/* atan2, 2 */
/* 161 */	Lhypot,		/* hypot, 2 */
/* 162 */	Llog_2,		/* logb, 2 */
/* 163 */	Lpfactorial,	/* pfactorial, 2 */
/* 164 */	Lbinomial,	/* binomial, 2 */
	};

three_args *f3[] = {
	NULL,
/* 1 */		Lmkecomplex,	/* mkecomplex, 3 */
/* 2 */		Lmkbecomplex,	/* mkbecomplex, 3 */
/* 3 */		Lmkecomplex,	/* !#ec, 3 */
/* 4 */		Lapply2,	/* apply2, 3 */
/* 5 */		Lfuncall2,	/* funcall2, 3 */
/* 6 */		Lerrorset,	/* errorset, 3 */
/* 7 */		Llist3,		/* list3, 3 */
/* 8 */		Lput,		/* put, 3 */
/* 9 */		Lputd,		/* putd, 3 */
/* 10 */	Lputv,		/* putv, 3 */
/* 11 */	Lmap2,		/* map2, 3 */
/* 12 */	Lmapc2,		/* mapc2, 3 */
/* 13 */	Lmapcar2,	/* mapcar2, 3 */
/* 14 */	Lmaplist2,	/* maplist2, 3 */
/* 15 */	Lmapcan2,	/* mapcan2, 3 */
/* 16 */	Lmapcon2,	/* mapcon2, 3 */
/* 17 */	Lmapv2,		/* mapv2, 3 */
/* 18 */	Lmkarray3,	/* mkarray3, 3 */
/* 19 */	Lmkiarray3,	/* mkiarray3, 3 */
/* 20 */	Lmkfarray3,	/* mkfarray3, 3 */
/* 21 */	Lmkdarray3,	/* mkdarray3, 3 */
/* 22 */	Lgetarray2,	/* getarray2, 3 */
/* 23 */	Lsetarray1,	/* setarray1, 3 */
/* 24 */	Lsubst,		/* subst, 3 */
/* 25 */	Lsethash,	/* sethash, 3 */
/* 26 */	Lvfprintf,	/* vfprintf, 3 */
/* 27 */	Lfprintf1,	/* fprintf1, 3 */
/* 28 */	Lacons,		/* acons, 3 */
	};

#endif
/*
 */
