/*
 * compare.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * Arithmetic functions. lessp, greaterp, eqn etc. 
 *      $Header: /Users/kako/asuka/src/RCS/compare.c,v 2.9 2016/09/06 05:50:02 kako Exp kako $
 */

static char const rcsid[] =
    "$Id: compare.c,v 2.9 2016/09/06 05:50:02 kako Exp kako $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"
#include "compare.h"
#include "mathop.h"

bool            boolerr1(Sex a);
bool            bool_er(Sex a, Sex b);
bool            bool_yes(Sex a, Sex b);
bool            bool_no(Sex a, Sex b);
bool            zerope(Sex a);
bool            zeropE(Sex a);
bool            zeropV(Sex a);
bool            zeropz(Sex a);
bool            zeropZ(Sex a);
bool            zeropl(Sex a);
bool            zeropY(Sex a);


typedef bool    bool_one(Sex);
typedef bool    bool_two(Sex, Sex);



bool
boolerr1(Sex a)
{
    aerror1("invalid argument to compare", a);
    return NO;
}

bool
bool_er(Sex a, Sex b)
{
    // lisp_printf("bool_er: numtype_a = %x, b=%x\n", numtype_of(a),
    // numtype_of(b));
    aerror2("invalid argument to compare", a, b);
    return NO;
}

bool
bool_yes(Sex a, Sex b)
{
    return YES;
}

bool
bool_no(Sex a, Sex b)
{
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
lessp2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return a < b;
    else if (is_bigneg(b))
	return YES;
    return NO;
}

bool
lessp2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return a < b;
    else if (is_bigneg(b))
	return NO;
    return YES;
}

bool
lessp2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return a < b;
	else if (is_bigneg(b))
	    return NO;
	else
	    return YES;
    } else if (is_fixnum(b)) {
	if (is_bigneg(a))
	    return YES;
	else
	    return NO;
    } else {
	return lessp2bb(a, b);
    }
}




bool
lessp2il(Sex a, Sex b)
{
    long            n = mkint(a);

    if (logical_length(b) != 12)
	return YES;
    if (n < 0)
	return YES;
    if ((unsigned long) n < logical_value(b)[0])
	return YES;
    return NO;
}

bool
lessp2li(Sex a, Sex b)
{
    long            n = mkint(b);

    if (logical_length(a) != 12)
	return NO;
    if (n < 0)
	return NO;
    if (logical_value(a)[0] < (unsigned long) n)
	return YES;
    return NO;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/



two_args       *lessp_table_1[] =
    { lessp2il, lessp2ib, lessp2im, lessp2ir, lessp2if,
    lessp2ie, bool_er, lessp2iv, lessp2id, bool_er,
    lessp2iE, bool_er, lessp2iV, bool_er, bool_er, lessp2iY
};

two_args       *lessp_table_2[] =
    { lessp2li, lessp2bi, lessp2mi, lessp2ri, lessp2fi,
    lessp2ei, bool_er, lessp2vi, lessp2di, bool_er,
    lessp2Ei, bool_er, lessp2Vi, bool_er, bool_er, lessp2Yi
};

two_args       *lessp_table[TYPEMAX][TYPEMAX] = {
    {lessp2ll, lessp2lb, lessp2lm, lessp2lr, lessp2lf,
     lessp2le, bool_er, lessp2lv, lessp2ld, bool_er,
     lessp2lE, bool_er, lessp2lV, bool_er, bool_er, lessp2lY}
    ,
    {lessp2bl, lessp2bb, lessp2bm, lessp2br, lessp2bf,
     lessp2be, bool_er, lessp2bv, lessp2bd, bool_er,
     lessp2bE, bool_er, lessp2bV, bool_er, bool_er, lessp2bY}
    ,
    {lessp2ml, lessp2mb, lessp2mm, lessp2mr, lessp2mf,
     lessp2me, bool_er, lessp2mv, lessp2md, bool_er,
     lessp2mE, bool_er, lessp2mV, bool_er, bool_er, lessp2mY}
    ,
    {lessp2rl, lessp2rb, lessp2rm, lessp2rr, lessp2rf,
     lessp2re, bool_er, lessp2rv, lessp2rd, bool_er,
     lessp2rE, bool_er, lessp2rV, bool_er, bool_er, lessp2rY}
    ,
    {lessp2fl, lessp2fb, lessp2fm, lessp2fr, lessp2ff,
     lessp2fe, bool_er, lessp2fv, lessp2fd, bool_er,
     lessp2fE, bool_er, lessp2fV, bool_er, bool_er, lessp2fY}
    ,
    {lessp2el, lessp2eb, lessp2em, lessp2er, lessp2ef,
     lessp2ee, bool_er, lessp2ev, lessp2ed, bool_er,
     lessp2eE, bool_er, lessp2eV, bool_er, bool_er, lessp2eY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er, lessp2cY}
    ,
    {lessp2vl, lessp2vb, lessp2vm, lessp2vr, lessp2vf,
     lessp2ve, bool_er, lessp2vv, lessp2vd, bool_er,
     lessp2vE, bool_er, lessp2vV, bool_er, bool_er, lessp2vY}
    ,
    {lessp2dl, lessp2db, lessp2dm, lessp2dr, lessp2df,
     lessp2de, bool_er, lessp2dv, lessp2dd, bool_er,
     lessp2dE, bool_er, lessp2dV, bool_er, bool_er, lessp2dY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er, lessp2zY}
    ,
    {lessp2El, lessp2Eb, lessp2Em, lessp2Er, lessp2Ef,
     lessp2Ee, bool_er, lessp2Ev, lessp2Ed, bool_er,
     lessp2EE, bool_er, lessp2EV, bool_er, bool_er, lessp2EY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er, lessp2ZY}
    ,
    {lessp2Vl, lessp2Vb, lessp2Vm, lessp2Vr, lessp2Vf,
     lessp2Ve, bool_er, lessp2Vv, lessp2Vd, bool_er,
     lessp2VE, bool_er, lessp2VV, bool_er, bool_er, lessp2VY}
    ,
    {lessp2Ml, lessp2Mb, lessp2Mm, lessp2Mr, lessp2Mf,
     lessp2Me, bool_er, lessp2Mv, lessp2Md, bool_er,
     lessp2ME, bool_er, lessp2MV, lessp2MM, bool_er, lessp2MY}
    ,
    {lessp2ql, lessp2qb, lessp2qm, lessp2qr, lessp2qf,
     lessp2qe, bool_er, lessp2qv, lessp2qd, bool_er,
     lessp2qE, bool_er, lessp2qV, lessp2qM, bool_er, lessp2qY}
    ,
    {lessp2Yl, lessp2Yb, lessp2Ym, lessp2Yr, lessp2Yf,
     lessp2Ye, bool_er, lessp2Yv, lessp2Yd, bool_er,
     lessp2YE, bool_er, lessp2YV, lessp2YM, bool_er, lessp2YY}
};



bool
lessp2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return a < b ? YES : NO;
	} else if (is_numhdr(b))
	    return lessp_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return lessp_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b)) {
	    return lessp_table[numtype_of(a)][numtype_of(b)] (a, b);
	}
    }
    return bool_er(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


two_args       *clessp_table_1[] =
    { lessp2il, lessp2ib, lessp2im, lessp2ir, lessp2if,
    lessp2ie, lessp2ic, lessp2iv, lessp2id, lessp2iz,
    lessp2iE, lessp2iZ, lessp2iV, lessp2iM, lessp2iq, lessp2iY
};

two_args       *clessp_table_2[] =
    { lessp2li, lessp2bi, lessp2mi, lessp2ri, lessp2fi,
    lessp2ei, lessp2ci, lessp2vi, lessp2di, lessp2zi,
    lessp2Ei, lessp2Zi, lessp2Vi, lessp2Mi, lessp2qi, lessp2Yi
};

two_args       *clessp_table[TYPEMAX][TYPEMAX] = {
    {lessp2ll, lessp2lb, lessp2lm, lessp2lr, lessp2lf,
     lessp2le, lessp2lc, lessp2lv, lessp2ld, lessp2lz,
     lessp2lE, lessp2lZ, lessp2lV, bool_er, bool_er, lessp2lY}
    ,
    {lessp2bl, lessp2bb, lessp2bm, lessp2br, lessp2bf,
     lessp2be, lessp2bc, lessp2bv, lessp2bd, lessp2bz,
     lessp2bE, lessp2bZ, lessp2bV, bool_er, bool_er, lessp2bY}
    ,
    {lessp2ml, lessp2mb, lessp2mm, lessp2mr, lessp2mf,
     lessp2me, lessp2mc, lessp2mv, lessp2md, lessp2mz,
     lessp2mE, lessp2mZ, lessp2mV, bool_er, bool_er, lessp2mY}
    ,
    {lessp2rl, lessp2rb, lessp2rm, lessp2rr, lessp2rf,
     lessp2re, lessp2rc, lessp2rv, lessp2rd, lessp2rz,
     lessp2rE, lessp2rZ, lessp2rV, bool_er, bool_er, lessp2rY}
    ,
    {lessp2fl, lessp2fb, lessp2fm, lessp2fr, lessp2ff,
     lessp2fe, lessp2fc, lessp2fv, lessp2fd, lessp2fz,
     lessp2fE, lessp2fZ, lessp2fV, bool_er, bool_er, lessp2fY}
    ,
    {lessp2el, lessp2eb, lessp2em, lessp2er, lessp2ef,
     lessp2ee, lessp2ec, lessp2ev, lessp2ed, lessp2ez,
     lessp2eE, lessp2eZ, lessp2eV, bool_er, bool_er, lessp2eY}
    ,
    {lessp2cl, lessp2cb, lessp2cm, lessp2cr, lessp2cf,
     lessp2ce, lessp2cc, lessp2cv, lessp2cd, lessp2cz,
     lessp2cE, lessp2cZ, lessp2cV, bool_er, bool_er, lessp2cY}
    ,
    {lessp2vl, lessp2vb, lessp2vm, lessp2vr, lessp2vf,
     lessp2ve, lessp2vc, lessp2vv, lessp2vd, lessp2vz,
     lessp2vE, lessp2vZ, lessp2vV, bool_er, bool_er, lessp2vY}
    ,
    {lessp2dl, lessp2db, lessp2dm, lessp2dr, lessp2df,
     lessp2de, lessp2dc, lessp2dv, lessp2dd, lessp2dz,
     lessp2dE, lessp2dZ, lessp2dV, bool_er, bool_er, lessp2dY}
    ,
    {lessp2zl, lessp2zb, lessp2zm, lessp2zr, lessp2zf,
     lessp2ze, lessp2zc, lessp2zv, lessp2zd, lessp2zz,
     lessp2zE, lessp2zZ, lessp2zV, bool_er, bool_er, lessp2zY}
    ,
    {lessp2El, lessp2Eb, lessp2Em, lessp2Er, lessp2Ef,
     lessp2Ee, lessp2Ec, lessp2Ev, lessp2Ed, lessp2Ez,
     lessp2EE, lessp2EZ, lessp2EV, bool_er, bool_er, lessp2EY}
    ,
    {lessp2Zl, lessp2Zb, lessp2Zm, lessp2Zr, lessp2Zf,
     lessp2Ze, lessp2Zc, lessp2Zv, lessp2Zd, lessp2Zz,
     lessp2ZE, lessp2ZZ, lessp2ZV, bool_er, bool_er, lessp2ZY}
    ,
    {lessp2Vl, lessp2Vb, lessp2Vm, lessp2Vr, lessp2Vf,
     lessp2Ve, lessp2Vc, lessp2Vv, lessp2Vd, lessp2Vz,
     lessp2VE, lessp2VZ, lessp2VV, bool_er, bool_er, lessp2VY}
    ,
    {lessp2Ml, lessp2Mb, lessp2Mm, lessp2Mr, lessp2Mf,
     lessp2Me, lessp2Mc, lessp2Mv, lessp2Md, lessp2Mz,
     lessp2ME, lessp2MZ, lessp2MV, bool_er, bool_er, lessp2MY}
    ,
    {lessp2ql, lessp2qb, lessp2qm, lessp2qr, lessp2qf,
     lessp2qe, lessp2qc, lessp2qv, lessp2qd, lessp2qz,
     lessp2qE, lessp2qZ, lessp2qV, bool_er, bool_er, lessp2qY}
    ,
    {lessp2Yl, lessp2Yb, lessp2Ym, lessp2Yr, lessp2Yf,
     lessp2Ye, lessp2Yc, lessp2Yv, lessp2Yd, lessp2Yz,
     lessp2YE, lessp2YZ, lessp2YV, bool_er, bool_er, lessp2YY}
};


Sex
clessp2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return a < b ? YES : NO;
	} else if (is_numhdr(b))
	    return clessp_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return clessp_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b))
	    return clessp_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return bool_er(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
Lzerop(Sex a)
{
    if (is_fixnum(a))
	return Bool_to_Sex(a == mkfxat(0));
    return Bool_to_Sex(zerop(a));
}

Sex
Lonep(Sex a)
{
    if (is_fixnum(a))
	return Bool_to_Sex(a == mkfxat(1));
    return Bool_to_Sex(onep(a));
}

Sex
Levenp(Sex a)
{
    switch ((long) a & TAG_BITS) {
    case TAG_FIXNUM:
	return ((long) a & 0x10) == 0 ? true : nil;
    case TAG_NUMBERS:
	if (is_bigint(a))
	    return (bignum_value(a)[0] & 1) == 0 ? true : nil;
    default:
	error("evenp: not integer");
	return nil;
    }
}

Sex
Loddp(Sex a)
{
    switch ((long) a & TAG_BITS) {
    case TAG_FIXNUM:
	return ((long) a & 0x10) != 0 ? true : nil;
    case TAG_NUMBERS:
	if (is_bigint(a))
	    return (bignum_value(a)[0] & 1) != 0 ? true : nil;
    default:
	error("oddp: not integer");
	return nil;
    }
}

Sex
Lminusp(Sex a)
{
    return (is_number(a) && minusp(a)) ? true : nil;
}

Sex
Lplusp(Sex a)
{
    return (is_number(a) && plusp(a)) ? true : nil;
}


Sex
Lappeq(Sex a, Sex b)
{
    bool            r;
    r = appeq2(a, b);
    return Bool_to_Sex(r);
}

Sex
Leqn(Sex a, Sex b)
{
    bool            r;
    r = numeq2(a, b);
    return Bool_to_Sex(r);
}

Sex
Llessp(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a < b);
    r = lessp2(a, b);
    return Bool_to_Sex(r);
}

Sex
Lgreaterp(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a > b);
    r = lessp2(b, a);
    return Bool_to_Sex(r);
}

Sex
Lgeq(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a >= b);
    r = lessp2(a, b);
    return Bool_to_Sex(!r);
}

Sex
Lleq(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a <= b);
    r = lessp2(b, a);
    return Bool_to_Sex(!r);
}

Sex
Lclessp(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a < b);
    r = clessp2(a, b);
    return Bool_to_Sex(r);
}

Sex
Lcgreaterp(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a > b);
    r = clessp2(b, a);
    return Bool_to_Sex(r);
}

Sex
Lcgeq(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a >= b);
    r = clessp2(a, b);
    return Bool_to_Sex(!r);
}

Sex
Lcleq(Sex a, Sex b)
{
    bool            r;
    if (is_fixnum(a) && is_fixnum(b))
	return Bool_to_Sex(a <= b);
    r = clessp2(b, a);
    return Bool_to_Sex(!r);
}


Sex
Lmax2(Sex a, Sex b)
{
    bool            w;
    if (is_fixnum(a) && is_fixnum(b))
	return a > b ? a : b;
    stackcheck();
    push2(a, b);
    w = lessp2(a, b);
    pop2(b, a);
    return w ? b : a;
}

Sex
Lmin2(Sex a, Sex b)
{
    bool            w;
    if (is_fixnum(a) && is_fixnum(b))
	return a > b ? b : a;
    stackcheck();
    push2(a, b);
    w = lessp2(a, b);
    pop2(b, a);
    return w ? a : b;
}

bool
zerope(Sex a)
{
    double          v = efloat_value(a);
    double          e = efloat_error(a);
    return (v <= e) && (v >= -e);
}


bool
zeropd(Sex a)
{
    long            n;
    a = bfloat_man(a);
    n = bnum_n(numlen(a)) - 1;
    if (bignum_value(a)[n] == 0)
	return YES;
    return NO;
}

bool
zeropE(Sex a)
{
    bf_error_type   d;
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    d = floatit(befloat_value(a));
    d = fabs(d);
    return d <= befloat_error(a);
#else
    d = current_b_precision - befloat_error(a);
    if (d < 0)
	return YES;
    a = befloat_value(a);
    a = bfloat_man(a);
    d = bnum_n(numlen(a)) - 1;
    if (bignum_value(a)[d] == 0)
	return YES;
    return NO;
#endif
}

bool
zeropV(Sex a)
{
    return plusp(binterval_upper(a)) && minusp(binterval_lower(a));
}

bool
zeropz(Sex a)
{
    return (ecomplex_real(a) == 0.0 && ecomplex_imag(a) == 0);
}

bool
zeropZ(Sex a)
{
    push(a);
    if (!zeropd(becomplex_real(a))) {
	popv(1);
	return NO;
    }
    pop(a);
    return zeropd(becomplex_imag(a));
}

bool
zeropl(Sex a)
{
    if (logical_length(a) == 12 && logical_value(a)[0] == 0)
	return YES;
    return NO;
}

bool
zeropY(Sex a)
{
    return fmodular_ivalue(a) == 0 ? YES : NO;
}


bool
zerop(a)
     Sex             a;
{
    if (is_fixnum(a)) {
	return (a == mkfxat(0));
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return zeropl(a);
	case N_BIGNUM:
	    return NO;
	case N_RATIONAL:
	    return NO;
	case N_FLOAT:
	    return (float_val(a) == 0.0);
	case N_COMPLEX:
	    return zerop(complex_real(a)) && zerop(complex_imag(a));
	case N_EFLOAT:
	    return zerope(a);
	case N_INTERVAL:
	    return interval_lower(a) <= 0.0 && interval_upper(a) >= 0.0;
	case N_BFLOAT:
	    return zeropd(a);
	case N_ECOMPLEX:
	    return zeropz(a);
	case N_BEFLOAT:
	    return zeropE(a);
	case N_BINTERVAL:
	    return zeropV(a);
	case N_BECOMPLEX:
	    return zeropZ(a);
	case N_MODULAR:
	    return modular_value(a) == 0;
	case N_BMODULAR:
	    return zerop(bmodular_value(a));
	case N_FMODULAR:
	    return fmodular_ivalue(a) == 0;
	}
    }
    return NO;
}


bool
onep(a)
     Sex             a;
{
    if (is_fixnum(a)) {
	return a == mkfxat(1);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return zerop(Lsub1(a));
	case N_BIGNUM:
	    return NO;
	case N_RATIONAL:
	    return NO;
	case N_FLOAT:
	    return (float_val(a) == 1.0);
	case N_COMPLEX:
	    return zerop(complex_imag(a)) && onep(complex_real(a));
	case N_EFLOAT:
	    return (efloat_value(a) - efloat_error(a) < 1.0)
		&& (efloat_value(a) + efloat_error(a) > 1.0);
	case N_INTERVAL:
	    return zerop(Lsub1(a));
	case N_BFLOAT:
	    return zerop(Lsub1(a));
	case N_ECOMPLEX:
	    return zerop(Lsub1(a));
	case N_BEFLOAT:
	    return zerop(Lsub1(a));
	case N_BINTERVAL:
	    return zerop(Lsub1(a));
	case N_BECOMPLEX:
	    return zerop(Lsub1(a));
	case N_MODULAR:
	    return modular_value(a) == 1;
	case N_BMODULAR:
	    return onep(bmodular_value(a));
	case N_FMODULAR:
	    return fmodular_dvalue(a) == 1.0;
	}
    }
    return NO;
}

bool
monep(a)
     Sex             a;
{
    if (is_fixnum(a)) {
	return a == mkfxat(-1);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return zerop(Ladd1(a));
	case N_BIGNUM:
	    return NO;
	case N_RATIONAL:
	    return NO;
	case N_FLOAT:
	    return (float_val(a) == -1.0);
	case N_COMPLEX:
	    return zerop(complex_imag(a)) && monep(complex_real(a));
	case N_EFLOAT:
	    return (efloat_value(a) - efloat_error(a) < -1.0)
		&& (efloat_value(a) + efloat_error(a) > -1.0);
	case N_INTERVAL:
	    return zerop(Ladd1(a));
	case N_BFLOAT:
	    return zerop(Ladd1(a));
	case N_ECOMPLEX:
	    return zerop(Ladd1(a));
	case N_BEFLOAT:
	    return zerop(Ladd1(a));
	case N_BINTERVAL:
	    return zerop(Ladd1(a));
	case N_BECOMPLEX:
	    return zerop(Ladd1(a));
	case N_MODULAR:
	    return modular_value(a) == -1;
	case N_BMODULAR:
	    return monep(bmodular_value(a));
	case N_FMODULAR:
	    return fmodular_dvalue(a) == -1.0;
	}
    }
    return NO;
}


/*
 * sign testing
 */


bool
minusp(a)
     Sex             a;
{
    bf_error_type   d;
    long            n;

    if (is_fixnum(a)) {
	n = mkint(a);
	return n < 0;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return NO;
	case N_BIGNUM:
	    return is_bigneg(a) ? YES : NO;
	case N_RATIONAL:
	    return minusp(rational_num(a));
	case N_FLOAT:
	    return (float_val(a) < 0.0);
	case N_COMPLEX:
	    return NO;
	case N_EFLOAT:
	    return (efloat_value(a) < -efloat_error(a));
	case N_INTERVAL:
	    return (interval_upper(a) < 0.0);
	case N_BFLOAT:
	    return !zeropd(a) && minusp(bfloat_man(a));
	case N_ECOMPLEX:
	    if (fabs(ecomplex_imag(a)) < ecomplex_error(a))
		return (ecomplex_error(a) > ecomplex_real(a));
	    return NO;
	case N_BEFLOAT:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    d = floatit(befloat_value(a));
	    return d < -befloat_error(a);
#else
	    d = current_b_precision;
	    return (minusp(befloat_value(a)) && (d > befloat_error(a)));
#endif
	case N_BINTERVAL:
	    return minusp(binterval_upper(a));
	case N_BECOMPLEX:
	    return NO;
	case N_FMODULAR:
	    return fmodular_dvalue(a) < 0;
	}
    }
    return NO;
}



bool
plusp(a)
     Sex             a;
{
    bf_error_type   d;
    long            n;

    if (is_fixnum(a)) {
	n = mkint(a);
	return n > 0;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return NO;
	case N_BIGNUM:
	    return is_bigneg(a) ? NO : YES;
	case N_RATIONAL:
	    return plusp(rational_num(a));
	case N_FLOAT:
	    return (float_val(a) > 0.0);
	case N_COMPLEX:
	    return NO;
	case N_EFLOAT:
	    return (efloat_value(a) > efloat_error(a));
	case N_INTERVAL:
	    return (interval_lower(a) > 0.0);
	case N_BFLOAT:
	    return !zeropd(a) && plusp(bfloat_man(a));
	case N_ECOMPLEX:
	    if (fabs(ecomplex_imag(a)) < ecomplex_error(a))
		return (ecomplex_error(a) < ecomplex_real(a));
	    return NO;
	case N_BEFLOAT:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    d = floatit(befloat_value(a));
	    return d > befloat_error(a);
#else
	    d = current_b_precision;
	    return (plusp(befloat_value(a)) && (d > befloat_error(a)));
#endif
	case N_BINTERVAL:
	    return plusp(binterval_lower(a));
	case N_BECOMPLEX:
	    return NO;
	case N_FMODULAR:
	    return fmodular_dvalue(a) > 0;
	}
    }
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


/*
 * Numeric equality - note that comparisons involving non-numbers are errors
 * here (unlike the position in eql, equal, equalp).  Also this must be coded
 * so that it never provokes garbage collection.
 */
bool
numeq2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return a == b;
    return NO;
}

bool
numeq2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return a == b;
    return NO;
}

bool
numeq2jj(Sex a, Sex b)
{
    if (is_fixnum(a))
	return a == b;
    else if (is_fixnum(b))
	return NO;
    else
	return numeq2bb(a, b);
}


bool
numeq2il(Sex a, Sex b)
{
    if (logical_length(b) == 12 &&
	logical_value(b)[0] == (unsigned long) mkint(a))
	return YES;
    return NO;
}




two_args       *numeq_table_1[] =
    { numeq2il, numeq2ib, numeq2im, numeq2ir, numeq2if,
    numeq2ie, numeq2ic, numeq2iv, numeq2id, numeq2iz,
    numeq2iE, numeq2iZ, numeq2iV, numeq2iM, numeq2iq, numeq2iY
};


two_args       *numeq_table[TYPEMAX][TYPEMAX] = {
    {numeq2ll, numeq2lb, numeq2lm, numeq2lr, numeq2lf,
     numeq2le, numeq2lc, numeq2lv, numeq2ld, numeq2lz,
     numeq2lE, numeq2lZ, numeq2lV, numeq2lM, numeq2lq, numeq2lY}
    ,
    {bool_er, numeq2bb, numeq2bm, numeq2br, numeq2bf,
     numeq2be, numeq2bc, numeq2bv, numeq2bd, numeq2bz,
     numeq2bE, numeq2bZ, numeq2bV, numeq2bM, numeq2bq, numeq2bY}
    ,
    {bool_er, bool_er, numeq2mm, numeq2mr, numeq2mf,
     numeq2me, numeq2mc, numeq2mv, numeq2md, numeq2mz,
     numeq2mE, numeq2mZ, numeq2mV, numeq2mM, numeq2mq, numeq2mY}
    ,
    {bool_er, bool_er, bool_er, numeq2rr, numeq2rf,
     numeq2re, numeq2rc, numeq2rv, numeq2rd, numeq2rz,
     numeq2rE, numeq2rZ, numeq2rV, bool_er, bool_er, numeq2rY}
    ,
    {bool_er, bool_er, bool_er, bool_er, numeq2ff,
     numeq2fe, numeq2fc, numeq2fv, numeq2fd, numeq2fz,
     numeq2fE, numeq2fZ, numeq2fV, numeq2fM, numeq2fq, numeq2fY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     numeq2ee, numeq2ec, numeq2ev, numeq2ed, numeq2ez,
     numeq2eE, numeq2eZ, numeq2eV, numeq2eM, numeq2eq, numeq2eY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, numeq2cc, numeq2cv, numeq2cd, numeq2cz,
     numeq2cE, numeq2cZ, numeq2cV, bool_er, bool_er, numeq2cY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, numeq2vv, numeq2vd, numeq2vz,
     numeq2vE, numeq2vZ, numeq2vV, numeq2vM, bool_er, numeq2vY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, numeq2dd, numeq2dz,
     numeq2dE, numeq2dZ, numeq2dV, numeq2dM, bool_er, numeq2dY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, numeq2zz,
     numeq2zE, numeq2zZ, numeq2zV, numeq2zM, bool_er, numeq2zY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     numeq2EE, numeq2EZ, numeq2EV, numeq2EM, bool_er, numeq2EY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, numeq2ZZ, numeq2ZV, numeq2ZM, bool_er, numeq2ZY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, numeq2VV, numeq2VM, bool_er, numeq2VY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, numeq2MM, numeq2Mq, numeq2MY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, numeq2qM, numeq2qq, numeq2qY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, numeq2YM, numeq2Yq, numeq2YY}
};


bool
numeq2(a, b)
     Sex             a,
                     b;
{
    if (a == b)
	return YES;
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return a == b ? YES : NO;
	} else if (is_numhdr(b))
	    return numeq_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return numeq_table_1[numtype_of(a)] (b, a);
	else if (is_numhdr(b)) {
	    long            i = numtype_of(a);
	    long            j = numtype_of(b);
	    if (i > j)
		return numeq_table[j][i] (b, a);
	    else
		return numeq_table[i][j] (a, b);
	}
    }
    return bool_er(a, b);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/


/*
 * Numeric equality - note that comparisons involving non-numbers are errors
 * here (unlike the position in eql, equal, equalp).  Also this must be coded
 * so that it never provokes garbage collection.
 */


two_args       *appeq_table_1[] =
    { appeq2il, appeq2ib, appeq2im, appeq2ir, appeq2if,
    appeq2ie, appeq2ic, appeq2iv, appeq2id, appeq2iz,
    appeq2iE, appeq2iZ, appeq2iV, appeq2iM, appeq2iq, appeq2iY
};


two_args       *appeq_table[TYPEMAX][TYPEMAX] = {
    {appeq2ll, appeq2lb, appeq2lm, appeq2lr, appeq2lf,
     appeq2le, appeq2lc, appeq2lv, appeq2ld, appeq2lz,
     appeq2lE, appeq2lZ, appeq2lV, bool_er, bool_er, appeq2lY}
    ,
    {bool_er, appeq2bb, appeq2bm, appeq2br, appeq2bf,
     appeq2be, appeq2bc, appeq2bv, appeq2bd, appeq2bz,
     appeq2bE, appeq2bZ, appeq2bV, bool_er, bool_er, appeq2bY}
    ,
    {bool_er, bool_er, appeq2mm, appeq2mr, appeq2mf,
     appeq2me, appeq2mc, appeq2mv, appeq2md, appeq2mz,
     appeq2mE, appeq2mZ, appeq2mV, bool_er, bool_er, appeq2mY}
    ,
    {bool_er, bool_er, bool_er, appeq2rr, appeq2rf,
     appeq2re, appeq2rc, appeq2rv, appeq2rd, appeq2rz,
     appeq2rE, appeq2rZ, appeq2rV, bool_er, bool_er, appeq2rY}
    ,
    {bool_er, bool_er, bool_er, bool_er, appeq2ff,
     appeq2fe, appeq2fc, appeq2fv, appeq2fd, appeq2fz,
     appeq2fE, appeq2fZ, appeq2fV, bool_er, bool_er, appeq2fY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     appeq2ee, appeq2ec, appeq2ev, appeq2ed, appeq2ez,
     appeq2eE, appeq2eZ, appeq2eV, bool_er, bool_er, appeq2eY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, appeq2cc, appeq2cv, appeq2cd, appeq2cz,
     appeq2cE, appeq2cZ, appeq2cV, bool_er, bool_er, appeq2cY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, appeq2vv, appeq2vd, appeq2vz,
     appeq2vE, appeq2vZ, appeq2vV, bool_er, bool_er, appeq2vY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, appeq2dd, appeq2dz,
     appeq2dE, appeq2dZ, appeq2dV, bool_er, bool_er, appeq2dY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, appeq2zz,
     appeq2zE, appeq2zZ, appeq2zV, bool_er, bool_er, appeq2zY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     appeq2EE, appeq2EZ, appeq2EV, bool_er, bool_er, appeq2EY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, appeq2ZZ, appeq2ZV, bool_er, bool_er, appeq2ZY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, appeq2VV, bool_er, bool_er, appeq2VY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, appeq2MM, bool_er, appeq2MY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, appeq2qq, appeq2qY}
    ,
    {bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, bool_er,
     bool_er, bool_er, bool_er, bool_er, appeq2Yq, appeq2YY}
};


bool
appeq2(a, b)
     Sex             a,
                     b;
{
    if (a == b)
	return YES;
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return a == b ? YES : NO;
	} else if (is_numhdr(b))
	    return appeq_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return appeq_table_1[numtype_of(a)] (b, a);
	else if (is_numhdr(b)) {
	    long            i = numtype_of(a);
	    long            j = numtype_of(b);
	    if (i > j)
		return appeq_table[j][i] (b, a);
	    else
		return appeq_table[i][j] (a, b);
	}
    }
    return bool_er(a, b);
}
