% Module COMPLR --- The Standard LISP compiler;
%                @(#) complr.rl 4.2 87/09/14 ;

eval!-when(compile, remflag('(rplacw), 'lose));
eval!-when(compile, flag('(modcmp), 'stayslow));

% Author: Griss and Hearn;
% Modified by: Fujio Kako;                                85-8-28;

comment machine dependent parts are in a separate file;

comment these include the macros described below and, in addition,
        an auxiliary function !&mkfunc which is required to pass
        functional arguments (input as function <func>) to the
        loader. in most cases, !&mkfunc may be defined as mkquote;

comment global flags used in this compiler:

!*module        indicates block compilation (a future extension of
                this compiler)
!*msg           indicates whether certain messages should be printed
!*nolinke       if on inhibits use of !*linke c-macro
!*ord           if on forces left-to-right argument evaluation
!*plap          if on causes lap output to be printed
!*r2i           if on causes recursion removal where possible;

global '(!*module !*msg !*nolinke !*ord !*plap !*r2i);

comment global variables used:

erfg!*          used by reduce to control error recovery
maxnargs        maximum number of arguments permitted;

global '(erfg!* maxnargs);

maxnargs := 13;         %Standard LISP limit;

comment fluid variables used:

alsts           alist of fluid parameters
codelist        code being built
condtail        simulated stack of position in the tail of a cond
dfprint!*       name of special definition process (or nil)
exit            label for !*exit jump
flagg           used in !&comtst, and in !&fixrest
freelst         list of free variables with bindings
golist          storage map for jump labels
iregs           initial register contents
iregs1          temporary placeholder for iregs for branch compilation
jmplist         list of locations in codelist of transfers
lblist          list of label words
llngth          cell whose car is length of frame
name            name of function being currently compiled
narg            number of arguments in function
regs            known current contents of registers as an alist with
                 elements  of form (<reg> . <contents>)
regs1           temporary placeholder for regs during branch compilation
slst            association list for stores which have not yet been used
stlst           list of active stores in function
stomap          storage map for variables
!&switch        boolean expression value flag - keeps track of nulls;

fluid '(alsts codelist condtail dfprint!* exit flagg freelst golist
        iregs iregs1 jmplist lblist llngth name narg regs regs1 slst
        stlst stomap !&switch);

comment c-macros used in this compiler;

comment the following c-macros must not change regs 1-maxnargs:

!*alloc n                allocate new stack frame of n words
!*dealloc n              deallocate above frame
!*entry name type nargs  entry point to function name of type type
                           with nargs args
!*exit                   exit to previously saved return address
!*store reg floc         store contents of reg (or nil) in floc
!*jump adr               unconditional jump
!*jumpc  adr exp type    jump to adr if exp is of type type
!*jumpnc adr exp type    jump to adr if exp is not of type type
!*jumpnil adr exp        jump on exp eq to nil
!*jumpt adr exp          jump on exp not eq to nil
!*jumpe adr exp1 exp2    jump on exp1 eq to exp2
!*jumpn adr exp1 exp2    jump on exp1 not eq to exp2
!*jumpgt adr exp1 exp2   jump on exp1 greater than exp2
!*jumpge adr exp1 exp2   jump on exp1 greater than or eq to exp2
!*jumplt adr exp1 exp2   jump on exp1 less than exp2
!*jumple adr exp1 exp2   jump on exp1 less than or eq to exp2
!*lbl adr                define label
!*lambind regs alst      bind free lambda vars in alst currently in regs
!*progbind alst          bind free prog vars in alst
!*freerstr alst          unbind free variables in alst

comment the following c-macro must only change specific register
        being loaded:

!*load reg exp           load exp into reg;

comment the following c-macros do not protect regs 1-maxnargs:

!*link fn type nargs reg     link to fn of type type with nargs args
!*linke fn type nargs  regn  link to fn of type type with nargs args
                             and exit removing frame of n words
!*code list                  this macro allows for the inclusion of a list
                               of c-macro expressions (or even explicit
                               assembly language) in a function definition;

flag('(!*alloc !*dealloc !*entry !*exit !*store !*jump !*jumpc !*jumpnc
       !*jumpnil !*jumpt !*jumpe !*jumpn 
       !*jumpge !*jumpgt !*jumple !*jumplt
       !*lbl !*lambind !*progbind
       !*freerstr !*load !*link !*linke !*code),
'mc);

comment general functions used in this compiler;


symbolic procedure rplacw(a,b); rplaca(rplacd(a,cdr b),car b);

comment the following two functions are used by the cons open
        coding. they should be defined in the interpreter if
        possible. they should only be compiled without a compfn
        for cons;


comment top level compiling functions;

symbolic procedure compile x;
   begin scalar exp;
      for each y in x do
           if null (exp := getd y) then lprim list(y,'undefined)
            else compd(y,car exp,cdr exp);
      return x
   end;

symbolic procedure compd(name,type,exp);
   begin
      if not flagp(type,'compile)
        then <<lprim list("UNCOMPILABLE FUNCTION",name,"OF TYPE",
                          type);
               return nil>>;
      if not atom exp
        then if !*module then modcmp(name,type,exp)
              else if dfprint!*
               then apply(dfprint!*,
                          list if type eq 'expr
                                 then 'de . (name . cdr exp)
                                else if type eq 'fexpr
                                 then 'df . (name . cdr exp)
                           %     else if type eq 'macro
                           %      then 'dm . (name . cdr exp)
                           %     else if type eq 'smacro
                           %      then 'ds . (name . cdr exp)
                                else list('putd,mkquote name,
                                           mkquote type,
                                           mkquote exp))
              else begin scalar x;
                      if flagp(type,'compile)
                        then put(name,'cfntype,list type);
                      x :=
                       list('!*entry,name,type,length cadr exp)
                         . !&comproc(exp,
                                     if flagp(type,'compile)
                                       then name);
                      if !*plap then <<
			   prin2t "/*";
			   for each y in x do print y;
			   prin2t " */";
		      >>;
                      lap x;
                      %this is the entry point to the assembler.  LAP
                      %must remove any preexisting function definition;
                      if (x := get(name,'cfntype))
                           and eqcar(getd name,car x)
                        then remprop(name,'cfntype)
                   end;
      return name
   end;

flag('(expr fexpr ),'compile);

symbolic procedure !&comproc(exp,name);
   %compiles a function body, returning the generated LAP;
   begin scalar codelist,flagg,iregs,iregs1,jmplist,lblist,
                llngth,regs,regs1,alsts,exit,slst,stlst,stomap,
                condtail,freelst,
                !&switch; integer narg;
      llngth := list 1;
      narg := 0;
      exit := !&genlbl();
      stomap := '((nil 1));
      codelist := list ('!*alloc . llngth);
      exp := !&pass1 exp;
      if length cadr exp>maxnargs
        then lprie list("TOO MANY ARGS FOR COMPILER IN",name);
      for each z in cadr exp do <<!&frame z;
                                  narg := narg + 1;
                                  if not nonlocal z
                                    then iregs :=
                                          nconc(iregs,
                                                list list(narg,z));
                                  regs :=
                                   nconc(regs,list list(narg,z))>>;
      if null regs then regs := list (1 . nil);
      alsts := !&freebind(cadr exp,t);
      !&pass2 caddr exp;
      !&freerst(alsts,0);
      !&pass3();
      rplaca(llngth,1 - car llngth);
      return codelist
   end;

symbolic procedure nonlocal x;
   if fluidp x then 'fluid else if globalp x then 'global else nil;

comment pass 1 of the compiler;

symbolic procedure !&pass1 exp; !&pa1(exp,nil);

symbolic procedure !&pa1(u,vbls);
   begin scalar x;
      return if atom u
               then if constantp u or u memq '(nil t) then mkquote u
                     else if u memq vbls then u
                     else if nonlocal u then u
                     else <<mknonlocal u; u>>
              else if not atom car u
               then !&pa1(car u,vbls) . !&palis(cdr u,vbls)
              else if x := get(car u,'pa1fn) then apply(x,list(u,vbls))
              else if x := get(car u,'cnewnam) then !&pa1(x . cdr u,vbls)
              else if (x := getd car u)
                        and car x eq 'macro
                        and not get(car u,'compfn)
               then !&pa1(apply(cdr x,list u),vbls)
              else if x and car x eq 'smacro and not get(car u,'compfn)
               then !&pa1(sublis(pair (caddr x,cdr u), cadddr x),vbls)
              else if x := get(car u,'cmacro)
               then !&pa1(sublis(pair(cadr x,cdr u),caddr x),vbls)
              else if !&cfntype car u eq 'fexpr
                        and not get(car u,'compfn)
               then list(car u,mkquote cdr u)
              else if getd car u		%There's another bug ???;
               then car u . !&palis(cdr u,vbls)	% do this. F.K 91-8-20;
              else if car u memq vbls or fluidp car u
               then list('apply,car u,!&palist(cdr u,vbls))
              else car u . !&palis(cdr u,vbls)
   end;

symbolic procedure !&paiden(u,vbls); u;

put('go,'pa1fn,'!&paiden);

put('quote,'pa1fn,'!&paiden);

put('code,'pa1fn,'!&paiden);

symbolic procedure !&pacond(u,vbls);
   'cond . for each z in cdr u
               collect list(!&pa1(car z,vbls),
                            !&pa1(!&mkprogn cdr z,vbls));

put('cond,'pa1fn,'!&pacond);

symbolic procedure !&pafunc(u,vbls);
   if atom cadr u then !&mkfunc cadr u
    else !&mkfunc compd(!&mknam name,'expr,cadr u);

put('function,'pa1fn,'!&pafunc);

symbolic procedure !&palamb(u,vbls);
   'lambda . list(cadr u,!&pa1(!&mkprogn cddr u,append(cadr u,vbls)));

put('lambda,'pa1fn,'!&palamb);

symbolic procedure !&palist(u,vbls); 'list . !&palis(u,vbls);

symbolic procedure !&paprog(u,vbls);
   'prog . (cadr u . !&paprog1(cddr u,append(cadr u,vbls)));

symbolic procedure !&paprog1(u,vbls);
   for each x in u collect if atom x then x else !&pa1(x,vbls);

put('prog,'pa1fn,'!&paprog);

symbolic procedure !&palis(u,vbls);
   for each x in u collect !&pa1(x,vbls);

symbolic procedure mknonlocal u;
   <<lprim list(u,"declared fluid"); fluid list u; list('fluid,u)>>;

symbolic procedure !&mknam u;
   %generates unique name for auxiliary function in U;
   intern strcat(u, gensym());

symbolic procedure !&mkprogn u;
   if null u or cdr u then 'progn . u else car u;

comment cmacro definitions for some functions;

comment we do not expand caaaar and similar functions, since fewer
        instructions are generated without open coding;

deflist('((caar (lambda (u) (car (car u))))
          (cadr (lambda (u) (car (cdr u))))
          (cdar (lambda (u) (cdr (car u))))
          (cddr (lambda (u) (cdr (cdr u))))
          (caaar (lambda (u) (car (car (car u)))))
          (caadr (lambda (u) (car (car (cdr u)))))
          (cadar (lambda (u) (car (cdr (car u)))))
          (caddr (lambda (u) (car (cdr (cdr u)))))
          (cdaar (lambda (u) (cdr (car (car u)))))
          (cdadr (lambda (u) (cdr (car (cdr u)))))
          (cddar (lambda (u) (cdr (cdr (car u)))))
          (cdddr (lambda (u) (cdr (cdr (cdr u)))))
          (not (lambda (u) (null u)))),'cmacro);

comment pass 2 of the compiler;

symbolic procedure !&pass2 exp; !&comval(exp,0);

symbolic procedure !&comval(exp,status);
   %computes code for value of EXP;
   if !&anyreg(exp,nil)
     then if status>1 then nil else !&lreg1(exp,status)
    else !&comval1(exp,stomap,status);

symbolic procedure !&comval1(exp,stomap,status);
   begin scalar x;
      if atom exp then if status<2 then !&lreg1(exp,status) else nil
       else if not atom car exp
        then if caar exp eq 'lambda
               then !&comply(car exp,cdr exp,status)
              else lprie list("INVALID FUNCTION",car exp)
       else if x := get(car exp,'compfn) then apply(x,list(exp,status))
       else if !*r2i and car exp eq name and status=0 and null freelst
        then !&comrec(exp,status)
       else if car exp eq 'lambda
        then lprie list("INVALID USE OF LAMBDA IN FUNCTION",name)
       else if car exp eq '!*code then !&attach exp
       else !&call(car exp,cdr exp,status);
      return nil
   end;

symbolic procedure !&anyreg(u,v);
   %determines if U can be loaded in any register;
   %!*ORD = T means force correct order, unless safe;
   if eqcar(u,'quote) then t
    else (atom u
          or idp car u and get(car u,'anyreg) and !&anyreg(cadr u,nil))
           and (null !*ord or !&anyregl v);

symbolic procedure !&anyregl u;
   null u or !&anyreg(car u,nil) and !&anyregl cdr u;

symbolic procedure !&call(fn,args,status);
   !&call1(fn,!&comlis args,status);

symbolic procedure !&call1(fn,args,status);
   %ARGS is reversed list of compiled arguments of FN;
   begin integer argno;
      argno := length args;
      !&loadargs(args,status);
      !&attach list('!*link,fn,!&cfntype fn,argno, 
			if status > 1 then nil else 1);
      if flagp(fn,'onereg) then regs := (1 . nil) . cdr regs
       else if flagp(fn,'tworeg)
        then regs := (1 . nil) . com!_delasc(2,cdr regs)
       else regs := list (1 . nil)
   end;

symbolic procedure com!_delasc(u,v);
   if null v then nil
    else if u=caar v then cdr v
    else car v . com!_delasc(u,cdr v);

symbolic procedure !&comlis exp;
   %returns reversed list of compiled arguments;
   begin scalar acused,y;
      while exp do
         <<if !&anyreg(car exp,cdr exp) then y := car exp . y
            else <<if acused then !&store1();
                   !&comval1(car exp,stomap,1);
                   acused := gensym();
                   regs := (1 . (acused . cdar regs)) . cdr regs;
                   y := acused . y>>;
           exp := cdr exp>>;
      return y
   end;

symbolic procedure !&store1; %Marks contents of register 1 for storage;
   begin scalar x;
      x := cadar regs;
      if null x or eqcar(x,'quote) then return nil
       else if not atsoc(x,stomap) then !&frame x;
      !&store0(x,1)
   end;

symbolic procedure !&comply(fn,args,status);
   begin scalar alsts,vars; integer i;
      vars := cadr fn;
      !&loadargs(!&comlis args,1);
      args := !&remvarl vars; % The stores that were protected;
      i := 1;
      for each v in vars do <<!&frame v;
                              regs := !&repasc(i,v,regs);
                              i := i + 1>>;
      alsts := !&freebind(vars,t); %Old fluid values saved;
      i := 1;
      for each v in vars do <<if not nonlocal v then !&store0(v,i);
                              i := i + 1>>;
      !&comval(caddr fn,status);
      !&freerst(alsts,status);
      !&rstvarl(vars,args)
   end;

symbolic procedure !&comrec(exp,status);
   begin scalar x,z;
      !&loadargs(!&comlis cdr exp,status);
      z := codelist;
      if null cdr z
        then lprie list("CIRCULAR DEFINITION FOR",car exp);
      while cddr z do z := cdr z;
      if caar z eq '!*lbl then x := cdar z
       else <<x := !&genlbl(); rplacd(z,list('!*lbl . x,cadr z))>>;
      !&attjmp x
   end;

symbolic procedure !&loadargs(args,status);
   begin integer n;
      n := length args;
      if n>maxnargs then lprie list("TOO MANY ARGUMENTS IN",name);
      if status>0 then !&clrregs();
      while args do
         <<!&lreg(n,car args,cdr args,status);
           n := n - 1;
           args := cdr args>>
   end;

symbolic procedure !&locate x;
   begin scalar y,vtype;
      if eqcar(x,'quote) then return list x
       else if y := !&rassoc(x,regs) then return list car y
       else if not atom x then return list (car x . !&locate cadr x)
       else if vtype := nonlocal x then return list list(vtype,x);
      while y := atsoc(x,slst) do slst := delete(y,slst);
      return if y := atsoc(x,stomap) then cdr y else list mknonlocal x
   end;

symbolic procedure !&lreg(reg,u,v,status);
   begin scalar x,y;
      if (x := assoc(reg,regs)) and u member cdr x then return nil
       else if (y := assoc(reg,iregs))
                 and (status>0 or !&memlis(cadr y,v))
        then <<!&store0(cadr y,reg); iregs := delete(y,iregs)>>;
      !&attach ('!*load . (reg . !&locate u));
      regs := !&repasc(reg,u,regs)
   end;

symbolic procedure !&lreg1(x,status); !&lreg(1,x,nil,status);

comment functions for handling non-local variables;

symbolic procedure !&freebind(vars,lambp);
   %bind FLUID variables in lambda or prog lists;
   %LAMBP is true for LAMBDA, false for PROG;
   begin scalar falst,fregs,x,y; integer i;
      i := 1;
      for each x in vars do <<if fluidp x
                                then <<%falst :=
                                       % (x . !&getffrm x) . falst;
                                       falst :=  x . falst;
                                       fregs := i . fregs>>
                               else if globalp x
                                then lprie list("CANNOT BIND GLOBAL ",
                                                x);
                              i := i + 1>>;
      if null falst then return nil;
	!&getffrm car falst;		%% dummy allocation of locals;;
      if lambp then !&attach list('!*lambind,fregs,falst)
       else !&attach list('!*progbind,falst);
      return falst
   end;

symbolic procedure !&freerst(alsts,status); %restores FLUID variables;
   if alsts then !&attach list('!*freerstr,alsts);

symbolic procedure !&attach u; codelist := u . codelist;

symbolic procedure !&store0(u,reg);
   %marks expression U in register REG for storage;
   begin scalar x;
      x := '!*store . (reg . !&getfrm u);
      stlst := x . stlst;
      !&attach x;
      if atom u
        then <<!&clrstr u; slst := (u . codelist) . slst>>
   end;

symbolic procedure !&clrstr var; %removes unneeded stores;
   begin scalar x;
      if condtail then return nil;
      x := atsoc(var,slst);
      if null x then return nil;
      stlst := !&deleq(cadr x,stlst);
      slst := !&deleq(x,slst);
      rplaca(cadr x,'!*noop)
   end;

comment functions for general tests;

symbolic procedure !&comtst(exp,labl);
   %compiles boolean expression EXP.
   %If EXP has the same value as !&SWITCH then branch to LABL,
   %otherwise fall through;
   %REGS/IREGS are active registers for fall through,
   %REGS1/IREGS1 for branch;
   begin scalar x;
      while eqcar(exp,'null) do
         <<!&switch := not !&switch; exp := cadr exp>>;
      if not atom exp and atom car exp and (x := get(car exp,'comtst))
        then apply(x,list(exp,labl))
       else <<if exp='(quote t)
                then if !&switch then !&attjmp labl else flagg := t
               else <<!&comval(exp,1);
                      !&attach list(if !&switch then '!*jumpt
                                     else '!*jumpnil,car labl,1);
                      !&addjmp codelist>>;
              regs1 := regs;
              iregs1 := iregs>>;
      if eqcar(car codelist,'!*jumpt)
        then regs := (1 . ('(quote nil) . cdar regs)) . cdr regs
       else if eqcar(car codelist,'!*jumpnil)
        then regs1 := (1 . ('(quote nil) . cdar regs1)) . cdr regs1
   end;

comment specific function open coding;

symbolic procedure !&comandor(exp,status);
   begin scalar fn,labl,iregsl,regsl;
      fn := car exp eq 'and;
      labl := !&genlbl();
      if status>1
        then begin scalar regs1;
                !&tstandor(exp,labl);
                regs := !&rmerge2(regs,regs1)
             end
       else begin
               if status>0 then !&clrregs();
               exp := cdr exp;
               while exp do
                  <<!&comval(car exp,if cdr exp then 1 else status);
                       %to allow for recursion on last entry;
                    iregsl := iregs . iregsl;
                    regsl := regs . regsl;
                    if cdr exp
                      then <<!&attach list(if fn then '!*jumpnil
                                            else '!*jumpt,car labl,1);
                             !&addjmp codelist>>;
                    exp := cdr exp>>;
               iregs := !&rmerge iregsl;
               regs := !&rmerge regsl
            end;
      !&attlbl labl
   end;

symbolic procedure !&tstandor(exp,labl);
   begin scalar flg,flg1,fn,lab2,regsl,regs1l,tailp;
      %FLG is initial switch condition;
      %FN is appropriate AND/OR case;
      %FLG1 determines appropriate switching state;
      flg := !&switch;
      !&switch := nil;
      fn := car exp eq 'and;
      flg1 := flg eq fn;
      exp := cdr exp;
      lab2 := !&genlbl();
      !&clrregs();
      while exp do
         <<!&switch := nil;
           if null cdr exp and flg1
             then <<if fn then !&switch := t;
                    !&comtst(car exp,labl);
                    regsl := regs . regsl;
                    regs1l := regs1 . regs1l>>
            else <<if not fn then !&switch := t;
                   if flg1
                     then <<!&comtst(car exp,lab2);
                            regsl := regs1 . regsl;
                            regs1l := regs . regs1l>>
                    else <<!&comtst(car exp,labl);
                           regsl := regs . regsl;
                           regs1l := regs1 . regs1l>>>>;
           if null tailp
             then <<condtail := nil . condtail; tailp := t>>;
           exp := cdr exp>>;
      !&attlbl lab2;
      regs := if not flg1 then car regsl else !&rmerge regsl;
      regs1 := if flg1 then car regs1l else !&rmerge regs1l;
      if tailp then condtail := cdr condtail;
      !&switch := flg
   end;

put('and,'compfn,'!&comandor);

put('or,'compfn,'!&comandor);

put('and,'comtst,'!&tstandor);

put('or,'comtst,'!&tstandor);

symbolic procedure !&comcond(exp,status);
   %compiles conditional expressions;
   %registers REGS and IREGS are set for dropping through,
   %REGS1 and IREGS1 are set for a branch;
   begin scalar iregs1,regs1,flagg,!&switch,lab1,lab2,regsl,iregsl,tailp;
      exp := cdr exp;
      lab1 := !&genlbl();
      if status>0 then !&clrregs();
      for each x in exp do <<lab2 := !&genlbl();
                             !&switch := nil;
                             if cdr x then !&comtst(car x,lab2)
                                         %update CONDTAIL;
                              else <<!&comval(car x,1);
                                     !&attach list('!*jumpnil,car lab2,1);
                                     !&addjmp codelist;
                                     iregs1 := iregs;
                                     regs1 := (1 . '(quote nil) .
                                                cdar regs) . cdr regs>>;
                             if null tailp
                               then <<condtail := nil . condtail;
                                      tailp := t>>;
                             !&comval(cadr x,status);
                                % Branch code;
                                %test if need jump to LAB1;
                             if not !&transferp car codelist
                               then <<!&attjmp lab1;
                                      iregsl := iregs . iregsl;
                                      regsl := regs . regsl>>;
                             regs := regs1;
            %restore register status for next iteration;
         iregs := iregs1;
         iregs1 := nil;
            %we do not need to set REGS1 to NIL since all !&COMTSTs
            %are required to set it;
         !&attlbl lab2>>;
      if null flagg and status<2
        then <<!&lreg1('(quote nil),status);
               iregs := !&rmerge1(iregs,iregsl);
               regs := !&rmerge1(regs,regsl)>>
       else if regsl
        then <<iregs := !&rmerge1(iregs,iregsl);
               regs := !&rmerge1(regs,regsl)>>;
      !&attlbl lab1;
      if tailp then condtail := cdr condtail
   end;

symbolic procedure !&rmerge u;
   if null u then nil else !&rmerge1(car u,cdr u);

symbolic procedure !&rmerge1(u,v);
   if null v then u else !&rmerge1(!&rmerge2(u,car v),cdr v);

symbolic procedure !&rmerge2(u,v);
   if null u or null v then nil
    else (lambda x;
             if x
               then (caar u . xn(cdar u,cdr x))
                      . !&rmerge2(cdr u,delete(x,v))
              else !&rmerge2(cdr u,v))
       assoc(caar u,v);

flag('(!*jump !*linke error),'transfer);

put('cond,'compfn,'!&comcond);

symbolic procedure !&comif(exp,status);
   %compiles conditional expressions;
   %registers REGS and IREGS are set for dropping through,
   %REGS1 and IREGS1 are set for a branch;
   begin scalar iregs1,regs1,flagg,!&switch,lab1,lab2,regsl,iregsl,tailp, x;
%print list("comif", exp);
      exp := cdr exp;
      lab1 := !&genlbl();
      if status>0 then !&clrregs();
	 x := car exp;
	 lab2:=!&genlbl();
	 !&switch :=nil;
	 if cadr exp then !&comtst(x, lab2)
           else <<!&comval(x, 1);
                  !&attach list('!*jumpnil, car lab2, 1);
                  !&addjmp codelist;
                  iregs1 := iregs;
                  regs1 := (1 . '(quote nil) . cdar regs) . cdr regs>>;
         if null tailp then <<condtail := nil . condtail;
                              tailp:=t>>;
         !&comval(cadr exp, status);
         if not !&transferp car codelist
           then <<!&attjmp lab1;
                  iregsl := iregs . iregsl;
                  regsl := regs . regsl>>;
         regs := regs1;
         iregs := iregs1;
         iregs1 := nil;
         !&attlbl lab2;
       x:=caddr exp;
       if x then <<
           !&switch:=nil;
           if null tailp then <<condtail := nil . condtail; tailp:=t>>;
           !&comval(x, status);
           if not !&transferp car codelist
             then <<!&attjmp lab1;
                    iregsl := iregs . iregsl;
                    regsl := regs . regsl>>;
           iregs := iregs1;
           iregs1 := nil>>;
      
      if null flagg and status<2
        then <<!&lreg1('(quote nil),status);
               iregs := !&rmerge1(iregs,iregsl);
               regs := !&rmerge1(regs,regsl)>>
       else if regsl
        then <<iregs := !&rmerge1(iregs,iregsl);
               regs := !&rmerge1(regs,regsl)>>;
      !&attlbl lab1;
      if tailp then condtail := cdr condtail
   end;

put('if,'compfn,'!&comif);

symbolic procedure !&comcons(exp,status);
   if null (exp := cdr exp) or null cdr exp or cddr exp
     then lprie "MISMATCH OF ARGUMENTS"
    else if cadr exp='(quote nil)
     then !&call('ncons,list car exp,status)
    else if eqcar(!&rassoc(cadr exp,regs),1)
        and !&anyreg(car exp,nil)
     then !&call1('xcons,!&comlis reverse exp,status)
    else if !&anyreg(cadr exp,nil) then !&call('cons,exp,status)
    else !&call1('xcons,reversip !&comlis exp,status);

put('cons,'compfn,'!&comcons);

symbolic procedure !&comgo(exp,status);
   <<!&clrregs();
     if status>2 then <<!&attjmp !&getlbl cadr exp; slst := nil>>
      else lprie list(exp,"INVALID")>>;

put('go,'compfn,'!&comgo);

symbolic procedure !&comlist(exp,status);
   %we only support explicit functions up to 5 arguments here;
   begin scalar m,n,fn;
      exp := cdr exp;
      m := min(maxnargs,5);
      n := length exp;
      if n=0 then !&lreg1('(quote nil),status)
       else if n>m then !&comval(!&comlist1 exp,status)
       else !&call(if n=1 then 'ncons
                    else if n=2 then 'list2
                    else if n=3 then 'list3
                    else if n=4 then 'list4
                    else 'list5,exp,status)
   end;


symbolic procedure !&comlist1 exp;
   if null exp then '(quote nil)
    else list('cons,car exp,'list . cdr exp);

put('list,'compfn,'!&comlist);

symbolic procedure !&pamap(u,vars);
   if eqcar(caddr u,'function)
     then (lambda x;
              list(car u,
                   !&pa1(cadr u,vars),
                   mkquote (if atom x then x else !&pa1(x,vars))))
       cadr caddr u
    else car u . !&palis(cdr u,vars);

put('map,'pa1fn,'!&pamap);

put('mapc,'pa1fn,'!&pamap);

put('mapcan,'pa1fn,'!&pamap);

put('mapcar,'pa1fn,'!&pamap);

put('mapcon,'pa1fn,'!&pamap);

put('maplist,'pa1fn,'!&pamap);

symbolic procedure !&commap(exp,status);
   begin scalar body,fn,lab1,lab2,lab3,tmp,mtype,result,slst1,var,x;
      body := cadr exp;
      fn := caddr exp;
      lab1 := !&genlbl();
      lab2 := !&genlbl();
      mtype :=
       if car exp memq '(mapcar maplist) then 'cons
        else if car exp memq '(mapcan mapcon)
               then <<lab3 := !&genlbl(); 'nconc>>
        else nil;
      !&clrregs();
      if mtype then <<!&frame (result := gensym());
                      if null lab3 then !&store0(result,nil)>>;
      !&frame (var := gensym());
      !&comval(body,1);
      regs := list list(1,var);
      if lab3 then <<!&store0(var,1); !&frame (tmp := gensym());
                     !&comval('(ncons 'nil),1);
                     !&store0(result,1); !&store0(tmp,1);
                     !&lreg1(var,1)>>;
      !&attjmp lab2;
      !&attlbl lab1;
      !&store0(var,1);
      x := if car exp memq '(map mapcon maplist) then var
            else list('car,var);
      if eqcar(fn,'quote) then fn := cadr fn
	else if atom fn then 
		fn := list('lambda, '(x), list('funcall1, fn, 'x));
      slst1 := slst; %to allow for store in function body;
      !&comval(list(fn,x),if mtype then 1 else 3);
      if mtype
        then <<if lab3 then <<!&attach list('!*jumpnil,car lab3,1);
                              !&addjmp codelist;
                              !&attach '(!*load 2 1);
                              !&lreg1(tmp,1);
                              !&store0(tmp,2);
                              !&attach '(!*link nconc expr 2 1);
                              !&attlbl lab3>>
                else <<!&lreg(2,result,nil,1);
                       !&attach '(!*link cons expr 2 1);
                       !&store0(result,1)>>;
               regs := list (1 . nil)>>;
      slst := xn(slst,slst1);
      !&comval(list('cdr,var),1);
      !&attlbl lab2;
      !&attach list('!*jumpt,car lab1,1);
      !&addjmp codelist;
      if mtype
        then !&comval(list(if lab3 then 'cdr else 'reversip,result),1)
       else regs := list list(1,mkquote nil)
   end;


put('map,'compfn,'!&commap);

put('mapc,'compfn,'!&commap);

put('mapcan,'compfn,'!&commap);

put('mapcar,'compfn,'!&commap);

put('mapcon,'compfn,'!&commap);

put('maplist,'compfn,'!&commap);

symbolic procedure !&comprog(exp,status); %compiles program blocks;
   begin scalar alsts,golist,pg,proglis,exit; integer i;
      proglis := cadr exp;
      exp := cddr exp;
      exit := !&genlbl();
      pg := !&remvarl proglis; %protect prog variables;
      for each x in proglis do !&frame x;
      alsts := !&freebind(proglis,nil);
      for each x in proglis do if not nonlocal x then !&store0(x,nil);
      for each x in exp do if atom x
                             then golist := (x . !&genlbl()) . golist;
      while exp do
         <<if atom car exp
             then <<!&clrregs();
                    !&attlbl !&getlbl car exp;
                    regs := list (1 . nil)>>
            else !&comval(car exp,if status>2 then 4 else 3);
           if null cdr exp
                and status<2
                and (atom(car(exp)) or not (caar(exp) memq '(go return)))
             then exp := list '(return (quote nil))
            else exp := cdr exp>>;
      !&attlbl exit;
      if cdr !&findlbl exit then regs := list (1 . nil);
      !&freerst(alsts,status);
      !&rstvarl(proglis,pg)
   end;

put('prog,'compfn,'!&comprog);

symbolic procedure !&remvarl vars;
   for each x in vars collect !&remvar x;

symbolic procedure !&remvar x;
   %removes references to variable X from IREGS and REGS
   %and protects SLST;
   <<!&remstores x; !&protect x>>;

symbolic procedure !&remstores x;
   begin
      for each y in iregs do if x eq cadr y
                               then <<!&store0(cadr y,car y);
                                      iregs := delete(y,iregs)>>;
      for each y in regs do while x member cdr y do
                               rplacd(y,!&deleq(x,cdr y))
   end;

symbolic procedure !&protect u;
   begin scalar x,y;
      while x := assoc(u,slst)
         do <<slst := !&deleq(x,slst); y := x . y>>;
      return y
   end;

symbolic procedure !&rstvarl(vars,lst);
   for each x in vars do
     <<!&remstores x; !&clrstr x; !&unprotect car lst; lst := cdr lst>>;

symbolic procedure !&unprotect val; %restores VAL to SLST;
   for each v in val do slst := v . slst;

symbolic procedure !&comprogn(exp,status);
   begin
      exp := cdr exp;
      if null exp then return nil;
      while cdr exp do
         <<!&comval(car exp,if status<2 then 2 else status);
           exp := cdr exp>>;
      !&comval(car exp,status)
   end;

put('prog2,'compfn,'!&comprogn);

%put('unwind!-protect,'compfn,'!&comprogn);

put('progn,'compfn,'!&comprogn);

symbolic procedure !&comreturn(exp,status);
   <<if status<4 or not !&anyreg(cadr exp,nil)
       then !&lreg1(car !&comlis list cadr exp,status);
     slst := nil;   % Since we don't keep track of which variables
                    % are bound in this block.
     !&attjmp exit>>;

put('return,'compfn,'!&comreturn);

symbolic procedure !&comsetq(exp,status);
   begin scalar x;
      exp := cdr exp;
      if status>1 and (null cadr exp or cadr exp='(quote nil))
        then !&store2(car exp,nil)
       else <<!&comval(cadr exp,1);
              !&store2(car exp,1);
              if x := !&rassoc(car exp,iregs)
                then iregs := delete(x,iregs);
              regs := (1 . (car exp . cdar regs)) . cdr regs>>
   end;

symbolic procedure !&remsetvar(u,v);
   %removes references to SETQ variable U from regs list V;
   if null v then nil
    else (caar v . !&rems1(u,cdar v)) . !&remsetvar(u,cdr v);

symbolic procedure !&rems1(u,v);
   %removes references to SETQ variable U from list V;
   if null v then nil
    else if smemq(u,car v) then !&rems1(u,cdr v)
    else car v . !&rems1(u,cdr v);

%symbolic procedure smemq(u,v);
%   %true if atom U is a member of V at any level (excluding
%   %quoted expressions);
%   if atom v then u eq v
%    else if car v eq 'quote then nil
%    else smemq(u,car v) or smemq(u,cdr v);

symbolic procedure !&store2(u,v);
   begin scalar vtype;
      regs := !&remsetvar(u,regs);
      if vtype := nonlocal u
        then !&attach list('!*store,v,list(vtype,u))
       else if not atsoc(u,stomap)
        then !&attach list('!*store,v,mknonlocal u)
       else !&store0(u,v)
   end;

put('setq,'compfn,'!&comsetq);
put('!:!=,'compfn,'!&comsetq);

comment specific test open coding;

symbolic procedure !&comeq(exp,labl);
   begin scalar u,v,w;
      u := cadr exp;
      v := caddr exp;
      if u member cdar regs then w := !&comeq1(v,u)
       else if v member cdar regs then w := !&comeq1(u,v)
       else if !&anyreg(v,nil) then <<!&comval(u,1); w := !&locate v>>
       else if !&anyreg(u,list v)
        then <<!&comval(v,1); w := !&locate u>>
       else <<u := !&comlis cdr exp; w := !&locate cadr u>>;
      !&attach ((if !&switch then '!*jumpe else '!*jumpn)
                    . (car labl . (1 . w)));
      iregs1 := iregs;
      regs1 := regs;
      !&addjmp codelist
   end;

symbolic procedure !&comeq1(u,v);
   if !&anyreg(u,list v) then !&locate u
    else <<!&comval(u,1); !&locate v>>;

put('eq,'comtst,'!&comeq);

symbolic procedure !&comneq(exp,labl);
   begin scalar u,v,w;
      u := cadr exp;
      v := caddr exp;
      if u member cdar regs then w := !&comeq1(v,u)
       else if v member cdar regs then w := !&comeq1(u,v)
       else if !&anyreg(v,nil) then <<!&comval(u,1); w := !&locate v>>
       else if !&anyreg(u,list v)
        then <<!&comval(v,1); w := !&locate u>>
       else <<u := !&comlis cdr exp; w := !&locate cadr u>>;
      !&attach ((if !&switch then '!*jumpn else '!*jumpe)
                    . (car labl . (1 . w)));
      iregs1 := iregs;
      regs1 := regs;
      !&addjmp codelist
   end;

%put('neq,'comtst,'!&comneq);


symbolic procedure !&comilessp(exp,labl);
   begin scalar u,v,w;
      u := !&comlis cdr exp;
      w := !&locate car u;
      v := !&locate cadr u;
      !&attach ((if !&switch then '!*jumplt else '!*jumpge)
                    . (car labl . (car v . w)));
      iregs1 := iregs;
      regs1 := regs;
      !&addjmp codelist
   end;
       
put('ilessp,'comtst,'!&comilessp);


symbolic procedure !&testfn(exp,labl);
   %generates c-macros !*JUMPC and !*JUMPNC;
   begin scalar x;
      if not (x := !&rassoc(cadr exp,regs)) then !&comval(cadr exp,1);
      !&clrregs();
      !&attach list(if !&switch then '!*jumpc else '!*jumpnc,
                    car labl,
                    if x then car x else 1,car exp);
      regs1 := regs;
      !&addjmp codelist
   end;

comment support functions;

symbolic procedure !&memlis(u,v);
   v and (!&memb(u,car v) or !&memlis(u,cdr v));

symbolic procedure !&memb(u,v);
   if atom v then u eq v else !&memb(u,cadr v);

symbolic procedure !&rassoc(u,v);
   if null v then nil
    else if u member cdar v then car v
    else !&rassoc(u,cdr v);

symbolic procedure !&repasc(reg,u,v);
   if null v then list list(reg,u)
    else if reg=caar v then list(reg,u) . cdr v
    else car v . !&repasc(reg,u,cdr v);

symbolic procedure !&clrregs; %store deferred values in IREGS;
   while iregs do
      <<!&store0(cadar iregs,caar iregs); iregs := cdr iregs>>;

symbolic procedure !&cfntype fn;
   begin scalar x;
      return if not atom fn then 'expr
              else if x := get(fn,'cfntype) then car x
              else if x := getd fn then car x
              else 'expr
   end;

symbolic procedure !&genlbl;
   begin scalar l;
      l := gensym();
      lblist := list l . lblist;
      return list l
   end;

symbolic procedure !&getlbl labl;
   begin scalar x;
      x := atsoc(labl,golist);
      if null x then lprie list(labl," - MISSING LABEL -");
      return cdr x
   end;

symbolic procedure !&findlbl lblst; assoc(car lblst,lblist);

symbolic procedure !&rechain(olbl,nlbl);
   % Fix OLBL to now point at NLBL;
   begin scalar x,y,uses;
      x := !&findlbl olbl;
      y := !&findlbl nlbl;
      rplaca(olbl,car nlbl); % FIX L VAR;
      uses := cdr x; % OLD USES;
      rplacd(x,nil);
      rplacd(y,append(uses,cdr y));
      for each x in uses do rplaca(cdr x,car nlbl)
   end;

symbolic procedure !&moveup u;
   if caadr u eq '!*jump
     then <<jmplist := !&deleq(cdr u,jmplist);
            rplacw(u,cdr u);
            jmplist := u . jmplist>>
    else rplacw(u,cdr u);

symbolic procedure !&attlbl lbl;
   if caar codelist eq '!*lbl then !&rechain(lbl,cdar codelist)
    else !&attach ('!*lbl . lbl);

symbolic procedure !&attjmp lbl;
   begin
      if caar codelist eq '!*lbl
        then <<!&rechain(cdar codelist,lbl);
               codelist := cdr codelist>>;
      if !&transferp car codelist then return nil;
      !&attach ('!*jump . lbl);
      !&addjmp codelist
   end;

symbolic procedure !&transferp x;
   flagp(if car x eq '!*link then cadr x else car x,'transfer);

symbolic procedure !&addjmp clist;
   begin scalar x;
      x := !&findlbl cdar clist;
      rplacd(x,car clist . cdr x);
      jmplist := clist . jmplist
   end;

symbolic procedure !&remjmp clist;
   begin scalar x;
      x := !&findlbl cdar clist;
      rplacd(x,!&deleq(car clist,cdr x));
      jmplist := !&deleq(clist,jmplist);
      !&moveup clist
   end;

symbolic procedure !&deleq(u,v);
   if null v then nil
    else if u eq car v then cdr v
    else car v . !&deleq(u,cdr v);

symbolic procedure !&frame u; %allocates space for U in frame;
   begin scalar z;
      stomap := list(u,z := cadar stomap - 1) . stomap;
      if z<car llngth then rplaca(llngth,z)
   end;

symbolic procedure !&getfrm u;
   (lambda x;
       if x then cdr x else lprie list("COMPILER ERROR: LOST VAR",u))
    atsoc(u,stomap);

symbolic procedure !&getffrm u;
   begin scalar x; x := !&getfrm u; freelst := x . freelst; return x
   end;

comment pass 3 of the compiler (post code generation fixups);

symbolic procedure !&pass3;
   begin scalar flagg; %remove spurious stores;
      for each j in slst do <<stlst := !&deleq(cadr j,stlst);
                              rplaca(cadr j,'!*noop)>>;
      !&fixchains();
      !&fixlinks();
      !&fixfrm();
      !&attlbl exit;
      if flagg
        then <<if not !*nolinke
                    and caar codelist eq '!*lbl
                    and caadr codelist eq '!*linke
                 then rplaca(cdr codelist,
                             list('!*link,cadadr codelist,
                                  cadr cdadr codelist,
                                  caddr cdadr codelist,
				  cadddr cdadr codelist));
                  %removes unnecessary !*LINKE;
               !&attach ('!*dealloc . llngth);
               !&attach list '!*exit>>;
      !&peepholeopt();
      !&fixrest()
   end;

symbolic procedure !&fixchains;
   begin scalar ejmps,ejmps1,p,q; %find any common chains of code;
      if not (car codelist='!*lbl . exit) then !&attlbl exit;
      codelist := cdr codelist;
      if not (car codelist='!*jump . exit) then !&attjmp exit;
      ejmps := reverse jmplist;
      while ejmps do
         begin
            p := car ejmps;
            ejmps := cdr ejmps;
            if caar p eq '!*jump
              then <<ejmps1 := ejmps;
                     while ejmps1 do
                        if car p=caar ejmps1 and cadr p=cadar ejmps1
                          then <<!&remjmp p;
                                 !&fixchn(p,cdar ejmps1);
                                 ejmps1 := nil>>
                         else ejmps1 := cdr ejmps1>>
         end
   end;

symbolic procedure !&fixlinks;
   %replace !*LINK by !*LINKE where appropriate;
   begin scalar ejmps,p,q;
      ejmps := jmplist;
      if not !*nolinke
        then while ejmps do
                begin
                   p := car ejmps;
                   q := cdr p;
                   ejmps := cdr ejmps;
		% there are compiler bugs; may be fixchain ???;
		% currently patch here, hope it works. F.K 91-8-20;
                   if caar p eq '!*lbl and cadar q eq car exit
                       then return flagg := t;
                   if not (cadar p eq car exit) then return nil
                    else if not (caar p eq '!*jump)
                              or not (caar q eq '!*link)
                     then return flagg := t;
                   rplacw(car q,
                          '!*linke
                            . (cadar q
                                 . (caddar q
                                      . (cadr cddar q 
					  . (caddr cddar q . llngth)))));
                   !&remjmp p
                end
       else flagg := t
   end;

symbolic procedure !&findblk(u,lbl);
   if null cdr u then nil
    else if caadr u eq '!*lbl and !&transferp caddr u then u
    else if get(caadr u,'negjmp) and cadadr u eq lbl then u
    else !&findblk(cdr u,lbl);

put('!*noop,'optfn,'!&moveup);

put('!*lbl,'optfn,'!&lblopt);

symbolic procedure !&lblopt u;
   begin scalar z;
      if cadar u eq cadadr u then return !&remjmp cdr u
       else if caadr u eq '!*jump
                 and (z := get(caaddr u,'negjmp))
                 and cadar u eq cadr caddr u
        then return <<z := z . (cadadr u . cddr caddr u);
                      !&remjmp cdr u;
                      !&remjmp cdr u;
                      rplacd(u,z . (cadr u . cddr u));
                      !&addjmp cdr u;
                      t>>
       else return nil
   end;

symbolic procedure !&peepholeopt;
   %'peep-hole' optimization for various cases;
   begin scalar x,z;
      z := codelist;
      while z do
         if not (x := get(caar z,'optfn)) or not apply(x,list z)
           then z := cdr z
   end;

symbolic procedure !&fixrest;
   %checks for various cases involving unique (and unused) labels
   %and sequences like (JUMPx lab) M1 ... Mn ... (LAB lab) M1 ... Mn
   %where Mi do not affect reg 1;
   begin scalar labs,tlabs,x,y,z,nn;
      while codelist do
         <<if caar codelist eq '!*lbl
             then <<!&lblopt codelist;
                    if cdr (z := !&findlbl cdar codelist)
                      then <<y := car codelist . y;
                             if null cddr z
                                  and !&transferp cadr z
                                  and caadr y eq '!*load
                                  and !&noloadp(cdadr y,
                                                cdr atsoc(cadr z,
                                                          jmplist))
                               then <<if
                                          not !&noloadp(cdadr y,
                                                        cdr codelist)
                                        then rplacw(cdr codelist,
                                                    cadr y
                                                    . cadr codelist
                                                      . cddr codelist);
                                      rplacw(cdr y,cddr y)>>
                              else <<if null cddr z
                                          and caadr codelist eq '!*jump
                                          and get(caadr z,'negjmp)
                                       then labs :=
                                             (cadr z . y) . labs;
                                     if !&transferp cadr codelist
                                       then tlabs :=
                                             (cadar y . y)
                                               . tlabs>>>>>>
            else if get(caar codelist,'negjmp)
                      and (z := atsoc(car codelist,labs))
               then if (nn:=cddar codelist) then
                    <<nn:=car nn;
                      while not atom nn and
                         (car nn eq 'car or car nn eq 'cdr) do
                        nn := cadr nn;
                      x := car codelist;
                      codelist := cdr codelist;
                      z := cddr z;
                      while car y= car z
                          and ((caar y eq '!*store
                                and not (caddar y = 1)
                                and not (caddar y = nn))
                              or (caar y eq '!*load
                                  and not (cadar y = 1)
                                  and not (cadar y = nn))) do
                           <<codelist := car y . codelist;
                             rplacw(z,cadr z . cddr z);
                             y := cdr y>>;
                      codelist := x . codelist;
                      y := x . y>>
              else <<x := car codelist;
                     codelist := cdr codelist;
                     z := cddr z;
                     while car y=car z
                            and ((caar y eq '!*store
                                  and not (caddar y = 1))
                                   or (caar y eq '!*load
                                        and not (cadar y=1))) do
                       <<codelist := car y . codelist;
                         rplacw(z,cadr z . cddr z);
                         y := cdr y>>;
                    codelist := x . codelist;
                    y := x . y>>
            else if caar codelist eq '!*jump
                      and (z := atsoc(cadar codelist,tlabs))
                      and (x :=
                            !&findblk(cdr codelist,
                                      if caar y eq '!*lbl then cadar y
                                       else nil))
             then begin scalar w;
                     if not (caadr x eq '!*lbl)
                       then <<if not (caar x eq '!*lbl)
                                then x :=
                                      cdr rplacd(x,
                                                 ('!*lbl . !&genlbl())
                                                   . cdr x);
                              w :=
                               get(caadr x,'negjmp)
                                 . (cadar x . cddadr x);
                              !&remjmp cdr x;
                              rplacd(x,w . (cadr x . cddr x));
                              !&addjmp cdr x>>
                      else x := cdr x;
                     w := nil;
                     repeat <<w := car y . w; y := cdr y>>
                        until y eq cdr z;
                     rplacd(x,nconc(w,cdr x));
                     !&remjmp codelist;
                     tlabs := nil; %since code chains have changed;
                     codelist := nil . (car y . codelist);
                     y := cdr y
                  end
            else y := car codelist . y;
           codelist := cdr codelist>>;
      codelist := y
   end;

symbolic procedure !&noloadp(args,instrs);
   %determines if a LOAD is not necessary in instruction stream;
   atom cadr args
     and (caar instrs eq '!*load and cdar instrs=args
            or caar instrs eq '!*store
                 and (cdar instrs=args
                        or not (caddar instrs=cadr args)
                             and !&noloadp(args,cdr instrs)));

symbolic procedure !&fixchn(u,v);
   begin scalar x;
      while car u=car v and cdr u do <<!&moveup u; v := cdr v>>;
      % CDR U check needed for case
      % (prog nil (print a) (go b) (print a) (go b) b)
      x := !&genlbl();
      if caar v eq '!*lbl then !&rechain(x,cdar v)
       else rplacw(v,('!*lbl . x) . (car v . cdr v));
      if caar u eq '!*lbl then <<!&rechain(cdar u,x); !&moveup u>>;
      if caar u eq '!*jump then return nil;
      rplacw(u,('!*jump . x) . (car u . cdr u));
      !&addjmp u
   end;

symbolic procedure !&fixfrm;
   begin scalar holes,lst,x,y,z; integer n;
      if null stlst and null freelst then return rplaca(llngth,1);
      n := 0;
      while not (n<car llngth) do
         <<y := nil;
           for each lst in stlst do if n=caddr lst
                                      then y := cddr lst . y;
           for each lst in freelst do if n=car lst then y := lst . y;
           if null y then holes := n . holes else z := (n . y) . z;
           n := n - 1>>;
      y := z;
      if caar z>car llngth then rplaca(llngth,caar z);
      while holes do
         <<while holes and car holes<car llngth do holes := cdr holes;
           if holes
             then <<holes := reversip holes;
                    for each x in cdar z do rplaca(x,car holes);
                    rplaca(llngth,
                           if null cdr z or car holes<caadr z
                             then car holes
                            else caadr z);
                    holes := reversip cdr holes;
                    z := cdr z>>>>;
      %now see if we can map frame to registers;
      n := if narg<3 then 3 else narg + 1;
      if freelst or null !&regp codelist or car llngth<n - maxnargs
        then return nil;
      for each x in stlst do rplacw(x,
                                    list('!*load,
                                         n - caddr x,
                                         if null cadr x
                                           then '(quote nil)
                                          else cadr x));
      while y do
         <<for each x in cdar y do not (car x>0)
                                     and rplaca(x,n - car x);
              %first test makes sure replacement only occurs once;
           y := cdr y>>;
      rplaca(llngth,1)
   end;

symbolic procedure !&regp u;
   %there is no test for !*LAMBIND/!*PROGBIND
   %since FREELST tested explicitly in !&FIXFRM;
   if null cdr u then t
    else if caar u memq '(!*load !*store)
          and numberp cadar u and cadar u>2
     then nil
    else if flagp(caadr u,'unknownuse)
              and
                   not (idp cadadr u
                          and (flagp(cadadr u,'onereg)
                                 or flagp(cadadr u,'tworeg))
                          or car u='!*jump . exit)
     then nil
    else !&regp cdr u;

flag('(!*code !*link !*linke),'unknownuse);

symbolic procedure !*code u;  eval u;

put('!*jumpn,'negjmp,'!*jumpe);

put('!*jumpe,'negjmp,'!*jumpn);

put('!*jumpnil,'negjmp,'!*jumpt);

put('!*jumpt,'negjmp,'!*jumpnil);

put('!*jumpc,'negjmp,'!*jumpnc);

put('!*jumpnc,'negjmp,'!*jumpc);

put('!*jumpge,'negjmp,'!*jumplt);
put('!*jumple,'negjmp,'!*jumpgt);
put('!*jumpgt,'negjmp,'!*jumple);
put('!*jumplt,'negjmp,'!*jumpge);

comment some arithmetic optimizations to reduce the amount of code
        generated;


% fns.c:    {"logxor", logxor_fn, -1},
% fns.c:    {"logand", logand_fn, -1},
% fns.c:    {"logor", logor_fn, -1},
% fns.c:    {"ilogxor", ilogxor_fn, -1},
% fns.c:    {"ilogand", ilogand_fn, -1},
% fns.c:    {"ilogor", ilogor_fn, -1},
% fns.c:    {"max", max_fn, -1},
% fns.c:    {"min", min_fn, -1},
% fns.c:    {"imax", imax_fn, -1},
% fns.c:    {"imin", imin_fn, -1},

symbolic procedure !&paplus(u, vars);
   !&pa1(expand(cdr u, 'plus2), vars);

put('plus,'pa1fn,'!&paplus);
put('!+,'pa1fn,'!&paplus);

symbolic procedure !&patimes(u, vars);
   !&pa1(expand(cdr u, 'times2), vars);

put('times,'pa1fn,'!&patimes);
put('!*,'pa1fn,'!&patimes);

symbolic procedure !&pamax(u, vars);
   !&pa1(expand(cdr u, 'max2), vars);

put('max,'pa1fn,'!&pamax);

symbolic procedure !&pamin(u, vars);
   !&pa1(expand(cdr u, 'min2), vars);

put('min,'pa1fn,'!&pamin);

symbolic procedure !&paimax(u, vars);
   !&pa1(expand(cdr u, 'imax2), vars);

put('imax,'pa1fn,'!&paimax);

symbolic procedure !&palogand(u, vars);
   !&pa1(expand(cdr u, 'logand2), vars);

put('logand,'pa1fn,'!&palogand);

symbolic procedure !&palogor(u, vars);
   !&pa1(expand(cdr u, 'logor2), vars);

put('logor,'pa1fn,'!&palogor);

symbolic procedure !&palogxor(u, vars);
   !&pa1(expand(cdr u, 'logxor), vars);

put('logxor,'pa1fn,'!&palogxor);

symbolic procedure !&pailogand(u, vars);
   !&pa1(expand(cdr u, 'ilogand2), vars);

put('ilogand,'pa1fn,'!&pailogand);

symbolic procedure !&pailogor(u, vars);
   !&pa1(expand(cdr u, 'ilogor2), vars);

put('ilogor,'pa1fn,'!&pailogor);

symbolic procedure !&pailogxor(u, vars);
   !&pa1(expand(cdr u, 'ilogxor2), vars);

put('ilogxor,'pa1fn,'!&pailogxor);

symbolic procedure !&paplus2(u,vars);
   if caddr u=1 then list('add1,!&pa1(cadr u,vars))
    else if cadr u=1 then list('add1,!&pa1(caddr u,vars))
    else 'plus2 . !&palis(cdr u,vars);

put('plus2,'pa1fn,'!&paplus2);

symbolic procedure !&padiff(u,vars);
   if caddr u=1 then list('sub1,!&pa1(cadr u,vars))
    else 'difference . !&palis(cdr u,vars);

put('difference,'pa1fn,'!&padiff);

symbolic procedure !&palessp(u,vars);
   if caddr u=0 then list('minusp,!&pa1(cadr u,vars))
    else 'lessp . !&palis(cdr u,vars);

put('lessp,'pa1fn,'!&palessp);

symbolic procedure !&paiplus2(u,vars);
   if caddr u=1 then list('iadd1,!&pa1(cadr u,vars))
    else if cadr u=1 then list('iadd1,!&pa1(caddr u,vars))
    else 'iplus2 . !&palis(cdr u,vars);

put('iplus2,'pa1fn,'!&paiplus2);

symbolic procedure !&paidiff(u,vars);
   if caddr u=1 then list('isub1,!&pa1(cadr u,vars))
    else 'idifference . !&palis(cdr u,vars);

put('idifference,'pa1fn,'!&paidiff);

symbolic procedure !&pailessp(u,vars);
   if caddr u=0 then list('minusp,!&pa1(cadr u,vars))
    else 'ilessp . !&palis(cdr u,vars);

put('ilessp,'pa1fn,'!&pailessp);

comment removing unnecessary calls to minus;

symbolic procedure !&paminus(u,vars);
   if eqcar(u := !&pa1(cadr u,vars),'quote) and numberp cadr u
     then mkquote ( - cadr u)
    else list('minus,u);

put('minus,'pa1fn,'!&paminus);
put('!-,'pa1fn,'!&paminus);

symbolic procedure !&paiminus(u,vars);
   if eqcar(u := !&pa1(cadr u,vars),'quote) and numberp cadr u
     then mkquote ( - cadr u)
    else list('iminus,u);

put('iminus,'pa1fn,'!&paiminus);

%        optimization of equal, member and assoc;
%                        by f.kako        85-10-17;

symbolic procedure !&paequal(u,vars);
   begin scalar x,y;
        x := cadr u;
        y := caddr u;
        return if !&eqvp x or !&eqvp y then 'eq . !&palis(cdr u, vars)
                else if x = 0 then list('zerop,!&pa1(y,vars))
                else if y = 0 then list('zerop,!&pa1(x,vars))
                else if x = 1 then list('onep,!&pa1(y,vars))
                else if y = 1 then list('onep,!&pa1(x,vars))
                else 'equal . !&palis(cdr u, vars)

  end;

symbolic procedure !&eqvp u;
   pairp u and eqcar(u,'quote) and idp cadr u;

put('equal,'pa1fn,'!&paequal);
put('!=,'pa1fn,'!&paequal);

symbolic procedure !&paneq(u,vars);
   list('null ,!&paequal(u, vars));

put('neq, 'pa1fn, '!&paneq);

symbolic procedure !&pamember(u,vars);
   (lambda(x,y);
        if !&eqvp x then 'memq
         else if not eqcar(y,'quote) then 'member
         else begin scalar a;
                 a := y := cadr y;
                 while y and a do <<a := idp car y; y :=  cdr y>>;
                 return if a then 'memq else 'member
              end)
         (cadr u, caddr u)
        . !&palis(cdr u, vars);

put('member,'pa1fn,'!&pamember);

symbolic procedure !&paassoc(u,vars);
   (lambda(x,y);
        if !&eqvp x then 'atsoc
         else if not eqcar(y,'quote) then 'assoc
         else begin scalar a;
                 a := t;
                 y := cadr y;
                 while y and a do <<a := idp caar y; y := cdr y>>;
                 return if a then 'atsoc else 'assoc
              end)
           (cadr u, caddr u)
        . !&palis(cdr u, vars);

put('assoc,'pa1fn,'!&paassoc);


%*******************************************************************;
%								    ;
%	compiler optimization for slisp/68000			    ;
%                                                                   ;
%                                                                   ;
%*******************************************************************;

%symbolic procedure !&palength(u,vars);
% optimize sequence of
%	LENGTH EXPLODE U;
%	to strlen u;
%   begin scalar x;
%     x := cadr u;
%     return if not atom x and car x eq 'explode
%	      then list('strlen,!&pa1(cadr x,vars))
%             else list('length, !&pa1(x,vars))
%   end;

%put('length, 'pa1fn, '!&palength);

symbolic procedure !&pacompress(u,vars);
% optimize sequence of
%	COMPRESS APPEND(EXPLODE X, EXPLODE Y) to strcat(x,y); 
%	COMPRESS NCONC(EXPLODE X, EXPLODE Y) to strcat(x,y);
%	COMPRESS EXPLODE x to sprin2 x;
   begin scalar x;
     x := cadr u;
     return if not atom x and (car x eq 'append or car x eq 'nconc)
	          and eqcar(cadr x,'explode) and eqcar(caddr x,'explode)
	      then list('strcat,!&pa1(cadadr x,vars), !&pa1(cadr caddr x,vars))
             else if not atom x and (car x eq 'append or car x eq 'nconc)
	          and eqcar(cadr x,'explode) and eqcar(caddr x,'explode)
	      then list('strcat,!&pa1(cadadr x,vars), !&pa1(cadr caddr x,vars))
	     else if not atom x and (car x eq 'explode)
	      then list('sprin2,!&pa1(cadr x,vars))
             else list('compress, !&pa1(x,vars))
   end;


put('compress, 'pa1fn, '!&pacompress);

%symbolic procedure !&paeqcar(u, vars);
%   list('and, list('null, list('atom, !&pa1(cadr u, vars))),
%              list('eq, list('car, !&pa1(cadr u, vars)), !&pa1(caddr u, vars)));
%
%put('eqcar, 'pa1fn, '!&paeqcar);


%flag('(rplacd!* rplaca!* reversip!* aconc!* nconc!*),'lose);

deflist('((rplacd!* rplacd) (rplaca!* rplaca) (reversip!* reversip)
       (aconc!* aconc) (nconc!* nconc)),'cnewnam);

%symbolic smacro procedure arrayp u; get(u,'rtype) eq 'array;

%symbolic smacro procedure simpcar u; simp car u;




end;  % of file
