
/*
 * const.c                (C) Copyright 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/const.c,v 2.2 2015/04/23 01:01:47 kako Exp $
 */

static char const rcsid[] =
    "$Id: const.c,v 2.2 2015/04/23 01:01:47 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"

#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)


Sex             intervalitp(Sex);
Sex             negatep(Sex);
Sex             modulusp(Sex, Sex);
Sex             smodulusp(Sex, Sex);

#define ONEP(x)                onep(x)
#define ZEROP(x)        zerop(x)


Sex             make_const(Sex v, Sex p, Sex d);
Sex             negatet(Sex a);
Sex             gnegatet(Sex a);
Sex             modulust(Sex a, Sex b);
Sex             gmodulust(Sex a, Sex b);
Sex             smodulust(Sex a, Sex b);
Sex             gsmodulust(Sex a, Sex b);

/***************************************************************/
/***************************************************************/
/***************************************************************/

void
print_const(Sex a, putchar_fn * fn)
{
    print_sex(const_var(a), fn);
}

/***************************************************************/
/***************************************************************/


Sex
make_const(Sex v, Sex p, Sex d)
{
    Sex             c;

    if (is_poly(v))
	v = poly_var(v);
    stackcheck();
    push2(v, d);
    c = getvector(TAG_VECTOR, T_TRNNUM, sizeof(Transendential));
    pop2(d, v);
    const_var(c) = v;
    const_pol(c) = v;
    const_def(c) = d;
    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/




/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
gplus2tn(Sex a, Sex b)
{
    return gplus2pn(const_pol(a), b);
}

Sex
gplus2nt(Sex a, Sex b)
{
    return gplus2pn(const_pol(b), a);
}


Sex
gplus2tt(Sex a, Sex b)
{
    return gplus2pp(const_pol(a), const_pol(b));
}

Sex
gdiff2tn(Sex a, Sex b)
{
    return gdiff2pn(const_pol(a), b);
}

Sex
gdiff2nt(Sex a, Sex b)
{
    return gdiff2pn(a, const_pol(b));
}

Sex
gdiff2tt(Sex a, Sex b)
{
    return gdiff2pp(const_pol(a), const_pol(b));
}


Sex
negatet(Sex a)
{
    return negatep(const_pol(a));
}

Sex
modulust(Sex a, Sex b)
{
    return modulusp(const_pol(a), b);
}

Sex
smodulust(Sex a, Sex b)
{
    return smodulusp(const_pol(a), b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
gtimes2tn(Sex p, Sex t)
{
    return gtimes2pn(const_pol(p), t);
}

Sex
gtimes2nt(Sex t, Sex p)
{
    return gtimes2np(t, const_pol(p));
}


Sex
gtimes2tt(Sex a, Sex b)
{
    return gtimes2pp(const_pol(a), const_pol(b));
}

Sex
gexpt2ti(Sex a, Sex b)
{
    return gexpt2pi(const_pol(a), b);
}

Sex
gquot2tn(Sex a, Sex b)
{
    return gquot2pn(const_pol(a), b);
}


Sex
gquot2nt(Sex a, Sex b)
{
    return gratio2np(a, const_pol(b));
}

Sex
gquot2tt(Sex a, Sex b)
{
    return gratio2pp(const_pol(a), const_pol(b));
}




/***************************************************************/
/***************************************************************/
/***************************************************************/




Sex
gplus2rt(Sex a, Sex b)
{
    return gplus2rp(a, const_pol(b));
}

Sex
gplus2tr(Sex a, Sex b)
{
    return gplus2pr(const_pol(a), b);
}

Sex
gdiff2rt(Sex a, Sex b)
{
    return gdiff2ra(a, const_pol(b));
}

Sex
gdiff2tr(Sex a, Sex b)
{
    return gdiff2pr(const_pol(a), b);
}

Sex
gtimes2rt(Sex a, Sex b)
{
    return gtimes2rp(a, const_pol(b));
}

Sex
gtimes2tr(Sex a, Sex b)
{
    return gtimes2pr(const_pol(a), b);
}

Sex
gquot2rt(Sex a, Sex b)
{
    return gratio2rp(a, const_pol(b));
}

Sex
gquot2tr(Sex a, Sex b)
{
    return gratio2pr(const_pol(a), b);
}

Sex
gratio2tn(Sex a, Sex b)
{
    return gratio2pn(const_pol(a), b);
}

Sex
gratio2nt(Sex a, Sex b)
{
    return gratio2np(a, const_pol(b));
}

Sex
gratio2rt(Sex a, Sex b)
{
    return gratio2rp(a, const_pol(b));
}

Sex
gratio2tt(Sex a, Sex b)
{
    return gratio2pp(const_pol(a), const_pol(b));
}

Sex
grem2nt(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
grem2tn(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
grem2tt(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
grem2rt(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
grem2tr(Sex a, Sex b)
{
    return grem2pr(const_pol(a), b);
}

Sex
gnegatet(Sex a)
{
    return gnegatep(const_pol(a));
}


Sex
gmodulust(Sex a, Sex b)
{
    return gmodulusp(const_pol(a), b);
}

Sex
gsmodulust(Sex a, Sex b)
{
    return gsmodulusp(const_pol(a), b);
}

/***************************************************************/
/***************************************************************/
Sex
ggcd2tn(Sex p, Sex t)
{
    return ggcd2pn(const_pol(p), t);
}

Sex
ggcd2nt(Sex t, Sex p)
{
    return ggcd2np(t, const_pol(p));
}

Sex
ggcd2tt(Sex a, Sex b)
{
    return ggcd2pp(const_pol(a), const_pol(b));
}
Sex
ggcd2rt(Sex a, Sex b)
{
    return ggcd2rp(a, const_pol(b));
}
Sex
ggcd2tr(Sex a, Sex b)
{
    return ggcd2pr(const_pol(a), b);
}
/***************************************************************/
