/*
 * dipoly.c             (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * distributed polynomial routine
 * $Header: /Users/kako/asuka/src/RCS/dipoly.c,v 2.29 2015/04/23 04:04:47 kako Exp $
 */

static char const rcsid[] =
    "$Id: dipoly.c,v 2.29 2015/04/23 04:04:47 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "dipoly.h"
#include "polynomial.h"


long            term_order = TERM_ORDER_LEX;

#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)


extern Sex      Lreorder(Sex a);

extern int      kernel_order_of_var(Sex x, Sex p);
extern Sex      Lclone(Sex a);
extern Sex      Lnreverse(Sex a);

extern Sex      negatep(Sex a);
extern Sex      Lnconc(Sex a, Sex b);
extern Sex      Mgcd(Sex a, Sex b);
extern Sex      Mlterm(Sex a);


Sex             Apoly2dist(Sex f, Sex p);
Sex             Anum2dist(Sex f, Sex p);
Sex             poly2dist_sub(Sex f, Sex v, Sex p, long tdeg, long order);
static Sex      order_vec_to_pol(Sex a, Sex v, long order);
Sex             Lpolynomial(Sex a);
Sex             Ldipoly(Sex a);
Sex             Llist_polynomial(Sex a);
Sex             Llist_dipoly(Sex a);
Sex             poly_to_dipoly(Sex a, Sex v);
static Sex      sort_dipoly(Sex a, long order);
static void     quick_sort_dipoly_term(Sex * a, long n, long m,
				       long order);
Sex             num_to_dipoly(Sex a, Sex v);
bool            check_dividablity(Sex a, Sex b, Sex s);

Sex             Lsetorder_lex(Sex a);
Sex             Lsetorder_tlex(Sex a);
Sex             Lsetorder_revlex(Sex a);
Sex             Lsetorder_trevlex(Sex a);
Sex             Lsetorder_gradings(Sex a, Sex b);
Sex             gplus2dd(Sex a, Sex b, long order);
Sex             gplus2dt(Sex a, Sex b, long order);
Sex             gdiff2dd(Sex a, Sex b, long order);
Sex             gdiff2dt(Sex a, Sex b, long order);
Sex             gtimes2_evec(Sex a, Sex b);
Sex             gtimes2dd(Sex a, Sex b, long order);
Sex             gtimes2dt(Sex a, Sex b);
Sex             expti_dipoly_term(Sex p, long n);
static Sex      lcm_dipoly_term(Sex a, Sex b, long order);
static long     compare_dipoly_term(Sex a, Sex b, long order);

Sex             all_variables_aux(Sex a, Sex v);
Sex             all_variables(Sex a);
extern Sex      Lmember(Sex a, Sex b);

long            gradings_matrix_m = 0;
long            gradings_matrix_n = 0;
long            gradings_matrix[1024];
/*
 */


/*************************************************************************/
/*************************************************************************/
/*************************************************************************/

#if 0
void
print_dipoly(Sex a, putchar_fn * fn)
{
    a = dipoly_pol(a);
    print_sex(a, fn);
}

#else

void
print_dipoly(Sex a, putchar_fn * fn)
{
    Sex             x,
                    p;
    long            first = YES;
    long            n,
                    i;
    int             flag;
    int             order;

    x = dipoly_order(a);
    order = mkint(car(x));
    x = cdr(x);

    n = list_length(x);
    p = dipoly_pol(a);
    stackcheck();
    push4(x, p, nil, nil);
    push(nil);
#define c stack[0]
#define y stack[-1]
#define a stack[-2]
#define p stack[-3]
#define x stack[-4]

    if (p == nil)
	fn('0');
    while (p != nil) {		/* 
				 * { [coef exp ...] ...}  
				 */
	flag = NO;
	a = car(p);
#if DIPOLY_COEFF == 0
	c = car(a);
	a = cdr(a);
#else
	c = vect_value(a, 0);
#endif
	if (is_poly(c)) {
	    if (!first)
		fn('+');
	    fn('(');
	    print_sex(c, fn);
	    fn(')');
	} else {
	    if (!first && !minuspP(c))
		fn('+');
	    if (c == mkfxat(1)) {
		flag = YES;
	    } else if (c == mkfxat(-1)) {
		flag = YES;
		fn('-');
	    } else {
		print_sex(c, fn);
	    }
	}

	y = x;
#if DIPOLY_COEFF == 0
	i = order / 2;
	for (; i < n + order / 2; i++) {
	    c = vect_value(a, i);
	    if (c != mkfxat(0)) {
		flag = NO;
		print_sex(car(y), fn);
		if (c != mkfxat(1)) {
		    fn('^');
		    print_sex(c, fn);
		}
	    }
	    y = cdr(y);
	}
#else
	i = order / 2 + 1;
	for (; i <= n + order / 2; i++) {
	    c = vect_value(a, i);
	    if (c != mkfxat(0)) {
		flag = NO;
		print_sex(car(y), fn);
		if (c != mkfxat(1)) {
		    fn('^');
		    print_sex(c, fn);
		}
	    }
	    y = cdr(y);
	}
#endif
	if (flag)
	    print_sex(mkfxat(1), fn);
	p = cdr(p);
	first = NO;
    }
#undef a
#undef c
#undef p
#undef x
    popv(5);
}

#endif




Sex
make_dipoly(Sex a, Sex v)
{
    Sex             d;
    stackcheck();
    push2(a, v);
    d = getvector(TAG_VECTOR, T_DIPOLY, sizeof(Distributed_Polynomial));
    pop2(v, a);
    // dipoly_var(d) = v;
    // dipoly_order(d) = mkfxat(term_order);
    dipoly_order(d) = v;
    dipoly_pol(d) = a;
    return d;
}

/*************************************************************************/
/*************************************************************************/

Sex
poly2dist_sub(Sex f, Sex v, Sex p, long tdeg, long order)
			// polynomial to distributed form
{
    int             n;
    Sex             ll;
    Sex             q;


    if (!is_poly(f)) {
#if DIPOLY_COEFF == 0
	if (order & 2)
	    vect_value(v, 0) = mkfxat(tdeg);
	v = cons(f, v);
#else
	vect_value(v, 0) = f;
	if (order & 2)
	    vect_value(v, 1) = mkfxat(tdeg);
#endif
	return cons(v, nil);
    }
    stackcheck();
    push4(f, p, v, nil);
    push(nil);
#define f stack[-4]
#define p stack[-3]
#define q stack[-2]
#define a stack[-1]
#define l stack[0]
    n = kernel_order_of_var(poly_var(f), p);
    if (n >= 0) {
#if DIPOLY_COEFF == 0
	n = n + order / 2;
#else
	n = n + order / 2 + 1;
#endif
	f = poly_pol(f);
	while (is_cons(f)) {
	    a = lterm(f);
	    ll = Lclone(q);
	    vect_value(ll, n) = power(a);
	    ll = poly2dist_sub(coeff(a), ll, p, tdeg + mkint(power(a)),
			       order);
	    l = Lnconc(l, ll);
	    f = reduct(f);
	}
    } else {
#if DIPOLY_COEFF == 0
	if (order & 2)
	    vect_value(q, 0) = mkfxat(tdeg);
	q = cons(f, q);
#else
	vect_value(q, 0) = f;
	if (order & 2)
	    vect_value(q, 1) = mkfxat(tdeg);
#endif
	l = cons(q, l);
    }
    l = Lnreverse(l);
#undef f
#undef p
#undef q
#undef a
#undef l
    q = stack[0];
    popv(5);
    return q;
}


Sex
Apoly2dist(Sex f, Sex p)	// polynomial to distributed form
{
    int             n;
    Sex             ll;
    Sex             q;
    Sex             save_order;
    int             tdeg;
    int             order;

    if (!is_cons(p) || !is_fixnum(car(p)))
	return aerror("poly2dist");
    order = mkint(car(p));
    p = cdr(p);
    n = list_length(p);

    save_order = kernel_order;
    kernel_order = p;
    stackcheck();
    push(save_order);
    push4(f, p, nil, nil);
    push(nil);
#define f stack[-4]
#define p stack[-3]
#define q stack[-2]
#define a stack[-1]
#define l stack[0]
    f = Lreorder(f);
#if DIPOLY_COEFF == 0
    q = make_vector(n + order / 2);
#else
    q = make_vector(n + 1 + order / 2);
#endif
    n = kernel_order_of_var(poly_var(f), p);
    tdeg = 0;
    if (n >= 0) {
#if DIPOLY_COEFF == 0
	n = n + order / 2;
#else
	n = n + order / 2 + 1;
#endif
	f = poly_pol(f);
	while (is_cons(f)) {
	    a = lterm(f);
	    ll = Lclone(q);
	    vect_value(ll, n) = power(a);
	    tdeg = mkint(power(a));
	    ll = poly2dist_sub(coeff(a), ll, p, tdeg, order);
	    l = Lnconc(l, ll);
	    f = reduct(f);
	}
    } else {
#if DIPOLY_COEFF == 0
	if (order & 2)
	    vect_value(q, 0) = mkfxat(tdeg);
	q = cons(f, q);
#else
	vect_value(q, 0) = f;
	if (order & 2)
	    vect_value(q, 1) = mkfxat(tdeg);
#endif
	l = cons(q, l);
    }
    l = Lnreverse(l);
#undef f
#undef p
#undef q
#undef a
#undef l
    q = stack[0];
    popv(5);
    pop(save_order);
    kernel_order = save_order;
    return q;
}



Sex
Anum2dist(Sex f, Sex p)		// number to distributed form
{
    int             n;
    int             order;

    if (!is_cons(p) || !is_fixnum(car(p)))
	return aerror("poly2dist");
    order = mkint(car(p));
    p = cdr(p);
    n = list_length(p);

    stackcheck();
    push2(f, nil);
#define f stack[-1]
#define q stack[-0]
#if DIPOLY_COEFF == 0
    q = make_vector(n + order / 2);
    q = cons(f, q);
#else
    q = make_vector(n + 1 + order / 2);
    vect_value(q, 0) = f;
#endif
    q = cons(q, nil);
#undef f
#undef q
    f = stack[0];
    popv(2);
    return f;
}


Sex
Aexpvec(Sex f)			// exponent vector of monomial
{
    long            i,
                    n;
    Sex             q;
    Sex             p;
    Sex            *pa,
                   *pb;
    long            order;

    if (is_dipoly(f)) {
	q = dipoly_order(f);
	order = mkint(car(q));
	q = cdr(q);
#if 0
	n = list_length(q) + order / 2;
#else
	n = list_length(q);
#endif
	push(f);
	q = make_vector(n);
	pop(f);
	f = dipoly_pol(f);
	if (f != nil) {
	    f = car(f);
#if 0
#if DIPOLY_COEFF == 0
	    f = cdr(f);
	    pa = &vect_value(f, 0);
#else
	    pa = &vect_value(f, 1);
#endif
#else
#if DIPOLY_COEFF == 0
	    f = cdr(f);
	    pa = &vect_value(f, 0 + order / 2);
#else
	    pa = &vect_value(f, 1 + order / 2);
#endif
#endif
	    pb = &vect_value(q, 0);
	    for (i = 0; i < n; i++)
		*pb++ = *pa++;
	}
	return q;
    }
    n = 0;
    q = f;
    while (is_poly(q)) {
	n++;
	q = coeff(lterm(poly_pol(q)));
    }

    push(f);
    p = all_variables(f);
    n = list_length(p);
    push(p);
    q = make_vector(n);
    pop2(p, f);
    while (is_poly(f)) {
	n = kernel_order_of_var(poly_var(f), p);
	f = poly_pol(f);
	f = lterm(f);
	vect_value(q, n) = power(f);
	f = coeff(f);
	n++;
    }
    return q;
}

Sex
Aexpvec2(Sex f, Sex p)		// exponent vector of monomial
{
    long            n;
    Sex             q;
    Sex             save_order;

    if (is_dipoly(f)) {
	push(p);
	f = Lpolynomial(f);
	pop(p);
    }
    n = list_length(p);

    save_order = kernel_order;
    kernel_order = p;
    push(save_order);
    push2(f, p);
    f = Lreorder(f);
    stack[-1] = f;
    q = make_vector(n);
    pop2(p, f);
    while (is_poly(f)) {
	n = kernel_order_of_var(poly_var(f), p);
	if (n >= 0) {
	    f = poly_pol(f);
	    f = lterm(f);
	    vect_value(q, n) = power(f);
	    f = coeff(f);
	} else
	    break;
    }
    pop(save_order);
    kernel_order = save_order;
    return q;
}

Sex
polynomial_to_distributed(Sex a, Sex v)
{
    push(v);
    if (is_number(a))
	a = Anum2dist(a, v);
    else
	a = Apoly2dist(a, v);
    pop(v);
    return make_dipoly(a, v);
}


static          Sex
order_vec_to_pol(Sex a, Sex v, long order)
{
    Sex             p,
                    q;
    long            i,
                    n;
    long            m;


    n = mkint(vect_dimx(a));
#if DIPOLY_COEFF == 0
    q = mkfxat(1);
#else
    q = vect_value(a, 0);
#endif
    push3(a, q, v);
#define a stack[-2]
#define q stack[-1]
#define w stack[-0]
#if DIPOLY_COEFF == 0
    i = order / 2;
#else
    i = order / 2 + 1;
#endif
    for (; i < n; i++) {
	m = vect_value(a, i);
	if (m != mkfxat(0)) {
	    p = gexpt(car(w), m);
	    q = gtimes2(p, q);
	}
	w = cdr(w);
    }
#undef a
#undef q
#undef w
    q = stack[-1];
    popv(3);
    return q;
}

Sex
Lpolynomial(Sex a)
{
    Sex             v;
    Sex             p;
    long            order;

    if (is_cons(a))
	return Llist_polynomial(a);
    if (!is_dipoly(a))
	return a;
    v = dipoly_order(a);
    order = mkint(car(v));
    v = cdr(v);

    a = dipoly_pol(a);
    if (!is_cons(a))
	return mkfxat(0);
    push3(a, v, nil);
#define a stack[-2]
#define w stack[-1]
#define q stack[-0]
#if DIPOLY_COEFF == 0
    q = order_vec_to_pol(cdar(a), w, order);
    q = gtimes2(q, caar(a));
#else
    q = order_vec_to_pol(car(a), w, order);
#endif
    a = cdr(a);
    while (is_cons(a)) {
#if DIPOLY_COEFF == 0
	p = order_vec_to_pol(cdar(a), w, order);
	p = gtimes2(p, caar(a));
#else
	p = order_vec_to_pol(car(a), w, order);
#endif
	q = gplus2(q, p);
	a = cdr(a);
    }
#undef a
#undef w
#undef q
    p = stack[0];
    popv(3);
    return p;
}


Sex
simplify_dipoly(Sex a)
{
    Sex             v;
    Sex             p;
    long            order;

    v = dipoly_order(a);
    order = mkint(car(v));
    v = cdr(v);

    a = dipoly_pol(a);
    if (!is_cons(a))
	return mkfxat(0);
    push3(a, v, nil);
#define a stack[-2]
#define w stack[-1]
#define q stack[-0]
#if DIPOLY_COEFF == 0
    q = order_vec_to_pol(cdar(a), w, order);
    q = gtimes2(q, caar(a));
#else
    q = order_vec_to_pol(car(a), w, order);
#endif
    a = cdr(a);
    while (is_cons(a)) {
#if DIPOLY_COEFF == 0
	p = order_vec_to_pol(cdar(a), w, order);
	p = gtimes2(p, caar(a));
#else
	p = order_vec_to_pol(car(a), w, order);
#endif
	q = gplus2(q, p);
	a = cdr(a);
    }
#undef a
#undef w
#undef q
    p = stack[0];
    popv(3);
    return p;
}

Sex
Llist_polynomial(Sex a)
{

    Sex             b;

    stackcheck();
    push2(a, nil);
#define a stack[-1]
#define c stack[0]
    while (a != nil) {
	b = car(a);
	b = Lpolynomial(b);
	c = cons(b, c);
	a = cdr(a);
    }
#undef a
#undef c
    a = stack[0];
    popv(2);
    return Lnreverse(a);
}


Sex
all_variables_aux(Sex a, Sex v)
{
    Sex             p;
    stackcheck();
    while (is_cons(a)) {
	p = coeff(lterm(a));
	a = reduct(a);
	if (is_poly(p)) {
	    push2(a, p);
	    if (Lmember(poly_var(p), v) == nil)
		v = cons(poly_var(p), v);
	    pop(p);
	    v = all_variables_aux(poly_pol(p), v);
	    pop(a);
	}
    }
    return v;
}

Sex
all_variables(Sex a)
{
    Sex             v;

    ASSERT(is_poly(a));
    v = nil;
    push(a);
    v = cons(poly_var(a), v);
    pop(a);
    v = all_variables_aux(poly_pol(a), v);
    return Lnreverse(v);
}

Sex
Ldipoly(Sex a)
{
    Sex             v;

    if (is_cons(a))
	return Llist_dipoly(a);
    if (is_dipoly(a)) {
	if (dipoly_kernel_order == dipoly_order(a))
	    return a;
	a = Lpolynomial(a);
    }
    if (!is_poly(a)) {
	if (is_number(a)) {
	    if (dipoly_kernel_order == nil) {
		warnings("kernel order is not defined", nil, nil);
		return a;
	    } else {
		a = Anum2dist(a, dipoly_kernel_order);
		return make_dipoly(a, dipoly_kernel_order);
	    }
	}
	warnings("cannot convert to dipoly", nil, nil);
	return a;
    }
    if (dipoly_kernel_order == nil) {
	push(a);
	v = all_variables(a);
	dipoly_kernel_order = cons(mkfxat(term_order), v);
	pop(a);
    }
    a = Apoly2dist(a, dipoly_kernel_order);
    a = sort_dipoly(a, term_order);
    return make_dipoly(a, dipoly_kernel_order);
}

Sex
Llist_dipoly(Sex a)
{

    Sex             b;


    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define c stack[-2]
#define v stack[-1]
#define s stack[0]
    if (dipoly_kernel_order == nil) {
	s = a;
	v = nil;
	while (s != nil) {
	    b = car(s);
	    s = cdr(s);
	    b = all_variables(b);
	    v = Lunion(v, b);
	}
	dipoly_kernel_order = cons(mkfxat(term_order), v);
    }
    while (a != nil) {
	b = car(a);
	b = Ldipoly(b);
	c = cons(b, c);
	a = cdr(a);
    }
#undef a
#undef c
#undef v
#undef s
    a = stack[-2];
    popv(4);
    return Lnreverse(a);
}


Sex
poly_to_dipoly(Sex a, Sex v)
{
    long            order;
    if (!is_cons(v) || !is_fixnum(car(v)))
	return aerror("dipoly");
    order = mkint(car(v));
    push(v);
    a = Apoly2dist(a, v);
    a = sort_dipoly(a, order);
    pop(v);
    return make_dipoly(a, v);
}

static          Sex
sort_dipoly(Sex a, long order)
{
    long            i;
    long            n;
    Sex             b;
    Sex            *p;

    b = a;
    n = 0;
    while (is_cons(b)) {
	n++;
	b = cdr(b);
    }
    push(a);
    b = make_vector(n);
    pop(a);
    p = &vect_value(b, 0);
    for (i = 0; i < n; i++) {
	*p++ = car(a);
	a = cdr(a);
    }
    p = &vect_value(b, 0);
    quick_sort_dipoly_term(p, 0L, n - 1, order);
    a = nil;
    for (i = 0; i < n; i++)
	a = cons(*p++, a);
    // return Lnreverse(a);
    return a;
}

long
compare_dipoly(Sex a, Sex b)
{
    Sex             c;
    int             order;
    long            o;
    Sex             p,
                    q;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	return aerror("compare_dipoly");
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
    while (a != nil) {
	if (b == nil)
	    return -1;
	p = car(a);
	q = car(b);
	o = compare_dipoly_term(p, q, order);
	if (o != 0)
	    return -o;
	a = cdr(a);
	b = cdr(b);
    }
    return b == nil ? 0 : 1;
}

static long
compare_dipoly_term(Sex a, Sex b, long order)
{
    long            i,
                    n;
    Sex            *pa,
                   *pb;
    long            r = 0;

#if DIPOLY_COEFF == 0
    a = cdr(a);
    b = cdr(b);
#endif
    n = mkint(vect_dimx(a));
    if (order & 1) {		/* reverse lex order */
#if DIPOLY_COEFF == 0
	pa = &vect_value(a, 0);
	pb = &vect_value(b, 0);
#else
	pa = &vect_value(a, 1);
	pb = &vect_value(b, 1);
#endif
	if (order & 2) {
	    r = -(*pa - *pb);
	    if (r != 0)
		return r;
	}
#if DIPOLY_COEFF == 0
	pa = pa + n - 1;
	pb = pb + n - 1;
	i = order / 2;
#else
	pa = pa + n - 2;
	pb = pb + n - 2;
	i = order / 2 + 1;
#endif
	for (; i < n; i++) {
	    if (*pa != *pb) {
		r = (*pa - *pb);
		break;
	    }
	    pa--;
	    pb--;
	}
	return r;
    } else {
#if DIPOLY_COEFF == 0
	pa = &vect_value(a, 0);
	pb = &vect_value(b, 0);
#else
	pa = &vect_value(a, 1);
	pb = &vect_value(b, 1);
#endif
	r = 0;
#if DIPOLY_COEFF == 0
	for (i = 0; i < n; i++) {
	    if (*pa != *pb) {
		r = -(*pa - *pb);
		break;
	    }
	    pa++;
	    pb++;
	}
#else
	for (i = 1; i < n; i++) {
	    if (*pa != *pb) {
		r = -(*pa - *pb);
		break;
	    }
	    pa++;
	    pb++;
	}
#endif
	return r;
    }
}

static void
quick_sort_dipoly_term(Sex * a, long left, long right, long order)
{
    long            l_save,
                    r_save;
    long            ll;
    Sex             pivot;

    l_save = left;
    r_save = right;
    pivot = a[left];

    while (left < right) {
	while (compare_dipoly_term(a[right], pivot, order) <= 0
	       && left < right)
	    right--;
	if (left != right) {
	    a[left] = a[right];
	    left++;
	}
	while (compare_dipoly_term(a[left], pivot, order) >= 0
	       && left < right)
	    left++;
	if (left != right) {
	    a[right] = a[left];
	    right--;
	}
    }
    a[left] = pivot;
    ll = left;
    left = l_save;
    right = r_save;
    if (left < ll)
	quick_sort_dipoly_term(a, left, ll - 1, order);
    if (right > ll)
	quick_sort_dipoly_term(a, ll + 1, right, order);
}

/*************************************************************************/
/*************************************************************************/
Sex
Lsetorder_lex(Sex a)
{
    Sex             old_order = dipoly_kernel_order;
    term_order = TERM_ORDER_LEX;
    dipoly_kernel_order = cons(mkfxat(TERM_ORDER_LEX), a);
    return old_order;
}


Sex
Lsetorder_tlex(Sex a)
{
    Sex             old_order = dipoly_kernel_order;
    term_order = TERM_ORDER_TLEX;
    dipoly_kernel_order = cons(mkfxat(TERM_ORDER_TLEX), a);
    return old_order;
}


Sex
Lsetorder_revlex(Sex a)
{
    Sex             old_order = dipoly_kernel_order;
    term_order = TERM_ORDER_REVLEX;
    dipoly_kernel_order = cons(mkfxat(TERM_ORDER_REVLEX), a);
    return old_order;
}


Sex
Lsetorder_trevlex(Sex a)
{
    Sex             old_order = dipoly_kernel_order;
    term_order = TERM_ORDER_TREVLEX;
    dipoly_kernel_order = cons(mkfxat(TERM_ORDER_TREVLEX), a);
    return old_order;
}

Sex
Lsetorder_gradings(Sex a, Sex b)
{
    Sex             old_order = dipoly_kernel_order;
    Sex             c;
    Sex            *p;
    long            dx,
                    dy;
    long            i,
                    k;
#if 0
    lisp_printf("setorder_gradings: %p", a);
    lisp_printf(" , vars = %p\n", b);
#endif

    if (!is_matrix(a))
	aerror1("setorder_gradings no matrix", a);
    if (b == nil)
	aerror("gradings: order not set");
    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    gradings_matrix_m = dx;
    gradings_matrix_n = dy;
    if (dx > dy || dy > 32)
	aerror("gradings order exceed max");
    if (dy != list_length(b))
	aerror("gradings size mismath");
    p = &matrix_value(a, 0);
    k = dx * dy;
    for (i = 0; i < k; i++) {
	gradings_matrix[i] = mkint(*p);
	p++;
    }


#if 0
    printf("grading matrix = ");
    for (i = 0; i < dx; i++) {
	if (i != 0)
	    printf("                 ");
	for (k = 0; k < dy; k++)
	    printf("  %4d ", gradings_matrix[i * dy + k]);
	printf("\n");
    }
#endif
    term_order = TERM_ORDER_GRADINGS;
    dipoly_kernel_order = cons(mkfxat(TERM_ORDER_GRADINGS), b);
    return old_order;
}

/*************************************************************************/
/*************************************************************************/

Sex
Adipoly_spoly(Sex a, Sex b)
{
    Sex             c;
    long            order;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("spoly");
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
    push4(a, b, c, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define v stack[-0]

    v = lcm_dipoly_term(car(a), car(b), order);
    // lisp_printf("spoly: v= %p\n", v);
    /*
     * lisp_printf("v = %p\n", v); lisp_printf("a = %p\n", a);
     * lisp_printf("b = %p\n", b); 
     */
    // lisp_printf("spoly: a = %p\n", a);
    if (a != nil)
	a = cdr(a);
    a = gtimes2dt(a, car(v));
    // lisp_printf("spoly: a*v = %p\n", a);
    // lisp_printf("spoly: b = %p\n", b);
    if (b != nil)
	b = cdr(b);
    b = gtimes2dt(b, cdr(v));
    // lisp_printf("spoly: b*v = %p\n", b);
    a = gdiff2dd(a, b, order);
    // lisp_printf("spoly: a-b = %p\n", a);
#undef a
#undef b
#undef c
#undef v
    a = stack[-3];
    c = stack[-1];
    popv(4);
    return make_dipoly(a, c);
}

static          Sex
lcm_dipoly_term(Sex a, Sex b, long order)
{
    long            i,
                    n;
    Sex             k,
                    l,
                    m;
    long            ta,
                    tb;
    push5(a, b, nil, nil, nil);
    push(nil);
#define a stack[-5]
#define b stack[-4]
#define p stack[-3]
#define c stack[-2]
#define d stack[-1]
#define g stack[0]
#if DIPOLY_COEFF == 0
    g = Mgcd(car(a), car(b));

    p = gquot2(car(b), g);
    g = gquot2(car(a), g);
    a = cdr(a);
    b = cdr(b);
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    d = make_vector(n);
    ta = 0;
    tb = 0;
    for (i = order / 2; i < n; i++) {
	k = vect_value(a, i);
	l = vect_value(b, i);
	m = k > l ? k : l;
	vect_value(c, i) = m - k + TAG_FIXNUM;
	ta = ta + m - k;
	vect_value(d, i) = m - l + TAG_FIXNUM;
	tb = tb + m - l;
    }
    if (order / 2) {
	vect_value(c, 0) = ta + TAG_FIXNUM;
	vect_value(d, 0) = tb + TAG_FIXNUM;
    }
    d = cons(g, d);
    c = cons(p, c);
    c = cons(c, d);
#else
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    d = make_vector(n);
    g = Mgcd(vect_value(a, 0), vect_value(b, 0));
    p = gquot2(vect_value(b, 0), g);
    vect_value(c, 0) = p;
    p = gquot2(vect_value(a, 0), g);
    vect_value(d, 0) = p;
    for (i = 1; i < n; i++) {
	k = vect_value(a, i);
	l = vect_value(b, i);
	m = k > l ? k : l;
	vect_value(c, i) = m - k + TAG_FIXNUM;
	vect_value(d, i) = m - l + TAG_FIXNUM;
    }
    c = cons(c, d);
#endif
#undef a
#undef b
#undef p
#undef c
#undef d
#undef g
    a = stack[-2];
    popv(6);
    return a;
}

Sex
Adipoly_reduce(Sex a, Sex b)
{
    Sex             bs;
    bs = b;

    stackcheck();
    push4(a, b, bs, nil);
#define a stack[-3]
#define b stack[-2]
#define bs stack[-1]
#define c stack[-0]
    while (bs != nil) {
	c = car(bs);
	bs = cdr(bs);
	c = Adipoly_reduce1(a, c);
	if (a != c) {
	    a = c;
	    bs = b;
	}
    }
#undef a
    a = stack[-3];
#undef b
#undef bs
#undef c
    popv(4);
    return a;
}

Sex
Adipoly_reduce1(Sex a, Sex b)
{
    Sex             as,
                    bs;
    Sex             c;
    long            i,
                    n,
                    order;
    Sex            *pa,
                   *pb,
                   *pq;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("reduce");

    as = dipoly_pol(a);
    if (as == nil)
	return a;
    as = car(as);
    bs = dipoly_pol(b);
    bs = car(bs);
#if DIPOLY_COEFF == 0
    as = cdr(as);
    bs = cdr(bs);
#endif
    n = mkint(vect_dimx(as));
#if DIPOLY_COEFF == 0
    pa = &vect_value(as, 0);
    pb = &vect_value(bs, 0);
    for (i = 0; i < n; i++) {
	if ((signed long) *pa++ < (signed long) *pb++)
	    return a;
    }
#else
    pa = &vect_value(as, 1);
    pb = &vect_value(bs, 1);
    for (i = 1; i < n; i++) {
	if ((signed long) *pa++ < (signed long) *pb++)
	    return a;
    }
#endif

    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);

    stackcheck();
    push5(a, b, as, bs, c);
    push2(nil, nil);
#define a stack[-6]
#define b stack[-5]
#define as stack[-4]
#define bs stack[-3]
#define c stack[-2]
#define p stack[-1]
#define q stack[-0]

    q = make_vector(n);
#if DIPOLY_COEFF == 0
    pa = &vect_value(as, 0);
    pb = &vect_value(bs, 0);
    pq = &vect_value(q, 0);
    for (i = 0; i < n; i++)
	*pq++ = *pa++ - *pb++ + TAG_FIXNUM;
    as = caar(a);
    bs = caar(b);
    p = Mgcd(as, bs);
    as = gquot2(as, p);
    bs = gquot2(bs, p);
    p = gratio2(as, bs);
    p = cons(p, q);
#else
    pa = &vect_value(as, 1);
    pb = &vect_value(bs, 1);
    pq = &vect_value(q, 1);
    for (i = 1; i < n; i++)
	*pq++ = *pa++ - *pb++ + TAG_FIXNUM;
    p = Mgcd(vect_value(as, 0), vect_value(bs, 0));
    as = gquot2(vect_value(as, 0), p);
    bs = gquot2(vect_value(bs, 0), p);
    p = gratio2(as, bs);
    vect_value(q, 0) = p;
#endif
    a = cdr(a);
    b = cdr(b);
    b = gtimes2dt(b, p);
    a = gdiff2dd(a, b, order);
    a = make_dipoly(a, c);
#undef a
#undef b
#undef as
#undef bs
#undef c
#undef p
#undef q
    a = stack[-6];
    popv(7);
    return a;
}

Sex
Adipoly_lt_reducible(Sex a, Sex b)
{
    long            n,
                    i;
    Sex            *pa,
                   *pb;
    long            order;
    long            m;
    if (dipoly_order(a) != dipoly_order(b))
	aerror("lt_reducible: order mismatch");

    order = mkint(car(dipoly_order(a)));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    n = mkint(vect_dimx(a));
    m = order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    for (i = m; i < n; i++) {
	if (*pa++ < *pb++)
	    return nil;
    }
#else
    a = car(a);
    b = car(b);
    n = mkint(vect_dimx(a));
    m = order / 2 + 1;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    for (i = m; i < n; i++) {
	if (*pa++ < *pb++)
	    return nil;
    }
#endif
    return true;
}

Sex
Adipoly_lt_divisible(Sex a, Sex b)
{
    long            n,
                    i;
    Sex            *pa,
                   *pb;
    long            order;
    long            m;

    if (dipoly_order(a) != dipoly_order(b))
	aerror("lt_divisible: order mismatch");

    order = mkint(car(dipoly_order(a)));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    n = mkint(vect_dimx(a));
    m = order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    for (i = m; i < n; i++) {
	if (*pa++ < *pb++)
	    return nil;
    }
#else
    a = car(a);
    b = car(b);
    n = mkint(vect_dimx(a));
    m = order / 2 + 1;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    for (i = m; i < n; i++) {
	if (*pa++ < *pb++)
	    return nil;
    }
#endif
    return true;
}

Sex
Adipoly_have_nontrivial_gcd(Sex a, Sex b)
{
    long            n,
                    i;
    Sex            *pa,
                   *pb;
    long            order;
    Sex             c;
    long            m;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("have_nontrivial_gcd: order mismatch");
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    n = mkint(vect_dimx(a));
    m = order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    for (i = m; i < n; i++) {
	if (*pa > mkfxat(0) && *pb > mkfxat(0))
	    return true;
	pa++;
	pb++;
    }
#else
    a = car(a);
    b = car(b);
    n = mkint(vect_dimx(a));
    m = 1 + order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    for (i = m; i < n; i++) {
	if (*pa > mkfxat(0) && *pb > mkfxat(0))
	    return true;
	pa++;
	pb++;
    }
#endif
    return nil;
}

#if 0
Sex
Adipoly_lt_lcm_ordervec(Sex a, Sex b)
{
    Sex             c;
    long            n,
                    i;
    Sex            *pa,
                   *pb,
                   *pc;
    Sex             ax,
                    bx;
    long            order;
    long            m;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("lt_expvec: order mismatch");
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    push2(a, b);
    m = order / 2;
    n = mkint(vect_dimx(a)) - m;
    c = make_vector(n);
    pop2(b, a);
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
#else
    a = car(a);
    b = car(b);
    push2(a, b);
    m = 1 + order / 2;
    n = mkint(vect_dimx(a)) - m;
    c = make_vector(n);
    pop2(b, a);
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
#endif
    return c;
}
#endif

Sex
Adipoly_lt_lcm_expvec(Sex a, Sex b)
{
    Sex             c;
    long            n,
                    i;
    Sex            *pa,
                   *pb,
                   *pc;
    Sex             ax,
                    bx;
    long            order;
    long            m;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("lt_expvec: order mismatch");
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    push2(a, b);
    m = order / 2;
    n = mkint(vect_dimx(a)) - m;
    c = make_vector(n);
    pop2(b, a);
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
#else
    a = car(a);
    b = car(b);
    push2(a, b);
    m = 1 + order / 2;
    n = mkint(vect_dimx(a)) - m;
    c = make_vector(n);
    pop2(b, a);
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
#endif
    return c;
}

bool
check_dividablity(Sex a, Sex b, Sex s)
{
    Sex            *pa,
                   *pb,
                   *ps;
    long            n;
    int             i;
    int             r = NO;
    Sex             d;
#if DIPOLY_COEFF == 0
    a = cdr(a);
    b = cdr(b);
    d = vect_dimx(a);
    if (d != vect_dimx(b) || d != vect_dimx(b))
	aerror("check_dividablity: size mismatch");
    n = mkint(d);
    pa = &vect_value(a, 0);
    pb = &vect_value(b, 0);
    ps = &vect_value(s, 0);
    for (i = 0; i < n; i++) {
	if (*pa < *pb)
	    return NO;
	*ps++ = *pa++ - *pb++ + TAG_FIXNUM;
    }
#else
    d = vect_dimx(a);
    if (d != vect_dimx(b) || d != vect_dimx(b))
	aerror("check_dividablity: size mismatch");
    n = mkint(d);
    pa = &vect_value(a, 1);
    pb = &vect_value(b, 1);
    ps = &vect_value(s, 1);
    for (i = 1; i < n; i++) {
	if (*pa < *pb)
	    return NO;
	*ps++ = *pa++ - *pb++ + TAG_FIXNUM;
    }
#endif
    return YES;
}



Sex
Adipoly_full_reduce(Sex a, Sex b)
{
    Sex             d;
    int             dirty = NO;
    long            order;

    d = dipoly_order(a);
    order = mkint(car(d));
    d = dipoly_pol(a);
    if (d == nil)
	return a;
    stackcheck();
    push(nil);
    push4(nil, nil, d, nil);
    push4(a, b, nil, nil);
#define s stack[-8]
#define g stack[-7]
#define f stack[-6]
#define d stack[-5]
#define e stack[-4]
#define a stack[-3]
#define b stack[-2]
#define bs stack[-1]
#define c stack[-0]
#if DIPOLY_COEFF == 0
    g = Lclone(cdar(d));
#else
    g = Lclone(car(d));
#endif
    while (YES) {
	bs = b;
	while (d != nil && bs != nil) {
	    // lisp_printf("d = %p\n", (d)); 
	    // lisp_printf("bs=%p\n", Mlterm(car(bs))); 
	    c = car(bs);
	    c = dipoly_pol(c);
	    bs = cdr(bs);
	    if (c == nil)
		continue;

	    f = car(c);
	    s = car(d);
	    if (check_dividablity(s, f, g)) {
#if DIPOLY_COEFF == 0
		f = car(f);
		s = car(s);
		f = gratio2(s, f);
		f = cons(f, g);
#else
		f = vect_value(f, 0);
		s = vect_value(s, 0);
		f = gratio2(f, s);
		vect_value(g, 0) = f;
		f = g;
#endif


		// lisp_printf("f = %p\n", f); 


		c = gtimes2dt(c, f);

		c = cdr(c);
		d = cdr(d);


		d = gdiff2dd(d, c, order);
		bs = b;
		dirty = YES;
	    }
	}
	if (d == nil)
	    break;
	e = cons(car(d), e);
	d = cdr(d);
    }
    if (dirty) {
	if (e == nil)
	    a = mkfxat(0);
	else {
	    e = Lnreverse(e);
	    a = make_dipoly(e, dipoly_order(a));
	}
    }
#undef a
#undef b
#undef bs
#undef c
#undef d
#undef e
#undef f
#undef g
#undef s
    a = stack[-3];
    popv(9);
    return a;
}

Sex
Adipoly_quotient(Sex a, Sex b)
{
    Sex             c;
    Sex             d;
    int             dirty = NO;
    long            order;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	return aerror("dipoly_quotient");
    order = mkint(car(c));
    d = dipoly_pol(a);
    if (d == nil)
	return mkfxat(0);
    b = dipoly_pol(b);
    if (b == nil)
	return mkfxat(0);
    stackcheck();
    push2(nil, nil);
    push4(nil, nil, d, nil);
    push3(a, b, c);
#define p stack[-8]
#define s stack[-7]
#define g stack[-6]
#define f stack[-5]
#define d stack[-4]
#define e stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
#if DIPOLY_COEFF == 0
    g = Lclone(cdar(d));
#else
    g = Lclone(car(d));
#endif
    while (d != nil) {
	/*
	 * lisp_printf("d = %p\n", d); lisp_printf("b=%p\n", b); 
	 */
	c = b;
	f = car(c);
	s = car(d);
	if (check_dividablity(s, f, g)) {
#if DIPOLY_COEFF == 0
	    f = car(f);
	    s = car(s);
	    f = gratio2(s, f);
	    p = Lclone(g);
	    f = cons(f, p);
#else
	    f = vect_value(f, 0);
	    s = vect_value(s, 0);
	    f = gratio2(f, s);
	    vect_value(g, 0) = f;
	    f = Lclone(g);
#endif

	    e = cons(f, e);
	    c = cdr(c);
	    d = cdr(d);
	    c = gtimes2dt(c, f);

	    d = gdiff2dd(d, c, order);
	} else {
	    d = cdr(d);
	}
    }
    if (e == nil)
	a = mkfxat(0);
    else {
	e = Lnreverse(e);
	a = make_dipoly(e, dipoly_order(a));
    }
#undef a
#undef b
#undef c
#undef d
#undef e
#undef f
#undef g
#undef s
#undef p
    a = stack[-2];
    popv(9);
    return a;
}


Sex
Adipoly_remainder(Sex a, Sex b)
{
    Sex             c;
    Sex             d;
    int             dirty = NO;
    long            order;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	return aerror("dipoly_remainder");
    order = mkint(car(c));
    d = dipoly_pol(a);
    if (d == nil)
	return a;
    b = dipoly_pol(b);
    if (b == nil)
	return a;
    stackcheck();
    push(nil);
    push4(nil, nil, d, nil);
    push3(a, b, c);
#define s stack[-7]
#define g stack[-6]
#define f stack[-5]
#define d stack[-4]
#define e stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
#if DIPOLY_COEFF == 0
    g = Lclone(cdar(d));
#else
    g = Lclone(car(d));
#endif
    while (d != nil) {

	/*
	 * lisp_printf("d = %p\n", d); lisp_printf("b=%p\n", b); 
	 */

	c = b;
	f = car(c);
	s = car(d);
	if (check_dividablity(s, f, g)) {
#if DIPOLY_COEFF == 0
	    f = car(f);
	    s = car(s);
	    f = gratio2(s, f);
	    f = cons(f, g);
#else
	    f = vect_value(f, 0);
	    s = vect_value(s, 0);
	    f = gratio2(f, s);
	    vect_value(g, 0) = f;
	    f = g;
#endif


	    c = cdr(c);
	    d = cdr(d);

	    c = gtimes2dt(c, f);

	    d = gdiff2dd(d, c, order);
	    dirty = YES;
	} else {
	    e = cons(car(d), e);
	    d = cdr(d);
	}
    }
    if (dirty) {
	if (e == nil)
	    a = mkfxat(0);
	else {
	    e = Lnreverse(e);
	    a = make_dipoly(e, dipoly_order(a));
	}
    }
#undef a
#undef b
#undef c
#undef d
#undef e
#undef f
#undef g
#undef s
    a = stack[-2];
    popv(8);
    return a;
}


Sex
Adipoly_reduce_set(Sex a)
{
    int             dirty = NO;
    long            n,
                    i;
    Sex            *pa;
    long            order;

    n = list_length(a);
    if (n == 0)
	return nil;
    order = mkint(car(dipoly_order(car(a))));


    stackcheck();
    push(nil);
    push4(nil, nil, nil, nil);
    push4(a, nil, nil, nil);
#define s stack[-8]
#define g stack[-7]
#define f stack[-6]
#define d stack[-5]
#define e stack[-4]
#define a stack[-3]
#define b stack[-2]
#define bs stack[-1]
#define c stack[-0]

    b = make_vector(n);

    c = a;
    pa = &vect_value(b, 0);
    for (i = 0; i < n; i++) {
	*pa++ = car(c);
	c = cdr(c);
    }
    d = dipoly_pol(car(a));
#if DIPOLY_COEFF == 0
    g = Lclone(cdar(d));
#else
    g = Lclone(car(d));
#endif
    while (YES) {
	bs = b;
	while (d != nil && bs != nil) {
	    c = car(bs);
	    c = dipoly_pol(c);
	    bs = cdr(bs);
	    if (c == nil)
		continue;

	    f = car(c);
	    s = car(d);
	    if (check_dividablity(s, f, g)) {
#if DIPOLY_COEFF == 0
		f = car(f);
		s = car(s);
		f = gratio2(s, f);
		f = cons(f, g);
#else
		f = vect_value(f, 0);
		s = vect_value(s, 0);
		f = gratio2(f, s);
		vect_value(g, 0) = f;
		f = g;
#endif

		c = cdr(c);
		d = cdr(d);

		c = gtimes2dt(c, f);

		d = gdiff2dd(d, c, order);
		bs = b;
		dirty = YES;
	    }
	}
	if (d == nil)
	    break;
	e = cons(car(d), e);
	d = cdr(d);
    }
    if (dirty) {
	if (e == nil)
	    a = mkfxat(0);
	else {
	    e = Lnreverse(e);
	    a = make_dipoly(e, dipoly_order(a));
	}
    }
#undef a
#undef b
#undef bs
#undef c
#undef d
#undef e
#undef f
#undef g
#undef s
    a = stack[-3];
    popv(9);
    return a;
}


/*************************************************************************/
/*************************************************************************/

Sex
Adipoly_LT_gcd(Sex a, Sex b)
{
    long            n,
                    i;
    Sex            *pa,
                   *pb,
                   *pc;
    long            order;
    Sex             ax,
                    bx;
    Sex             c;
    long            m;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("lt_gcd: order mismatch");
    order = mkint(car(c));
    push(c);
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    push2(a, b);
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    pop2(b, a);
    m = order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, m);
    for (i = m; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? bx : ax;
    }
    if (order & 2) {
	m = 0;
	for (i = 1; i < n; i++)
	    m += *(--pc) - TAG_FIXNUM;
	*(--pc) = m + TAG_FIXNUM;
    }
    c = cons(mkfxat(1), c);
#else
    a = car(a);
    b = car(b);
    push2(a, b);
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    pop2(b, a);
    m = 1 + order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, m);
    for (i = m; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? bx : ax;
    }
    if (order & 2) {
	m = 0;
	for (i = 2; i < n; i++)
	    m += *(--pc) - TAG_FIXNUM;
	*(--pc) = m + TAG_FIXNUM;
    }
    vect_value(c, 0) = mkfxat(1);
#endif
    c = cons(c, nil);
    pop(a);
    return make_dipoly(c, a);
}

Sex
Adipoly_LT_lcm(Sex a, Sex b)
{
    Sex             c;
    long            n,
                    i;
    Sex            *pa,
                   *pb,
                   *pc;
    Sex             ax,
                    bx;
    long            order;
    long            m;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("lt_lcm: order mismatch");
    push(c);
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
#if DIPOLY_COEFF == 0
    a = cdar(a);
    b = cdar(b);
    push2(a, b);
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    pop2(b, a);
    m = order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, m);
    for (i = m; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
    if (order & 2) {
	m = 0;
	for (i = 1; i < n; i++)
	    m += *(--pc) - TAG_FIXNUM;
	*(--pc) = m + TAG_FIXNUM;
    }
    c = cons(mkfxat(1), c);
#else
    a = car(a);
    b = car(b);
    push2(a, b);
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    pop2(b, a);
    m = 1 + order / 2;
    pa = &vect_value(a, m);
    pb = &vect_value(b, m);
    pc = &vect_value(c, m);
    for (i = m; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
    if (order & 2) {
	m = 0;
	for (i = 2; i < n; i++)
	    m += *(--pc) - TAG_FIXNUM;
	*(--pc) = m + TAG_FIXNUM;
    }
    vect_value(c, 0) = mkfxat(1);
#endif
    c = cons(c, nil);
    pop(a);
    return make_dipoly(c, a);
}

/*************************************************************************/
/*************************************************************************/

Sex
Aexpvec_lcm(Sex a, Sex b)
{
    Sex             c;
    long            n,
                    i;
    Sex            *pa,
                   *pb,
                   *pc;
    Sex             ax,
                    bx;

    if (vect_dimx(a) != vect_dimx(b)) {
#if 0
	lisp_printf("dim a = 0x%08x\n", vect_dimx(a));
	lisp_printf("dim b = 0x%08x\n", vect_dimx(b));
#endif
	aerror("expvec_lcm: size mismatch");
    }
    n = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(n);
    pop2(b, a);
    pa = &vect_value(a, 0);
    pb = &vect_value(b, 0);
    pc = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax > bx ? ax : bx;
    }
    return c;
}

Sex
Aexpvec_gcd(Sex a, Sex b)
{
    Sex             c;
    long            n,
                    i;
    Sex            *pa,
                   *pb,
                   *pc;
    Sex             ax,
                    bx;

    if (vect_dimx(a) != vect_dimx(b))
	aerror("expvec_gcd: size mismatch");
    n = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(n);
    pop2(b, a);
    pa = &vect_value(a, 0);
    pb = &vect_value(b, 0);
    pc = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax = *pa++;
	bx = *pb++;
	*pc++ = ax < bx ? ax : bx;
    }
    return c;
}

Sex
Aexpvec_div(Sex a, Sex b)
{
    long            n,
                    i;
    Sex            *pa,
                   *pb;

    if (vect_dimx(a) != vect_dimx(b))
	aerror("expvec_div: size mismatch");
    n = mkint(vect_dimx(a));
    pa = &vect_value(a, 0);
    pb = &vect_value(b, 0);
    for (i = 0; i < n; i++) {
	if (*pa++ < *pb++)
	    return nil;
    }
    return true;
}


/*************************************************************************/
/*************************************************************************/

Sex
gplus2Dn(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    push(a);
    b = num_to_dipoly(b, dipoly_order(a));
    pop(a);
    return gplus2DD(a, b);
}


Sex
gplus2nD(Sex a, Sex b)
{
    if (zerop(a))
	return b;
    push(b);
    a = num_to_dipoly(a, dipoly_order(b));
    pop(b);
    return gplus2DD(a, b);
}


Sex
gplus2DD(Sex a, Sex b)
{
    Sex             c;
    long            o;
    long            order;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("gplus2DD");
    order = mkint(car(c));
    push(a);
    a = dipoly_pol(a);
    b = dipoly_pol(b);
    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a) && is_cons(b)) {
	d = car(a);
	e = car(b);
	o = compare_dipoly_term(d, e, order);
	if (o < 0) {
	    c = cons(d, c);
	    a = cdr(a);
	} else if (o > 0) {
	    c = cons(e, c);
	    b = cdr(b);
	} else {
#if DIPOLY_COEFF == 0
	    d = gplus2(car(d), car(e));
#else
	    d = gplus2(vect_value(d, 0), vect_value(e, 0));
#endif
	    if (!gzerop(d)) {
#if DIPOLY_COEFF == 0
		e = cons(d, cdr(e));
#else
		e = Lclone(e);
		vect_value(e, 0) = d;
#endif
		c = cons(e, c);
	    }
	    a = cdr(a);
	    b = cdr(b);
	}
    }
    while (is_cons(a)) {
	c = cons(car(a), c);
	a = cdr(a);
    }
    while (is_cons(b)) {
	c = cons(car(b), c);
	b = cdr(b);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));

}

Sex
gdiff2DD(Sex a, Sex b)
{
    Sex             c;
    long            o;
    long            order;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	aerror("gdiff2DD");
    order = mkint(car(c));
    stackcheck();
    push(a);
    a = dipoly_pol(a);
    b = dipoly_pol(b);
    c = nil;
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a) && is_cons(b)) {
#if 0
	lisp_printf("gdiff2 : a = %p\n", a);
	lisp_printf("gdiff2 : b = %p\n", b);
#endif
	d = car(a);
	e = car(b);
	o = compare_dipoly_term(d, e, order);
	if (o < 0) {
	    c = cons(d, c);
#if 0
	    lisp_printf("gdiff2 a: c = %p\n", c);
#endif
	    a = cdr(a);
	} else if (o > 0) {
#if DIPOLY_COEFF == 0
	    d = gnegate(car(e));
	    d = cons(d, cdr(e));
#else
	    d = Lclone(e);
	    e = vect_value(e, 0);
	    e = gnegate(e);
	    vect_value(d, 0) = e;
#endif
	    c = cons(d, c);
#if 0
	    lisp_printf("gdiff2 A: c = %p\n", c);
#endif
	    b = cdr(b);
	} else {
#if DIPOLY_COEFF == 0
	    d = gdiff2(car(d), car(e));
#else
	    d = gdiff2(vect_value(d, 0), vect_value(e, 0));
#endif
	    if (!gzerop(d)) {
#if DIPOLY_COEFF == 0
		e = cons(d, cdr(e));
#else
		e = Lclone(e);
		vect_value(e, 0) = d;
#endif
		c = cons(e, c);
#if 0
		lisp_printf("gdiff2 b: c = %p\n", c);
#endif
	    }
	    a = cdr(a);
	    b = cdr(b);
	}
    }
    while (is_cons(a)) {
	c = cons(car(a), c);
#if 0
	lisp_printf("gdiff2 c: c = %p\n", c);
#endif
	a = cdr(a);
    }
    while (is_cons(b)) {
	d = car(b);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gnegate(e);
	d = cons(e, cdr(d));
#else
	e = vect_value(d, 0);
	e = gnegate(e);
	d = Lclone(d);
	vect_value(d, 0) = e;
#endif
	c = cons(d, c);
#if 0
	lisp_printf("gdiff2 d: c = %p\n", c);
#endif
	b = cdr(b);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}

Sex
gnegateD(Sex a)
{
    Sex             c;
    long            o;

    push(a);
    a = dipoly_pol(a);
    c = nil;
    stackcheck();
    push2(a, c);
    push2(nil, nil);
#define a stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a)) {
	d = car(a);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gnegate(e);
	d = cons(e, cdr(d));
#else
	e = vect_value(d, 0);
	e = gnegate(e);
	d = Lclone(d);
	vect_value(d, 0) = e;
#endif
	c = cons(d, c);
	a = cdr(a);
    }
#undef a
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(4);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}

Sex
gmodulusD(Sex a, Sex b)
{
    Sex             c;
    long            o;

    push(a);
    a = dipoly_pol(a);
    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a)) {
	d = car(a);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gmodulus(e, b);
	if (!gzerop(e)) {
	    d = cons(e, cdr(d));
	    c = cons(d, c);
	}
#else
	e = vect_value(d, 0);
	e = gmodulus(e, b);
	if (!gzerop(e)) {
	    d = Lclone(d);
	    vect_value(d, 0) = e;
	    c = cons(d, c);
	}
#endif
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}


Sex
gsmodulusD(Sex a, Sex b)
{
    Sex             c;
    long            o;

    push(a);
    a = dipoly_pol(a);
    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a)) {
	d = car(a);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gsmodulus(e, b);
	if (!gzerop(e)) {
	    d = cons(e, cdr(d));
	    c = cons(d, c);
	}
#else
	e = vect_value(d, 0);
	e = gsmodulus(e, b);
	if (!gzerop(e)) {
	    d = Lclone(d);
	    vect_value(d, 0) = e;
	    c = cons(d, c);
	}
#endif
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}


Sex
num_to_dipoly(Sex a, Sex v)
{
    long            n;
    stackcheck();
    push2(a, v);
#if DIPOLY_COEFF == 0
    n = list_length(v) - 1;
    v = make_vector(n);
    a = stack[-1];
    v = cons(a, v);
#else
    n = list_length(v);
    v = make_vector(n);
    a = stack[-1];
    vect_value(v, 0) = a;
#endif
    a = cons(v, nil);
    v = stack[0];
    popv(2);
    return make_dipoly(a, v);
}


bool
dipoly_is_number(Sex a)
{
    Sex             c;
    long            i,
                    n;

    c = dipoly_pol(a);
    if (c == nil)
	return YES;
    c = car(c);
#if DIPOLY_COEFF == 0
    a = car(c);
    c = cdr(c);
    n = mkint(vect_dimx(c));
    for (i = 0; i < n; i++) {
	if (vect_value(c, i) != mkfxat(0))
	    return NO;
    }
#else
    if (!is_number(vect_value(c, 0)))
	return NO;
    n = mkint(vect_dimx(c));
    for (i = 1; i < n; i++) {
	if (vect_value(c, i) != mkfxat(0))
	    return NO;
    }
#endif
    return YES;
}




Sex
gdiff2Dn(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    push(a);
    b = num_to_dipoly(b, dipoly_order(a));
    pop(a);
    return gdiff2DD(a, b);
}



Sex
gdiff2nD(Sex a, Sex b)
{
    if (zerop(a))
	return gnegateD(b);
    stackcheck();
    push(b);
    a = num_to_dipoly(a, dipoly_order(b));
    pop(b);
    return gdiff2DD(a, b);
}






Sex
gtimes2Dn(Sex a, Sex b)
{
    Sex             c;
    long            o;

    stackcheck();
    push(a);
    a = dipoly_pol(a);
    c = nil;
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a)) {
	d = car(a);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gtimes2(e, b);
	if (!gzerop(e)) {
	    d = cons(e, cdr(d));
	    c = cons(d, c);
	}
#else
	e = vect_value(d, 0);
	e = gtimes2(e, b);
	if (!gzerop(e)) {
	    d = Lclone(d);
	    vect_value(d, 0) = e;
	    c = cons(d, c);
	}
#endif
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}


Sex
gtimes2nD(Sex t, Sex p)
{
    return gtimes2Dn(p, t);
}

Sex
gplus2dd(Sex a, Sex b, long order)
{
    Sex             c;
    long            o;

    if (a == nil)
	return b;
    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a) && is_cons(b)) {
	d = car(a);
	e = car(b);
	o = compare_dipoly_term(d, e, order);
	if (o < 0) {
	    c = cons(d, c);
	    a = cdr(a);
	} else if (o > 0) {
	    c = cons(e, c);
	    b = cdr(b);
	} else {
#if DIPOLY_COEFF == 0
	    d = gplus2(car(d), car(e));
#else
	    d = gplus2(vect_value(d, 0), vect_value(e, 0));
#endif
	    if (!gzerop(d)) {
#if DIPOLY_COEFF == 0
		e = cons(d, cdr(e));
#else
		e = Lclone(e);
		vect_value(e, 0) = d;
#endif
		c = cons(e, c);
	    }
	    a = cdr(a);
	    b = cdr(b);
	}
    }
    while (is_cons(a)) {
	c = cons(car(a), c);
	a = cdr(a);
    }
    while (is_cons(b)) {
	c = cons(car(b), c);
	b = cdr(b);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    return c;
}

Sex
gplus2dt(Sex a, Sex b, long order)
{
    Sex             c;
    long            o;

    if (a == nil)
	return cons(b, nil);
    c = nil;
    stackcheck();
    push4(a, b, c, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[-0]
    while (is_cons(a)) {
	d = car(a);
	o = compare_dipoly_term(d, b, order);
	if (o < 0) {
	    c = cons(d, c);
	    a = cdr(a);
	} else if (o > 0) {
	    c = cons(b, c);
	    b = nil;
	    break;
	} else {
#if DIPOLY_COEFF == 0
	    d = gplus2(car(d), car(b));
	    if (!gzerop(d)) {
		b = cons(d, cdr(b));
		c = cons(b, c);
	    }
#else
	    d = gplus2(vect_value(d, 0), vect_value(b, 0));
	    if (!gzerop(d)) {
		b = Lclone(b);
		vect_value(b, 0) = d;
		c = cons(b, c);
	    }
#endif
	    a = cdr(a);
	    b = nil;
	    break;
	}
    }
    if (b != nil) {
	c = cons(b, c);
    }
    while (is_cons(a)) {
	c = cons(car(a), c);
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef d
    c = stack[-1];
    popv(4);
    c = Lnreverse(c);
    return c;
}

Sex
gdiff2dd(Sex a, Sex b, long order)
{
    Sex             c;
    long            o;

    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a) && is_cons(b)) {
	d = car(a);
	e = car(b);
	o = compare_dipoly_term(d, e, order);
	if (o < 0) {
	    c = cons(d, c);
	    a = cdr(a);
	} else if (o > 0) {
#if DIPOLY_COEFF == 0
	    d = gnegate(car(e));
	    e = cons(d, cdr(e));
#else
	    e = Lclone(e);
	    d = vect_value(e, 0);
	    d = gnegate(d);
	    vect_value(e, 0) = d;
#endif
	    c = cons(e, c);
	    b = cdr(b);
	} else {
#if DIPOLY_COEFF == 0
	    d = gdiff2(car(d), car(e));
#else
	    d = gdiff2(vect_value(d, 0), vect_value(e, 0));
#endif
	    if (!gzerop(d)) {
#if DIPOLY_COEFF == 0
		e = cons(d, cdr(e));
#else
		e = Lclone(e);
		vect_value(e, 0) = d;
#endif
		c = cons(e, c);
	    }
	    a = cdr(a);
	    b = cdr(b);
	}
    }
    while (is_cons(a)) {
	c = cons(car(a), c);
	a = cdr(a);
    }
    while (is_cons(b)) {
	e = car(b);
#if DIPOLY_COEFF == 0
	d = gnegate(car(e));
	e = cons(d, cdr(e));
#else
	e = Lclone(e);
	d = vect_value(e, 0);
	d = gnegate(d);
	vect_value(e, 0) = d;
#endif
	c = cons(e, c);
	b = cdr(b);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    return c;
}

Sex
gdiff2dt(Sex a, Sex b, long order)
{
    Sex             c;
    long            o;

    push2(a, b);
#define a stack[-1]
#define b stack[0]
#if DIPOLY_COEFF == 0
    c = gnegate(car(b));
    b = cons(c, cdr(b));
#else
    b = Lclone(b);
    c = vect_value(b, 0);
    c = gnegate(c);
    vect_value(b, 0) = c;
#endif
#undef a
#undef b
    pop2(b, a);
    return gplus2dt(a, b, order);
}

Sex
gtimes2_evec(Sex a, Sex b)
{
    long            i;
    long            k,
                    l;
    long            n;
    Sex             p;

    stackcheck();
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
#if DIPOLY_COEFF == 0
    p = gtimes2(car(a), car(b));
    a = cdr(a);
    b = cdr(b);
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    for (i = 0; i < n; i++) {
	k = (long) vect_value(a, i);
	l = (long) vect_value(b, i);
	vect_value(c, i) = (Sex) (k + l - TAG_FIXNUM);
    }
    c = cons(p, c);
#else
    n = mkint(vect_dimx(a));
    c = make_vector(n);
    p = gtimes2(vect_value(a, 0), vect_value(b, 0));
    vect_value(c, 0) = p;
    for (i = 1; i < n; i++) {
	k = (long) vect_value(a, i);
	l = (long) vect_value(b, i);
	vect_value(c, i) = (Sex) (k + l - TAG_FIXNUM);
    }
#endif
#undef a
#undef b
#undef c
    p = stack[0];
    popv(3);
    return p;
}

Sex
gtimes2dd(Sex a, Sex b, long order)
{
    Sex             q;

    stackcheck();
    push4(a, b, nil, nil);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define p stack[-1]
#define r stack[-0]
    r = nil;
    while (is_cons(a)) {
	p = nil;
	c = b;
	while (is_cons(c)) {
	    q = gtimes2_evec(car(a), car(c));
	    p = cons(q, p);
	    c = cdr(c);
	}
	p = Lnreverse(p);
	r = gplus2dd(r, p, order);
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef p
#undef r
    q = stack[0];
    popv(5);
    return q;
}


Sex
gtimes2dt(Sex a, Sex b)
{
    Sex             p;

    stackcheck();
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define r stack[-0]
    r = nil;
    while (is_cons(a)) {
	p = gtimes2_evec(car(a), b);
	r = cons(p, r);
	a = cdr(a);
    }
#undef a
#undef b
#undef r
    p = stack[0];
    popv(3);
    return Lnreverse(p);
}



Sex
gtimes2DD(Sex a, Sex b)
{
    Sex             q;
    long            order;

    q = dipoly_order(a);
    if (q != dipoly_order(b))
	return aerror2("gtimes2DD", a, b);
    order = mkint(car(q));
    stackcheck();
    push(a);
    a = dipoly_pol(a);
    b = dipoly_pol(b);

    push4(a, b, nil, nil);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define p stack[-1]
#define r stack[-0]
    r = nil;
    while (is_cons(a)) {
	p = nil;
	c = b;
	while (is_cons(c)) {
	    q = gtimes2_evec(car(a), car(c));
	    p = cons(q, p);
	    c = cdr(c);
	}
	p = Lnreverse(p);
	r = gplus2dd(r, p, order);
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef p
#undef r
    q = stack[0];
    popv(5);
    pop(a);
    a = dipoly_order(a);
    return make_dipoly(q, a);
}


Sex
expti_dipoly_term(Sex p, long n)
{
    Sex             a;
    long            m;
    long            i,
                    k;

#if DIPOLY_COEFF == 0
    a = car(p);
    p = cdr(p);
    push(a);
    p = Lclone(p);
    pop(a);
    m = mkint(vect_dimx(p));
    push(p);
    a = gexpt(a, mkfxat(n));
    pop(p);
    for (i = 0; i < m; i++) {
	k = mkint(vect_value(p, i));
	k = k * n;
	vect_value(p, i) = mkfxat(k);
    }
    p = cons(a, p);
#else
    p = Lclone(p);
    m = mkint(vect_dimx(p));
    push(p);
    a = vect_value(p, 0);
    a = gexpt(a, mkfxat(n));
    pop(p);
    vect_value(p, 0) = a;
    for (i = 1; i < m; i++) {
	k = mkint(vect_value(p, i));
	k = k * n;
	vect_value(p, i) = mkfxat(k);
    }
#endif
    return p;
}

Sex
exptiD(Sex a, long n, long order)
{
    Sex             p;
    long            i;

    if (a == nil || n == 1)
	return nil;

    p = car(a);
    a = cdr(a);
    stackcheck();
    push2(a, p);
    push5(nil, nil, nil, nil, nil);
#define a stack[-6]
#define p stack[-5]
#define c stack[-4]
#define f stack[-3]
#define q stack[-2]
#define r stack[-1]
#define s stack[ 0]

    c = expti_dipoly_term(p, n);
    s = cons(c, nil);

    if (a != nil) {
	f = mkfxat(n);
	r = a;
	for (i = 1; i < n; i++) {
	    c = expti_dipoly_term(p, n - i);
#if DIPOLY_COEFF == 0
	    q = car(c);
	    q = gtimes2(q, f);
	    c = cons(q, cdr(c));
#else
	    q = vect_value(c, 0);
	    q = gtimes2(q, f);
	    vect_value(c, 0) = q;
#endif
	    f = gtimes2(f, mkfxat(n - i));
	    f = gquot2(f, mkfxat(i + 1));
	    if (!zerop(q)) {
		q = gtimes2dt(r, c);
		s = gplus2dd(s, q, order);
	    }
	    r = gtimes2dd(r, a, order);
	}
	s = gplus2dd(s, r, order);
    }
#undef a
#undef p
#undef c
#undef f
#undef q
#undef r
#undef s
    p = stack[0];
    popv(7);
    return p;
}

Sex
gexpt2Di(Sex a, Sex b)
{
    Sex             x;
    long            n = mkint(b);
    long            order;

    if (n == 0)
	return mkfxat(1);
    else if (n == 1)
	return a;
    x = dipoly_order(a);
    order = mkint(car(x));
    a = dipoly_pol(a);
    stackcheck();
    push(x);
    a = exptiD(a, n, order);
    pop(x);
    return make_dipoly(a, x);
}




Sex
gquot2Dn(Sex a, Sex b)
{
    Sex             c;
    long            o;

    push(a);
    a = dipoly_pol(a);
    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a)) {
	d = car(a);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gratio2(e, b);
	d = cons(e, cdr(d));
#else
	e = vect_value(d, 0);
	e = gratio2(e, b);
	d = Lclone(d);
	vect_value(d, 0) = e;
#endif
	c = cons(d, c);
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}



Sex
gquot2nD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gquot2(a, b);
}


Sex
gquot2DD(Sex a, Sex b)
{
#if 0
    push(b);
    a = Lpolynomial(a);
    pop(b);
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gquot2(a, b);
#endif
    return Adipoly_quotient(a, b);
}




Sex
gplus2rD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gplus2(a, b);
}

Sex
gplus2Dr(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gplus2(a, b);
}

Sex
gdiff2rD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gdiff2(a, b);
}

Sex
gdiff2Dr(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gdiff2(a, b);
}

Sex
gtimes2rD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gtimes2(a, b);
}

Sex
gtimes2Dr(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gtimes2(a, b);
}

Sex
gquot2rD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gquot2(a, b);
}

Sex
gquot2Dr(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gquot2(a, b);
}


Sex
gratio2Dn(Sex a, Sex b)
{
    Sex             c;
    long            o;

    push(a);
    a = dipoly_pol(a);
    c = nil;
    stackcheck();
    push3(a, b, c);
    push2(nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[-0]
    while (is_cons(a)) {
	d = car(a);
#if DIPOLY_COEFF == 0
	e = car(d);
	e = gratio2(e, b);
	d = cons(e, cdr(d));
#else
	e = vect_value(d, 0);
	e = gratio2(e, b);
	d = Lclone(d);
	vect_value(d, 0) = e;
#endif
	c = cons(d, c);
	a = cdr(a);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    c = stack[-2];
    popv(5);
    c = Lnreverse(c);
    pop(a);
    return make_dipoly(c, dipoly_order(a));
}


Sex
gratio2nD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gratio2(a, b);
}


Sex
gratio2DD(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gratio2(a, b);
}

Sex
grem2nD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gremaind2(a, b);
}

Sex
grem2Dn(Sex a, Sex b)
{
    return mkfxat(0);
#if 0
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gremaind2(a, b);
#endif
}

Sex
grem2DD(Sex a, Sex b)
{
#if 0
    push(a);
    b = Lpolynomial(b);
    pop(a);
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gremaind2(a, b);
#endif
    return Adipoly_remainder(a, b);
}

Sex
grem2rD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return gremaind2(a, b);
}

Sex
grem2Dr(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gremaind2(a, b);
}


bool
lessp2DD(Sex a, Sex b)
{
    Sex             c;
    int             order;

    c = dipoly_order(a);
    if (c != dipoly_order(b))
	return aerror("lessp2DD");
    order = mkint(car(c));
    a = dipoly_pol(a);
    b = dipoly_pol(b);
    if (a == nil) {
	if (b == nil)
	    return NO;
	return YES;
    }
    if (b == nil)
	return NO;
    a = car(a);
    b = car(b);
    if (compare_dipoly_term(a, b, order) > 0)
	return YES;
    return NO;
}



bool
numeq2DD(Sex a, Sex b)
{
    if (a == b)
	return YES;
    if (dipoly_order(a) != dipoly_order(b))
	return NO;
    if (equal(a, b) == nil)
	return NO;
    return YES;
}

Sex
gplus2vD(Sex a, Sex b)
{
    return gplus2pD(a, b);
}
Sex
gplus2aD(Sex a, Sex b)
{
    return gplus2nD(a, b);
}
Sex
gplus2tD(Sex a, Sex b)
{
    return gplus2nD(a, b);
}
Sex
gplus2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return gplus2DD(a, b);
}
Sex
gplus2Dv(Sex a, Sex b)
{
    return gplus2Dp(a, b);
}
Sex
gplus2Da(Sex a, Sex b)
{
    return gplus2Dn(a, b);
}
Sex
gplus2Dt(Sex a, Sex b)
{
    return gplus2Dn(a, b);
}

Sex
gplus2Dp(Sex a, Sex b)
{
    push(a);
    a = dipoly_order(a);
    b = poly_to_dipoly(b, a);
    pop(a);
    return gplus2DD(a, b);
}





Sex
gplus2TD(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2RD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vD(Sex a, Sex b)
{
    return gdiff2pD(a, b);
}
Sex
gdiff2aD(Sex a, Sex b)
{
    return gdiff2nD(a, b);
}
Sex
gdiff2tD(Sex a, Sex b)
{
    return gdiff2nD(a, b);
}
Sex
gdiff2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return gdiff2DD(a, b);
}
Sex
gdiff2Dv(Sex a, Sex b)
{
    return gdiff2Dp(a, b);
}
Sex
gdiff2Da(Sex a, Sex b)
{
    return gdiff2Dn(a, b);
}
Sex
gdiff2Dt(Sex a, Sex b)
{
    return gdiff2Dn(a, b);
}
Sex
gdiff2Dp(Sex a, Sex b)
{
    push(a);
    a = dipoly_order(a);
    b = poly_to_dipoly(b, a);
    pop(a);
    return gdiff2DD(a, b);
}
Sex
gdiff2DA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2DV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2DM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2DT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2DR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2DS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2VD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2MD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TD(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2RD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vD(Sex a, Sex b)
{
    return gtimes2pD(a, b);
}
Sex
gtimes2aD(Sex a, Sex b)
{
    return gtimes2nD(a, b);
}
Sex
gtimes2tD(Sex a, Sex b)
{
    return gtimes2nD(a, b);
}
Sex
gtimes2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return gtimes2DD(a, b);
}
Sex
gtimes2Dv(Sex a, Sex b)
{
    return gtimes2Dp(a, b);
}
Sex
gtimes2Da(Sex a, Sex b)
{
    return gtimes2Dn(a, b);
}
Sex
gtimes2Dt(Sex a, Sex b)
{
    return gtimes2Dn(a, b);
}
Sex
gtimes2Dp(Sex a, Sex b)
{
    push(a);
    a = dipoly_order(a);
    b = poly_to_dipoly(b, a);
    pop(a);
    return gtimes2DD(a, b);
}
Sex
gtimes2DA(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2DT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2DR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2DS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2AD(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2TD(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2RD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vD(Sex a, Sex b)
{
    return gquot2pD(a, b);
}
Sex
gquot2aD(Sex a, Sex b)
{
    return gquot2nD(a, b);
}
Sex
gquot2tD(Sex a, Sex b)
{
    return gquot2nD(a, b);
}
Sex
gquot2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return gquot2DD(a, b);
}
Sex
gquot2Dv(Sex a, Sex b)
{
    return gquot2Dp(a, b);
}
Sex
gquot2Da(Sex a, Sex b)
{
    return gquot2Dn(a, b);
}
Sex
gquot2Dt(Sex a, Sex b)
{
    return gquot2Dn(a, b);
}
Sex
gquot2Dp(Sex a, Sex b)
{
    push(a);
    a = dipoly_order(a);
    b = poly_to_dipoly(b, a);
    pop(a);
    return gquot2DD(a, b);
}
Sex
gquot2DA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2DV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2DM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2DT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2DR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2DS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2MD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TD(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2RD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LD(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vD(Sex a, Sex b)
{
    return gratio2pD(a, b);
}
Sex
gratio2aD(Sex a, Sex b)
{
    return gratio2nD(a, b);
}
Sex
gratio2tD(Sex a, Sex b)
{
    return gratio2nD(a, b);
}
Sex
gratio2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return gratio2DD(a, b);
}















Sex
gratio2Dp(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return gratio2(a, b);
}






























Sex
grem2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return grem2DD(a, b);
}















Sex
grem2Dp(Sex a, Sex b)
{
    push(a);
    a = dipoly_order(a);
    b = poly_to_dipoly(b, a);
    pop(a);
    return grem2DD(a, b);
}


Sex
ggcd2DD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    push(b);
    a = Lpolynomial(a);
    a = ggcd2(a, b);
    return a;
}

Sex
ggcd2Dr(Sex a, Sex b)
{
    push(b);
    a = Lpolynomial(a);
    pop(b);
    return ggcd2(a, b);
}

Sex
ggcd2Dn(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    return mkfxat(1);
}

Sex
ggcd2nD(Sex a, Sex b)
{
    if (zerop(a))
	return b;
    return mkfxat(1);
}


Sex
ggcd2vD(Sex a, Sex b)
{
    return ggcd2pD(a, b);
}
Sex
ggcd2aD(Sex a, Sex b)
{
    return ggcd2nD(a, b);
}
Sex
ggcd2tD(Sex a, Sex b)
{
    return ggcd2nD(a, b);
}
Sex
ggcd2pD(Sex a, Sex b)
{
    push(b);
    b = dipoly_order(b);
    a = poly_to_dipoly(a, b);
    pop(b);
    return ggcd2DD(a, b);
}
Sex
ggcd2Dv(Sex a, Sex b)
{
    return ggcd2Dp(a, b);
}
Sex
ggcd2Da(Sex a, Sex b)
{
    return ggcd2Dn(a, b);
}
Sex
ggcd2Dt(Sex a, Sex b)
{
    return ggcd2Dn(a, b);
}
Sex
ggcd2Dp(Sex a, Sex b)
{
    push(a);
    a = dipoly_order(a);
    b = poly_to_dipoly(b, a);
    pop(a);
    return ggcd2DD(a, b);
}

Sex
ggcd2rD(Sex a, Sex b)
{
    push(a);
    b = Lpolynomial(b);
    pop(a);
    return ggcd2(a, b);
}
