/*
 * eval.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/eval.c,v 2.31 2018/03/02 08:39:42 kako Exp $
 */

static char const rcsid[] =
    "$Id: eval.c,v 2.31 2018/03/02 08:39:42 kako Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"


#define caar(p) car(car(p))
#define cddr(p) cdr(cdr(p))
#define cdar(p) cdr(car(p))


long            need_trace = NO;
long            trace_count = 1;
/*
 * long             trace_flag = SYM_TRACED | SYM_PROC;
 */
long            trace_flag = 0;
long            trace_level = 10;
extern long     run_mode;
static Sex      setf_fn(Sex args, Sex env);
static bool     case_match(Sex v, Sex s);

/***************************************************************/
/***************************************************************/

Sex             Lde_fn(Sex exp);
Sex             Ldf_fn(Sex exp);
Sex             Ldm_fn(Sex exp);
Sex             Lds_fn(Sex exp);
Sex             Lcatch_fn(Sex exp);
Sex             Lthrow_fn(Sex exp);
Sex             throw_fn(Sex exp, Sex env);
Sex             with_timeout(Sex args, Sex env);
Sex             unwind_protect(Sex args, Sex env);
Sex             return_fn(Sex args, Sex env);
Sex             aref_setf_fn(Sex var, Sex ind, Sex val);

/***************************************************************/
/***************************************************************/

// #define EVAL_DEBUG 1

Sex
Lnreverse(Sex a)
{
    Sex             b = nil;
    while (is_cons(a)) {
	Sex             c = a;
	a = cdr(a);
	cdr(c) = b;
	b = c;
    }
    return b;
}


#ifdef EVAL_DEBUG
Sex             myeval(Sex u, Sex env);

Sex
eval(Sex u, Sex env)
{
    Sex            *sav,
                    v;
    sav = stack;
    stackcheck();
    push(u);
    v = myeval(u, env);
    pop(u);
#ifdef EVAL_DEBUG
    push(v);
    lisp_printf("eval(%p) = ", u);
    v = stack[0];
    lisp_printf("%p\n", v);
    pop(v);
#endif

    if (sav != stack) {
	lisp_printf("error at stack level %08x != %08x\n", sav, stack);
	push2(u, v);
	Lprint(u);
	v = stack[0];
	Lprint(v);
	pop2(v, u);
    }
    return v;
}
#else
#define myeval eval
#endif


Sex
myeval(Sex u, Sex env)
{

    Sex             fn,
                    args;
    long            nargs;
    long            i;

    /*
     * push2(u, env); lisp_printf("eval: %p\n", u); pop2(env, u); 
     */

    stackcheck();
    if (!is_cons(u)) {
	if (is_symbol(u)) {
#ifdef StandardLisp
	    while (is_cons(env)) {
		Sex             c = car(env);
		if (u == car(c))
		    return cdr(c);
		env = cdr(env);
	    }
#endif

	    if (value(u) != undefined_var) {

		return value(u);
	    }
#ifdef ASUKA
	    else if ((run_mode & RUN_MASK) == RUN_ASUKA)
		return make_polynomial(u, mkfxat(1), mkfxat(1));
#endif
	    else
		error1("variable not defined", u);
	}
	return u;
    }
    fn = car(u);
    args = cdr(u);
    if (is_symbol(fn)) {
	switch (qtag(fn) & PROC_MASK) {
	case PROC_SMACRO:
	    push(env);
	    fn = (Sex) proc(fn);
	    u = sublis(car(fn), args, car(cdr(fn)));
	    pop(env);
	    return eval(u, env);
	case PROC_CMACRO:
	    push(env);
	    push(u);
	    fn = apply_lambda(fn, Lget(fn, macro_symbol), 1, env);
	    pop(env);
	    return eval(fn, env);
	case PROC_MACRO:
	    push(env);
	    push(u);
	    fn = apply_lambda(fn, (Sex) proc(fn), 1, env);
	    pop(env);
	    return eval(fn, env);
	case PROC_CSMACRO:
	case PROC_SUBR:
	    nargs = 0;
	    while (is_cons(args)) {
		Sex             w;
		push3(fn, args, env);
		w = car(args);
		w = eval(w, env);
		pop3(env, args, fn);
		if (nargs < MAX_NARGS) {
		    push(w);	/* 
				 * args build up on the Lisp stack 
				 */
		    nargs++;
		}
		args = cdr(args);
	    }
	    if (nargs > MAX_NARGS) {
		push3(fn, args, env);
		error1("eval: too many arguments", fn);
	    }
	    return apply_n(fn, nargs, env);
	case PROC_FSUBR:
	    if ((trace_flag & qtag(fn)) != 0) {
		Sex             w;
		stackcheck();
		push3(fn, args, env);
		lisp_printf("\n*     %04d+:%P\n", trace_count++, fn);
		lisp_printf("         Arg %d: %P\n", 1, args);
		pop2(env, args);
		w = ((Special *) proc(fn)) (args, env);
		pop(fn);
		push(w);
		lisp_printf("\n*     %04d-:%P\n", --trace_count, fn);
		lisp_printf("@        Value : %P\n", stack[0]);
		pop(w);
		return w;
	    } else
		return ((Special *) proc(fn)) (args, env);
	case PROC_EXPR:
	    nargs = 0;
	    while (is_cons(args)) {
		Sex             w;
		push3(fn, args, env);
		w = car(args);
		w = eval(w, env);
		pop3(env, args, fn);
		if (nargs < MAX_NARGS) {
		    push(w);	/* 
				 * args build up on the Lisp stack 
				 */
		    nargs++;
		}
		args = cdr(args);
	    }
	    if (nargs > MAX_NARGS) {
		error1("eval: too many arguments", fn);
	    }
	    return apply_lambda(fn, (Sex) (proc(fn)), nargs, env);
	case PROC_FEXPR:
	    push2(args, env);
	    return apply_lambda(fn, (Sex) (proc(fn)), 2, env);
	case PROC_SCODE:
	case PROC_CODE:
	    nargs = 0;
	    while (is_cons(args)) {
		Sex             w;
		push3(fn, args, env);
		w = car(args);
		w = eval(w, env);
		pop3(env, args, fn);
		if (nargs < MAX_NARGS) {
		    push(w);	/* 
				 * args build up on the Lisp stack 
				 */
		    nargs++;
		}
		args = cdr(args);
	    }
	    if (nargs > MAX_NARGS) {
		push3(fn, args, env);
		error1("eval: too many arguments", fn);
	    }
	    for (i = 1; i <= nargs; i++)
		bytecode_regs[i] = stack[i - nargs];
	    popv(nargs);
	    // printf("eval: 0x%08x\n", fn);
	    return interpreter(bcode(fn), nargs, nil);
	case PROC_FCODE:
	    if ((trace_flag & qtag(fn)) != 0) {
		Sex             w;
		stackcheck();
		push3(fn, args, env);
		lisp_printf("\n*     %04d+:%P\n", trace_count++, fn);
		lisp_printf("         Arg %d: %P\n", 1, args);
		pop2(env, args);
		w = interpreter(bcode(fn), args, env);
		pop(fn);
		push(w);
		lisp_printf("\n*     %04d-:%P\n", --trace_count, fn);
		lisp_printf("@        Value : %P\n", stack[0]);
		pop(w);
		return w;
	    } else
		return interpreter(bcode(fn), args, env);
	case PROC_MCODE:
	    push(env);
	    push(u);
	    fn = apply_lambda(fn, Lget(fn, macro_symbol), 1, env);
	    pop(env);
	    return eval(fn, env);
	default:
	    error1("eval: undefined function", fn);
	}
    } else if (is_cons(fn)) {
	if (car(fn) == lambda) {
	    nargs = 0;
	    while (is_cons(args)) {
		Sex             w;
		push3(fn, args, env);
		w = car(args);
		w = eval(w, env);
		pop3(env, args, fn);
		if (nargs < MAX_NARGS) {
		    push(w);	/* 
				 * args build up on the Lisp stack 
				 */
		    nargs++;
		}
		args = cdr(args);
	    }
	    if (nargs > MAX_NARGS) {
		error1("eval: too many arguments", fn);
	    }
	    return apply_lambda(nil, cdr(fn), nargs, env);
	} else
	    error1("eval: invalid function", fn);
    } else
	error1("eval: invalid function", fn);
    return u;			/* 
				 * otherwith returns u 
				 */
}


Sex
apply_lambda(Sex fn, Sex def, long nargs, Sex env)
{
    Sex            *next_arg;
    long            args_left = nargs;
    Sex             ret;
    Sex            *bsave;
    long            i;

    if (!is_cons(def)) {
	popv(nargs);
	return (nil);		/* 
				 * Should never happen  
				 */
    }
    stackcheck();
    next_arg = &stack[1 - nargs];	/* 
					 * Points to arg1                
					 */
    push(fn);
    push3(car(def),		/* 
				 * bvl                           
				 */
	  cdr(def),		/* 
				 * body                          
				 */
	  env);
    push2(nil, nil);		/* 
				 * v, w                      
				 */
#define w           stack[0]
#define v           stack[-1]
#define env         stack[-2]
#define body        stack[-3]
#define bvl         stack[-4]
    if ((trace_flag & qtag(fn)) != 0) {
	stackcheck();
	lisp_printf("\n*     %04d+:%P\n", trace_count++, fn);
	for (i = 0; i < nargs; i++)
	    lisp_printf("         Arg %d: %P\n", i + 1, next_arg[i]);
    }
#ifdef StandardLisp
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	v = car(bvl);
	if (args_left == 0)
	    error1("too few args", fn);
	args_left--;

	if (!is_symbol(v))
	    error1("apply_lambda: bad arg list", v);
	v = cons(v, *next_arg++);
	env = cons(v, env);
    }


    if (args_left != 0)
	error2("too many args", fn, mkfxat(args_left));

    ret = progn_fn(body, env);
#else
    bsave = bstack;
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	v = car(bvl);
	if (args_left == 0)
	    error1("too few args", fn);
	args_left--;

	if (!is_symbol(v))
	    error1("apply_lambda: bad arg list", v);
	pushb(v);
	bstackcheck();
	value(v) = *next_arg++;
    }


    if (args_left != 0)
	error2("too many args", fn, mkfxat(args_left));

    ret = progn_fn(body, env);
    while (bstack > bsave)
	popb();
#endif
    popv(5);
    pop(fn);
    popv(nargs);
    push(ret);
    if ((trace_flag & qtag(fn)) != 0) {
	lisp_printf("\n*     %04d-:%P\n", --trace_count, fn);
	lisp_printf(">        Value : %P\n", stack[0]);
    }
    pop(ret);
    return ret;
#undef w
#undef v
#undef env
#undef body
#undef bvl
}

Sex
Leval(A)
     Sex             A;
{
    return eval(A, nil);
}

Sex
evlis(Sex A, Sex env)
{
    Sex             r;
    stackcheck();
    r = nil;
    while (is_cons(A)) {
	push3(env, cdr(A), r);
	A = car(A);
	A = eval(A, env);
	pop(r);
	r = cons(A, r);
	pop2(A, env);
    }
    return Lnreverse(r);
}


Sex
Levlis(A)
     Sex             A;
{
    Sex             r;
    stackcheck();
    r = nil;
    while (is_cons(A)) {
	push2(cdr(A), r);
	A = car(A);
	A = eval(A, nil);
	pop(r);
	r = cons(A, r);
	pop(A);
    }
    return Lnreverse(r);
}

Sex
apply(Sex fn, Sex args, Sex env)
{
    long            i;

    i = 0;
    while (is_cons(args)) {
	if (i < MAX_NARGS) {
	    stackcheck();
	    push(car(args));
	} else {
	    error1("too many arguments", fn);
	}
	args = cdr(args);
	i++;
    }
    return apply_n(fn, i, env);
}

Sex
Lapply(Sex fn, Sex args)
{
    long            i;

    i = 0;
    while (is_cons(args)) {
	if (i < MAX_NARGS) {
	    stackcheck();
	    push(car(args));
	} else {
	    error1("too many arguments", fn);
	}
	args = cdr(args);
	i++;
    }
    return apply_n(fn, i, nil);
}


Sex
Lapply0(Sex fn)
{
    stackcheck();
    return apply_n(fn, 0, nil);
}

Sex
Lapply1(Sex fn, Sex a)
{
    push(a);
    stackcheck();
    return apply_n(fn, 1, nil);
}

Sex
Lapply2(Sex fn, Sex a, Sex b)
{
    push2(a, b);
    stackcheck();
    return apply_n(fn, 2, nil);
}

Sex
Lapply3(Sex fn, Sex a, Sex b, Sex c)
{
    push3(a, b, c);
    stackcheck();
    return apply_n(fn, 3, nil);
}

Sex
Lfuncall0(Sex fn)
{
    stackcheck();
    return apply_n(fn, 0, nil);
}

Sex
Lfuncall1(Sex fn, Sex a1)
{
    push(a1);
    stackcheck();
    return apply_n(fn, 1, nil);
}

Sex
Lfuncall2(Sex fn, Sex a1, Sex a2)
{
    push2(a1, a2);
    stackcheck();
    return apply_n(fn, 2, nil);
}

Sex
Lfuncall3(Sex fn, Sex a1, Sex a2, Sex a3)
{
    push3(a1, a2, a3);
    stackcheck();
    return apply_n(fn, 3, nil);
}

Sex
Lfuncall4(Sex fn, Sex a1, Sex a2, Sex a3, Sex a4)
{
    push4(a1, a2, a3, a4);
    stackcheck();
    return apply_n(fn, 4, nil);
}

Sex
Lfuncall5(Sex fn, Sex a1, Sex a2, Sex a3, Sex a4, Sex a5)
{
    push5(a1, a2, a3, a4, a5);
    stackcheck();
    return apply_n(fn, 5, nil);
}

Sex
Lfuncall(Sex fn, Sex args)
{
    long            nargs;

    nargs = 0;
    stackcheck();
    while (is_cons(args)) {
	Sex             w;
	/*
	 * push3(fn, args, env);
	 */
	w = car(args);
	/*
	 * w = eval(w, env); pop3(env, args, fn);
	 */
	if (nargs < MAX_NARGS) {
	    push(w);		/* 
				 * args build up on the Lisp stack 
				 */
	    nargs++;
	}
	args = cdr(args);
    }
    if (nargs > MAX_NARGS) {
	error1("funcall: too many arguments", fn);
    }
    return apply_n(fn, nargs, nil);
}


Sex
sublis(Sex x, Sex y, Sex z)
{
    stackcheck();
    if (x == nil)
	if (y == nil)
	    return z;
	else
	    error("smacro argument too short");
    else if (y == nil)
	error("smacro argument too short");
    push2(x, y);
    z = subla(car(x), car(y), z);
    pop2(y, x);
    return sublis(cdr(x), cdr(y), z);
}

Sex
subla(Sex x, Sex y, Sex z)
{
    Sex             a1,
                    a2;

    if (x == z)
	return y;
    else if (!is_cons(z))
	return z;
    stackcheck();
    push3(x, y, z);
    a1 = subla(x, y, car(z));
    push(a1);
    a2 = subla(stack[-3], stack[-2], cdr(stack[-1]));
    pop(a1);
    popv(3);
    return cons(a1, a2);
}


Sex
undefined_proc(Sex fname)
{
    return error1("undef: function not defined", fname);
}



Sex
apply_n(Sex fn, long nargs, Sex env)
{
    Sex             def;
    long            i;

    if (is_symbol(fn)) {
	if ((qtag(fn) & PROC_MASK) == 0)
	    aerror1(" funtion not defined ", fn);
	if (nargs != qnargs(fn)) {
	    if (nargs < qnargs(fn))
		aerror1("subr -- too few arguments", fn);
	    else if (nargs > qnargs(fn))
		aerror1("subr -- too many arguments", fn);
	}
	if ((trace_flag & qtag(fn)) != 0) {
	    stackcheck();
	    push(fn);
	    lisp_printf("\n*     %04d+:%P\n", trace_count++, fn);
	    for (i = 0; i < nargs; i++)
		lisp_printf("         Arg %d: %P\n", i + 1,
			    stack[-nargs + i]);
	    pop(fn);
	}
	switch (qtag(fn) & PROC_MASK) {
	case PROC_EXPR:
	    def = apply_lambda(fn, (Sex) proc(fn), nargs, env);
	    if ((trace_flag & qtag(fn)) != 0) {
		lisp_printf("\n*     %04d-:%P\n", --trace_count, fn);
		lisp_printf("&        Value : %P\n", def);
	    }
	    return def;
	case PROC_CSMACRO:
	case PROC_SUBR:
	    switch (nargs) {
	    default:		/* 
				 * can never arise  
				 */
		aerror("too many args in a function call");
	    case 0:
		def = (*proc0(fn)) ();
		break;
	    case 1:
		def = (*proc1(fn)) (stack[0]);
		break;
	    case 2:
		def = (*proc2(fn)) (stack[-1], stack[0]);
		break;
	    case 3:
		def = (*proc3(fn)) (stack[-2], stack[-1], stack[0]);
		break;
	    case 4:
		def = (*proc4(fn)) (stack[-3], stack[-2],
				    stack[-1], stack[0]);
		break;
	    case 5:
		def = (*proc5(fn)) (stack[-4], stack[-3],
				    stack[-2], stack[-1], stack[0]);
		break;
	    case 6:
		def = (*proc6(fn)) (stack[-5], stack[-4],
				    stack[-3], stack[-2], stack[-1],
				    stack[0]);
		break;
	    case 7:
		def = (*proc7(fn)) (stack[-6], stack[-5],
				    stack[-4], stack[-3], stack[-2],
				    stack[-1], stack[0]);
		break;
	    case 8:
		def = (*proc8(fn)) (stack[-7], stack[-6],
				    stack[-5], stack[-4], stack[-3],
				    stack[-2], stack[-1], stack[0]);
		break;
	    case 9:
		def = (*proc9(fn)) (stack[-8], stack[-7],
				    stack[-6], stack[-5], stack[-4],
				    stack[-3], stack[-2], stack[-1],
				    stack[0]);
		break;
	    case 10:
		def = (*proc10(fn)) (stack[-9], stack[-8],
				     stack[-7], stack[-6], stack[-5],
				     stack[-4], stack[-3], stack[-2],
				     stack[-1], stack[0]);
		break;
	    case 11:
		def = (*proc11(fn)) (stack[-10],
				     stack[-9], stack[-8], stack[-7],
				     stack[-6], stack[-5], stack[-4],
				     stack[-3], stack[-2], stack[-1],
				     stack[0]);
		break;
	    case 12:
		def = (*proc12(fn)) (stack[-11],
				     stack[-10], stack[-9], stack[-8],
				     stack[-7], stack[-6], stack[-5],
				     stack[-4], stack[-3], stack[-2],
				     stack[-1], stack[0]);
		break;
	    case 13:
		def = (*proc13(fn)) (stack[-12],
				     stack[-11], stack[-10], stack[-9],
				     stack[-8], stack[-7], stack[-6],
				     stack[-5], stack[-4], stack[-3],
				     stack[-2], stack[-1], stack[0]);
		break;
	    case 14:
		def = (*proc14(fn)) (stack[-13],
				     stack[-12], stack[-11], stack[-10],
				     stack[-9], stack[-8], stack[-7],
				     stack[-6], stack[-5], stack[-4],
				     stack[-3], stack[-2], stack[-1],
				     stack[0]);
		break;
	    case 15:
		def = (*proc15(fn)) (stack[-14],
				     stack[-13], stack[-12], stack[-11],
				     stack[-10], stack[-9], stack[-8],
				     stack[-7], stack[-6], stack[-5],
				     stack[-4], stack[-3], stack[-2],
				     stack[-1], stack[0]);
		break;
	    }
	    if ((trace_flag & qtag(fn)) != 0) {
		lisp_printf("\n*     %04d-:%P\n", --trace_count, fn);
		lisp_printf("#        Value : %P\n", def);
	    }
	    popv(nargs);
	    return def;
	case PROC_CODE:
	    if (nargs < 0 || nargs > MAX_NARGS)
		aerror("too many args in a function call");
	    for (i = 1; i <= nargs; i++)
		bytecode_regs[i] = stack[i - nargs];

	    def = interpreter(bcode(fn), nargs, env);

	    if ((trace_flag & qtag(fn)) != 0) {
		lisp_printf("\n*     %04d-:%P\n", --trace_count, fn);
		lisp_printf("#        Value : %P\n", def);
	    }
	    popv(nargs);
	    return def;
	}
    } else if (!is_cons(fn)) {
	popv(nargs);
	return error1("bad function", fn);
    }
    if ((def = car(fn)) == lambda)
	return apply_lambda(nil, cdr(fn), nargs, env);
    else if (def == funarg) {
	def = cdr(fn);
	if (is_cons(def))
	    return apply_lambda(nil, cdr(def), nargs, car(def));
    }
    popv(nargs);
    return error1("bad apply", fn);
}


static          Sex
and_fn(Sex args, Sex env)
{
    stackcheck();
    if (!is_cons(args))
	return true;
    for (;;) {
	Sex             v = car(args);
	args = cdr(args);
	if (!is_cons(args))
	    return eval(v, env);
	push2(args, env);
	v = eval(v, env);
	pop2(env, args);
	if (v == nil)
	    return nil;
    }
}


static          Sex
block_fn(Sex args, Sex env)
{
    Sex             p;
    if (!is_cons(args))
	return nil;
    stackcheck();
    push3(car(args),		/* 
				 * my_tag                               
				 */
	  cdr(args),		/* 
				 * args                                 
				 */
	  env);
#define env    stack[0]
#define args   stack[-1]
#define my_tag stack[-2]
    my_tag = cons(mkfxat(0), my_tag);
    env = cons(my_tag, env);
    p = nil;
    while (is_cons(args)) {
	p = car(args);
	p = eval(p, env);
	args = cdr(args);
    }
    popv(3);
    return p;
#undef env
#undef args
#undef my_tag
}

jmp_buf        *catch_buf = NULL;

#ifdef ASUKA

static          Sex
catch_fn(Sex args, Sex env)
{
    Sex             tag;
    Sex             v;
    jmp_buf         my_buf;
    jmp_buf        *saved_buf = catch_buf;
    double          trap_save;
    Sex            *stack_save = stack,
	*bstack_save = bstack;

    if (!is_cons(args))
	return true;
    stackcheck();
    push2(args, env);
    tag = car(args);
    trap_save = trap_time;
    trap_time = 0.0;
    tag = catch_tags = cons(tag, catch_tags);
    trap_time = trap_save;
    pop2(env, args);
    push(tag);
    catch_buf = &my_buf;
    stack_save = stack;
    if (setjmp(my_buf) == 0) {
	v = progn_fn(cdr(args), env);
	// v = true;
    } else {
	while (bstack > bstack_save)
	    popb();
	stack = stack_save;
	v = exit_val;
    }
    catch_buf = saved_buf;
    pop(tag);
    catch_tags = cdr(tag);
    car(tag) = tag;
    cdr(tag) = nil;
    return v;
}


Sex
throw_fn(Sex args, Sex env)
{
    Sex             tag,
                    p;
    if (!is_cons(args))
	return aerror("throw");
    stackcheck();
    tag = car(args);
    args = cdr(args);
    push2(args, env);
    pop2(env, args);
    // lisp_printf("throw: catch_tag = %p\n", catch_tags);
    for (p = catch_tags; p != nil; p = cdr(p)) {
	if (tag == car(p))
	    goto tag_found;
    }
    if (tag != any_symbol)
	return aerror("throw: tag not found");
  tag_found:
    if (is_cons(args)) {
	push(p);
	tag = car(args);
	tag = eval(tag, env);
	pop(p);
	exit_val = tag;
    } else {
	exit_val = nil;
    }
    exit_tag = tag;
    exit_reason = EXIT_THROW;
    longjmp(*catch_buf, 1);
    return nil;
}

#else

static          Sex
catch_fn(Sex args, Sex env)
{
    Sex             tag;
    Sex             v = nil;
    jmp_buf         my_buf;
    jmp_buf        *saved_buf = catch_buf;
    double          trap_save;
    Sex            *stack_save = stack,
	*bstack_save = bstack;

    if (!is_cons(args))
	return nil;
    stackcheck();
    push2(args, env);
    tag = car(args);
    tag = eval(tag, env);
    trap_save = trap_time;
    trap_time = 0;
    tag = catch_tags = cons(tag, catch_tags);
    trap_time = trap_save;
    pop2(env, args);
    push(tag);
    catch_buf = &my_buf;
    stack_save = stack;
    if (setjmp(my_buf) == 0) {
	v = progn_fn(cdr(args), env);
    } else {
	while (bstack > bstack_save)
	    popb();
	stack = stack_save;
    }
    catch_buf = saved_buf;
    pop(tag);
    catch_tags = cdr(tag);
    car(tag) = tag;
    cdr(tag) = nil;
    return v;
}


Sex
throw_fn(Sex args, Sex env)
{
    Sex             tag,
                    p;
    if (!is_cons(args))
	return aerror("throw");
    stackcheck();
    tag = car(args);
    args = cdr(args);
    push2(args, env);
    tag = eval(tag, env);
    pop2(env, args);
    // lisp_printf("throw: catch tag = %p\n", catch_tag);
    for (p = catch_tags; p != nil; p = cdr(p)) {
	if (tag == car(p))
	    goto tag_found;
    }
    return aerror("throw: tag not found");
  tag_found:
    if (is_cons(args)) {
	push(p);
	tag = car(args);
	tag = eval(tag, env);
	pop(p);
	exit_val = tag;
    } else {
	exit_val = nil;
    }
    exit_tag = tag;
    longjmp(*catch_buf, 1);
    return nil;
}

#endif

Sex
Lcatch_fn(Sex u)
{
    return catch_fn(u, nil);
}

Sex
Lthrow_fn(Sex u)
{
    return throw_fn(u, nil);
}

Sex
with_timeout(Sex args, Sex env)
{
    Sex             n,
                    u;
    double          tv;
    stackcheck();
    if (!is_cons(args))
	return aerror("with_timeout");
    n = car(args);
    if (!is_number(n))
	return aerror("with_timeout");
    tv = floatit(n) * 0.001;
    args = cdr(args);
    if (!is_cons(args))
	return aerror("with_timeout");
    push(env);
    args = cons(time_out_symbol, args);
    pop(env);
    tv = tv + get_time();
    trap_time = tv;
    return catch_fn(args, env);
}

#if 0
Sex
unwind_protect(Sex args, Sex env)
{
    Sex             r;
    Sex            *save,
                   *bsave;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    long            trace_save = trace_count;


    if (!is_cons(args))
	return nil;

    save = stack;
    bsave = bstack;
    stackcheck();

    r = nil;
    for (;;) {
	if (!is_cons(args))
	    break;

	r = car(args);
	args = cdr(args);
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    errorset_buffer = saved_buffer;
	    trace_count = trace_save;
	    r = nil;
	} else {
	    errorset_buffer = &this_level;
	    push2(args, env);
	    r = eval(r, env);
	    pop2(env, args);
	}
    }
    errorset_buffer = saved_buffer;
    return r;
}
#else
Sex
unwind_protect(Sex args, Sex env)
{
    Sex             f;
    if (!is_cons(args))
	return nil;
    stackcheck();
    f = nil;
    for (;;) {
	f = car(args);
	args = cdr(args);
	if (!is_cons(args))
	    break;
	push2(args, env);
	eval(f, env);
	pop2(env, args);
	if (exit_reason) {
	    return nil;
	}
    }
    return eval(f, env);
}
#endif

static          Sex
cond_fn(Sex args, Sex env)
{
    stackcheck();
    while (is_cons(args)) {
	Sex             p = car(args);
	if (is_cons(p)) {
	    Sex             p1;
	    push2(args, env);
	    p1 = car(p);
	    p1 = eval(p1, env);
	    pop2(env, args);
	    if (p1 != nil)
		return progn_fn(cdr(car(args)), env);
	}
	args = cdr(args);
    }
    return nil;
}


static          Sex
de_fn(Sex args, Sex env)
{
    Sex             name,
                    nargs;
    long            n;

    name = car(args);
    if (!is_symbol(name)) {
	error1("de: not a symbol", name);
    } else if (qtag(name) & SYM_PROC) {
	messages("subr redefined", name, expr_symbol);
    }
    qtag(name) = PROC_EXPR | (qtag(name) & ~PROC_MASK);
    args = cdr(args);
    nargs = car(args);
    for (n = 0; nargs != nil; n++, nargs = cdr(nargs));
    qnargs(name) = n;
    proc(name) = (lisp_proc *) args;
    return name;
}

static          Sex
dm_fn(Sex args, Sex env)
{
    Sex             name;
    name = car(args);
    if (!is_symbol(name)) {
	error1("dm: not a symbol", name);
    } else if (qtag(name) & SYM_PROC) {
	messages("macro redefined", name, macro_symbol);
    }
    qtag(name) = PROC_MACRO | (qtag(name) & ~PROC_MASK);
    args = cdr(args);
    qnargs(name) = 1;
    proc(name) = (lisp_proc *) args;
    return name;
}

static          Sex
ds_fn(Sex args, Sex env)
{
    Sex             name;
    Sex             nargs;
    long            n;

    name = car(args);
    if (!is_symbol(name)) {
	error1("ds: not a symbol", name);
    } else if (qtag(name) & SYM_PROC) {
	messages("smacro redefined", name, smacro_symbol);
    }
    qtag(name) = PROC_SMACRO | (qtag(name) & ~PROC_MASK);
    args = cdr(args);
    nargs = car(args);
    for (n = 0; nargs != nil; n++, nargs = cdr(nargs));
    qnargs(name) = n;
    proc(name) = (lisp_proc *) args;
    return name;
}

static          Sex
df_fn(Sex args, Sex env)
{
    Sex             name,
                    w;
    name = car(args);
    if (!is_symbol(name)) {
	error1("df: not a symbol", name);
    } else if (qtag(name) & SYM_PROC) {
	messages("fsubr redefined", name, fexpr_symbol);
    }
    qtag(name) = PROC_FEXPR | (qtag(name) & ~PROC_MASK);
    args = cdr(args);
    stackcheck();
    push2(name, args);
    w = Llist2(car(car(args)), alist_symbol);
    pop(args);
    args = cons(w, cdr(args));
    pop(name);
    qnargs(name) = 2;
    proc(name) = (lisp_proc *) args;
    return name;
}

Sex
Lde_fn(Sex exp)
{
    return de_fn(exp, nil);
}

Sex
Ldf_fn(Sex exp)
{
    return df_fn(exp, nil);
}

Sex
Ldm_fn(Sex exp)
{
    return dm_fn(exp, nil);
}

Sex
Lds_fn(Sex exp)
{
    return ds_fn(exp, nil);
}

static          Sex
eval_when_fn(Sex args, Sex env)
{
    Sex             situations;
    if (!is_cons(args))
	return nil;
    situations = car(args);
    args = cdr(args);
    while (is_cons(situations)) {
	if (car(situations) == eval_symbol)
	    return progn_fn(args, env);
	situations = cdr(situations);
    }
    return nil;
}


/*
 * Sex function_fn(Sex args, Sex env) { if (is_cons(args) && cdr(args) ==
 * nil) { args = car(args); if (is_cons(args) && car(args) == lambda) args 
 * = Llist2star(funarg, env, cdr(args)); return args; } return
 * aerror("function"); } 
 */
Sex
function_fn(Sex args, Sex env)
{
    if (is_cons(args) && cdr(args) == nil)
	return car(args);
    return aerror("function");
}


static          Sex
go_fn(Sex args, Sex env)
{
    Sex             p,
                    tag;
    if (!is_cons(args))
	return aerror("go");
    else
	tag = car(args);
    for (p = env; is_cons(p); p = cdr(p)) {
	Sex             w = car(p);
	if (!is_cons(w))
	    continue;
	if (car(w) == mkfxat(1) && eql(car(cdr(w)), tag)) {
	    p = w;
	    goto tag_found;
	}
    }
    return error1("go tag not found", tag);
  tag_found:
    exit_reason = EXIT_GO;
    exit_tag = tag;
    exit_val = p;
    return nil;
}

static          Sex
if_fn(Sex args, Sex env)
{
    Sex             p = nil,
	tr = nil,
	fs = nil;
    /*
     * lisp_printf("if %p\n", args); 
     */
    if (!is_cons(args))
	return aerror("if");
    p = car(args);
    args = cdr(args);
    if (!is_cons(args))
	return aerror("if");
    tr = car(args);
    args = cdr(args);
    if (!is_cons(args))
	fs = nil;
    else {
	fs = car(args);
	args = cdr(args);
	if (args != nil)
	    return aerror("if");
    }
    stackcheck();
    push3(fs, tr, env);
    p = eval(p, env);
    pop3(env, tr, fs);
    if (p == nil)
	return eval(fs, env);	/* 
				 * tail call on result      
				 */
    else
	return eval(tr, env);	/* 
				 * ... passing back values  
				 */
}

#ifdef ASUKA
static          bool
case_match(Sex v, Sex s)
{
    if (!is_cons(s)) {
	return geq2(v, s);
    }
    push2(v, s);
    while (is_cons(s)) {
	s = stack[0];
	v = stack[-1];
	if (geq2(v, car(s))) {
	    popv(2);
	    return YES;
	}
	s = stack[0];
	s = cdr(s);
	stack[0] = s;
    }
    popv(2);
    return NO;
}

static          Sex
switch_fn(Sex args, Sex env)
{
    Sex             v,
                    p;

    if (!is_cons(args))
	return aerror("switch");
    v = car(args);
    args = cdr(args);
    stackcheck();
    push2(env, v);
    // args = car(args);
    args = Lnreverse(args);
    lisp_printf("args = %p\n", args);
    push3(args, nil, nil);

#define env stack[-4]
#define v stack[-3]
#define args stack[-2]
#define exp stack[-1]
#define body stack[0]

    v = eval(v, env);
    p = nil;
    while (is_cons(args)) {
	exp = car(args);
	lisp_printf("tag = %p\n", exp);
	body = cadr(exp);
	lisp_printf("body = %p\n", body);
	exp = car(exp);
	args = cdr(args);
	lisp_printf("exp = %p\n", exp);
	exp = eval(exp, env);
	lisp_printf("exp = %p\n", exp);

	if (case_match(v, exp)) {
	    lisp_printf("body = %p\n", body);
	    while (is_cons(body)) {
		if (car(body) != nil) {
		    p = eval(car(body), env);
		}
		body = cdr(body);
	    }
	    popv(5);
	    return p;
	}

    }
#undef env
#undef v
#undef args
#undef exp
#undef body
    popv(5);
    return nil;
}

static          Sex
case_fn(Sex args, Sex env)
{
    Sex             v,
                    p;
    Sex             e;

    if (!is_cons(args))
	return aerror("case");
    v = car(args);
    e = cdr(args);
    args = car(e);
    e = cdar(e);

    stackcheck();
    push3(env, v, e);
    // args = car(args);
    args = Lnreverse(args);
    lisp_printf("args = %p\n", args);
    push3(args, nil, nil);
    e = stack[-3];
    lisp_printf("e = %p\n", e);

#define env stack[-5]
#define v stack[-4]
#define e stack[-3]
#define args stack[-2]
#define exp stack[-1]
#define body stack[0]

    v = eval(v, env);
    p = nil;
    while (is_cons(args)) {
	exp = car(args);
	// lisp_printf("tag = %p\n", exp);
	body = cadr(exp);
	// lisp_printf("body = %p\n", body);
	exp = car(exp);
	args = cdr(args);
	// lisp_printf("exp = %p\n", exp);
	exp = eval(exp, env);
	// lisp_printf("exp = %p\n", exp);

	if (case_match(v, exp)) {
	    // lisp_printf("body = %p\n", body);
	    while (is_cons(body)) {
		if (car(body) != nil) {
		    p = eval(car(body), env);
		}
		body = cdr(body);
	    }
	    popv(6);
	    return p;
	}

    }
    if (e != nil)
	p = eval(e, env);
    else
	p = nil;
#undef env
#undef v
#undef e
#undef args
#undef exp
#undef body
    popv(6);
    return p;
}
#endif

static          Sex
while_fn(Sex args, Sex env)
{
    Sex             p = nil,
	exp = nil,
	body = nil;
    if (!is_cons(args))
	return aerror("wihle");
    exp = car(args);		/* condition */
    args = cdr(args);
    if (!is_cons(args))
	return aerror("while");
    body = car(args);		/* body */
    args = cdr(args);
    if (args != nil)
	return aerror("while");
    stackcheck();
    push4(body, exp, env, nil);
    for (;;) {
	exp = stack[-2];
	env = stack[-1];
	p = eval(exp, env);
	if (p == nil)
	    break;
	p = stack[-3];
	stack[0] = p;
	while (p != nil) {
	    env = stack[-1];
	    eval(car(p), env);
	    p = stack[0];
	    p = cdr(p);
	    stack[0] = p;
	}
    }
    popv(4);
    return nil;
}

static          Sex
repeat_fn(Sex args, Sex env)
{
    Sex             p = nil,
	exp = nil,
	body = nil;
    if (!is_cons(args))
	return aerror("repeat");
    exp = car(args);		/* condition */
    args = cdr(args);
    if (!is_cons(args))
	return aerror("repeat");
    body = car(args);		/* body */
    args = cdr(args);
    if (args != nil)
	return aerror("repeat");
    stackcheck();
    push4(body, exp, env, nil);
    for (;;) {
	p = stack[-3];
	stack[0] = p;
	while (p != nil) {
	    env = stack[-1];
	    eval(car(p), env);
	    p = stack[0];
	    p = cdr(p);
	    stack[0] = p;
	}
	exp = stack[-2];
	env = stack[-1];
	p = eval(exp, env);
	if (p != nil)
	    break;
    }
    popv(4);
    return nil;
}

static          Sex
for_fn(Sex args, Sex env)
{
    Sex             q = nil,
	exp = nil,
	body = nil;


    if (!is_cons(args))
	return aerror("for");
    exp = car(args);		/* for list */
    args = cdr(args);
    if (!is_cons(args))
	return aerror("for");
    body = car(args);		/* body */
    args = cdr(args);
    if (args != nil)
	return aerror("for");
    stackcheck();
    push4(body, exp, env, nil);
    push4(nil, nil, nil, nil);
#define body  stack[-7]
#define exp   stack[-6]
#define env   stack[-5]
#define p     stack[-4]
#define ivar  stack[-3]
#define iinit stack[-2]
#define ilim  stack[-1]
#define istep stack[ 0]

    ivar = car(exp);
    iinit = eval(cadr(exp), nil);
    exp = cddr(exp);
    ilim = eval(car(exp), nil);
    istep = eval(cadr(exp), nil);
    if (istep == nil)
	istep = mkfxat(1);
    for (;;) {
#ifdef DEBUG
	lisp_printf("for: %p %p %p\n", iinit, ilim, istep);
#endif
	q = Llist2(ivar, iinit);
	setf_fn(q, env);
	if (!plusp(istep)) {
	    if (lessp2(iinit, ilim))
		break;
	} else {
	    if (lessp2(ilim, iinit))
		break;
	}

	p = body;
	while (p != nil) {
	    eval(car(p), env);
	    p = cdr(p);
	}
	iinit = plus2(iinit, istep);
    }
#undef body
#undef exp
#undef env
#undef p
#undef ivar
#undef iinit
#undef ilim
#undef istep
    popv(8);
    return nil;
}

static          Sex
foreach_fn(Sex args, Sex env)
{
    Sex             var = nil,
	exp = nil,
	body = nil;
    long            flag = 0;


    if (!is_cons(args))
	return aerror("foreach 1");
    var = car(args);		/* for list */
    if (!is_symbol(var) || var == nil || var == true)
	return aerror("bad variable in foreach statement");
    args = cdr(args);
    if (!is_cons(args))
	return aerror("foreach 2");
    exp = car(args);		/* body */
    args = cdr(args);
    if (!is_cons(args))
	return aerror("foreach 3");
    body = car(args);
    args = cdr(args);
    if (!is_cons(args))
	return aerror("foreach 4");
    args = car(args);
    if (args != nil)
	flag = 1;
    stackcheck();
    push5(body, exp, env, var, nil);
#define body  stack[-4]
#define exp   stack[-3]
#define env   stack[-2]
#define var   stack[-1]
#define p     stack[-0]

    exp = eval(exp, nil);
    while (exp != nil) {
	value(var) = flag ? exp : car(exp);
	p = body;
	while (p != nil) {
	    eval(car(p), env);
	    p = cdr(p);
	}
	exp = cdr(exp);
    }
#undef body
#undef exp
#undef env
#undef var
#undef p
    popv(5);
    return nil;
}



static          Sex
let_fn(Sex args, Sex env)
{
    Sex             bvl,
                    body;
    Sex             bodyx,
                    env1x;
    Sex            *bsave;

    if (!is_cons(args))
	return nil;
    bvl = car(args);
    body = cdr(args);

    stackcheck();
    push3(bvl, body, env);
    push5(nil, nil, env, nil, nil);
#define specenv    stack[-1]
#define env1       stack[-2]
#define p          stack[-3]
#define q          stack[-4]
#define env        stack[-5]
#define body       stack[-6]
#define bvl        stack[-7]

#ifdef StandardLisp
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	Sex             z;
	q = car(bvl);
	if (is_cons(q)) {
	    z = cdr(q);
	    q = car(q);
	    if (is_cons(z))
		z = car(z);
	    else
		z = nil;
	} else
	    z = nil;
	if (!is_symbol(q)) {
	    Sex             qq = q;
	    popv(8);
	    return (error1("bad arg list", qq));
	} else {
	    if (z != nil) {
		z = eval(z, env);
	    }
	    z = cons(q, z);
	    env1 = cons(z, env1);
	}
    }


    bodyx = progn_fn(body, env1);
#else
    bsave = bstack;
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	Sex             z;
	q = car(bvl);
	if (is_cons(q)) {
	    z = cdr(q);
	    q = car(q);
	    if (is_cons(z))
		z = car(z);
	    else
		z = nil;
	} else
	    z = nil;
	if (!is_symbol(q)) {
	    Sex             qq = q;
	    popv(8);
	    return (error1("bad arg list", qq));
	} else {
	    if (z != nil) {
		z = eval(z, env);
	    }
	    z = cons(q, z);
	    specenv = cons(z, specenv);
	}
    }


    if (specenv == nil) {
	bodyx = body;
	env1x = env1;
	popv(8);
	return (progn_fn(bodyx, env1x));
    }
    for (p = specenv; p != nil; p = cdr(p)) {
	Sex             w = car(p),
	    v = car(w),
	    z = cdr(w);
	Sex             old = value(v);
	value(v) = z;
	cdr(w) = old;
    }

    bodyx = progn_fn(body, env1);
    for (p = specenv; p != nil; p = cdr(p)) {
	Sex             w = car(p),
	    v = car(w),
	    z = cdr(w);
	value(v) = z;
    }
#endif

    popv(8);
    return (bodyx);
#undef specenv
#undef env1
#undef p
#undef q
#undef env
#undef body
#undef bvl
}


static          Sex
letstar_fn(Sex args, Sex env)
{
    Sex             bodyx;
    Sex            *bsave;

    if (!is_cons(args))
	return nil;
    stackcheck();
    push3(car(args), cdr(args), env);
    push2(nil, nil);		/* 
				 * p, q                      
				 */
#define p          stack[0]
#define q          stack[-1]
#define env        stack[-2]
#define body       stack[-3]
#define bvl        stack[-4]
#ifdef StandardLisp
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	Sex             z;
	q = car(bvl);
	if (is_cons(q)) {
	    z = cdr(q);
	    q = car(q);
	    if (is_cons(z))
		z = car(z);
	    else
		z = nil;
	} else
	    z = nil;
	if (!is_symbol(q)) {
	    error1("let*: bad arg list", q);
	} else {
	    if (z != nil)
		z = eval(z, env);
	    z = cons(q, z);
	    env = cons(z, env);
	}
    }
    bodyx = progn_fn(body, env);
#else
    bsave = bstack;
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	Sex             z;
	q = car(bvl);
	if (is_cons(q)) {
	    z = cdr(q);
	    q = car(q);
	    if (is_cons(z))
		z = car(z);
	    else
		z = nil;
	} else
	    z = nil;
	if (!is_symbol(q)) {
	    error1("let*: bad arg list", q);
	} else {
	    if (z != nil)
		z = eval(z, env);
	    pushb(q);
	    bstackcheck();
	    value(q) = z;
	}
    }
    bodyx = progn_fn(body, env);
    while (bstack > bsave)
	popb();
#endif
    popv(5);
    return (bodyx);

#undef p
#undef q
#undef env
#undef body
#undef bvl
}


static          Sex
plus_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (plus) => 0  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	if (is_fixnum(r) && is_fixnum(w)) {
	    long            c = mkint(r) + mkint(w);
	    long            w1 = c & fix_mask;
	    if (w1 == 0 || w1 == fix_mask)
		r = mkfxat(c);
	    else
		r = plus2(r, w);
	} else
	    r = plus2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
times_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(1);	/* 
				 * * (times) => 1  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = times2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}


static          Sex
list_fn(Sex args, Sex env)
{
    Sex             w1,
                    w2,
                    w3,
                    r = nil;
    if (!is_cons(args))
	return nil;		/* 
				 * * (list)  
				 */
    w1 = car(args);
    args = cdr(args);
    if (!is_cons(args)) {	/* 
				 * * (list w1)  
				 */
	w1 = eval(w1, env);
	w1 = Lncons(w1);
	return w1;
    }
    w2 = car(args);
    args = cdr(args);
    stackcheck();
    if (!is_cons(args)) {	/* 
				 * * (list w1 w2)  
				 */
	push2(env, w2);
	w1 = eval(w1, env);
	w2 = stack[0];
	stack[0] = w1;
	w2 = eval(w2, stack[-1]);
	w1 = Llist2(stack[0], w2);
	popv(2);
	return w1;
    }
    w3 = car(args);
    args = cdr(args);
    if (!is_cons(args)) {	/* 
				 * * (list w1 w2 w3)  
				 */
	push3(env, w2, w3);
	w1 = eval(w1, env);
	w2 = stack[-1];
	stack[-1] = w1;
	w2 = eval(w2, stack[-2]);
	w3 = stack[0];
	stack[0] = w2;
	w3 = eval(w3, stack[-2]);
	w3 = Lncons(w3);
	w1 = Llist2star(stack[-1], stack[0], w3);
	popv(3);
	return w1;
    }
    push4(args, env, w1, w2);
    w3 = eval(w3, env);
    push(w3);
    w2 = eval(stack[-1], stack[-3]);
    stack[-1] = w2;
    w1 = eval(stack[-2], stack[-3]);
    r = Lncons(w1);
    pop2(w3, w2);
    r = Llist2star(w3, w2, r);
    pop3(w1, env, args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = cons(w, r);
	pop2(args, env);
	args = cdr(args);
    }
    return Lnreverse(r);
}

static          Sex
liststar_fn(Sex args, Sex env)
{
    Sex             r = nil;
    if (!is_cons(args))
	return aerror("list*");
    stackcheck();
    do {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = cons(w, r);
	pop2(args, env);
	args = cdr(args);
    }
    while (is_cons(args));
    args = car(r);
    r = cdr(r);
    while (r != nil) {
	Sex             c = r;
	r = cdr(r);
	cdr(c) = args;
	args = c;
    }
    return args;
}


static          Sex
or_fn(Sex args, Sex env)
{
    if (!is_cons(args))
	return nil;
    stackcheck();
    for (;;) {
	Sex             v = car(args);
	args = cdr(args);
	if (!is_cons(args))
	    return eval(v, env);
	push2(args, env);
	v = eval(v, env);
	pop2(env, args);
	if (v != nil)
	    return v;
    }
}

static          Sex
prog_fn(Sex args, Sex env)
{
    Sex             p;
    if (!is_cons(args) || !is_cons(cdr(args)))
	return nil;
    stackcheck();
    push3(nil, args, env);
#define env    stack[0]
#define args   stack[-1]
#define my_tag stack[-2]
    my_tag = cons(mkfxat(0), nil);
    env = cons(my_tag, env);
    p = prog_fn_1(car(args), cdr(args), env);
    popv(3);
    return p;
#undef env
#undef args
#undef my_tag
}

Sex
prog_fn_1(Sex bvl, Sex body, Sex env)
{
    Sex             bodyx;
    Sex            *bsave;
    stackcheck();
    push3(bvl, body, env);
    push5(nil, nil, env, nil, nil);
#define specenv    stack[-1]
#define env1       stack[-2]
#define p          stack[-3]
#define q          stack[-4]
#define env        stack[-5]
#define body       stack[-6]
#define bvl        stack[-7]

#ifdef StandardLisp
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	Sex             v;
	v = car(bvl);
	if (!is_symbol(v))
	    error1("prog: bad prog list", v);
	v = cons(v, nil);
	env = cons(v, env);
    }

    bodyx = tagbody_fn(body, env);
#else
    bsave = bstack;
    for (; is_cons(bvl); bvl = cdr(bvl)) {
	Sex             v;
	v = car(bvl);
	if (!is_symbol(v))
	    error1("prog: bad prog list", v);
	pushb(v);
	bstackcheck();
	value(v) = nil;
    }

    bodyx = tagbody_fn(body, env);
    while (bstack > bsave)
	popb();
#endif

    popv(8);
    return (bodyx);



#undef specenv
#undef env1
#undef p
#undef q
#undef env
#undef body
#undef bvl
}


Sex
progn_fn(Sex args, Sex env)
{
    Sex             f;
    if (!is_cons(args))
	return nil;
    stackcheck();
    f = nil;
    for (;;) {
	f = car(args);
	args = cdr(args);
	if (!is_cons(args))
	    break;
	push2(args, env);
	eval(f, env);
	pop2(env, args);
	if (exit_reason) {
	    return nil;
	}
    }
    return eval(f, env);
}

static          Sex
prog1_fn(Sex args, Sex env)
{
    Sex             f;
    if (!is_cons(args))
	return nil;
    stackcheck();
    push2(args, env);
    f = car(args);
    f = eval(f, env);
    pop2(env, args);
    push(f);
    for (;;) {
	args = cdr(args);
	if (!is_cons(args))
	    break;
	push2(args, env);
	eval(car(args), env);
	pop2(env, args);
    }
    pop(f);
    return f;
}

Sex
prog2_fn(Sex args, Sex env)
{
    Sex             f;
    if (!is_cons(args))
	return nil;		/* 
				 * * (prog2) -> nil  
				 */
    stackcheck();
    push2(args, env);
    eval(car(args), env);
    pop2(env, args);
    args = cdr(args);
    if (!is_cons(args))
	return nil;		/* 
				 * * (prog2 x) -> nil  
				 */
    push2(args, env);
    f = car(args);
    f = eval(f, env);		/* 
				 * * second arg  
				 */
    pop2(env, args);
    push(f);
    for (;;) {
	args = cdr(args);
	if (!is_cons(args))
	    break;
	push2(args, env);
	eval(car(args), env);
	pop2(env, args);
    }
    pop(f);
    return f;
}


Sex
quote_fn(Sex args, Sex env)
{
    if (is_cons(args) && cdr(args) == nil)
	return car(args);
    return aerror("quote");
}

Sex
return_fn(Sex args, Sex env)
{
    Sex             p;
    stackcheck();
    for (p = env; is_cons(p); p = cdr(p)) {
	Sex             w = car(p);
	if (!is_cons(w))
	    continue;
	if (car(w) == mkfxat(0) && cdr(w) == nil) {
	    goto tag_found;
	}
    }
    return error1("return outside the block", nil);
  tag_found:
    if (is_cons(args)) {
	p = car(args);
	env = eval(p, env);
	exit_val = env;
    } else {
	exit_val = nil;
    }
    exit_reason = EXIT_RETURN;
    exit_tag = nil;
    return nil;
}

static          Sex
setq_fn(Sex args, Sex env)
{
    Sex             var,
                    val = nil;
    stackcheck();
    if (is_cons(args)) {
	var = car(args);
	if (!is_symbol(var) || var == nil || var == true)
	    return aerror("setq (bad variable)");
	args = cdr(args);
	if (is_cons(args)) {
	    push3(args, env, var);
	    val = car(args);
	    val = eval(val, env);
	    pop3(var, env, args);
	    args = cdr(args);
	} else
	    val = nil;
#ifdef StandardLisp
	while (is_cons(env)) {
	    Sex             c = car(env);
	    if (var == car(c)) {
		cdr(c) = val;
		return val;
	    }
	    env = cdr(env);
	}
#endif
	value(var) = val;
	/*
	 *      qtag(var) |= SYM_FLUID;
	 *      lisp_printf("   **** %s declared fluid\n", pname(var)); 
	 */
    }
    return val;
}

Sex
aref_setf_fn(Sex var, Sex ind, Sex val)
{
    push4(var, ind, val, nil);
#define var stack[-3]
#define ind stack[-2]
#define val stack[-1]
#define x   stack[0]
    ind = eval(ind, nil);
    while (ind != nil) {
	x = car(ind);
	ind = cdr(ind);
	var = eval(var, nil);
	if (ind == nil)
	    break;
	var = Lgetv(var, x);
    }
    val = eval(val, nil);
    ind = x;
#undef var
#undef ind
#undef val
#undef x
    popv(1);
    pop3(val, ind, var);
    return Lputv(var, ind, val);
}


static          Sex
setff_fn(Sex var, Sex val)
{
    Sex             v,
                    a;
    v = cadr(var);
    if (!is_symbol(v) || v == nil || v == true)
	aerror("setf (bad variable)");
    a = car(var);
#ifdef ASUKA
    if (a == aref_symbol || a == array_symbol)
	return aref_setf_fn(cadr(var), car(cdr(cdr(var))), val);
#else
    if (a == aref_symbol)
	return aref_setf_fn(cadr(var), car(cdr(cdr(var))), val);
#endif
    v = value(v);
    if (!is_cons(v))
	aerror("setf (bad construction)");
    if (a == car_symbol)
	car(v) = val;
    else if (a == cdr_symbol)
	cdr(v) = val;
    else if (a == caar_symbol)
	car(car(v)) = val;
    else if (a == cadr_symbol)
	car(cdr(v)) = val;
    else if (a == cdar_symbol)
	cdr(car(v)) = val;
    else if (a == cddr_symbol)
	cdr(cdr(v)) = val;
    else
	aerror("setf (unknown setter)");
    return val;
}

static          Sex
setf_fn(Sex args, Sex env)
{
    Sex             var,
                    val = nil;

    stackcheck();
    if (is_cons(args)) {
	var = car(args);
	push(var);
	args = cdr(args);
	if (is_cons(args)) {
	    push3(args, env, var);
	    val = car(args);
	    val = eval(val, env);
	    pop3(var, env, args);
	    args = cdr(args);
	} else
	    val = nil;
	pop(var);

	if (is_symbol(var)) {
	    if (var == nil || var == true)
		aerror("setf (bad variable)");
#ifdef StandardLisp
	    while (is_cons(env)) {
		Sex             c = car(env);
		if (var == car(c)) {
		    cdr(c) = val;
		    return val;
		}
		env = cdr(env);
	    }
#endif
	    value(var) = val;
	} else
	    return setff_fn(var, val);
    }
    return val;
}


Sex
tagbody_fn(Sex args, Sex env)
{
    Sex             f,
                    p,
                    my_env;
    stackcheck();
    f = nil;
    push2(env, args);
    for (p = args; is_cons(p); p = cdr(p)) {
	Sex             w = car(p);
	if (!is_cons(w)) {
	    Sex             w1;
	    push3(f, p, env);
	    w1 = cons(mkfxat(1), p);
	    pop(env);
	    env = cons(w1, env);
	    pop2(p, f);
	}
    }
    pop(args);
    for (p = args;; p = cdr(p)) {
	if (!is_cons(p)) {
	    pop(my_env);
	    /*
	     * while (env != my_env) { car(car(env)) = mkfxat(2); env = cdr(env); } 
	     */
	    return nil;
	}
	if (is_cons(f = car(p)) && f != nil) {
	    push3(p, env, f);
	    eval(f, env);
	    if (exit_reason) {
		if (exit_reason == EXIT_GO) {
		    exit_reason = 0;
		    pop3(f, env, p);
		    p = cdr(exit_val);
		    continue;
		} else if (exit_reason == EXIT_RETURN) {
		    exit_reason = 0;
		    popv(4);
		    return exit_val;
		}
	    }
	    pop3(f, env, p);
	}
    }
}



static          Sex
unless_fn(Sex args, Sex env)
{
    Sex             w;
    if (!is_cons(args))
	return nil;
    stackcheck();
    push2(args, env);
    w = car(args);
    w = eval(w, env);
    pop2(env, args);
    if (w != nil)
	return nil;
    else
	return progn_fn(cdr(args), env);
}





void
unwind_bind(Sex * entry_stack)
{
    Sex            *sp = bstack;
    Sex             v;
    while (sp != entry_stack) {
	v = *sp--;
        //lisp_printf("unwind_bind: %p\n", v);
	value(v) = *sp--;
    }
    bstack = sp;
}


jmp_buf        *errorset_buffer = NULL;

Sex
Lerrorset(Sex form, Sex fg1, Sex fg2)
{
    Sex             r;
    Sex            *save,
                   *bsave;
    unsigned long   flags = miscflags;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    long            trace_save = trace_count;

#if 0
	stackcheck();
	push3(form, fg1, fg2);
	lisp_printf("errorset form=%p\n", form);
	pop3(fg2,fg1,form);
#endif
    miscflags &= ~(MESSAGES_FLAG | BACKTRACE_FLAG);
    if (fg1 != nil)
	miscflags |= MESSAGES_FLAG;
    if (fg2 != nil)
	miscflags |= BACKTRACE_FLAG;
    save = stack;
    bsave = bstack;
    stackcheck();
    if (setjmp(this_level)) {
	unwind_bind(bsave);
	stack = save;
	errorset_buffer = saved_buffer;
	trace_count = trace_save;
	return nil;
    }
    errorset_buffer = &this_level;
    r = eval(form, env);
    errorset_buffer = saved_buffer;
    miscflags = (flags & ~GC_MSG_BITS) | (miscflags & GC_MSG_BITS);
    r = Lncons(r);
    return r;
}



Sex
Lerrorset1(Sex form)
{
    return Lerrorset(form, nil, nil);
}




static          Sex
when_fn(Sex args, Sex env)
{
    Sex             w;
    if (!is_cons(args))
	return nil;
    stackcheck();
    push2(args, env);
    w = car(args);
    w = eval(w, env);
    pop2(env, args);
    if (w == nil)
	return nil;
    else
	return progn_fn(cdr(args), env);
}



setup_type const eval_setup[] = {
    {"apply", (lisp_proc *) Lapply, 2},
    {"apply0", (lisp_proc *) Lapply0, 1},
    {"apply1", (lisp_proc *) Lapply1, 2},
    {"apply2", (lisp_proc *) Lapply2, 3},
    {"apply3", (lisp_proc *) Lapply3, 4},
    {"evlis", (lisp_proc *) Levlis, 1},
    {"funcall", (lisp_proc *) Lfuncall, 2},
    {"funcall0", (lisp_proc *) Lfuncall0, 1},
    {"funcall1", (lisp_proc *) Lfuncall1, 2},
    {"funcall2", (lisp_proc *) Lfuncall2, 3},
    {"funcall3", (lisp_proc *) Lfuncall3, 4},
    {"funcall4", (lisp_proc *) Lfuncall4, 5},
    {"funcall5", (lisp_proc *) Lfuncall5, 6},
    {"verbos", (lisp_proc *) Lverbos, 1},
    {"errorset1", (lisp_proc *) Lerrorset1, 1},
    {"errorset", (lisp_proc *) Lerrorset, 3},
    {"reclaim", (lisp_proc *) Lgc, 0},
    {"gc", (lisp_proc *) Lgc, 0},
    {"and", (lisp_proc *) and_fn, -1},
    {"catch", (lisp_proc *) catch_fn, -1},
#ifdef ASUKA
    {"switch", (lisp_proc *) switch_fn, -1},
    {"case", (lisp_proc *) case_fn, -1},
#endif
    {"cond", (lisp_proc *) cond_fn, -1},
    {"eval-when", (lisp_proc *) eval_when_fn, -1},
    {"function", (lisp_proc *) function_fn, -1},
    {"go", (lisp_proc *) go_fn, -1},
    {"if", (lisp_proc *) if_fn, -1},
#ifdef ASUKA
    {"while", (lisp_proc *) while_fn, -1},
    {"repeat", (lisp_proc *) repeat_fn, -1},
    {"for", (lisp_proc *) for_fn, -1},
    {"foreach", (lisp_proc *) foreach_fn, -1},
#endif
    {"let*", (lisp_proc *) letstar_fn, -1},
    {"~block", (lisp_proc *) block_fn, -1},
    {"de", (lisp_proc *) de_fn, -1},
    {"dm", (lisp_proc *) dm_fn, -1},
    {"ds", (lisp_proc *) ds_fn, -1},
    {"df", (lisp_proc *) df_fn, -1},
    {"~let", (lisp_proc *) let_fn, -1},
    {"or", (lisp_proc *) or_fn, -1},
    {"prog", (lisp_proc *) prog_fn, -1},
    {"prog1", (lisp_proc *) prog1_fn, -1},
    /*
     * {"prog2", (lisp_proc *)prog2_fn, -1},
     */
    /*
     * {"progn",     (lisp_proc *)progn_fn, -1}, 
     */
    /*
     * {"quote",     (lisp_proc *)quote_fn, -1}, 
     */
    {"return", (lisp_proc *) return_fn, -1},
    {"setq", (lisp_proc *) setq_fn, -1},
    {"setf", (lisp_proc *) setf_fn, -1},
    {"tagbody", (lisp_proc *) tagbody_fn, -1},
    {"catch", (lisp_proc *) catch_fn, -1},
    {"throw", (lisp_proc *) throw_fn, -1},
    {"with-timeout", (lisp_proc *) with_timeout, -1},
    {"with_timeout", (lisp_proc *) with_timeout, -1},
    {"unwind-protect", (lisp_proc *) unwind_protect, -1},
    {"unless", (lisp_proc *) unless_fn, -1},
    {"when", (lisp_proc *) when_fn, -1},
    {"list", (lisp_proc *) list_fn, -1},
    {"list*", (lisp_proc *) liststar_fn, -1},
    {"plus", (lisp_proc *) plus_fn, -1},
    {"times", (lisp_proc *) times_fn, -1},
    {NULL, 0, 0}
};
