/*
 * factor.c      (C) Copyright 2015 by Fujio Kako, Nara Women's University.
 * Polynomial Factorization Functions.
 *      $Header: /Users/kako/asuka/src/RCS/factor.c,v 2.19 2016/03/28 06:02:37 kako Exp kako $
 */

static char const rcsid[] =
    "$Id: factor.c,v 2.19 2016/03/28 06:02:37 kako Exp kako $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "entries.h"
#include "misc.h"
#include "fns.h"
#include "polynomial.h"
#include "tri.h"
#include "garith.h"

#define cddr(x)	cdr(cdr(x))
#define cdar(x)	cdr(car(x))

static Sex      Afactor(Sex a);
Sex             Asquarefree(Sex a);
Sex             Asquarefree(Sex a);
static Sex      Azfactor(Sex a);
static Sex      factorlist2prefix(Sex a);

extern Sex      Lncons(Sex a);
extern Sex      Lnconc(Sex a, Sex b);
extern Sex      Laconc(Sex a, Sex b);
extern Sex      Lclone(Sex a);
extern Sex      Lmkarray1(Sex n);
extern Sex      Alist2array(Sex a, Sex b);

static Sex      factor(Sex);
static Sex      SquareFree(Sex);
static Sex      SquareFree2(Sex);

static Sex      Afactor_mod(Sex f);
static Sex      Afactor_ff(Sex f);
static Sex      Asquarefree2(Sex a);
static bool     square_free_mod(Sex a, Sex p);
static Sex      factor_mod(Sex f, Sex p);
static Sex      sqfree(Sex a);
static Sex      ddd_factor(Sex a);
static Sex      berlekamp(Sex f, Sex p);
static Sex      MignotteBound(Sex f);
static Sex      uhensel(Sex f, Sex p, Sex g, Sex a, Sex b);
static Sex      FormMatrix(Sex f, Sex q);
static Sex      NullSpaceBase(Sex c);
static Sex      Factor_mod(Sex V, long k, Sex q, Sex f);
static Sex      isqrt(Sex s);
static bool     zeroRow(Sex M, long n, long i);
static Sex      Combine_Factors(Sex F, Sex U, Sex q, long n, long m);
static Sex      Combine_nest(Sex F, Sex U, Sex q, long k, long c, long n,
			     long m, Sex a);

static Sex      eea(Sex a, Sex b);
static Sex      multieea(Sex a, Sex x);
static Sex      diophant(Sex a1, Sex a2, Sex c, Sex x);
static Sex      zfactor(Sex a);
static Sex      mcfactors(Sex a, Sex L);
static Sex      mcfactor(long a, long p);
static Sex      big_mcfactor(Sex a, long i);




extern Sex      Amcoeff(Sex a);
extern Sex      Lexpt(Sex a, Sex n);
extern Sex      gexpt(Sex a, Sex n);
extern Sex      Mpp(Sex a);
extern Sex      Mvar(Sex a);
extern Sex      Mdeg(Sex a);
extern Sex      Adeg(Sex f, Sex x);
extern Sex      Lrandom(Sex a);
extern Sex      Lnext_prime(Sex p);
extern Sex      Mlcoeff(Sex a);
extern Sex      Mlpower(Sex a);
extern Sex      Mlterm(Sex a);
extern Sex      Alcoeff(Sex a, Sex x);
extern Sex      Alpower(Sex a, Sex x);
extern Sex      Alterm(Sex a, Sex x);
extern Sex      Amkmodular(Sex a, Sex p);
extern Sex      Aunmodular(Sex a);
extern Sex      Adiff(Sex a, Sex c);
extern Sex      new_vector(long n);
extern Sex      Lident(Sex a);
extern Sex      Labs(Sex a);
extern Sex      Arowvector(Sex m, Sex i);
extern Sex      Acoeff(Sex a, Sex b);
extern Sex      make_matrix(Sex n, Sex m);
extern Sex      Llessp(Sex a, Sex b);
extern long     int_gcd(long a, long b);
extern Sex      gcdjj(Sex a, Sex b);

extern long     last_prime;
extern long     last_prime_squared;
long            maxtrys = 10;
long            confidence = 40;

/**************************************************************/
/**************************************************************/

static          Sex
Afactor(Sex g)
{
    g = factor(g);
    g = factorlist2prefix(g);
    return g;
}


static          Sex
factorlist2prefix(Sex g)
{
    Sex             g1;
    push2(g, nil);
#define g stack[-1]
#define p stack[0]
    p = nil;
    while (g != nil) {
//lisp_printf("g=%p\n", g);
	g1 = car(g);
	g = cdr(g);
#if 0
	g1 = cons(ASUKA_CHAR_OF('^'), g1);
#else
	if (cadr(g1) == mkfxat(1))
	    g1 = car(g1);
	else
	    g1 = cons(ASUKA_CHAR_OF('^'), g1);
#endif
	p = cons(g1, p);
    }
    if(cdr(p) != nil) {
        p = Lnreverse(p);
        p = cons(ASUKA_CHAR_OF('*'), p);
    } else
	p = car(p);
#undef g
    g = new_prefix();
    prefix_form(g) = p;
#undef p
    popv(2);
    return g;
}

Sex
Asquarefree(Sex a)
{
    return SquareFree(a);
}

Sex
Asquarefree2(Sex a)
{
    return SquareFree2(a);
}

static          Sex
Azfactor(Sex a)
{
    a = zfactor(a);
    a = factorlist2prefix(a);
    return a;
}

static          Sex
zfactor(Sex a)
{
    Sex             q;
    long            n,
                    m;
    long            i;
    Sex             j;
    stackcheck();
    push(nil);
#define L stack[0]
    if (is_fixnum(a)) {
	n = mkint(a);
	m = 0;
	while (n % 2 == 0) {
	    m++;
	    n /= 2;
	}
	if (m) {
	    q = Lncons(mkfxat(m));
	    q = cons(mkfxat(2), q);
	    L = Laconc(L, q);
	}
	m = 0;
	while (n % 3 == 0) {
	    m++;
	    n /= 3;
	}
	if (m) {
	    q = Lncons(mkfxat(m));
	    q = cons(mkfxat(3), q);
	    L = Laconc(L, q);
	}
	i = 3;
	while (i * i < n) {
	    i += 2;
	    m = 0;
	    while (n % i == 0) {
		m++;
		n /= i;
	    }
	    if (m) {
		q = Lncons(mkfxat(m));
		q = cons(mkfxat(i), q);
		L = Laconc(L, q);
	    }
	}
	if (n != 1) {
	    q = Lncons(mkfxat(1));
	    q = cons(mkfxat(n), q);
	    L = Laconc(L, q);
	}
	pop(q);
	return q;
    }
#undef L
    push3(a, nil, nil);
#define n stack[-2]
#define p stack[-1]
#define L stack[0]
    m = 0;
    while (Levenp(n) != nil) {
	m++;
	n = quot2ji(n, mkfxat(2));
    }
    if (m) {
	q = Lncons(mkfxat(m));
	q = cons(mkfxat(2), q);
	L = Laconc(L, q);
    }
    p = prime_list;
    while (p != nil) {
	j = car(p);
	p = cdr(p);
	m = 0;
	while (rem2ji(n, j) == mkfxat(0)) {
	    m++;
	    n = quot2ji(n, j);
	}
	if (m) {
	    q = Lncons(mkfxat(m));
	    q = cons(j, q);
	    L = Laconc(L, q);
	}
    }
    if (n != mkfxat(1)) {
	q = mcfactors(n, nil);
	// q = Lncons(mkfxat(1));
	// q = cons(n, q);
	// L = Laconc(L, q);
	L = Lnconc(L, q);
    }
    q = L;
    popv(3);
#undef n
#undef p
#undef L
    return q;
}


static          Sex
mcfactors(Sex a, Sex L)
{
    long            i;
    stackcheck();
    push2(a, L);
#define a stack[-1]
#define L stack[0]
    if (Lprimep(a) != nil) {
	a = Llist2(a, mkfxat(1));
	L = Laconc(L, a);
    } else {
	for (i = 1; i < maxtrys; i++) {
	    if (is_fixnum(a))
		a = mcfactor(mkint(a), i);
	    else
		a = big_mcfactor(a, i);

	    if (is_cons(a))
		break;
	}
	if (is_cons(a)) {
	    if (Llessp(car(a), cdr(a)) != nil) {
		L = mcfactors(car(a), L);
		L = mcfactors(cdr(a), L);
	    } else {
		L = mcfactors(cdr(a), L);
		L = mcfactors(car(a), L);
	    }
	} else {
	    a = Llist2(a, mkfxat(1));
	    L = Laconc(L, a);
	}
    }
#undef a
#undef L
    a = stack[-1];
    popv(2);
    return a;
}

static          Sex
mcfactor(long n, long p)
{
    long            g,
                    k,
                    m,
                    q,
                    r,
                    x,
                    y,
                    ys;
    long            ms;
    long            i;

    m = 20;
    y = 0;
    r = q = 1;
    do {
	x = y;
	for (i = 1; i <= r; i++) {
	    y = umulrem(y, y, n);
	    y = (y + p) % n;
	}
	k = 0;
	do {
	    ys = y;
	    ms = m < (r - k) ? m : r - k;
	    for (i = 1; r <= ms; i++) {
		y = umulrem(y, y, n);
		y = (y + p) % n;
		q = umulrem(q, labs(x - y), n);
	    }
	    g = int_gcd(q, n);
	    k = k + m;
	} while (k < r && g <= 1);
	r = 2 * r;
    } while (g <= 1);
    if (g == n) {
	do {
	    ys = umulrem(ys, ys, n);
	    ys = (ys + p) % n;
	    g = int_gcd(labs(y - ys), n);
	} while (g <= 1);
    }
    if (g == n)
	return mkfxat(n);
    m = n / g;
    return cons(mkfxat(g), mkfxat(m));
}



static          Sex
big_mcfactor(Sex n, long p)
{
    Sex             a;
    long            i,
                    k,
                    m,
                    r;
    long            ms;

    stackcheck();
    push(n);
    pushv(5);
#define n stack[-5]
#define q stack[-4]
#define x stack[-3]
#define y stack[-2]
#define ys stack[-1]
#define g stack[-0]
    m = 20;
    y = mkfxat(0);
    r = 1;
    q = mkfxat(1);
    do {
	x = y;
	for (i = 1; i <= r; i++) {
	    y = times2jj(y, y);
	    y = plus2ji(y, mkfxat(p));
	    y = rem2jj(y, n);
	}
	k = 0;
	do {
	    ys = y;
	    ms = m < (r - k) ? m : r - k;
	    for (i = 1; i <= ms; i++) {
		y = times2jj(y, y);
		y = plus2ji(y, mkfxat(p));
		y = rem2jj(y, n);
		q = times2jj(q, labs(x - y));
		q = rem2jj(q, n);
	    }
	    g = gcdjj(q, n);
	    k = k + m;
	} while (k < r && (g == mkfxat(0) || g == mkfxat(1)));
	r = 2 * r;
    } while (g == mkfxat(0) || g == mkfxat(1));
    if (Leqn(g, n) != nil) {
	do {
	    ys = times2jj(ys, ys);
	    ys = plus2ji(ys, mkfxat(p));
	    ys = rem2jj(ys, n);
	    g = gcdjj(labs(y - ys), n);
	} while (g == mkfxat(0) || g == mkfxat(1));
    }
    if (Leqn(q, n) != nil) {
	a = n;
    } else {
	m = quot2jj(n, g);
	a = cons(g, m);
    }
#undef n
#undef q
#undef x
#undef y
#undef ys
#undef g
    popv(6);
    return a;
}




/**************************************************************/
/**************************************************************/
#if 0
static          Sex
Afactor_mod(Sex f)
{
    long            p;
    Sex             a;

    stackcheck();
    push(f);
#define f stack[-0]

    p = 3;
    while (!square_free_mod(f, mkfxat(p)))
	p = Lnext_prime(mkfxat(p));
    a = factor_mod(f, mkfxat(p));
    popv(1);
#undef f
    return a;
}
#endif

#if 0
static          Sex
Afactor_ff(Sex f)
{
    Sex             p;
    stackcheck();
    push5(f, nil, nil, nil, nil);
    push3(nil, nil, nil);
#define f stack[-7]
#define l stack[-6]
#define r stack[-5]
#define s stack[-4]
#define t stack[-3]
#define a stack[-2]
#define b stack[-1]
#define e stack[-0]
    l = sqfree(f);
    r = nil;
    while (t != nil) {
	s = car(t);
	a = car(s);
	e = cadr(s);
	t = cdr(t);
	t = ddd_factor(a);
	s = Llist2(b, e);
	r = Lappend(r, s);
    }
    p = r;
    popv(8);
#undef f
#undef l
#undef r
#undef s
#undef t
#undef a
#undef b
#undef e
    return p;
}
#endif

static          Sex
univariate_factor(Sex F)
{
    Sex             s;
    long            i,
                    n;
    stackcheck();
    push4(F, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
#define F stack[-13]
#define H stack[-12]
#define x stack[-10]
#define f stack[-9]
#define g stack[-8]
#define h stack[-7]
#define p stack[-6]
#define m stack[-5]
#define b stack[-3]
#define j stack[-2]
#define I stack[-1]
#define Hs stack[-0]
    p = Mcont(F);
    F = gratio2(F, p);
    x = Mvar(F);
    H = SquareFree(F);
    n = Llength(H);
    I = nil;
    if (!onep(p)) {
	// s = Llist2(p, mkfxat(1));
	// I = Lncons(s);
	I = zfactor(p);
    }
    Hs = H;
    // lisp_printf("factor: Hs = %p\n", Hs);
    while (Hs != nil) {
	j = car(Hs);
	Hs = cdr(Hs);
	f = car(j);
	if (!is_poly(f))
	    continue;
	if (Mdeg(f) < mkfxat(2)) {
	    I = Laconc(I, j);
	    continue;
	}
	p = Lrandom(mkfxat(100));
	while (1) {
	    p = Lnext_prime(p);
	    m = Mlcoeff(f);
	    if (gzerop(gremaind2(f, p)))
		continue;
	    g = Amkmodular(f, p);
	    s = Adiff(g, x);
	    s = Mgcd(g, s);
	    if (!is_poly(s))
		break;
	}
	// lisp_printf("factor: prime = %p\n", p);
	g = berlekamp(f, p);
	// lisp_printf("factor: g=%p\n", g);
	n = mkint(Llength(g));
	if (n == 1) {
	    s = Llist2(car(g), cadr(j));
	    I = Laconc(I, s);
	} else {
	    b = MignotteBound(f);
	    // lisp_printf("mignotte bound = %p\n", b);
	    h = uhensel(f, p, g, nil, b);
	    // lisp_printf("hense = %p\n", h);
	    n = mkint(car(array_dim(h)));
	    for (i = 0; i < n; i++) {
		s = array_value(h, i);
		s = Mpp(s);
		s = Llist2(s, cadr(j));
		I = Laconc(I, s);
	    }
	}
    }

    s = I;
    popv(14);
#undef F
#undef H
#undef x
#undef f
#undef g
#undef h
#undef p
#undef m
#undef b
#undef j
#undef I
#undef Hs
    return s;
}

/**************************************************************/
/**************************************************************/
static          bool
square_free_mod(Sex f, Sex q)
{
    bool            r;
    stackcheck();
    push4(f, q, nil, nil);
#define f stack[-3]
#define q stack[-2]
#define x stack[-1]
#define g stack[-0]
    f = Amkmodular(f, q);
    x = Mvar(f);
    g = Adiff(f, x);
    g = Mgcd(f, g);
    r = !is_poly(g);
    popv(4);
#undef f
#undef q
#undef x
#undef g
    return r;
}

static          Sex
SquareFree(Sex a)
{
    Sex             s;
    long            i;
    stackcheck();
    push(a);
    pushv(8);
#define a stack[-8]
#define o stack[-7]
#define b stack[-6]
#define c stack[-5]
#define w stack[-4]
#define x stack[-3]
#define y stack[-2]
#define z stack[-1]
#define g stack[-0]
    i = 1;
    o = nil;
    x = Mvar(a);
    b = Adiff(a, x);
    c = Mgcd(a, b);
    w = gquot2(a, c);
    while (is_poly(c)) {
	y = Mgcd(w, c);
	z = gquot2(w, y);
	s = Llist2(z, mkfxat(i));
	o = cons(s, o);
	i++;
	w = y;
	c = gquot2(c, y);
    }
    s = Llist2(w, mkfxat(i));
    o = cons(s, o);
#undef a
    a = o;
#undef o
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
#undef g
    popv(9);
    return a;
}

static          Sex
SquareFree2(Sex a)
{
    Sex             s;
    long            i;
    stackcheck();
    push(a);
    pushv(8);
#define a stack[-8]
#define o stack[-7]
#define b stack[-6]
#define c stack[-5]
#define w stack[-4]
#define x stack[-3]
#define y stack[-2]
#define z stack[-1]
#define g stack[-0]
    i = 1;
    o = nil;
    x = Mvar(a);
    b = Adiff(a, x);
    c = Mgcd(a, b);
    if (c == mkfxat(1))
	w = a;
    else {
	w = gquot2(a, c);
	y = gquot2(b, c);
	s = Adiff(w, x);
	z = gdiff2(y, s);
	while (!gzerop(z)) {
	    g = Mgcd(w, z);
	    s = Llist2(g, mkfxat(i));
	    o = cons(s, o);
	    i++;
	    w = gquot2(w, g);
	    y = gquot2(z, g);
	    s = Adiff(w, x);
	    z = gdiff2(y, s);
	}
    }
    s = Llist2(w, mkfxat(i));
    o = cons(s, o);
#undef a
    a = o;
#undef o
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
#undef g
    popv(9);
    return a;
}


static          Sex
squarerootFF(Sex a, Sex q)
{
    Sex             s,
                    u;
    long            i,
                    j,
                    n;
    long            k;

    k = mkint(q);
    stackcheck();
    push(a);
    pushv(5);
#define a stack[-6]
#define b stack[-4]
#define c stack[-3]
#define x stack[-2]
#define u stack[-1]
    x = Mvar(a);
    c = Amcoeff(a);
    n = mkint(Llength(c));
    n = (n - 1) / k;
    b = mkint(0);
    for (i = 0; i <= n; i++) {
	u = gexpt(x, mkfxat(i));
	s = array_value(c, i * k);
	s = gtimes2(s, u);
	b = gplus2(b, s);
    }
    s = b;
#undef a
#undef b
#undef c
#undef x
#undef u
    popv(6);
    return s;
}

static          Sex
SquareFreeFF(Sex a, Sex q)
{
    Sex             s;
    long            i;
    stackcheck();
    push2(a, q);
    pushv(7);
#define a stack[-8]
#define q stack[-7]
#define o stack[-6]
#define b stack[-5]
#define c stack[-4]
#define w stack[-3]
#define x stack[-2]
#define y stack[-1]
#define z stack[-0]
    i = 1;
    o = nil;
    x = Mvar(a);
    a = Amkmodular(a, q);
    b = Adiff(a, x);
    if (b != mkfxat(0)) {
	c = Mgcd(a, b);
	w = gquot2(a, c);
	while (w != mkfxat(1)) {
	    y = Mgcd(w, c);
	    z = gquot2(w, y);
	    s = Llist2(z, mkfxat(i));
	    o = cons(s, o);
	    i++;
	    w = y;
	    c = gquot2(c, y);
	}
	if (c != mkfxat(1)) {
	    c = squarerootFF(c, q);
	    s = SquareFreeFF(c, q);
	    s = Llist2(s, q);
	    o = cons(s, o);
	}
    } else {
	a = squarerootFF(c, q);
	s = SquareFreeFF(a, q);
	s = Llist2(s, q);
	o = cons(s, o);
    }
#undef a
#undef q
    a = o;
#undef o
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
    popv(9);
    return a;
}
/**************************************************************/
/**************************************************************/
static          Sex
berlekamp(Sex f, Sex q)
{
    Sex             s;
    long            i,
                    k,
                    m;
    long            n;
    stackcheck();
    push2(f, q);
    pushv(13);
#define f stack[-14]
#define q stack[-13]
#define v stack[-12]
#define g stack[-11]
#define factors stack[-10]
#define prime_factors stack[-9]
#define h stack[-4]
#define Q stack[-3]
#define K stack[-2]
#define V stack[-1]
#define hh stack[-0]

    n = mkint(Mdeg(f));
    Q = FormMatrix(f, q);
    s = Lident(mkfxat(n));
    s = gdiff2(Q, s);
    v = NullSpaceBase(s);
    // lisp_printf("nullspace = %p\n", v);
    k = mkint(Llength(v));
    h = Mvar(f);
    h = symbol_to_polynomial(h);
    V = new_vector(k);
    hh = new_vector(n);
    for (i = 0; i < n; i++) {
	s = gexpt(h, mkfxat(i));
	vect_value(hh, i) = s;
    }
    for (i = 1; i <= k; i++) {
	s = gtimes2(hh, car(v));
	s = Amkmodular(s, q);
	vect_value(V, i - 1) = s;
	v = cdr(v);
    }
    // lisp_printf("V=%p\n", V);
    factors = Lncons(f);
    prime_factors = nil;
    m = mkint(Llength(prime_factors)) + mkint(Llength(factors));
    while (m < k) {
	// lisp_printf("m = %d, k = %d\n", m, k);
	f = car(factors);
	factors = cdr(factors);
	// lisp_printf("berlekamp: f=%p\n", f);
	// lisp_printf("k = %d\n", k);
	// lisp_printf("factors = %p\n", factors);
	g = Factor_mod(V, k, q, f);
	// lisp_printf("factor mod g = %p\n", g);
	if (g != nil) {
	    factors = Lappend(factors, g);
	    m = m + mkint(Llength(g));
	} else {
	    s = Lncons(f);
	    // prime_factors = Laconc(prime_factors, f);
	    prime_factors = Lappend(prime_factors, s);
	    // m = m + mkint(Llength(f));
	}
	m = mkint(Llength(prime_factors)) + mkint(Llength(factors));
    }
    s = Lappend(factors, prime_factors);
#undef f
#undef q
#undef v
#undef g
#undef factors
#undef prime_factors
#undef n
#undef h
#undef Q
#undef K
#undef V
#undef hh
    popv(15);
    return s;
}

static          Sex
Factor_mod(Sex V, long k, Sex q, Sex f)
{
    long            i,
                    j,
                    l;
    Sex             s;

    l = mkint(q);
    stackcheck();
    push3(V, q, f);
    pushv(3);
#define V stack[-5]
#define q stack[-4]
#define f stack[-3]
#define g stack[-2]
    f = Amkmodular(f, q);
    for (i = 1; i <= k; i++) {
	s = vect_value(V, i - 1);
	if (!is_poly(s))
	    continue;
	for (j = 0; j < l; j++) {
	    s = vect_value(V, i - 1);
	    s = gdiff2(s, mkfxat(j));
	    g = Mgcd(s, f);
	    if (is_poly(g)) {
		s = gratio2(f, g);
		if (is_poly(s)) {
		    s = Llist2(g, s);
		    popv(6);
		    return s;
		}
	    }
	}
    }
    popv(6);
#undef V
#undef q
#undef f
#undef g
    return nil;
}

static          Sex
FormMatrix(Sex F, Sex q)
{
    Sex             s;
    long            i,
                    j;
    long            n,
                    m;
    stackcheck();
    push2(F, q);
    pushv(9);
#define F stack[-10]
#define q stack[-9]
#define f stack[-6]
#define k stack[-4]
#define a stack[-3]
#define vq stack[-2]
#define v stack[-1]
#define M stack[-0]
    f = Amkmodular(F, q);
    v = Mvar(F);
    n = mkint(Adeg(f, v));
    M = make_matrix(n, n);
    k = mkfxat(1);
    // matrix_value(M, 1,1 ) = mkfxat(1);
    matrix_value(M, 0) = mkfxat(1);
    vq = make_polynomial(v, mkfxat(1), q);
    if (n >= q)
	vq = gremaind2(vq, f);
    for (i = 1; i < n; i++) {
	k = gtimes2(k, vq);
	k = gremaind2(k, f);
	a = Acoeff(k, v);
	m = mkint(car(array_dim(a)));
	for (j = 1; j <= m; j++) {
	    s = array_value(a, j - 1);
	    matrix_value(M, n * (i) + j - 1) = s;
	}
    }
    s = M;
#undef F
#undef q
#undef f
#undef k
#undef a
#undef vq
#undef v
#undef M
    popv(11);
    return s;
}

static          Sex
NullSpaceBase(Sex M)
{
    Sex             s,
                    u;
    long            n,
                    m,
                    i,
                    j,
                    k,
                    l;

    stackcheck();
    push(M);
    pushv(8);
#define M stack[-8]
#define v stack[-7]
#define J stack[-5]
#define p stack[-4]
    // s = Asize(M);
    // m = mkint(car(s));
    // n = mkint(cadr(s));
    m = mkint(matrix_dimx(M));
    n = mkint(matrix_dimy(M));
    for (k = 1; k <= n; k++) {
	for (i = k; i <= n; i++) {
	    s = matrix_value(M, (k - 1) * m + i - 1);
	    if (!gzerop(s))
		break;
	}
	if (i <= n) {
	    p = matrix_value(M, (i - 1) * n + i - 1);
	    for (l = 1; l <= n; l++) {
		s = matrix_value(M, (l - 1) * n + i - 1);
		s = gratio2(s, p);
		matrix_value(M, (l - 1) * n + i - 1) = s;
	    }
	    for (j = 1; j <= n; j++) {
		J = matrix_value(M, (j - 1) * n + k - 1);
		matrix_value(M, (j - 1) * n + k - 1) =
		    matrix_value(M, (j - 1) * n + i - 1);
		matrix_value(M, (j - 1) * n + i - 1) = J;
	    }
	    for (i = 1; i <= n; i++) {
		if (i == k)
		    continue;

		p = matrix_value(M, (k - 1) * n + i - 1);
		for (l = k; l <= n; l++) {
		    s = matrix_value(M, (l - 1) * n + k - 1);
		    s = gtimes2(s, p);
		    u = matrix_value(M, (l - 1) * n + i - 1);
		    s = gdiff2(u, s);
		    matrix_value(M, (l - 1) * n + i - 1) = s;
		}
	    }
	}
    }
    v = nil;
    for (i = 1; i <= n; i++) {
	s = matrix_value(M, (i - 1) * n + i - 1);
	s = gdiff2(s, mkfxat(1));
	matrix_value(M, (i - 1) * n + i - 1) = s;
    }
    for (i = 1; i <= n; i++) {
	if (!zeroRow(M, n, i)) {
	    s = Arowvector(M, mkfxat(i));
	    v = Laconc(v, s);
	}
    }
    s = v;
#undef M
#undef v
#undef J
#undef p
    popv(9);
    return s;
}

static          bool
zeroRow(Sex M, long n, long i)
{
    long            k;
    Sex             s;

    stackcheck();
    push(M);
    for (k = 1; k <= n; k++) {
	M = stack[0];
	s = matrix_value(M, (i - 1) * n + k - 1);
	if (!gzerop(s)) {
	    popv(1);
	    return NO;
	}
    }
    popv(1);
    return YES;
}


/**************************************************************/
/**************************************************************/

/*
 * univariate hensel construction 
 */

static          Sex
MignotteBound(Sex f)
{
    Sex             r;

    if (!is_poly(f))
	return Labs(f);

    stackcheck();
    push(f);
    pushv(6);
#define f stack[-6]
#define x stack[-5]
#define a stack[-4]
#define i stack[-3]
#define n stack[-2]
#define s stack[-1]
#define m stack[-0]



    x = Mvar(f);
    m = Mdeg(f);
    a = Acoeff(f, x);
    n = mkint(car(array_dim(a)));
    s = mkfxat(0);
    for (i = 0; i < n; i++) {
	r = array_value(a, i);
	r = gtimes2(r, r);
	s = gplus2(s, r);
    }

    s = isqrt(s);
    r = array_value(a, n - 1);
    r = quot2(s, r);
    s = plus2(r, mkfxat(1));
    m = quot2(plus2(m, mkfxat(1)), mkfxat(2));
    r = Lexpt(mkfxat(2), m);
    r = times2(s, r);
#undef f
#undef x
#undef a
#undef i
#undef n
#undef s
#undef m
    popv(7);
    return r;
}


static          Sex
isqrt(Sex m)
{				/* integer square root of m */
    Sex             s;
    int             j;
    if (m == mkfxat(0))
	return m;
    stackcheck();
    push(m);
    pushv(3);
#define m stack[-3]
#define x stack[-2]
#define x1 stack[-1]
#define bk stack[-0]
    j = number_log2(m) + 2;
    j = j / 2;
    x1 = mkfxat(j);
    x = Lexpt(mkfxat(2), x1);
    s = quot2(m, x);
    bk = diff2(x, s);
    while (plusp(bk)) {
	s = plus2(bk, mkfxat(1));
	s = quot2(s, mkfxat(2));
	x = diff2(x, s);
	s = quot2(m, x);
	bk = Adiff(x, s);
    }
    s = x;
    popv(4);
#undef m
#undef x
#undef x1
#undef bk
    return s;
}

static          Sex
iroot(Sex m, Sex n)
{				/* integer n-th root of m */
    Sex             s;
    long            l;
    if (m == mkfxat(0))
	return m;
    stackcheck();
    push2(m, n);
    pushv(4);
#define m stack[-5]
#define n stack[-4]
#define n1 stack[-3]
#define x stack[-2]
#define x1 stack[-1]
#define bk stack[-0]
    l = number_log2(m) + 2;
    x1 = quot2(mkfxat(l), n);
    x = Lexpt(mkfxat(2), x1);
    n1 = diff2(n, mkfxat(1));
    x1 = Lexpt(x, n1);

    s = quot2(m, x1);
    bk = diff2(x, s);
    while (plusp(bk)) {
	s = diff2(n, mkfxat(1));
	s = plus2(bk, mkfxat(1));
	bk = quot2(s, n);
	x = diff2(x, bk);
	x1 = Lexpt(x, n1);
	s = quot2(m, x1);
	bk = Adiff(x, s);
    }
    s = x;
    popv(6);
#undef m
#undef n
#undef n1
#undef x
#undef x1
#undef bk
    return s;
}


/*
 * Univariate Hensel Lifting Algorithm 
 */


static          Sex
EEA(Sex a, Sex b, Sex x)
{
    Sex             u;
    long            n;

    stackcheck();
    push3(a, b, x);
    pushv(14);
#define a stack[-16]
#define b stack[-15]
#define x stack[-14]
#define s stack[-13]
#define t stack[-12]
#define c stack[-11]
#define d stack[-10]
#define c1 stack[-0]
#define c2 stack[-8]
#define d1 stack[-7]
#define d2 stack[-6]
#define r stack[-5]
#define r1 stack[-4]
#define r2 stack[-3]
#define v stack[-2]
#define q stack[-1]
#define g stack[-0]
    u = Alcoeff(a, x);
    c = gratio2(a, u);
    u = Alcoeff(b, x);
    d = gratio2(b, u);
    c1 = d2 = mkfxat(1);
    d1 = c2 = mkfxat(0);

    n = 0;
    while (!zerop(d)) {
	n++;
	q = grquotient2(c, d, x);
	u = gtimes2(q, d);
	r = gdiff2(c, u);
	u = gtimes2(q, d1);
	r1 = gdiff2(c1, u);
	u = gtimes2(q, d2);
	r2 = gdiff2(c2, u);
	c = d;
	c1 = d1;
	c2 = d2;
	d = r;
	d1 = r1;
	d2 = r2;
    }
    /*
     * normalize GCD 
     */
    u = Alcoeff(c, x);
    g = gratio2(c, u);
    v = Alcoeff(c, x);
    s = Alcoeff(a, x);
    s = gtimes2(s, v);
    t = Alcoeff(b, x);
    t = gtimes2(t, v);
    s = gratio2(c1, s);
    t = gratio2(c2, t);
    u = Llist3(g, s, t);
    popv(17);
#undef a
#undef b
#undef x
#undef s
#undef t
#undef c
#undef d
#undef c1
#undef c2
#undef d1
#undef d2
#undef r
#undef r1
#undef r2
#undef v
#undef q
#undef g
    return u;
}

static          Sex
replace_lc(Sex f, Sex a, Sex x)
{
    Sex             u;
    stackcheck();
    push3(f, a, x);
#define f stack[-2]
#define a stack[-1]
#define x stack[-0]
    u = Alpower(f, x);
    a = gtimes2(a, u);
    u = Alterm(f, x);
    f = gdiff2(f, u);
    u = gplus2(f, a);
#undef f
#undef a
#undef x
    popv(3);
    return u;
}

static          Sex
multivariate_replace_lc(Sex f, Sex a)
{
    Sex             u;
    stackcheck();
    push3(f, a, nil);
#define f stack[-2]
#define a stack[-1]
#define c stack[-0]
    c = Mlcoeff(f);
    if (is_poly(c)) {
	c = multivariate_replace_lc(c, a);
    } else {
	c = a;
    }
    u = Mlpower(f);
    c = gtimes2(c, u);
    u = Mlterm(f);
    f = gdiff2(f, u);
    u = gplus2(f, c);
    popv(3);
#undef f
#undef a
#undef c
    return u;
}

static          Sex
UnivariateHensel(Sex a, Sex p, Sex u1, Sex w1, Sex B, Sex gamma)
{
    Sex             z;

    stackcheck();
    push5(a, p, u1, w1, B);
    push(gamma);
    pushv(18);
#define a stack[-23]
#define p stack[-22]
#define u1 stack[-21]
#define w1 stack[-20]
#define B stack[-19]
#define gamma stack[-18]
    // local aa, x;
    // local alpha, delta, sigma, tau, modulus;
#define aa stack[-17]
#define x stack[-16]
#define alpha stack[-15]
#define delta stack[-14]
#define sigma stack[-13]
#define tau stack[-12]
#define modulus stack[-11]
    // local e, g, q, r, s, t;
    // local u, w, sh, th;
    // local c;
#define e stack[-10]
#define g stack[-9]
#define q stack[-8]
#define r stack[-7]
#define s stack[-6]
#define t stack[-5]
#define u stack[-4]
#define w stack[-3]
#define sh stack[-2]
#define th stack[-1]
#define c stack[-0]

    x = Mvar(a);

    /*
     * step 1 
     */

    alpha = Alcoeff(a, x);
    if (gamma == nil)
	gamma = alpha;
    a = gtimes2(a, gamma);
    s = Alcoeff(u1, x);
    t = Alcoeff(w1, x);
    u1 = Amkmodular(u1, p);
    w1 = Amkmodular(w1, p);
    u1 = gtimes2(gamma, u1);
    u1 = gratio2(u1, s);
    w1 = gtimes2(alpha, w1);
    w1 = gratio2(w1, t);

    /*
     * step 2
     */

    g = EEA(u1, w1, x);
    s = cadr(g);
    t = cadr(cdr(g));
    g = car(g);
    /*
     * step 3 
     */
    z = Aunmodular(u1);
    u = replace_lc(z, gamma, x);
    z = Aunmodular(w1);
    w = replace_lc(z, alpha, x);
    z = gtimes2(u, w);
    e = gdiff2(a, z);
    modulus = p;
    B = gtimes2(B, mkfxat(2));
    B = gtimes2(B, gamma);
    /*
     * step 4 
     */
    while (!zerop(e) && Llessp(modulus, B) != nil) {
	/*
	 * write "modulus = ", modulus, ", error = ", e;
	 */
	/*
	 * step 4.1 
	 */
	c = gquot2(e, modulus);
	sh = gtimes2(s, c);
	th = gtimes2(t, c);
	g = grdivide2(sh, w1, x);
	q = car(g);
	r = cdr(g);
	sigma = r;
	z = gtimes2(q, u1);
	tau = gplus2(th, z);
	/*
	 * step 4.2 
	 */
	z = Aunmodular(tau);
	z = gtimes2(z, modulus);
	u = gplus2(u, z);
	z = Aunmodular(sigma);
	z = gtimes2(z, modulus);
	w = gplus2(w, z);
	z = gtimes2(u, w);
	e = gdiff2(a, z);
	modulus = gtimes2(modulus, p);
    }
    /*
     * step 5 
     */
    if (zerop(e)) {
	delta = Mcont(u);
	u = gquot2(u, delta);
	z = gquot2(gamma, delta);
	w = gquot2(w, z);
	z = Llist3(u, w, modulus);
    } else
	z = nil;
    popv(24);
#undef a
#undef p
#undef u1
#undef w1
#undef B
#undef gamma
    // local aa, x;
    // local alpha, delta, sigma, tau, modulus;
#undef aa
#undef x
#undef alpha
#undef delta
#undef sigma
#undef tau
#undef modulus
    // local e, g, q, r, s, t;
    // local u, w, sh, th;
    // local c;
#undef e
#undef g
#undef q
#undef r
#undef s
#undef t
#undef u
#undef w
#undef sh
#undef th
#undef c
    return z;
}





static          Sex
multieea(Sex a, Sex x)
{
    Sex             u;
    long            j,
                    n;
    stackcheck();
    push2(a, x);
    pushv(6);
#define a stack[-7]
#define x stack[-6]
#define q stack[-4]
#define s stack[-3]
#define b stack[-1]
#define c stack[-0]
    // begin local n, q, s, j;
    // local b, c;

    // lisp_printf("multieea a = %p, %p\n", a, x);
    n = mkint(car(array_dim(a)));
    q = Lmkarray1(mkfxat(n));
    s = Lmkarray1(mkfxat(n));


    u = array_value(a, n - 1);
    array_value(q, n - 2) = u;
    for (j = n - 3; j >= 0; j--) {
	u = gtimes2(array_value(a, j + 1), array_value(q, j + 1));
	array_value(q, j) = u;
    }

    b = mkfxat(1);
    for (j = 0; j < n - 1; j++) {
	c = diophant(array_value(q, j), array_value(a, j), b, x);
	array_value(s, j) = cadr(c);
	b = car(c);
    }
    array_value(s, n - 1) = b;
    u = s;
    popv(8);
#undef a
#undef x
#undef q
#undef s
#undef b
#undef c

    return u;
}


static          Sex
diophant(Sex a1, Sex a2, Sex c, Sex x)
{
    Sex             u;
    stackcheck();
    push4(a1, a2, c, x);
    pushv(4);
#define a1 stack[-7]
#define a2 stack[-6]
#define c stack[-5]
#define x stack[-4]
#define s1 stack[-3]
#define s2 stack[-2]
#define s stack[-1]
#define d stack[-0]
    // begin local s1, s2, s, d;

    // lisp_printf("eea: %p, %p\n", a2, a1);
    s = eea(a2, a1);
    u = gtimes2(c, car(s));
    d = grdivide2(u, a1, x);
    s1 = cdr(d);

    a1 = gtimes2(c, cadr(s));
    a2 = gtimes2(a2, car(d));
    s2 = gplus2(a1, a2);
    u = Llist2(s1, s2);
    popv(8);
#undef a1
#undef a2
#undef c
#undef x
#undef s1
#undef s2
#undef s
#undef d
    return u;
}


static          Sex
eea(Sex a, Sex b)
{
    Sex             u;
    long            n;
    stackcheck();
    push2(a, b);
    pushv(14);

    // begin local s, t;
    // local c, d, c1, c2, d1, d2;
    // local r, r1, r2;
    // local n;
    // local q, g;
#define a stack[-15]
#define b stack[-14]
#define s stack[-13]
#define t stack[-12]
#define c stack[-11]
#define d stack[-10]
#define c1 stack[-9]
#define c2 stack[-8]
#define d1 stack[-7]
#define d2 stack[-6]
#define r stack[-5]
#define r1 stack[-4]
#define r2 stack[-3]
#define q stack[-1]
#define g stack[-0]

    u = Mlcoeff(a);
    c = gratio2(a, u);
    u = Mlcoeff(b);
    d = gratio2(b, u);
    c1 = d2 = mkfxat(1);
    d1 = c2 = mkfxat(0);

    n = 0;
    while (!zerop(d)) {
	n++;
	q = gquot2(c, d);
	u = gtimes2(q, d);
	r = gdiff2(c, u);
	u = gtimes2(q, d1);
	r1 = gdiff2(c1, u);
	u = gtimes2(q, d2);
	r2 = gdiff2(c2, u);
	c = d;
	c1 = d1;
	c2 = d2;
	d = r;
	d1 = r1;
	d2 = r2;
    }
    /*
     * normalize GCD 
     */
    if (is_poly(g)) {
	aerror("eea: arguments have common divider");
    }

    d = Mlcoeff(c);
    d1 = Mlcoeff(a);
    d2 = Mlcoeff(b);
    d1 = gtimes2(d, d1);
    d2 = gtimes2(d, d2);
    s = gratio2(c1, d1);
    t = gratio2(c2, d2);
    u = Llist2(s, t);

    popv(16);
#undef a
#undef b
#undef s
#undef t
#undef c
#undef d
#undef c1
#undef c2
#undef d1
#undef d2
#undef r
#undef r1
#undef r2
#undef q
#undef g
    return u;
}



static          Sex
uhensel(Sex a, Sex p, Sex u, Sex w, Sex B)
{
    Sex             z;
    long            n,
                    i,
                    j,
                    k;
    stackcheck();
    push5(a, p, u, w, B);
#define a stack[-29]
#define p stack[-28]
#define u stack[-27]
#define w stack[-26]
#define B stack[-25]
    pushv(25);
    // local aa, x;
    // local alpha, gamma, delta, sigma, tau, modulus;
#define aa stack[-24]
#define x stack[-23]
#define alpha stack[-22]
#define gamma stack[-21]
#define delta stack[-20]
#define sigma stack[-19]
#define tau stack[-18]
#define modulus stack[-17]
    // local e, g, q, r, s, t;
#define e stack[-16]
#define g stack[-15]
#define q stack[-14]
#define r stack[-13]
#define s stack[-12]
#define t stack[-11]
    // local sh, th;
#define sh stack[-10]
#define th stack[-9]
    // local c;
    // local U, W, S, T, n, i, j, k;
#define c stack[-8]
#define U stack[-7]
#define W stack[-6]
#define S stack[-5]
#define T stack[-4]

    x = Mvar(a);
    z = Llength(u);
    n = mkint(z);
    U = Lmkarray1(mkfxat(n));
    W = Lmkarray1(mkfxat(n));
    Alist2array(U, u);
    if (w != nil)
	Alist2array(W, w);
    /*
     * write "uhensel: p=", p; write "modulus of u is " ,
     * lisp:modular_modulus(lcoeff(U[0],x)); 
     */

    /*
     * step 1 
     */

    alpha = Mlcoeff(a);
    if (w == nil) {
	for (i = 0; i < n; i++) {
	    array_value(W, i) = alpha;
	}
    }
    gamma = mkfxat(1);
    for (i = 0; i < n; i++) {
	s = array_value(W, i);
	gamma = gtimes2(gamma, s);
	t = array_value(U, i);
	r = Mlcoeff(t);
	t = Amkmodular(t, p);
	t = gtimes2(s, t);
	t = gratio2(t, r);
	array_value(U, i) = t;
    }
    gamma = gratio2(gamma, alpha);
    a = gtimes2(a, gamma);


    /*
     * step 2
     */

    S = multieea(U, x);		/* U is modular poly */
    // lisp_printf("step2 U=%p\n", U);
    // lisp_printf("step2 S=%p\n", S);

    T = Lmkarray1(mkfxat(n));
    /*
     * step 3 
     */
    e = mkfxat(1);
    for (i = 0; i < n; i++) {
	s = array_value(U, i);
	t = array_value(W, i);
	s = Aunmodular(s);
	s = replace_lc(s, t, x);
	array_value(U, i) = s;
	e = gtimes2(e, s);
    }
    e = gdiff2(a, e);
    /*
     * write "e = ", e; 
     */
    modulus = p;
    /*
     * step 4 
     */
    B = gtimes2(B, mkfxat(2));
    B = gtimes2(B, gamma);
    while (!gzerop(e) && Llessp(modulus, B) != nil) {
	/*
	 * write "modulus = ", modulus, ", error = ", e;
	 */
	/*
	 * step 4 
	 */

	c = gratio2(e, modulus);
	/*
	 * write "c=", c; 
	 */
	th = mkfxat(0);
	for (i = 0; i < n; i++) {
	    // lisp_printf("i = %d, c = %p\n", i, c);
	    sh = gtimes2(array_value(S, i), c);
	    g = grdivide2(sh, array_value(U, i), x);
	    th = gplus2(th, car(g));
	    array_value(T, i) = cdr(g);
	    z = Aunmodular(array_value(T, i));
	    z = gtimes2(z, modulus);
	    z = gplus2(array_value(U, i), z);
	    array_value(U, i) = z;
	}
	if (!gzerop(th)) {
	    z = gtimes2(th, array_value(U, 0));
	    z = gplus2(array_value(T, 0), z);
	    array_value(T, 0) = z;
	    z = gtimes2(th, array_value(U, 0));
	    z = Aunmodular(z);
	    z = gtimes2(z, modulus);
	    z = gplus2(array_value(U, 0), z);
	    array_value(U, 0) = z;
	}
	e = mkfxat(1);
	for (i = 0; i < n; i++) {
	    e = gtimes2(e, array_value(U, i));
	}
	e = gdiff2(a, e);
	modulus = times2(modulus, p);
    }
    /*
     * step 5 
     */
    // lisp_printf("step5: e=%p\n", e);
    if (gzerop(e)) {
	gamma = mkfxat(1);
	for (i = 0; i < n; i++) {
	    s = array_value(U, i);
	    delta = Mcont(s);
	    s = gquot2(s, delta);
	    s = gsmodulus(s, modulus);
	    array_value(U, i) = s;
	    gamma = gtimes2(gamma, delta);
	}
	s = array_value(U, 0);
	s = gtimes2(s, gamma);
	// s = gsmodulus(s, modulus);
	array_value(U, 0) = s;
	// lisp_printf("U=%p\n", U);
    } else {
	for (i = 1; i <= n / 2; i++) {
	    if (a == mkfxat(1))
		break;
	    // lisp_printf("a = %p, U = %p\n", a, U);
	    r = Combine_Factors(a, U, modulus, i, n);
	    // lisp_printf("r = %p\n", r);
	    if (r != nil) {
		s = r;
		while (s != nil) {
		    a = gratio2(a, car(s));
		    s = cdr(s);
		}
	    }
	    T = Lappend(T, r);
	}

	/*
	 * write "U=", U; T := {}; for i:=0 to n-1 do if gremainder(a,
	 * U[i]) = 0 then T := cons(U[i], T); U[i] := 1; fi; od; for i:= 0 
	 * to n-1 do if U[i] = 1 then continue; fi; for j:= i+1 to n-1 do
	 * if U[j] = 1 then continue; fi; S:= (U[i] * U[j]) smod modulus;
	 * if gremainder(a, S) = 0 then T:=cons(S, T); U[i] := 1; U[j] :=
	 * 1; break; fi; od; od; for i:= 0 to n-1 do if U[i] = 1 then
	 * continue; fi; for j:= i+1 to n-1 do if U[i] = 1 then break;
	 * elif U[j] = 1 then continue; fi; for k:=j+1 to n-1 do if U[k] = 
	 * 1 then continue; fi;
	 * 
	 * S:= (U[i] * U[j] * U[k]) smod modulus; if gremainder(a, S) = 0
	 * then T:=cons(S, T); U[i] := 1; U[j] := 1; U[k] := 1; break; fi;
	 * od; od; od; 
	 */
	S = mkfxat(1);
	for (i = 0; i < n; i++) {
	    S = gtimes2(S, array_value(U, i));
	}
	S = gsmodulus(S, modulus);
	T = cons(S, T);
	z = Llength(T);
	U = Lmkarray1(z);
	Alist2array(U, T);
    }

    z = U;
    popv(30);
#undef a
#undef p
#undef u
#undef w
#undef B
#undef aa
#undef x
#undef alpha
#undef gamma
#undef delta
#undef sigma
#undef tau
#undef modulus
#undef e
#undef g
#undef q
#undef r
#undef s
#undef t
#undef sh
#undef th
#undef c
#undef U
#undef W
#undef S
#undef T
    return z;
}


static          Sex
Combine(Sex F, Sex q, long n, long m)
{
    Sex             s;
    long            i;
    stackcheck();
    push4(F, q, nil, nil);
#define F stack[-3]
#define q stack[-2]
#define U stack[-1]
#define c stack[0]
    U = Lmkarray1(mkfxat(m));
    c = make_symbol("c");
    for (i = 0; i < m; i++) {
	c = gplus2(c, mkfxat(1));
	array_value(U, i) = c;
    }
    s = Combine_Factors(F, U, q, n, m);
#undef F
#undef q
#undef U
#undef c
    popv(4);
    return s;
}

static          Sex
Combine_Factors(Sex F, Sex U, Sex q, long n, long m)
{
    Sex             r;
    long            k,
                    c;
    stackcheck();
    push3(F, U, q);
    push(nil);
#define F stack[-3]
#define U stack[-2]
#define q stack[-1]
#define a stack[0]

    a = Lmkarray1(mkfxat(n));
    k = 0;
    c = 1;
    r = Combine_nest(F, U, q, k, c, n, m, a);
    popv(4);
#undef F
#undef U
#undef q
#undef a
    return r;
}


static          Sex
Combine_nest(Sex F, Sex U, Sex q, long k, long c, long n, long m, Sex a)
{
    long            i,
                    j;

    Sex             u;

    stackcheck();
    push4(F, U, q, a);
    pushv(3);
#define F stack[-6]
#define U stack[-5]
#define q stack[-4]
#define a stack[-3]
#define s stack[-2]
#define r stack[-1]
#define rr stack[-0]
    r = nil;
    for (i = k + 1; i <= m - n + c; i++) {
	if (array_value(U, i - 1) == mkfxat(1))
	    continue;
	array_value(a, c - 1) = mkfxat(i - 1);
	if (n != c) {
	    rr = Combine_nest(F, U, q, i, c + 1, n, m, a);
	    r = Lappend(r, rr);
	} else {
	    s = mkfxat(1);
	    for (j = 0; j < n; j++) {
		u = array_value(a, j);
		u = array_value(U, mkint(u));
		if (u == mkfxat(1)) {
		    s = mkfxat(1);
		    break;
		}
		s = gtimes2(s, u);
		s = gsmodulus(s, q);
	    }
	    if (s == mkfxat(1))
		continue;

	    if (gremaind2(F, s) == mkfxat(0)) {
		r = cons(s, r);
		for (j = 0; j < n; j++) {
		    u = array_value(a, j);
		    array_value(U, mkint(u)) = mkfxat(1);
		}
	    }
	}
    }
    u = r;
#undef F
#undef U
#undef q
#undef a
#undef s
#undef r
#undef rr
    popv(7);
    return u;
}


/*
 * f:=(x^3+3*x+1)*(x^2+x+1); q:=17; 
 */

/*
 * g:=berlekamp(f, q); h:=UnivariateHensel(f, q, first(g), second(g), q^4, 
 * nil); 
 */
/**************************************************************/
/**************************************************************/

setup_type const factor_setup[] = {

    {"factor", (lisp_proc *) Afactor, 1},
    {"zfactor", (lisp_proc *) Azfactor, 1},
    {"berlekamp", (lisp_proc *) berlekamp, 2},
    {"squarefree", (lisp_proc *) SquareFree, 1},
    {"FormMatrix", (lisp_proc *) FormMatrix, 2},
    {"factorlist2prefix", (lisp_proc *) factorlist2prefix, 1},
    {NULL, 0, 0}
};
