/**************************************************************/
/*
 * F4 algorithm 
 */
/**************************************************************/


static          Sex
F4(Sex F)
{
    Sex             r;
    long            d;
    stackcheck();
    push(F);
    pushv(8);
    // local G, F0, d, P, Pd, Ld, Fd, h;
#define F stack[-8]
#define G stack[-7]
#define F0 stack[-6]
#define P stack[-5]
#define Pd stack[-4]
#define Ld stack[-3]
#define Fd stack[-2]
#define h stack[-1]

    G = F;
    F0 = F;
    d = 0;
    P = makePair1(G);
    while (P != nil) {
	d++;
	// lisp_printf("F4: loop %d\n", d);
	Pd = Sel(P);
	P = Asetdiff(P, Pd);
	h = Left2(Pd);
	r = Right2(Pd);
	Ld = Lunion(h, r);
	Fd = F4_Reduction(Ld, G);
	while (Fd != nil) {
	    h = car(Fd);
	    h = makePair2(h, G);
	    P = Lunion(P, h);
	    h = car(Fd);
	    Fd = cdr(Fd);
	    h = Lncons(h);
	    G = Lunion(G, h);
	}
    }
    r = G;
    popv(9);

#undef F
#undef G
#undef F0
#undef P
#undef Pd
#undef Ld
#undef Fd
#undef h
    return r;
}

static          Sex
F4_HT(Sex f)
{
    f = Ldipoly(f);
    f = Mlpower(f);
    f = Lpolynomial(f);
    return f;
}

static          Sex
F4_HT2(Sex F)
{				/* takousiki no HT */
    Sex             r;
    stackcheck();
    push(F);
    pushv(2);
    // local f,L;
#define F stack[-2]
#define f stack[-1]
#define L stack[-0]
    while (F != nil) {
	f = car(F);
	F = cdr(F);
	f = F4_HT(f);
	f = Lncons(f);
	L = Amerge(L, f);
    }
    r = L;
    popv(3);
#undef F
#undef f
#undef L
    return r;
}

static          Sex
F4_HM(Sex f)
{
    f = Ldipoly(f);
    f = Mlterm(f);
    f = Lpolynomial(f);
    return f;
}


static          Sex
F4_HC(Sex f)
{
    f = Ldipoly(f);
    f = Mlcoeff(f);
    f = Lpolynomial(f);
    return f;
}

static          Sex
makeP(Sex g1, Sex g2)
{
    Sex             r;
    stackcheck();
    push2(g1, g2);
    pushv(5);

    // begin local l, t1, t2;
    // local h1, h2;
#define g1 stack[-6]
#define g2 stack[-5]
#define l stack[-4]
#define t1 stack[-3]
#define t2 stack[-2]
#define h1 stack[-1]
#define h2 stack[-0]
    h1 = F4_HT(g1);
    h2 = F4_HT(g2);
    l = Mlcm(h1, h2);
    t1 = F4_HC(g2);
    t1 = gtimes2(t1, l);
    t1 = gratio2(t1, h1);
    t2 = F4_HC(g1);
    t2 = gtimes2(t2, l);
    t2 = gratio2(t2, h2);
    r = Llist5(l, t1, g1, t2, g2);
    popv(7);
#undef g1
#undef g2
#undef l
#undef t1
#undef t2
#undef h1
#undef h2
    return r;
}


static          Sex
makePair1(Sex G)
{
    Sex             r;
    stackcheck();
    push(G);
    pushv(7);

    // begin local L, g1, g2, G2;
    // local lcm, t1, t2;
#define G stack[-7]
#define L stack[-6]
#define g1 stack[-5]
#define g2 stack[-4]
#define G2 stack[-3]
#define lcm stack[-2]
#define t1  stack[-1]
#define t2  stack[-0]
    while (G != nil) {
	g1 = car(G);
	G = cdr(G);
	G2 = G;
	while (G2 != nil) {
	    g2 = car(G2);
	    G2 = cdr(G2);
	    g2 = makeP(g1, g2);
	    g2 = Lncons(g2);
	    L = Lappend(L, g2);
	}
    }
    r = L;
    popv(8);
#undef G
#undef L
#undef g1
#undef g2
#undef G2
#undef lcm
#undef t1
#undef t2
    return r;
}


static          Sex
makePair2(Sex h, Sex G)
{
    Sex             r;
    stackcheck();
    push2(h, G);
    pushv(2);
    // begin local L, g1;
#define h stack[-3]
#define G stack[-2]
#define L stack[-1]
#define g1 stack[-0]
    while (G != nil) {
	g1 = car(G);
	G = cdr(G);
	g1 = makeP(h, g1);
	g1 = Lncons(g1);
	L = Lappend(L, g1);
    }
    r = L;
    popv(4);
#undef h
#undef G
#undef L
#undef g1
    return r;
}


static          Sex
Left(Sex Pd)
{
    Sex             r;
    Pd = cdr(Pd);
    r = Llist2(car(Pd), cadr(Pd));
    return r;
}

static          Sex
Right(Sex Pd)
{
    Sex             r;
    Pd = cddr(Pd);
    Pd = cdr(Pd);
    r = Llist2(car(Pd), cadr(Pd));
    return r;
}

static          Sex
Left2(Sex Pd)
{
    Sex             s;
    stackcheck();
    push(Pd);
    pushv(2);
#define Pd stack[-2]
#define p stack[-1]
#define r stack[-0]
    while (Pd != nil) {
	p = car(Pd);
	Pd = cdr(Pd);
	// p = Left(p);
	p = cdr(p);
	p = Llist2(car(p), cadr(p));
	r = cons(p, r);
    }
    s = Lnreverse(r);
    popv(3);
#undef Pd
#undef p
#undef r
    return s;
}

static          Sex
Right2(Sex Pd)
{
    Sex             s;
    stackcheck();
    push(Pd);
    pushv(2);
#define Pd stack[-2]
#define p stack[-1]
#define r stack[-0]
    while (Pd != nil) {
	p = car(Pd);
	Pd = cdr(Pd);
	// p = Right(p);
	p = cddr(p);
	p = cdr(p);
	p = Llist2(car(p), cadr(p));
	r = cons(p, r);
    }
    s = Lnreverse(r);
    popv(3);
#undef Pd
#undef p
#undef r
    return s;
}


/*
 * 10.23 
 */

static          Sex
make_monolist(Sex f)
{				/* tannkousiki no list tukuru */
    Sex             r;
    stackcheck();
    push(f);
    pushv(2);
#define f stack[-2]
#define a stack[-1]
#define L stack[-0]
    f = Ldipoly(f);
    while (!gzerop(f)) {
	a = Mlpower(f);		/* f no sentouukou */
	f = Mreduct(f);
	L = cons(a, L);
    }
    r = Lnreverse(L);
    r = Lpolynomial(r);
    popv(3);
#undef f
#undef a
#undef L
    return r;
}


static          Sex
make_coefflist(Sex f)
{				/* keisu no list tukuru */
    Sex             r;

    stackcheck();
    push(f);
    pushv(4);
#define f stack[-4]
#define c stack[-3]
#define a stack[-2]
#define L stack[-1]
#define C stack[-0]
    f = Ldipoly(f);
    while (!gzerop(f)) {
	c = Mlcoeff(f);
	a = Mlpower(f);		/* f no sentouukou */
	f = Mreduct(f);
	L = cons(a, L);
	C = cons(c, C);
    }
    L = Lnreverse(L);
    C = Lnreverse(C);
    L = Lpolynomial(L);
    C = Lpolynomial(C);
    r = Llist2(L, C);
    popv(5);
#undef f
#undef c
#undef a
#undef L
#undef C
    return r;
}

static          Sex
make_monolist_from_list(Sex F)
{				/* takousiki no list no tannkousiki no
				 * list tukuru */
    Sex             r;
    stackcheck();
    push(F);
    pushv(3);

    // local L, l, f;
#define F stack[-3]
#define L stack[-2]
#define l stack[-1]
#define f stack[-0]

    while (F != nil) {
	f = car(F);
	F = cdr(F);
	l = make_monolist(f);
	L = Amerge(L, l);
    }
    r = L;
    popv(4);
#undef L
#undef F
#undef l
#undef f
    return r;
}

static          Sex
make_coeffmatrix(Sex F)
{				/* keisuugyouretu tukuru */
    long            i,
                    n,
                    m;
    Sex             r;

    stackcheck();
    push(F);
    pushv(10);
    // local c,A,m,L,f,a,i,n,l,ll;
#define F stack[-10]
#define c stack[-9]
#define A stack[-8]
#define L stack[-7]
#define f stack[-5]
#define a stack[-5]
#define l stack[-4]
#define ll stack[-3]
    n = mkint(Llength(F));
    L = make_monolist_from_list(F);
    m = mkint(Llength(L));
    A = make_matrix(n, m);
    i = 0;
    while (F != nil) {
	f = car(F);
	F = cdr(F);
	c = make_coefflist(f);
	l = car(c);
	c = cadr(c);
	i = i + 1;

	n = 1;
	ll = L;
	while (l != nil) {
	    a = car(l);
	    while (ll != nil) {
		if (Lequal(a, car(ll)) != nil)
		    break;
		ll = cdr(ll);
		n = n + 1;
	    }


	    matrix_value(A, (i - 1) * m + n - 1) = car(c);
	    l = cdr(l);
	    c = cdr(c);
	}
    }
    r = Llist2(A, L);
    popv(11);
#undef F
#undef c
#undef A
#undef L
#undef f
#undef a
#undef l
#undef ll
    return r;
}

/*
 * 10.30 
 */





static          Sex
symbolic_preprocessing(Sex L, Sex G)
{				/* m top reducible modulo G */
    Sex             r;

    stackcheck();
    push2(L, G);
    pushv(10);
    // local F,D,TF,M,g;
    // local a, mp, m, l;
    // H
#define L stack[-11]
#define G stack[-10]
#define F stack[-9]
#define D stack[-8]
#define TF stack[-7]
#define M stack[-6]
#define g stack[-5]
#define a stack[-4]
#define mp stack[-3]
#define m stack[-2]
#define l stack[-1]
#define H stack[-0]

    while (L != nil) {
	l = car(L);
	L = cdr(L);
	l = gtimes2(car(l), cadr(l));
	F = cons(l, F);
    }
    F = Lnreverse(F);

    D = F4_HT2(F);
    TF = make_monolist_from_list(F);

    M = Asetdiff(TF, D);


    while (M != nil) {
	m = car(M);
	M = cdr(M);
	l = Lncons(m);
	D = Amerge(D, l);
	H = G;
	while (H != nil) {
	    g = car(H);
	    H = cdr(H);
	    mp = wari(m, g);
	    if (mp != nil) {
		g = gtimes2(mp, g);
		l = Lncons(g);
		F = Lappend(F, l);
		// D =merge(D, {F4_HT(g)});
		l = make_monolist(g);
		TF = Amerge(TF, l);
		// M =setdiff(TF, D); 
		// M =merge(M, make_monolist(g));
		break;
	    }
	}
	M = Asetdiff(TF, D);
    }
    r = F;
    popv(12);
#undef L
#undef G
#undef F
#undef D
#undef TF
#undef M
#undef g
#undef a
#undef mp
#undef m
#undef l
#undef H
    return r;
}

/*
 * 11.6
 */

static          Sex
F4_Reduction(Sex L, Sex G)
{				/* L,G wo kanyakusuru */
    Sex             r;
    long            i,
                    n,
                    m;

    stackcheck();
    push2(L, G);
    pushv(10);
    // local F,n,m;
    // local h, f, i, l, M, Fg;
    // local V;
#define L stack[-11]
#define G stack[-10]
#define F stack[-9]
#define h stack[-8]
#define f stack[-7]
#define M stack[-6]
#define Fg stack[-5]
#define V stack[-4]
#define l stack[-3]
    F = symbolic_preprocessing(L, G);
    Fg = make_coeffmatrix(F);
    M = car(Fg);
    L = cadr(Fg);
    V = Anewvector(L);
    r = Asize(M);
    m = mkint(cadr(r));
    n = mkint(car(r));
    M = row_echelon_form(M, n, m);	/* betu de tukutta */
    V = gtimes2(M, V);

    l = nil;
    L = F4_HT2(F);
    for (i = 1; i <= n; i++) {
	f = vect_value(V, i - 1);
	if (!gzerop(f)) {
	    h = F4_HT(f);
	    if (Lmember(h, L) == nil) {
		f = Lncons(f);
		l = Lappend(l, f);
	    }
	}
    }
    r = l;
    popv(12);
#undef L
#undef G
#undef F
#undef h
#undef f
#undef M
#undef Fg
#undef V
#undef l
    return r;
}
/*
 * L:={{d,a*b*c+a*b*d+a*c*d+b*c*d},{1,a*b*c*d-1}}; G:={a*b*c*d-1}; 
 */

static          Sex
Sel(Sex P)
{
    Sex             r;
    long            d,
                    d1;
    stackcheck();
    push(P);
    pushv(4);
    // local p, d, d1, L;
    // Q;
#define P stack[-4]
#define p stack[-3]
#define L stack[-2]
#define Q stack[-1]

    d = mkint(Atdeg(caar(P)));

    Q = P;
    while (Q != nil) {
	p = car(Q);
	Q = cdr(Q);
	d1 = mkint(Atdeg(car(p)));
	if (d1 < d)
	    d = d1;
    }

    while (P != nil) {
	p = car(P);
	P = cdr(P);

	d1 = mkint(Atdeg(car(p)));
	if (d == d1) {
	    p = Lncons(p);
	    L = Lappend(L, p);
	}
    }
    r = L;
    popv(5);
#undef P
#undef p
#undef L
#undef Q
    return r;
}




static          Sex
row_echelon_form(Sex a, long n, long m)
{				/* a matrix of size n*m */
    long            i,
                    j,
                    k,
                    l;
    long            nm;

    stackcheck();
    push(a);
    pushv(3);
    // local i,j,k,l;
    // local a11;
    // local v1, v2;
    // local nm;
#define a stack[-3]
#define a11 stack[-2]
#define v1 stack[-1]
#define v2 stack[-0]
    nm = n > m ? m : n;
    l = 1;
    for (i = 1; i <= m; i++) {
	if (l > n)
	    break;
	if (zerop(matrix_value(a, (l - 1) * m + i - 1))) {
	    for (k = l + 1; k <= n; k++) {
		if (!zerop(matrix_value(a, (k - 1) * m + i - 1))) {
		    /*
		     * kgyo to igyo wo irekae
		     */

		    // swap_row(a,l,k);
		    for (j = 1; j <= m; j++) {
			v1 = matrix_value(a, (l - 1) * m + j - 1);
			v2 = matrix_value(a, (k - 1) * m + j - 1);
			matrix_value(a, (k - 1) * m + j - 1) = v1;
			matrix_value(a, (l - 1) * m + j - 1) = v2;
		    }
		    break;
		}
	    }
	    if (k > n)
		continue;
	}

	/*
	 * v1 =rowvector(a, l);
	 * v1 =v1/a[l,i];
	 * setrow(a, l, v1); 
	 */
	a11 = matrix_value(a, (l - 1) * m + i - 1);
	if (!gonep(a11)) {
	    for (j = 1; j <= m; j++) {
		v1 = matrix_value(a, (l - 1) * m + j - 1);
		v1 = gratio2(v1, a11);
		matrix_value(a, (l - 1) * m + j - 1) = v1;
	    }
	}

	/*
	 * v1:=rowvector(a, l);
	 */

	for (j = l + 1; j <= n; j++) {
	    /*
	     * v2 =rowvector(a, j); v2 =v2-v1*a[j,i]; setrow(a, j, v2); 
	     */
	    a11 = matrix_value(a, (j - 1) * m + i - 1);
	    if (!gzerop(a11)) {
		for (k = 1; k <= m; k++) {
		    v1 = matrix_value(a, (l - 1) * m + k - 1);
		    v1 = gtimes2(a11, v1);
		    v2 = matrix_value(a, (j - 1) * m + k - 1);
		    v2 = gdiff2(v2, v1);
		    matrix_value(a, (j - 1) * m + k - 1) = v2;
		}
	    }
	}
	l++;
    }
#undef a
    a = stack[-3];
    popv(4);
#undef a11
#undef v1
#undef v2
    return a;
}

static          Sex
wari(Sex f, Sex g)
{
    Sex             a;

    stackcheck();
    push(g);
    f = LT(f);
    g = stack[0];
    stack[0] = f;
    g = LT(g);
    pop(f);
    a = gratio2(f, g);
    if (is_number(a) || is_poly(a))
	return a;
    return nil;
}

static          Sex
LT(Sex f)
{
    f = Ldipoly(f);
    f = Mlterm(f);
    f = Lpolynomial(f);
    return f;
}
