/**************************************************************/
/*
 * Fauger F5 algorithm 
 */
/**************************************************************/


#define POLY(f)	car(cddr(f))
#define HT(f)	Mlpower(f)
#define HC(f)	Mlcoeff(f)
#define HM(f)	Mlterm(f)

static long     F5_N;
Sex             F5_Rules;



static          Sex
poly_list(Sex R)
{
    Sex             s;
    stackcheck();
    push(R);
    pushv(2);

    // begin local l, r;
#define R stack[-2]
#define l stack[-1]
#define r stack[-0]
    while (R != nil) {
	r = car(R);
	R = cdr(R);
	r = car(cddr(r));
	r = Lncons(r);
	l = Lappend(l, r);
    }
    s = l;
    popv(3);
#undef R
#undef l
#undef r
    return s;
}

static          Sex
check_dipoly_list(Sex F)
{

    stackcheck();
    push(F);
    pushv(3);
    // local o, f, l;
#define F stack[-3]
#define o stack[-2]
#define f stack[-1]
#define l stack[-0]

    if (!is_cons(F)) {
	// if debug then error("not list", F);
	popv(4);
	return nil;
    }
    f = car(F);
    F = cdr(F);
    if (!is_dipoly(f)) {
	// if debug then error("not dipoly", f);
	popv(4);
	return nil;
    }
    o = dipoly_order(f);
    l = Lgetdipolyorder();
    if (Lequal(o, l) == nil) {
	// if debug then error("order mismatch", o);fi;
	popv(4);
	return nil;
    }
    while (F != nil) {
	f = car(F);
	F = cdr(F);
	if (!is_dipoly(f)) {
	    // if debug then error("not dipoly", f);fi;
	    popv(4);
	    return nil;
	}
	l = dipoly_order(f);
	if (Lequal(o, l) == nil) {
	    // if debug then error("order mismatch", f);
	    popv(4);
	    return nil;
	}
    }
    popv(4);
#undef F
#undef o
#undef f
#undef l
    return true;
}

static          Sex
F5(Sex F)
{
    long            i,
                    m;
    Sex             s;

    stackcheck();
    push(F);
    pushv(6);

    // begin local G, rm, m;
    // local i, fm, r;
    // H
#define F stack[-6]
#define G stack[-5]
#define rm stack[-4]
#define fm stack[-3]
#define r stack[-2]
#define H stack[-1]


    if (check_dipoly_list(F) == nil) {
	s = F;
	popv(7);
	return aerror1("not dipoly list", s);
    }

    m = list_length(F);
    ResetRules(m);

    /*
     * G = make_vector(m+1); for(i = 1; i <= m+1; i++) { vect_value(G,
     * i-1) = nil; } 
     */

    F = Lreverse(F);
    fm = car(F);

    s = HC(fm);
    fm = gratio2(fm, s);
    F = cdr(F);
    rm = Llist3(mkfxat(1), mkfxat(m), fm);
    G = Lncons(rm);
#ifdef F5_DEBUG
    // write "f5: G", m, "= ", G[m];
    lisp_printf("f5: G%d=%p\n", m, G);
#endif
    for (i = m - 1; i >= 1; i--) {
	fm = car(F);
	s = HC(fm);
	fm = gratio2(fm, s);
	F = cdr(F);
	G = F5_inc(i, fm, G);
#ifdef F5_DEBUG
	// write "f5: G", i, "= ", G[i];
	lisp_printf("f5: G%d = %p\n", i, G);
#endif
    }
    s = poly_list(G);
    popv(7);
#undef F
#undef G
#undef rm
#undef fm
#undef r
#undef H
    return s;
}

static          Sex
F5_inc(long i, Sex f, Sex G)
{
    Sex             u;
    long            d,
                    d2;

    push2(f, G);
    pushv(13);

    // begin local ri;
    // local r, g, s;
    // local F, P, P2, Pd, Gp, L;
    // local p, Rd, d;
#define f stack[-14]
#define G stack[-13]
#define ri stack[-12]
#define r stack[-11]
#define g stack[-10]
#define s stack[-9]
#define F stack[-8]
#define P stack[-7]
#define P2 stack[-6]
#define Pd stack[-5]
#define Gp stack[-4]
#define L stack[-3]
#define p stack[-2]
#define Rd stack[-1]
#define H stack[0]

    u = Ldipoly(mkfxat(1));
    ri = Llist3(u, mkfxat(i), f);
    s = Lncons(ri);
    Gp = G;
    G = Lappend(G, s);
    L = nil;
    H = Gp;
    while (H != nil) {
	r = car(H);
	H = cdr(H);
	g = CritPair(ri, r, i, Gp);
	if (g != nil) {
	    L = Laconc(L, g);
	}
    }
    P = sort_by_degree(L);

    while (P != nil) {
	r = car(P);
	d = mkint(Atdeg(car(r)));

#ifdef F5_DEBUG
	// write "d=", d;
	// write "critical pairs = ", P;
	lisp_printf("d=%d\ncritical pairs = %p\n", d, P);
#endif

	Pd = nil;
	H = P;
	while (H != nil) {
	    g = car(H);
	    H = cdr(H);
	    d2 = mkint(Atdeg(car(g)));
	    if (d2 == d)
		Pd = Laconc(Pd, g);
	}
#ifdef F5_DEBUG
	// write "P",d,"=", Pd;
	lisp_printf("P%d=%p\n", d, Pd);
#endif
	P = Asetdiff(P, Pd);
	F = Spol(Pd);
#ifdef F5_DEBUG
	// write "call reduction: F=", F;
	lisp_printf("call reduction: F=%p\n", F);
#endif
	Rd = Reduction(F, G, i, Gp);
#ifdef F5_DEBUG
	// write "R",d,"=", Rd;
	lisp_printf("R%d=%p\n", d, Rd);
#endif
	L = nil;
	while (Rd != nil) {
	    r = car(Rd);
	    Rd = cdr(Rd);
	    H = G;
	    while (H != nil) {
		g = car(H);
		H = cdr(H);
		s = CritPair(r, g, i, Gp);
		if (s != nil) {
		    s = Lncons(s);
		    L = Lappend(L, s);
		} else {
#ifdef F5_DEBUG
		    lisp_printf("pair %p and %p is rejected\n", r, g);
#endif
		}
	    }
	    /*
	     * P:=append(P, L);
	     */
	    r = Lncons(r);
	    G = Lappend(G, r);
	}
	L = sort_by_degree(L);
#ifdef F5_DEBUG
	// write "L = ", L;
	// write "G = ", Gp;
	lisp_printf("L=%p\n", L);
	lisp_printf("G=%p\n", G);
#endif
	P = merge_by_degree(P, L);
	// P:=sort_by_degree(P);
    }
    u = G;
    popv(15);
#undef f
#undef G
#undef ri
#undef r
#undef g
#undef s
#undef F
#undef P
#undef P2
#undef Pd
#undef Gp
#undef L
#undef p
#undef Rd
#undef H
    return u;
}


static          Sex
compare_signature2(Sex u1, Sex r1, Sex u2, Sex r2)
{
    Sex             s;
    long            s1,
                    s2;

    s1 = mkint(cadr(r1));
    s2 = mkint(cadr(r2));
    if (s1 > s2)
	return true;
    if (s1 < s2)
	return nil;
    stackcheck();
    push2(u2, r2);
    u1 = gtimes2(u1, car(r1));
    u2 = stack[-1];
    r2 = stack[-0];
    stack[0] = u1;
    u2 = gtimes2(u2, car(r2));
    u1 = stack[0];
    popv(2);
    if (compare(u1, u2) < 0)
	return true;
    return nil;
}

static          Sex
top_reducible(Sex t, Sex G)
{
    Sex             s;
    stackcheck();
    push2(t, G);
    // begin local g, r, s;
#define t stack[-1]
#define G stack[-0]
#ifdef F5_DEBUG
    lisp_printf("top_reducible: G=%p\n", G);
#endif
    while (G != nil) {
	s = car(G);
	G = cdr(G);
	s = car(cddr(s));
	s = ht_div(t, s);
	if (s != nil) {
	    popv(2);
	    return true;
	}
    }
    popv(2);
#undef t
#undef G
    return nil;
}

static          Sex
CritPair(Sex r1, Sex r2, long k, Sex G)
{
    Sex             r;
    stackcheck();
    push3(r1, r2, G);
    r = CritPair2(r1, r2, k, G);
    pop3(G, r2, r1);
    push(r);
#ifdef F5_DEBUG
    lisp_printf("Pair(%p, %p) --> %p\n", r1, r2, r);
#endif
    pop(r);
    return r;
}


static          Sex
CritPair2(Sex r1, Sex r2, long k, Sex G)
{
    Sex             r,
                    rtmp;
    long            k1,
                    k2;

    stackcheck();
    push3(r1, r2, G);
    pushv(14);

    // begin local p1, p2, t;
    // local u1, u2, s1, s2;
    // local u, v, k2, t2, k1, t1;
    // local rtmp;
#define r1 stack[-16]
#define r2 stack[-15]
#define G stack[-14]
#define p1 stack[-13]
#define p2 stack[-12]
#define t stack[-11]
#define u1 stack[-10]
#define u2 stack[-9]
#define s1 stack[-8]
#define s2 stack[-7]
#define u stack[-6]
#define v stack[-5]
#define t2 stack[-4]
#define t1 stack[-3]

    p1 = car(cddr(r1));
    p2 = car(cddr(r2));
    p1 = HT(p1);
    p2 = HT(p2);
    t = Mlcm(p1, p2);
    u1 = gquot2(t, p1);
    u2 = gquot2(t, p2);
    u1 = Ldipoly(u1);
    u2 = Ldipoly(u2);
    if (compare_signature2(u1, r1, u2, r2) != nil) {
	rtmp = r1;
	r1 = r2;
	r2 = rtmp;
	rtmp = u1;
	u1 = u2;
	u2 = rtmp;
    }
    t1 = car(r1);
    k1 = mkint(cadr(r1));
    t2 = car(r2);
    k2 = mkint(cadr(r2));
#ifdef F5a
    t1 = gtimes2(u1, t1);
    t2 = gtimes2(u2, t2);


#ifdef F5_DEBUG
    lisp_printf("cp (%d, %d): G=%p\n", k1, k2, G);
#endif

    if (top_reducible(t1, vect_value(G, k1)) != nil) {
	popv(17);
	return nil;
    }
    if (top_reducible(t2, vect_value(G, k2)) != nil) {
	popv(17);
	return nil;
    }
#else
    if (k1 > k) {
	popv(17);
	return nil;
    }
    t1 = gtimes2(t1, u1);
    t2 = NF(t1, G);
#ifdef F5_DEBUG
    lisp_printf("NF(%p, %p) ==> %p\n", t1, G, t2);
#endif
    if (Lequal(t1, t2) == nil) {
	popv(17);
	return nil;
    }
    if (k2 == k) {
	t1 = gtimes2(t2, u2);
	t2 = NF(t1, G);
#ifdef F5_DEBUG
	lisp_printf("NF(%p, %p) ==> %p\n", u, G, t2);
#endif
	if (Lequal(t1, t2) == nil) {
	    popv(17);
	    return nil;
	}
    }
#endif
    t = Ldipoly(t);
    r = Llist5(t, u1, r1, u2, r2);
    popv(17);
#undef r1
#undef r2
#undef G
#undef p1
#undef p2
#undef t
#undef u1
#undef u2
#undef s1
#undef s2
#undef u
#undef v
#undef t2
#undef t1
    return r;
}


static          Sex
Spol(Sex P)
{
    Sex             s;

    stackcheck();
    push(P);
    pushv(10);

    // begin local F;
    // local ri, rj, r;
    // local i, j, ul, vl, tl;
    // local p;
    // global N;
    // u, v;
#define P stack[-10]
#define F stack[-9]
#define ri stack[-8]
#define rj stack[-7]
#define r stack[-6]
#define ul stack[-5]
#define vl stack[-4]
#define tl stack[-3]
#define p stack[-2]
#define u stack[-1]
#define v stack[-0]
    while (P != nil) {
	p = car(P);
	P = cdr(P);
	tl = car(p);
	p = cdr(p);
	ul = car(p);
	p = cdr(p);
	ri = car(p);
	p = cdr(p);
	vl = car(p);
	p = cdr(p);
	rj = car(p);
	if (IsRewritten(ul, ri) == nil) {
	    if (IsRewritten(vl, rj) == nil) {
		s = POLY(ri);
		s = HC(s);
		vl = gtimes2(vl, s);
		s = POLY(rj);
		s = HC(s);
		vl = gratio2(vl, s);

		v = POLY(rj);
		v = gtimes2(vl, v);
		u = POLY(ri);
		u = gtimes2(ul, u);
		vl = gdiff2(u, v);
		s = HC(vl);
		vl = gquot2(vl, s);
		F5_N++;
		u = car(ri);
		u = gtimes2(ul, u);
		r = Llist3(u, cadr(ri), vl);
#ifdef F5_DEBUG
		lisp_printf("Spol: addrules %d\n", F5_N);
#endif
		AddRules(r, F5_N);
		r = Lncons(r);
		F = Lappend(F, r);
	    }
	}
    }
    // s = sort_by_signature(F);
    s = sort_by_degree(F);
    popv(11);
#undef P
#undef F
#undef ri
#undef rj
#undef r
#undef ul
#undef vl
#undef tl
#undef p
#undef u
#undef v
    return s;
}


static          Sex
NF(Sex f, Sex F)
{
    Sex             s,
                    r;
    if (f == nil)
	return mkfxat(0);

    stackcheck();
    push2(f, F);
    pushv(2);

    // begin local g, r;
#define f stack[-3]
#define F stack[-2]
#define g stack[-1]
    while (F != nil) {
	r = car(F);
	F = cdr(F);
	r = car(cddr(r));
	g = cons(r, g);
    }
    s = Lnreverse(g);
    s = Areduce(f, s);
    popv(4);
#undef f
#undef F
#undef g
    return s;
}

static          Sex
NF2(Sex f, Sex F)
{
    long            i,
                    j,
                    n;
    Sex             s;

    stackcheck();
    push2(f, F);
    pushv(3);

    // begin local r, g, n;
    // local i, j;
    // local c;
#define f stack[-4]
#define F stack[-3]
#define r stack[-2]
#define g stack[-1]
#define c stack[-0]

    n = list_length(F);
    g = mkarray(n);

    for (i = 0; i < n; i++) {
	array_value(g, i) = car(cddr(car(F)));
	F = cdr(F);
    }
    r = mkfxat(0);
    while (!gzerop(f)) {
	for (j = 0; j < n; j++) {
	    c = ht_div(f, array_value(g, j));
	    if (c != nil)
		break;
	}
	if (c == nil) {
	    c = HM(f);
	    r = gplus2(r, c);
	    f = gdiff2(f, c);
	} else {
	    c = Ldipoly(c);
	    c = gtimes2(array_value(g, j), c);
	    f = gdiff2(f, c);
	}
    }
    s = r;
    popv(5);
#undef f
#undef F
#undef r
#undef g
#undef c
    return s;
}

static          Sex
ht_div(Sex f, Sex g)
{
    Sex             r;
    // local r;
    stackcheck();
    push(g);
    f = HM(f);
    g = stack[0];
    stack[0] = f;
    g = HM(g);
    pop(f);
    r = gratio2(f, g);
    if (is_number(r) || is_poly(r))
	return Ldipoly(r);
    if (is_dipoly(r))
	return r;
    return nil;
}




static          Sex
ResetRules(long m)
{
    long            i;
    Sex            *p;

    F5_Rules = mkarray(m);
    F5_N = m;
    p = &array_value(F5_Rules, 0);
    for (i = 0; i < m; i++)
	*p++ = nil;

    return nil;
}


static          Sex
AddRules(Sex r, long n)
{				/* r = {t, i, p} */
    /*
     * t: term 
     */
    /*
     * i: index Fi 
     */
    /*
     * p: polynomial 
     */
    long            i;

    // begin local i;
    // global Rules;
    i = mkint(cadr(r));
#ifdef F5_DEBUG
    lisp_printf("add rules[%d] = {%p, %d}\n", i, r, n);
#endif
    r = Llist3(car(r), r, mkfxat(n));
    r = cons(r, array_value(F5_Rules, i - 1));
    array_value(F5_Rules, i - 1) = r;
    return nil;
}

static          Sex
Rewritten(Sex u, Sex rk)
{
    Sex             s;
    long            i;

    stackcheck();
    push2(u, rk);
    pushv(4);

    // begin local L, i, t, ti, d, l;
    // local ut;
    // global Rules;
#define u stack[-5]
#define rk stack[-4]
#define L stack[-3]
#define t stack[-2]
#define d stack[-1]
#define l stack[-0]


    i = mkint(cadr(rk));
    t = car(rk);
    t = gtimes2(t, u);
    L = array_value(F5_Rules, i - 1);
#ifdef F5_DEBUG
    // write "Rewritten: rule = ", L;
    lisp_printf("Rewritten: rule = %p\n", L);
#endif
    while (L != nil) {
	l = car(L);
	L = cdr(L);
	s = car(l);
	d = term_div(t, s);
	if (d != nil) {
	    rk = cadr(l);
	    s = Llist2(d, rk);
	    popv(6);
	    return s;
	}
    }
    s = Llist2(u, rk);
    popv(6);
#undef u
#undef rk
#undef L
#undef t
#undef d
#undef l
    return s;
}


static          Sex
term_div(Sex f, Sex g)
{

    Sex             a;
    // begin local a;
    a = gratio2(f, g);
    if (is_number(a) || is_poly(a))
	return Ldipoly(a);
    return nil;
}

static          Sex
IsRewritten(Sex u, Sex rk)
{
    Sex             rl;
    Sex             l,
                    r;

    stackcheck();

    push(u);

    // begin local l, rl;
    // local r;
    push(rk);
    l = Rewritten(u, rk);
    pop(rk);
    push2(rk, l);
    rl = cadr(l);
    r = Lequal(rl, rk) == nil ? true : nil;
    pop3(l, rk, u);
#ifdef F5_DEBUG
    lisp_printf("isRewritten: {%p, %p} --> %p, %p\n", u, rk, l, r);
#endif
    return r;
}


static          Sex
merge_by_degree(Sex u, Sex v)
{
    Sex             s;

    stackcheck();
    push2(u, v);
    pushv(3);

    // begin local a, b, w;
#define u stack[-4]
#define v stack[-3]
#define a stack[-2]
#define b stack[-1]
#define w stack[-0]
    w = nil;
    while (u != nil && v != nil) {
	a = car(u);
	b = car(v);
	if (compare_by_degree(a, b)) {
	    w = cons(a, w);
	    u = cdr(u);
	    continue;
	} else {
	    w = cons(b, w);
	    v = cdr(v);
	    continue;
	}
    }
    s = Lnreverse(w);
    s = Lappend(s, u);
    s = Lappend(s, v);
    popv(5);
#undef u
#undef v
#undef a
#undef b
#undef w
    return s;
}


static          Sex
sort_by_signature(Sex u)
{
    long            i,
                    n;
    Sex             s;

    n = list_length(u);
    if (n <= 1)
	return u;
    stackcheck();
    push2(u, nil);
    // local a, i, n, s;
#define u stack[-1]
#define a stack[-0]
    a = mkarray(n);
    Alist2array(a, u);
    quick_sort_by_signature(a, 0, n - 1);
    s = nil;
    for (i = n - 1; i >= 0; i--) {
	s = cons(array_value(a, i), s);
    }
    popv(2);
#undef u
#undef a
    return s;
}


static          Sex
quick_sort_by_signature(Sex a, long l, long r)
{
    long            ls,
                    rs,
                    ll;
    Sex             s;

    stackcheck();
    push2(a, nil);
    // begin local ls, rs; 
    // local pivot;
    // local ll;
#define a stack[-1]
#define pivot stack[0]


    rs = r;
    ls = l;
    pivot = array_value(a, l);

    while (l < r) {
	while (compare_by_signature(pivot, array_value(a, r)) && l < r) {
	    r = r - 1;
	}
	if (r != l) {
	    array_value(a, l) = array_value(a, r);
	    l = l + 1;
	}
	while (compare_by_signature(array_value(a, l), pivot) && l < r) {
	    l = l + 1;
	}
	if (l != r) {
	    array_value(a, r) = array_value(a, l);
	    r = r - 1;
	}
    }
    array_value(a, l) = pivot;
    ll = l;
    l = ls;
    r = rs;
    if (l < ll) {
	quick_sort_by_signature(a, l, ll - 1);
    }
    if (r > ll) {
	quick_sort_by_signature(a, ll + 1, r);
    }
    s = a;
    popv(2);
#undef a
#undef pivot
    return s;
}


static          Sex
sort_by_degree(Sex u)
{
    long            i,
                    n;
    Sex             s;

    n = list_length(u);
    if (n <= 1)
	return u;
    stackcheck();
    push2(u, nil);
#define u stack[-1]
#define a stack[0]
    // local a, i, n, s;

    a = mkarray(n);
    Alist2array(a, u);
    quick_sort_by_degree(a, 0, n - 1);
    s = nil;
    for (i = n - 1; i >= 0; i--) {
	s = cons(array_value(a, i), s);
    }
#undef u
#undef a
    popv(2);
    return s;
}


static          Sex
quick_sort_by_degree(Sex a, long l, long r)
{
    long            ls,
                    rs,
                    ll;
    Sex             s;

    stackcheck();
    push2(a, nil);

    // begin local ls, rs; 
    // local pivot;
    // local ll;
#define a stack[-1]
#define pivot stack[0]


    rs = r;
    ls = l;
    pivot = array_value(a, l);

    while (l < r) {
	while (compare_by_degree(pivot, array_value(a, r)) && l < r) {
	    r = r - 1;
	}
	if (r != l) {
	    array_value(a, l) = array_value(a, r);
	    l = l + 1;
	}
	while (compare_by_degree(array_value(a, l), pivot) && l < r) {
	    l = l + 1;
	}
	if (l != r) {
	    array_value(a, r) = array_value(a, l);
	    r = r - 1;
	}
    }
    array_value(a, l) = pivot;
    ll = l;
    l = ls;
    r = rs;
    if (l < ll)
	quick_sort_by_degree(a, l, ll - 1);
    if (r > ll)
	quick_sort_by_degree(a, ll + 1, r);
    s = a;
    popv(2);
#undef a
#undef pivot
    return s;
}

static          bool
compare_by_degree(Sex u, Sex v)
{
    long            t1,
                    t2;
    t1 = mkint(Atdeg(car(u)));
    t2 = mkint(Atdeg(car(v)));
    // write "compare ", u, " and ", v;
    return t1 < t2 ? YES : NO;
}

static          bool
compare_by_signature(Sex u, Sex v)
{
    long            ui,
                    vi;
    Sex             s;
    ui = mkint(cadr(u));
    vi = mkint(cadr(v));
    if (ui > vi)
	return YES;
    if (ui < vi)
	return NO;
    if (compare(car(u), car(v)) < 0)
	return YES;
    return NO;
}

static long
compare_signature(Sex u, Sex v)
	/*
	 * 1: u > v -1: u < v 0: u = v 
	 */
{

    long            ui,
                    vi;

    ui = mkint(cadr(u));
    vi = mkint(cadr(v));
    if (ui < vi)
	return 1;
    if (ui > vi)
	return -1;
    u = car(u);
    v = car(v);
    return compare(u, v);
}

static          Sex
minimal_by_signature(Sex u)
{
    Sex             s;

    if (u == nil)
	return u;
    stackcheck();
    push3(u, nil, nil);
#define u stack[-2]
#define v stack[-1]
#define w stack[0]

    // write "minimal_by_signature: u=", u;
    v = car(u);
    u = cdr(u);
    while (u != nil) {
	w = car(u);
	u = cdr(u);
	if (compare_signature(v, w) > 0)
	    v = w;
    }
    s = v;
    popv(3);
#undef u
#undef v
#undef w
    return s;
}


static          Sex
Reduction(Sex Todo, Sex Gp, long k, Sex G)
{
    long            vj;
    Sex             r;

    stackcheck();
    push3(Todo, Gp, G);
    pushv(5);
    // local Done, h, s, rk;
    // local To;
#define Todo stack[-7]
#define Gp stack[-6]
#define G stack[-5]
#define Done stack[-4]
#define h stack[-3]
#define s stack[-2]
#define rk stack[-1]
#define To stack[-0]

    To = Todo;
    Done = nil;
    while (Todo != nil) {
	rk = minimal_by_signature(Todo);
	Todo = Ldelete(rk, Todo);
	/*
	 * h:=NF_with_signature(rk, Gp);
	 */
	// vj = mkint(cadr(rk));
	h = NF(car(cddr(rk)), G);
	// write "NF(", third(rk), " , ", G[second(rk)+1], ")-->", h;
	rk = Llist3(car(rk), cadr(rk), h);
	s = Lappend(Gp, Done);
	s = TopReduction(rk, s, k, G);
	/*
	 * write "reduction:", hh;
	 */
	Done = Lappend(Done, car(s));
	Todo = Lappend(Todo, cadr(s));
    }
#ifdef F5_DEBUG
    // write "Reduction: ", To, " --> ", Done;
    lisp_printf("Reduction: %p --> %p\n", To, Done);
#endif
    r = Done;
    popv(8);
#undef Todo
#undef Gp
#undef G
#undef Done
#undef h
#undef s
#undef rk
#undef To
    return r;
}

static          Sex
IsReducible(Sex r0, Sex G, long k, Sex Gp)
{
    long            k0,
                    kj;
    Sex             r;

    stackcheck();
    push3(r0, G, Gp);
    pushv(7);

    // begin
    // local t, tp, rj;
    // local vj, kj, u;
#define r0 stack[-9]
#define G stack[-8]
#define Gp stack[-7]
#define t0 stack[-6]
#define u2 stack[-5]
#define tj stack[-4]
#define vj stack[-3]
#define rj stack[-2]
#define u stack[-1]
#define u1 stack[-0]

    t0 = car(r0);
    k0 = mkint(cadr(r0));
    // t = HT(t);
    while (G != nil) {
	rj = car(G);
	G = cdr(G);
	tj = car(rj);
	kj = mkint(cadr(rj));

	u = gratio2(HM(car(cddr(r0))), HM(car(cddr(rj))));
	if (!is_poly(u) && !is_number(u))
	    continue;
	u = Ldipoly(u);
	u1 = gtimes2(u, tj);
	u2 = NF(u1, Gp);
	if (Lequal(u1, u2) == nil)
	    continue;
	if (IsRewritten(u, rj) != nil)
	    continue;
	if (kj == k0 && Lequal(u1, t0) != nil)
	    continue;
	r = rj;
	popv(10);
	return r;
    }
    popv(10);
#undef r0
#undef Gp
#undef G
#undef t0
#undef u2
#undef tj
#undef vj
#undef rj
#undef u
#undef u1
    return mkfxat(0);
}

static          Sex
TopReduction(Sex r0, Sex G, long k, Sex Gp)
{
    Sex             r;
    long            k0,
                    k1;

    stackcheck();
    push3(r0, Gp, G);
    pushv(6);
    // begin
    // local p, u;
    // local q, rj, J, r0;
    // global N;
#define r0 stack[-8]
#define Gp stack[-7]
#define G stack[-6]
#define p stack[-5]
#define u stack[-4]
#define q stack[-3]
#define rj stack[-2]
#define rp stack[-1]
#define rk stack[-0]

    // write "TopReduction: ", rk, ", by ", Gp;
    p = car(cddr(r0));
    if (gzerop(p)) {
	aerror("the system is not a regular sequence");
	r = Llist2(nil, nil);
	popv(9);
	return r;
    }
    rp = IsReducible(r0, G, k, Gp);
    if (rp == mkfxat(0)) {
	r = HC(p);
	p = gratio2(p, r);
	r0 = Llist3(car(r0), cadr(r0), p);
	r = Lncons(r0);
	r = Llist2(r, nil);
	popv(9);
	return r;
    }

    rk = rp;
    u = HT(p);
    q = car(cddr(rk));
    r = HT(q);
    u = gratio2(u, r);
    u = Ldipoly(u);

    r = gtimes2(u, q);
    p = gdiff2(p, r);

    k0 = mkint(cadr(r0));
    k1 = mkint(cadr(rk));
    if (k1 > k0) {
	rk = Llist3(car(rk), mkfxat(k0), p);
	rk = Lncons(rk);
	r = Llist2(nil, rk);
	popv(9);
	return r;
    }
    u = gtimes2(u, car(rk));
    if (k0 == k1) {
	if (compare(u, car(r0)) < 0) {
	    r0 = Llist3(car(r0), mkfxat(k0), p);
	    r0 = Lncons(r0);
	    r = Llist2(nil, r0);
	    popv(9);
	    return r;
	}
    }
    F5_N++;
    rk = Llist3(u, mkfxat(k1), p);

#ifdef F5_DEBUG
    lisp_printf("TopReduction: addrules %d\n", F5_N);
#endif
    AddRules(rk, F5_N);
    rk = Llist2(rk, r0);
    r = Llist2(nil, rk);
    popv(9);
#undef rk
#undef Gp
#undef G
#undef p
#undef u
#undef q
#undef rj
#undef rp
#undef r0
    return r;
}


static          Sex
NF_with_signature(Sex f, Sex F)
{
    long            i,
                    j,
                    n;
    Sex             u;

    stackcheck();
    push2(f, F);
    pushv(10);

    // begin local r, g, n;
    // local i, j;
    // local c;
    // local rj, rt, gt, gj;
    // s;
#define f stack[-11]
#define F stack[-10]
#define r stack[-9]
#define g stack[-8]
#define c stack[-7]
#define rj stack[-6]
#define rt stack[-5]
#define gt stack[-4]
#define gj stack[-3]
#define s stack[-2]

    /*
     * write "nf_with_signature: f=", f, ", F=", F;
     */

    n = list_length(F);
    g = mkarray(n);

    for (i = 0; i < n; i++) {
	array_value(g, i) = car(F);
	F = cdr(F);
    }

    rt = car(f);
    f = cdr(f);
    rj = car(f);

    f = cadr(f);
    r = mkfxat(0);

    while (!gzerop(f)) {
	for (j = 0; j < n; j++) {
	    s = array_value(g, j);
	    c = ht_div(f, car(cddr(s)));
	    if (c != nil)
		break;
	}
	if (c == nil) {
	    c = HM(f);
	    r = gplus2(r, c);
	    f = gdiff2(f, c);
	} else {

	    c = Ldipoly(c);
	    gt = gtimes2(car(s), c);
	    s = cdr(s);
	    gj = car(s);
	    s = cadr(s);
	    s = gtimes2(s, c);
	    f = gdiff2(f, s);
	    if (rj > gj) {
		rj = gj;
		rt = gt;
	    } else if (rj == gj && compare(rt, gt) < 0)
		rt = gt;
	}
    }
    /*
     * write "nf_with_signature: r=", r;
     */
    u = Llist3(rt, rj, r);
    popv(12);
#undef f
#undef F
#undef r
#undef g
#undef c
#undef rj
#undef rt
#undef gt
#undef gj
#undef s
    return u;
}

/**************************************************************/
/**************************************************************/
