/*
 * factor.c      (C) Copyright 2015 by Fujio Kako, Nara Women's University.
 * Polynomial Factorization Functions.
 *      $Header: /Users/kako/asuka/src/RCS/factor.c,v 2.30 2016/09/06 05:50:02 kako Exp $
 */

static char const rcsid[] =
    "$Id: factor.c,v 2.30 2016/09/06 05:50:02 kako Exp $";

// #define DEBUG

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "entries.h"
#include "misc.h"
#include "fns.h"
#include "polynomial.h"
#include "tri.h"
#include "garith.h"

#define cddr(x)	cdr(cdr(x))
#define cdar(x)	cdr(car(x))

static Sex      univariate_factor(Sex j);
static Sex      multivariate_factor(Sex j, Sex v, long n);
static Sex      factor1(Sex F);
static Sex      factor2(Sex F);
static Sex      mhensel(Sex a, Sex xv, Sex av, Sex p, Sex pl, Sex l, Sex u,
			Sex lcU);
static Sex      mdiophant(Sex a, Sex c, Sex xv, Sex av, long n, long d,
			  Sex p, Sex pk, Sex k);
static Sex      udiophant(Sex a, Sex x, Sex c, Sex p, Sex pk, Sex k);
static Sex      MultiTermEEAlift(Sex a, Sex x, Sex p, Sex pk, Sex k);
static Sex      EEAlift(Sex a, Sex b, Sex x, Sex p, Sex k);
static Sex      Combine_Multivariate_Factors(Sex F, Sex U, Sex lcU, Sex q,
					     Sex r, Sex y, long n, long m);
static Sex      Combine_Multivariate_nest(Sex F, Sex U, Sex lcU, Sex q,
					  Sex r, Sex y, long k, long c,
					  long n, long m, int *a);

extern Sex      my_gcd_mgcd(Sex f, Sex g);

static Sex      Afactor(Sex a);
Sex             Asquarefree(Sex a);
static Sex      Azfactor(Sex a);
Sex             factorlist2prefix(Sex a);

extern Sex      Lncons(Sex a);
extern Sex      Lnconc(Sex a, Sex b);
extern Sex      Laconc(Sex a, Sex b);
extern Sex      Lclone(Sex a);
extern Sex      Lmkarray1(Sex n);
extern Sex      Alist2array(Sex a, Sex b);

static Sex      factor(Sex);
static Sex      SquareFree(Sex);
static Sex      SquareFree2(Sex);

static Sex      Afactor_mod(Sex f);
static Sex      Afactor_ff(Sex f);
static Sex      Asquarefree2(Sex a);
bool            square_free_mod(Sex a, Sex p);
static Sex      factor_mod(Sex f, Sex p);
static Sex      sqfree(Sex a);
Sex             ddd_factor(Sex a);
static Sex      berlekamp(Sex f, Sex p);
static Sex      MignotteBound(Sex f);
static Sex      uhensel(Sex f, Sex p, Sex g, Sex a, Sex b);
static Sex      FormMatrix(Sex f, Sex q);
static Sex      NullSpaceBase(Sex c);
static Sex      Factor_mod(Sex V, long k, Sex q, Sex f);
static Sex      isqrt(Sex s);
static Sex      iroot(Sex s, Sex b);
static bool     zeroRow(Sex M, long n, long i);
static Sex      Combine_Factors(Sex F, Sex U, Sex q, long n, long m);
static Sex      Combine_nest(Sex F, Sex U, Sex q, long k, long c, long n,
			     long m, Sex a);

static Sex      eea(Sex a, Sex b);
static Sex      multieea(Sex a, Sex x);
static Sex      diophant(Sex a1, Sex a2, Sex c, Sex x);
static Sex      zfactor(Sex a);
static Sex      mcfactors(Sex a, Sex L);
static Sex      mcfactor(long a, long p);
static Sex      big_mcfactor(Sex a, long i);




extern Sex      Amcoeff(Sex a);
extern Sex      Lexpt(Sex a, Sex n);
extern Sex      gexpt(Sex a, Sex n);
extern Sex      Mpp(Sex a);
extern Sex      Mvar(Sex a);
extern Sex      Mdeg(Sex a);
extern Sex      Adeg(Sex f, Sex x);
extern Sex      Lrandom(Sex a);
extern Sex      Lnext_prime(Sex p);
extern Sex      Mlcoeff(Sex a);
extern Sex      Mlpower(Sex a);
extern Sex      Mlterm(Sex a);
extern Sex      Alcoeff(Sex a, Sex x);
extern Sex      Alpower(Sex a, Sex x);
extern Sex      Alterm(Sex a, Sex x);
extern Sex      Amkmodular(Sex a, Sex p);
extern Sex      Aunmodular(Sex a);
extern Sex      Adiff(Sex a, Sex c);
extern Sex      new_vector(long n);
extern Sex      Lident(Sex a);
extern Sex      Labs(Sex a);
extern Sex      Arowvector(Sex m, Sex i);
extern Sex      Acoeff(Sex a, Sex b);
extern Sex      make_matrix(Sex n, Sex m);
extern Sex      Llessp(Sex a, Sex b);
extern long     int_gcd(long a, long b);
extern Sex      gcdjj(Sex a, Sex b);

extern long     last_prime;
extern long     last_prime_squared;
long            maxtrys = 10;
long            confidence = 40;

/**************************************************************/
/**************************************************************/

static          Sex
Afactor(Sex g)
{
    g = factor1(g);
    g = factorlist2prefix(g);
    return g;
}


Sex
factorlist2prefix(Sex g)
{
    Sex             g1;
    push2(g, nil);
#define g stack[-1]
#define p stack[0]
    p = nil;
    while (g != nil) {
#ifdef DEBUG
	lisp_printf("g=%p\n", g);
#endif
	g1 = car(g);
	g = cdr(g);
#if 0
	g1 = cons(ASUKA_CHAR_OF('^'), g1);
#else
	if (cadr(g1) == mkfxat(1))
	    g1 = car(g1);
	else
	    g1 = cons(ASUKA_CHAR_OF('^'), g1);
#endif
	p = cons(g1, p);
    }
    if (cdr(p) != nil) {
	p = Lnreverse(p);
	p = cons(ASUKA_CHAR_OF('*'), p);
    } else {
	p = car(p);
    }
#undef g
    g = new_prefix();
    prefix_form(g) = p;
#undef p
    popv(2);
    return g;
}

Sex
Asquarefree(Sex a)
{
    return SquareFree(a);
}

Sex
Asquarefree2(Sex a)
{
    return SquareFree2(a);
}

static          Sex
Azfactor(Sex a)
{
    a = zfactor(a);
    a = factorlist2prefix(a);
    return a;
}

static          Sex
zfactor(Sex a)
{
    Sex             q;
    long            n,
                    m;
    long            i;
    Sex             j;
    stackcheck();
#define L stack[0]
    if (is_fixnum(a)) {
	push(nil);
	n = mkint(a);
	m = 0;
	while (n % 2 == 0) {
	    m++;
	    n /= 2;
	}
	if (m) {
	    q = Lncons(mkfxat(m));
	    q = cons(mkfxat(2), q);
	    L = Laconc(L, q);
	}
	m = 0;
	while (n % 3 == 0) {
	    m++;
	    n /= 3;
	}
	if (m) {
	    q = Lncons(mkfxat(m));
	    q = cons(mkfxat(3), q);
	    L = Laconc(L, q);
	}
	i = 3;
	while (i * i < n) {
	    i += 2;
	    m = 0;
	    while (n % i == 0) {
		m++;
		n /= i;
	    }
	    if (m) {
		q = Lncons(mkfxat(m));
		q = cons(mkfxat(i), q);
		L = Laconc(L, q);
	    }
	}
	if (n != 1) {
	    q = Lncons(mkfxat(1));
	    q = cons(mkfxat(n), q);
	    L = Laconc(L, q);
	} else if (L == nil) {
	    q = Llist2(mkfxat(1), mkfxat(1));
	    L = Lncons(q);
	}
	pop(q);
	return q;
    }
#undef L
    push3(a, nil, nil);
#define n stack[-2]
#define p stack[-1]
#define L stack[0]
    m = 0;
    while (Levenp(n) != nil) {
	m++;
	n = quot2ji(n, mkfxat(2));
    }
    if (m) {
	q = Lncons(mkfxat(m));
	q = cons(mkfxat(2), q);
	L = Laconc(L, q);
    }
    p = prime_list;
    while (p != nil) {
	j = car(p);
	p = cdr(p);
	m = 0;
	while (rem2ji(n, j) == mkfxat(0)) {
	    m++;
	    n = quot2ji(n, j);
	}
	if (m) {
	    q = Lncons(mkfxat(m));
	    q = cons(j, q);
	    L = Laconc(L, q);
	}
    }
    if (n != mkfxat(1)) {
	q = mcfactors(n, nil);
	// q = Lncons(mkfxat(1));
	// q = cons(n, q);
	// L = Laconc(L, q);
	L = Lnconc(L, q);
    }
    q = L;
    popv(3);
#undef n
#undef p
#undef L
    return q;
}


static          Sex
mcfactors(Sex a, Sex L)
{
    long            i;
    stackcheck();
    push2(a, L);
#define a stack[-1]
#define L stack[0]
    if (Lprimep(a) != nil) {
	a = Llist2(a, mkfxat(1));
	L = Laconc(L, a);
    } else {
	for (i = 1; i < maxtrys; i++) {
	    if (is_fixnum(a))
		a = mcfactor(mkint(a), i);
	    else
		a = big_mcfactor(a, i);

	    if (is_cons(a))
		break;
	}
	if (is_cons(a)) {
	    if (Llessp(car(a), cdr(a)) != nil) {
		L = mcfactors(car(a), L);
		L = mcfactors(cdr(a), L);
	    } else {
		L = mcfactors(cdr(a), L);
		L = mcfactors(car(a), L);
	    }
	} else {
	    a = Llist2(a, mkfxat(1));
	    L = Laconc(L, a);
	}
    }
#undef a
#undef L
    a = stack[-1];
    popv(2);
    return a;
}

static          Sex
mcfactor(long n, long p)
{
    long            g,
                    k,
                    m,
                    q,
                    r,
                    x,
                    y,
                    ys;
    long            ms;
    long            i;

    m = 20;
    y = 0;
    r = q = 1;
    do {
	x = y;
	for (i = 1; i <= r; i++) {
	    y = umulrem(y, y, n);
	    y = (y + p) % n;
	}
	k = 0;
	do {
	    ys = y;
	    ms = m < (r - k) ? m : r - k;
	    for (i = 1; r <= ms; i++) {
		y = umulrem(y, y, n);
		y = (y + p) % n;
		q = umulrem(q, labs(x - y), n);
	    }
	    g = int_gcd(q, n);
	    k = k + m;
	} while (k < r && g <= 1);
	r = 2 * r;
    } while (g <= 1);
    if (g == n) {
	do {
	    ys = umulrem(ys, ys, n);
	    ys = (ys + p) % n;
	    g = int_gcd(labs(y - ys), n);
	} while (g <= 1);
    }
    if (g == n)
	return mkfxat(n);
    m = n / g;
    return cons(mkfxat(g), mkfxat(m));
}



static          Sex
big_mcfactor(Sex n, long p)
{
    Sex             a;
    long            i,
                    k,
                    m,
                    r;
    long            ms;

    stackcheck();
    push(n);
    pushv(5);
#define n stack[-5]
#define q stack[-4]
#define x stack[-3]
#define y stack[-2]
#define ys stack[-1]
#define g stack[-0]
    m = 20;
    y = mkfxat(0);
    r = 1;
    q = mkfxat(1);
    do {
	x = y;
	for (i = 1; i <= r; i++) {
	    y = times2jj(y, y);
	    y = plus2ji(y, mkfxat(p));
	    y = rem2jj(y, n);
	}
	k = 0;
	do {
	    ys = y;
	    ms = m < (r - k) ? m : r - k;
	    for (i = 1; i <= ms; i++) {
		y = times2jj(y, y);
		y = plus2ji(y, mkfxat(p));
		y = rem2jj(y, n);
		g = diff2jj(x, y);
		g = absj(g);
		q = times2jj(q, g);
		q = rem2jj(q, n);
	    }
	    g = gcdjj(q, n);
	    k = k + m;
	} while (k < r && (g == mkfxat(0) || g == mkfxat(1)));
	r = 2 * r;
    } while (g == mkfxat(0) || g == mkfxat(1));
    if (Leqn(g, n) != nil) {
	do {
	    ys = times2jj(ys, ys);
	    ys = plus2ji(ys, mkfxat(p));
	    ys = rem2jj(ys, n);
	    g = diff2jj(y, ys);
	    g = absj(g);
	    g = gcdjj(g, n);
	} while (g == mkfxat(0) || g == mkfxat(1));
    }
    if (Leqn(q, n) != nil) {
	a = n;
    } else {
	m = quot2jj(n, g);
	a = cons(g, m);
    }
#undef n
#undef q
#undef x
#undef y
#undef ys
#undef g
    popv(6);
    return a;
}




/**************************************************************/
/**************************************************************/
#if 0
static          Sex
Afactor_mod(Sex f)
{
    long            p;
    Sex             a;

    stackcheck();
    push(f);
#define f stack[-0]

    p = 3;
    while (!square_free_mod(f, mkfxat(p)))
	p = Lnext_prime(mkfxat(p));
    a = factor_mod(f, mkfxat(p));
    popv(1);
#undef f
    return a;
}
#endif

#if 0
static          Sex
Afactor_ff(Sex f)
{
    Sex             p;
    stackcheck();
    push5(f, nil, nil, nil, nil);
    push3(nil, nil, nil);
#define f stack[-7]
#define l stack[-6]
#define r stack[-5]
#define s stack[-4]
#define t stack[-3]
#define a stack[-2]
#define b stack[-1]
#define e stack[-0]
    l = sqfree(f);
    r = nil;
    while (t != nil) {
	s = car(t);
	a = car(s);
	e = cadr(s);
	t = cdr(t);
	t = ddd_factor(a);
	s = Llist2(b, e);
	r = Lappend(r, s);
    }
    p = r;
    popv(8);
#undef f
#undef l
#undef r
#undef s
#undef t
#undef a
#undef b
#undef e
    return p;
}
#endif

static          Sex
factor2(Sex F)
{
    Sex             s;
    long            i,
                    n;
    stackcheck();
    push4(F, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
#define F stack[-13]
#define H stack[-12]
#define x stack[-10]
#define f stack[-9]
#define g stack[-8]
#define h stack[-7]
#define p stack[-6]
#define m stack[-5]
#define b stack[-3]
#define j stack[-2]
#define I stack[-1]
#define Hs stack[-0]
    p = Mcont(F);
    F = gratio2(F, p);
    x = Mvar(F);
    H = SquareFree(F);
    n = Llength(H);
    I = nil;
    if (!onep(p)) {
	// s = Llist2(p, mkfxat(1));
	// I = Lncons(s);
	I = zfactor(p);
    }
    Hs = H;
    // lisp_printf("factor: Hs = %p\n", Hs);
    while (Hs != nil) {
	j = car(Hs);
	Hs = cdr(Hs);
	f = car(j);
	if (!is_poly(f))
	    continue;
	if (Mdeg(f) < mkfxat(2)) {
	    I = Laconc(I, j);
	    continue;
	}
	p = Lrandom(mkfxat(100));
	while (1) {
	    p = Lnext_prime(p);
	    m = Mlcoeff(f);
	    if (gzerop(gremaind2(f, p)))
		continue;
	    g = Amkmodular(f, p);
	    s = Adiff(g, x);
	    s = Mgcd(g, s);
	    if (!is_poly(s))
		break;
	}
	// lisp_printf("factor: prime = %p\n", p);
	g = berlekamp(f, p);
	// lisp_printf("factor: g=%p\n", g);
	n = mkint(Llength(g));
	if (n == 1) {
	    s = Llist2(car(g), cadr(j));
	    I = Laconc(I, s);
	} else {
	    b = MignotteBound(f);
	    // lisp_printf("mignotte bound = %p\n", b);
	    h = uhensel(f, p, g, nil, b);
	    h = car(h);
	    // lisp_printf("hense = %p\n", h);
	    n = mkint(car(array_dim(h)));
	    for (i = 0; i < n; i++) {
		s = array_value(h, i);
		s = Mpp(s);
		s = Llist2(s, cadr(j));
		I = Laconc(I, s);
	    }
	}
    }

    s = I;
    popv(14);
#undef F
#undef H
#undef x
#undef f
#undef g
#undef h
#undef p
#undef m
#undef b
#undef j
#undef I
#undef Hs
    return s;
}

/**************************************************************/
/**************************************************************/
bool
square_free_mod(Sex f, Sex q)
{
    bool            r;
    stackcheck();
    push4(f, q, nil, nil);
#define f stack[-3]
#define q stack[-2]
#define x stack[-1]
#define g stack[-0]
    f = Amkmodular(f, q);
    x = Mvar(f);
    g = Adiff(f, x);
    g = Mgcd(f, g);
    r = !is_poly(g);
    popv(4);
#undef f
#undef q
#undef x
#undef g
    return r;
}

static          Sex
SquareFree(Sex a)
{
    Sex             s;
    long            i;
    stackcheck();
    push(a);
    pushv(8);
#define a stack[-8]
#define o stack[-7]
#define b stack[-6]
#define c stack[-5]
#define w stack[-4]
#define x stack[-3]
#define y stack[-2]
#define z stack[-1]
#define g stack[-0]
    i = 1;
    o = nil;
    x = Mvar(a);
    b = Adiff(a, x);
    c = my_gcd_mgcd(a, b);
    w = gquot2(a, c);
    while (is_poly(c)) {
	y = my_gcd_mgcd(w, c);
	z = gquot2(w, y);
	if (!gonep(z)) {
	    s = Llist2(z, mkfxat(i));
	    o = cons(s, o);
	}
	i++;
	w = y;
	c = gquot2(c, y);
    }
    if (!gonep(w)) {
	s = Llist2(w, mkfxat(i));
	o = cons(s, o);
    }
#undef a
    a = o;
#undef o
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
#undef g
    popv(9);
    return a;
}

static          Sex
SquareFree2(Sex a)
{
    Sex             s;
    long            i;
    stackcheck();
    push(a);
    pushv(8);
#define a stack[-8]
#define o stack[-7]
#define b stack[-6]
#define c stack[-5]
#define w stack[-4]
#define x stack[-3]
#define y stack[-2]
#define z stack[-1]
#define g stack[-0]
    i = 1;
    o = nil;
    x = Mvar(a);
    b = Adiff(a, x);
    c = Mgcd(a, b);
    if (c == mkfxat(1))
	w = a;
    else {
	w = gquot2(a, c);
	y = gquot2(b, c);
	s = Adiff(w, x);
	z = gdiff2(y, s);
	while (!gzerop(z)) {
	    g = Mgcd(w, z);
	    s = Llist2(g, mkfxat(i));
	    o = cons(s, o);
	    i++;
	    w = gquot2(w, g);
	    y = gquot2(z, g);
	    s = Adiff(w, x);
	    z = gdiff2(y, s);
	}
    }
    s = Llist2(w, mkfxat(i));
    o = cons(s, o);
#undef a
    a = o;
#undef o
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
#undef g
    popv(9);
    return a;
}


static          Sex
squarerootFF(Sex a, Sex q)
{
    Sex             s,
                    u;
    long            i,
                    j,
                    n;
    long            k;

    k = mkint(q);
    stackcheck();
    push(a);
    pushv(5);
#define a stack[-6]
#define b stack[-4]
#define c stack[-3]
#define x stack[-2]
#define u stack[-1]
    x = Mvar(a);
    c = Amcoeff(a);
    n = mkint(Llength(c));
    n = (n - 1) / k;
    b = mkint(0);
    for (i = 0; i <= n; i++) {
	u = gexpt(x, mkfxat(i));
	s = array_value(c, i * k);
	s = gtimes2(s, u);
	b = gplus2(b, s);
    }
    s = b;
#undef a
#undef b
#undef c
#undef x
#undef u
    popv(6);
    return s;
}

static          Sex
SquareFreeFF(Sex a, Sex q)
{
    Sex             s;
    long            i;
    stackcheck();
    push2(a, q);
    pushv(7);
#define a stack[-8]
#define q stack[-7]
#define o stack[-6]
#define b stack[-5]
#define c stack[-4]
#define w stack[-3]
#define x stack[-2]
#define y stack[-1]
#define z stack[-0]
    i = 1;
    o = nil;
    x = Mvar(a);
    a = Amkmodular(a, q);
    b = Adiff(a, x);
    if (b != mkfxat(0)) {
	c = Mgcd(a, b);
	w = gquot2(a, c);
	while (w != mkfxat(1)) {
	    y = Mgcd(w, c);
	    z = gquot2(w, y);
	    s = Llist2(z, mkfxat(i));
	    o = cons(s, o);
	    i++;
	    w = y;
	    c = gquot2(c, y);
	}
	if (c != mkfxat(1)) {
	    c = squarerootFF(c, q);
	    s = SquareFreeFF(c, q);
	    s = Llist2(s, q);
	    o = cons(s, o);
	}
    } else {
	a = squarerootFF(c, q);
	s = SquareFreeFF(a, q);
	s = Llist2(s, q);
	o = cons(s, o);
    }
#undef a
#undef q
    a = o;
#undef o
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
    popv(9);
    return a;
}
/**************************************************************/
/**************************************************************/
static          Sex
berlekamp(Sex f, Sex q)
{
    Sex             s;
    long            i,
                    k,
                    m;
    long            n;
    stackcheck();
    push2(f, q);
    pushv(13);
#define f stack[-14]
#define q stack[-13]
#define v stack[-12]
#define g stack[-11]
#define factors stack[-10]
#define prime_factors stack[-9]
#define h stack[-4]
#define Q stack[-3]
#define K stack[-2]
#define V stack[-1]
#define hh stack[-0]

    n = mkint(Mdeg(f));
    Q = FormMatrix(f, q);
    s = Lident(mkfxat(n));
    s = gdiff2(Q, s);
    v = NullSpaceBase(s);
    // lisp_printf("nullspace = %p\n", v);
    k = mkint(Llength(v));
    h = Mvar(f);
    h = symbol_to_polynomial(h);
    V = new_vector(k);
    hh = new_vector(n);
    for (i = 0; i < n; i++) {
	s = gexpt(h, mkfxat(i));
	vect_value(hh, i) = s;
    }
    for (i = 1; i <= k; i++) {
	s = gtimes2(hh, car(v));
	s = Amkmodular(s, q);
	vect_value(V, i - 1) = s;
	v = cdr(v);
    }
    // lisp_printf("V=%p\n", V);
    factors = Lncons(f);
    prime_factors = nil;
    m = mkint(Llength(prime_factors)) + mkint(Llength(factors));
    while (m < k) {
	// lisp_printf("m = %d, k = %d\n", m, k);
	f = car(factors);
	factors = cdr(factors);
	// lisp_printf("berlekamp: f=%p\n", f);
	// lisp_printf("k = %d\n", k);
	// lisp_printf("factors = %p\n", factors);
	g = Factor_mod(V, k, q, f);
	// lisp_printf("factor mod g = %p\n", g);
	if (g != nil) {
	    factors = Lappend(factors, g);
	    m = m + mkint(Llength(g));
	} else {
	    s = Lncons(f);
	    // prime_factors = Laconc(prime_factors, f);
	    prime_factors = Lappend(prime_factors, s);
	    // m = m + mkint(Llength(f));
	}
	m = mkint(Llength(prime_factors)) + mkint(Llength(factors));
    }
    s = Lappend(factors, prime_factors);
#undef f
#undef q
#undef v
#undef g
#undef factors
#undef prime_factors
#undef n
#undef h
#undef Q
#undef K
#undef V
#undef hh
    popv(15);
    return s;
}

static          Sex
Factor_mod(Sex V, long k, Sex q, Sex f)
{
    long            i,
                    j,
                    l;
    Sex             s;

    l = mkint(q);
    stackcheck();
    push3(V, q, f);
    pushv(3);
#define V stack[-5]
#define q stack[-4]
#define f stack[-3]
#define g stack[-2]
    // lisp_printf("Factor_mod: f=%p\n", f);
    f = Amkmodular(f, q);
    for (i = 1; i <= k; i++) {
	s = vect_value(V, i - 1);
	if (!is_poly(s))
	    continue;
	for (j = 0; j < l; j++) {
	    s = vect_value(V, i - 1);
	    s = gdiff2(s, mkfxat(j));
	    g = Mgcd(s, f);
	    if (is_poly(g)) {
		s = gratio2(f, g);
		if (is_poly(s)) {
		    s = Llist2(g, s);
		    popv(6);
		    return s;
		}
	    }
	}
    }
    popv(6);
#undef V
#undef q
#undef f
#undef g
    return nil;
}

static          Sex
FormMatrix(Sex F, Sex q)
{
    Sex             s;
    long            i,
                    j;
    long            n,
                    m;
    stackcheck();
    push2(F, q);
    pushv(9);
#define F stack[-10]
#define q stack[-9]
#define f stack[-6]
#define k stack[-4]
#define a stack[-3]
#define vq stack[-2]
#define v stack[-1]
#define M stack[-0]
    f = Amkmodular(F, q);
    v = Mvar(F);
    n = mkint(Adeg(f, v));
    M = make_matrix(n, n);
    k = mkfxat(1);
    // matrix_value(M, 1,1 ) = mkfxat(1);
    matrix_value(M, 0) = mkfxat(1);
    vq = make_polynomial(v, mkfxat(1), q);
    if (n >= q)
	vq = gremaind2(vq, f);
    for (i = 1; i < n; i++) {
	k = gtimes2(k, vq);
	k = gremaind2(k, f);
	a = Acoeff(k, v);
	m = mkint(car(array_dim(a)));
	for (j = 1; j <= m; j++) {
	    s = array_value(a, j - 1);
	    matrix_value(M, n * (i) + j - 1) = s;
	}
    }
    s = M;
#undef F
#undef q
#undef f
#undef k
#undef a
#undef vq
#undef v
#undef M
    popv(11);
    return s;
}

static          Sex
NullSpaceBase(Sex M)
{
    Sex             s,
                    u;
    long            n,
                    m,
                    i,
                    j,
                    k,
                    l;

    stackcheck();
    push(M);
    pushv(8);
#define M stack[-8]
#define v stack[-7]
#define J stack[-5]
#define p stack[-4]
    // s = Asize(M);
    // m = mkint(car(s));
    // n = mkint(cadr(s));
    m = mkint(matrix_dimx(M));
    n = mkint(matrix_dimy(M));
    for (k = 1; k <= n; k++) {
	for (i = k; i <= n; i++) {
	    // s = matrix_value(M, (k - 1) * m + i - 1);
	    s = matrix_value(M, (k - 1) * n + i - 1);
	    if (!gzerop(s))
		break;
	}
	if (i <= n) {
	    // p = matrix_value(M, (i - 1) * n + i - 1);
	    p = matrix_value(M, (k - 1) * n + i - 1);
	    for (l = 1; l <= n; l++) {
		s = matrix_value(M, (l - 1) * n + i - 1);
		s = gratio2(s, p);
		matrix_value(M, (l - 1) * n + i - 1) = s;
	    }
	    for (j = 1; j <= n; j++) {
		J = matrix_value(M, (j - 1) * n + k - 1);
		matrix_value(M, (j - 1) * n + k - 1) =
		    matrix_value(M, (j - 1) * n + i - 1);
		matrix_value(M, (j - 1) * n + i - 1) = J;
	    }
	    for (i = 1; i <= n; i++) {
		if (i == k)
		    continue;

		p = matrix_value(M, (k - 1) * n + i - 1);
		for (l = k; l <= n; l++) {
		    s = matrix_value(M, (l - 1) * n + k - 1);
		    s = gtimes2(s, p);
		    u = matrix_value(M, (l - 1) * n + i - 1);
		    s = gdiff2(u, s);
		    matrix_value(M, (l - 1) * n + i - 1) = s;
		}
	    }
	}
    }
    v = nil;
    for (i = 1; i <= n; i++) {
	s = matrix_value(M, (i - 1) * n + i - 1);
	s = gdiff2(s, mkfxat(1));
	matrix_value(M, (i - 1) * n + i - 1) = s;
    }
    for (i = 1; i <= n; i++) {
	if (!zeroRow(M, n, i)) {
	    s = Arowvector(M, mkfxat(i));
	    v = Laconc(v, s);
	}
    }
    s = v;
#undef M
#undef v
#undef J
#undef p
    popv(9);
    return s;
}

static          bool
zeroRow(Sex M, long n, long i)
{
    long            k;
    Sex             s;

    stackcheck();
    push(M);
    for (k = 1; k <= n; k++) {
	M = stack[0];
	s = matrix_value(M, (i - 1) * n + k - 1);
	if (!gzerop(s)) {
	    popv(1);
	    return NO;
	}
    }
    popv(1);
    return YES;
}


/**************************************************************/
/**************************************************************/

/*
 * univariate hensel construction 
 */

static          Sex
MignotteBound(Sex f)
{
    Sex             r;
    long            i,
                    n;

    if (!is_poly(f))
	return Labs(f);

    stackcheck();
    push(f);
    pushv(6);
#define f stack[-6]
#define x stack[-5]
#define a stack[-4]
    // #define i stack[-3]
    // #define n stack[-2]
#define s stack[-1]
#define m stack[-0]



    x = Mvar(f);
    m = Mdeg(f);
    a = Acoeff(f, x);
    n = mkint(car(array_dim(a)));
    s = mkfxat(0);
    for (i = 0; i < n; i++) {
	r = array_value(a, i);
	r = gtimes2(r, r);
	s = gplus2(s, r);
    }

    // lisp_printf("calculate isqrt of %p\n", s);
    s = isqrt(s);
    r = array_value(a, n - 1);
    r = quot2(s, r);
    s = plus2(r, mkfxat(1));
    m = quot2(plus2(m, mkfxat(1)), mkfxat(2));
    r = Lexpt(mkfxat(2), m);
    r = times2(s, r);
#undef f
#undef x
#undef a
    // #undef i
    // #undef n
#undef s
#undef m
    popv(7);
    return r;
}


static          Sex
isqrt(Sex m)
{				/* integer square root of m */
    Sex             s;
    int             j;
    if (m == mkfxat(0))
	return m;
    stackcheck();
    push(m);
    pushv(3);
#define m stack[-3]
#define x stack[-2]
#define x1 stack[-1]
#define bk stack[-0]
    j = number_log2(m) + 2;
    j = j / 2;
    x1 = mkfxat(j);
    x = Lexpt(mkfxat(2), x1);
    s = quot2(m, x);
    bk = diff2(x, s);
    while (plusp(bk)) {
	// lisp_printf("x = %p, bk=%p\n", x, bk);
	s = plus2(bk, mkfxat(1));
	s = quot2(s, mkfxat(2));
	x = diff2(x, s);
	s = quot2(m, x);
	bk = diff2(x, s);
    }
    s = x;
    popv(4);
#undef m
#undef x
#undef x1
#undef bk
    return s;
}

static          Sex
iroot(Sex m, Sex n)
{				/* integer n-th root of m */
    Sex             s;
    long            l;
    if (m == mkfxat(0))
	return m;
    stackcheck();
    push2(m, n);
    pushv(4);
#define m stack[-5]
#define n stack[-4]
#define n1 stack[-3]
#define x stack[-2]
#define x1 stack[-1]
#define bk stack[-0]
    l = number_log2(m) + 2;
    x1 = quot2(mkfxat(l), n);
    x = Lexpt(mkfxat(2), x1);
    n1 = diff2(n, mkfxat(1));
    x1 = Lexpt(x, n1);

    s = quot2(m, x1);
    bk = diff2(x, s);
    while (plusp(bk)) {
	s = diff2(n, mkfxat(1));
	s = plus2(bk, mkfxat(1));
	bk = quot2(s, n);
	x = diff2(x, bk);
	x1 = Lexpt(x, n1);
	s = quot2(m, x1);
	bk = diff2(x, s);
    }
    s = x;
    popv(6);
#undef m
#undef n
#undef n1
#undef x
#undef x1
#undef bk
    return s;
}


/*
 * Univariate Hensel Lifting Algorithm 
 */


static          Sex
EEA(Sex a, Sex b, Sex x)
{
    Sex             u;
    long            n;

    stackcheck();
    push3(a, b, x);
    pushv(14);
#define a stack[-16]
#define b stack[-15]
#define x stack[-14]
#define s stack[-13]
#define t stack[-12]
#define c stack[-11]
#define d stack[-10]
#define c1 stack[-9]
#define c2 stack[-8]
#define d1 stack[-7]
#define d2 stack[-6]
#define r stack[-5]
#define r1 stack[-4]
#define r2 stack[-3]
#define v stack[-2]
#define q stack[-1]
#define g stack[-0]
    u = Alcoeff(a, x);
    c = gratio2(a, u);
    u = Alcoeff(b, x);
    d = gratio2(b, u);
    c1 = d2 = mkfxat(1);
    d1 = c2 = mkfxat(0);

    n = 0;
    while (!gzerop(d)) {
	n++;
	q = grquotient2(c, d, x);
	u = gtimes2(q, d);
	r = gdiff2(c, u);
	u = gtimes2(q, d1);
	r1 = gdiff2(c1, u);
	u = gtimes2(q, d2);
	r2 = gdiff2(c2, u);
	c = d;
	c1 = d1;
	c2 = d2;
	d = r;
	d1 = r1;
	d2 = r2;
    }
    /*
     * normalize GCD 
     */
    v = Alcoeff(c, x);
    g = gratio2(c, v);
    // v = Alcoeff(c, x);
    s = Alcoeff(a, x);
    s = gtimes2(s, v);
    t = Alcoeff(b, x);
    t = gtimes2(t, v);
    v = gratio2(c1, s);
    s = v;
    v = gratio2(c2, t);
    t = v;
    u = Llist3(g, s, t);
    popv(17);
#undef a
#undef b
#undef x
#undef s
#undef t
#undef c
#undef d
#undef c1
#undef c2
#undef d1
#undef d2
#undef r
#undef r1
#undef r2
#undef v
#undef q
#undef g
    return u;
}

static          Sex
replace_lc(Sex f, Sex a, Sex x)
{
    Sex             u;
    stackcheck();
    push3(f, a, x);
#define f stack[-2]
#define a stack[-1]
#define x stack[-0]
    u = Alpower(f, x);
    a = gtimes2(a, u);
    u = Alterm(f, x);
    f = gdiff2(f, u);
    u = gplus2(f, a);
#undef f
#undef a
#undef x
    popv(3);
    return u;
}

static          Sex
multivariate_replace_lc(Sex f, Sex a)
{
    Sex             u;
    stackcheck();
    push3(f, a, nil);
#define f stack[-2]
#define a stack[-1]
#define c stack[-0]
    c = Mlcoeff(f);
    if (is_poly(c)) {
	c = multivariate_replace_lc(c, a);
    } else {
	c = a;
    }
    u = Mlpower(f);
    c = gtimes2(c, u);
    u = Mlterm(f);
    f = gdiff2(f, u);
    u = gplus2(f, c);
    popv(3);
#undef f
#undef a
#undef c
    return u;
}

static          Sex
UnivariateHensel(Sex a, Sex p, Sex u1, Sex w1, Sex B, Sex gamma)
{
    Sex             z;

    stackcheck();
    push5(a, p, u1, w1, B);
    push(gamma);
    pushv(18);
#define a stack[-23]
#define p stack[-22]
#define u1 stack[-21]
#define w1 stack[-20]
#define B stack[-19]
#define gamma stack[-18]
    // local aa, x;
    // local alpha, delta, sigma, tau, modulus;
#define aa stack[-17]
#define x stack[-16]
#define alpha stack[-15]
#define delta stack[-14]
#define sigma stack[-13]
#define tau stack[-12]
#define modulus stack[-11]
    // local e, g, q, r, s, t;
    // local u, w, sh, th;
    // local c;
#define e stack[-10]
#define g stack[-9]
#define q stack[-8]
#define r stack[-7]
#define s stack[-6]
#define t stack[-5]
#define u stack[-4]
#define w stack[-3]
#define sh stack[-2]
#define th stack[-1]
#define c stack[-0]

    x = Mvar(a);

    /*
     * step 1 
     */

    alpha = Alcoeff(a, x);
    if (gamma == nil)
	gamma = alpha;
    a = gtimes2(a, gamma);
    s = Alcoeff(u1, x);
    t = Alcoeff(w1, x);
    u1 = Amkmodular(u1, p);
    w1 = Amkmodular(w1, p);
    u1 = gtimes2(gamma, u1);
    u1 = gratio2(u1, s);
    w1 = gtimes2(alpha, w1);
    w1 = gratio2(w1, t);

    /*
     * step 2
     */

    g = EEA(u1, w1, x);
    s = cadr(g);
    t = cadr(cdr(g));
    g = car(g);
    /*
     * step 3 
     */
    z = Aunmodular(u1);
    u = replace_lc(z, gamma, x);
    z = Aunmodular(w1);
    w = replace_lc(z, alpha, x);
    z = gtimes2(u, w);
    e = gdiff2(a, z);
    modulus = p;
    B = gtimes2(B, mkfxat(2));
    B = gtimes2(B, gamma);
    /*
     * step 4 
     */
    while (!zerop(e) && Llessp(modulus, B) != nil) {
	/*
	 * write "modulus = ", modulus, ", error = ", e;
	 */
	/*
	 * step 4.1 
	 */
	c = gquot2(e, modulus);
	sh = gtimes2(s, c);
	th = gtimes2(t, c);
	g = grdivide2(sh, w1, x);
	q = car(g);
	r = cdr(g);
	sigma = r;
	z = gtimes2(q, u1);
	tau = gplus2(th, z);
	/*
	 * step 4.2 
	 */
	z = Aunmodular(tau);
	z = gtimes2(z, modulus);
	u = gplus2(u, z);
	z = Aunmodular(sigma);
	z = gtimes2(z, modulus);
	w = gplus2(w, z);
	z = gtimes2(u, w);
	e = gdiff2(a, z);
	modulus = gtimes2(modulus, p);
    }
    /*
     * step 5 
     */
    if (zerop(e)) {
	delta = Mcont(u);
	u = gquot2(u, delta);
	z = gquot2(gamma, delta);
	w = gquot2(w, z);
	z = Llist3(u, w, modulus);
    } else
	z = nil;
    popv(24);
#undef a
#undef p
#undef u1
#undef w1
#undef B
#undef gamma
    // local aa, x;
    // local alpha, delta, sigma, tau, modulus;
#undef aa
#undef x
#undef alpha
#undef delta
#undef sigma
#undef tau
#undef modulus
    // local e, g, q, r, s, t;
    // local u, w, sh, th;
    // local c;
#undef e
#undef g
#undef q
#undef r
#undef s
#undef t
#undef u
#undef w
#undef sh
#undef th
#undef c
    return z;
}





static          Sex
multieea(Sex a, Sex x)
{
    Sex             u;
    long            j,
                    n;
    stackcheck();
    push2(a, x);
    pushv(6);
#define a stack[-7]
#define x stack[-6]
#define q stack[-4]
#define s stack[-3]
#define b stack[-1]
#define c stack[-0]
    // begin local n, q, s, j;
    // local b, c;

    // lisp_printf("multieea a = %p, %p\n", a, x);
    n = mkint(car(array_dim(a)));
    q = mkarray(n);
    s = mkarray(n);


    u = array_value(a, n - 1);
    array_value(q, n - 2) = u;
    for (j = n - 3; j >= 0; j--) {
	u = gtimes2(array_value(a, j + 1), array_value(q, j + 1));
	array_value(q, j) = u;
    }

    b = mkfxat(1);
    for (j = 0; j < n - 1; j++) {
	c = diophant(array_value(q, j), array_value(a, j), b, x);
	array_value(s, j) = cadr(c);
	b = car(c);
    }
    array_value(s, n - 1) = b;
    u = s;
    popv(8);
#undef a
#undef x
#undef q
#undef s
#undef b
#undef c

    return u;
}


static          Sex
diophant(Sex a1, Sex a2, Sex c, Sex x)
{
    Sex             u;
    stackcheck();
    push4(a1, a2, c, x);
    pushv(4);
#define a1 stack[-7]
#define a2 stack[-6]
#define c stack[-5]
#define x stack[-4]
#define s1 stack[-3]
#define s2 stack[-2]
#define s stack[-1]
#define d stack[-0]
    // begin local s1, s2, s, d;

    // lisp_printf("eea: %p, %p\n", a2, a1);
    s = eea(a2, a1);
    u = gtimes2(c, car(s));
    d = grdivide2(u, a1, x);
    s1 = cdr(d);

    a1 = gtimes2(c, cadr(s));
    a2 = gtimes2(a2, car(d));
    s2 = gplus2(a1, a2);
    u = Llist2(s1, s2);
    popv(8);
#undef a1
#undef a2
#undef c
#undef x
#undef s1
#undef s2
#undef s
#undef d
    return u;
}


static          Sex
eea(Sex a, Sex b)
{
    Sex             u;
    long            n;
    stackcheck();
    push2(a, b);
    pushv(14);

    // begin local s, t;
    // local c, d, c1, c2, d1, d2;
    // local r, r1, r2;
    // local n;
    // local q, g;
#define a stack[-15]
#define b stack[-14]
#define s stack[-13]
#define t stack[-12]
#define c stack[-11]
#define d stack[-10]
#define c1 stack[-9]
#define c2 stack[-8]
#define d1 stack[-7]
#define d2 stack[-6]
#define r stack[-5]
#define r1 stack[-4]
#define r2 stack[-3]
#define q stack[-1]
#define g stack[-0]

    u = Mlcoeff(a);
    c = gratio2(a, u);
    u = Mlcoeff(b);
    d = gratio2(b, u);
    c1 = d2 = mkfxat(1);
    d1 = c2 = mkfxat(0);

    n = 0;
    while (!zerop(d)) {
	n++;
	q = gquot2(c, d);
	u = gtimes2(q, d);
	r = gdiff2(c, u);
	u = gtimes2(q, d1);
	r1 = gdiff2(c1, u);
	u = gtimes2(q, d2);
	r2 = gdiff2(c2, u);
	c = d;
	c1 = d1;
	c2 = d2;
	d = r;
	d1 = r1;
	d2 = r2;
    }
    /*
     * normalize GCD 
     */
    if (is_poly(g)) {
	aerror("eea: arguments have common divider");
    }

    d = Mlcoeff(c);
    d1 = Mlcoeff(a);
    d2 = Mlcoeff(b);
    d1 = gtimes2(d, d1);
    d2 = gtimes2(d, d2);
    s = gratio2(c1, d1);
    t = gratio2(c2, d2);
    u = Llist2(s, t);

    popv(16);
#undef a
#undef b
#undef s
#undef t
#undef c
#undef d
#undef c1
#undef c2
#undef d1
#undef d2
#undef r
#undef r1
#undef r2
#undef q
#undef g
    return u;
}



static          Sex
uhensel(Sex a, Sex p, Sex u, Sex w, Sex B)
{
    Sex             z;
    long            n,
                    i,
                    j,
                    k;
    long            m;
    stackcheck();
    push5(a, p, u, w, B);
#define a stack[-29]
#define p stack[-28]
#define u stack[-27]
#define w stack[-26]
#define B stack[-25]
    pushv(25);
    // local aa, x;
    // local alpha, gamma, delta, sigma, tau, modulus;
#define aa stack[-24]
#define x stack[-23]
#define alpha stack[-22]
#define gamma stack[-21]
#define delta stack[-20]
#define sigma stack[-19]
#define tau stack[-18]
#define modulus stack[-17]
    // local e, g, q, r, s, t;
#define e stack[-16]
#define g stack[-15]
#define q stack[-14]
#define r stack[-13]
#define s stack[-12]
#define t stack[-11]
    // local sh, th;
#define sh stack[-10]
#define th stack[-9]
    // local c;
    // local U, W, S, T, n, i, j, k;
#define c stack[-8]
#define U stack[-7]
#define W stack[-6]
#define S stack[-5]
#define T stack[-4]

    x = Mvar(a);
    z = Llength(u);
    n = mkint(z);
    U = mkarray(n);
    W = mkarray(n);
    Alist2array(U, u);
    if (w != nil)
	Alist2array(W, w);
    /*
     * write "uhensel: p=", p; write "modulus of u is " ,
     * lisp:modular_modulus(lcoeff(U[0],x)); 
     */

    /*
     * step 1 
     */

    alpha = Mlcoeff(a);
    if (w == nil) {
	for (i = 0; i < n; i++) {
	    array_value(W, i) = alpha;
	}
    }
    gamma = mkfxat(1);
    for (i = 0; i < n; i++) {
	s = array_value(W, i);
	gamma = gtimes2(gamma, s);
	t = array_value(U, i);
	r = Mlcoeff(t);
	t = Amkmodular(t, p);
	t = gtimes2(s, t);
	t = gratio2(t, r);
	array_value(U, i) = t;
    }
    gamma = gratio2(gamma, alpha);
    a = gtimes2(a, gamma);


    /*
     * step 2
     */

    S = multieea(U, x);		/* U is modular poly */
    // lisp_printf("step2 U=%p\n", U);
    // lisp_printf("step2 S=%p\n", S);

    T = mkarray(n);
    /*
     * step 3 
     */
    e = mkfxat(1);
    for (i = 0; i < n; i++) {
	s = array_value(U, i);
	t = array_value(W, i);
	s = Aunmodular(s);
	s = replace_lc(s, t, x);
	array_value(U, i) = s;
	e = gtimes2(e, s);
    }
    e = gdiff2(a, e);
    /*
     * write "e = ", e; 
     */
    modulus = p;
    m = 1;
    /*
     * step 4 
     */
    B = times2(B, mkfxat(2));
    B = times2(B, gamma);
    B = Labs(B);
    while (!gzerop(e) && Llessp(modulus, B) != nil) {
	/*
	 * write "modulus = ", modulus, ", error = ", e;
	 */
	/*
	 * step 4 
	 */

	c = gratio2(e, modulus);
	/*
	 * write "c=", c; 
	 */
	th = mkfxat(0);
	for (i = 0; i < n; i++) {
	    // lisp_printf("i = %d, c = %p\n", i, c);
	    sh = gtimes2(array_value(S, i), c);
	    g = grdivide2(sh, array_value(U, i), x);
	    th = gplus2(th, car(g));
	    array_value(T, i) = cdr(g);
	    z = Aunmodular(array_value(T, i));
	    z = gtimes2(z, modulus);
	    z = gplus2(array_value(U, i), z);
	    array_value(U, i) = z;
	}
	if (!gzerop(th)) {
	    z = gtimes2(th, array_value(U, 0));
	    z = gplus2(array_value(T, 0), z);
	    array_value(T, 0) = z;
	    z = gtimes2(th, array_value(U, 0));
	    z = Aunmodular(z);
	    z = gtimes2(z, modulus);
	    z = gplus2(array_value(U, 0), z);
	    array_value(U, 0) = z;
	}
	e = mkfxat(1);
	for (i = 0; i < n; i++) {
	    e = gtimes2(e, array_value(U, i));
	}
	e = gdiff2(a, e);
	modulus = times2(modulus, p);
	m++;
    }
    /*
     * step 5 
     */
    // lisp_printf("step5: e=%p\n", e);
    if (gzerop(e)) {
#if 0
	gamma = mkfxat(1);
	for (i = 0; i < n; i++) {
	    s = array_value(U, i);
	    delta = Mcont(s);
	    s = gquot2(s, delta);
	    s = gsmodulus(s, modulus);
	    array_value(U, i) = s;
	    gamma = gtimes2(gamma, delta);
	}
	s = array_value(U, 0);
	s = gtimes2(s, gamma);
	// s = gsmodulus(s, modulus);
	array_value(U, 0) = s;
	// lisp_printf("U=%p\n", U);
#endif
	while (Llessp(modulus, B) != nil) {
	    modulus = times2(modulus, p);
	    m++;
	}
    } else {
	for (i = 1; i <= n / 2; i++) {
	    if (a == mkfxat(1))
		break;
	    // lisp_printf("a = %p, U = %p\n", a, U);
	    r = Combine_Factors(a, U, modulus, i, n);
	    // lisp_printf("r = %p\n", r);
	    if (r != nil) {
		s = r;
		while (s != nil) {
		    a = gratio2(a, car(s));
		    s = cdr(s);
		}
	    }
	    T = Lappend(T, r);
	}

	/*
	 * write "U=", U; T := {}; for i:=0 to n-1 do if gremainder(a,
	 * U[i]) = 0 then T := cons(U[i], T); U[i] := 1; fi; od; for i:= 0 
	 * to n-1 do if U[i] = 1 then continue; fi; for j:= i+1 to n-1 do
	 * if U[j] = 1 then continue; fi; S:= (U[i] * U[j]) smod modulus;
	 * if gremainder(a, S) = 0 then T:=cons(S, T); U[i] := 1; U[j] :=
	 * 1; break; fi; od; od; for i:= 0 to n-1 do if U[i] = 1 then
	 * continue; fi; for j:= i+1 to n-1 do if U[i] = 1 then break;
	 * elif U[j] = 1 then continue; fi; for k:=j+1 to n-1 do if U[k] = 
	 * 1 then continue; fi;
	 * 
	 * S:= (U[i] * U[j] * U[k]) smod modulus; if gremainder(a, S) = 0
	 * then T:=cons(S, T); U[i] := 1; U[j] := 1; U[k] := 1; break; fi;
	 * od; od; od; 
	 */
	S = mkfxat(1);
	for (i = 0; i < n; i++) {
	    S = gtimes2(S, array_value(U, i));
	}
	S = gsmodulus(S, modulus);
	if (!gonep(S))
	    T = cons(S, T);
	z = Llength(T);
	U = mkarray(mkint(z));
	Alist2array(U, T);
    }

    z = Llist3(U, modulus, mkfxat(m));
    // z = U;
    popv(30);
#undef a
#undef p
#undef u
#undef w
#undef B
#undef aa
#undef x
#undef alpha
#undef gamma
#undef delta
#undef sigma
#undef tau
#undef modulus
#undef e
#undef g
#undef q
#undef r
#undef s
#undef t
#undef sh
#undef th
#undef c
#undef U
#undef W
#undef S
#undef T
    return z;
}


static          Sex
Combine(Sex F, Sex q, long n, long m)
{
    Sex             s;
    long            i;
    stackcheck();
    push4(F, q, nil, nil);
#define F stack[-3]
#define q stack[-2]
#define U stack[-1]
#define c stack[0]
    U = mkarray(m);
    c = make_symbol("c");
    for (i = 0; i < m; i++) {
	c = gplus2(c, mkfxat(1));
	array_value(U, i) = c;
    }
    s = Combine_Factors(F, U, q, n, m);
#undef F
#undef q
#undef U
#undef c
    popv(4);
    return s;
}

static          Sex
Combine_Factors(Sex F, Sex U, Sex q, long n, long m)
{
    Sex             r;
    long            k,
                    c;
    stackcheck();
    push3(F, U, q);
    push(nil);
#define F stack[-3]
#define U stack[-2]
#define q stack[-1]
#define a stack[0]

    a = mkarray(n);
    k = 0;
    c = 1;
    r = Combine_nest(F, U, q, k, c, n, m, a);
    popv(4);
#undef F
#undef U
#undef q
#undef a
    return r;
}


static          Sex
Combine_nest(Sex F, Sex U, Sex q, long k, long c, long n, long m, Sex a)
{
    long            i,
                    j;

    Sex             u;

    stackcheck();
    push4(F, U, q, a);
    pushv(3);
#define F stack[-6]
#define U stack[-5]
#define q stack[-4]
#define a stack[-3]
#define s stack[-2]
#define r stack[-1]
#define rr stack[-0]
    r = nil;
    for (i = k + 1; i <= m - n + c; i++) {
	if (array_value(U, i - 1) == mkfxat(1))
	    continue;
	array_value(a, c - 1) = mkfxat(i - 1);
	if (n != c) {
	    rr = Combine_nest(F, U, q, i, c + 1, n, m, a);
	    r = Lappend(r, rr);
	} else {
	    s = mkfxat(1);
	    for (j = 0; j < n; j++) {
		u = array_value(a, j);
		u = array_value(U, mkint(u));
		if (u == mkfxat(1)) {
		    s = mkfxat(1);
		    break;
		}
		s = gtimes2(s, u);
		s = gsmodulus(s, q);
	    }
	    if (s == mkfxat(1))
		continue;

	    if (gremaind2(F, s) == mkfxat(0)) {
		r = cons(s, r);
		for (j = 0; j < n; j++) {
		    u = array_value(a, j);
		    array_value(U, mkint(u)) = mkfxat(1);
		}
	    }
	}
    }
    u = r;
#undef F
#undef U
#undef q
#undef a
#undef s
#undef r
#undef rr
    popv(7);
    return u;
}


/**************************************************************/
/**************************************************************/

/*
 * multivariate factorization 
 */

/*
 * Multivariate Hensel Lifting Algorithm 
 */

static          Sex
TaylorCoeff(Sex f, Sex x, long k, Sex a)
{
    long            i;
    stackcheck();
    push3(f, x, a);
#define f stack[-2]
#define x stack[-1]
    // #define k stack[-1]
#define a stack[-0]
    for (i = 0; i < k; i++)
	f = Adiff(f, x);
    f = Asub(f, x, a);
    a = Lfactorial(mkfxat(k));
    f = gratio2(f, a);
#undef f
#undef x
    // #undef k
#undef a
    f = stack[-2];
    popv(3);
    return f;
}

static          Sex
print_e(Sex e, Sex x, Sex v, Sex pl)
{
    long            n;
    stackcheck();
    push4(e, x, v, pl);
    push3(nil, nil, nil);
#define e stack[-6]
#define x stack[-5]
#define v stack[-4]
#define pl stack[-3]
#define q stack[-2]
#define r stack[-1]
#define y stack[-0]
    n = 0;
    y = symbol_to_polynomial(x);
    y = gdiff2(y, v);
    while (!gzerop(e)) {
	q = grdivide2(e, y, x);
	r = cdr(q);
	r = gsmodulus(r, pl);
	e = car(q);
	lisp_printf("(%p)^%d*(%p) + ", y, n, r);
	n++;
    }
    lisp_printf("\n");
#undef e
#undef x
#undef v
#undef pl
#undef q
#undef r
#undef y
    popv(7);
    return nil;
}


static          Sex
mhensel(Sex a, Sex xv, Sex av, Sex p, Sex pl, Sex l, Sex u, Sex lcU)
{
    Sex             z;
    long            i,
                    j,
                    k,
                    m,
                    n,
                    nu,
                    maxdeg;
    long            h;

    ASSERT(is_array(xv) && is_array(av));
    ASSERT(is_array(u));

    stackcheck();
    push5(a, xv, av, p, pl);
    push3(l, u, lcU);
    pushv(16);
#define a stack[-23]
#define xv stack[-22]
#define av stack[-21]
#define p stack[-20]
#define pl stack[-19]
#define l stack[-18]
#define u stack[-17]
#define lcU stack[-16]

#define A stack[-15]
#define U stack[-14]
#define U1 stack[-13]
#define e stack[-12]
#define s stack[-11]
#define dU stack[-10]
#define c stack[-9]
#define monomial stack[-8]
#define t stack[-7]
#define x stack[-6]
#define d stack[-5]
#define alpha stack[-4]
#define r stack[-3]
#define Tc stack[-2]
#define T stack[-1]
#define b stack[-0]
    /*
     * begin local A, U, U1; local i, j, k, m; local n, nu, maxdeg; local
     * e; local s, dU, c, monomial, t, x, d; local alpha; local r, Tc, T,
     * b; 
     */

    n = mkint(car(array_dim(xv)));
    nu = mkint(car(array_dim(u)));
    A = mkarray(n);
    alpha = Alcoeff(a, array_value(xv, 0));
    c = alpha;
    z = gexpt(alpha, mkfxat(nu - 1));
    z = gtimes2(a, z);
    array_value(A, n - 1) = z;

    for (j = n - 1; j >= 1; j--) {
	z = Asub(array_value(A, j), array_value(xv, j),
		 array_value(av, j));
	z = gsmodulus(z, pl);
	array_value(A, j - 1) = z;
	c = Asub(c, array_value(xv, j), array_value(av, j));
    }

#ifdef DEBUG
    lisp_printf("A=%p\n", A);
    lisp_printf("u=%p\n", u);
#endif
    maxdeg = 0;

    for (i = 1; i < n; i++) {
	h = mkint(Adeg(a, array_value(xv, i)));
	if (maxdeg < h)
	    maxdeg = h;
    }
    nu = mkint(car(array_dim(u)));
    lcU = mkarray(nu);
    for (i = 0; i < nu; i++) {
	z = Mpp(array_value(u, i));
	array_value(u, i) = z;
	d = Alcoeff(z, array_value(xv, 0));
	d = Ldivide(c, d);
	if (gzerop(cdr(d))) {
	    d = car(d);
	    array_value(lcU, i) = alpha;
	    z = gtimes2(array_value(u, i), d);
	    array_value(u, i) = z;
	} else {
	    aerror("can't divide alpha by lc(u)");
	}
    }
    U = Lclone(u);

    /*
     * step 2
     */
    for (j = 1; j < n; j++) {
	U1 = Lclone(U);
#ifdef DEBUG
	lisp_printf("before replace u=%p\n", U);
#endif
	monomial = mkfxat(1);
	for (m = 0; m < nu; m++) {
	    c = array_value(lcU, m);
	    if (!gonep(c)) {
		for (i = j + 1; i < n; i++) {
		    c = Asub(c, array_value(xv, i), array_value(av, i));
		}
		z = replace_lc(array_value(U, m), c, array_value(xv, 0));
		array_value(U, m) = z;
	    }
	}
#ifdef DEBUG
	lisp_printf("replaced u=%p\n", U);
#endif
	e = mkfxat(1);
	for (i = 0; i < nu; i++) {
	    e = gtimes2(e, array_value(U, i));
	    e = gsmodulus(e, pl);
	}
	e = gdiff2(array_value(A, j), e);
	e = gsmodulus(e, pl);

	z = Adeg(array_value(A, j), array_value(xv, j));
	h = mkint(z);

#ifdef DEBUG
	lisp_printf("h=%d\n", h);
#endif
	for (k = 1; k <= h; k++) {
#ifdef DEBUG
	    lisp_printf("k=%d, e=%p\n", k, e);
	    print_e(e, array_value(xv, j), array_value(av, j), pl);
#endif
	    if (gzerop(e))
		break;
	    z = symbol_to_polynomial(array_value(xv, j));
	    z = gdiff2(z, array_value(av, j));
	    monomial = gtimes2(monomial, z);
	    c = TaylorCoeff(e, array_value(xv, j), k, array_value(av, j));
	    c = gsmodulus(c, pl);
#ifdef DEBUG
	    lisp_printf("c=%p\n", c);
#endif
	    if (!gzerop(c)) {
		dU = mdiophant(U1, c, xv, av, j, maxdeg, p, pl, l);
#ifdef DEBUG
		lisp_printf("U1 = %p\n", U1);
		lisp_printf("dU = %p\n", dU);
#endif
		e = mkfxat(1);
		for (m = 0; m < nu; m++) {
		    s = grremaind2(array_value(dU, m),
				   array_value(A, j - 1),
				   array_value(av, 0));
		    s = gtimes2(s, monomial);
		    z = gplus2(array_value(U, m), s);
		    z = gsmodulus(z, pl);
		    array_value(U, m) = z;
		    z = gtimes2(e, array_value(U, m));
		    e = gsmodulus(z, pl);
		}
		z = gdiff2(array_value(A, j), e);
		e = gsmodulus(z, pl);
#ifdef DEBUG
		lisp_printf("U=%p\n", U);
		lisp_printf("e=%p\n", e);
#endif
	    }
	}
	if (gzerop(e))
	    continue;
	T = nil;
	Tc = nil;
	b = array_value(A, j);
	m = nu / 2;
	for (i = 1; i <= m; i++) {
	    if (gonep(b))
		break;
	    r = Combine_Multivariate_Factors(array_value(A, j), U, lcU, pl,
					     monomial, array_value(xv, j),
					     i, nu);
	    if (r != nil) {
		T = Lappend(T, car(r));
		Tc = Lappend(Tc, cdr(r));
		r = car(r);
		while (r != nil) {
		    s = car(r);
		    r = cdr(r);
		    b = gquot2(b, s);
		}
	    }
	}
	if (!gonep(b)) {
	    T = cons(b, T);
	    r = mkfxat(1);
	    for (i = 0; i < nu; i++) {
		r = gtimes2(r, array_value(lcU, i));
	    }

	    e = Tc;
	    while (e != nil) {
		s = car(e);
		e = cdr(e);
		r = gquot2(r, s);
	    }
	    Tc = cons(r, Tc);
	}
#ifdef DEBUG
	lisp_printf("T = %p\n", T);
	lisp_printf("Tc = %p\n", Tc);
#endif
	nu = mkint(Llength(T));
	if (nu == 1) {
	    z = mkarray(1);
	    array_value(z, 0) = array_value(A, 0);
	    popv(24);
	    return z;
	}
	U = mkarray(nu);
	lcU = mkarray(nu);
	Alist2array(U, T);
	Alist2array(lcU, Tc);
	e = mkfxat(1);
	for (i = 0; i < nu; i++) {
	    e = gtimes2(e, array_value(U, i));
	}
	e = gdiff2(array_value(A, j), e);
    }

    /*
     * step 3 
     */
    if (zerop(e))
	z = U;
    else
	z = nil;
#undef a
#undef xv
#undef av
#undef p
#undef pl
#undef l
#undef u
#undef lcU

#undef A
#undef U
#undef U1
#undef e
#undef s
#undef dU
#undef c
#undef monomial
#undef t
#undef x
#undef d
#undef alpha
#undef r
#undef Tc
#undef T
#undef b
    popv(24);
    return z;
}


static          Sex
Combine_Multivariate_Factors(Sex F, Sex U, Sex lcU, Sex q, Sex r, Sex y,
			     long n, long m)
{
    // Sex a;
    long            k;

    int            *a,
                   *p;
    Sex             b;
    stackcheck();
    /*
     * begin local a; local k, c; local l; 
     */
    /*
     * debug "C_f:", U, n, m; 
     */

    a = (int *) malloc(n * sizeof(int));
    p = a;
    for (k = 0; k < n; k++)
	*p++ = 0;

    b = Combine_Multivariate_nest(F, U, lcU, q, r, y, 0, 1, n, m, a);
    free(a);
    return b;
}

static          Sex
Combine_Multivariate_nest(Sex F, Sex U, Sex lcU, Sex q, Sex r, Sex y,
			  long k, long c, long n, long m, int *a)
{
    long            d;
    long            i,
                    j;
    stackcheck();
    push5(F, U, lcU, q, r);
    push2(y, nil);
    pushv(4);
#define F stack[-10]
#define U stack[-9]
#define lcU stack[-8]
#define q stack[-7]
#define r stack[-6]
#define y stack[-5]
    // #define a stack[-4]
#define s stack[-3]
#define l stack[-2]
#define lc stack[-1]
#define rr stack[-0]
    /*
     * begin local i, j; local s, l, lc, rr; 
     */
    /*
     * debug "c_n: ", U, k, c, n, m, a; 
     */

    l = nil;
    lc = nil;
    for (i = k + 1; i <= m - n + c; i++) {
	if (gonep(array_value(U, i - 1)))
	    continue;
	// array_value(a, c - 1) = mkfxat(i - 1);
	a[c - 1] = i - 1;
	if (n != c) {
	    rr = Combine_Multivariate_nest(F, U, lcU, q, r, y, i, c + 1, n,
					   m, a);
	    l = Lappend(l, car(rr));
	    lc = Lappend(lc, cdr(rr));
	} else {
	    s = mkfxat(1);
	    for (j = 0; j < n; j++) {
		d = a[j];	// mkint(array_value(a, j));
		if (gonep(array_value(U, d))) {
		    s = mkfxat(1);
		    break;
		}
		s = gtimes2(s, array_value(U, d));
		s = gpremaind2(s, r, y);
		s = gsmodulus(s, q);
	    }
	    if (gonep(s))
		continue;

	    if (gzerop(gremaind2(F, s))) {
		l = cons(s, l);
		s = mkfxat(1);
		for (j = 0; j < n; j++) {
		    d = a[j];	// mkint(array_value(a, j));
		    s = gtimes2(s, array_value(lcU, d));
		    array_value(U, d) = mkfxat(1);
		}
		lc = cons(s, lc);
	    }

	}
    }
#undef F
    F = cons(l, lc);
#undef U
#undef lcU
#undef q
#undef r
#undef y
    // #undef a
#undef s
#undef l
#undef lc
#undef rr
    popv(11);
    return F;
}



/*
 * multivariate polynomial diophantine euqations 
 */

static          Sex
mdiophant(Sex a, Sex c, Sex xv, Sex av, long n, long d, Sex p, Sex pk,
	  Sex k)
{
    long            m;
    long            r;
    long            i,
                    j;
    Sex             z;

    stackcheck();
    push5(a, c, xv, av, nil);
    push4(nil, p, pk, k);
    pushv(16);
#define a stack[-24]
#define c stack[-23]
#define xv stack[-22]
#define av stack[-21]
    // #define n stack[-20]
    // #define d stack[-19]
#define p stack[-18]
#define pk stack[-17]
#define k stack[-16]
    // #define m stack[-15]
    // #define j stack[-14]
    // #define r stack[-13]
#define x stack[-12]
#define v stack[-11]
#define A stack[-10]
#define b stack[-9]
#define anew stack[-8]
#define cnew stack[-7]
#define sigma stack[-6]
#define monomial stack[-5]
#define cm stack[-4]
#define e stack[-3]
#define es stack[-2]
#define ds stack[-1]
    // #define i stack[-0]
    /*
     * begin local m, j, r; local x, v, A, b; local anew, cnew, sigma,
     * monomial, cm; local e, es, ds; local i; 
     */

    /*
     * step 1 
     */
    r = mkint(car(array_dim(a)));
    b = mkarray(r);

    x = array_value(xv, n - 1);
    v = array_value(av, n - 1);

    if (n > 1) {
	/*
	 * step 2.1 
	 */
	A = mkfxat(1);
	for (i = 0; i < r; i++) {
	    A = gtimes2(A, array_value(a, i));
	}
	for (j = 0; j < r; j++) {
	    z = gquot2(A, array_value(a, j));
	    array_value(b, j) = z;
	}

	anew = mkarray(r);
	for (j = 0; j < r; j++) {
	    z = Asub(array_value(a, j), x, v);
	    array_value(anew, j) = z;
	}
	cnew = Asub(c, x, v);
	sigma = mdiophant(anew, cnew, xv, av, n - 1, d, p, pk, k);
	e = mkfxat(0);
	for (j = 0; j < r; j++) {
	    z = gtimes2(array_value(sigma, j), array_value(b, j));
	    e = gplus2(e, z);
	}
	e = gdiff2(c, e);
	e = gmodulus(e, pk);
	monomial = mkfxat(1);
	for (m = 1; m <= d; m++) {
	    if (zerop(e))
		break;
	    z = symbol_to_polynomial(x);
	    z = gdiff2(z, v);
	    monomial = gtimes2(monomial, z);
	    cm = e;
	    for (j = 1; j <= m; j++)
		cm = Adiff(cm, x);
	    cm = Asub(cm, x, v);
	    z = Lfactorial(mkfxat(m));
	    cm = gratio2(cm, z);
	    if (!gzerop(cm)) {
		ds = mdiophant(anew, cm, xv, av, n - 1, d, p, pk, k);
		es = mkfxat(0);
		for (j = 0; j < r; j++) {
		    z = gtimes2(monomial, array_value(ds, j));
		    z = gplus2(array_value(sigma, j), z);
		    array_value(sigma, j) = z;
		    z = gtimes2(array_value(ds, j), monomial);
		    z = gtimes2(z, array_value(b, j));
		    es = gplus2(es, z);
		}
		e = gdiff2(e, es);
		e = gmodulus(e, pk);
	    }
	}

    } else {
	/*
	 * step 2.2 Univariate case 
	 */

	sigma = mkarray(r);
	cm = Acoeff(c, x);
	n = mkint(Adeg(c, x));
	for (m = 0; m <= n; m++) {
	    if (gzerop(array_value(cm, m)))
		continue;
	    z = symbol_to_polynomial(x);
	    z = gexpt(z, mkfxat(m));
	    ds = udiophant(a, x, z, p, pk, k);
#ifdef DEBUG
	    // lisp_printf("udiophant retusn %p\n", ds);
#endif
	    // debug "udiophant returns ", ds;
	    for (j = 0; j < r; j++) {
		z = gtimes2(array_value(ds, j), array_value(cm, m));
		z = gplus2(array_value(sigma, j), z);
		array_value(sigma, j) = z;
	    }
#ifdef DEBUG
	    // lisp_printf("sigma = %p\n", sigma);
#endif
	    // debug "sigma = ", sigma;
	}
	/*
	 * if c <> 0 then sigma := udiophant(a, x, c, p, pk, k); for j:=0
	 * to r-1 do sigma[j] := sigma[j] * c; od; else array sigma[r];
	 * fi; 
	 */
    }

    for (j = 0; j < r; j++) {
	z = gmodulus(array_value(sigma, j), pk);
	array_value(sigma, j) = z;
    }
    z = sigma;
    popv(25);
    return z;
#undef a
#undef c
#undef xv
#undef av
#undef n
    // #undef d
#undef p
#undef pk
#undef k
    // #define m 
    // #define j 
    // #define r 
#undef x
#undef v
#undef A
#undef b
#undef anew
#undef cnew
#undef sigma
#undef monomial
#undef cm
#undef e
#undef es
#undef ds
    // #define i 
}



static          Sex
udiophant(Sex a, Sex x, Sex c, Sex p, Sex pk, Sex k)
{
    long            r;
    long            j;
    Sex             z;

    stackcheck();
    push5(a, x, c, p, pk);
    push(k);
    pushv(8);
#define a stack[-13]
#define x stack[-12]
#define c stack[-11]
#define p stack[-10]
#define pk stack[-9]
#define k stack[-8]
    // #define r stack[-7]
#define q stack[-6]
#define s stack[-5]
    // #define j stack[-4]
#define s1 stack[-3]
#define s2 stack[-2]
#define result stack[-1]
#define y stack[-0]
    /*
     * ; begin local r, q, s, j, s1, s2, result; local y; 
     */

    r = mkint(car(array_dim(a)));
    result = mkarray(r);
    // y:=symbol_to_polynomial(x);

    if (r > 2) {
	s = MultiTermEEAlift(a, x, p, pk, k);
	for (j = 0; j < r; j++) {
	    /*
	     * s1:=mkmodular(y^m*s[j], pk); 
	     */
	    s1 = gtimes2(array_value(s, j), c);
	    s1 = gmodulus(s1, pk);
	    s1 = grremaind2(s1, array_value(a, j), x);
	    array_value(result, j) = s1;
	}
    } else {
	s = EEAlift(array_value(a, 1), array_value(a, 0), x, p, k);
#ifdef DEBUG
	// lisp_printf("EEAlift = %p\n", s);
#endif
	s1 = car(s);
	s2 = cadr(s);
	/*
	 * s1:=mkmodular(y^m*s1, pk); result[0] :=s1; q:=rquotient(s1,
	 * a[0], x); s2:=mkmodular(y^m*s2, pk); s2:=s2+q*a[1]; result[1]
	 * := s2; 
	 */
	s1 = gtimes2(s1, c);
	s1 = gmodulus(s1, pk);
	q = grquotient2(s1, array_value(a, 0), x);
	q = gmodulus(q, pk);
	z = gtimes2(q, array_value(a, 0));
	s1 = gdiff2(s1, z);
	s1 = gmodulus(s1, pk);
	z = gtimes2(q, array_value(a, 1));
	s2 = gtimes2(s2, c);
	z = gplus2(s2, z);
	s2 = gmodulus(z, pk);
	array_value(result, 0) = s1;
	array_value(result, 1) = s2;
    }
    z = result;
#undef a
#undef x
#undef c
#undef p
#undef pk
#undef k
    // #undef r 
#undef q
#undef s
    // #undef j 
#undef s1
#undef s2
#undef result
#undef y
    popv(14);
    return z;
}


static          Sex
MultiTermEEAlift(Sex a, Sex x, Sex p, Sex pk, Sex k)
{
    long            r,
                    j;
    Sex             z;
    stackcheck();
    push5(a, x, p, pk, k);
    pushv(8);
#define a stack[-12]
#define x stack[-11]
#define p stack[-10]
#define pk stack[-9]
#define k stack[-8]
#define s stack[-7]
#define q stack[-6]
#define b stack[-5]
#define anew stack[-4]
#define xv stack[-3]
#define av stack[-2]
#define sigma stack[-1]
    /*
     * begin local s, q, b; local r, j; local anew, xv, av, sigma; 
     */

    r = mkint(car(array_dim(a)));
    q = mkarray(r - 1);
    s = mkarray(r);
    anew = mkarray(2);
    xv = mkarray(1);
    av = mkarray(1);
    array_value(xv, 0) = x;
    array_value(av, 0) = mkfxat(0);

    array_value(q, r - 2) = array_value(a, r - 1);
    for (j = r - 3; j >= 0; j--) {
	z = gtimes2(array_value(a, j + 1), array_value(q, j + 1));
	array_value(q, j) = z;
    }

    b = mkfxat(1);
    for (j = 1; j < r; j++) {
	array_value(anew, 0) = array_value(q, j - 1);
	array_value(anew, 1) = array_value(a, j - 1);
	sigma = mdiophant(anew, b, xv, av, 1, 0, p, pk, k);

	b = array_value(sigma, 0);
	array_value(s, j - 1) = array_value(sigma, 1);
    }
    array_value(s, r - 1) = b;
    z = s;
    popv(13);
#undef a
#undef x
#undef p
#undef pk
#undef k
#undef s
#undef q
#undef b
#undef anew
#undef xv
#undef av
#undef sigma
    return z;
}

static          Sex
EEAlift(Sex a, Sex b, Sex x, Sex p, Sex k)
{
    long            j;
    long            l;
    Sex             z;

    l = mkint(k);
    stackcheck();
    push5(a, b, x, p, k);
    pushv(12);
#define a stack[-16]
#define b stack[-15]
#define x stack[-14]
#define p stack[-13]
#define k stack[-12]
#define ap stack[-11]
#define bp stack[-10]
#define t stack[-9]
#define s stack[-8]
#define tp stack[-7]
#define sp stack[-6]
#define modulus stack[-5]
#define e stack[-4]
#define c stack[-3]
#define q stack[-2]
#define sigma stack[-1]
#define tau stack[-0]
    /*
     * ; begin local ap, bp, t, s, tp, sp; local modulus, e, c, q, sigma,
     * tau; local j; 
     */

    ap = Amkmodular(a, p);
    bp = Amkmodular(b, p);
    // lisp_printf("call EEA with %p, %p, %p\n", ap, bp, x);
    s = EEA(ap, bp, x);
    // lisp_printf("EEA=%p\n", s);
    tp = caddr(s);
    sp = cadr(s);
    s = Aunmodular(sp);
    t = Aunmodular(tp);
    modulus = p;
    for (j = 1; j < l; j++) {
	z = gtimes2(t, b);
	e = gtimes2(s, a);
	z = gdiff2(mkfxat(1), z);
	e = gdiff2(z, e);
	c = gquot2(e, modulus);
	c = Amkmodular(c, p);
	sigma = gtimes2(sp, c);
	tau = gtimes2(tp, c);
	q = grdivide2(sigma, bp, x);
	sigma = cdr(q);
	q = car(q);
	z = gtimes2(q, ap);
	tau = gplus2(tau, z);
	z = Aunmodular(sigma);
	z = gtimes2(z, modulus);
	s = gplus2(s, z);
	z = Aunmodular(tau);
	z = gtimes2(z, modulus);
	t = gplus2(t, z);
	modulus = gtimes2(modulus, p);
    }

    z = Llist2(s, t);
    popv(17);
    return z;
#undef a
#undef b
#undef x
#undef p
#undef k
#undef ap
#undef bp
#undef t
#undef s
#undef tp
#undef sp
#undef modulus
#undef e
#undef c
#undef q
#undef sigma
#undef tau
}




/*
 * univariate polynomial factorization 
 */


static          Sex
factor1(Sex F)
{
    long            n;
    Sex             s;
    stackcheck();
    if (!is_poly(F)) {
#if 0
	s = Llist2(F, mkfxat(1));
	s = Lncons(s);
	return s;
#endif
	return zfactor(F);
    }
    push(F);
    pushv(13);
#define F stack[-13]
#define H stack[-12]
    // #define n stack[-11]
#define x stack[-10]
#define v stack[-9]
#define f stack[-8]
#define g stack[-7]
#define h stack[-6]
#define p stack[-5]
#define m stack[-4]
#define b stack[-3]
    // #define Hs stack[-2]
#define j stack[-1]
#define I stack[-0]
    /*
     * begin local H, n, x, v; local f, g, h, p, m; local b; local i, j;
     * local I; 
     */

    p = Mcont(F);
    F = gquot2(F, p);
    x = poly_var(F);
    H = SquareFree(F);
#ifdef DEBUG
    lisp_printf("factor1: sq H = %p\n", H);
#endif
    I = nil;
    if (!gonep(p)) {
	// write "p=", p;
	if (!is_poly(p)) {
#if 1
	    I = zfactor(p);
#else
	    I = Llist2(p, mkfxat(1));
	    I = Lncons(I);
#endif
	} else {
	    I = factor1(p);
	}
    }
    while (H != nil) {
	j = car(H);
	H = cdr(H);
	f = car(j);
	// write "f=", f;
	v = Aunknowns(f);
	n = mkint(Llength(v));
	switch (n) {
	case 0:
	    break;
	case 1:
	    j = univariate_factor(j);
	    break;
	default:
	    j = multivariate_factor(j, v, n);
	    break;
	}
	I = Lappend(j, I);
    }
    s = I;
    popv(14);
    return s;
#undef F
#undef H
    // #undef n
#undef x
#undef v
#undef f
#undef g
#undef h
#undef p
#undef m
#undef b
    // #undef i 
#undef j
#undef I
}


static          Sex
multivariate_factor(Sex j, Sex v, long n)
{
    long            k;
    long            i;
    // long nu;
    Sex             s;
    stackcheck();
    push3(j, v, nil);
    pushv(20);
#define j stack[-22]
#define v stack[-21]
    // #define n stack[-20]
    // #define i stack[-19]
#define I stack[-18]
#define f stack[-17]
#define g stack[-16]
#define m stack[-15]
#define p stack[-14]
#define h stack[-13]
    // #define b stack[-12]
#define x stack[-11]
#define a stack[-10]
#define b stack[-9]
    // #define k stack[-8]
#define q stack[-7]
#define c stack[-6]
#define B stack[-5]
#define lc stack[-4]
    // #define nu stack[-3]
#define pl stack[-2]
#define l stack[-1]
#define alpha stack[-0]
    /*
     * ; begin local i, I, f, g, m, p; local h, b, x, a, b; local k, q, c, 
     * B, lc, nu; local pl, l, alpha; 
     */

    f = car(j);
    if (!is_poly(f) || mkint(Mdeg(f)) < 2) {
	s = Lncons(j);
	popv(23);
	return s;
    }
    a = mkarray(n);
    b = mkarray(n);
    Alist2array(a, v);
    x = array_value(a, 0);
    m = Mlcoeff(f);
    g = f;
    for (k = 1; k < n; k++) {
	p = mkfxat(2);
	while (1) {
	    q = Lrandom(p);
	    s = quot2(p, mkfxat(2));
	    q = diff2(q, s);
	    s = array_value(a, k);
	    h = Asub(g, s, q);
	    s = array_value(a, k);
	    c = Asub(m, s, q);
	    p = plus2(p, mkfxat(2));
	    if (gzerop(c))
		continue;
	    s = Adiff(h, x);
	    s = my_gcd_mgcd(h, s);
	    if (!is_poly(s))
		break;
	}
	array_value(b, k) = q;
	m = c;
	g = h;
    }

    p = Lrandom(mkfxat(100));
    while (1) {
	p = Lnext_prime(p);
	m = Mlcoeff(f);
	c = gmodulus(m, p);
	if (gzerop(c))
	    continue;
	c = Mlcoeff(g);
	c = gmodulus(c, p);
	if (gzerop(c))
	    continue;
	h = Amkmodular(g, p);
	s = Adiff(h, x);
	s = Mgcd(h, s);
	if (!is_poly(s))
	    break;
    }

    h = berlekamp(g, p);
    // nu = mkint(Llength(h));
    if (Llength(h) == mkfxat(1)) {
	I = Lncons(j);
    } else {
	I = nil;
	B = MignotteBound(g);
	h = uhensel(g, p, h, nil, B);
	pl = cadr(h);
	l = caddr(h);
	h = car(h);
	m = Asize(h);
	if (car(m) == mkfxat(1)) {
	    I = Lncons(j);
	} else {
	    // h:=array2list(h);
	    /*
	     * alpha:=mlcoeff(f); m:=length(h); array lc[m]; 
	     */
	    lc = nil;
	    /*
	     * lc[0]:=alpha; for i:=1 to m-1 do lc[i] := alpha;
	     * f:=f*alpha; od; 
	     */
	    h = mhensel(f, a, b, p, pl, l, h, lc);

	    n = mkint(car(array_dim(h)));
	    for (i = 0; i < n; i++) {
		m = array_value(h, i);
		m = Mpp(m);
		s = Llist2(m, cadr(j));
		I = cons(s, I);
	    }
	}
    }


    s = I;
    popv(23);
    return s;
#undef j
#undef v
    // #define n stack[-20]
    // #undef i 
#undef I
#undef f
#undef g
#undef m
#undef p
#undef h
    // #define b 
#undef x
#undef a
#undef b
    // #define k stack[-8]
#undef q
#undef c
#undef B
#undef lc
    // #undef nu
#undef pl
#undef l
#undef alpha
}



static          Sex
univariate_factor(Sex j)
{
    long            d;
    long            i;
    long            n;
    Sex             s;
    stackcheck();
    push(j);
    pushv(10);
#define j stack[-10]
#define I stack[-8]
#define f stack[-7]
#define g stack[-6]
#define m stack[-5]
#define p stack[-4]
#define h stack[-3]
#define b stack[-2]
#define x stack[-1]
    /*
     * begin local i, I, f, g, m, p; local h, b, x, n; 
     */

    f = car(j);
    if (!is_poly(f) || mkint(Mdeg(f)) < 2) {
	s = Lncons(j);
	popv(11);
	return s;
    }
    p = Lrandom(mkfxat(100));
    x = poly_var(f);
    while (1) {
	p = Lnext_prime(p);
	m = Mlcoeff(f);
	if (zerop(remaind2(m, p)))
	    continue;
	g = Amkmodular(f, p);
	s = Adiff(g, x);
	s = Mgcd(g, s);
	if (!is_poly(s))
	    break;
    }
    g = berlekamp(f, p);
    d = mkint(Llength(g));
    if (d == 1) {
	s = Llist2(car(g), cadr(j));
	I = Lncons(s);
    } else {
	I = nil;
	b = MignotteBound(f);
	h = uhensel(f, p, g, nil, b);
	h = car(h);
	n = mkint(car(array_dim(h)));
	for (i = 0; i < n; i++) {
	    f = array_value(h, i);
	    f = Mpp(f);
	    s = Llist2(f, cadr(j));
	    I = cons(s, I);
	}
    }
    s = I;
#undef j
#undef I
#undef f
#undef g
#undef m
#undef p
#undef h
#undef b
#undef x
    popv(11);
    return s;
}
/**************************************************************/
/**************************************************************/

setup_type const factor_setup[] = {

    {"factor", (lisp_proc *) Afactor, 1},
    {"zfactor", (lisp_proc *) Azfactor, 1},
    {"berlekamp", (lisp_proc *) berlekamp, 2},
    {"squarefree", (lisp_proc *) SquareFree, 1},
    {"FormMatrix", (lisp_proc *) FormMatrix, 2},
    {"factorlist2prefix", (lisp_proc *) factorlist2prefix, 1},
    {"isqrt", (lisp_proc *) isqrt, 1},
    {"iroot", (lisp_proc *) iroot, 2},
    {NULL, 0, 0}
};
