/*
 * fmodular.c      (C) Copyright 2012 by Fujio Kako, Nara Women's University.
 * short int and modular arithmetics
 *      $Header: /Users/kako/asuka/src/RCS/fmodular.c,v 2.1 2012/10/21 13:32:48 kako Exp $
 */

static char const rcsid[] =
    "$Id: fmodular.c,v 2.1 2012/10/21 13:32:48 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"
#include "mathop.h"
#include "compare.h"
#include "float.h"

/***************************************************************/
/***************************************************************/

#if _LP64
static long     fmod_modulus = 288230376151711813L;
#else
static long     fmod_modulus = 67108879L;
#endif

void            print_fmodular(Sex u, putchar_fn * fn);
Sex             Lset_fmodulus(Sex a);
Sex             make_fmodular(double d, long n);
Sex             Lfmodular(Sex a);

long            fmodular(Sex a);
long            fmodulari(Sex a);
long            fmodularb(Sex a);
long            fmodularr(Sex a);

extern long     recipim(long n, long m);
extern long     mult2mm(long a, long b, long m);

/***************************************************************/
/***************************************************************/

void
print_fmodular(Sex u, putchar_fn * fn)
{
    double          d;
    long            x;
    d = fmodular_dvalue(u);
    x = fmodular_ivalue(u);
    fn('#');
    fn('f');
    fn('[');
    print_float(d, fn);
    fn(',');
    print_int(x, fn);
    fn(']');
}


Sex
Lset_fmodulus(Sex a)
{
    long            n,
                    m;
    if (!is_fixnum(a))
	error1("not a int", a);
    n = mkint(a);
    if (n <= 0)
	error1("negative interger", a);
    m = fmod_modulus;
    fmod_modulus = n;
    return mkfxat(m);
}

Sex
Lfmodular(Sex a)
{
    long            n;
    double          d;

    push(a);
    d = floatit(a);
    pop(a);
    a = Lmodular2(a, fmod_modulus);
    n = modular_value(a);
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = d;
    return a;
}


long
fmodulari(Sex a)
{
    long            n;
    n = mkint(a) % fmod_modulus;
    if (n < 0)
	n = n + fmod_modulus;
    return n;
}

long
fmodularb(Sex a)
{
    long            n;
    a = remaind2(a, mkfxat(fmod_modulus));
    n = mkint(a);
    if (n < 0)
	n = n + fmod_modulus;
    return n;
}

long
fmodularr(Sex a)
{
    long            n,
                    k;
    push(a);
    a = rational_num(a);
    a = remaind2(a, mkfxat(fmod_modulus));
    n = mkint(a);
    if (n < 0)
	n = n + fmod_modulus;
    pop(a);
    a = rational_den(a);
    a = remaind2(a, mkfxat(fmod_modulus));
    k = mkint(a);
    k = recipim(k, fmod_modulus);
    k = mult2mm(k, n, fmod_modulus);
    return k;
}

long
fmodular(Sex a)
{
    long            n,
                    k;
    long            m = fmod_modulus;

    if (is_fixnum(a)) {
	n = mkint(a) % m;
	if (n < 0)
	    n = n + m;
	return n;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    a = remaind2(a, mkfxat(m));
	    n = mkint(a);
	    if (n < 0)
		n = n + m;
	    return n;
	case N_RATIONAL:
	    push(a);
	    a = rational_num(a);
	    a = remaind2(a, mkfxat(m));
	    n = mkint(a);
	    if (n < 0)
		n = n + m;
	    pop(a);
	    a = rational_den(a);
	    a = remaind2(a, mkfxat(m));
	    k = mkint(a);
	    k = recipim(k, m);
	    k = mult2mm(k, n, m);
	    return k;
	}
    }
    return error1("bad argument to fmodular ", a);
}


Sex
make_fmodular(double d, long n)
{
    Sex             a;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = d;
    return a;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
plus2iY(Sex a, Sex b)
{
    long            n = mkint(a);
    long            m = fmodular_ivalue(b);
    double          wb = fmodular_dvalue(b);
    wb = wb + n;
    n = n + m;
    n = n % fmod_modulus;
    if (n < 0)
	n = n + fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wb;
    return a;
}

Sex
plus2Yi(Sex b, Sex a)
{
    return plus2iY(a, b);
}


Sex
plus2YY(Sex a, Sex b)
{
    long            n = fmodular_ivalue(a) + fmodular_ivalue(b);
    double          wc = fmodular_dvalue(a) + fmodular_dvalue(b);
    if (n > fmod_modulus)
	n = n - fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wc;
    return a;
}


Sex
plus2Yb(Sex a, Sex b)
{
    return plus2bY(b, a);
}
Sex
plus2Yr(Sex a, Sex b)
{
    return plus2rY(b, a);
}

Sex
plus2bY(Sex a, Sex b)
{
    long            n = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);
    d = d + floatitb(a);
    n = n + fmodularb(a);
    if (n > fmod_modulus)
	n = n - fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = d;
    return a;
}


Sex
plus2rY(Sex a, Sex b)
{
    long            m = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);

    d = d + floatitr(a);
    m = m + fmodularr(a);
    if (m > fmod_modulus)
	m = m - fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = d;
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
diff2iY(Sex a, Sex b)
{
    long            n = mkint(a);
    long            m = fmodular_ivalue(b);
    double          wb = fmodular_dvalue(b);
    wb = n - wb;
    n = n - m;
    n = n % fmod_modulus;
    if (n < 0)
	n = n + fmod_modulus;
    a = new_fmodular();

    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wb;

    return a;
}

Sex
diff2Yi(Sex b, Sex a)
{
    long            n = mkint(a);
    long            m = fmodular_ivalue(b);
    double          wb = fmodular_dvalue(b);
    wb = wb - n;
    n = m - n;
    n = n % fmod_modulus;
    if (n < fmod_modulus)
	n = n + fmod_modulus;
    a = new_fmodular();

    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wb;

    return a;
}


Sex
diff2YY(Sex a, Sex b)
{
    long            n = fmodular_ivalue(a) - fmodular_ivalue(b);
    double          wc = fmodular_dvalue(a) - fmodular_dvalue(b);
    if (n < 0)
	n = n + fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wc;
    return a;
}

Sex
diff2Yb(Sex b, Sex a)
{
    long            n = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);
    d = d - floatitb(a);
    a = remaind2(a, mkfxat(fmod_modulus));
    n = n - mkint(a);
    if (n < 0)
	n += fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = d;
    return a;
}


Sex
diff2Yr(Sex b, Sex a)
{
    long            m = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);
    double          w;
    long            n;

    n = fmodularr(a);
    w = floatitr(a);
    d = d - w;
    m = n - m;
    if (m < 0)
	m = m + fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = d;
    return a;
}



Sex
diff2bY(Sex a, Sex b)
{
    long            n = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);
    d = -d + floatitb(a);
    a = remaind2(a, mkfxat(fmod_modulus));
    n = -n + mkint(a);
    if (n < 0)
	n = n + fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = d;
    return a;
}


Sex
diff2rY(Sex a, Sex b)
{
    long            m = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);

    m = fmodularr(a) - m;
    d = floatitr(a) - d;
    if (m < 0)
	m = m + fmod_modulus;
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = d;
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
times2Yi(Sex b, Sex a)
{
    long            n = mkint(a);
    long            m = fmodular_ivalue(b);
    double          wa = fmodular_dvalue(b);
    wa = wa * n;
    n = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wa;
    return a;
}

Sex
times2iY(Sex a, Sex b)
{
    return times2Yi(b, a);
}

Sex
times2bY(Sex a, Sex b)
{
    long            n = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);
    long            m;

    d = floatitb(a) * d;

    m = fmodularb(a);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = d;
    return a;
}


Sex
times2rY(Sex a, Sex b)
{
    long            n = fmodular_ivalue(b);
    double          d = fmodular_dvalue(b);
    long            m;

    d = floatitr(a) * d;
    m = fmodularr(a);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = d;
    return a;
}



Sex
times2YY(Sex a, Sex b)
{
    double          wc = fmodular_dvalue(a) * fmodular_dvalue(b);
    long            n = fmodular_ivalue(a);
    long            m = fmodular_ivalue(b);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_dvalue(a) = wc;
    fmodular_ivalue(a) = m;
    return a;
}


Sex
times2Yb(Sex a, Sex b)
{
    return times2bY(b, a);
}

Sex
times2Yr(Sex a, Sex b)
{
    return times2rY(b, a);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
quot2iY(Sex a, Sex b)
{
    double          wa = mkint(a);
    double          wb = fmodular_dvalue(b);
    long            n = fmodular_ivalue(b);
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    n = recipim(n, fmod_modulus);
    n = mult2mm(mkint(a), n, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = n;
    fmodular_dvalue(a) = wa / wb;
    return a;
}


Sex
quot2Yi(Sex a, Sex b)
{
    double          wa = fmodular_dvalue(a);
    long            m = fmodular_ivalue(a);
    long            n = mkint(b);

    if (n == 0)
	return aerror2("bad arg for quotient", a, b);
    wa = wa / n;
    n = recipim(n, fmod_modulus);
    m = mult2mm(n, m, fmod_modulus);

    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = wa;
    return a;
}


Sex
quot2YY(Sex a, Sex b)
{
    double          wa = fmodular_dvalue(a);
    double          wb = fmodular_dvalue(b);
    long            n,
                    m;
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    n = fmodular_ivalue(a);
    m = fmodular_ivalue(b);
    m = recipim(m, fmod_modulus);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = wa / wb;
    return a;
}

Sex
quot2Yb(Sex a, Sex b)
{
    double          wa = fmodular_dvalue(a);
    double          wb;
    long            n,
                    m;
    wb = floatitb(b);
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    n = fmodular_ivalue(a);
    m = fmodularb(b);

    m = recipim(m, fmod_modulus);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = wa / wb;
    return a;
}

Sex
quot2Yr(Sex a, Sex b)
{
    double          wa = fmodular_dvalue(a);
    double          wb;
    long            n,
                    m;
    wb = floatitr(b);
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    n = fmodular_ivalue(a);
    m = fmodularr(b);
    m = recipim(m, fmod_modulus);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = wa / wb;
    return a;
}

Sex
quot2bY(Sex a, Sex b)
{
    double          wa;
    double          wb = fmodular_dvalue(b);
    long            n,
                    m;
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    wa = floatitb(a);
    n = fmodularb(a);
    m = fmodular_ivalue(b);
    m = recipim(m, fmod_modulus);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = wa / wb;
    return a;
}

Sex
quot2rY(Sex a, Sex b)
{
    double          wa;
    double          wb = fmodular_dvalue(b);
    long            n,
                    m;

    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    wa = floatitr(a);
    n = fmodularr(a);
    m = fmodular_ivalue(b);
    m = recipim(m, fmod_modulus);
    m = mult2mm(n, m, fmod_modulus);
    a = new_fmodular();
    fmodular_ivalue(a) = m;
    fmodular_dvalue(a) = wa / wb;
    return a;
}

Sex
negateY(Sex a)
{
    double          d = fmodular_dvalue(a);
    long            n = fmodular_ivalue(a);

    a = new_fmodular();
    fmodular_ivalue(a) = fmod_modulus - n;
    fmodular_dvalue(a) = -d;
    return a;
}
/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
lessp2iY(Sex a, Sex b)
{
    return ((double) mkint(a) < fmodular_dvalue(b));
}



bool
lessp2bY(Sex a, Sex b)
{
    double          x,
                    y;

    y = fmodular_dvalue(b);
    x = floatitb(a);
    return (x < y);
}

bool
lessp2rY(Sex a, Sex b)
{
    return (floatitr(a) < fmodular_dvalue(b));
}


bool
lessp2Yi(Sex a, Sex b)
{
    return (fmodular_dvalue(a) < (double) mkint(b));
}

bool
lessp2Yb(Sex a, Sex b)
{
    double          x,
                    y;

    x = fmodular_dvalue(a);
    y = floatitb(b);
    return (x < y);
}


bool
lessp2Yr(Sex a, Sex b)
{
    return (fmodular_dvalue(a) < floatitr(b));
}

bool
lessp2YY(Sex a, Sex b)
{
    return (fmodular_dvalue(a) < fmodular_dvalue(b));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
numeq2Yi(Sex b, Sex a)
{
    return ((double) mkint(a) == fmodular_dvalue(b));
}

bool
numeq2Yb(Sex b, Sex a)
{
    return (floatitb(a) == fmodular_dvalue(b));
}

bool
numeq2Yr(Sex b, Sex a)
{
    return (floatitr(a) == fmodular_dvalue(b));
}




bool
numeq2iY(Sex a, Sex b)
{
    return ((double) mkint(a) == fmodular_dvalue(b));
}

bool
numeq2bY(Sex a, Sex b)
{
    return (floatitb(a) == fmodular_dvalue(b));
}

bool
numeq2rY(Sex a, Sex b)
{
    return (floatitr(a) == fmodular_dvalue(b));
}



bool
numeq2YY(Sex a, Sex b)
{
    return (fmodular_dvalue(a) == fmodular_dvalue(b));
}

/***************************************************************/
/***************************************************************/
