/*
 * fns.c                (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 *  Basic functions.
 * $Header: /Users/kako/asuka/src/RCS/fns.c,v 2.29 2016/09/06 05:50:02 kako Exp $
 */

static char const rcsid[] =
    "$Id: fns.c,v 2.29 2016/09/06 05:50:02 kako Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "kctype.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"

#define INCLUDE_FEND

#include "fns.h"

#ifdef ASUKA
extern Sex      Lsetorder(Sex a);
#endif

#ifdef DEBUG

#define show_usage(r, n) \
{ \
	printf("cell use 0x%08lx -- 0x%08lx\n", (long)r, (long)(r + n));\
}
#else
#define show_usage(r, n)
#endif


extern long     number_of_gc_called;

extern Sex      Adebug(Sex a);
extern Sex      Aenable(Sex a);
extern Sex      Adisable(Sex a);

/*****************************************************************************/
/*
 * Some basic functions                                                 
 */
/****************************************************************************
 */



Sex
Lcar(A)
     Sex             A;
{
    return car(A);
}

Sex
Lcdr(A)
     Sex             A;
{
    return cdr(A);
}

Sex
Lrplaca(A, Q)
     Sex             A,
                     Q;
{
    car(A) = Q;
    return A;
}

Sex
Lrplacd(A, Q)
     Sex             A,
                     Q;
{
    cdr(A) = Q;
    return A;
}



Sex
Latom(Sex p)
{
    if (is_number(p))
	return true;
    else if (is_symbol(p))
	return true;
    else if (is_string(p))
	return true;
    return nil;
    // return Bool_to_Sex(!is_cons(p));
}

Sex
Leq(Sex a, Sex b)
{
    return Bool_to_Sex(a == b);
}

Sex
Lpnth(Sex u, Sex n)
{
    long            i;
    if (u == nil || !is_fixnum(n))
	error2("nth: invalid argument", u, n);
    i = mkint(n);
    while (i-- > 1)
	u = cdr(u);
    return u;
}

Sex
Lnth(Sex u, Sex n)
{
    long            i;
    if (u == nil || !is_fixnum(n))
	error2("nth: invalid argument", u, n);
    i = mkint(n);
    while (i-- > 1)
	u = cdr(u);
    return car(u);
}

Sex
Llistp(Sex p)
{
    while (is_cons(p))
	p = cdr(p);
    if (p == nil)
	return true;
    return nil;
}

Sex
Lnumberp(Sex p)
{
    return Bool_to_Sex(is_number(p));
}



Sex
Lfloatp(Sex p)
{
    return is_float(p) ? true : nil;
}


Sex
Lget(Sex a, Sex b)
{
    Sex             pl,
                    w;
    if (!is_symhdr(a))
	return nil;
    pl = plist(a);
    while (pl != nil) {
	w = car(pl);
	if (car(w) == b)
	    return cdr(w);
	pl = cdr(pl);
    }
    return nil;
}

Sex
Lput(Sex a, Sex b, Sex c)
{
    Sex             pl;
    if (!is_symhdr(a))
	return c;
    pl = plist(a);
    while (pl != nil) {
	Sex             w = car(pl);
	if (car(w) == b) {
	    cdr(w) = c;
	    return c;
	} else
	    pl = cdr(pl);
    }
    stackcheck();
    push2(a, c);
    b = Lacons(b, c, plist(a));
    pop2(c, a);
    plist(a) = b;
    return c;
}

Sex
Lremprop(Sex a, Sex b)
{
    Sex             pl,
                    prevp;
    if (!is_symhdr(a))
	return nil;
    prevp = nil;
    pl = plist(a);
    while (pl != nil) {
	Sex             w = car(pl);
	if (car(w) == b) {
	    pl = cdr(pl);
	    if (prevp == nil)
		plist(a) = pl;
	    else
		cdr(prevp) = pl;
	    return cdr(w);
	}
	prevp = pl;
	pl = cdr(prevp);
    }
    return nil;
}


/*****************************************************************************/
/*
 * Storage allocation.                                                  
 */
/*****************************************************************************/


Sex
cons(Sex a, Sex b)
{
    Cons_Cell      *r;
    Sex             s;
    r = cellpointer - 1;
    show_usage(r, 1);
    car(r) = a;
    cdr(r) = b;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	s = reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
#ifdef DEBUG
	if (car(s) & 1) {
	    printf("cons: s = 0x%08lx = (0x%08lx . 0x%08lx)\n", s, car(s),
		   cdr(s));
	    printf("cons: a = 0x%08lx , b = 0x%08lx\n", a, b);
	}
#endif
	return s;
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
rcons(Sex a, Sex b)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    show_usage(r, 1);
    cdr(a) = (Sex) r;
    car(r) = b;
    cdr(r) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Lxcons(Sex b, Sex a)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    show_usage(r, 1);
    car(r) = a;
    cdr(r) = b;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Lncons(Sex a)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    show_usage(r, 1);
    car(r) = a;
    cdr(r) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Llist2(Sex a, Sex b)
{
    Cons_Cell      *r;
    r = cellpointer - 2;
    show_usage(r, 2);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Llist2star(Sex a, Sex b, Sex c)
{
    Cons_Cell      *r;
    r = cellpointer - 2;
    show_usage(r, 2);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = c;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Lacons(Sex a, Sex b, Sex c)
{
    Cons_Cell      *r;
    r = cellpointer - 2;
    show_usage(r, 2);
    car(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    cdr(r) = c;
    car(r + 1) = a;
    cdr(r + 1) = b;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Llist3(Sex a, Sex b, Sex c)
{
    Cons_Cell      *r;
    r = cellpointer - 3;
    show_usage(r, 3);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}


Sex
Llist4(Sex a, Sex b, Sex c, Sex d)
{
    Cons_Cell      *r;
    r = cellpointer - 4;
    show_usage(r, 4);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Llist5(Sex a, Sex b, Sex c, Sex d, Sex e)
{
    Cons_Cell      *r;
    r = cellpointer - 5;
    show_usage(r, 5);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = (Sex) ((char *) (r + 4) + TAG_CONS);
    car(r + 4) = e;
    cdr(r + 4) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}


#ifdef ASUKA
Sex
Llist6(Sex a, Sex b, Sex c, Sex d, Sex e, Sex f)
{
    /*
     * Sex s; 
     */
    Cons_Cell      *r;
    r = cellpointer - 6;
    show_usage(r, 6);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = (Sex) ((char *) (r + 4) + TAG_CONS);
    car(r + 4) = e;
    cdr(r + 4) = (Sex) ((char *) (r + 5) + TAG_CONS);
    car(r + 5) = f;
    cdr(r + 5) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}

Sex
Llist7(Sex a, Sex b, Sex c, Sex d, Sex e, Sex f, Sex g)
{
    /*
     * Sex s; 
     */
    Cons_Cell      *r;
    r = cellpointer - 7;
    show_usage(r, 7);
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = (Sex) ((char *) (r + 4) + TAG_CONS);
    car(r + 4) = e;
    cdr(r + 4) = (Sex) ((char *) (r + 5) + TAG_CONS);
    car(r + 5) = f;
    cdr(r + 5) = (Sex) ((char *) (r + 6) + TAG_CONS);
    car(r + 6) = g;
    cdr(r + 6) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit) {
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    } else
	return (Sex) ((char *) r + TAG_CONS);
}
#endif

Sex
Lcaar(Sex a)
{
    return car(car(a));
}

Sex
Lcadr(Sex a)
{
    return car(cdr(a));
}

Sex
Lcdar(Sex a)
{
    a = car(a);
    return cdr(a);
}

Sex
Lcddr(Sex a)
{
    a = cdr(a);
    return cdr(a);
}

Sex
Lcaaar(Sex a)
{
    a = car(a);
    a = car(a);
    return car(a);
}

Sex
Lcaadr(Sex a)
{
    a = cdr(a);
    a = car(a);
    return car(a);
}

Sex
Lcadar(Sex a)
{
    a = car(a);
    a = cdr(a);
    return car(a);
}

Sex
Lcaddr(Sex a)
{
    a = cdr(a);
    a = cdr(a);
    return car(a);
}

Sex
Lcdaar(Sex a)
{
    a = car(a);
    a = car(a);
    return cdr(a);
}

Sex
Lcdadr(Sex a)
{
    a = cdr(a);
    a = car(a);
    return cdr(a);
}

Sex
Lcddar(Sex a)
{
    a = car(a);
    a = cdr(a);
    return cdr(a);
}

Sex
Lcdddr(Sex a)
{
    a = cdr(a);
    a = cdr(a);
    return cdr(a);
}

Sex
Lcaaaar(Sex a)
{
    a = car(a);
    a = car(a);
    a = car(a);
    return car(a);
}

Sex
Lcaaadr(Sex a)
{
    a = cdr(a);
    a = car(a);
    a = car(a);
    return car(a);
}

Sex
Lcaadar(Sex a)
{
    a = car(a);
    a = cdr(a);
    a = car(a);
    return car(a);
}

Sex
Lcaaddr(Sex a)
{
    a = cdr(a);
    a = cdr(a);
    a = car(a);
    return car(a);
}

Sex
Lcadaar(Sex a)
{
    a = car(a);
    a = car(a);
    a = cdr(a);
    return car(a);
}

Sex
Lcadadr(Sex a)
{
    a = cdr(a);
    a = car(a);
    a = cdr(a);
    return car(a);
}

Sex
Lcaddar(Sex a)
{
    a = car(a);
    a = cdr(a);
    a = cdr(a);
    return car(a);
}

Sex
Lcadddr(Sex a)
{
    a = cdr(a);
    a = cdr(a);
    a = cdr(a);
    return car(a);
}

Sex
Lcdaaar(Sex a)
{
    a = car(a);
    a = car(a);
    a = car(a);
    return cdr(a);
}

Sex
Lcdaadr(Sex a)
{
    a = cdr(a);
    a = car(a);
    a = car(a);
    return cdr(a);
}

Sex
Lcdadar(Sex a)
{
    a = car(a);
    a = cdr(a);
    a = car(a);
    return cdr(a);
}

Sex
Lcdaddr(Sex a)
{
    a = cdr(a);
    a = cdr(a);
    a = car(a);
    return cdr(a);
}

Sex
Lcddaar(Sex a)
{
    a = car(a);
    a = car(a);
    a = cdr(a);
    return cdr(a);
}

Sex
Lcddadr(Sex a)
{
    a = cdr(a);
    a = car(a);
    a = cdr(a);
    return cdr(a);
}

Sex
Lcdddar(Sex a)
{
    a = car(a);
    a = cdr(a);
    a = cdr(a);
    return cdr(a);
}

Sex
Lcddddr(Sex a)
{
    a = cdr(a);
    a = cdr(a);
    a = cdr(a);
    return cdr(a);
}


Sex
Lidp(Sex a)
{
    return Bool_to_Sex(is_symhdr(a));
}


Sex
Lpairp(Sex a)
{
    return Bool_to_Sex(is_cons(a));
}

Sex
Lconstantp(Sex a)
{
#if 0
    return Bool_to_Sex(is_number(a));
#else
    return Bool_to_Sex(is_number(a) || is_vechdr(a) || is_binhdr(a));
#endif
}



Sex
Lnumtype(Sex p)
{
    if (!is_numhdr(p))
	return nil;
    if (is_bigneg(p))
	return mkfxat(numtype_of(p));
    return mkfxat(numtype_of(p));
}

Sex
Lgnumtype(Sex p)
{
    if (!is_vechdr(p) || (vechdr(p) & 0x5f2) != 0x5f2)
	return nil;
    return mkfxat(gnumtype_of(p));
}

Sex
Lbigp(Sex p)
{
    return is_bigint(p) ? true : nil;
}

Sex
Lfixnump(Sex p)
{
    return is_fixnum(p) ? true : nil;
}

Sex
Lfixp(Sex p)
{
    switch (((long) p) & TAG_BITS) {
    case TAG_FIXNUM:
	return true;
    case TAG_NUMBERS:
	switch (numtype_of(p)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return true;
	}
    }
    return nil;
}

Sex
Llogicalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_LOGICAL)
	return true;
    return nil;
}


Sex
Lrationalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_RATIONAL)
	return true;
    return nil;
}

Sex
Lbfloatp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BFLOAT)
	return true;
    return nil;
}


Sex
Lefloatp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_EFLOAT)
	return true;
    return nil;
}


Sex
Lbefloatp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BEFLOAT)
	return true;
    return nil;
}

Sex
Lcomplexp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_COMPLEX)
	return true;
    return nil;
}

Sex
Lecomplexp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_ECOMPLEX)
	return true;
    return nil;
}


Sex
Lbecomplexp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BECOMPLEX)
	return true;
    return nil;
}

Sex
Lintervalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_INTERVAL)
	return true;
    return nil;
}

Sex
Lbintervalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BINTERVAL)
	return true;
    return nil;
}



bool
stringp(Sex a)
{
    if (is_string(a))
	return YES;
    return NO;
}

Sex
Lstringp(Sex a)
{
    if (!is_string(a))
	return nil;
    else
	return true;
}

Sex
Lvectorp(Sex a)
{
    if (!is_simple_vector(a))
	return nil;
    return true;
}

Sex
Larrayp(Sex a)
{
    if (!is_array(a))
	return nil;
    return true;
}

Sex
make_sov(Sex els)
{
    Sex             obj;
    long            size,
                    i;

    size = list_length(els);
    stackcheck();
    push(els);
    obj = getvector(TAG_VECTOR, T_SIMPLE, bnum_len(size));
    pop(els);
    for (i = 0; i < size; i++) {
	vector_value(obj, i) = car(els);
	els = cdr(els);
    }
    return obj;
}


static          Sex
Lpackagep(Sex a)
{
    if (!is_package(a))
	return nil;
    return true;
}

static          Sex
Lpackage_cnt(Sex a)
{
    if (!is_package(a))
	return nil;
    return package_cnt(a);
}

static          Sex
Lhashp(Sex a)
{
    if (!is_hashtable(a))
	return nil;
    return true;
}


Sex
Lflagp(Sex a, Sex b)
{
    Sex             pl,
                    w;
    if (!is_symhdr(a))
	return nil;
    pl = plist(a);
    while (pl != nil) {
	w = car(pl);
	if (car(w) == b)
	    return true;
	pl = cdr(pl);
    }
    return nil;
}

Sex
Lflagpcar(Sex a, Sex b)
{
    Sex             pl,
                    w;
    if (a != nil) {
	if (!is_cons(a))
	    return nil;
	a = car(a);
	if (!is_symhdr(a))
	    return nil;
    }
    pl = plist(a);
    while (pl != nil) {
	w = car(pl);
	if (car(w) == b)
	    return true;
	pl = cdr(pl);
    }
    return nil;
}

Sex
Lflag(Sex a, Sex b)
{
    stackcheck();
    while (is_cons(a)) {
	Sex             v = car(a),
	    pl;
	a = cdr(a);
	if (!is_symhdr(v))
	    continue;
	push2(a, b);
	pl = plist(v);
	while (pl != nil) {
	    Sex             w = car(pl);
	    if (car(w) == b) {
		cdr(w) = true;
		goto already_flagged;
	    } else
		pl = cdr(pl);
	}
	push(v);
	b = Lacons(b, true, plist(v));
	pop(v);
	plist(v) = b;
      already_flagged:
	pop2(b, a);
    }
    return nil;
}

Sex
Lremflag(Sex a, Sex b)
{
    while (is_cons(a)) {
	Sex             pl,
	                prevp,
	                v = car(a);
	a = cdr(a);
	if (!is_symhdr(v))
	    continue;
	prevp = nil;
	pl = plist(v);
	while (pl != nil) {
	    Sex             w = car(pl);
	    if (car(w) == b) {
		pl = cdr(pl);
		if (prevp == nil)
		    plist(v) = pl;
		else
		    cdr(prevp) = pl;
		break;
	    }
	    prevp = pl;
	    pl = cdr(prevp);
	}
    }
    return nil;
}


Sex
Lmkquote(Sex a)
{
    a = Llist2(quote_symbol, a);
    return a;
}

Sex
Lpair(Sex a, Sex b)
{
    Sex             r = nil;
    stackcheck();
    while (is_cons(a) && is_cons(b)) {
	push2(a, b);
	r = Lacons(car(a), car(b), r);
	pop2(b, a);
	a = cdr(a);
	b = cdr(b);
    }
    a = nil;
    while (r != nil) {
	b = cdr(r);
	cdr(r) = a;
	a = r;
	r = b;
    }
    return a;
}

static long
memqcount(Sex a, Sex b)
/*
 * Counts how many times a is a member of the list b
 */
{
    long            r = 0;
    if (is_symhdr(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    if (a == car(b))
		r++;
	    b = cdr(b);
	}
	return r;
    }
    while (is_cons(b)) {
	Sex             cb = car(b);
	if (a == cb)
	    r++;
	b = cdr(b);
    }
    return r;
}

static long
membercount(Sex a, Sex b)
/*
 * Counts how many times a is a member of the list b
 */
{
    long            r = 0;
    if (is_symhdr(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    if (a == car(b))
		r++;
	    b = cdr(b);
	}
	return r;
    }
    while (is_cons(b)) {
	Sex             cb = car(b);
	if (a == cb || equal(a, cb) != nil)
	    r++;
	b = cdr(b);
    }
    return r;
}

Sex
Lintersection(Sex a, Sex b)
{
    Sex             r = nil,
	w;
    long            n1,
                    n2;
    stackcheck();
    push(b);
    while (is_cons(a)) {
	push2(a, r);
	w = Lmember(car(a), stack[-2]);
	if (w != nil) {
	    n1 = membercount(car(stack[-1]), stack[0]);
	    if (n1 != 0) {
		n2 = membercount(car(stack[-1]), stack[-2]);
		if (n2 > n1)
		    n1 = 0;
	    }
	    if (n1 == 0) {
		pop(r);
		a = stack[0];
		r = cons(car(a), r);
		pop(a);
	    } else
		pop2(r, a);
	} else
	    pop2(r, a);
	a = cdr(a);
    }
    popv(1);
    a = nil;
    while (is_cons(r)) {
	b = r;
	r = cdr(r);
	cdr(b) = a;
	a = b;
    }
    return a;
}

Sex
Lintersectionq(Sex a, Sex b)
{
    Sex             r = nil,
	w;
    long            n1,
                    n2;
    stackcheck();
    push(b);
    while (is_cons(a)) {
	push2(a, r);
	w = Lmemq(car(a), stack[-2]);
	if (w != nil) {
	    n1 = memqcount(car(stack[-1]), stack[0]);
	    if (n1 != 0) {
		n2 = memqcount(car(stack[-1]), stack[-2]);
		if (n2 > n1)
		    n1 = 0;
	    }
	    if (n1 == 0) {
		pop(r);
		a = stack[0];
		r = cons(car(a), r);
		pop(a);
	    } else
		pop2(r, a);
	} else
	    pop2(r, a);
	a = cdr(a);
    }
    popv(1);
    a = nil;
    while (is_cons(r)) {
	b = r;
	r = cdr(r);
	cdr(b) = a;
	a = b;
    }
    return a;
}

Sex
Lunion(Sex a, Sex b)
{
    stackcheck();
    while (is_cons(a)) {
	Sex             c;
	push2(a, b);
	c = Lmember(car(a), b);
	pop(b);
	if (c == nil) {
	    b = cons(car(stack[0]), b);
	}
	pop(a);
	a = cdr(a);
    }
    return b;
}

Sex
Lunionq(Sex a, Sex b)
{
    stackcheck();
    while (is_cons(a)) {
	Sex             c;
	push2(a, b);
	c = Lmemq(car(a), b);
	pop(b);
	if (c == nil) {
	    b = cons(car(stack[0]), b);
	}
	pop(a);
	a = cdr(a);
    }
    return b;
}

Sex
Lerror(Sex msg, Sex a)
{
    push2(msg, a);
    if (miscflags & MESSAGES_FLAG) {
	lisp_printf("\n *** error: %P, ", stack[-1]);
	lisp_printf("%P\n", stack[0]);
    }
    popv(2);
    longjmp(*errorset_buffer, 2);
    // return nil;
}

Sex
Lerror0()
{
    miscflags &= ~(MESSAGES_FLAG);
    return nil;
}

Sex
Lerror1()
{
    if (miscflags & MESSAGES_FLAG) {
	lisp_printf("\n *** error: \n");
    }
    longjmp(*errorset_buffer, 2);
    // return nil;
}

Sex
Lerror2(Sex msg)
{
    if (miscflags & MESSAGES_FLAG) {
	lisp_printf("\n *** error: %P\n", msg);
    }
    longjmp(*errorset_buffer, 2);
    // return nil;
}

Sex
Lstop(Sex code)
{
    if (!is_fixnum(code))
	return aerror("stop");
    return nil;
}

Sex
Lfluid(Sex a)
{
    Sex             q,
                    b;

    b = a;
    stackcheck();
    while (is_cons(b)) {
	q = car(b);
	if (!is_symhdr(q)) {
	    push2(a, b);
	    warnings("non symbol", q, nil);
	    pop2(b, a);
	    /*
	     * return aerror("fluid"); 
	     */
	} else {
	    qtag(q) |= SYM_FLUID;
	    if (value(q) == undefined_var)
		value(q) = nil;
	}
	b = cdr(b);
    }
    return a;
}

Sex
Lglobal(Sex a)
{
    Sex             q,
                    b;
    b = a;
    while (is_cons(b)) {
	q = car(b);
	if (!is_symhdr(q))
	    return aerror("global");
	qtag(q) |= (SYM_FLUID | SYM_GLOBAL);
	if (value(q) == undefined_var)
	    value(q) = nil;
	b = cdr(b);
    }
    return a;
}

Sex
Lmake_fluid(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    qtag(a) |= SYM_FLUID;
    if (value(a) == undefined_var)
	value(a) = nil;
    return a;
}

Sex
Lmake_special(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    qtag(a) |= SYM_FLUID;
    if (value(a) == undefined_var)
	value(a) = nil;
    return a;
}

Sex
Lmake_global(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    qtag(a) |= SYM_FLUID | SYM_GLOBAL;
    if (value(a) == undefined_var)
	value(a) = nil;
    return a;
}

Sex
Lunfluid(Sex a)
{
    long            t;
    Sex             b,
                    q;

    b = a;
    while (is_cons(b)) {
	q = car(b);
	b = cdr(b);
	if (!is_symhdr(q))
	    continue;
	t = qtag(q) & ~(SYM_FLUID | SYM_GLOBAL);
	value(q) = undefined_var;
	qtag(q) = t;
    }
    return a;
}

Sex
Lmake_unfluid(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    qtag(a) &= ~(SYM_FLUID | SYM_GLOBAL);
    value(a) = undefined_var;
    return a;
}

Sex
Lmake_unglobal(Sex a)
{
    long            t;
    if (!is_symhdr(a))
	return nil;
    t = qtag(a) & ~(SYM_GLOBAL);
    if ((t & SYM_FLUID) == 0)
	value(a) = nil;
    return a;
}

Sex
Lmake_unspecial(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    qtag(a) &= ~(SYM_FLUID | SYM_GLOBAL);
    value(a) = undefined_var;
    return a;
}


Sex
Lunglobal(Sex a)
{
    long            t;
    Sex             b,
                    q;

    b = a;
    while (is_cons(b)) {
	q = car(b);
	b = cdr(b);
	if (!is_symhdr(q))
	    continue;
	t = qtag(q) & ~(SYM_GLOBAL);
	if ((t & SYM_FLUID) == 0)
	    value(q) = nil;
	qtag(q) = t;
    }
    return a;
}


Sex
Lis_special_var(Sex a)
{
    if (is_symhdr(a) && (qtag(a) & SYM_FLUID) != 0)
	return true;
    return nil;
}

Sex
Lfluidp(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    /*
     * else if ((qtag(a) & (SYM_FLUID | SYM_GLOBAL)) == SYM_FLUID) 
     */
    else if ((qtag(a) & SYM_FLUID) != 0)
	return true;
    else
	return nil;
}

Sex
Lglobalp(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    else if ((qtag(a) & SYM_GLOBAL) != 0)
	return true;
    else
	return nil;
}

Sex
Lboundp(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    else if ((qtag(a) & SYM_FLUID) == 0)
	return nil;
    else
	return true;
}

Sex
Lsymbol_value(Sex a)
{
    if (!is_symhdr(a))
	return a;
    else
	return value(a);
}


Sex
Lset(Sex a, Sex b)
{
    if (!is_symhdr(a) || a == nil || a == true) {
	return aerror("set");
    }
    value(a) = b;
    return b;
}


Sex
Lcodep(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    if ((qtag(a) & PROC_CCODE) == 0)
	return nil;
    else
	return true;
}


Sex
Lbytecodep(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    if ((qtag(a) & PROC_BCODE) == 0)
	return nil;
    else
	return true;
}

Sex
getvector(long tag, long type, long size)
{

    while (YES) {
	char           *r = (char *) fwdpointer;
	long            alloc_size = (long) alignit(size);

	if (alloc_size > (long) fwdlimit - (long) r) {
	    (void) reclaim(nil, GC_FWD, alloc_size);
	    continue;
	}
	fwdpointer = (Sex *) (r + alloc_size);
	*((Tag *) r) = type | TAG_HDRS;
	*((long *) (r + SEX)) = size;
	return (Sex) (r + tag);
    }
}

Sex
getvector_init(long n, Sex k)
{
    Sex             p;
    push(k);
    n = alignit(n);
    p = getvector(TAG_VECTOR, T_SIMPLE, n);
    pop(k);
    n = bnum_n(n);
    while (n > 0) {
	n--;
	vector_value(p, n) = k;
    }
    return p;
}

double          base_time = 0.0,
    io_time = 0.0,
    gc_time = 0.0,
    trap_time = 0.0;

Sex
Lfloat_time()
{
    double          clock;

    clock = get_time();
    clock = clock - base_time - io_time - gc_time;
    return make_float(clock);
}


Sex
Ltime()
{
    double          clock;

    clock = get_time();
    clock = clock - base_time - io_time - gc_time;
    return make_float((clock * 1000));
}

Sex
Lgctime()
{
    // return mkfxat((long) (gc_time * 1000));
    return make_float(gc_time * 1000);
}

Sex
Lgccalled()
{
    // return mkfxat((long) (gc_time * 1000));
    return mkfxat(number_of_gc_called);
}

Sex
Ldate_and_time()
{
    Sex             w;
    time_t          t = time(NULL);
    char            today[32];
    strcpy(today, ctime(&t));
    w = make_string2(today, 24);
    return w;
}

Sex
Ldate()
{
    Sex             w;
    time_t          t = time(NULL);
    char            today[32];
    strcpy(today, ctime(&t));
    w = make_string2(today, 24);
    return w;
}

Sex
Ldatestamp()
{
    Sex             w;
    time_t          n = time(NULL);
    if (n >= fix_min && n <= fix_max)
	return mkfxat(n);
    return mkbigat1s(1, n);
}

Sex
deleqip(Sex a, Sex l)
{
    Sex             w,
                    r;
    if (l == nil)
	return nil;
    if (car(l) == a)
	return cdr(l);
    r = l;
    while (w = l, (l = cdr(l)) != nil) {
	if (car(l) == a) {
	    cdr(w) = cdr(l);
	    return r;
	}
    }
    return r;
}


Sex
Lputd(Sex a, Sex b, Sex c)
{
    long            ptype;

    if (!is_symhdr(a))
	return aerror1("putd", a);
    if (b == expr_symbol || b == subr_symbol) {
	ptype = PROC_EXPR;
    } else if (b == fexpr_symbol || b == fsubr_symbol) {
	ptype = PROC_FEXPR;
    } else if (b == macro_symbol) {
	ptype = PROC_MACRO;
    } else if (b == smacro_symbol) {
	ptype = PROC_SMACRO;
    } else {
	return aerror1("putd", b);
    }
    if (is_symhdr(c)) {
	if ((qtag(c) & SYM_PROC) == 0)
	    return aerror1("putd", c);
	qtag(a) = (qtag(a) & ~PROC_MASK) | (qtag(c) & PROC_MASK);
	proc(a) = proc(c);
	qnargs(a) = qnargs(c);
	return a;
    } else if (!is_cons(c)) {
	return aerror1("putd", c);
    } else if (is_fixnum(car(c))) {
	Sex             d = cdr(c);

	if (is_logical(d)) {
	    ptype |= PROC_CCODE;
	    proc(a) = (lisp_proc *) logical_value(d)[0];
	} else if (is_fixnum(d)) {
	    ptype |= PROC_BCODE;
	    bcode(a) = mkint(d);
	} else {
	    aerror1("putd", c);
	}
	qnargs(a) = mkint(car(c));
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else if (car(c) == lambda) {
	Sex             bvl;
	long            nargs;
	c = cdr(c);
	bvl = car(c);
	nargs = 0;
	while (is_cons(bvl)) {
	    nargs++;
	    bvl = cdr(bvl);
	}
	proc(a) = (lisp_proc *) c;
	qnargs(a) = nargs;
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else if (car(b) == funarg) {
	Sex             bvl;
	long            nargs;
	c = cdr(c);
	bvl = car(c);
	nargs = 0;
	while (is_cons(bvl)) {
	    nargs++;
	    bvl = cdr(bvl);
	}
	proc(a) = (lisp_proc *) c;
	qnargs(a) = nargs;
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else
	return aerror1("putd", b);
}



Sex
Lgetd(Sex a)
{
    Tag             h;
    Sex             type;
    long            narg;
    if (!is_symhdr(a))
	return aerror1("getd", a);
    h = qtag(a);
    switch (h & PROC_MASK) {
    case PROC_SCODE:
    case PROC_CSMACRO:
	a = cons(lambda, Lget(a, smacro_symbol));
	type = smacro_symbol;
	break;
    case PROC_SMACRO:
	a = cons(lambda, (Sex) proc(a));
	type = smacro_symbol;
	break;
    case PROC_MCODE:
    case PROC_CMACRO:
	a = cons(lambda, Lget(a, macro_symbol));
	type = macro_symbol;
	break;
    case PROC_MACRO:
	a = cons(lambda, (Sex) proc(a));
	type = macro_symbol;
	break;
    case PROC_EXPR:
	a = cons(lambda, (Sex) proc(a));
	type = expr_symbol;
	break;
    case PROC_FEXPR:
	a = cons(lambda, (Sex) proc(a));
	type = fexpr_symbol;
	break;
    case PROC_SUBR:
	narg = qnargs(a);
	a = mklogical((long) proc(a));
	a = cons(mkfxat(narg), a);
	type = expr_symbol;
	break;
    case PROC_FSUBR:
	narg = qnargs(a);
	a = mklogical((long) proc(a));
	a = cons(mkfxat(narg), a);
	type = fexpr_symbol;
	break;
    case PROC_CODE:
	narg = qnargs(a);
	a = mkfxat((long) bcode(a));
	a = cons(mkfxat(narg), a);
	type = expr_symbol;
	break;
    case PROC_FCODE:
	narg = qnargs(a);
	a = mkfxat((long) bcode(a));
	a = cons(mkfxat(narg), a);
	type = fexpr_symbol;
	break;
    default:
	return nil;
    }
    return cons(type, a);
}


Sex
Lmovd(Sex a, Sex b)
{
    b = Lgetd(b);
    if (is_cons(b))
	Lputd(a, car(b), cdr(b));
    return b;
}


Sex
Lis_macro(Sex a)
{
    Tag             h;
    if (!is_symhdr(a))
	return nil;
    h = qtag(a);
    switch (h & PROC_MASK) {
    case PROC_MCODE:
    case PROC_CMACRO:
    case PROC_MACRO:
	return true;
    case PROC_SCODE:
    case PROC_CSMACRO:
    case PROC_SMACRO:
	return true;
    }
    return nil;
}

Sex
Lis_special(Sex a)
{
    Tag             h;
    if (!is_symhdr(a))
	return nil;
    h = qtag(a);
    switch (h & PROC_MASK) {
    case PROC_FEXPR:
    case PROC_FSUBR:
    case PROC_FCODE:
	return true;
    }
    return nil;
}


Sex
Lget_number_of_args(Sex a)
{
    Tag             h;
    long            narg;
    if (!is_symhdr(a))
	return nil;
    h = qtag(a);
    switch (h & PROC_MASK) {
    case PROC_SCODE:
    case PROC_CSMACRO:
    case PROC_SMACRO:
    case PROC_EXPR:
    case PROC_SUBR:
    case PROC_CODE:
	break;
    case PROC_CMACRO:
    case PROC_MACRO:
    case PROC_MCODE:
    case PROC_FEXPR:
    case PROC_FSUBR:
    case PROC_FCODE:
    default:
	return nil;
    }
    narg = qnargs(a);
    return mkfxat(narg);
}

Sex
Lput_proc(Sex a, Sex b, Sex c)
{
    long            ptype;

    if (!is_symhdr(a))
	return aerror1("putd", a);
    if (b == expr_symbol) {
	ptype = PROC_EXPR;
    } else if (b == fexpr_symbol) {
	ptype = PROC_FEXPR;
    } else if (b == macro_symbol) {
	ptype = PROC_MACRO;
    } else if (b == smacro_symbol) {
	ptype = PROC_SMACRO;
    } else {
	return aerror1("putd", b);
    }
    if (is_symhdr(c)) {
	if ((qtag(c) & SYM_PROC) == 0)
	    return aerror1("putd", c);
	qtag(a) = (qtag(a) & ~PROC_MASK) | (qtag(c) & PROC_MASK);
	proc(a) = proc(c);
	qnargs(a) = qnargs(c);
	return a;
    } else if (!is_cons(c)) {
	return aerror1("putd", c);
    } else if (is_fixnum(car(c))) {
	Sex             d = cdr(c);
	if (!is_logical(d))
	    aerror1("putd", c);
	ptype |= PROC_CCODE;
	proc(a) = (lisp_proc *) logical_value(d)[0];
	qnargs(a) = mkint(car(c));
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else if (car(c) == lambda) {
	Sex             bvl;
	long            nargs;
	c = cdr(c);
	bvl = car(c);
	nargs = 0;
	while (is_cons(bvl)) {
	    nargs++;
	    bvl = cdr(bvl);
	}
	proc(a) = (lisp_proc *) c;
	qnargs(a) = nargs;
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else if (car(b) == funarg) {
	Sex             bvl;
	long            nargs;
	c = cdr(c);
	bvl = car(c);
	nargs = 0;
	while (is_cons(bvl)) {
	    nargs++;
	    bvl = cdr(bvl);
	}
	proc(a) = (lisp_proc *) c;
	qnargs(a) = nargs;
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else
	return aerror1("putd", b);
}



Sex
Lget_proc(Sex a, Sex b)
{
    Tag             h;
    Sex             type;
    long            narg;
    if (!is_symhdr(a))
	return aerror1("get-proc", a);
    if (b == csmacro_symbol) {
	a = cons(lambda, Lget(a, smacro_symbol));
	type = smacro_symbol;
    } else if (b == smacro_symbol) {
	a = cons(lambda, (Sex) proc(a));
	type = smacro_symbol;
    } else if (b == mcode_symbol || b == macro_symbol) {
	a = cons(lambda, Lget(a, macro_symbol));
	type = macro_symbol;
    } else if (b == macro_symbol) {
	a = cons(lambda, (Sex) proc(a));
	type = macro_symbol;
    } else if (b == expr_symbol) {
	a = cons(lambda, (Sex) proc(a));
	type = expr_symbol;
    } else if (b == fexpr_symbol) {
	a = cons(lambda, (Sex) proc(a));
	type = fexpr_symbol;
    } else if (b == subr_symbol) {
	narg = qnargs(a);
	a = mklogical((long) proc(a));
	a = cons(mkfxat(narg), a);
	type = expr_symbol;
    } else if (b == fsubr_symbol) {
	narg = qnargs(a);
	a = mklogical((long) proc(a));
	a = cons(mkfxat(narg), a);
	type = fexpr_symbol;
    } else if (b == code_symbol) {
	narg = qnargs(a);
	a = mkfxat((long) bcode(a));
	a = cons(mkfxat(narg), a);
	type = expr_symbol;
    } else if (b == fcode_symbol) {
	narg = qnargs(a);
	a = mkfxat((long) bcode(a));
	a = cons(mkfxat(narg), a);
	type = fexpr_symbol;
    } else
	return nil;
    return cons(type, a);
}

Sex
Lremd(Sex a)
{
    Sex             res;
    if (!is_symhdr(a))
	return aerror1("remd", a);
    res = Lgetd(a);
    if (res == nil)
	return nil;
    qtag(a) = qtag(a) & ~(SYM_TRACED | PROC_MASK);
    proc(a) = (lisp_proc *) undefined_proc;
    qnargs(a) = 0;
    return res;
}


Sex
Ltraceall(Sex a)
{
    if (a == nil)
	trace_flag &= ~(SYM_TRACED | SYM_PROC);
    else
	trace_flag |= (SYM_TRACED | SYM_PROC);
    return a;
}

Sex
Ltrace(Sex a)
{
    Sex             q,
                    m;

    if (is_cons(a)) {
	q = a;
	while (is_cons(q)) {
	    m = car(q);
	    if (is_symhdr(m) && (qtag(m) & SYM_PROC) != 0)
		qtag(m) |= SYM_TRACED;
	    q = cdr(q);
	}
	trace_flag |= SYM_TRACED;
    } else if (is_fixnum(a)) {
	trace_level = mkint(a);
    } else if (a == nil)
	trace_flag &= ~(SYM_TRACED | SYM_PROC);
    else
	trace_flag |= SYM_TRACED;
    return a;
}

Sex
Luntrace(Sex a)
{
    Sex             q,
                    m;

    if (is_cons(a)) {
	q = a;
	while (is_cons(q)) {
	    m = car(q);
	    if (is_symhdr(m) && (qtag(m) & SYM_PROC) != 0)
		qtag(m) &= ~SYM_TRACED;
	    q = cdr(q);
	}
    } else if (a == nil)
	trace_flag |= SYM_TRACED;
    else
	trace_flag &= ~(SYM_TRACED | SYM_PROC);
    return a;
}

Sex
Lmacro_function(Sex a)
{
    if (!is_symhdr(a))
	return nil;
    else if ((qtag(a) & PROC_MASK) != PROC_MACRO)
	return nil;
    else
	return cons(lambda, (Sex) proc(a));
}



Sex
Lsymbol_length(Sex a)
{
    if (!is_symhdr(a))
	return aerror1("symbol-name", a);
    return mkfxat(symbol_length(a));
}

Sex
Lsymbol_name(Sex a)
{
    if (!is_symhdr(a))
	return aerror1("symbol-name", a);
    return make_string2(pname(a), symbol_length(a));
}

Sex
Lplist(Sex a)
{
    if (!is_symhdr(a))
	return aerror1("plist", a);
    return plist(a);
}


Sex
Latom_tag(Sex a)
{
    if (!is_symhdr(a))
	return aerror1("plist", a);
    return mklogical(qtag(a));
}

Sex
Lnumber_tag(Sex a)
{
    if (!is_numhdr(a))
	return aerror1("number-tag", a);
    return mklogical(numhdr(a));
}

Sex
Lnumber_length(Sex a)
{
    if (!is_numhdr(a))
	return aerror1("number-tag", a);
    return mkfxat(bignum_length(a));
}

Sex
Lmkvect(Sex n)
{
    long            n1;
    if (!is_fixnum(n) || (n1 = mkint(n)) < 0)
	return aerror1("mkvect", n);
    n1 = bnum_len(n1 + 1);	/* index = 0 .. n) */
    return getvector_init(n1, nil);
}


Sex
Larraydim(Sex a)
{
    if (!is_array(a))
	return mkfxat(0);
    return array_dim(a);
}

long
checkarray1(Sex n1)
{
    long            d1;
    if (!is_fixnum(n1) || (d1 = mkint(n1)) <= 0)
	return aerror1("checkarray1", n1);
    return d1;
}

long
checkarray2(Sex n1, Sex n2)
{
    long            d1,
                    d2;
    long            n;
    if (!is_fixnum(n1) || (d1 = mkint(n1)) <= 0
	|| !is_fixnum(n2) || (d2 = mkint(n2)) <= 0)
	return aerror2("checkarray2", n1, n2);
    n = d1 * d2;
    if (n <= 0)
	return aerror2("checkarray2", n1, n2);
    return n;
}


long
checkarray3(Sex n1, Sex n2, Sex n3)
{
    long            n;
    long            d1,
                    d2,
                    d3;
    if (!is_fixnum(n1) || (d1 = mkint(n1)) <= 0
	|| !is_fixnum(n2) || (d2 = mkint(n2)) <= 0
	|| !is_fixnum(n3) || (d3 = mkint(n3)) <= 0)
	return aerror3("checkarray3", n1, n2, n3);
    n = d1 * d2 * d3;
    if (n <= 0)
	return aerror3("checkarray3", n1, n2, n3);
    return n;
}

Sex
mkarray(long n)
{
    long            i;
    Sex             p;
    Sex             d;
    Sex            *m;
    if (n <= 0)
	return aerror1("mkarray", mkfxat(n));
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (n - 1) * sizeof(Sex));
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    d = mkfxat(n);
    d = Lncons(d);
    pop(p);
    array_dim(p) = d;
    return p;
}

Sex
Lmkarray1(Sex n1)
{
    long            i,
                    d1;
    Sex             p;
    Sex            *m;
    d1 = checkarray1(n1);
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (d1 - 1) * sizeof(Sex));
    m = &array_value(p, 0);
    for (i = 0; i < d1; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    n1 = mkfxat(d1);
    n1 = Lncons(n1);
    pop(p);
    array_dim(p) = n1;
    /*
     * printf("mkarray: len = %d\n", veclen(p)); 
     */
    return p;
}

Sex
Lmkarray2(Sex n1, Sex n2)
{
    long            i,
                    n;
    Sex             p;
    Sex            *m;
    n = checkarray2(n1, n2);
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (n - 1) * sizeof(Sex));
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    n2 = Lncons(n2);
    n1 = cons(n1, n2);
    pop(p);
    array_dim(p) = n1;
    return p;
}

Sex
Lmkarray3(Sex n1, Sex n2, Sex n3)
{
    long            i,
                    n;
    Sex             p;
    Sex            *m;
    n = checkarray3(n1, n2, n3);
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (n - 1) * sizeof(Sex));
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    n3 = Lncons(n3);
    n2 = cons(n2, n3);
    n1 = cons(n1, n2);
    pop(p);
    array_dim(p) = n1;
    return p;
}

Sex
mkarray_fn(Sex args, Sex env)
{
    long            i;
    long            n;
    long            dim[10];
    Sex             p,
                    r;
    Sex            *m;

    if (!is_cons(args))
	return aerror1("mkarray", args);
    stackcheck();
    n = 1;
    for (i = 0; i < 10; i++) {
	push2(args, env);
	r = car(args);
	r = eval(r, env);
	if (!is_fixnum(r))
	    return aerror1("mkarray", r);
	dim[i] = mkint(r);
	n = n * dim[i];
	if (n < 0 || n > 0x0fffffff)
	    return aerror1("mkarray", r);
	pop2(env, args);
	args = cdr(args);
	if (!is_cons(args))
	    break;
    }
    r = nil;
    for (; i >= 0; i--)
	r = cons(mkfxat(dim[i]), r);
    push(r);
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (n - 1) * sizeof(Sex));
    pop(r);
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    array_dim(p) = r;
    return p;
}


Sex
Lmkiarray1(Sex n1)
{
    long            n,
                    i;
    int            *m;
    Sex             p;

    n = checkarray1(n1);
    p = getvector(TAG_BINARY, T_IARRAY, n * sizeof(int) + HDR);
    m = &iarray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

Sex
Lmkiarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    int            *m;
    Sex             p;

    n = checkarray2(n1, n2);
    p = getvector(TAG_BINARY, T_IARRAY2, n * sizeof(int) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &iarray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

Sex
Lmkiarray3(Sex n1, Sex n2, Sex n3)
{
    long            n,
                    i;
    int            *m;
    Sex             p;

    n = checkarray3(n1, n2, n3);
    p = getvector(TAG_BINARY, T_IARRAY3, n * sizeof(int) + HDR + SEX * 2);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &iarray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

Sex
Lmklarray1(Sex n1)
{
    long            n,
                    i;
    long           *m;
    Sex             p;

    n = checkarray1(n1);
    p = getvector(TAG_BINARY, T_IARRAY, n * sizeof(long) + HDR);
    m = &larray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

Sex
Lmklarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    long           *m;
    Sex             p;

    n = checkarray2(n1, n2);
    p = getvector(TAG_BINARY, T_IARRAY2, n * sizeof(long) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &larray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

Sex
Lmklarray3(Sex n1, Sex n2, Sex n3)
{
    long            n,
                    i;
    long           *m;
    Sex             p;

    n = checkarray3(n1, n2, n3);
    p = getvector(TAG_BINARY, T_IARRAY3, n * sizeof(long) + HDR + SEX * 2);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &larray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

Sex
Lmkfarray1(Sex n1)
{
    long            n,
                    i;
    float          *m;
    Sex             p;

    n = checkarray1(n1);
    p = getvector(TAG_BINARY, T_SARRAY, n * sizeof(float) + HDR);
    m = &sarray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}
Sex
Lmkfarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    float          *m;
    Sex             p;

    n = checkarray2(n1, n2);
    p = getvector(TAG_BINARY, T_SARRAY2, n * sizeof(float) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &sarray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

Sex
Lmkfarray3(Sex n1, Sex n2, Sex n3)
{
    long
                    n,
                    i;
    float          *m;
    Sex             p;

    n = checkarray3(n1, n2, n3);
    p = getvector(TAG_BINARY, T_SARRAY3,
		  n * sizeof(float) + HDR + SEX * 2);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &sarray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

Sex
Lmkdarray1(Sex n1)
{
    long            n,
                    i;
    double         *m;
    Sex             p;

    n = checkarray1(n1);
    p = getvector(TAG_BINARY, T_DARRAY, n * sizeof(double) + HDR);
    m = &darray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

Sex
Lmkdarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    double         *m;
    Sex             p;

    n = checkarray2(n1, n2);
    p = getvector(TAG_BINARY, T_DARRAY2, n * sizeof(double) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &darray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

Sex
Lmkdarray3(Sex n1, Sex n2, Sex n3)
{
    long            n,
                    i;
    double         *m;
    Sex             p;

    n = checkarray3(n1, n2, n3);
    p = getvector(TAG_BINARY, T_DARRAY3,
		  n * sizeof(double) + HDR + SEX * 2);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &darray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}


static          Sex
list_getarray1(Sex v, long d)
{
    long            i;
    for (i = 1; i < d; i++) {
	if (!is_cons(v))
	    aerror("list_getarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_getarray");
    return car(v);
}

static          Sex
list_getarray2(Sex v, long d1, long d2)
{
    long            i;
    for (i = 1; i < d1; i++) {
	if (!is_cons(v))
	    aerror("list_getarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_getarray");
    v = car(v);
    for (i = 1; i < d2; i++) {
	if (!is_cons(v))
	    aerror("list_getarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_getarray");
    return car(v);
}

static          Sex
list_getarray3(Sex v, long d1, long d2, long d3)
{
    long            i;
    for (i = 1; i < d1; i++) {
	if (!is_cons(v))
	    aerror("list_getarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_getarray");
    v = car(v);
    for (i = 1; i < d2; i++) {
	if (!is_cons(v))
	    aerror("list_getarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_getarray");
    v = car(v);
    for (i = 1; i < d3; i++) {
	if (!is_cons(v))
	    aerror("list_getarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_getarray");
    return car(v);
}

static          Sex
list_setarray1(Sex v, Sex x, long d)
{
    long            i;
    for (i = 1; i < d; i++) {
	if (!is_cons(v))
	    aerror("list_setarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_setarray");
    car(v) = x;
    return x;
}

static          Sex
list_setarray2(Sex v, Sex x, long d1, long d2)
{
    long            i;
    for (i = 1; i < d1; i++) {
	if (!is_cons(v))
	    aerror("list_setarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_setarray");
    v = car(v);
    for (i = 1; i < d2; i++) {
	if (!is_cons(v))
	    aerror("list_setarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_setarray");
    car(v) = x;
    return x;
}

static          Sex
list_setarray3(Sex v, Sex x, long d1, long d2, long d3)
{
    long            i;
    for (i = 1; i < d1; i++) {
	if (!is_cons(v))
	    aerror("list_setarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_setarray");
    v = car(v);
    for (i = 1; i < d2; i++) {
	if (!is_cons(v))
	    aerror("list_setarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_setarray");
    v = car(v);
    for (i = 1; i < d3; i++) {
	if (!is_cons(v))
	    aerror("list_setarray");
	v = cdr(v);
    }
    if (!is_cons(v))
	aerror("list_setarray");
    car(v) = x;
    return x;
}

Sex
Lgetarray1(Sex v, Sex n1)
{
    long            d1,
                    hl;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0)
	return aerror1("getarray1", n1);
    if (is_array(v)) {
	p = array_dim(v);

	if (list_length(p) != 1)
	    return aerror1("getarray not 1d array ", p);
	hl = mkint(car(p));
	if (d1 >= hl)
	    return aerror1("getarray index range", n1);
	else
	    return array_value(v, d1);
    }
#ifdef ASUKA
    if (is_vect(v)) {
	hl = mkint(matrix_dimx(v));
	if (d1 > hl || d1 == 0)
	    return aerror1("getarray index range", n1);
	return vect_value(v, d1 - 1);
    }
    if (is_binhdr(v)) {
	long            i;
	double          f;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_1)
	    return aerror1("getarray1 not array ", v);
	hl = binlen(v) - HDR;
	switch (type_of_header(i)) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    i = iarray_value(v, d1);
	    return mkfxat(i);
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    i = larray_value(v, d1);
	    return mkfxat(i);
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    f = sarray_value(v, d1);
	    return make_float(f);
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    f = darray_value(v, d1);
	    return make_float(f);
	default:
	    return aerror1("getarray not array ", v);
	}
    }
    if (is_cons(v))
	return list_getarray1(v, d1);
#endif
    return aerror1("getarray", v);
}

Sex
Lgetarray2(Sex v, Sex n1, Sex n2)
{
    long            d1,
                    d2,
                    hl;
    long            sy;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0)
	return aerror2("getarray2", n1, n2);
    if (is_array(v)) {
	p = array_dim(v);

	if (list_length(p) != 2)
	    return aerror1("getarray2 not 2d array", p);
	sy = mkint(cadr(p));
	if (d1 >= mkint(car(p)) || d2 >= sy)
	    return aerror2("getarray2 index range", n1, n2);
	d1 = d1 * sy + d2;
	return array_value(v, d1);
    }
#ifdef ASUKA
    if (is_matrix(v)) {
	sy = mkint(matrix_dimy(v));
	if (d1 > mkint(matrix_dimx(v)) || d2 > sy || d1 == 0 || d2 == 0)
	    return aerror2("getarray2 index range", n1, n2);
	d1 = d1 * sy + d2 - sy - 1;
	return matrix_value(v, d1);
    }
    if (is_binhdr(v)) {
	long            i;
	double          f;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_2)
	    return aerror1("getarray2 not array ", v);
	sy = barray_dimy(v);
	if (d2 >= sy)
	    return aerror2("getarray2 index range", n1, n2);
	hl = binlen(v) - HDR - SEX;
	hl = hl / sy;
	switch (type_of_header(i)) {
	case T_IARRAY2:
	    hl = hl / sizeof(int);
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    d1 = d1 * sy + d2;
	    i = iarray2_value(v, d1);
	    return mkfxat(i);
	case T_LARRAY2:
	    hl = hl / sizeof(long);
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    d1 = d1 * sy + d2;
	    i = larray2_value(v, d1);
	    return mkfxat(i);
	case T_SARRAY2:
	    hl = hl / sizeof(float);
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    d1 = d1 * sy + d2;
	    f = sarray2_value(v, d1);
	    return make_float(f);
	case T_DARRAY2:
	    hl = hl / sizeof(double);
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    d1 = d1 * sy + d2;
	    f = darray2_value(v, d1);
	    return make_float(f);
	default:
	    return aerror1("getarray2 not array ", v);
	}
    }
    if (is_cons(v))
	return list_getarray2(v, d1, d2);
#endif
    return aerror1("getarray2", v);
}

Sex
Lgetarray3(Sex v, Sex n1, Sex n2, Sex n3)
{
    long            d1,
                    d2,
                    d3,
                    hl;
    long            sy,
                    sz;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0
	|| !is_fixnum(n3) || (d3 = mkint(n3) - A_OFF) < 0)
	return aerror3("getarray3", n1, n2, n3);
    if (is_array(v)) {
	p = array_dim(v);

	if (list_length(p) != 3)
	    return aerror1("getarray3 not a 3d array ", p);
	sy = mkint(cadr(p));
	sz = mkint(cadr(cdr(p)));
	if (d2 >= sy || d3 >= sz)
	    return aerror2("getarray3 index range", n2, n3);
	hl = mkint(car(p));
	if (d1 >= mkint(car(p)))
	    return aerror1("getarray3 index range", n1);
	d1 = (d1 * sy + d2) * sz + d3;
	return array_value(v, d1);
    }
#ifdef ASUKA
    if (is_binhdr(v)) {
	long            i;
	double          f;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_3)
	    return aerror1("getarray3 not array ", v);
	hl = binlen(v) - HDR - SEX * 2;
	sy = barray_dimy(v);
	sz = barray_dimz(v);
	if (d2 >= sy || d3 >= sz)
	    return aerror3("getarray3 index range", n1, n2, n3);
	switch (type_of_header(i)) {
	case T_IARRAY3:
	    hl = hl / sizeof(int);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray3 index range", n1, n2);
	    }
	    i = iarray3_value(v, d1);
	    return mkfxat(i);
	case T_LARRAY3:
	    hl = hl / sizeof(long);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray3 index range", n1, n2);
	    }
	    i = larray3_value(v, d1);
	    return mkfxat(i);
	case T_SARRAY3:
	    hl = hl / sizeof(float);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray3 index range", n1, n2);
	    }
	    f = sarray3_value(v, d1);
	    return make_float(f);
	case T_DARRAY3:
	    hl = hl / sizeof(double);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray3 index range", n1, n2);
	    }
	    f = darray3_value(v, d1);
	    return make_float(f);
	default:
	    return aerror1("getarray3 not array ", v);
	}
    }
    if (is_cons(v))
	return list_getarray3(v, d1, d2, d3);
#endif
    return aerror1("getarray3", v);
}

Sex
getarray_fn(Sex arg, Sex env)
{
    long            n;
    Sex             a1,
                    a2,
                    a3;
    Sex             v;

    stackcheck();
    n = list_length(arg);

    push2(arg, env);
    push4(nil, nil, nil, nil);
#define arg stack[-5]
#define env stack[-4]
#define vv stack[-3]
#define n1 stack[-2]
#define n2 stack[-1]
#define n3 stack[0]

    if (n >= 2 && n <= 4) {
	vv = eval(car(arg), env);
	arg = cdr(arg);
	switch (n) {
	case 2:
	    a1 = eval(car(arg), env);
	    v = vv;
	    popv(6);
	    return Lgetarray1(v, a1);
	case 3:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    a2 = eval(car(arg), env);
	    a1 = n1;
	    v = vv;
	    popv(6);
	    return Lgetarray2(v, a1, a2);
	case 4:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    n2 = eval(car(arg), env);
	    arg = cdr(arg);
	    a3 = eval(car(arg), env);
	    a2 = n2;
	    a1 = n1;
	    v = vv;
	    popv(6);
	    return Lgetarray3(v, a1, a2, a3);
	}
    }
#undef arg
#undef env
#undef vv
#undef n1
#undef n2
#undef n3
    return nil;
}

Sex
Lsetarray1(Sex v, Sex x, Sex n1)
{
    long            d1,
                    hl;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0)
	return aerror1("setarray1", n1);
    if (is_array(v)) {
	p = array_dim(v);
	if (list_length(p) != 1)
	    return aerror1("setarray not a one dimensional array", v);
	hl = mkint(car(p));
	if (d1 >= hl) {
	    return aerror1("setarray index range", n1);
	}
	array_value(v, d1) = x;
	return x;
    }
#ifdef ASUKA
    if (is_vect(v)) {
	hl = mkint(vect_dimx(v));
	if (d1 > hl || d1 == 0)
	    return aerror1("setarray index range", n1);
	vect_value(v, d1 - 1) = x;
	return x;
    }
    if (is_binhdr(v)) {
	long            i;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_1)
	    return aerror1("setarray not vector ", v);
	hl = binlen(v) - HDR;
	switch (type_of_header(i)) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    iarray_value(v, d1) = mkint(x);
	    return x;
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    larray_value(v, d1) = mkint(x);
	    return x;
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    sarray_value(v, d1) = (float) floatit(x);
	    return x;
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    darray_value(v, d1) = floatit(x);
	    return x;
	default:
	    return aerror1("setarray not vector ", v);
	}
    }
    if (is_cons(v))
	return list_setarray1(v, x, d1);
#endif
    return aerror1("setarray not vector ", v);
}


Sex
Lsetarray2(Sex v, Sex x, Sex n1, Sex n2)
{
    long            d1,
                    d2,
                    hl;
    long            sy;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0)
	return aerror2("setarray2", n1, n2);
    if (is_array(v)) {
	p = array_dim(v);
	if (list_length(p) != 2)
	    return aerror1("setarray not a two dimensional array", v);
	hl = mkint(car(p));
	sy = mkint(cadr(p));
	if (d1 >= hl || d2 >= sy) {
	    return aerror2("setarray index range", n1, n2);
	}
	d1 = d1 * sy + d2;
	array_value(v, d1) = x;
	return x;
    }
#ifdef ASUKA
    if (is_matrix(v)) {
	sy = mkint(matrix_dimy(v));
	hl = mkint(matrix_dimx(v));
	if (d1 == 0 || d2 == 0 || d1 > hl || d2 > sy)
	    return aerror2("setarray index range", n1, n2);
	d1 = d1 * sy + d2 - sy - 1;
	matrix_value(v, d1) = x;
	return x;
    }
    if (is_binhdr(v)) {
	long            i;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_2)
	    return aerror1("setarray not vector ", v);
	hl = binlen(v) - HDR - SEX;
	sy = barray_dimy(v);
	if (d2 >= sy)
	    return aerror2("setarray index range", n1, n2);
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY2:
	    hl = hl / sizeof(int);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    iarray2_value(v, d1) = mkint(x);
	    return x;
	case T_LARRAY2:
	    hl = hl / sizeof(long);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    larray2_value(v, d1) = mkint(x);
	    return x;
	case T_SARRAY2:
	    hl = hl / sizeof(float);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    sarray2_value(v, d1) = (float) floatit(x);
	    return x;
	case T_DARRAY2:
	    hl = hl / sizeof(double);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    darray2_value(v, d1) = floatit(x);
	    return x;
	default:
	    return aerror1("setarray not vector ", v);
	}
    }
    if (is_cons(v))
	return list_setarray2(v, x, d1, d2);
#endif
    return aerror1("setarray not vector ", v);
}


Sex
Lsetarray3(Sex v, Sex x, Sex n1, Sex n2, Sex n3)
{
    long            d1,
                    d2,
                    d3,
                    hl;
    long            sy,
                    sz;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0
	|| !is_fixnum(n3) || (d3 = mkint(n3) - A_OFF) < 0)
	return aerror3("setarray3", n1, n2, n3);
    if (is_array(v)) {
	p = array_dim(v);
	if (list_length(p) != 3)
	    return aerror1("setarray not a three dimensional array", v);
	hl = mkint(car(p));
	sy = mkint(cadr(p));
	sz = mkint(cadr(cdr(p)));
	if (d1 >= hl || d2 >= sy || d3 >= sz)
	    return aerror3("setarray index range", n1, n2, n3);
	d1 = (d1 * sy + d2) * sz + d3;
	array_value(v, d1) = x;
	return x;
    }
#ifdef ASUKA
    if (is_binhdr(v)) {
	long            i;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_3)
	    return aerror1("setarray3 not vector ", v);
	sy = barray_dimy(v);
	sz = barray_dimz(v);
	if (d2 >= sy || d3 >= sz)
	    return aerror3("setarray index range", n1, n2, n3);
	hl = binlen(v) - HDR - SEX * 2;
	switch (type_of_header(i)) {
	case T_IARRAY3:
	    hl = hl / sizeof(int);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    iarray3_value(v, d1) = mkint(x);
	    return x;
	case T_LARRAY3:
	    hl = hl / sizeof(long);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    larray3_value(v, d1) = mkint(x);
	    return x;
	case T_SARRAY3:
	    hl = hl / sizeof(float);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    sarray3_value(v, d1) = (float) floatit(x);
	    return x;
	case T_DARRAY3:
	    hl = hl / sizeof(double);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    darray3_value(v, d1) = floatit(x);
	    return x;
	default:
	    return aerror1("setarray not vector ", v);
	}
    }
    if (is_cons(v))
	return list_setarray3(v, x, d1, d2, d3);
#endif
    return aerror1("setarray not vector ", v);
}



Sex
setarray_fn(Sex arg, Sex env)
{
    long            n;
    Sex             a1,
                    a2,
                    a3;
    Sex             b,
                    c;

    stackcheck();
    n = list_length(arg);
    push2(arg, env);
    push5(nil, nil, nil, nil, nil);
#define arg stack[-6]
#define env stack[-5]
#define v stack[-4]
#define x stack[-3]
#define n1 stack[-2]
#define n2 stack[-1]
#define n3 stack[0]
    if (n >= 3 && n <= 5) {
	v = eval(car(arg), env);
	arg = cdr(arg);
	x = eval(car(arg), env);
	arg = cdr(arg);
	switch (n) {
	case 3:
	    a1 = eval(car(arg), env);
	    b = v;
	    c = x;
	    popv(7);
	    return Lsetarray1(b, c, a1);
	case 4:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    a2 = eval(car(arg), env);
	    a1 = n1;
	    b = v;
	    c = x;
	    popv(7);
	    return Lsetarray2(b, c, a1, a2);
	case 5:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    n2 = eval(car(arg), env);
	    arg = cdr(arg);
	    a3 = eval(car(arg), env);
	    a2 = n2;
	    a1 = n1;
	    b = v;
	    c = x;
	    popv(7);
	    return Lsetarray3(b, c, a1, a2, a3);
	}
    }
#undef arg
#undef env
#undef v
#undef x
#undef n1
#undef n2
#undef n3
    return nil;
}

Sex
Lputv(Sex v, Sex n, Sex x)
{
    long            n1,
                    hl;

    if (!is_fixnum(n))
	return aerror1("putv offset not fixnum", n);
    n1 = mkint(n);
    if (is_simple_vector(v)) {
	hl = bnum_n(veclen(v));
	if (n1 < 0 || n1 >= hl) {
	    return aerror1("putv index range", n);
	}
	vector_value(v, n1) = x;
	return x;
    }
#if 0
    if (is_binhdr(v)) {
	hl = binlen(v) - HDR;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (n1 < 0 || n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    iarray_value(v, n1) = mkint(x);
	    return x;
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (n1 < 0 || n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    larray_value(v, n1) = mkint(x);
	    return x;
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (n1 < 0 || n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    sarray_value(v, n1) = (float) floatit(x);
	    return x;
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (n1 < 0 || n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    darray_value(v, n1) = floatit(x);
	    return x;
	default:
	    return aerror1("putv not vector ", v);
	}
    }
#endif
    return aerror1("putv not vector ", v);
}

Sex
Laref(Sex v, Sex n)
{
    Sex             m;
    long            n1,
                    hl;
    while (is_cons(n)) {
	m = car(n);
	if (!is_simple_vector(v))
	    return aerror1("aref", v);
	else if (!is_fixnum(m))
	    return aerror1("aref offset not fixnum", m);
	hl = bnum_n(veclen(v));
	n1 = mkint(m);
	if (n1 < 0 || n1 >= hl)
	    return aerror1("aref index range", m);
	else
	    v = vector_value(v, n1);
	n = cdr(n);
    }
    return v;
}

Sex
Lgetv(Sex v, Sex n)
{
    long            n1,
                    hl;
    long            i;
    double          f;

    if (!is_fixnum(n) || (n1 = mkint(n)) < 0)
	return aerror1("getv", n);

    if (is_simple_vector(v)) {
	hl = bnum_n(veclen(v));
	if (n1 >= hl)
	    return aerror1("getv index range", n);
	else
	    return vector_value(v, n1);
    }
#if 0

    if (is_binhdr(v)) {
	hl = binlen(v) - HDR;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    i = iarray_value(v, n1);
	    return mkfxat(i);
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    i = larray_value(v, n1);
	    return mkfxat(i);
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    f = sarray_value(v, n1);
	    return make_float(f);
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    f = darray_value(v, n1);
	    return make_float(f);
	default:
	    return aerror1("getv not vector ", v);
	}
    }
#endif
    return aerror1("getv", v);
}


Sex
Lupbv(Sex v)
{
    Tag             h;
    long            n;
    if ((is_simple_vector(v))) {
	h = vechdr(v);
	n = bnum_n(veclen(v)) - 1;
	return mkfxat(n);
    }
#if 0
    if (is_binhdr(v)) {
	n = binlen(v) - HDR;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY:
	    n = n / sizeof(int);
	    return mkfxat(n);
	case T_LARRAY:
	    n = n / sizeof(long);
	    return mkfxat(n);
	case T_SARRAY:
	    n = n / sizeof(float);
	    return mkfxat(n);
	case T_DARRAY:
	    n = n / sizeof(double);
	    return mkfxat(n);
	default:
	    return nil;
	}
    }
#endif
    return nil;
}

/*******************************************************/
/*******************************************************/
/*******************************************************/

Sex
Lmkvector(Sex a)
{
    long            i;
    long            n;
    Sex             p;
    Sex            *m;

    n = mkint(a);
    if (n < 0 || n > 0x0fffffff)
	return aerror1("mkvector", a);

    p = getvector(TAG_VECTOR, T_VECT,
		  sizeof(Vector) + (n - 1) * sizeof(Sex));
    vect_dimx(p) = mkfxat(n);
    m = &vect_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKVECTOR_INIT_VALUE;
    }
    return p;
}

Sex
Lmkmatrix(Sex a, Sex b)
{
    long            i;
    long            n;
    long            dim[2];
    Sex             p;
    Sex            *m;

    dim[0] = mkint(a);
    dim[1] = mkint(b);
    n = dim[0] * dim[1];
    if (n < 0 || n > 0x0fffffff)
	return aerror2("mkmatrix", a, b);

    p = getvector(TAG_VECTOR, T_MATRIX,
		  sizeof(Matrix) + (n - 1) * sizeof(Sex));
    matrix_dimx(p) = mkfxat(dim[0]);
    matrix_dimy(p) = mkfxat(dim[1]);
    m = &matrix_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKMATRIX_INIT_VALUE;
    }
    return p;
}


/*******************************************************/
/*******************************************************/
/*******************************************************/

bool
eql(Sex a, Sex b)
{
    Tag             h;
    long            hh;

  again:
    if (a == b)
	return YES;
    if ((((long) a ^ (long) b) & TAG_BITS) != 0)
	return NO;
    switch ((long) a & TAG_BITS) {
    case TAG_NUMBERS:
	if ((h = numhdr(a)) != numhdr(b) || (hh = numlen(a)) != numlen(b))
	    return NO;
	switch (numhdr_of(h)) {
	case N_FLOAT:
	case N_BIGNUM:
	case N_EFLOAT:
	case N_INTERVAL:
	case N_ECOMPLEX:
	    hh = hh - TAG_NUMBERS;
	    while (hh > (HDR - TAG_NUMBERS)) {
		hh -= SEX;
		if (*(Sex *) ((char *) a + hh) !=
		    *(Sex *) ((char *) b + hh))
		    return NO;
	    }
	    return YES;
	case N_BFLOAT:
	    if (numvec(a, 0) != numvec(b, 0))
		return NO;
	    a = numvec(a, 1);
	    b = numvec(b, 1);
	    goto again;
	case N_BEFLOAT:
	    if (numvec(a, 1) != numvec(b, 1))
		return NO;
	    a = numvec(a, 0);
	    b = numvec(b, 0);
	    goto again;
	case N_BECOMPLEX:
	    if (eql(numvec(a, 2), numvec(b, 2)) == NO)
		return NO;
	case N_RATIONAL:
	case N_COMPLEX:
	case N_BINTERVAL:
	    if (eql(numvec(a, 1), numvec(b, 1)) == NO)
		return NO;
	case N_MODULAR:
	    a = numvec(a, 0);
	    b = numvec(b, 0);
	    goto again;
	}
	return NO;
    default:
	return NO;
    }
}

static          Sex
numequal(Sex a, Sex b)
{
    Tag             ha;
    long            l;

  again:
    if (a == b)
	return true;
    if (is_fixnum(a)) {
	if (!is_fixnum(b))
	    return nil;
	return Bool_to_Sex(a == b);
    } else if (is_fixnum(b))
	return nil;


    if ((ha = numhdr(a)) != numhdr(b))
	return nil;
    switch (numhdr_of(ha)) {
    case N_BFLOAT:
	if (numvec(a, 0) != numvec(b, 0))
	    return nil;
	a = numvec(a, 1);
	b = numvec(b, 1);
	goto again;
    case N_BEFLOAT:
	if (numvec(a, 1) != numvec(b, 1))
	    return nil;
    case N_MODULAR:
	a = numvec(a, 0);
	b = numvec(b, 0);
	goto again;
    case N_BECOMPLEX:
	if (numequal(numvec(a, 2), numvec(b, 2)) == nil)
	    return nil;
    case N_RATIONAL:
    case N_COMPLEX:
    case N_BINTERVAL:
	if (numequal(numvec(a, 1), numvec(b, 1)) == nil)
	    return nil;
	a = numvec(a, 0);
	b = numvec(b, 0);
	goto again;
    default:
	if ((l = numlen(a)) != numlen(b))
	    return nil;
	l = bnum_n(l);
	while ((l--) != 0)
	    if (numvec(a, l) != numvec(b, l))
		return nil;
	return true;

    }
}

static          Sex
strequal(Sex a, Sex b)
{
    Tag             ha = binhdr(a),
	hb = binhdr(b);
    long            l;

    if (ha != hb || (l = binlen(a)) != binlen(b))
	return nil;
    l = bnum_n(l + SEX - 1);
    while ((l--) != 0)
	if (object_value(a, l) != object_value(b, l))
	    return nil;
    return true;
}

static          Sex
vecequal(Sex a, Sex b)
{
    Tag             ha = vechdr(a),
	hb = vechdr(b);
    long            l;

    if (ha != hb || (l = veclen(a)) != veclen(b))
	return nil;
    l = bnum_n(l + SEX - 1);
    while ((l--) != 0) {
	Sex             ea = vector_value(a, l);
	Sex             eb = vector_value(b, l);
	if (ea == eb)
	    continue;
	if (equal(ea, eb) == nil)
	    return nil;
    }
    return true;
}



Sex
equal(Sex a, Sex b)
{
    long            ta;

    while (YES) {
	if ((((long) a ^ (long) b) & TAG_BITS) != 0)
	    return nil;
	switch (ta = (long) a & TAG_BITS) {
	case TAG_CONS:
	    if (!is_cons(b))
		return nil;
	    else {
		Sex             ca = car(a),
		    cb = car(b);
		if (ca == cb) {
		    a = cdr(a);
		    b = cdr(b);
		    if (a == b)
			return true;
		    continue;
		}
		if (equal(ca, cb) != true)
		    return nil;
		a = cdr(a);
		b = cdr(b);
		if (a == b)
		    return true;
		continue;
	    }
	case TAG_FIXNUM:
	case TAG_HDRS:
	case TAG_SYMBOL:
	    return nil;
	case TAG_NUMBERS:
	    return numequal(a, b);
	case TAG_VECTOR:
	    return vecequal(a, b);
	case TAG_BINARY:
	    return strequal(a, b);
	}
    }
}


Sex
Leql(Sex a, Sex b)
{
    if (a == b)
	return true;
    return Bool_to_Sex(eql(a, b));
}

Sex
Leqcar(Sex a, Sex b)
{
    if (!is_cons(a) || car(a) != b)
	return nil;
    return true;
}

Sex
Lequal(Sex a, Sex b)
{
    if (a == b)
	return true;
    else
	return equal(a, b);
}

Sex
Lneq(Sex a, Sex b)
{
    if (a == b)
	return nil;
    if (equal(a, b) == nil)
	return true;
    return nil;
}

Sex
Lnull(Sex a)
{
    return Bool_to_Sex(a == nil);
}

Sex
Lendp(Sex a)
{
    if (a == nil)
	return true;
    else if (is_cons(a))
	return nil;
    else
	return error1("endp: bad argument", a);
}


Sex
Lreverse(Sex a)
{
    Sex             r;
    stackcheck();
    r = nil;
    push(nil);
    while (is_cons(a)) {
	stack[0] = cdr(a);
	r = cons(car(a), r);
	a = stack[0];
    }
    popv(1);
    return r;
}


Sex
Lassoc(Sex a, Sex b)
{
    if (is_symhdr(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    Sex             c = car(b);
	    if (is_cons(c) && a == car(c))
		return c;
	    b = cdr(b);
	}
	return nil;
    }
    while (is_cons(b)) {
	Sex             c = car(b);
	if (is_cons(c)) {
	    Sex             cc = car(c);
	    if (a == cc || equal(a, cc) != nil)
		return c;
	}
	b = cdr(b);
    }
    return nil;
}

Sex
Lrassoc(Sex a, Sex b)
{
    while (is_cons(b)) {
	Sex             c = car(b);
	if (is_cons(c)) {
	    Sex             cc = cdr(c);
	    if (a == cc || equal(a, cc) != nil)
		return c;
	}
	b = cdr(b);
    }
    return nil;
}

Sex
Latsoc(Sex a, Sex b)
{
    Sex             c;

    while (is_cons(b)) {
	c = car(b);
	if (is_cons(c) && a == car(c))
	    return c;
	b = cdr(b);
    }
    return nil;
}

Sex
Lmember(Sex a, Sex b)
{
    if (is_symhdr(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    if (a == car(b))
		return b;
	    b = cdr(b);
	}
	return nil;
    }
    while (is_cons(b)) {
	Sex             cb = car(b);
	if (a == cb || equal(a, cb) != nil)
	    return b;
	b = cdr(b);
    }
    return nil;
}


Sex
Lmemq(Sex a, Sex b)
{
    while (is_cons(b)) {
	if (a == car(b))
	    return b;
	b = cdr(b);
    }
    return nil;
}


Sex
Lsetdiff(Sex a, Sex b)
{
    Sex             r = nil;
    Sex             c;
    stackcheck();
    push2(a, b);
    push(r);
    while (is_cons(stack[-2])) {
	c = Lmember(car(stack[-2]), stack[-1]);
	if (c == nil) {
	    stack[0] = cons(car(stack[-2]), stack[0]);
	}
	stack[-2] = cdr(stack[-2]);
    }
    pop(r);
    popv(2);
    return Lnreverse(r);
}

Sex
Lsetdiffq(Sex a, Sex b)
{
    Sex             r = nil;
    Sex             c;
    stackcheck();
    push2(a, b);
    push(r);
    while (is_cons(stack[-2])) {
	c = Lmemq(car(stack[-2]), stack[-1]);
	if (c == nil) {
	    stack[0] = cons(car(stack[-2]), stack[0]);
	}
	stack[-2] = cdr(stack[-2]);
    }
    pop(r);
    popv(2);
    return Lnreverse(r);
}

Sex
Lsetadd(Sex a, Sex b)
{
    if (Lmember(a, b) == nil)
	return cons(a, b);
    return b;
}

Sex
Lsetaddq(Sex a, Sex b)
{
    Sex             c;
    if (Lmemq(a, b) == nil)
	return cons(a, b);
    return b;
}

static          bool
smemq(Sex a, Sex b)
{
    Sex             w;
    while (is_cons(b)) {
	w = car(b);
	if (w == quote_symbol)
	    return NO;
	else if (smemq(a, w))
	    return YES;
	else
	    b = cdr(b);
    }
    return (a == b);
}

Sex
Lsmemq(Sex a, Sex b)
{
    return Bool_to_Sex(smemq(a, b));
}


long
list_length(Sex a)
{
    long            n;
    n = 0;
    while (is_cons(a)) {
	a = cdr(a);
	n++;
    }
    return n;
}


Sex
Llength(Sex a)
{
    long            n;
    n = 0;
    if (a != nil && !is_cons(a))
	error("length: not a list");
    while (is_cons(a)) {	/* 
				 * nil is not a cons  
				 */
	a = cdr(a);
	n++;
    }
    return mkfxat(n);
}


Sex
Lappend(Sex a, Sex b)
{
    Sex             r = nil;

    stackcheck();
    push(b);
    push(nil);
    while (is_cons(a)) {
	stack[0] = cdr(a);
	r = cons(car(a), r);
	a = stack[0];
    }
    popv(1);
    pop(b);
    while (r != nil) {
	a = cdr(r);
	cdr(r) = b;
	b = r;
	r = a;
    }
    return b;
}



Sex
Ldelete(Sex a, Sex b)
{
    Sex             r;
    Sex             q;
    stackcheck();
    push2(a, b);
    r = nil;
    if (is_symhdr(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    q = car(b);
	    if (q == stack[-1]) {
		b = cdr(b);
		break;
	    }
	    stack[0] = cdr(b);
	    r = cons(car(b), r);
	    b = stack[0];
	}
    } else {
	while (is_cons(b)) {
	    q = car(b);
	    if (q == a || equal(q, a) != nil) {
		b = cdr(b);
		break;
	    }
	    stack[0] = cdr(b);
	    r = cons(car(b), r);
	    b = stack[0];
	    a = stack[-1];
	}
    }
    popv(2);
    while (r != nil) {
	q = cdr(r);
	cdr(r) = b;
	b = r;
	r = q;
    }
    return b;
}


Sex
Ldeleq(Sex a, Sex b)
{
    Sex             r;
    Sex             q;
    stackcheck();
    push2(a, b);
    r = nil;
    while (is_cons(b)) {
	q = car(b);
	if (q == stack[-1]) {
	    b = cdr(b);
	    break;
	}
	stack[0] = cdr(b);
	r = cons(car(b), r);
	b = stack[0];
    }
    popv(2);
    while (r != nil) {
	q = cdr(r);
	cdr(r) = b;
	b = r;
	r = q;
    }
    return b;
}

Sex
Laconc(Sex a, Sex b)
{
    Sex             c;
    Sex             next;

    stackcheck();
    push(a);
    b = cons(b, nil);
    pop(a);
    if (!is_cons(a))
	return b;
    c = a;
    while (YES) {
	next = cdr(c);
	if (!is_cons(next)) {
	    cdr(c) = b;
	    return a;
	} else {
	    c = next;
	}
    }
}


Sex
Lnconc(Sex a, Sex b)
{
    Sex             c;
    Sex             next;
    if (a == nil)
	return b;
    if (!is_cons(a))
	error("first arg. of nconc is not a list");
    c = a;
    while (YES) {
	next = cdr(c);
	if (next == nil) {
	    cdr(c) = b;
	    return a;
	} else if (!is_cons(next)) {
	    error("first arg. of nconc is not a list");
	} else {
	    c = next;
	}
    }
}


static          Sex
Lsubstq(Sex a, Sex b, Sex c)
{
    Sex             w;
    if (c == b)
	return a;
    if (!is_cons(c)) {
	return c;
    }
    stackcheck();
    push3(a, b, c);
    w = Lsubstq(a, b, car(c));
    pop2(c, b);
    a = stack[0];
    stack[0] = w;
    w = Lsubstq(a, b, cdr(c));
    pop(a);
    return cons(a, w);
}

Sex
Lsubst(Sex a, Sex b, Sex c)
{
    Sex             w;
    if (c == b)
	return a;
    if (is_symhdr(b) || is_fixnum(b))
	return Lsubstq(a, b, c);
    stackcheck();
    push3(a, b, c);
    if (c == b || equal(c, b) != nil) {
	popv(2);
	pop(a);
	return a;
    }
    if (!is_cons(stack[0])) {
	pop(c);
	popv(2);
	return c;
    }
    w = Lsubst(stack[-2], stack[-1], car(stack[0]));
    pop2(c, b);
    a = stack[0];
    stack[0] = w;
    w = Lsubst(a, b, cdr(c));
    pop(a);
    return cons(a, w);
}


Sex
Lsublis(Sex al, Sex x)
{
    Sex             a1,
                    a2;
    stackcheck();
    push5(al, x, al, nil, nil);
#define carx stack[0]
#define cdrx stack[-1]
#define w    stack[-2]
#define x    stack[-3]
#define al   stack[-4]
    while (YES) {
	if (!is_cons(w)) {
	    if (!is_cons(x)) {
		a1 = x;
		popv(5);
		return a1;
	    }
	    carx = Lsublis(al, car(x));
	    cdrx = Lsublis(al, cdr(x));
	    if (carx == car(x) && cdrx == cdr(x)) {
		a1 = x;
		popv(5);
		return a1;
	    } else {
		a1 = cons(carx, cdrx);
		popv(5);
		return a1;
	    }
	}
	a1 = car(w);
	if (is_cons(a1)) {
	    a2 = car(a1);
	    if (a2 == x || equal(a2, x) != nil) {
		a1 = cdr(a1);
		popv(5);
		return a1;
	    }
	}
	w = cdr(w);
    }
}
#undef carx
#undef cdrx
#undef w
#undef x
#undef al

Sex
Lsubla(Sex al, Sex x)
{
    Sex             a1;

    stackcheck();
    push5(al, x, al, nil, nil);
#define carx stack[0]
#define cdrx stack[-1]
#define w    stack[-2]
#define x    stack[-3]
#define al   stack[-4]
    while (YES) {
	if (!is_cons(w)) {
	    if (!is_cons(x)) {
		a1 = x;
		popv(5);
		return a1;
	    }
	    carx = Lsubla(al, car(x));
	    cdrx = Lsubla(al, cdr(x));
	    if (carx == car(x) && cdrx == cdr(x)) {
		a1 = x;
		popv(5);
		return a1;
	    } else {
		a1 = cons(carx, cdrx);
		popv(5);
		return a1;
	    }
	}
	a1 = car(w);
	if (is_cons(a1)) {
	    if (car(a1) == x) {
		a1 = cdr(a1);
		popv(5);
		return a1;
	    }
	}
	w = cdr(w);
    }
}
#undef carx
#undef cdrx
#undef w
#undef x
#undef al

Sex
Lquit()
{
    exit(0);
    return nil;
}

Sex
Lbye()
{
    exit(0);
    return nil;
}


#ifdef INCLUDE_FEND

/*****************************************************************************/
/*
 * Standard lisp functions.
 */
/*****************************************************************************/



/*
 * Code for deflist 
 */

Sex
Ldeflist(Sex u, Sex ind)
{
    Sex             x,
                    y;

    /*
     * if (compile_flag) return nil; 
     */
    if (u == nil)
	return nil;
    stackcheck();
    push3(u, ind, nil);
#define u stack[-2]
#define ind stack[-1]
#define v stack[0]
    // lisp_printf("deflist: %p, %p\n", u, ind);
    v = car(u);
    Lput(car(v), ind, car(cdr(v)));
    x = Ldeflist(cdr(u), ind);
    y = car(v);
    popv(3);
    return cons(y, x);
#undef v
#undef u
#undef ind
}


Sex
Lexpand(Sex l, Sex fn)
{
    if (l == nil)
	return nil;
    else if (cdr(l) == nil)
	return car(l);
    else {
	Sex             r;
	stackcheck();
	push2(fn, l);
	r = Lexpand(cdr(l), fn);
	pop2(l, fn);
	return Llist3(fn, car(l), r);
    }
}


static          Sex
logxor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = logxor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
logand_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = logand2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
logor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = logor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
ilogxor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lilogxor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
ilogand_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lilogand2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
ilogor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lilogor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}


static          Sex
max_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lmax2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
min_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lmin2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}


static          Sex
imax_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Limax2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
imin_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (min) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Limin2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}




/*
 * Code for copy 
 */

Sex
Lcopy(Sex u)
{
    Sex             v;

    if (!is_cons(u))
	return u;
    stackcheck();
    push2(nil, nil);
    stack[-1] = cdr(u);
    stack[0] = Lcopy(car(u));
    u = Lcopy(stack[-1]);
    v = stack[0];
    popv(2);
    return cons(v, u);
}

/*
 * toplevel list or array copy 
 */

Sex
Lclone(Sex u)
{
    Sex             p;
    if (!is_cons(u))
	return array_copy(u);
    stackcheck();
    push3(nil, nil, u);
#define p  stack[-2]
#define pp stack[-1]
#define u  stack[ 0]
    p = cons(nil, nil);
    pp = p;
    while (is_cons(u)) {
	p = rcons(p, car(u));
	u = cdr(u);
    }
#undef p
    p = cdr(pp);
#undef pp
#undef u
    popv(3);
    return p;
}

Sex
array_copy(Sex u)
{
    long            len;
    long            i;
    Sex             obj;

    switch ((long) (u) & TAG_BITS) {
    case TAG_VECTOR:
	len = veclen(u);
	push(u);
	obj = getvector(TAG_VECTOR, T_SIMPLE, len);
	pop(u);
	vechdr(obj) = vechdr(u);
	len = (len - HDR) / SEX;
	for (i = 0; i < len; i++)
	    vector_value(obj, i) = vector_value(u, i);
	return obj;
    case TAG_BINARY:
	len = binlen(u);
	push(u);
	obj = getvector(TAG_BINARY, T_IARRAY, len);
	pop(u);
	binhdr(obj) = binhdr(u);
	len = (len - HDR) / SEX;
	for (i = 0; i < len; i++)
	    larray_value(obj, i) = larray_value(u, i);
	return obj;
    default:
	return error1("array_copy", u);
    }
}



/*
 * Code for xn 
 */

Sex
Lxn(Sex v, Sex u)
{
    Sex             w;
    stackcheck();
    push(v);
#define v stack[0]
    while (v != nil) {
	if (Lmember(car(v), u) == nil) {
	    v = cdr(v);
	    continue;
	}
	w = Ldelete(car(v), u);
	w = Lxn(cdr(v), w);
	u = car(v);
	popv(1);
	return cons(u, w);
    }
    popv(1);
    return nil;
#undef v
}





Sex
Llast(Sex u)
{
    while (is_cons(u)) {
	if (cdr(u) == nil)
	    return car(u);
	u = cdr(u);
    }
    return u;
}





Sex
Lnum(Sex n)
{
    if (is_numhdr(n) && type_of_header(numhdr(n)) == T_RATIONAL)
	return rational_num(n);
    else
	return error("not a rational number");
}

Sex
Lden(Sex n)
{
    if (is_numhdr(n) && type_of_header(numhdr(n)) == T_RATIONAL)
	return rational_den(n);
    else
	return error("not a rational number");
}





Sex
Lprintprompt(Sex u)
{
    return Lsetpchar(u);
}

#endif

Sex
Lprintaddr(Sex x)
{
    return mkfxat(x);
}

#ifdef ASUKA
Sex
Lmode_lisp()
{
    current_package = value(current_sugar_package);
    current_charvec = package_charvec(current_package);
    return nil;
}
Sex
Lmode_asuka()
{
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    return nil;
}
#endif



setup_type const fns_setup[] = {
    {"acons", (lisp_proc *) Lacons, 3},
    {"atom", (lisp_proc *) Latom, 1},
    {"boundp", (lisp_proc *) Lboundp, 1},
    {"car", (lisp_proc *) Lcar, 1},
    {"cdr", (lisp_proc *) Lcdr, 1},
    {"caar", (lisp_proc *) Lcaar, 1},
    {"cadr", (lisp_proc *) Lcadr, 1},
    {"cdar", (lisp_proc *) Lcdar, 1},
    {"cddr", (lisp_proc *) Lcddr, 1},
    {"caaar", (lisp_proc *) Lcaaar, 1},
    {"caadr", (lisp_proc *) Lcaadr, 1},
    {"cadar", (lisp_proc *) Lcadar, 1},
    {"caddr", (lisp_proc *) Lcaddr, 1},
    {"cdaar", (lisp_proc *) Lcdaar, 1},
    {"cdadr", (lisp_proc *) Lcdadr, 1},
    {"cddar", (lisp_proc *) Lcddar, 1},
    {"cdddr", (lisp_proc *) Lcdddr, 1},
    {"caaaar", (lisp_proc *) Lcaaaar, 1},
    {"caaadr", (lisp_proc *) Lcaaadr, 1},
    {"caadar", (lisp_proc *) Lcaadar, 1},
    {"caaddr", (lisp_proc *) Lcaaddr, 1},
    {"cadaar", (lisp_proc *) Lcadaar, 1},
    {"cadadr", (lisp_proc *) Lcadadr, 1},
    {"caddar", (lisp_proc *) Lcaddar, 1},
    {"cadddr", (lisp_proc *) Lcadddr, 1},
    {"cdaaar", (lisp_proc *) Lcdaaar, 1},
    {"cdaadr", (lisp_proc *) Lcdaadr, 1},
    {"cdadar", (lisp_proc *) Lcdadar, 1},
    {"cdaddr", (lisp_proc *) Lcdaddr, 1},
    {"cddaar", (lisp_proc *) Lcddaar, 1},
    {"cddadr", (lisp_proc *) Lcddadr, 1},
    {"cdddar", (lisp_proc *) Lcdddar, 1},
    {"cddddr", (lisp_proc *) Lcddddr, 1},

    {"codep", (lisp_proc *) Lcodep, 1},
    {"bytecodep", (lisp_proc *) Lbytecodep, 1},
    {"constantp", (lisp_proc *) Lconstantp, 1},
    {"date", (lisp_proc *) Ldate, 0},
    {"date-and-time", (lisp_proc *) Ldate_and_time, 0},
    {"datestamp", (lisp_proc *) Ldatestamp, 0},
    {"error", (lisp_proc *) Lerror, 2},
    {"error1", (lisp_proc *) Lerror1, 0},
    {"error2", (lisp_proc *) Lerror2, 1},
    {"numtype", (lisp_proc *) Lnumtype, 1},
    {"gnumtype", (lisp_proc *) Lgnumtype, 1},
    {"fixp", (lisp_proc *) Lfixp, 1},
    {"bigp", (lisp_proc *) Lbigp, 1},
    {"logicalp", (lisp_proc *) Llogicalp, 1},
    {"floatp", (lisp_proc *) Lfloatp, 1},
    {"bfloatp", (lisp_proc *) Lbfloatp, 1},
    {"rationalp", (lisp_proc *) Lrationalp, 1},
    {"complexp", (lisp_proc *) Lcomplexp, 1},
    {"complp", (lisp_proc *) Lcomplexp, 1},
    {"ecomplexp", (lisp_proc *) Lecomplexp, 1},
    {"becomplexp", (lisp_proc *) Lbecomplexp, 1},
    {"efloatp", (lisp_proc *) Lefloatp, 1},
    {"befloatp", (lisp_proc *) Lbefloatp, 1},
    {"intervalp", (lisp_proc *) Lintervalp, 1},
    {"bintervalp", (lisp_proc *) Lbintervalp, 1},
    {"fluidp", (lisp_proc *) Lfluidp, 1},
    {"is_special_var", (lisp_proc *) Lis_special_var, 1},
    {"gctime", (lisp_proc *) Lgctime, 0},
    {"gccalled", (lisp_proc *) Lgccalled, 0},
    {"get", (lisp_proc *) Lget, 2},
    {"globalp", (lisp_proc *) Lglobalp, 1},
    {"list2", (lisp_proc *) Llist2, 2},
    {"list3", (lisp_proc *) Llist3, 3},
    {"list4", (lisp_proc *) Llist4, 4},
    {"list5", (lisp_proc *) Llist5, 5},
    {"global", (lisp_proc *) Lglobal, 1},
    {"fluid", (lisp_proc *) Lfluid, 1},
    {"make_fluid", (lisp_proc *) Lmake_fluid, 1},
    {"make_global", (lisp_proc *) Lmake_global, 1},
    {"mkquote", (lisp_proc *) Lmkquote, 1},
    {"cons", (lisp_proc *) cons, 2},
    {"xcons", (lisp_proc *) Lxcons, 2},
    {"ncons", (lisp_proc *) Lncons, 1},
    {"numberp", (lisp_proc *) Lnumberp, 1},
    {"pair", (lisp_proc *) Lpair, 2},
    {"put", (lisp_proc *) Lput, 3},
    {"remprop", (lisp_proc *) Lremprop, 2},
    {"rplaca", (lisp_proc *) Lrplaca, 2},
    {"rplacd", (lisp_proc *) Lrplacd, 2},
    {"stop", (lisp_proc *) Lstop, 1},
    {"time", (lisp_proc *) Ltime, 0},
    {"float-time", (lisp_proc *) Lfloat_time, 0},
    {"union", (lisp_proc *) Lunion, 2},
    {"unionq", (lisp_proc *) Lunionq, 2},
    {"intersection", (lisp_proc *) Lintersection, 2},
    {"intersectionq", (lisp_proc *) Lintersectionq, 2},
    {"setdiff", (lisp_proc *) Lsetdiff, 2},
    {"setdiffq", (lisp_proc *) Lsetdiffq, 2},
    {"setadd", (lisp_proc *) Lsetadd, 2},
    {"setaddq", (lisp_proc *) Lsetaddq, 2},

    {"unglobal", (lisp_proc *) Lunglobal, 1},
    {"unfluid", (lisp_proc *) Lunfluid, 1},
    {"flag", (lisp_proc *) Lflag, 2},
    {"flagp", (lisp_proc *) Lflagp, 2},
    {"flagpcar", (lisp_proc *) Lflagpcar, 2},
    {"flagp**", (lisp_proc *) Lflagp, 2},
    {"get*", (lisp_proc *) Lget, 2},
    {"idp", (lisp_proc *) Lidp, 1},
    {"pairp", (lisp_proc *) Lpairp, 1},
    {"listp", (lisp_proc *) Llistp, 1},
    {"remflag", (lisp_proc *) Lremflag, 2},
    {"set", (lisp_proc *) Lset, 2},
    {"stringp", (lisp_proc *) Lstringp, 1},
    {"vectorp", (lisp_proc *) Lvectorp, 1},
    {"arrayp", (lisp_proc *) Larrayp, 1},
    {"append", (lisp_proc *) Lappend, 2},
    {"eq", (lisp_proc *) Leq, 2},
    {"eqcar", (lisp_proc *) Leqcar, 2},
    {"eql", (lisp_proc *) Leql, 2},
    {"equal", (lisp_proc *) Lequal, 2},
    {"endp", (lisp_proc *) Lendp, 1},
    {"getd", (lisp_proc *) Lgetd, 1},
    {"movd", (lisp_proc *) Lmovd, 2},
    {"length", (lisp_proc *) Llength, 1},
    {"putd", (lisp_proc *) Lputd, 3},
    {"get_number_of_args", (lisp_proc *) Lget_number_of_args, 1},
    {"is_macro", (lisp_proc *) Lis_macro, 1},
    {"is_special", (lisp_proc *) Lis_special, 1},
    {"remd", (lisp_proc *) Lremd, 1},
    {"trace", (lisp_proc *) Ltrace, 1},
    {"untrace", (lisp_proc *) Luntrace, 1},
    {"symbol-name", (lisp_proc *) Lsymbol_name, 1},
    {"id2string", (lisp_proc *) Lsymbol_name, 1},
    {"symbol-length", (lisp_proc *) Lsymbol_length, 1},
    {"plist", (lisp_proc *) Lplist, 1},
    {"putv", (lisp_proc *) Lputv, 3},
    {"upbv", (lisp_proc *) Lupbv, 1},
    {"memq", (lisp_proc *) Lmemq, 2},
    {"nconc", (lisp_proc *) Lnconc, 2},
    {"aconc", (lisp_proc *) Laconc, 2},
    {"map", (lisp_proc *) Lmap, 2},
    {"mapc", (lisp_proc *) Lmapc, 2},
    {"mapcar", (lisp_proc *) Lmapcar, 2},
    {"maplist", (lisp_proc *) Lmaplist, 2},
    {"mapcan", (lisp_proc *) Lmapcan, 2},
    {"mapcon", (lisp_proc *) Lmapcon, 2},
    {"mapv", (lisp_proc *) Lmapv, 2},
    {"map2", (lisp_proc *) Lmap2, 3},
    {"mapc2", (lisp_proc *) Lmapc2, 3},
    {"mapcar2", (lisp_proc *) Lmapcar2, 3},
    {"maplist2", (lisp_proc *) Lmaplist2, 3},
    {"mapcan2", (lisp_proc *) Lmapcan2, 3},
    {"mapcon2", (lisp_proc *) Lmapcon2, 3},
    {"mapv2", (lisp_proc *) Lmapv2, 3},
    {"neq", (lisp_proc *) Lneq, 2},
    {"not", (lisp_proc *) Lnull, 1},
    {"null", (lisp_proc *) Lnull, 1},
    {"reverse", (lisp_proc *) Lreverse, 1},
    {"smemq", (lisp_proc *) Lsmemq, 2},
    {"subla", (lisp_proc *) Lsubla, 2},
    {"sublis", (lisp_proc *) Lsublis, 2},
    {"delete", (lisp_proc *) Ldelete, 2},
    {"deleq", (lisp_proc *) Ldeleq, 2},
    {"rassoc", (lisp_proc *) Lrassoc, 2},
    {"assoc", (lisp_proc *) Lassoc, 2},
    {"assoc**", (lisp_proc *) Lassoc, 2},
    {"atsoc", (lisp_proc *) Latsoc, 2},
    {"assq", (lisp_proc *) Latsoc, 2},
    {"getv", (lisp_proc *) Lgetv, 2},
    {"aref", (lisp_proc *) Laref, 2},
    {"member", (lisp_proc *) Lmember, 2},
    {"member**", (lisp_proc *) Lmember, 2},
    {"mkvect", (lisp_proc *) Lmkvect, 1},
    {"arraydim", (lisp_proc *) Larraydim, 1},
    {"mkarray", (lisp_proc *) mkarray_fn, -1},
    {"mkarray1", (lisp_proc *) Lmkarray1, 1},
    {"mkarray2", (lisp_proc *) Lmkarray2, 2},
    {"mkarray3", (lisp_proc *) Lmkarray3, 3},
    {"mkiarray", (lisp_proc *) Lmkiarray1, 1},
    {"mkivect", (lisp_proc *) Lmkiarray1, 1},
    {"mkiarray2", (lisp_proc *) Lmkiarray2, 2},
    {"mkiarray3", (lisp_proc *) Lmkiarray3, 3},
    {"mkfvect", (lisp_proc *) Lmkfarray1, 1},
    {"mkfarray2", (lisp_proc *) Lmkfarray2, 2},
    {"mkfarray3", (lisp_proc *) Lmkfarray3, 3},
    {"mkdvect", (lisp_proc *) Lmkdarray1, 1},
    {"mkdarray2", (lisp_proc *) Lmkdarray2, 2},
    {"mkdarray3", (lisp_proc *) Lmkdarray3, 3},
    {"getarray", (lisp_proc *) getarray_fn, -1},
    {"getarray1", (lisp_proc *) Lgetarray1, 2},
    {"getarray2", (lisp_proc *) Lgetarray2, 3},
    {"getarray3", (lisp_proc *) Lgetarray3, 4},
    {"setarray", (lisp_proc *) setarray_fn, -1},
    {"setarray1", (lisp_proc *) Lsetarray1, 3},
    {"setarray2", (lisp_proc *) Lsetarray2, 4},
    {"setarray3", (lisp_proc *) Lsetarray3, 5},
    {"mkmatrix", (lisp_proc *) Lmkmatrix, 2},
    {"mkvector", (lisp_proc *) Lmkvector, 1},
    {"reversip", (lisp_proc *) Lnreverse, 1},
    {"subst", (lisp_proc *) Lsubst, 3},
    {"quit", (lisp_proc *) Lquit, 0},
    {"bye", (lisp_proc *) Lbye, 0},
#ifdef INCLUDE_FEND
    {"logxor", (lisp_proc *) logxor_fn, -1},
    {"logand", (lisp_proc *) logand_fn, -1},
    {"logor", (lisp_proc *) logor_fn, -1},
    {"ilogxor", (lisp_proc *) ilogxor_fn, -1},
    {"ilogand", (lisp_proc *) ilogand_fn, -1},
    {"ilogor", (lisp_proc *) ilogor_fn, -1},
    {"max", (lisp_proc *) max_fn, -1},
    {"min", (lisp_proc *) min_fn, -1},
    {"imax", (lisp_proc *) imax_fn, -1},
    {"imin", (lisp_proc *) imin_fn, -1},
    {"deflist", (lisp_proc *) Ldeflist, 2},
    {"expand", (lisp_proc *) Lexpand, 2},
    {"copy", (lisp_proc *) Lcopy, 1},
    {"clone", (lisp_proc *) Lclone, 1},
    {"xn", (lisp_proc *) Lxn, 2},
    {"last", (lisp_proc *) Llast, 1},
    {"num", (lisp_proc *) Lnum, 1},
    {"den", (lisp_proc *) Lden, 1},
    {"printprompt", (lisp_proc *) Lprintprompt, 1},
#endif
    {"printaddr", (lisp_proc *) Lprintaddr, 1},
    {"traceall", (lisp_proc *) Ltraceall, 1},
#ifdef MS_DOS
    {"int10h", (lisp_proc *) Lint10h, 4},
    {"int10hr", (lisp_proc *) Lint10hr, 4},
#endif
    {"atom-tag", (lisp_proc *) Latom_tag, 1},
    {"number-tag", (lisp_proc *) Lnumber_tag, 1},
    {"number-length", (lisp_proc *) Lnumber_length, 1},
    {"nth", (lisp_proc *) Lnth, 2},
    {"pnth", (lisp_proc *) Lpnth, 2},
    {"package_cnt", (lisp_proc *) Lpackage_cnt, 2},
    {"hashp", (lisp_proc *) Lhashp, 1},
    {"packagep", (lisp_proc *) Lpackagep, 1},
#ifdef ASUKA
    {"lisp", (lisp_proc *) Lmode_lisp, 0},
    {"asuka", (lisp_proc *) Lmode_asuka, 0},
    {"setorder", (lisp_proc *) Lsetorder, 1},
    {"DEBUG", (lisp_proc *) Adebug, 1},

    {"enable", (lisp_proc *) Aenable, -1},
    {"disable", (lisp_proc *) Adisable, -1},
#endif
    {NULL, 0, 0}
};
