/*
 * garith.c             (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * gal arithmetic routine
 * $Header: /Users/kako/asuka/src/RCS/garith.c,v 2.7 2015/12/09 03:44:28 kako Exp $
 */
static char const rcsid[] =
    "$Id: garith.c,v 2.7 2015/12/09 03:44:28 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"
#include "mathop.h"
#include "garith.h"


#ifdef DEBUG
Sex             xgratio2(Sex a, Sex b);
#endif

Sex
gmather(Sex a, Sex b)
{
    return aerror2("gal arithmetic operation", a, b);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/

two_args       *gplus_table_1[] =
    { gmather, gplus2nv, gplus2na, gplus2nt, gplus2np, gplus2nD,
    gplus2nA, gplus2nr, gplus2nV, gplus2nM, gplus2nT, gplus2nP,
    gplus2nR, gplus2nS, gplus2nL, gplus2nX
};

two_args       *gplus_table_2[] =
    { gmather, gplus2vn, gplus2an, gplus2tn, gplus2pn, gplus2Dn,
    gplus2An, gplus2rn, gplus2Vn, gplus2Mn, gplus2Tn, gplus2Pn,
    gplus2Rn, gplus2Sn, gplus2Ln, gplus2Xn
};

two_args       *gplus_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, gplus2vv, gplus2va, gplus2vt, gplus2vp, gplus2vD,
     gplus2vA, gplus2vr, gplus2vV, gplus2vM, gplus2vT, gplus2vP,
     gplus2vR, gplus2vS, gplus2vL, gplus2vX}
    ,
    {gmather, gplus2av, gplus2aa, gplus2at, gplus2ap, gplus2aD,
     gplus2aA, gplus2ar, gplus2aV, gplus2aM, gplus2aT, gplus2aP,
     gplus2aR, gplus2aS, gplus2aL, gplus2aX}
    ,
    {gmather, gplus2tv, gplus2ta, gplus2tt, gplus2tp, gplus2tD,
     gplus2tA, gplus2tr, gplus2tV, gplus2tM, gplus2tT, gplus2tP,
     gplus2tR, gplus2tS, gplus2tL, gplus2tX}
    ,
    {gmather, gplus2pv, gplus2pa, gplus2pt, gplus2pp, gplus2pD,
     gplus2pA, gplus2pr, gplus2pV, gplus2pM, gplus2pT, gplus2pP,
     gplus2pR, gplus2pS, gplus2pL, gplus2pX}
    ,
    {gmather, gplus2Dv, gplus2Da, gplus2Dt, gplus2Dp, gplus2DD,
     gplus2DA, gplus2Dr, gplus2DV, gplus2DM, gplus2DT, gplus2DP,
     gplus2DR, gplus2DS, gplus2DL, gplus2DX}
    ,
    {gmather, gplus2Av, gplus2Aa, gplus2At, gplus2Ap, gplus2AD,
     gplus2AA, gplus2Ar, gplus2AV, gplus2AM, gplus2AT, gplus2AP,
     gplus2AR, gplus2AS, gplus2AL, gplus2AX}
    ,
    {gmather, gplus2rv, gplus2ra, gplus2rt, gplus2rp, gplus2rD,
     gplus2rA, gplus2rr, gplus2rV, gplus2rM, gplus2rT, gplus2rP,
     gplus2rR, gplus2rS, gplus2rL, gplus2rX}
    ,
    {gmather, gplus2Vv, gplus2Va, gplus2Vt, gplus2Vp, gplus2VD,
     gplus2VA, gplus2Vr, gplus2VV, gplus2VM, gplus2VT, gplus2VP,
     gplus2VR, gplus2VS, gplus2VL, gplus2VX}
    ,
    {gmather, gplus2Mv, gplus2Ma, gplus2Mt, gplus2Mp, gplus2MD,
     gplus2MA, gplus2Mr, gplus2MV, gplus2MM, gplus2MT, gplus2MP,
     gplus2MR, gplus2MS, gplus2ML, gplus2MX}
    ,
    {gmather, gplus2Tv, gplus2Ta, gplus2Tt, gplus2Tp, gplus2TD,
     gplus2TA, gplus2Tr, gplus2TV, gplus2TM, gplus2TT, gplus2TP,
     gplus2TR, gplus2TS, gplus2TL, gplus2TX}
    ,
    {gmather, gplus2Pv, gplus2Pa, gplus2Pt, gplus2Pp, gplus2PD,
     gplus2PA, gplus2Pr, gplus2PV, gplus2PM, gplus2PT, gplus2PP,
     gplus2PR, gplus2PS, gplus2PL, gplus2PX}
    ,
    {gmather, gplus2Rv, gplus2Ra, gplus2Rt, gplus2Rp, gplus2RD,
     gplus2RA, gplus2Rr, gplus2RV, gplus2RM, gplus2RT, gplus2RP,
     gplus2RR, gplus2RS, gplus2RL, gplus2RX}
    ,
    {gmather, gplus2Sv, gplus2Sa, gplus2St, gplus2Sp, gplus2SD,
     gplus2SA, gplus2Sr, gplus2SV, gplus2SM, gplus2ST, gplus2SP,
     gplus2SR, gplus2SS, gplus2SL, gplus2SX}
    ,
    {gmather, gplus2Lv, gplus2La, gplus2Lt, gplus2Lp, gplus2LD,
     gplus2LA, gplus2Lr, gplus2LV, gplus2LM, gplus2LT, gplus2LP,
     gplus2LR, gplus2LS, gplus2LL, gplus2LX}
    ,
    {gmather, gplus2Xv, gplus2Xa, gplus2Xt, gplus2Xp, gplus2XD,
     gplus2XA, gplus2Xr, gplus2XV, gplus2XM, gplus2XT, gplus2XP,
     gplus2XR, gplus2XS, gplus2XL, gplus2XX}
};




Sex
gplus2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return plus2(a, b);
	else if (is_vechdr(b))
	    return gplus_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return gplus_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return gplus_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("gplus2", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



two_args       *gdiff_table_1[] =
    { gmather, gdiff2nv, gdiff2na, gdiff2nt, gdiff2np, gdiff2nD,
    gdiff2nA, gdiff2nr, gdiff2nV, gdiff2nM, gdiff2nT, gdiff2nP,
    gdiff2nR, gdiff2nS, gdiff2nL, gdiff2nX
};

two_args       *gdiff_table_2[] =
    { gmather, gdiff2vn, gdiff2an, gdiff2tn, gdiff2pn, gdiff2Dn,
    gdiff2An, gdiff2rn, gdiff2Vn, gdiff2Mn, gdiff2Tn, gdiff2Pn,
    gdiff2Rn, gdiff2Sn, gdiff2Ln, gdiff2Xn
};

two_args       *gdiff_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, gdiff2vv, gdiff2va, gdiff2vt, gdiff2vp, gdiff2vD,
     gdiff2vA, gdiff2vr, gdiff2vV, gdiff2vM, gdiff2vT, gdiff2vP,
     gdiff2vR, gdiff2vS, gdiff2vL, gdiff2vX}
    ,
    {gmather, gdiff2av, gdiff2aa, gdiff2at, gdiff2ap, gdiff2aD,
     gdiff2aA, gdiff2ar, gdiff2aV, gdiff2aM, gdiff2aT, gdiff2aP,
     gdiff2aR, gdiff2aS, gdiff2aL, gdiff2aX}
    ,
    {gmather, gdiff2tv, gdiff2ta, gdiff2tt, gdiff2tp, gdiff2tD,
     gdiff2tA, gdiff2tr, gdiff2tV, gdiff2tM, gdiff2tT, gdiff2tP,
     gdiff2tR, gdiff2tS, gdiff2tL, gdiff2tX}
    ,
    {gmather, gdiff2pv, gdiff2pa, gdiff2pt, gdiff2pp, gdiff2pD,
     gdiff2pA, gdiff2pr, gdiff2pV, gdiff2pM, gdiff2pT, gdiff2pP,
     gdiff2pR, gdiff2pS, gdiff2pL, gdiff2pX}
    ,
    {gmather, gdiff2Dv, gdiff2Da, gdiff2Dt, gdiff2Dp, gdiff2DD,
     gdiff2DA, gdiff2Dr, gdiff2DV, gdiff2DM, gdiff2DT, gdiff2DP,
     gdiff2DR, gdiff2DS, gdiff2DL, gdiff2DX}
    ,
    {gmather, gdiff2Av, gdiff2Aa, gdiff2At, gdiff2Ap, gdiff2AD,
     gdiff2AA, gdiff2Ar, gdiff2AV, gdiff2AM, gdiff2AT, gdiff2AP,
     gdiff2AR, gdiff2AS, gdiff2AL, gdiff2AX}
    ,
    {gmather, gdiff2rv, gdiff2ra, gdiff2rt, gdiff2rp, gdiff2rD,
     gdiff2rA, gdiff2rr, gdiff2rV, gdiff2rM, gdiff2rT, gdiff2rP,
     gdiff2rR, gdiff2rS, gdiff2rL, gdiff2rX}
    ,
    {gmather, gdiff2Vv, gdiff2Va, gdiff2Vt, gdiff2Vp, gdiff2VD,
     gdiff2VA, gdiff2Vr, gdiff2VV, gdiff2VM, gdiff2VT, gdiff2VP,
     gdiff2VR, gdiff2VS, gdiff2VL, gdiff2VX}
    ,
    {gmather, gdiff2Mv, gdiff2Ma, gdiff2Mt, gdiff2Mp, gdiff2MD,
     gdiff2MA, gdiff2Mr, gdiff2MV, gdiff2MM, gdiff2MT, gdiff2MP,
     gdiff2MR, gdiff2MS, gdiff2ML, gdiff2MX}
    ,
    {gmather, gdiff2Tv, gdiff2Ta, gdiff2Tt, gdiff2Tp, gdiff2TD,
     gdiff2TA, gdiff2Tr, gdiff2TV, gdiff2TM, gdiff2TT, gdiff2TP,
     gdiff2TR, gdiff2TS, gdiff2TL, gdiff2TX}
    ,
    {gmather, gdiff2Pv, gdiff2Pa, gdiff2Pt, gdiff2Pp, gdiff2PD,
     gdiff2PA, gdiff2Pr, gdiff2PV, gdiff2PM, gdiff2PT, gdiff2PP,
     gdiff2PR, gdiff2PS, gdiff2PL, gdiff2PX}
    ,
    {gmather, gdiff2Rv, gdiff2Ra, gdiff2Rt, gdiff2Rp, gdiff2RD,
     gdiff2RA, gdiff2Rr, gdiff2RV, gdiff2RM, gdiff2RT, gdiff2RP,
     gdiff2RR, gdiff2RS, gdiff2RL, gdiff2RX}
    ,
    {gmather, gdiff2Sv, gdiff2Sa, gdiff2St, gdiff2Sp, gdiff2SD,
     gdiff2SA, gdiff2Sr, gdiff2SV, gdiff2SM, gdiff2ST, gdiff2SP,
     gdiff2SR, gdiff2SS, gdiff2SL, gdiff2SX}
    ,
    {gmather, gdiff2Lv, gdiff2La, gdiff2Lt, gdiff2Lp, gdiff2LD,
     gdiff2LA, gdiff2Lr, gdiff2LV, gdiff2LM, gdiff2LT, gdiff2LP,
     gdiff2LR, gdiff2LS, gdiff2LL, gdiff2LX}
    ,
    {gmather, gdiff2Xv, gdiff2Xa, gdiff2Xt, gdiff2Xp, gdiff2XD,
     gdiff2XA, gdiff2Xr, gdiff2XV, gdiff2XM, gdiff2XT, gdiff2XP,
     gdiff2XR, gdiff2XS, gdiff2XL, gdiff2XX}
};




Sex
gdiff2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return diff2(a, b);
	else if (is_vechdr(b))
	    return gdiff_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return gdiff_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return gdiff_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("gdiff2", a, b);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/



two_args       *gtimes_table_1[] =
    { gmather, gtimes2nv, gtimes2na, gtimes2nt, gtimes2np, gtimes2nD,
    gtimes2nA, gtimes2nr, gtimes2nV, gtimes2nM, gtimes2nT, gtimes2nP,
    gtimes2nR, gtimes2nS, gtimes2nL, gtimes2nX
};

two_args       *gtimes_table_2[] =
    { gmather, gtimes2vn, gtimes2an, gtimes2tn, gtimes2pn, gtimes2Dn,
    gtimes2An, gtimes2rn, gtimes2Vn, gtimes2Mn, gtimes2Tn, gtimes2Pn,
    gtimes2Rn, gtimes2Sn, gtimes2Ln, gtimes2Xn
};

two_args       *gtimes_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, gtimes2vv, gtimes2va, gtimes2vt, gtimes2vp, gtimes2vD,
     gtimes2vA, gtimes2vr, gtimes2vV, gtimes2vM, gtimes2vT, gtimes2vP,
     gtimes2vR, gtimes2vS, gtimes2vL, gtimes2vX}
    ,
    {gmather, gtimes2av, gtimes2aa, gtimes2at, gtimes2ap, gtimes2aD,
     gtimes2aA, gtimes2ar, gtimes2aV, gtimes2aM, gtimes2aT, gtimes2aP,
     gtimes2aR, gtimes2aS, gtimes2aL, gtimes2aX}
    ,
    {gmather, gtimes2tv, gtimes2ta, gtimes2tt, gtimes2tp, gtimes2tD,
     gtimes2tA, gtimes2tr, gtimes2tV, gtimes2tM, gtimes2tT, gtimes2tP,
     gtimes2tR, gtimes2tS, gtimes2tL, gtimes2tX}
    ,
    {gmather, gtimes2pv, gtimes2pa, gtimes2pt, gtimes2pp, gtimes2pD,
     gtimes2pA, gtimes2pr, gtimes2pV, gtimes2pM, gtimes2pT, gtimes2pP,
     gtimes2pR, gtimes2pS, gtimes2pL, gtimes2pX}
    ,
    {gmather, gtimes2Dv, gtimes2Da, gtimes2Dt, gtimes2Dp, gtimes2DD,
     gtimes2DA, gtimes2Dr, gtimes2DV, gtimes2DM, gtimes2DT, gtimes2DP,
     gtimes2DR, gtimes2DS, gtimes2DL, gtimes2DX}
    ,
    {gmather, gtimes2Av, gtimes2Aa, gtimes2At, gtimes2Ap, gtimes2AD,
     gtimes2AA, gtimes2Ar, gtimes2AV, gtimes2AM, gtimes2AT, gtimes2AP,
     gtimes2AR, gtimes2AS, gtimes2AL, gtimes2AX}
    ,
    {gmather, gtimes2rv, gtimes2ra, gtimes2rt, gtimes2rp, gtimes2rD,
     gtimes2rA, gtimes2rr, gtimes2rV, gtimes2rM, gtimes2rT, gtimes2rP,
     gtimes2rR, gtimes2rS, gtimes2rL, gtimes2rX}
    ,
    {gmather, gtimes2Vv, gtimes2Va, gtimes2Vt, gtimes2Vp, gtimes2VD,
     gtimes2VA, gtimes2Vr, gtimes2VV, gtimes2VM, gtimes2VT, gtimes2VP,
     gtimes2VR, gtimes2VS, gtimes2VL, gtimes2VX}
    ,
    {gmather, gtimes2Mv, gtimes2Ma, gtimes2Mt, gtimes2Mp, gtimes2MD,
     gtimes2MA, gtimes2Mr, gtimes2MV, gtimes2MM, gtimes2MT, gtimes2MP,
     gtimes2MR, gtimes2MS, gtimes2ML, gtimes2MX}
    ,
    {gmather, gtimes2Tv, gtimes2Ta, gtimes2Tt, gtimes2Tp, gtimes2TD,
     gtimes2TA, gtimes2Tr, gtimes2TV, gtimes2TM, gtimes2TT, gtimes2TP,
     gtimes2TR, gtimes2TS, gtimes2TL, gtimes2TX}
    ,
    {gmather, gtimes2Pv, gtimes2Pa, gtimes2Pt, gtimes2Pp, gtimes2PD,
     gtimes2PA, gtimes2Pr, gtimes2PV, gtimes2PM, gtimes2PT, gtimes2PP,
     gtimes2PR, gtimes2PS, gtimes2PL, gtimes2PX}
    ,
    {gmather, gtimes2Rv, gtimes2Ra, gtimes2Rt, gtimes2Rp, gtimes2RD,
     gtimes2RA, gtimes2Rr, gtimes2RV, gtimes2RM, gtimes2RT, gtimes2RP,
     gtimes2RR, gtimes2RS, gtimes2RL, gtimes2RX}
    ,
    {gmather, gtimes2Sv, gtimes2Sa, gtimes2St, gtimes2Sp, gtimes2SD,
     gtimes2SA, gtimes2Sr, gtimes2SV, gtimes2SM, gtimes2ST, gtimes2SP,
     gtimes2SR, gtimes2SS, gtimes2SL, gtimes2SX}
    ,
    {gmather, gtimes2Lv, gtimes2La, gtimes2Lt, gtimes2Lp, gtimes2LD,
     gtimes2LA, gtimes2Lr, gtimes2LV, gtimes2LM, gtimes2LT, gtimes2LP,
     gtimes2LR, gtimes2LS, gtimes2LL, gtimes2LX}
    ,
    {gmather, gtimes2Xv, gtimes2Xa, gtimes2Xt, gtimes2Xp, gtimes2XD,
     gtimes2XA, gtimes2Xr, gtimes2XV, gtimes2XM, gtimes2XT, gtimes2XP,
     gtimes2XR, gtimes2XS, gtimes2XL, gtimes2XX}
};



Sex
gtimes2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return times2(a, b);
	else if (is_vechdr(b))
	    return gtimes_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return gtimes_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return gtimes_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("gtimes2", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/




two_args       *gquot_table_1[] =
    { gmather, gquot2nv, gquot2na, gquot2nt, gquot2np, gquot2nD,
    gquot2nA, gquot2nr, gquot2nV, gquot2nM, gquot2nT, gquot2nP,
    gquot2nR, gquot2nS, gquot2nL, gquot2nX
};

two_args       *gquot_table_2[] =
    { gmather, gquot2vn, gquot2an, gquot2tn, gquot2pn, gquot2Dn,
    gquot2An, gquot2rn, gquot2Vn, gquot2Mn, gquot2Tn, gquot2Pn,
    gquot2Rn, gquot2Sn, gquot2Ln, gquot2Xn
};

two_args       *gquot_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, gquot2vv, gquot2va, gquot2vt, gquot2vp, gquot2vD,
     gquot2vA, gquot2vr, gquot2vV, gquot2vM, gquot2vT, gquot2vP,
     gquot2vR, gquot2vS, gquot2vL, gquot2vX}
    ,
    {gmather, gquot2av, gquot2aa, gquot2at, gquot2ap, gquot2aD,
     gquot2aA, gquot2ar, gquot2aV, gquot2aM, gquot2aT, gquot2aP,
     gquot2aR, gquot2aS, gquot2aL, gquot2aX}
    ,
    {gmather, gquot2tv, gquot2ta, gquot2tt, gquot2tp, gquot2tD,
     gquot2tA, gquot2tr, gquot2tV, gquot2tM, gquot2tT, gquot2tP,
     gquot2tR, gquot2tS, gquot2tL, gquot2tX}
    ,
    {gmather, gquot2pv, gquot2pa, gquot2pt, gquot2pp, gquot2pD,
     gquot2pA, gquot2pr, gquot2pV, gquot2pM, gquot2pT, gquot2pP,
     gquot2pR, gquot2pS, gquot2pL, gquot2pX}
    ,
    {gmather, gquot2Dv, gquot2Da, gquot2Dt, gquot2Dp, gquot2DD,
     gquot2DA, gquot2Dr, gquot2DV, gquot2DM, gquot2DT, gquot2DP,
     gquot2DR, gquot2DS, gquot2DL, gquot2DX}
    ,
    {gmather, gquot2Av, gquot2Aa, gquot2At, gquot2Ap, gquot2AD,
     gquot2AA, gquot2Ar, gquot2AV, gquot2AM, gquot2AT, gquot2AP,
     gquot2AR, gquot2AS, gquot2AL, gquot2AX}
    ,
    {gmather, gquot2rv, gquot2ra, gquot2rt, gquot2rp, gquot2rD,
     gquot2rA, gquot2rr, gquot2rV, gquot2rM, gquot2rT, gquot2rP,
     gquot2rR, gquot2rS, gquot2rL, gquot2rX}
    ,
    {gmather, gquot2Vv, gquot2Va, gquot2Vt, gquot2Vp, gquot2VD,
     gquot2VA, gquot2Vr, gquot2VV, gquot2VM, gquot2VT, gquot2VP,
     gquot2VR, gquot2VS, gquot2VL, gquot2VX}
    ,
    {gmather, gquot2Mv, gquot2Ma, gquot2Mt, gquot2Mp, gquot2MD,
     gquot2MA, gquot2Mr, gquot2MV, gquot2MM, gquot2MT, gquot2MP,
     gquot2MR, gquot2MS, gquot2ML, gquot2MX}
    ,
    {gmather, gquot2Tv, gquot2Ta, gquot2Tt, gquot2Tp, gquot2TD,
     gquot2TA, gquot2Tr, gquot2TV, gquot2TM, gquot2TT, gquot2TP,
     gquot2TR, gquot2TS, gquot2TL, gquot2TX}
    ,
    {gmather, gquot2Pv, gquot2Pa, gquot2Pt, gquot2Pp, gquot2PD,
     gquot2PA, gquot2Pr, gquot2PV, gquot2PM, gquot2PT, gquot2PP,
     gquot2PR, gquot2PS, gquot2PL, gquot2PX}
    ,
    {gmather, gquot2Rv, gquot2Ra, gquot2Rt, gquot2Rp, gquot2RD,
     gquot2RA, gquot2Rr, gquot2RV, gquot2RM, gquot2RT, gquot2RP,
     gquot2RR, gquot2RS, gquot2RL, gquot2RX}
    ,
    {gmather, gquot2Sv, gquot2Sa, gquot2St, gquot2Sp, gquot2SD,
     gquot2SA, gquot2Sr, gquot2SV, gquot2SM, gquot2ST, gquot2SP,
     gquot2SR, gquot2SS, gquot2SL, gquot2SX}
    ,
    {gmather, gquot2Lv, gquot2La, gquot2Lt, gquot2Lp, gquot2LD,
     gquot2LA, gquot2Lr, gquot2LV, gquot2LM, gquot2LT, gquot2LP,
     gquot2LR, gquot2LS, gquot2LL, gquot2LX}
    ,
    {gmather, gquot2Xv, gquot2Xa, gquot2Xt, gquot2Xp, gquot2XD,
     gquot2XA, gquot2Xr, gquot2XV, gquot2XM, gquot2XT, gquot2XP,
     gquot2XR, gquot2XS, gquot2XL, gquot2XX}
};




Sex
gquot2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return quot2(a, b);
	else if (is_vechdr(b))
	    return gquot_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b)) {
	    if (b == mkfxat(1))
		return a;
	    return gquot_table_2[gnumtype_of(a)] (a, b);
	} else if (is_vechdr(b))
	    return gquot_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("gquot2", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


two_args       *gratio_table_1[] =
    { gmather, gratio2nv, gratio2na, gratio2nt, gratio2np, gratio2nD,
    gratio2nA, gratio2nr, gratio2nV, gratio2nM, gratio2nT, gratio2nP,
    gratio2nR, gratio2nS, gratio2nL, gratio2nX
};

two_args       *gratio_table_2[] =
    { gmather, gratio2vn, gratio2an, gratio2tn, gratio2pn, gratio2Dn,
    gratio2An, gratio2rn, gratio2Vn, gratio2Mn, gratio2Tn, gratio2Pn,
    gratio2Rn, gratio2Sn, gratio2Ln, gratio2Xn
};

two_args       *gratio_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, gratio2vv, gratio2va, gratio2vt, gratio2vp, gratio2vD,
     gratio2vA, gratio2vr, gratio2vV, gratio2vM, gratio2vT, gratio2vP,
     gratio2vR, gratio2vS, gratio2vL, gratio2vX}
    ,
    {gmather, gratio2av, gratio2aa, gratio2at, gratio2ap, gratio2aD,
     gratio2aA, gratio2ar, gratio2aV, gratio2aM, gratio2aT, gratio2aP,
     gratio2aR, gratio2aS, gratio2aL, gratio2aX}
    ,
    {gmather, gratio2tv, gratio2ta, gratio2tt, gratio2tp, gratio2tD,
     gratio2tA, gratio2tr, gratio2tV, gratio2tM, gratio2tT, gratio2tP,
     gratio2tR, gratio2tS, gratio2tL, gratio2tX}
    ,
    {gmather, gratio2pv, gratio2pa, gratio2pt, gratio2pp, gratio2pD,
     gratio2pA, gratio2pr, gratio2pV, gratio2pM, gratio2pT, gratio2pP,
     gratio2pR, gratio2pS, gratio2pL, gratio2pX}
    ,
    {gmather, gratio2Dv, gratio2Da, gratio2Dt, gratio2Dp, gratio2DD,
     gratio2DA, gratio2Dr, gratio2DV, gratio2DM, gratio2DT, gratio2DP,
     gratio2DR, gratio2DS, gratio2DL, gratio2DX}
    ,
    {gmather, gratio2Av, gratio2Aa, gratio2At, gratio2Ap, gratio2AD,
     gratio2AA, gratio2Ar, gratio2AV, gratio2AM, gratio2AT, gratio2AP,
     gratio2AR, gratio2AS, gratio2AL, gratio2AX}
    ,
    {gmather, gratio2rv, gratio2ra, gratio2rt, gratio2rp, gratio2rD,
     gratio2rA, gratio2rr, gratio2rV, gratio2rM, gratio2rT, gratio2rP,
     gratio2rR, gratio2rS, gratio2rL, gratio2rX}
    ,
    {gmather, gratio2Vv, gratio2Va, gratio2Vt, gratio2Vp, gratio2VD,
     gratio2VA, gratio2Vr, gratio2VV, gratio2VM, gratio2VT, gratio2VP,
     gratio2VR, gratio2VS, gratio2VL, gratio2VX}
    ,
    {gmather, gratio2Mv, gratio2Ma, gratio2Mt, gratio2Mp, gratio2MD,
     gratio2MA, gratio2Mr, gratio2MV, gratio2MM, gratio2MT, gratio2MP,
     gratio2MR, gratio2MS, gratio2ML, gratio2MX}
    ,
    {gmather, gratio2Tv, gratio2Ta, gratio2Tt, gratio2Tp, gratio2TD,
     gratio2TA, gratio2Tr, gratio2TV, gratio2TM, gratio2TT, gratio2TP,
     gratio2TR, gratio2TS, gratio2TL, gratio2TX}
    ,
    {gmather, gratio2Pv, gratio2Pa, gratio2Pt, gratio2Pp, gratio2PD,
     gratio2PA, gratio2Pr, gratio2PV, gratio2PM, gratio2PT, gratio2PP,
     gratio2PR, gratio2PS, gratio2PL, gratio2PX}
    ,
    {gmather, gratio2Rv, gratio2Ra, gratio2Rt, gratio2Rp, gratio2RD,
     gratio2RA, gratio2Rr, gratio2RV, gratio2RM, gratio2RT, gratio2RP,
     gratio2RR, gratio2RS, gratio2RL, gratio2RX}
    ,
    {gmather, gratio2Sv, gratio2Sa, gratio2St, gratio2Sp, gratio2SD,
     gratio2SA, gratio2Sr, gratio2SV, gratio2SM, gratio2ST, gratio2SP,
     gratio2SR, gratio2SS, gratio2SL, gratio2SX}
    ,
    {gmather, gratio2Lv, gratio2La, gratio2Lt, gratio2Lp, gratio2LD,
     gratio2LA, gratio2Lr, gratio2LV, gratio2LM, gratio2LT, gratio2LP,
     gratio2LR, gratio2LS, gratio2LL, gratio2LX}
    ,
    {gmather, gratio2Xv, gratio2Xa, gratio2Xt, gratio2Xp, gratio2XD,
     gratio2XA, gratio2Xr, gratio2XV, gratio2XM, gratio2XT, gratio2XP,
     gratio2XR, gratio2XS, gratio2XL, gratio2XX}
};


#ifdef DEBUG
Sex
gratio2(a, b)
     Sex             a,
                     b;
{
    Sex             c;
    lisp_printf("gratio2(%p", a);
    lisp_printf(", %p) = ", b);
    c = xgratio2(a, b);
    lisp_printf("%p\n", c);
    return c;
}
#else
#define xgratio2 gratio2
#endif

Sex
xgratio2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return ratio2(a, b);
	else if (is_vechdr(b))
	    return gratio_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return gratio_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return gratio_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("gratio2", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


two_args       *grem_table_1[] =
    { gmather, grem2nv, grem2na, grem2nt, grem2np, grem2nD,
    grem2nA, grem2nr, grem2nV, grem2nM, grem2nT, grem2nP,
    grem2nR, grem2nS, grem2nL, grem2nX
};

two_args       *grem_table_2[] =
    { gmather, grem2vn, grem2an, grem2tn, grem2pn, grem2Dn,
    grem2An, grem2rn, grem2Vn, grem2Mn, grem2Tn, grem2Pn,
    grem2Rn, grem2Sn, grem2Ln, grem2Xn
};

two_args       *grem_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, grem2vv, grem2va, grem2vt, grem2vp, grem2vD,
     grem2vA, grem2vr, grem2vV, grem2vM, grem2vT, grem2vP,
     grem2vR, grem2vS, grem2vL, grem2vX}
    ,
    {gmather, grem2av, grem2aa, grem2at, grem2ap, grem2aD,
     grem2aA, grem2ar, grem2aV, grem2aM, grem2aT, grem2aP,
     grem2aR, grem2aS, grem2aL, grem2aX}
    ,
    {gmather, grem2tv, grem2ta, grem2tt, grem2tp, grem2tD,
     grem2tA, grem2tr, grem2tV, grem2tM, grem2tT, grem2tP,
     grem2tR, grem2tS, grem2tL, grem2tX}
    ,
    {gmather, grem2pv, grem2pa, grem2pt, grem2pp, grem2pD,
     grem2pA, grem2pr, grem2pV, grem2pM, grem2pT, grem2pP,
     grem2pR, grem2pS, grem2pL, grem2pX}
    ,
    {gmather, grem2Dv, grem2Da, grem2Dt, grem2Dp, grem2DD,
     grem2DA, grem2Dr, grem2DV, grem2DM, grem2DT, grem2DP,
     grem2DR, grem2DS, grem2DL, grem2DX}
    ,
    {gmather, grem2Av, grem2Aa, grem2At, grem2Ap, grem2AD,
     grem2AA, grem2Ar, grem2AV, grem2AM, grem2AT, grem2AP,
     grem2AR, grem2AS, grem2AL, grem2AX}
    ,
    {gmather, grem2rv, grem2ra, grem2rt, grem2rp, grem2rD,
     grem2rA, grem2rr, grem2rV, grem2rM, grem2rT, grem2rP,
     grem2rR, grem2rS, grem2rL, grem2rX}
    ,
    {gmather, grem2Vv, grem2Va, grem2Vt, grem2Vp, grem2VD,
     grem2VA, grem2Vr, grem2VV, grem2VM, grem2VT, grem2VP,
     grem2VR, grem2VS, grem2VL, grem2VX}
    ,
    {gmather, grem2Mv, grem2Ma, grem2Mt, grem2Mp, grem2MD,
     grem2MA, grem2Mr, grem2MV, grem2MM, grem2MT, grem2MP,
     grem2MR, grem2MS, grem2ML, grem2MX}
    ,
    {gmather, grem2Tv, grem2Ta, grem2Tt, grem2Tp, grem2TD,
     grem2TA, grem2Tr, grem2TV, grem2TM, grem2TT, grem2TP,
     grem2TR, grem2TS, grem2TL, grem2TX}
    ,
    {gmather, grem2Pv, grem2Pa, grem2Pt, grem2Pp, grem2PD,
     grem2PA, grem2Pr, grem2PV, grem2PM, grem2PT, grem2PP,
     grem2PR, grem2PS, grem2PL, grem2PX}
    ,
    {gmather, grem2Rv, grem2Ra, grem2Rt, grem2Rp, grem2RD,
     grem2RA, grem2Rr, grem2RV, grem2RM, grem2RT, grem2RP,
     grem2RR, grem2RS, grem2RL, grem2RX}
    ,
    {gmather, grem2Sv, grem2Sa, grem2St, grem2Sp, grem2SD,
     grem2SA, grem2Sr, grem2SV, grem2SM, grem2ST, grem2SP,
     grem2SR, grem2SS, grem2SL, grem2SX}
    ,
    {gmather, grem2Lv, grem2La, grem2Lt, grem2Lp, grem2LD,
     grem2LA, grem2Lr, grem2LV, grem2LM, grem2LT, grem2LP,
     grem2LR, grem2LS, grem2LL, grem2LX}
    ,
    {gmather, grem2Xv, grem2Xa, grem2Xt, grem2Xp, grem2XD,
     grem2XA, grem2Xr, grem2XV, grem2XM, grem2XT, grem2XP,
     grem2XR, grem2XS, grem2XL, grem2XX}
};




Sex
gremaind2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return remaind2(a, b);
	else if (is_vechdr(b))
	    return grem_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return grem_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return grem_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("grem2", a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


two_args       *gexpt_table_2[] =
    { gmather, gexpt2vi, gexpt2ai, gexpt2ti, gexpt2pi, gexpt2Di,
    gexpt2Ai, gexpt2ri, gexpt2Vi, gexpt2Mi, gexpt2Ti, gexpt2Pi,
    gexpt2Ri, gexpt2Si, gexpt2Li, gexpt2Xi
};

Sex
gexpt(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return Lexpt(a, b);
    } else if (is_vechdr(a)) {
	if (is_fixnum(b))
	    return gexpt_table_2[gnumtype_of(a)] (a, b);
    }
    return aerror2("gexpt", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/







Sex
gplus2nA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2nV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2nM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2nT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2nR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2nS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2nL(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gplus2An(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Vn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Mn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Tn(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2Rn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Sn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Ln(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2vR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2vL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2av(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2at(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2aA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2aV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2aM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2aT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2aR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2aS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2aL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2ta(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2tA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2tV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2tM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2tT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2tR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2tS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2tL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2pt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2pA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2pV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2pM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2pT(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2pR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2pS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2pL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gplus2Av(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Aa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2At(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Ap(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Ar(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2AL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rv(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gplus2rA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2rL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Vv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Vp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2VA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Vr(Sex a, Sex b)
{
    return gmather(a, b);
}

/*
 * Sex  gplus2VV(Sex a, Sex b) { return gmather(a, b); } 
 */

Sex
gplus2VM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2VT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2VR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2VS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2VL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Mv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Ma(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Mt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Mp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2MA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Mr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2MV(Sex a, Sex b)
{
    return gmather(a, b);
}

Sex
gplus2MT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2MR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2MS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2ML(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Ta(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Tt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2TL(Sex a, Sex b)
{
    return gmather(a, b);
}









Sex
gplus2Rv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Ra(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Rt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Rp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Rr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2RL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Sv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Sa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2St(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Sp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Sr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2ST(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2SL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Lv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2La(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Lt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Lp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2Lr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gplus2LL(Sex a, Sex b)
{
    return gmather(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/






Sex
gdiff2nA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2nV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2nM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2nT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2nR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2nS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2nL(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gdiff2An(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Vn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Mn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Tn(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2Rn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Sn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Ln(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2vL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2av(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2at(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2aA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2aV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2aM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2aT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2aR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2aS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2aL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2ta(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2tV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2tL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2pt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2pL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2DL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Av(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Aa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2At(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Ap(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Ar(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2AL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rv(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gdiff2rA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2rL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Vv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Vp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2VA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Vr(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2VM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2VT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2VR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2VS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2VL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Mv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Ma(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Mt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Mp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2MA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Mr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2MV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2MT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2MR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2MS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2ML(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Ta(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Tt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2TL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gdiff2Rv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Ra(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Rt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Rp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Rr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2RL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Sv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Sa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2St(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Sp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Sr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2ST(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2SL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Lv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2La(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Lt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Lp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2Lr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gdiff2LL(Sex a, Sex b)
{
    return gmather(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/






Sex
gtimes2nA(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2nT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2nR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2nS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2nL(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2An(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2Tn(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2Rn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Sn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Ln(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vr(Sex a, Sex b)
{
    return gmather(a, b);
}







Sex
gtimes2vT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2vL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2av(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2at(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2aA(Sex a, Sex b)
{
    return gmather(a, b);
}















Sex
gtimes2aT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2aR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2aS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2aL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2ta(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2tA(Sex a, Sex b)
{
    return gmather(a, b);
}















Sex
gtimes2tT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2tR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2tS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2tL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2pt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2pA(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2pT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2pR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2pS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2pL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2DL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Av(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Aa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2At(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Ap(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2AA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Ar(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2AT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2AR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2AS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2AL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2rv(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2rA(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gtimes2rT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2rR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2rS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2rL(Sex a, Sex b)
{
    return gmather(a, b);
}








































Sex
gtimes2VT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2VR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2VS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2VL(Sex a, Sex b)
{
    return gmather(a, b);
}








































Sex
gtimes2MT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2MR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2MS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2ML(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Ta(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Tt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2TL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gtimes2Rv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Ra(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Rt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Rp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Rr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2RL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Sv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Sa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2St(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Sp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Sr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2ST(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2SL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Lv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2La(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Lt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Lp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2Lr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gtimes2LL(Sex a, Sex b)
{
    return gmather(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/






Sex
gquot2nA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2nV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2nM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2nT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2nR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2nS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2nL(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gquot2An(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Vn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Mn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Tn(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2Rn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Sn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Ln(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2vL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2av(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2at(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2aA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2aV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2aM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2aT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2aR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2aS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2aL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2ta(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2tV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2tL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2pt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2pL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2DL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Av(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Aa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2At(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Ap(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Ar(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2AL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rv(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gquot2rA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2rL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Vv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Vp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2VL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Mv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Ma(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Mt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Mp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2MA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Mr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2MV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2MM(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2MT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2MR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2MS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2ML(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Ta(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Tt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2TL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gquot2Rv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Ra(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Rt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Rp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Rr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2RL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Sv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Sa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2St(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Sp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Sr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2ST(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2SL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Lv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2La(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Lt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Lp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2Lr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gquot2LL(Sex a, Sex b)
{
    return gmather(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/






Sex
gratio2nA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2nV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2nT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2nR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2nS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2nL(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gratio2An(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gratio2Tn(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2Rn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Sn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Ln(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2vT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2vL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2av(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2at(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2aA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2aV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2aT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2aR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2aS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2aL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2ta(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2tT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2tL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2pt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2pA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2pV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2pT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2pR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2pS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2pL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2DL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Av(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Aa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2At(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Ap(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2AA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Ar(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2AV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2AT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2AR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2AS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2AL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2rv(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gratio2rA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2rV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2rT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2rR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2rS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2rL(Sex a, Sex b)
{
    return gmather(a, b);
}






























Sex
gratio2VV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2VT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2VR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2VS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2VL(Sex a, Sex b)
{
    return gmather(a, b);
}






























Sex
gratio2MV(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2MT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2MR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2MS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2ML(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Ta(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Tt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2TL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gratio2Rv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Ra(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Rt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Rp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Rr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2RL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Sv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Sa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2St(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Sp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2SA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Sr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2SV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2SM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2ST(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2SR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2SS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2SL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Lv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2La(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Lt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Lp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2Lr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gratio2LL(Sex a, Sex b)
{
    return gmather(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/






Sex
grem2nA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2nV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2nM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2nT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2nR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2nS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2nL(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
grem2An(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Vn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Mn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Tn(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2Rn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Sn(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Ln(Sex a, Sex b)
{
    return gmather(a, b);
}

Sex
grem2va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2vL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2av(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2at(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2aA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2aV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2aM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2aT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2aR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2aS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2aL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2ta(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tA(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2tV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2tL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2pt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2pL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2DL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Av(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Aa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2At(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Ap(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Ar(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2AL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rv(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
grem2rA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2rL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Vv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Va(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Vt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Vp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Vr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2VL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Mv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Ma(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Mt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Mp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2MA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Mr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2MV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2MM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2MT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2MR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2MS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2ML(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Tv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Ta(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Tt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Tp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Tr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2TL(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
grem2Rv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Ra(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Rt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Rp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Rr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2RL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Sv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Sa(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2St(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Sp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2SA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Sr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2SV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2SM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2ST(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2SR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2SS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2SL(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Lv(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2La(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Lt(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Lp(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LA(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2Lr(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LV(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LM(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LT(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LR(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LS(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
grem2LL(Sex a, Sex b)
{
    return gmather(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/











Sex
gexpt2Ai(Sex a, Sex b)
{
    return gmather(a, b);
}










Sex
gexpt2Ti(Sex a, Sex b)
{
    return gmather(a, b);
}





Sex
gexpt2Ri(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gexpt2Si(Sex a, Sex b)
{
    return gmather(a, b);
}
Sex
gexpt2Li(Sex a, Sex b)
{
    return gmather(a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
gnegate(Sex a)
{
    if (is_number(a))
	return negate(a);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_ALGNUM:
	    return gnegatea(a);
	case N_TRNNUM:
	    return gnegatet(a);
	case N_POLY:
	    return gnegatep(a);
	case N_RPOLY:
	    return gnegater(a);
	case N_DIPOLY:
	    return gnegateD(a);
	case N_VECT:
	    return gnegateV(a);
	case N_MATRIX:
	    return gnegateM(a);
	}
    }
    return aerror1("gnegate", a);
}

/***************************************************************/
/***************************************************************/

Sex
gmodulus(Sex a, Sex b)
{
    if (!is_integer(b))
	aerror2("gmodulus", a, b);
    if (is_number(a))
	return modulus(a, b);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_ALGNUM:
	    return gmodulusa(a, b);
	case N_TRNNUM:
	    return gmodulust(a, b);
	case N_POLY:
	    return gmodulusp(a, b);
	case N_RPOLY:
	    return gmodulusr(a, b);
	case N_DIPOLY:
	    return gmodulusD(a, b);
	case N_VECT:
	    return gmodulusV(a, b);
	case N_MATRIX:
	    return gmodulusM(a, b);
	}
    }
    return aerror2("gmodulus", a, b);
}



Sex
gsmodulus(Sex a, Sex b)
{
    if (!is_integer(b))
	aerror2("gsmodulus", a, b);
    if (is_number(a))
	return smodulus(a, b);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_ALGNUM:
	    return gsmodulusa(a, b);
	case N_TRNNUM:
	    return gsmodulust(a, b);
	case N_POLY:
	    return gsmodulusp(a, b);
	case N_RPOLY:
	    return gsmodulusr(a, b);
	case N_DIPOLY:
	    return gsmodulusD(a, b);
	case N_VECT:
	    return gsmodulusV(a, b);
	case N_MATRIX:
	    return gsmodulusM(a, b);
	}
    }
    return aerror2("gsmodulus", a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/



two_args       *ggcd_table_1[] =
    { gmather, ggcd2nv, ggcd2na, ggcd2nt, ggcd2np, ggcd2nD,
    ggcd2nA, ggcd2nr, ggcd2nV, ggcd2nM, ggcd2nT, ggcd2nP,
    ggcd2nR, ggcd2nS, ggcd2nL, ggcd2nX
};

two_args       *ggcd_table_2[] =
    { gmather, ggcd2vn, ggcd2an, ggcd2tn, ggcd2pn, ggcd2Dn,
    ggcd2An, ggcd2rn, ggcd2Vn, ggcd2Mn, ggcd2Tn, ggcd2Pn,
    ggcd2Rn, ggcd2Sn, ggcd2Ln, ggcd2Xn
};

two_args       *ggcd_table[GTYPEMAX][GTYPEMAX] = {
    {gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather, gmather, gmather,
     gmather, gmather, gmather, gmather}
    ,
    {gmather, ggcd2vv, ggcd2va, ggcd2vt, ggcd2vp, ggcd2vD,
     ggcd2vA, ggcd2vr, ggcd2vV, ggcd2vM, ggcd2vT, ggcd2vP,
     ggcd2vR, ggcd2vS, ggcd2vL, ggcd2vX}
    ,
    {gmather, ggcd2av, ggcd2aa, ggcd2at, ggcd2ap, ggcd2aD,
     ggcd2aA, ggcd2ar, ggcd2aV, ggcd2aM, ggcd2aT, ggcd2aP,
     ggcd2aR, ggcd2aS, ggcd2aL, ggcd2aX}
    ,
    {gmather, ggcd2tv, ggcd2ta, ggcd2tt, ggcd2tp, ggcd2tD,
     ggcd2tA, ggcd2tr, ggcd2tV, ggcd2tM, ggcd2tT, ggcd2tP,
     ggcd2tR, ggcd2tS, ggcd2tL, ggcd2tX}
    ,
    {gmather, ggcd2pv, ggcd2pa, ggcd2pt, ggcd2pp, ggcd2pD,
     ggcd2pA, ggcd2pr, ggcd2pV, ggcd2pM, ggcd2pT, ggcd2pP,
     ggcd2pR, ggcd2pS, ggcd2pL, ggcd2pX}
    ,
    {gmather, ggcd2Dv, ggcd2Da, ggcd2Dt, ggcd2Dp, ggcd2DD,
     ggcd2DA, ggcd2Dr, ggcd2DV, ggcd2DM, ggcd2DT, ggcd2DP,
     ggcd2DR, ggcd2DS, ggcd2DL, ggcd2DX}
    ,
    {gmather, ggcd2Av, ggcd2Aa, ggcd2At, ggcd2Ap, ggcd2AD,
     ggcd2AA, ggcd2Ar, ggcd2AV, ggcd2AM, ggcd2AT, ggcd2AP,
     ggcd2AR, ggcd2AS, ggcd2AL, ggcd2AX}
    ,
    {gmather, ggcd2rv, ggcd2ra, ggcd2rt, ggcd2rp, ggcd2rD,
     ggcd2rA, ggcd2rr, ggcd2rV, ggcd2rM, ggcd2rT, ggcd2rP,
     ggcd2rR, ggcd2rS, ggcd2rL, ggcd2rX}
    ,
    {gmather, ggcd2Vv, ggcd2Va, ggcd2Vt, ggcd2Vp, ggcd2VD,
     ggcd2VA, ggcd2Vr, ggcd2VV, ggcd2VM, ggcd2VT, ggcd2VP,
     ggcd2VR, ggcd2VS, ggcd2VL, ggcd2VX}
    ,
    {gmather, ggcd2Mv, ggcd2Ma, ggcd2Mt, ggcd2Mp, ggcd2MD,
     ggcd2MA, ggcd2Mr, ggcd2MV, ggcd2MM, ggcd2MT, ggcd2MP,
     ggcd2MR, ggcd2MS, ggcd2ML, ggcd2MX}
    ,
    {gmather, ggcd2Tv, ggcd2Ta, ggcd2Tt, ggcd2Tp, ggcd2TD,
     ggcd2TA, ggcd2Tr, ggcd2TV, ggcd2TM, ggcd2TT, ggcd2TP,
     ggcd2TR, ggcd2TS, ggcd2TL, ggcd2TX}
    ,
    {gmather, ggcd2Pv, ggcd2Pa, ggcd2Pt, ggcd2Pp, ggcd2PD,
     ggcd2PA, ggcd2Pr, ggcd2PV, ggcd2PM, ggcd2PT, ggcd2PP,
     ggcd2PR, ggcd2PS, ggcd2PL, ggcd2PX}
    ,
    {gmather, ggcd2Rv, ggcd2Ra, ggcd2Rt, ggcd2Rp, ggcd2RD,
     ggcd2RA, ggcd2Rr, ggcd2RV, ggcd2RM, ggcd2RT, ggcd2RP,
     ggcd2RR, ggcd2RS, ggcd2RL, ggcd2RX}
    ,
    {gmather, ggcd2Sv, ggcd2Sa, ggcd2St, ggcd2Sp, ggcd2SD,
     ggcd2SA, ggcd2Sr, ggcd2SV, ggcd2SM, ggcd2ST, ggcd2SP,
     ggcd2SR, ggcd2SS, ggcd2SL, ggcd2SX}
    ,
    {gmather, ggcd2Lv, ggcd2La, ggcd2Lt, ggcd2Lp, ggcd2LD,
     ggcd2LA, ggcd2Lr, ggcd2LV, ggcd2LM, ggcd2LT, ggcd2LP,
     ggcd2LR, ggcd2LS, ggcd2LL, ggcd2LX}
    ,
    {gmather, ggcd2Xv, ggcd2Xa, ggcd2Xt, ggcd2Xp, ggcd2XD,
     ggcd2XA, ggcd2Xr, ggcd2XV, ggcd2XM, ggcd2XT, ggcd2XP,
     ggcd2XR, ggcd2XS, ggcd2XL, ggcd2XX}
};



Sex
ggcd2(a, b)
     Sex             a,
                     b;
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return gcd(a, b);
	else if (is_vechdr(b))
	    return ggcd_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return ggcd_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return ggcd_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    }
    return aerror2("ggcd2", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
/***************************************************************/
/***************************************************************/
