
/*
 * gcd.c      (C) Copyright 2015 by Fujio Kako, Nara Women's University.
 * Greatest Common Divisor Functions.
 *      $Header: /Users/kako/asuka/src/RCS/gcd.c,v 2.57 2016/03/31 07:22:05 kako Exp kako $
 */

static char const rcsid[] =
    "$Id: gcd.c,v 2.57 2016/03/31 07:22:05 kako Exp kako $";

#define DEBUG
#define USE_HEURISTIC

// #define GCD_TEST

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "entries.h"
#include "misc.h"
#include "fns.h"
#include "polynomial.h"
#include "tri.h"
#include "garith.h"

// #define gquot2 wquo

#define cddr(x)	cdr(cdr(x))
#define cdar(x)	cdr(car(x))
#define caar(x)	car(car(x))


#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)
#define lcoeff(a)	caar(a)
#define lpower(a)	cdar(a)
#define ZEROP(a)	zerop(a)


static Sex      Afactor(Sex a);
Sex             Asquarefree(Sex a);
Sex             Asquarefree(Sex a);


extern Sex      wuremainder(Sex a, Sex b);

extern Sex      Lncons(Sex a);
extern Sex      Lnconc(Sex a, Sex b);
extern Sex      Laconc(Sex a, Sex b);
extern Sex      Lclone(Sex a);
extern Sex      Lmkarray1(Sex n);
extern Sex      Alist2array(Sex a, Sex b);
extern bool     glessp2(Sex a, Sex b);

Sex             my_gcd_mgcd(Sex, Sex);
static Sex      my_gcd_ezgcd(Sex, Sex);
static Sex      my_gcd_ezgcd1(Sex, Sex);
static Sex      my_gcd_ezgcd2(Sex, Sex);
static Sex      my_gcd_ezgcd3(Sex, Sex);
static Sex      my_gcd_ezgcd4(Sex, Sex);

static Sex      Afactor_mod(Sex f);
static Sex      Afactor_ff(Sex f);
static Sex      Asquarefree2(Sex a);
static bool     square_free_mod(Sex a, Sex p);
static Sex      factor_mod(Sex f, Sex p);
static Sex      sqfree(Sex a);
static Sex      ddd_factor(Sex a);
static Sex      berlekamp(Sex f, Sex p);
static Sex      MignotteBound(Sex f);
static Sex      uhensel(Sex f, Sex p, Sex g, Sex a, Sex b);
static Sex      FormMatrix(Sex f, Sex q);
static Sex      NullSpaceBase(Sex c);
static Sex      Factor_mod(Sex V, Sex k, Sex q, Sex f);
static Sex      isqrt(Sex s);
static bool     zeroRow(Sex M, long n, long i);
static Sex      Combine_Factors(Sex F, Sex U, Sex q, long n, long m);
static Sex      Combine_nest(Sex F, Sex U, Sex q, long k, long c, long n,
			     long m, Sex a);

static Sex      eea(Sex a, Sex b);
static Sex      multieea(Sex a, Sex x);
static Sex      diophant(Sex a1, Sex a2, Sex c, Sex x);




extern Sex      Amcoeff(Sex a);
extern Sex      Lexpt(Sex a, Sex n);
extern Sex      gexpt(Sex a, Sex n);
extern Sex      Mpp(Sex a);
extern Sex      Mvar(Sex a);
extern Sex      Mdeg(Sex a);
extern Sex      Adeg(Sex f, Sex x);
extern Sex      Lrandom(Sex a);
extern Sex      Lnext_prime(Sex p);
extern Sex      Mlcoeff(Sex a);
extern Sex      Mlpower(Sex a);
extern Sex      Mlterm(Sex a);
extern Sex      Alcoeff(Sex a, Sex x);
extern Sex      Alpower(Sex a, Sex x);
extern Sex      Alterm(Sex a, Sex x);
extern Sex      Amkmodular(Sex a, Sex p);
extern Sex      Aunmodular(Sex a);
extern Sex      Adiff(Sex a, Sex c);
extern Sex      new_vector(long n);
extern Sex      Lident(Sex a);
extern Sex      Labs(Sex a);
extern Sex      Arowvector(Sex m, Sex i);
extern Sex      Acoeff(Sex a, Sex b);
extern Sex      make_matrix(Sex n, Sex m);
extern Sex      Llessp(Sex a, Sex b);

static Sex      mgcd(Sex A, Sex B);
static Sex      MGCD(Sex A, Sex B, Sex x, Sex r);
static Sex      NewPrime(Sex la, Sex lb);
static Sex      NewLargePrime(void);
static Sex      norm_inf(Sex f);
static Sex      icont(Sex f);
static Sex      PGCD(Sex A, Sex B, Sex x, long k, Sex p);
static Sex      PolyInterp(Sex q, Sex h, Sex a, Sex f, Sex x, Sex p);
static Sex      conts(Sex f, Sex x, long k);
static Sex      conts_modp(Sex f, Sex x, long k);
static Sex      lcoeffs(Sex f, Sex x, long k);
static Sex      NewRandom(Sex p, Sex g);
static Sex      ugcd_modp(Sex A, Sex B);	/* A and B is in Zp[x] */
static Sex      gcd_modp(Sex A, Sex B);	/* A and B is in Zp[x1, ...] */
static Sex      ugcd(Sex A, Sex B);	/* A and B is in Z[x] */
#if 0
static Sex      igcd(Sex a, Sex b);	/* gcd of integers a and b */
#endif

extern Sex      Aunknowns(Sex A);
extern Sex      Lunion(Sex A, Sex B);
extern Sex      Lmin2(Sex a, Sex b);
extern Sex      Lmax2(Sex a, Sex b);
extern Sex      Asub(Sex f, Sex x, Sex v);

static Sex      ezgcd(Sex A, Sex B);
static Sex      EZGCD(Sex A, Sex B, Sex x, long k);
static Sex      NewEvaluationPoint(Sex a, Sex b, Sex x, long k, Sex p);
static Sex      NewEvaluationPoint1(Sex a, Sex b, Sex x, long k, Sex p);
static Sex      NewEvaluationPoint2(Sex a, Sex b, Sex x, long k, Sex p);
static int      new_evaluation_point = 2;
static int      use_heuristic = 1;
static int      ezgcd_debug = 0;
#if 0
static Sex      SGCD(Sex a, Sex b, Sex x, Sex v, Sex p);
static Sex      SquareFree(Sex a);
#endif
static Sex      EZ_LIFT(Sex U, Sex C, Sex H, Sex x, Sex b, Sex p, Sex lc,
			Sex lu);
static Sex      Mignotte(Sex f);
static Sex      MignotteBound2(Sex f, Sex g, Sex x);
static Sex      isqrt(Sex m);	/* integer square root of m */
static Sex      conts2(Sex f, Sex x, long k);
static Sex      EEA(Sex a, Sex b, Sex x);
static Sex      replace_lc(Sex f, Sex a, Sex x);
static Sex      multivariate_replace_lc(Sex f, Sex a);
static Sex      UnivariateHensel(Sex a, Sex p, Sex u1, Sex w1, Sex B,
				 Sex gamma);
static Sex      TaylerCoeff(Sex f, Sex x, long k, Sex a);
static void     print_e(Sex f, Sex v, Sex p);
static Sex      MultivariateHensel(Sex a, Sex x, Sex v, Sex p, Sex pl,
				   long l, Sex u, Sex w, Sex lcu, Sex lcw);
static Sex      MultivariateDiophant(Sex u, Sex w, Sex c, Sex xv, Sex av,
				     long n, long d, Sex p, Sex pk,
				     long k);
static Sex      UnivariateDiophant(Sex u, Sex w, Sex x, Sex c, Sex p,
				   Sex pk, long k);
static Sex      EEAlift(Sex a, Sex b, Sex x, Sex p, Sex pk, long k);

static Sex      cont(Sex a);
static Sex      pp(Sex a);

static Sex      cont_mgcd(Sex a);
static Sex      pp_mgcd(Sex a);
static Sex      cont_ezgcd(Sex a);
static Sex      pp_ezgcd(Sex a);
static Sex      cont_sprs(Sex a);
static Sex      pp_sprs(Sex a);
static Sex      my_gcd_sprs(Sex a, Sex b);
static Sex      cont_pprs(Sex a);
static Sex      pp_pprs(Sex a);
static Sex      my_gcd_pprs(Sex a, Sex b);

static Sex      cont_rprs(Sex a);
static Sex      pp_rprs(Sex a);
static Sex      my_gcd_rprs(Sex a, Sex b);
static Sex      my_gcd_rprs_aux(Sex a, Sex b);

static Sex      cont_euclid(Sex a);
static Sex      pp_euclid(Sex a);
static Sex      my_gcd_euclid(Sex a, Sex b);

static Sex      my_gcd_heuristic(Sex a, Sex b);
static Sex      my_gcd_heuristic_aux(Sex a, Sex b);
static Sex      my_gcd_combine1(Sex a, Sex b);
static Sex      my_gcd_combine2(Sex a, Sex b);
static Sex      quotient_by_int(Sex a, Sex b);
static Sex      remainder_by_int2(Sex a, Sex b, Sex c);
static Sex      remainder_by_int(Sex a, Sex b);
static bool     dividable_by_int(Sex a, Sex b);
static Sex      wrem(Sex a, Sex b);
static Sex      wrem_sub(Sex u, Sex a, long ad, Sex v, Sex b, long bd);
static Sex      wquo(Sex a, Sex b);
static Sex      wquo_sub(Sex u, Sex v, Sex b, long bd);

Sex             star_evaluation_points;

/**************************************************************/
/**************************************************************/

#define WREM
#define gquot2 wquo
#define wuremainder wrem

static          Sex
cont_mgcd(Sex a)
{				/* 
				 * GCD
				 */
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
	p = my_gcd_mgcd(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}

static          Sex
pp_mgcd(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont_mgcd(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}


Sex
my_gcd_mgcd(Sex a, Sex b)
{
    stackcheck();
    if (is_number(a)) {
	if (is_number(b))
	    return gcd(a, b);
	push(a);
	b = icont(b);
	pop(a);
	return gcd(a, b);
    }
    if (is_number(b)) {
	push(b);
	a = icont(a);
	pop(b);
	return gcd(a, b);
    }

    return mgcd(a, b);
}

static          Sex
cont_ezgcd(Sex a)
{				/* 
				 * GCD
				 */
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
#ifdef  USE_HEURISTIC
	p = use_heuristic ? my_gcd_heuristic(p, q) : Mgcd(p, q);
#else
	p = Mgcd(p, q);
#endif
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}

static          Sex
pp_ezgcd(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont_ezgcd(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}

static          Sex
my_gcd_ezgcd1(Sex a, Sex b)
{
    new_evaluation_point = 1;
    use_heuristic = 0;
    return my_gcd_ezgcd(a, b);
}

static          Sex
my_gcd_ezgcd2(Sex a, Sex b)
{
    new_evaluation_point = 2;
    use_heuristic = 0;
    return my_gcd_ezgcd(a, b);
}

static          Sex
my_gcd_ezgcd3(Sex a, Sex b)
{
    new_evaluation_point = 1;
    use_heuristic = 1;
    return my_gcd_ezgcd(a, b);
}

static          Sex
my_gcd_ezgcd4(Sex a, Sex b)
{
    new_evaluation_point = 2;
    use_heuristic = 1;
    return my_gcd_ezgcd(a, b);
}

static          Sex
my_gcd_ezgcd(Sex a, Sex b)
{
    stackcheck();
    if (is_number(a)) {
	if (is_number(b))
	    return gcd(a, b);
	push(a);
	b = icont(b);
	pop(a);
	return gcd(a, b);
    }
    if (is_number(b)) {
	push(b);
	a = icont(a);
	pop(b);
	return gcd(a, b);
    }
    return ezgcd(a, b);
}
/**************************************************************/
/**************************************************************/

/*
 * Modular GCD algorithm 
 */

static          Sex
mgcd(Sex A, Sex B)
{
    Sex             b;
    stackcheck();
    push2(A, B);
    pushv(4);
#define A stack[-5]
#define B stack[-4]
#define v stack[-3]
#define a stack[-2]
#define n stack[-1]
#define p stack[-0]
    // begin local v, a, n, p;
    a = Aunknowns(A);
    b = Aunknowns(B);
    v = Lunion(a, b);
    n = Llength(v);
    a = Lmkarray1(n);
    Alist2array(a, v);
    b = MGCD(A, B, a, n);
    popv(6);
#undef A
#undef B
#undef v
#undef a
#undef n
#undef p
    return b;
}



static          Sex
MGCD(Sex A, Sex B, Sex x, Sex r)
{
    long            n,
                    m,
                    k;
    Sex             u;
    long            da,
                    db;

    k = mkint(r);

    stackcheck();
    push4(A, B, x, r);
    pushv(16);
#define A stack[-19]
#define B stack[-18]
#define x stack[-17]
#define r stack[-16]
#define a stack[-15]
#define b stack[-14]
#define c stack[-13]
#define g stack[-12]
#define q stack[-11]
#define H stack[-10]
#define p stack[-9]
#define Ap stack[-8]
#define Bp stack[-7]
#define Cp stack[-6]
#define gp stack[-5]
#define aa stack[-4]
#define bb stack[-3]
#define s stack[-2]
#define lim stack[-1]
#define C stack[-0]
    // local a, b, c, g, q, H;
    // local p, Ap, Bp, Cp, gp, m, n;
    // local s;
    // local lim;
    // local C;

    a = icont(A);
    b = icont(B);
    A = gquot2(A, a);
    B = gquot2(B, b);
    c = gcd(a, b);
    aa = lcoeffs(A, x, k);
    bb = lcoeffs(B, x, k);
    g = gcd(aa, bb);
    q = mkfxat(1);
    H = mkfxat(1);
    da = mkint(Adeg(A, array_value(x, k - 1)));
    db = mkint(Adeg(B, array_value(x, k - 1)));
    n = da > db ? db : da;
    aa = norm_inf(A);
    bb = norm_inf(B);
    aa = Lmin2(aa, bb);
    bb = expti(mkfxat(2), n);
    lim = Labs(g);
    lim = gtimes2(lim, bb);
    lim = gtimes2(lim, aa);

    while (1) {
	p = NewPrime(g, q);


	Ap = Amkmodular(A, p);
	Bp = Amkmodular(B, p);
	gp = Amkmodular(g, p);
	/*
	 * write "calling PGCD with k=", k;
	 */
	// if(ezgcd_debug)lisp_printf("calling PGCD with k=%d\n", k);
	Cp = PGCD(Ap, Bp, x, k, p);
	// if(ezgcd_debug)lisp_printf("return from PGCD\n");
	/*
	 * write "return from PGCD";
	 */

	u = Adeg(Cp, array_value(x, k - 1));
	m = mkint(u);

	/*
	 * normalize Cp 
	 */
	aa = lcoeffs(Cp, x, k);
	Cp = gquot2(Cp, aa);
	Cp = gtimes2(Cp, gp);


	if (m < n) {
	    /*
	     * write "mgcd: bad homology, m=", m, ", n=", n;; 
	     */
	    // if(ezgcd_debug)lisp_printf("mgcd: bad homology, m=%d,
	    // n=%d\n", m, n);
	    q = p;

	    H = Aunmodular(Cp);
	    n = m;
	} else if (m == n) {

	    s = Amkmodular(q, p);
	    s = gquot2(mkfxat(1), s);
	    aa = gdiff2(Cp, H);
	    s = gtimes2(aa, s);

	    aa = Aunmodular(s);
	    aa = gtimes2(q, aa);
	    H = gplus2(H, aa);
	    q = gtimes2(q, p);
	}


	if (lessp2(lim, q)) {
	    s = icont(H);
	    C = gratio2(H, s);
	    aa = gremaind2(A, C);
	    if (gzerop(aa)) {
		bb = gremaind2(B, C);
		if (gzerop(bb)) {
		    u = gtimes2(c, C);
		    break;
		}
	    }
	} else if (m == 0) {
	    s = icont(H);
	    C = gratio2(H, s);
	    aa = gremaind2(A, C);
	    if (gzerop(aa)) {
		bb = gremaind2(B, C);
		if (gzerop(bb)) {
		    u = gtimes2(c, C);
		    break;
		}
	    }
	    u = c;
	    break;
	}
    }
    popv(20);
#undef A
#undef B
#undef x
#undef r
#undef a
#undef b
#undef c
#undef g
#undef q
#undef H
#undef p
#undef Ap
#undef Bp
#undef Cp
#undef gp
#undef aa
#undef bb
#undef s
#undef lim
#undef C
    return u;
}

static          Sex
NewPrime(Sex la, Sex lb)
{
    Sex             r;
    stackcheck();
    push3(la, lb, nil);
#define la stack[-2]
#define lb stack[-1]
#define p stack[-0]
    p = mkfxat(100000);
    p = Lrandom(p);
    p = plus2(p, mkfxat(10000));
    p = Lnext_prime(p);
    while (dividable_by_int(la, p) || dividable_by_int(lb, p)) {
	p = Lnext_prime(p);
    }
    r = p;
#undef la
#undef lb
#undef p
    popv(3);
    return r;
}

static          Sex
NewLargePrime()
{
    Sex             r;
    long            n;
    stackcheck();
    push(nil);
#define p stack[0]
    p = mkfxat(100000);
    p = Lrandom(p);
    p = plus2(p, mkfxat(10000));
    p = Lnext_prime(p);
#undef p
    pop(r);
    return r;
}

static          Sex
norm_inf(Sex f)
{
    Sex             r;
    Sex             v;
    long            i,
                    n;
    if (!is_poly(f))
	return Labs(f);

    // if(ezgcd_debug)lisp_printf("norm: f=%p\n", f);
    stackcheck();
    push(f);
    pushv(3);
#define f stack[-3]
#define c stack[-2]
#define m stack[-1]
#define a stack[-0]
    // begin local c, i, n, m;
    f = poly_pol(f);
    m = norm_inf(lcoeff(f));
    f = cdr(f);
    while (f != nil) {
	a = norm_inf(lcoeff(f));
	m = Lmax2(m, a);
	f = cdr(f);
    }

    /*
     * v = Mvar(f); c = Acoeff(f, v); n = mkint(car(array_dim(c))); m =
     * norm_inf(array_value(c, n - 1)); for (i = 0; i < n - 1; i++) { a =
     * array_value(c, i); if (!gzerop(a)) { a = norm_inf(a); m = Lmax2(m,
     * a); } } 
     */
    r = m;
    popv(4);
#undef f
#undef c
#undef m
#undef a
    return r;
}

static          Sex
icont(Sex f)
{
    long            i,
                    n;
    Sex             a;
    if (!is_poly(f))
	return f;
    stackcheck();
    push(f);
    pushv(4);
    // begin local c, i, p, g;
#define f stack[-4]
#define c stack[-3]
#define p stack[-2]
#define g stack[-1]
    c = Amcoeff(f);
    n = mkint(car(Asize(c)));
    g = icont(array_value(c, n - 1));
    for (i = 0; i < n - 1; i++) {
	if (onep(g)) {
	    popv(5);
	    return mkfxat(1);
	}
	p = array_value(c, i);
	if (!zerop(p)) {
	    p = icont(p);
	    g = gcd(g, p);
	}
    }
    a = g;
    popv(5);
#undef f
#undef c
#undef p
#undef g
    return a;
}



static          Sex
PGCD(Sex A, Sex B, Sex x, long k, Sex p)
{
    Sex             r;
    long            i,
                    j;
    long            n,
                    m,
                    lim;
    long            aa,
                    bb;
    Sex             C;

    if (k == 1) {
	C = my_gcd_sprs(A, B);
	n = mkint(Mdeg(C));
	if (n == 0)
	    C = mkfxat(1);
	return C;
    }
    stackcheck();
    push4(A, B, x, p);
    pushv(14);

#define A stack[-17]
#define B stack[-16]
#define x stack[-15]
#define p stack[-14]
#define a stack[-13]
#define b stack[-12]
#define qb stack[-11]
#define hb stack[-10]
#define c stack[-9]
#define C stack[-7]
#define Cb stack[-6]
#define q stack[-5]
#define H stack[-4]
#define g stack[-3]
#define gb stack[-2]
#define Bb stack[-1]
#define Ab stack[-0]
    // begin local i, j;
    // local a, b, qb, hb;
    // local c, C, Cb;
    // local m, n, lim;
    // local q, H, g;
    // local gb, Bb, Ab;



    a = conts_modp(A, x, k - 1);	/* A is in Zp[xk][x1, ..., xk-1] */
    b = conts_modp(B, x, k - 1);
    A = gquot2(A, a);
    B = gquot2(B, b);
    c = ugcd_modp(a, b);
    Ab = lcoeffs(A, x, k - 1);
    Bb = lcoeffs(B, x, k - 1);
    g = ugcd_modp(Ab, Bb);
    q = mkfxat(1);
    H = mkfxat(1);

    aa = mkint(Adeg(A, array_value(x, k - 1)));
    bb = mkint(Adeg(B, array_value(x, k - 1)));
    n = aa > bb ? bb : aa;
    lim = n + mkint(Adeg(g, array_value(x, k - 1)));

    aa = mkint(Adeg(A, array_value(x, k - 2)));
    bb = mkint(Adeg(B, array_value(x, k - 2)));
    n = aa > bb ? bb : aa;

    while (1) {
	while (1) {
	    b = Lrandom(p);
	    b = Amkmodular(b, p);
	    if (zerop(Asub(g, array_value(x, k - 1), b)))
		continue;
	    if (zerop(Asub(q, array_value(x, k - 1), b)))
		continue;
	    break;
	}
	// if(ezgcd_debug)lisp_printf("pgcd loop: b = %p\n", b);
	Ab = Asub(A, array_value(x, k - 1), b);
	Bb = Asub(B, array_value(x, k - 1), b);
	/*
	 * write "call PGCD in PGCD with k=", k-1; write "A=", Ab; write
	 * "B=", Bb; 
	 */
	// lisp_printf("calling PGCD with k=%d, A=%p, B=%p\n", k-1, Ab,
	// Bb);
	Cb = PGCD(Ab, Bb, x, k - 1, p);
	// lisp_printf("return from PGCD with C=%p, H=%p\n", Cb, H);
	/*
	 * write "return from pgcd"; write "C=", Cb; write "H=", H; 
	 */

	m = mkint(Adeg(Cb, array_value(x, k - 2)));

	gb = Asub(g, array_value(x, k - 1), b);
	r = lcoeffs(Cb, x, k - 1);
	Cb = gquot2(Cb, r);
	Cb = gtimes2(Cb, gb);
	// lisp_printf("Cb = %p\n", Cb);
	// lisp_printf("q = %p\n", q);

	/*
	 * write "pgcd: m=", m, ", n=", n;; 
	 */
	// lisp_printf("pgcd: m=%d, n=%d\n", m, n);
	if (m < n) {
	    /*
	     * write "pgcd: bad homology, m=", m, ", n=", n;; 
	     */
	    // lisp_printf("pgcd: bad homology, m=%d, n=%d\n", m, n);
	    /*
	     * q:=1; H:=1; 
	     */

	    q = symbol_to_polynomial(array_value(x, k - 1));
	    q = gdiff2(q, b);
	    H = Cb;

	    n = m;
	} else if (m == n) {
	    /*
	     * q := PolyInterp(q, H, b, Cb, x[k-1], p); H:=second(q);
	     * q:=first(q); 
	     */
	    qb = Asub(q, array_value(x, k - 1), b);
	    hb = Asub(H, array_value(x, k - 1), b);
	    // lisp_printf("qb = %p, hb = %p\n", qb, hb);
	    qb = gratio2(q, qb);
	    hb = gdiff2(Cb, hb);
	    qb = gtimes2(qb, hb);
	    // lisp_printf("H(before) = %p\n", H);
	    H = gplus2(H, qb);
	    // lisp_printf("H(after) = %p\n", H);
	    qb = symbol_to_polynomial(array_value(x, k - 1));
	    qb = gdiff2(qb, b);
	    q = gtimes2(q, qb);
	}
	/*
	 * write "lcoeff H = ", lcoeffs(H, x, k-1); write "g=", g; write
	 * "g-lcoeff(h) = ", lcoeffs(H, x, k-1) - g; 
	 */
	qb = lcoeffs(H, x, k - 1);
	// lisp_printf("lcoeff H=%p\n", qb);
	// lisp_printf("g=%p\n", g);
	qb = gdiff2(qb, g);
	// lisp_printf("g-lcoeff(h) = %p\n", qb);
	if (gzerop(qb)) {
	    qb = conts_modp(H, x, k - 1);
	    // lisp_printf("icont = %p\n", qb);
	    C = gquot2(H, qb);

	    /*
	     * write "C=", C; write "A % C = ", gremainder(A, C); write "B 
	     * % C = ", gremainder(B, C); 
	     */
	    // lisp_printf("C=%p\nA %% C = %p\nB%%C = %p\n", C,
	    // gremaind2(A, C), gremaind2(B, C));
	    if (zerop(gremaind2(A, C))) {
		if (zerop(gremaind2(B, C))) {
		    r = gtimes2(c, C);
		    popv(18)
			return r;
		}
	    }
#if 0
	    if (m == 0) {
		r = c;
		popv(18)
		    return r;
	    }
#endif
	}
    }
#undef A
#undef B
#undef x
#undef p
#undef a
#undef b
#undef qb
#undef hb
#undef c
#undef C
#undef Cb
#undef q
#undef H
#undef g
#undef gb
#undef Bb
#undef Ab
}

static          Sex
PolyInterp(Sex q, Sex h, Sex a, Sex f, Sex x, Sex p)
{
    Sex             r;
    stackcheck();
    push5(q, h, a, f, x);
    push(p);
    pushv(2);
#define q stack[-7]
#define h stack[-6]
#define a stack[-5]
#define f stack[-4]
#define x stack[-3]
#define p stack[-2]
#define qa stack[-1]
#define ha stack[-0]

    qa = Asub(q, x, a);
    ha = Asub(h, x, a);
    ha = gdiff2(f, ha);
    qa = gquot2(q, qa);
    qa = gtimes2(qa, ha);
    h = gplus2(h, qa);
    qa = gdiff2(x, a);
    q = gtimes2(q, qa);
    r = Llist2(q, h);
    popv(8);
#undef q
#undef h
#undef a
#undef f
#undef x
#undef p
#undef qa
#undef ha
    return r;
}






static          Sex
conts(Sex f, Sex x, long k)
{
    long            i,
                    n;
    Sex             r;

    if (is_number(f)) {
	f = Labs(f);
	return f;
    } else if (k == 0)
	return f;

    stackcheck();
    push2(f, x);
    pushv(3);
#define f stack[-4]
#define x stack[-3]
#define c stack[-2]
#define g stack[-1]
#define gg stack[-0]
    c = Acoeff(f, array_value(x, k - 1));
    n = mkint(car(array_dim(c)));
    g = array_value(c, n - 1);
    g = conts(g, x, k - 1);
    for (i = 0; i <= n - 2; i++) {
	if (gonep(g)) {
	    g = mkfxat(1);
	    break;
	}
	gg = conts(array_value(c, i), x, k - 1);
	if (gonep(gg)) {
	    g = mkfxat(1);
	    break;
	} else if (!gzerop(gg)) {
	    // lisp_printf("conts: gcd(%p, %p)\n", g, gg);
#ifdef USE_HEURISTIC
	    g = use_heuristic ? my_gcd_heuristic(g, gg) : Mgcd(g, gg);
#else
	    g = Mgcd(g, gg);
#endif
	    // g = my_gcd_mgcd(g, gg);
	}
    }
    r = g;
    popv(5);
#undef f
#undef x
#undef c
#undef g
#undef gg
    return r;
}

static          Sex
conts_modp(Sex f, Sex x, long k)
{
    long            i,
                    n;
    Sex             r;

    if (is_number(f)) {
	f = Labs(f);
	return f;
    } else if (k == 0)
	return f;

    stackcheck();
    push2(f, x);
    pushv(3);
#define f stack[-4]
#define x stack[-3]
#define c stack[-2]
#define g stack[-1]
#define gg stack[-0]
    c = Acoeff(f, array_value(x, k - 1));
    n = mkint(car(array_dim(c)));
    g = array_value(c, n - 1);
    g = conts_modp(g, x, k - 1);
    for (i = 0; i <= n - 2; i++) {
	if (gonep(g)) {
	    g = mkfxat(1);
	    break;
	}
	gg = conts_modp(array_value(c, i), x, k - 1);
	if (gonep(gg)) {
	    g = mkfxat(1);
	    break;
	} else if (!gzerop(gg)) {
	    // lisp_printf("conts: gcd(%p, %p)\n", g, gg);
	    g = gcd_modp(g, gg);
	}
    }
    r = g;
    popv(5);
#undef f
#undef x
#undef c
#undef g
#undef gg
    return r;
}


static          Sex
lcoeffs(Sex f, Sex x, long k)
{
    long            i;
    Sex             r;

    stackcheck();
    push2(f, x);
#define f stack[-1]
#define x stack[-0]
    for (i = 0; i < k; i++) {
	f = Alcoeff(f, array_value(x, i));
	if (!is_poly(f))
	    break;
    }
    r = f;
#undef f
#undef x
    popv(2);
    return r;
}




static          Sex
NewRandom(Sex p, Sex g)
{
    Sex             r;
    stackcheck();
    push2(p, g);
    pushv(3);
#define p stack[-4]
#define g stack[-3]
#define x stack[-2]
#define a stack[-1]
#define b stack[-0]
    // begin local x;
    // local a, b;
    x = Mvar(g);
    while (1) {
	b = Lrandom(p);
	a = Asub(g, x, b);
	if (!gzerop(a))
	    break;
    }
    r = b;
    popv(5);
#undef p
#undef g
#undef x
#undef a
#undef b
    return r;
}


static          Sex
ugcd(Sex A, Sex B)
{				/* A and B is in Zp[x] */
    Sex             s;
    if (gzerop(A))
	return B;
    if (gzerop(B))
	return A;
    if (!is_poly(A) || !is_poly(B))
	return mkfxat(1);
    stackcheck();
    // if(ezgcd_debug)lisp_printf("ugcd: %p, %p\n", A, B);
    push2(A, B);
    pushv(2);
#define A stack[-3]
#define B stack[-2]
#define r stack[-1]
#define x stack[-0]
    // begin local r, x;

    x = Mvar(A);
    // if(ezgcd_debug)lisp_printf("ugcd: mvar = %p\n", x);
    r = grremaind2(A, B, x);
    while (is_poly(r)) {
	A = B;
	B = r;
	r = grremaind2(A, B, x);
    }
    if (!zerop(r))
	B = mkfxat(1);
    // if(ezgcd_debug)lisp_printf("ugcd return %p\n", B);
    s = B;
    popv(4);
#undef A
#undef B
#undef r
#undef x
    return s;
}


static          Sex
ugcd_modp(Sex A, Sex B)
{				/* A and B is in Zp[x] */
    Sex             s;
    if (gzerop(A))
	return B;
    if (gzerop(B))
	return A;
    if (!is_poly(A) || !is_poly(B))
	return mkfxat(1);
    stackcheck();
    // if(ezgcd_debug)lisp_printf("ugcd: %p, %p\n", A, B);
    push2(A, B);
    pushv(2);
#define A stack[-3]
#define B stack[-2]
#define r stack[-1]
#define x stack[-0]
    // begin local r, x;

    x = Mvar(A);
    // if(ezgcd_debug)lisp_printf("ugcd: mvar = %p\n", x);
    r = grremaind2(A, B, x);
    while (is_poly(r)) {
	A = B;
	B = r;
	r = grremaind2(A, B, x);
    }
    if (!zerop(r))
	B = mkfxat(1);
    // if(ezgcd_debug)lisp_printf("ugcd return %p\n", B);
    s = B;
    popv(4);
#undef A
#undef B
#undef r
#undef x
    return s;
}

static          Sex
gcd_modp(Sex A, Sex B)
{				/* A and B is in Zp[x1, ...] */
    return my_gcd_euclid(A, B);
}


#if 0
static          Sex
igcd(Sex a, Sex b)
{				/* gcd of integers a and b */
    Sex             s;
    if (zerop(a))
	return b;
    if (zerop(b))
	return a;
    stackcheck();
    push2(a, b);
    push(nil);
#define a stack[-2]
#define b stack[-1]
#define r stack[-0]
    // begin local r;
    r = remaind2(a, b);
    while (!zerop(r)) {
	a = b;
	b = r;
	r = remaind2(a, b);
    }
    s = b;
    popv(3);
#undef a
#undef b
#undef r
    return s;
}
#endif

/*
 * A:=9*x^5+2*x^4*y*z-189*x^3*y^3*z + 117*x^3*y*z + 3*x^3 -42*x^2*y^4*z^2+ 
 26*x^2*y^2*z^3 + 18*x^2 -63*x*y^3*z + 39*x*y*z^2 + 4*x*y*z + 6;
 * B:=6*x^6-126*x^4*y^3*z + 78*x^4*y*z^2 + x^4*y + x^4*z + 13*x^3 -
 21*x^2*y^4*z -21*x^2*y^3*z^2 + 13*x^2*y^2*z^2 + 13*x^2 *y * z^3 -
 21*x*y^3 * z + 13*x*y*z^2 + 2*x*y + 2*x*z +2;
 * C:=z^3+z^2+z;
 * MGCD(6*A*C, 9*B*C, xx, 3);
 */

/**************************************************************/
/**************************************************************/

/*
 * Extended Zassenhaus GCD algorithm 
 */

// import "hensel.m";

static          Sex
ezgcd(Sex A, Sex B)
{
    Sex             r;
    long            n;
    stackcheck();
    push2(A, B);
    pushv(3);
#define A stack[-4]
#define B stack[-3]
#define v stack[-2]
#define a stack[-1]
    stack[0] = star_evaluation_points;
    star_evaluation_points = nil;
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("ezgcd: %p , %p\n", A, B);
#endif

    // begin local v, a, n, p;
    v = Aunknowns(B);
    r = Aunknowns(A);
    v = Lunion(r, v);
    n = mkint(Llength(v));
    a = mkarray(n);
    Alist2array(a, v);
    r = EZGCD(A, B, a, n);
    r = car(cddr(r));
    star_evaluation_points = stack[0];
    popv(5);
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("ezgcd returns %p\n", r);
#endif

#undef A
#undef B
#undef v
#undef a
    return r;
}


static          Sex
NewEvaluationPoint(Sex a, Sex b, Sex x, long k, Sex p)
{
    if (new_evaluation_point == 1)
	return NewEvaluationPoint1(a, b, x, k, p);
    else
	return NewEvaluationPoint2(a, b, x, k, p);
}


static          Sex
NewEvaluationPoint2(Sex a, Sex b, Sex x, long k, Sex p)
{
    Sex             r;
    long            i;
    long            j;
    long            l;
    long            n;

    if (k == 1) {
	return mkarray(1);
    }
    stackcheck();
    push4(a, b, x, p);
    pushv(5);
#define a stack[-8]
#define b stack[-7]
#define x stack[-6]
#define p stack[-5]
#define B stack[-4]
#define ca stack[-3]
#define cb stack[-2]
#define c stack[-1]
#define y stack[-0]
    // local j, B, ca, cb;

    if (star_evaluation_points == nil) {
	c = mkarray(k);
    } else {
	l = list_length(star_evaluation_points);
	l = my_random(l - 1);
	c = star_evaluation_points;
	for (; l > 0; l--)
	    c = cdr(c);
	c = Lclone(car(c));
    }
    while (1) {
	j = my_random(k - 1) + 1;
	if (j > k - 1)
	    continue;
	i = my_random(2);
	n = mkint(array_value(c, j));
	array_value(c, j) = mkfxat(n + 1 - 2 * i);
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("new eval point2: c[%d] = %d\n", j, n + 1 - 2 * i);
#endif
	ca = Asub(a, array_value(x, 1), array_value(c, 1));
	cb = Asub(b, array_value(x, 1), array_value(c, 1));
	for (j = 2; j < k; j++) {
	    ca = Asub(ca, array_value(x, j), array_value(c, j));
	    ca = remainder_by_int(ca, p);
	    if (zerop(ca))
		break;
	    cb = Asub(cb, array_value(x, j), array_value(c, j));
	    cb = remainder_by_int(cb, p);
	    if (zerop(cb))
		break;
	}
	if (!zerop(ca) && !zerop(cb)
	    && Lmember(c, star_evaluation_points) == nil) {
	    star_evaluation_points = cons(c, star_evaluation_points);
	    r = c;
	    break;
	}
    }
    popv(9);
#undef a
#undef b
#undef x
#undef p
#undef B
#undef ca
#undef cb
#undef c
#undef y
    return r;
}



static          Sex
NewEvaluationPoint1(Sex a, Sex b, Sex x, long k, Sex p)
{
    Sex             r;
    long            j;
    long            n;
    stackcheck();
    push4(a, b, x, p);
    pushv(5);
#define a stack[-8]
#define b stack[-7]
#define x stack[-6]
#define p stack[-5]
#define B stack[-4]
#define ca stack[-3]
#define cb stack[-2]
#define c stack[-1]
#define y stack[-0]
    // local j, B, ca, cb;

    B = mkarray(k);
    for (j = 1; j < k; j++) {
	y = array_value(x, j);
#if 1
	do {
	    do {
		c = Lrandom(p);
		ca = Asub(a, y, c);
		ca = remainder_by_int(ca, p);
	    } while (zerop(ca));
	    cb = Asub(b, y, c);
	    cb = remainder_by_int(cb, p);
	} while (zerop(cb));
#else
	c = mkfxat(0);
	ca = Asub(a, y, mkfxat(0));
	cb = Asub(b, y, mkfxat(0));
	while (zerop(ca) || zerop(cb)) {
	    c = Lrandom(p);
	    ca = Asub(a, y, c);
	    ca = remainder_by_int(ca, p);
	    cb = Asub(b, y, c);
	    cb = remainder_by_int(cb, p);
	}
#endif
	array_value(B, j) = c;
	a = ca;
	b = cb;
    }
    r = B;
    popv(9);
#undef a
#undef b
#undef x
#undef p
#undef B
#undef ca
#undef cb
#undef c
#undef y
    return r;
}



static          Sex
EZGCD(Sex A, Sex B, Sex x, long k)
{
    Sex             r;
    long            i,
                    j;
    long            d,
                    d2;
    long            n;
    stackcheck();
    push3(A, B, x);
    pushv(27);

    // local a, b, g, bb, bbb;
    // local la, lb, p, p2;
    // local Ai, Bi, Ci, d, Ai2, Bi2, Ci2, d2;
    // local i, U, Hi, c, ci, C, H;
    // local lc, lu;
#define A stack[-29]
#define B stack[-28]
#define x stack[-27]
#define q stack[-26]
#define lu	stack[-25]
#define lc	stack[-24]
#define a stack[-23]
#define b stack[-22]
#define g stack[-21]
#define bb stack[-20]
#define bbb stack[-19]
#define la stack[-18]
#define lb stack[-17]
#define p stack[-16]
#define p2 stack[-15]
#define Ai stack[-14]
#define Bi stack[-13]
#define Ci stack[-12]
#define Ai2 stack[-10]
#define Bi2 stack[-9]
#define Ci2 stack[-8]
#define U stack[-6]
#define Hi stack[-5]
#define c stack[-4]
#define ci stack[-3]
#define C stack[-2]
#define H stack[-1]

    a = conts2(A, x, 1);
    b = conts2(B, x, 1);
    A = gquot2(A, a);
    B = gquot2(B, b);
#ifdef USE_HEURISTIC
    g = use_heuristic ? my_gcd_heuristic(a, b) : Mgcd(a, b);
#else
    g = Mgcd(a, b);
#endif
    a = gquot2(a, g);
    b = gquot2(b, g);

#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("ezgcd: x=%p\n", x);
#endif

    la = lcoeffs(A, x, 1);
    lb = lcoeffs(B, x, 1);
#ifdef USE_HEURISTIC
    lc = use_heuristic ? my_gcd_heuristic(la, lb) : Mgcd(la, lb);
#else
    lc = Mgcd(la, lb);
#endif

#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("EZGCD: la=%p, lb=%p\n", la, lb);
#endif

    p = NewPrime(la, lb);

    bb = NewEvaluationPoint(la, lb, x, k, p);

    Ai = Amkmodular(A, p);
    Bi = Amkmodular(B, p);
    for (i = 1; i < k; i++) {
	Ai = Asub(Ai, array_value(x, i), array_value(bb, i));
	Bi = Asub(Bi, array_value(x, i), array_value(bb, i));
    }

    Ci = ugcd_modp(Ai, Bi);

    d = mkint(Adeg(Ci, array_value(x, 0)));

    /*
     * double check 
     */

    n = 0;
    while (1) {
	n++;
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("ezgcd loop %d\n", n);
#endif
	if (d == 0) {
	    A = gtimes2(A, a);
	    B = gtimes2(B, b);
	    r = Llist3(A, B, g);
	    popv(30);
	    return r;
	}

	p2 = NewPrime(la, lb);

	bbb = NewEvaluationPoint(la, lb, x, k, p2);

	Ai2 = Amkmodular(A, p2);
	Bi2 = Amkmodular(B, p2);
	for (i = 1; i < k; i++) {
	    Ai2 = Asub(Ai2, array_value(x, i), array_value(bbb, i));
	    Bi2 = Asub(Bi2, array_value(x, i), array_value(bbb, i));
	}

	Ci2 = ugcd_modp(Ai2, Bi2);
	d2 = mkint(Adeg(Ci2, array_value(x, 0)));

	if (d2 < d) {
	    p = p2;
	    Ai = Ai2;
	    Bi = Bi2;
	    Ci = Ci2;
	    d = d2;
	    bb = bbb;
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("ezgcd loop d2 (%d) < d(%d) \n", d2, d);
#endif
	    continue;		/* do the double check */
	} else if (d2 > d) {
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("ezgcd loop d2 ( %d ) > d(%d) \n", d2, d);
#endif
	    continue;		/* do the double check */
	}

	i = mkint(Adeg(B, array_value(x, 0)));
	if (d == i) {
	    r = gremaind2(A, B);
	    if (gzerop(r)) {
		A = gquot2(A, B);
		a = gtimes2(a, A);
		g = gtimes2(B, g);
		r = Llist3(a, b, g);
		popv(30);
		return r;
	    } else {
		d--;
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf("ezgcd division failed decrece d to %d\n",
				d);
#endif
		continue;	/* do double check */
	    }
	}

	Ai = gquot2(Ai, Ci);
	Bi = gquot2(Bi, Ci);
	r = ugcd_modp(Ai, Ci);
	if (gonep(r)) {
	    U = A;
	    Hi = Ai;
	    // c = a;
	    lu = la;
	} else {
	    r = ugcd_modp(Bi, Ci);
	    if (gonep(r)) {
		U = B;
		Hi = Bi;
		// c = b;
		lu = lb;
	    } else {
		/*
		 * return SGCD(A, B, x, bb, p); 
		 */
		Hi = Ai;
		j = mkint(p);
		for (i = 1; i < j; i++) {
		    Hi = gplus2(Hi, Bi);
		    r = ugcd_modp(Hi, Ci);
		    if (gonep(r))
			break;
		}
		if (i == j)
		    aerror0("ezgcd failed");
		U = gtimes2(mkfxat(i), B);
		U = gplus2(A, U);
		c = conts2(U, x, 1);
		// c = gquot2(c, g);
		U = gquot2(U, c);
		lu = lcoeffs(U, x, 1);
	    }
	}

	// U = gtimes2(c, U);
	// ci = c;
	// for (i = 1; i < k; i++) {
	// ci = Asub(ci, array_value(x, i), array_value(bb, i));
	// }
	// Ci = gtimes2(ci, Ci);
	C = EZ_LIFT(U, Ci, Hi, x, bb, p, lc, lu);
	if (C == nil) {
	    p = p2;
	    Ai = Ai2;
	    Bi = Bi2;
	    Ci = Ci2;
	    d = d2;
	    bb = bbb;
	    continue;
	}
	H = cadr(C);
	q = conts(H, x, 1);	// f. k. 2016.2.5
	H = gquot2(H, q);	// f. k. 2016.2.5
	C = car(C);
	q = conts(C, x, 1);	// f. k. 2016.2.5
	C = gquot2(C, q);	// f. k. 2016.2.5
	q = gtimes2(C, H);
	r = gdiff2(U, q);
	if (!gzerop(r)) {
	    r = gplus2(U, q);
	    if (!gzerop(r)) {
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf("ezgcd ez_lift failed. r = %p\n", r);
#endif
		p = p2;
		Ai = Ai2;
		Bi = Bi2;
		Ci = Ci2;
		d = d2;
		bb = bbb;
		continue;
		continue;	/* do the double check */
	    }
	}

	/*
	 * Final check 
	 */
	// C = Mpp(C);
	// C = pp_ezgcd(C);
	r = gremaind2(B, C);
	if (!gzerop(r)) {
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("ezgcd division failed r = %p\n", r);
#endif
	    p = p2;
	    Ai = Ai2;
	    Bi = Bi2;
	    Ci = Ci2;
	    d = d2;
	    bb = bbb;
	    continue;		/* try again */
	}
	r = gremaind2(A, C);
	if (!gzerop(r)) {
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("ezgcd division failed r = %p\n", r);
#endif
	    p = p2;
	    Ai = Ai2;
	    Bi = Bi2;
	    Ci = Ci2;
	    d = d2;
	    bb = bbb;
	    continue;		/* try again too */
	}
	a = gtimes2(a, A);
	a = gquot2(a, C);
	b = gtimes2(b, B);
	b = gquot2(b, C);
	g = gtimes2(g, C);
	r = Llist3(a, b, g);
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("ezgcd:: r = %p\n", r);
#endif
	popv(30);
	return r;
    }
#undef A
#undef B
#undef x
#undef q
#undef lu
#undef lc
#undef a
#undef b
#undef g
#undef bb
#undef bbb
#undef la
#undef lb
#undef p
#undef p2
#undef Ai
#undef Bi
#undef Ci
#undef Ai2
#undef Bi2
#undef Ci2
#undef U
#undef Hi
#undef c
#undef ci
#undef C
#undef H
}




#if 0
static          Sex
SGCD(Sex a, Sex b, Sex x, Sex v, Sex p)
{
    long            i,
                    n;
    Sex             u;

    push5(a, b, x, v, p);
    pushv(7);

    // begin local g, s, r;
    // local i, n, f, c;
#define a stack[-11]
#define b stack[-10]
#define x stack[-9]
#define v stack[-8]
#define p stack[-7]
#define g stack[-6]
#define s stack[-5]
#define r stack[-4]
#define f stack[-3]
#define c stack[-2]

    r = mkfxat(1);
    s = SquareFree(a);
    while (s != nil) {
	c = car(s);
	s = cdr(s);
	f = car(c);
	n = mkint(cadr(c));
	for (i = 1; i < n; i++) {
	    g = ezgcd(f, b);
	    if (gonep(g))
		break;
	    b = gquot2(b, g);
	    r = gtimes2(r, g);
	}
	if (!is_poly(b))
	    break;
    }
    u = r;
    popv(12);
#undef a
#undef b
#undef x
#undef v
#undef p
#undef g
#undef s
#undef r
#undef f
#undef c
    return u;
}

static          Sex
SquareFree(Sex a)
{
    long            i;
    Sex             r;

    stackcheck();
    push(a);
    pushv(9);
    // begin local i, out, b, c, w, x, y, z, g;
#define a stack[-9]
#define out stack[-8]
#define b stack[-7]
#define c stack[-6]
#define w stack[-5]
#define x stack[-4]
#define y stack[-3]
#define z stack[-2]
#define g stack[-1]

    i = 1;
    out = nil;
    x = Mvar(a);
    b = Adiff(a, x);
    g = ezgcd(a, b);
    if (!gonep(g)) {
	while (1) {
	    a = gquot2(a, g);
	    y = gquot2(b, g);
	    r = Adiff(a, x);
	    b = gdiff2(y, r);
	    if (gzerop(b))
		break;
	    g = ezgcd(a, b);
	    r = Llist2(g, mkfxat(i));
	    out = cons(r, out);
	    i++;
	}
    }
    r = Llist2(a, mkfxat(i));
    r = cons(r, out);
    popv(10);
#undef a
#undef out
#undef b
#undef c
#undef w
#undef x
#undef y
#undef z
#undef g
    return r;
}
#endif

static          Sex
EZ_LIFT(Sex U, Sex C, Sex H, Sex x, Sex b, Sex p, Sex lc, Sex lu)
{
    long            n;
    long            i;
    Sex             r;

    stackcheck();
    push5(U, C, H, x, b);
    push3(p, lc, lu);
    pushv(5);

    // local a, bound;
    // local i, n, u;
#define U stack[-12]
#define C stack[-11]
#define H stack[-10]
#define x stack[-9]
#define b stack[-8]
#define p stack[-7]
#define lc stack[-6]
#define lu stack[-5]
#define c stack[-4]
#define a stack[-3]
#define bound stack[-2]
#define u stack[-1]
#define pk stack[-0]

    a = U;
    c = lc;
    n = mkint(car(Asize(x)));
    for (i = n - 1; i >= 1; i--) {
	a = Asub(a, array_value(x, i), array_value(b, i));
	c = Asub(c, array_value(x, i), array_value(b, i));
    }
    if (zerop(remaind2(c, p))) {
	popv(13);
	return nil;
    }
    bound = MignotteBound2(a, C, array_value(x, 0));
    a = UnivariateHensel(a, p, C, H, bound, c);
    u = car(a);
    c = cadr(a);
    if (n == 1) {
	if (gzerop(gpremaind2(U, u, array_value(x, 0)))) {
	    r = Llist2(u, c);
	    popv(13);
	    return r;
	}
	aerror0("ez_lift failed (1)");
    }
    pk = cddr(a);
    n = mkint(cadr(pk));
    pk = car(pk);
    r = MultivariateHensel(U, x, b, p, pk, n, u, c, lc, lu);
    // if (r == nil)
    // aerror0("ez_lift failed (2)");
    popv(13);
#undef U
#undef C
#undef H
#undef x
#undef b
#undef p
#undef c
#undef lc
#undef lu
#undef a
#undef bound
#undef u
#undef pk
    return r;
}

static          Sex
Mignotte(Sex f)
{
    Sex             r;
    if (!is_poly(f)) {
	f = gtimes2(f, f);
	return f;
    }
    stackcheck();
    push(f);
    pushv(2);
    // local s, a;
#define f stack[-2]
#define s stack[-1]
#define a stack[-0]
    f = poly_pol(f);
    s = mkfxat(0);
    while (f != nil) {
	a = lterm(f);
	a = coeff(a);
	if (is_poly(a))
	    a = Mignotte(a);
	else
	    a = times2(a, a);
	s = plus2(s, a);
	f = reduct(f);
    }
    r = s;
    popv(3);
#undef f
#undef s
#undef a
    return r;
}


static          Sex
MignotteBound2(Sex f, Sex g, Sex x)
{
    Sex             r;
    long            d;

    if (!is_poly(f))
	return Labs(f);
    stackcheck();
    push2(f, x);
    d = mkint(Adeg(g, x));
    push(nil);
    // local s;
#define f stack[-2]
#define x stack[-1]
#define s stack[-0]

    s = Mignotte(f);
    s = isqrt(s);
    r = expti(mkfxat(2), d);
    r = gtimes2(s, r);
    popv(3);
#undef f
#undef x
#undef s
    return r;
}


static          Sex
isqrt(Sex m)
{				/* integer square root of m */
    Sex             s;
    long            d;

    if (m == mkfxat(0))
	return m;
    stackcheck();
    push(m);
    pushv(3);

    // begin local x, x1, bk;
#define m stack[-3]
#define x stack[-2]
#define bk stack[-0]
    d = number_log2(m);
    d = (d + 2) / 2;
    x = expti(mkfxat(2), d);
    bk = quot2(m, x);
    bk = diff2(x, bk);
    while (plusp(bk)) {
	s = plus2(bk, mkfxat(1));
	s = quot2(s, mkfxat(2));
	x = diff2(x, s);
	s = quot2(m, x);
	bk = diff2(x, s);
    }
    s = x;
    popv(4);
#undef m
#undef x
#undef bk
    return s;
}

static          Sex
conts2(Sex f, Sex x, long k)
{
    long            i,
                    n;
    Sex             s;
    if (is_number(f)) {
	f = Labs(f);
	return f;
    } else if (k == 0)
	return f;

    stackcheck();
    push2(f, x);
    pushv(4);

    // begin local i, c, n, g, gg;
#define f stack[-5]
#define x stack[-4]
#define c stack[-3]
#define g stack[-2]
#define gg stack[-1]

    c = Acoeff(f, array_value(x, k - 1));
    n = mkint(car(Asize(c)));
    g = array_value(c, n - 1);
    g = conts2(g, x, k - 1);
    for (i = 0; i < n - 1; i++) {
	if (gonep(g)) {
	    g = mkfxat(1);
	    break;
	}
	gg = conts2(array_value(c, i), x, k - 1);
	if (gonep(gg)) {
	    g = mkfxat(1);
	    break;
	}
	if (!gzerop(gg))
#ifdef USE_HEURISTIC
	    // g = my_gcd_combine1(g, gg);
	    g = use_heuristic ? my_gcd_heuristic(g, gg) : Mgcd(g, gg);
#else
	    g = Mgcd(g, gg);
#endif
    }
    s = g;
    popv(6);
#undef f
#undef x
#undef c
#undef g
#undef gg
    return s;
}






/**************************************************************/
/**************************************************************/


/*
 * Univariate Hensel Lifting Algorithm 
 */

static          Sex
EEA(Sex a, Sex b, Sex x)
{
    long            n;
    Sex             u;

    stackcheck();
    push3(a, b, x);
    pushv(13);
    // local c, d, c1, c2, d1, d2;
    // local r, r1, r2;
    // local n;
    // local q, g;
#define a stack[-15]
#define b stack[-14]
#define x stack[-13]
#define c stack[-12]
#define d stack[-11]
#define c1 stack[-10]
#define c2 stack[-9]
#define d1 stack[-8]
#define d2 stack[-7]
#define r stack[-6]
#define r1 stack[-5]
#define r2 stack[-4]
#define q stack[-3]
#define g stack[-2]
    c = Ainit(a, x);
    d = Ainit(b, x);
    c = gquot2(a, c);
    d = gquot2(b, d);

    c1 = d2 = mkfxat(1);
    d1 = c2 = mkfxat(0);

    n = 0;
    while (!gzerop(d)) {
	n++;
	q = grquotient2(c, d, x);
	u = gtimes2(q, d);
	r = gdiff2(c, u);
	u = gtimes2(q, d1);
	r1 = gdiff2(c1, u);
	u = gtimes2(q, d2);
	r2 = gdiff2(c2, u);
	c = d;
	c1 = d1;
	c2 = d2;
	d = r;
	d1 = r1;
	d2 = r2;
    }
    /*
     * normalize GCD 
     */
    g = Ainit(c, x);
    g = gquot2(c, g);
    a = Ainit(a, x);
    b = Ainit(b, x);
    c = Ainit(c, x);
    a = gtimes2(a, c);
    b = gtimes2(b, c);
    c1 = gquot2(c1, a);
    c2 = gquot2(c2, b);
    u = Llist3(g, c1, c2);
    popv(16);
#undef a
#undef b
#undef x
#undef c
#undef d
#undef c1
#undef c2
#undef d1
#undef d2
#undef r
#undef r1
#undef r2
#undef q
#undef g
    return u;
}


static          Sex
replace_lc(Sex f, Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push3(f, a, x);
    push2(nil, nil);
#define f stack[-4]
#define a stack[-3]
#define x stack[-2]
#define b stack[-1]
#define c stack[-0]
    b = Alterm(f, x);
    c = Alpower(f, x);
    c = gtimes2(a, c);
    f = gdiff2(f, b);
    r = gplus2(f, c);
    popv(5);
#undef f
#undef a
#undef x
#undef b
#undef c
    return r;
}


static          Sex
multivariate_replace_lc(Sex f, Sex a)
{
    Sex             r;
    stackcheck();
    push3(f, a, nil);
    // begin local c;
#define f stack[-2]
#define a stack[-1]
#define c stack[-0]

    c = Mlcoeff(f);
    if (is_poly(c)) {
	a = multivariate_replace_lc(c, a);
    }
    r = Mlterm(f);
    c = gdiff2(f, r);
    r = Mlpower(f);
    r = gtimes2(a, r);
    r = gplus2(c, r);
    popv(3);
#undef f
#undef a
#undef c
    return r;
}


static          Sex
UnivariateHensel(Sex a, Sex p, Sex u1, Sex w1, Sex B, Sex gamma)
{
    Sex             z;
    long            n;

    stackcheck();
    push5(a, p, u1, w1, B);
    push(gamma);
    pushv(18);
#define a stack[-23]
#define p stack[-22]
#define u1 stack[-21]
#define w1 stack[-20]
#define B stack[-19]
#define gamma stack[-18]
    // local aa, x;
    // local alpha, delta, sigma, tau, modulus;
#define aa stack[-17]
#define x stack[-16]
#define alpha stack[-15]
#define delta stack[-14]
#define sigma stack[-13]
#define tau stack[-12]
#define modulus stack[-11]
    // local e, g, q, r, s, t;
    // local u, w, sh, th;
    // local c;
#define e stack[-10]
#define g stack[-9]
#define q stack[-8]
#define r stack[-7]
#define s stack[-6]
#define t stack[-5]
#define u stack[-4]
#define w stack[-3]
#define sh stack[-2]
#define th stack[-1]
#define c stack[-0]

#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("UH: a = %p\n", a);
#endif
    x = Mvar(a);

    /*
     * step 1 
     */

    alpha = Alcoeff(a, x);
    if (gamma == nil)
	gamma = alpha;
    a = gtimes2(a, gamma);
    s = Alcoeff(u1, x);
    t = Alcoeff(w1, x);
    u1 = Amkmodular(u1, p);
    w1 = Amkmodular(w1, p);
    u1 = gtimes2(gamma, u1);
    u1 = gratio2(u1, s);
    w1 = gtimes2(alpha, w1);
    w1 = gratio2(w1, t);

    /*
     * step 2
     */

    g = EEA(u1, w1, x);
    s = cadr(g);
    t = cadr(cdr(g));
    g = car(g);
    /*
     * step 3 
     */
    z = Aunmodular(u1);
    u = replace_lc(z, gamma, x);
    z = Aunmodular(w1);
    w = replace_lc(z, alpha, x);
    z = gtimes2(u, w);
    e = gdiff2(a, z);
    modulus = p;
    B = gtimes2(B, mkfxat(2));
    B = gtimes2(B, gamma);
    B = Labs(B);
    /*
     * step 4 
     */
    n = 1;
    while (!zerop(e) && Llessp(modulus, B) != nil) {
	/*
	 * write "modulus = ", modulus, ", error = ", e;
	 */
	/*
	 * step 4.1 
	 */
	c = gquot2(e, modulus);
	sh = gtimes2(s, c);
	th = gtimes2(t, c);
	g = grdivide2(sh, w1, x);
	q = car(g);
	r = cdr(g);
	sigma = r;
	z = gtimes2(q, u1);
	tau = gplus2(th, z);
	/*
	 * step 4.2 
	 */
	z = Aunmodular(tau);
	z = gtimes2(z, modulus);
	u = gplus2(u, z);
	z = Aunmodular(sigma);
	z = gtimes2(z, modulus);
	w = gplus2(w, z);
	z = gtimes2(u, w);
	e = gdiff2(a, z);
	modulus = gtimes2(modulus, p);
	n++;
    }
    /*
     * step 5 
     */
    if (zerop(e)) {
	delta = Mcont(u);
	u = gquot2(u, delta);
	z = gquot2(gamma, delta);
	w = gquot2(w, z);
	z = Llist4(u, w, modulus, mkfxat(n));
    } else
	z = nil;
    popv(24);
#undef a
#undef p
#undef u1
#undef w1
#undef B
#undef gamma
    // local aa, x;
    // local alpha, delta, sigma, tau, modulus;
#undef aa
#undef x
#undef alpha
#undef delta
#undef sigma
#undef tau
#undef modulus
    // local e, g, q, r, s, t;
    // local u, w, sh, th;
    // local c;
#undef e
#undef g
#undef q
#undef r
#undef s
#undef t
#undef u
#undef w
#undef sh
#undef th
#undef c
    return z;
}


/*
 * Multivariate Hensel Lifting Algorithm 
 */

static          Sex
TaylerCoeff(Sex f, Sex x, long k, Sex a)
{
    long            i;
    stackcheck();
    push3(f, x, a);
#define f stack[-2]
#define x stack[-1]
#define a stack[0]

    for (i = 1; i <= k; i++)
	f = Adiff(f, x);
    f = Asub(f, x, a);
    a = Lfactorial(mkfxat(k));
    f = gratio2(f, a);
#undef a
    a = f;
    popv(3);
#undef f
#undef x
    return a;
}

static void
print_e(Sex e, Sex v, Sex pl)
{
    long            n;
    stackcheck();
    push3(e, v, pl);
    push2(nil, nil);
#define e stack[-4]
#define v stack[-3]
#define pl stack[-2]
#define q stack[-1]
#define r stack[-0]
    n = 0;
    while (is_poly(e)) {
	q = grdivide2(e, v, Mvar(v));
	r = cdr(q);
	r = gsmodulus(r, pl);
	e = car(q);
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("(%p)^%d*(%p)+", v, n, r);
#endif
	n++;
    }
    if (!gzerop(e)) {
	if (ezgcd_debug)
	    lisp_printf("%p\n", e);
    } else {
	if (ezgcd_debug)
	    lisp_printf("\n");
    }
    popv(5);
#undef e
#undef v
#undef pl
#undef q
#undef r
}

static          Sex
MultivariateHensel(Sex a, Sex x, Sex v, Sex p, Sex pl, long l, Sex u,
		   Sex w, Sex lcu, Sex lcw)
{
    long            i,
                    j,
                    k,
                    m,
                    n;
    long            d;
    long            maxdeg;
    Sex             r;

    stackcheck();
    push5(a, x, v, p, pl);
    push4(u, w, lcu, lcw);
    pushv(18);
    // begin local A, u1, w1, u2, w2, dU;
    // local i, j, k, m;
    // local n, nu, maxdeg;
    // local e;
    // local s, du, dw, c, monomial, t, d;
    // local llu, llw, u3, w3, cc;
#define a stack[-26]
#define x stack[-25]
#define v stack[-24]
#define p stack[-23]
#define pl stack[-22]
#define u stack[-21]
#define w stack[-20]
#define lcu stack[-19]
#define lcw stack[-18]
#define A stack[-17]
#define u1 stack[-16]
#define w1 stack[-15]
#define u2 stack[-14]
#define w2 stack[-13]
#define dU stack[-12]
#define e stack[-11]
#define s stack[-10]
#define du stack[-9]
#define dw stack[-8]
#define c stack[-7]
#define monomial stack[-6]
#define t stack[-5]
    // #define d stack[-7]
#define llu stack[-4]
#define llw stack[-3]
#define q stack[-2]
#define u3 stack[-2]		/* use same place as q */
#define w3 stack[-1]
#define cc stack[-0]

    n = mkint(car(Asize(x)));
    A = mkarray(n);

#ifdef DEBUG
    if (ezgcd_debug) {
	lisp_printf("MH: lcu = %p, lcw = %p\n", lcu, lcw);
	lisp_printf("MH: p=%p, pl=%p, x= %p, v=%p\n", p, pl, x, v);
	// lisp_printf("u=%p + ...\n", Mlterm(u));
	// lisp_printf("w=%p + ...\n", Mlterm(w));
	lisp_printf("u=%p \n", (u));
	lisp_printf("w=%p \n", (w));
    }
#endif

    llu = lcu;
    llw = lcw;
    for (j = n - 1; j >= 1; j--) {
	llu = Asub(llu, array_value(x, j), array_value(v, j));
	llw = Asub(llw, array_value(x, j), array_value(v, j));
    }
#ifdef DEBUG
    if (ezgcd_debug) {
	lisp_printf("lcu = %p\n", llu);
	lisp_printf("lcw = %p\n", llw);
    }
#endif

    // c = Minit(u);
    c = Ainit(u, array_value(x, 0));	// f.k. 2016.2.5
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("lcoeff(u) = %p\n", c);
#endif
    q = Ldivide(llu, c);
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("llu/lcoeff(u) = %p\n", q);
#endif
    if (zerop(cdr(q))) {
	q = car(q);
#ifdef STRATEGY_AA
	u = gtimes2(u, q);
	a = gtimes2(a, lcu);
	if (minusp(llu))
	    u = gnegate(u);
	cc = mkfxat(1);
	q = Labs(q);
	while (lessp2(cc, q)) {
	    cc = times2(cc, p);
	    l++;
	}
	pl = gtimes2(pl, cc);
#else
	if (onep(q)) {
	    lcw = gquot2(lcw, lcu);
	    llw = gquot2(llw, llu);
	} else if (monep(q)) {
	    lcw = gquot2(lcw, lcu);
	    llw = gquot2(llw, llu);
	    u = gnegate(u);
	} else {
	    u = gtimes2(u, q);
	    a = gtimes2(a, lcu);
	    // if(minusp(llu)) {u = gnegate(u); lcu = gnegate(lcu);} //
	    // ??? f.k. 2016.2.6
	    cc = mkfxat(1);
	    q = Labs(q);
	    while (lessp2(cc, q)) {
		cc = times2(cc, p);
		l++;
	    }
	    pl = gtimes2(pl, cc);

	}
#endif
#ifdef DEBUG
	if (ezgcd_debug) {
	    lisp_printf("u = %p\n", u);
	    lisp_printf("w = %p\n", w);
	    lisp_printf("a = %p\n", a);
	    lisp_printf("llu = %p, llw = %p, lcu = %p, lcw = %p\n", llu,
			llw, lcu, lcw);
	}
#endif
    } else {
	aerror0("can't divide llu by lcoeff(u)");
    }
    // c = Minit(w);
    c = Ainit(w, array_value(x, 0));	// f.k. 2016.2.5
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("lcoeff(w) = %p\n", c);
#endif
    q = Ldivide(llw, c);
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("llw/lcoeff(w) = %p\n", q);
#endif
    if (zerop(cdr(q))) {
	q = car(q);
	w = gtimes2(w, q);
#ifdef DEBUG
	if (ezgcd_debug) {
	    lisp_printf("u = %p\n", u);
	    lisp_printf("w = %p\n", w);
	    lisp_printf("a = %p\n", a);
	    lisp_printf("lcu = %p, lcw = %p\n", lcu, lcw);
	}
#endif
    } else {
	aerror0("can't divide llw by lcoeff(w)");
    }


    array_value(A, n - 1) = a;

#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("A[%d] = %p\n", n - 1, a);
#endif
    for (j = n - 1; j >= 1; j--) {
	r = Asub(array_value(A, j), array_value(x, j), array_value(v, j));
	r = gmodulus(r, pl);
	array_value(A, j - 1) = r;
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("A[%d] = %p\n", j - 1, r);
#endif
    }

    maxdeg = 0;

    for (i = 1; i < n; i++) {
	d = mkint(Adeg(a, array_value(x, i)));
	if (maxdeg < d)
	    maxdeg = d;
    }
#ifdef DEBUG
    r = gtimes2(u, w);
    r = gdiff2(array_value(A, 0), r);
    r = gmodulus(r, pl);
    if (ezgcd_debug)
	lisp_printf("A[0] - u*w = %p\n", r);
#endif
    u1 = u;
    w1 = w;

    /*
     * step 2
     */
    for (j = 1; j < n; j++) {
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("\nMh: j=%d\n\n", j);
#endif
	u2 = u1;
	w2 = w1;
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("u2=%p\n", u2);
	if (ezgcd_debug)
	    lisp_printf("w2=%p\n", w2);
#endif
	monomial = mkfxat(1);
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("replace u1 of %p ", Mlterm(u1));
#endif
	if (is_poly(lcu)) {
	    c = lcu;
	    for (i = j + 1; i < n; i++) {
		r = Asub(c, array_value(x, i), array_value(v, i));
		c = gmodulus(r, pl);
	    }
	    u1 = replace_lc(u1, c, array_value(x, 0));
	} else if (!onep(lcu)) {
	    u1 = replace_lc(u1, lcu, array_value(x, 0));
	}
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("by %p \n", Mlterm(u1));
	if (ezgcd_debug)
	    lisp_printf("replace w1 of %p ", Mlterm(w1));
#endif
	if (is_poly(lcw)) {
	    c = lcw;
	    for (i = j + 1; i < n; i++) {
		r = Asub(c, array_value(x, i), array_value(v, i));
		c = gmodulus(r, pl);
	    }
	    w1 = replace_lc(w1, c, array_value(x, 0));
	} else if (!onep(lcw)) {
	    w1 = replace_lc(w1, lcw, array_value(x, 0));
	}
#ifdef DEBUG
	if (ezgcd_debug)
	    lisp_printf("by %p \n", Mlterm(w1));
#endif
	e = gtimes2(u1, w1);
	e = gdiff2(array_value(A, j), e);
	e = gsmodulus(e, pl);
	d = mkint(Adeg(array_value(A, j), array_value(x, j))) + 1;
	for (k = 1; k <= d; k++) {
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("\nMh: k=%d\n\n", k);
#endif
	    if (gzerop(e))
		break;
	    r = symbol_to_polynomial(array_value(x, j));
	    r = gdiff2(r, array_value(v, j));
	    monomial = gtimes2(monomial, r);
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("multi-hensel: k=%d, pl=%p\n", k, pl);
	    if (ezgcd_debug)
		lisp_printf("multi-hensel: monomial = %p\n", monomial);
	    print_e(e, r, pl);
#endif
	    c = TaylerCoeff(e, array_value(x, j), k, array_value(v, j));
	    c = gsmodulus(c, pl);
	    if (!gzerop(c)) {
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf
			("MH call MD with u2=%p,\n   w2 = %p,\n   c = %p\n",
			 u2, w2, c);
#endif
		dU = MultivariateDiophant(u2, w2, c, x, v, j, maxdeg, p,
					  pl, l);
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf("MD returns %p\n", dU);
#endif
		// e = 1;
		s = gtimes2(car(dU), monomial);
		u1 = gplus2(u1, s);
		u1 = gsmodulus(u1, pl);
		// e = gtimes2(e, u1);
		// e = gsmodulus(u1, pl);
		s = gtimes2(cadr(dU), monomial);
		w1 = gplus2(w1, s);
		w1 = gsmodulus(w1, pl);
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf("Mhensel dU = %p\n", dU);
		// if(ezgcd_debug)lisp_printf("u1=%p + ...\n",
		// Mlterm(u1));
		// if(ezgcd_debug)lisp_printf("w1=%p + ...\n",
		// Mlterm(w1));
		if (ezgcd_debug)
		    lisp_printf("u1=%p \n", (u1));
		if (ezgcd_debug)
		    lisp_printf("w1=%p \n", (w1));
#endif
		// e = gtimes2(e, w1);
		e = gtimes2(u1, w1);
		e = gsmodulus(e, pl);
		e = gdiff2(array_value(A, j), e);
		e = gsmodulus(e, pl);
	    }
	}
	if (!gzerop(e)) {
	    // aerror1("Multivariate Hensel lifting failed", e);
	    // cc = Mcont(u1);
	    cc = conts(u1, x, 1);	// f.k. 2016.2.5
	    // if(ezgcd_debug)lisp_printf("cont(u1) = %p\n", cc);
	    if (!gonep(cc)) {
		w3 = gtimes2(w1, cc);
		if (!gonep(monomial)) {
		    w3 = grremaind2(w3, monomial, array_value(x, j));
		}
		w3 = gsmodulus(w3, pl);
		u3 = gquot2(u1, cc);
		u3 = gsmodulus(u3, pl);
		e = gtimes2(u3, w3);
		e = gsmodulus(e, pl);
		e = gdiff2(array_value(A, j), e);
		e = gsmodulus(e, pl);
		if (gzerop(e)) {
		    u1 = u3;
		    w1 = w3;
		    continue;
		}
	    }
	    // cc = Mcont(w1);
	    cc = conts(w1, x, 1);	// f.k. 2016.2.5
	    // if(ezgcd_debug)lisp_printf("cont(w1) = %p\n", cc);
	    if (!gonep(cc)) {
		u3 = gtimes2(u1, cc);
		if (!gonep(monomial))
		    u3 = grremaind2(u3, monomial, array_value(x, j));
		u3 = gsmodulus(u3, pl);
		w3 = gquot2(w1, cc);	// f.k. 2016.2.9
		w3 = gsmodulus(w3, pl);
		e = gtimes2(u3, w3);
		e = gsmodulus(e, pl);
		e = gdiff2(array_value(A, j), e);
		e = gsmodulus(e, pl);
		if (gzerop(e)) {
		    u1 = u3;
		    w1 = w3;
		    continue;
		}
	    }


	}
    }

    /*
     * step 3 
     */

    u1 = gsmodulus(u1, pl);
    w1 = gsmodulus(w1, pl);
#ifdef DEBUG
    // if(ezgcd_debug)lisp_printf("u1=%p + ...\n", (u1));
    // if(ezgcd_debug)lisp_printf("w1=%p + ...\n", (w1));
    if (ezgcd_debug)
	lisp_printf("u1=%p \n", (u1));
    if (ezgcd_debug)
	lisp_printf("w1=%p \n", (w1));
#endif
    e = gtimes2(u1, w1);
    e = gdiff2(a, e);
    if (gzerop(e)) {
	// lcu = Mcont(u1);
	// lcu = cont_ezgcd(u1);
	// u1 = gquot2(u1, lcu);
	u1 = pp_ezgcd(u1);
	// lcw = Mcont(w1);
	// lcw = cont_ezgcd(w1);
	// w1 = gquot2(w1, lcw);
	w1 = pp_ezgcd(w1);
	r = Llist2(u1, w1);
    } else
	r = nil;
    popv(27);
#undef a
#undef x
#undef v
#undef p
#undef pl
#undef u
#undef w
#undef lcu
#undef lcw
#undef A
#undef u1
#undef w1
#undef u2
#undef w2
#undef dU
#undef e
#undef s
#undef du
#undef dw
#undef c
#undef monomial
#undef t
#undef llu
#undef llw
#undef q
#undef u3
#undef w3
#undef cc
    // #undef d
    return r;
}


/*
 * multivariate polynomial diophantine euqations 
 */

static          Sex
MultivariateDiophant(Sex u, Sex w, Sex c, Sex xv, Sex av, long n, long d,
		     Sex p, Sex pk, long k)
{
    long            m,
                    j;
    Sex             res;

    stackcheck();
    push5(u, w, c, xv, av);
    push2(p, pk);
    pushv(21);


    // local m, j, r;
    // local x, v, A, b;
    // local anew, cnew, sigma, monomial, cm;
    // local e, es, ds;
    // local i;
    // local unew, wnew, bu, bw;
    // local s1, s2;
#define u stack[-27]
#define w stack[-26]
#define c stack[-25]
#define xv stack[-24]
#define av stack[-23]
#define p stack[-22]
#define pk stack[-21]
#define x stack[-20]
#define v stack[-19]
#define A stack[-18]
#define b stack[-17]
#define r stack[-16]
#define anew stack[-15]
#define cnew stack[-14]
#define sigma stack[-13]
#define monomial stack[-12]
#define cm stack[-11]
#define e stack[-10]
#define es stack[-9]
#define unew stack[-8]
#define wnew stack[-7]
#define bu stack[-6]
#define bw stack[-5]
#define s1 stack[-4]
#define s2 stack[-3]
#define ds stack[-2]

    /*
     * step 1 
     */

#ifdef DEBUG_MD
    if (ezgcd_debug) {
	lisp_printf("\nMD: n = %d\n\n", n);
	lisp_printf("MD: u=%p\n", u);
	lisp_printf("MD: w=%p\n", w);
	lisp_printf("MD: c=%p\n", c);
    }
#endif

    x = array_value(xv, n - 1);
    v = array_value(av, n - 1);


    if (n > 1) {
	/*
	 * step 2.1 
	 */

	unew = Asub(u, x, v);
	wnew = Asub(w, x, v);

	cnew = Asub(c, x, v);
	sigma =
	    MultivariateDiophant(unew, wnew, cnew, xv, av, n - 1, d, p, pk,
				 k);
	s1 = car(sigma);
	s2 = cadr(sigma);
	e = gtimes2(s1, w);
	r = gtimes2(s2, u);
	e = gplus2(e, r);
	r = gdiff2(c, e);
	e = gmodulus(r, pk);
	monomial = mkfxat(1);
	for (m = 1; m <= d; m++) {
#ifdef DEBUG_MD
	    if (ezgcd_debug) {
		lisp_printf("\nMD: m = %d\n\n", m);
	    }
#endif
	    if (gzerop(e))
		break;
	    r = symbol_to_polynomial(x);
	    r = gdiff2(r, v);
	    monomial = gtimes2(monomial, r);
	    cm = e;
	    for (j = 1; j <= m; j++)
		cm = Adiff(cm, x);
	    cm = Asub(cm, x, v);
	    r = Lfactorial(mkfxat(m));
	    cm = gratio2(cm, r);
	    if (!gzerop(cm)) {
		ds = MultivariateDiophant(unew, wnew, cm, xv, av, n - 1, d,
					  p, pk, k);
		r = gtimes2(car(ds), monomial);
		s1 = gplus2(s1, r);
		r = gtimes2(car(ds), monomial);
		es = gtimes2(r, w);
		ds = gtimes2(cadr(ds), monomial);

		s2 = gplus2(s2, ds);
		r = gtimes2(ds, u);
		es = gplus2(es, r);
		r = gdiff2(e, es);
		e = gmodulus(r, pk);
	    }
	}

    } else {
	/*
	 * step 2.2 Univariate case 
	 */
	ds = UnivariateDiophant(u, w, x, c, p, pk, k);
	s1 = car(ds);
	s2 = cadr(ds);
    }

    s1 = gmodulus(s1, pk);
    s2 = gmodulus(s2, pk);
    res = Llist2(s1, s2);
    popv(28);
#undef u
#undef w
#undef c
#undef xv
#undef av
#undef p
#undef pk
#undef x
#undef v
#undef A
#undef b
#undef r
#undef anew
#undef cnew
#undef sigma
#undef monomial
#undef cm
#undef e
#undef es
#undef unew
#undef wnew
#undef bu
#undef bw
#undef s1
#undef s2
    return res;
}



static          Sex
UnivariateDiophant(Sex u, Sex w, Sex x, Sex c, Sex p, Sex pk, long k)
{
    Sex             r;

    stackcheck();
    push5(u, w, x, c, p);
    push(pk);
    pushv(4);
#define u stack[-9]
#define w stack[-8]
#define x stack[-7]
#define c stack[-6]
#define p stack[-5]
#define pk stack[-4]
#define q stack[-3]
#define s stack[-2]
#define s1 stack[-1]
#define s2 stack[-0]

    // local q, s, j, s1, s2;
    s = EEAlift(w, u, x, p, pk, k);
    s1 = car(s);
    s2 = cadr(s);
    s1 = gtimes2(c, s1);
    s1 = gmodulus(s1, pk);

    q = grquotient2(s1, u, x);
    q = gmodulus(q, pk);

    r = gtimes2(q, u);
    r = gdiff2(s1, r);
    r = gmodulus(r, pk);
    s1 = r;
    s2 = gtimes2(s2, c);
    r = gtimes2(q, w);
    r = gplus2(s2, r);
    r = gmodulus(r, pk);
    s2 = r;
    r = Llist2(s1, s2);
    popv(10);
#undef u
#undef w
#undef x
#undef c
#undef p
#undef pk
#undef q
#undef s
#undef s1
#undef s2
    return r;
}


static          Sex
EEAlift(Sex a, Sex b, Sex x, Sex p, Sex pk, long k)
{
    long            j;
    Sex             r;

    stackcheck();
    push5(a, b, x, p, pk);
    pushv(13);

#define a stack[-17]
#define b stack[-16]
#define x stack[-15]
#define p stack[-14]
#define pk stack[-13]
#define ap stack[-12]
#define bp stack[-11]
#define t stack[-10]
#define s stack[-9]
#define tp stack[-8]
#define sp stack[-7]
#define modulus stack[-6]
#define e stack[-5]
#define c stack[-4]
#define q stack[-3]
#define sigma stack[-2]
#define tau stack[-1]
#define sa stack[0]
    // begin local ap, bp, t, s, tp, sp;
    // local modulus, e, c, q, sigma, tau;
    // local j;

    ap = Amkmodular(a, p);
    bp = Amkmodular(b, p);
    s = EEA(ap, bp, x);
    tp = car(cddr(s));
    sp = cadr(s);
    s = Aunmodular(sp);
    t = Aunmodular(tp);
    modulus = p;
    for (j = 1; j < k; j++) {
	sa = gtimes2(s, a);
	r = gtimes2(t, b);
	sa = gplus2(sa, r);
	e = gdiff2(mkfxat(1), sa);
	c = gratio2(e, modulus);
	c = Amkmodular(c, p);
	sigma = gtimes2(sp, c);
	tau = gtimes2(tp, c);
	q = grdivide2(sigma, bp, x);
	sigma = cdr(q);
	q = car(q);
	r = gtimes2(q, ap);
	tau = gplus2(tau, r);
	r = Aunmodular(sigma);
	r = gtimes2(r, modulus);
	s = gplus2(s, r);
	r = Aunmodular(tau);
	r = gtimes2(r, modulus);
	t = gplus2(t, r);
	modulus = gtimes2(modulus, p);
    }
    r = Llist2(s, t);
    popv(18);
#undef a
#undef b
#undef x
#undef p
#undef pk
#undef ap
#undef bp
#undef t
#undef s
#undef tp
#undef sp
#undef modulus
#undef e
#undef c
#undef q
#undef sigma
#undef tau
#undef sa
    return r;
}


/**************************************************************/
/**************************************************************/

static          Sex
cont(Sex a)
{
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
	p = ggcd2(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}

static          Sex
pp(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont(a);
    pop(a);
    p = gquot2(a, p);
    return p;
}

/**************************************************************/
/**************************************************************/


static          Sex
cont_pprs(Sex a)
{				/* 
				 * GCD
				 */
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
	p = my_gcd_pprs(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}

static          Sex
pp_pprs(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont_pprs(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}



#ifdef GCD_TEST
static          Sex
my_gcd_pprs(Sex a, Sex b)
{				/* 
				 * GCD primitive prs
				 */
    Sex             g;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (a == mkfxat(0))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (b == mkfxat(0))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    push(a);
	    b = cont(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return ggcd2(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont(a);
	    pop(b);
	    return ggcd2(a, b);
	} else if (o == 0) {
	    int             n = 1;
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push(nil);
#define a stack[-5]
#define b stack[-4]
#define p stack[-3]
#define q stack[-2]
#define g1 stack[-1]
#define g2 stack[0]

	    g1 = cont(a);
	    g2 = cont(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = ggcd2(g1, g2);

	    g1 = wuremainder(p, q);
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, p);
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, q);
#endif
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		p = q;
		q = g1;
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf(" p%d: %p\n", n++, g1);
#endif
		g1 = wuremainder(p, q);
		g1 = pp(g1);
	    }
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, g1);
#endif
	    if (g1 == mkfxat(0)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = pp(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
	    popv(6);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#else
static          Sex
my_gcd_pprs(Sex a, Sex b)
{				/* 
				 * GCD primitive prs
				 */
    Sex             g;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (a == mkfxat(0))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont_pprs(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (b == mkfxat(0))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont_pprs(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    push(a);
	    b = cont_pprs(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return my_gcd_pprs(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont_pprs(a);
	    pop(b);
	    return my_gcd_pprs(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push(nil);
#define a stack[-5]
#define b stack[-4]
#define p stack[-3]
#define q stack[-2]
#define g1 stack[-1]
#define g2 stack[0]

	    g1 = cont_pprs(a);
	    g2 = cont_pprs(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = my_gcd_pprs(g1, g2);

	    g1 = wuremainder(p, q);
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		p = q;
		q = g1;
		g1 = wuremainder(p, q);
		g1 = pp_pprs(g1);
	    }
	    if (g1 == mkfxat(0)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = pp_pprs(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
	    popv(6);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#endif

static          Sex
cont_rprs(Sex a)
{				/* 
				 * GCD
				 */
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil && p != mkfxat(1)) {
	q = lcoeff(r);
	p = my_gcd_rprs(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}


static          Sex
pp_rprs(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont_rprs(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}

#if 0
static          Sex
my_gcd_rprs(Sex a, Sex b)
{
    stackcheck();
    push3(a, b, nil);
#define  a stack[-2]
#define  b stack[-1]
#define  c stack[-0]
    c = my_gcd_rprs_aux(a, b);
#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("gcd_rprs(%p, %p) ==> %p\n", a, b, c);
#endif
#undef a
#undef b
#undef c
    a = stack[0];
    popv(3);
    return a;
}
#endif

#ifdef GCD_TEST
static          Sex
my_gcd_rprs(Sex a, Sex b)
{				/* 
				 * GCD reduced prs
				 */
    Sex             g;
    long            o,
                    d;
    long            d1,
                    d2;

#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("gcd_rprs: %p,", a);
    if (ezgcd_debug)
	lisp_printf("%p\n", b);
#endif

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (a == mkfxat(0))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (b == mkfxat(0))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));
	// if(ezgcd_debug)lisp_printf("compare %p and %p ==> %d\n",
	// Mvar(a), Mvar(b), o);
	if (o > 0) {
	    push(a);
	    b = cont(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return ggcd2(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont(a);
	    pop(b);
	    return ggcd2(a, b);
	} else if (o == 0) {
	    int             n = 1;
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push2(nil, nil);
#define a stack[-6]
#define b stack[-5]
#define p stack[-4]
#define q stack[-3]
#define g1 stack[-2]
#define g2 stack[-1]
#define e1 stack[0]

	    if (lpower(poly_pol(a)) < lpower(poly_pol(b))) {
		p = a;
		a = b;
		b = p;
	    }
	    g1 = cont(a);
	    g2 = cont(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = ggcd2(g1, g2);
	    // if(ezgcd_debug)lisp_printf("g2=%p\n", g2);

	    g1 = wuremainder(p, q);
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, p);
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, q);
#endif
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		d1 = mkint(lpower(poly_pol(p))) -
		    mkint(lpower(poly_pol(q))) + 1;
		e1 = gexpt(lcoeff(poly_pol(q)), mkfxat(d1));
		p = q;
		q = g1;
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf(" p%d: %p\n", n++, g1);
#endif
		// if(ezgcd_debug)lisp_printf("r=%p\n", q);
		g1 = wuremainder(p, q);
		g1 = gquot2(g1, e1);
	    }
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, g1);
#endif
	    if (g1 == mkfxat(0)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = pp(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
#undef e1
	    popv(7);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#else
static          Sex
my_gcd_rprs(Sex a, Sex b)
{				/* 
				 * GCD reduced prs
				 */
    Sex             g;
    long            o,
                    d;
    long            d1,
                    d2;

#ifdef DEBUG
    if (ezgcd_debug)
	lisp_printf("gcd_rprs: %p,", a);
    if (ezgcd_debug)
	lisp_printf("%p\n", b);
#endif

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (a == mkfxat(0))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont_rprs(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (b == mkfxat(0))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont_rprs(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));
	// if(ezgcd_debug)lisp_printf("compare %p and %p ==> %d\n",
	// Mvar(a), Mvar(b), o);
	if (o > 0) {
	    push(a);
	    b = cont_rprs(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return my_gcd_rprs(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont_rprs(a);
	    pop(b);
	    return my_gcd_rprs(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push2(nil, nil);
#define a stack[-6]
#define b stack[-5]
#define p stack[-4]
#define q stack[-3]
#define g1 stack[-2]
#define g2 stack[-1]
#define e1 stack[0]

	    if (lpower(poly_pol(a)) < lpower(poly_pol(b))) {
		p = a;
		a = b;
		b = p;
	    }
	    g1 = cont_rprs(a);
	    g2 = cont_rprs(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = my_gcd_rprs(g1, g2);
	    // if(ezgcd_debug)lisp_printf("g2=%p\n", g2);

	    g1 = wuremainder(p, q);
	    // if(ezgcd_debug)lisp_printf("p=%p\n", p);
	    // if(ezgcd_debug)lisp_printf("q=%p\n", q);
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		d1 = mkint(lpower(poly_pol(p))) -
		    mkint(lpower(poly_pol(q))) + 1;
		e1 = gexpt(lcoeff(poly_pol(q)), mkfxat(d1));
		p = q;
		q = g1;
		// if(ezgcd_debug)lisp_printf("r=%p\n", q);
		g1 = wuremainder(p, q);
		g1 = gquot2(g1, e1);
	    }
	    if (g1 == mkfxat(0)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = pp_rprs(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
#undef e1
	    popv(7);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#endif



static          Sex
cont_sprs(Sex a)
{				/* 
				 * GCD
				 */
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
	p = my_gcd_sprs(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}

static          Sex
pp_sprs(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont_sprs(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}


#ifdef GCD_TEST
static          Sex
my_gcd_sprs(Sex a, Sex b)
{				/* 
				 * GCD subresultant prs
				 */
    Sex             g;
    Sex             s;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (ZEROP(a))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (ZEROP(b))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    push(a);
	    b = cont(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return ggcd2(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont(a);
	    pop(b);
	    return ggcd2(a, b);
	} else if (o == 0) {
	    int             n = 0;
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push4(nil, nil, nil, nil);
#define a stack[-8]
#define b stack[-7]
#define p stack[-6]
#define q stack[-5]
#define r stack[-4]
#define g2 stack[-3]
#define e1 stack[-2]
#define s1 stack[-1]
#define t1 stack[0]

	    if (power(lterm(poly_pol(a))) < power(lterm(poly_pol(b)))) {
		p = a;
		a = b;
		b = p;
	    }
	    r = cont(a);
	    g2 = cont(b);
	    p = gquot2(a, r);
	    q = gquot2(b, g2);
	    g2 = ggcd2(r, g2);

	    r = wuremainder(p, q);
	    d1 = mkint(lpower(poly_pol(p))) - mkint(lpower(poly_pol(q)));
	    s1 = mkfxat(1);
	    t1 = mkfxat(1);
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("p%d: %p\n", n++, p);
	    if (ezgcd_debug)
		lisp_printf("p%d: %p\n", n++, q);
#endif
	    while (is_poly(r) && poly_var(a) == poly_var(r)) {
		e1 = gexpt(t1, mkfxat(d1 - 1));
		t1 = gexpt(lcoeff(poly_pol(q)), mkfxat(d1));	// psi_2
		t1 = gquot2(t1, e1);
		d1 = mkint(lpower(poly_pol(q))) -
		    mkint(lpower(poly_pol(r)));
		s1 = gexpt(t1, mkfxat(d1));
		s1 = gtimes2(s1, lcoeff(poly_pol(q)));
		p = q;
		q = r;
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf("p%d: %p\n", n++, r);
#endif
		r = wuremainder(p, q);
		r = gquot2(r, s1);
	    }
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf("p%d: %p\n", n++, r);
#endif

	    if (ZEROP(r)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    r = pp(q);
		else
		    r = mkfxat(1);
	    } else
		r = mkfxat(1);

	    g = gtimes2(r, g2);
#undef a
#undef b
#undef p
#undef q
#undef r
#undef g2
#undef e
#undef s1
#undef t1
	    popv(9);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#else
static          Sex
my_gcd_sprs(Sex a, Sex b)
{				/* 
				 * GCD subresultant prs
				 */
    Sex             g;
    Sex             s;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (ZEROP(a))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont_sprs(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (ZEROP(b))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont_sprs(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    push(a);
	    b = cont_sprs(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return my_gcd_sprs(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont_sprs(a);
	    pop(b);
	    return my_gcd_sprs(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push4(nil, nil, nil, nil);
#define a stack[-8]
#define b stack[-7]
#define p stack[-6]
#define q stack[-5]
#define r stack[-4]
#define g2 stack[-3]
#define e1 stack[-2]
#define s1 stack[-1]
#define t1 stack[0]

	    if (power(lterm(poly_pol(a))) < power(lterm(poly_pol(b)))) {
		p = a;
		a = b;
		b = p;
	    }
	    r = cont_sprs(a);
	    g2 = cont_sprs(b);
	    p = gquot2(a, r);
	    q = gquot2(b, g2);
	    g2 = my_gcd_sprs(r, g2);

	    r = wuremainder(p, q);
	    d1 = mkint(lpower(poly_pol(p))) - mkint(lpower(poly_pol(q)));
	    s1 = mkfxat(1);
	    t1 = mkfxat(1);
	    while (is_poly(r) && poly_var(a) == poly_var(r)) {
		e1 = gexpt(t1, mkfxat(d1 - 1));
		t1 = gexpt(lcoeff(poly_pol(q)), mkfxat(d1));	// psi_2
		t1 = gquot2(t1, e1);
		d1 = mkint(lpower(poly_pol(q))) -
		    mkint(lpower(poly_pol(r)));
		s1 = gexpt(t1, mkfxat(d1));
		s1 = gtimes2(s1, lcoeff(poly_pol(q)));
		p = q;
		q = r;
		r = wuremainder(p, q);
		r = gquot2(r, s1);
	    }

	    if (ZEROP(r)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    r = pp_sprs(q);
		else
		    r = mkfxat(1);
	    } else
		r = mkfxat(1);

	    g = gtimes2(r, g2);
#undef a
#undef b
#undef p
#undef q
#undef r
#undef g2
#undef e
#undef s1
#undef t1
	    popv(9);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#endif



static          Sex
cont_euclid(Sex a)
{				/* 
				 * GCD
				 */
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
	p = my_gcd_euclid(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}

static          Sex
pp_euclid(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = cont_euclid(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}


#ifdef  GCD_TEST
static          Sex
my_gcd_euclid(Sex a, Sex b)
{				/* 
				 * GCD subresultant prs
				 */
    Sex             g;
    Sex             s;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (ZEROP(a))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (ZEROP(b))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    push(a);
	    b = cont(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return ggcd2(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont(a);
	    pop(b);
	    return ggcd2(a, b);
	} else if (o == 0) {
	    int             n = 1;
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push4(nil, nil, nil, nil);
#define a stack[-8]
#define b stack[-7]
#define p stack[-6]
#define q stack[-5]
#define g1 stack[-4]
#define g2 stack[-3]

	    if (lpower(poly_pol(a)) < lpower(poly_pol(b))) {
		p = a;
		a = b;
		b = p;
	    }
	    g1 = cont(a);
	    g2 = cont(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = ggcd2(g1, g2);

	    g1 = wuremainder(p, q);
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, p);
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, q);
#endif
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		p = q;
		q = g1;
#ifdef DEBUG
		if (ezgcd_debug)
		    lisp_printf(" p%d: %p\n", n++, g1);
#endif
		g1 = wuremainder(p, q);
	    }
#ifdef DEBUG
	    if (ezgcd_debug)
		lisp_printf(" p%d: %p\n", n++, g1);
#endif
	    if (ZEROP(g1)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = pp(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
	    popv(9);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#else
static          Sex
my_gcd_euclid(Sex a, Sex b)
{				/* 
				 * GCD subresultant prs
				 */
    Sex             g;
    Sex             s;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (ZEROP(a))
	    return b;
	push(a);
	while (is_poly(b))
	    b = cont_euclid(b);
	pop(a);
	return gcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (ZEROP(b))
	    return a;
	push(b);
	while (is_poly(a))
	    a = cont_euclid(a);
	pop(b);
	return gcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    push(a);
	    b = cont_euclid(b);	/* 
				 * bηgcd
				 */
	    pop(a);
	    return my_gcd_euclid(a, b);
	} else if (o < 0) {
	    push(b);
	    a = cont_euclid(a);
	    pop(b);
	    return my_gcd_euclid(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push4(nil, nil, nil, nil);
#define a stack[-8]
#define b stack[-7]
#define p stack[-6]
#define q stack[-5]
#define g1 stack[-4]
#define g2 stack[-3]

	    if (lpower(poly_pol(a)) < lpower(poly_pol(b))) {
		p = a;
		a = b;
		b = p;
	    }
	    g1 = cont_euclid(a);
	    g2 = cont_euclid(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = my_gcd_euclid(g1, g2);

	    g1 = wuremainder(p, q);
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		p = q;
		q = g1;
		g1 = wuremainder(p, q);
	    }
	    if (ZEROP(g1)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = pp_euclid(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
	    popv(9);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}
#endif

/**************************************************************/
/**************************************************************/

static          bool
dividable_by_int(Sex a, Sex b)
{
    Sex             c;
    bool            d = YES;
    ASSERT(is_integer(b));
    if (!is_poly(a)) {
	a = remaind2(a, b);
	if (zerop(a))
	    return YES;
	return NO;
    }

    stackcheck();
    push2(poly_pol(a), b);
#define a stack[-1]
#define b stack[-0]
    while (a != nil) {
	c = lterm(a);
	c = coeff(c);
	d = dividable_by_int(c, b);
	if (d == NO)
	    break;
	a = reduct(a);
    }
#undef a
#undef b
    popv(2);
    return d;
}

static          Sex
remainder_by_int(Sex a, Sex b)
{
    Sex             d,
                    p,
                    x;

    stackcheck();
    if (!is_poly(a)) {
	p = remaind2(a, b);
	return p;
    }
    push3(poly_pol(a), b, poly_var(a));
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = remainder_by_int(coeff(c), b);
	if (!gzerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef x
#undef p
#undef c
    p = stack[-2];
    x = stack[-3];
    popv(6);
    return make_poly(x, cdr(p));
}
static          Sex
remainder_by_int2(Sex a, Sex b, Sex q)
{
    Sex             d,
                    p,
                    x;

    stackcheck();
    if (!is_poly(a)) {
#if 0
	p = remaind2(a, b);
#else
	push2(q, b);
	p = remaind2(a, b);
	push(p);
	q = stack[-2];
	if (!lessp2(p, q)) {
	    p = stack[0];
	    b = stack[-1];
	    p = diff2(p, b);
	}
	popv(3);
#endif
	return p;
    }
    push4(poly_pol(a), b, poly_var(a), q);
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-6]
#define b stack[-5]
#define x stack[-4]
#define q stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = remainder_by_int2(coeff(c), b, q);
	if (!gzerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef x
#undef q
#undef p
#undef c
    p = stack[-2];
    x = stack[-4];
    popv(7);
    return make_poly(x, cdr(p));
}

static          Sex
quotient_by_int(Sex a, Sex b)
{
    Sex             d,
                    p,
                    x;
    if (!is_poly(a))
	return quot2(a, b);
    stackcheck();
    push3(poly_pol(a), b, poly_var(a));
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = quotient_by_int(coeff(c), b);
	// lisp_printf("quotient_by_int(%p, %p) ==> %p\n", coeff(c), b,
	// d);
	if (!gzerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	    // if(ezgcd_debug)lisp_printf("p=%p\n", p);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef x
#undef p
#undef c
    p = stack[-2];
    x = stack[-3];
    popv(6);
    return make_poly(x, cdr(p));
}


#ifdef HURISTIC_2
#define HURISTIC_REPEAT 10
#else
#define HURISTIC_REPEAT 6
#endif

static int      heuristic_miss_count;

static          Sex
my_gcd_heuristic(Sex a, Sex b)
{
    heuristic_miss_count = 0;
    a = my_gcd_heuristic_aux(a, b);
    // if(heuristic_miss_count != 0)
    // printf("heuristic miss count = %d\n", heuristic_miss_count);
    return a;
}


static          Sex
my_gcd_heuristic_aux(Sex a, Sex b)
{
    long            i,
                    k;
    long            na,
                    nb;
    Sex             r;

    stackcheck();
    if (!is_poly(a)) {
	if (!is_poly(b))
	    return gcd(a, b);
	push(a);
	while (is_poly(b))
	    b = cont_sprs(b);
	pop(a);
	return gcd(a, b);
    } else if (!is_poly(b)) {
	push(b);
	while (is_poly(a))
	    a = cont_sprs(a);
	pop(b);
	return gcd(a, b);
    }
    push5(a, b, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push2(nil, nil);
#define a stack[-11]
#define b stack[-10]
#define x stack[-9]
#define y stack[-8]
#define c stack[-7]
#define d stack[-6]
#define g stack[-5]
#define ag stack[-4]
#define bg stack[-3]
#define cg stack[-2]
#define gi stack[-1]
#define ei stack[-0]
    x = all_variables(a);
    y = all_variables(b);
    x = Lunion(x, y);
    if (x == nil)
	return gcd(a, b);
    x = car(x);
    c = Amax_coeff(a);
    d = Amax_coeff(b);
    if (lessp2(d, c))
	c = d;
    c = plus2(c, c);
    y = c;
    g = plus2(c, mkfxat(2));
    for (i = 0; i < HURISTIC_REPEAT; i++) {
	na = mkint(Adeg(a, x));
	nb = mkint(Adeg(b, x));
	if (na > nb)
	    na = nb;
	if (number_log2(g) * na > 5000) {
	    break;
	}
	ag = poly_subst(a, x, g);
	bg = poly_subst(b, x, g);

	// lisp_printf("entering gcd_heuristic(%p, %p, %p)\n", ag, bg ,
	// g);
	cg = my_gcd_heuristic_aux(ag, bg);
	// lisp_printf("gcd_heuristic(%p, %p) => %p\n", ag, bg, cg);
	c = mkfxat(0);
	for (k = 0; !gzerop(cg); k++) {
	    gi = remainder_by_int2(cg, g, y);
	    // lisp_printf("cg = (%p) - (%p)\n", cg, gi);
	    cg = gdiff2(cg, gi);
	    if (k == 0)
		ei = mkfxat(1);
	    else
		ei = make_polynomial(x, mkfxat(1), mkfxat(k));
	    gi = gtimes2(gi, ei);
	    c = gplus2(c, gi);
	    cg = quotient_by_int(cg, g);
	}
	// if(ezgcd_debug)lisp_printf("gcd (%p, %p) = %p\n", a, b ,c);
	if (!zerop(g) && gzerop(gremaind2(a, c))
	    && gzerop(gremaind2(b, c))) {
	    r = c;
	    popv(12);
	    return r;
	}
#ifdef HURISTIC_2
	g = times2(g, mkfxat(73794));
	g = remaind2(g, mkfxat(27011));
#else
	// if(ezgcd_debug)lisp_printf("g = %p\n", g);
	g = plus2(g, g);
	g = plus2(g, mkfxat(1));
#endif
    }
    heuristic_miss_count++;
    r = ggcd2(a, b);
    popv(12);
#undef a
#undef b
#undef x
#undef y
#undef c
#undef d
#undef g
#undef ag
#undef bg
#undef cg
#undef gi
#undef ei
    return r;
}

static          Sex
my_gcd_combine1(Sex a, Sex b)
{
    Sex             h = nil;
    Sex             c;
    long            tdeg_f,
                    tdeg_g,
                    nvars_f,
                    nvars_g,
                    nterms_f,
                    nterms_g;
    double          max_coeff_f,
                    max_coeff_g;
    long            xdeg_f,
                    ydeg_f,
                    xdeg_g,
                    ydeg_g;
    int             nvars;

    stackcheck();
    push5(a, b, nil, nil, nil);
#define f stack[-4]
#define g stack[-3]
#define u stack[-2]
#define w stack[-1]
#define x stack[-0]
    tdeg_f = mkint(Atdeg(f));
    tdeg_g = mkint(Atdeg(g));
    u = Aunknowns(f);
    w = Aunknowns(g);
    nvars_f = list_length(u);
    nvars_g = list_length(w);
    nterms_f = mkint(Anterms(f));
    nterms_g = mkint(Anterms(g));
    u = Amax_coeff(f);
    // max_coeff_f = log(floatit(u));
    max_coeff_f = floatit(u);
    u = Amax_coeff(g);
    // max_coeff_g = log(floatit(u));
    max_coeff_g = floatit(u);
    u = Lunion(u, w);
    nvars = list_length(u);
    if (nvars >= 2) {
	xdeg_f = mkint(Adeg(f, car(u)));
	xdeg_g = mkint(Adeg(g, car(u)));
	ydeg_f = mkint(Adeg(f, cadr(u)));
	ydeg_g = mkint(Adeg(g, cadr(u)));
    } else if (nvars >= 1) {
	xdeg_f = mkint(Adeg(f, car(u)));
	xdeg_g = mkint(Adeg(g, car(u)));
	ydeg_f = 0;
	ydeg_g = 0;
    } else {
	xdeg_f = 0;
	xdeg_g = 0;
	ydeg_f = 0;
	ydeg_g = 0;
    }
    if (max_coeff_f <= 91.0) {
	if (nvars_f <= 7) {
	    if (max_coeff_g <= 5.435) {
		if (max_coeff_f <= 2.0) {
		    if (xdeg_f <= 4)
			h = my_gcd_sprs(f, g);
		    else if (nterms_g <= 42) {
			if (nvars_f <= 5) {
			    if (nterms_f <= 19)
				h = my_gcd_sprs(f, g);
			    else
				h = my_gcd_heuristic(f, g);
			} else {
			    h = my_gcd_sprs(f, g);
			}
		    } else
			h = my_gcd_heuristic(f, g);
		} else {
		    h = my_gcd_heuristic(f, g);
		}
	    } else {
		h = my_gcd_sprs(f, g);
	    }
	} else {
	    h = my_gcd_sprs(f, g);
	}
    } else {
	if (nvars_f <= 1) {
	    if (tdeg_f <= 38)
		h = my_gcd_heuristic(f, g);
	    else
		h = my_gcd_sprs(f, g);
	} else {
	    if (xdeg_g <= 10)
		h = my_gcd_mgcd(f, g);
	    else
		h = my_gcd_heuristic(f, g);
	}
    }
#undef f
#undef g
#undef u
#undef w
#undef x
    popv(5);
    return h;
}


static          Sex
my_gcd_combine2(Sex a, Sex b)
{
    Sex             h = nil;
    Sex             c;
    long            tdeg_f,
                    tdeg_g,
                    nvars_f,
                    nvars_g,
                    nterms_f,
                    nterms_g;
    double          max_coeff_f,
                    max_coeff_g;
    long            xdeg_f,
                    ydeg_f,
                    xdeg_g,
                    ydeg_g;
    int             nvars;
    double          td;
    long            tl;

    stackcheck();
    push5(a, b, nil, nil, nil);
#define f stack[-4]
#define g stack[-3]
#define u stack[-2]
#define w stack[-1]
#define x stack[-0]
    tdeg_f = mkint(Atdeg(f));
    tdeg_g = mkint(Atdeg(g));
    u = Aunknowns(f);
    w = Aunknowns(g);
    nvars_f = list_length(u);
    nvars_g = list_length(w);
    nterms_f = mkint(Anterms(f));
    nterms_g = mkint(Anterms(g));
    u = Amax_coeff(f);
    // max_coeff_f = log(floatit(u));
    max_coeff_f = floatit(u);
    u = Amax_coeff(g);
    // max_coeff_g = log(floatit(u));
    max_coeff_g = floatit(u);
    u = Lunion(u, w);
    if (tdeg_g > tdeg_f) {
	tl = tdeg_f;
	tdeg_f = tdeg_g;
	tdeg_g = tl;
    }
    if (nvars_g > nvars_f) {
	tl = nvars_f;
	nvars_f = nvars_g;
	nvars_g = tl;
    }
    if (nterms_g > nterms_f) {
	tl = nterms_f;
	nterms_f = nterms_g;
	nterms_g = tl;
    }
    if (max_coeff_g > max_coeff_f) {
	td = max_coeff_f;
	max_coeff_f = max_coeff_g;
	max_coeff_g = td;
    }
    nvars = list_length(u);
    if (nvars >= 2) {
	xdeg_f = mkint(Adeg(f, car(u)));
	xdeg_g = mkint(Adeg(g, car(u)));
	if (xdeg_g > xdeg_f) {
	    tl = xdeg_f;
	    xdeg_f = xdeg_g;
	    xdeg_g = tl;
	}
	ydeg_f = mkint(Adeg(f, cadr(u)));
	ydeg_g = mkint(Adeg(g, cadr(u)));
	if (ydeg_g > ydeg_f) {
	    tl = ydeg_f;
	    ydeg_f = ydeg_g;
	    ydeg_g = tl;
	}
    } else if (nvars >= 1) {
	xdeg_f = mkint(Adeg(f, car(u)));
	xdeg_g = mkint(Adeg(g, car(u)));
	if (xdeg_g > xdeg_f) {
	    tl = xdeg_f;
	    xdeg_f = xdeg_g;
	    xdeg_g = tl;
	}
	ydeg_f = 0;
	ydeg_g = 0;
    } else {
	xdeg_f = 0;
	xdeg_g = 0;
	ydeg_f = 0;
	ydeg_g = 0;
    }
    if (nvars_f <= 7) {
	if (max_coeff_g <= 5.0) {
	    if (xdeg_g <= 9) {
		if (xdeg_g <= 6)
		    h = my_gcd_heuristic(f, g);
		else {
		    if (xdeg_f <= 7)
			h = my_gcd_sprs(f, g);
		    else
			h = my_gcd_heuristic(f, g);
		}
	    } else {
		h = my_gcd_sprs(f, g);
	    }
	} else {
	    if (max_coeff_f <= 27.0) {
		h = my_gcd_sprs(f, g);
	    } else {
		if (nvars_f <= 1) {
		    if (nterms_f <= 40)
			h = my_gcd_heuristic(f, g);
		    else
			h = my_gcd_sprs(f, g);
		} else {
		    if (nvars_f <= 2) {
			if (max_coeff_g <= 8472)
			    h = my_gcd_heuristic(f, g);
			else
			    h = my_gcd_mgcd(f, g);
		    } else {
			if (tdeg_f <= 8)
			    h = my_gcd_sprs(f, g);
			else
			    h = my_gcd_mgcd(f, g);
		    }

		}
	    }
	}
    } else {
	h = my_gcd_sprs(f, g);
    }
#undef f
#undef g
#undef u
#undef w
#undef x
    popv(5);
    return h;
}
/**************************************************************/
/**************************************************************/
static          Sex
wquo_int(Sex a, Sex b)
{
    Sex             d,
                    p,
                    x;
    if (!is_poly(a)) {
	d = Ldivide(a, b);
	if (!zerop(cdr(d)))
	    aerror1("wquo: can't divide", d);
	return car(d);
    }

    stackcheck();
    push3(poly_pol(a), b, poly_var(a));
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = wquo_int(coeff(c), b);
	if (!gzerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef x
#undef p
#undef c
    p = stack[-2];
    x = stack[-3];
    popv(6);
    return make_poly(x, cdr(p));
}


static          Sex
wquo_aux(Sex a, Sex b)
{
    Sex             d,
                    p,
                    x;
    if (!is_poly(a)) {
	return aerror2("wquo: can't divide", a, b);
    }

    stackcheck();
    push3(poly_pol(a), b, poly_var(a));
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = wquo(coeff(c), b);
	if (!gzerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef x
#undef p
#undef c
    p = stack[-2];
    x = stack[-3];
    popv(6);
    return make_poly(x, cdr(p));
}

static          Sex
wquo(Sex a, Sex b)
{
    Sex             d;
    Sex             x,
                    y;
    long            o;
    Sex             u,
                    v,
                    w;

    stackcheck();
    if (gzerop(a))
	return a;
    if (!is_poly(b)) {
	return wquo_int(a, b);
    }
    if (!is_poly(a)) {
	return aerror2("wquo: can't divide", a, b);
    }
    x = poly_var(a);
    y = poly_var(b);
    o = compare_var(x, y);
    if (o > 0) {
	return aerror2("wquo: can't divide", a, b);
    } else if (o < 0) {
	return wquo_aux(a, b);
    }
    u = poly_pol(a);
    v = poly_pol(b);
    ASSERT(is_fixnum(lpower(u)));
    ASSERT(is_fixnum(lpower(v)));
    w = nil;
    push5(a, b, x, u, v);
    push5(w, nil, nil, nil, nil);
#define u stack[-6]
#define v stack[-5]
#define w stack[-4]
#define lu stack[-3]
#define lv stack[-2]
#define lw stack[-1]
#define e stack[-0]
    while (u != nil && lpower(u) >= lpower(v)) {
	lu = lcoeff(u);
	lv = lcoeff(v);
	lw = wquo(lu, lv);
	lw = mkterm(lw, lpower(u) - lpower(v) + TAG_FIXNUM);
	w = cons(lw, w);
#ifdef WREM
	u = wquo_sub(u, v, coeff(lw), mkint(power(lw)));
#else
	e = times2tp(lw, reduct(v));
	u = diff2pp(reduct(u), e);
#endif
    }
#undef u
#undef v
#undef w
#undef lu
#undef lv
#undef lw
#undef e
    w = stack[-4];
    w = Lnreverse(w);
    x = stack[-7];
    popv(10);
    return make_poly(x, w);
}

static          Sex
wrem(Sex a, Sex b)
{
    Sex             d;
    Sex             x,
                    y;
    long            o;
    Sex             u,
                    v,
                    w;
    long            n;

    stackcheck();
    if (!is_poly(b)) {
	return mkfxat(0);
    }
    if (!is_poly(a)) {
	return a;
    }
    x = poly_var(a);
    y = poly_var(b);
    o = compare_var(x, y);
    if (o > 0) {
	return a;
    } else if (o < 0) {
	return mkfxat(0);
    }
    u = poly_pol(a);
    v = poly_pol(b);
    ASSERT(is_fixnum(lpower(u)));
    ASSERT(is_fixnum(lpower(v)));
    w = nil;
    push5(a, b, x, u, v);
    push5(w, nil, nil, nil, nil);
#define u stack[-6]
#define v stack[-5]
#define lu stack[-3]
#define lv stack[-2]
#define lw stack[-1]
#define e stack[-0]
    n = mkint(lpower(u)) - mkint(lpower(v)) + 1;
    lv = lcoeff(v);
    while (u != nil && lpower(u) >= lpower(v)) {
	// if(ezgcd_debug)lisp_printf("u = %p\n", u);
	// if(ezgcd_debug)lisp_printf("v = %p\n", v);
	// if(ezgcd_debug)lisp_printf("w = %p\n", w);

	n--;

#ifdef WREM
	lu = lcoeff(u);
	u = wrem_sub(u, lv, 0, v, lu, mkint(lpower(u)) - mkint(lpower(v)));
#else
	lu = lcoeff(u);
	lw = mkterm(lu, lpower(u) - lpower(v) + TAG_FIXNUM);
	e = mkterm(lv, mkfxat(0));
	u = times2tp(e, reduct(u));
	e = times2tp(lw, reduct(v));
	u = diff2pp(u, e);
#endif
    }
    if (n > 0) {
	lv = gexpt(lv, mkfxat(n));
	u = times2tp(mkterm(lv, mkfxat(0)), u);
    }
#undef u
#undef v
#undef lu
#undef lv
#undef lw
#undef e
    x = stack[-7];
    u = stack[-6];
    popv(10);
    return make_poly(x, u);
}

static          Sex
wquo_sub(Sex u, Sex v, Sex b, long bd)	// u * a - v * b --> u;
{
    long            ud,
                    vd;

    stackcheck();
    u = reduct(u);
    v = reduct(v);
    push3(u, v, b);
    push5(nil, nil, nil, nil, nil);
#define u stack[-7]
#define v stack[-6]
#define b stack[-5]
#define uc stack[-4]
#define vc stack[-3]
#define lu stack[-2]
#define lv stack[-1]
#define w stack[-0]

    w = nil;
    while (u != nil && v != nil) {
	lu = lterm(u);
	lv = lterm(v);
	uc = coeff(lu);
	ud = mkint(power(lu));
	vc = coeff(lv);
	vd = mkint(power(lv));
	vd = vd + bd;
	if (ud > vd) {
	    w = cons(lu, w);
	    u = reduct(u);
	} else if (ud == vd) {
	    vc = gtimes2(vc, b);
	    uc = gdiff2(uc, vc);
	    if (!gzerop(uc))
		w = cons(mkterm(uc, mkfxat(ud)), w);
	    u = reduct(u);
	    v = reduct(v);
	} else {
	    vc = gtimes2(vc, b);
	    vc = gnegate(vc);
	    w = cons(mkterm(vc, mkfxat(vd)), w);
	    v = reduct(v);
	}
    }
    while (u != nil) {
	lu = lterm(u);
	w = cons(lu, w);
	u = reduct(u);
    }
    while (v != nil) {
	lv = lterm(v);
	vc = coeff(lv);
	vd = mkint(power(lv));
	vd = vd + bd;
	vc = gtimes2(vc, b);
	vc = gnegate(vc);
	w = cons(mkterm(vc, mkfxat(vd)), w);
	v = reduct(v);
    }
    w = Lnreverse(w);
#undef u
#undef v
#undef b
#undef uc
#undef vc
#undef lu
#undef lv
#undef w
    u = stack[0];
    popv(8);
    return u;
}


static          Sex
wrem_sub(Sex u, Sex a, long ad, Sex v, Sex b, long bd)	// u * a - v * b
							// --> u;
{
    long            ud,
                    vd;

    stackcheck();
    u = reduct(u);
    v = reduct(v);
    push4(u, a, v, b);
    push5(nil, nil, nil, nil, nil);
#define u stack[-8]
#define a stack[-7]
#define v stack[-6]
#define b stack[-5]
#define uc stack[-4]
#define vc stack[-3]
#define lu stack[-2]
#define lv stack[-1]
#define w stack[-0]

    w = nil;
    while (u != nil && v != nil) {
	lu = lterm(u);
	lv = lterm(v);
	uc = coeff(lu);
	ud = mkint(power(lu));
	vc = coeff(lv);
	vd = mkint(power(lv));
	vd = vd + bd;
	ud = ud + ad;
	if (ud > vd) {
	    uc = gtimes2(uc, a);
	    w = cons(mkterm(uc, mkfxat(ud)), w);
	    u = reduct(u);
	} else if (ud == vd) {
	    vc = gtimes2(vc, b);
	    uc = gtimes2(uc, a);
	    uc = gdiff2(uc, vc);
	    if (!gzerop(uc))
		w = cons(mkterm(uc, mkfxat(ud)), w);
	    u = reduct(u);
	    v = reduct(v);
	} else {
	    vc = gtimes2(vc, b);
	    vc = gnegate(vc);
	    w = cons(mkterm(vc, mkfxat(vd)), w);
	    v = reduct(v);
	}
    }
    while (u != nil) {
	lu = lterm(u);
	uc = coeff(lu);
	ud = mkint(power(lu));
	ud = ud + ad;
	uc = gtimes2(uc, a);
	w = cons(mkterm(uc, mkfxat(ud)), w);
	u = reduct(u);
    }
    while (v != nil) {
	lv = lterm(v);
	vc = coeff(lv);
	vd = mkint(power(lv));
	vd = vd + bd;
	vc = gtimes2(vc, b);
	vc = gnegate(vc);
	w = cons(mkterm(vc, mkfxat(vd)), w);
	v = reduct(v);
    }
    w = Lnreverse(w);
#undef u
#undef a
#undef v
#undef b
#undef uc
#undef vc
#undef lu
#undef lv
#undef w
    u = stack[0];
    popv(9);
    return u;
}



/**************************************************************/
/**************************************************************/

static          Sex
Lezgcd_debug(Sex a)
{
    if (a == nil)
	ezgcd_debug = 0;
    else
	ezgcd_debug = 1;
    return nil;
}

/**************************************************************/
/**************************************************************/

setup_type const gcd_setup[] = {
    {"ezgcd_debug", (lisp_proc *) Lezgcd_debug, 1},
    {"gcd_mgcd", (lisp_proc *) my_gcd_mgcd, 2},
    {"gcd_ezgcd", (lisp_proc *) my_gcd_ezgcd, 2},
    {"gcd_ezgcd1", (lisp_proc *) my_gcd_ezgcd1, 2},
    {"gcd_ezgcd2", (lisp_proc *) my_gcd_ezgcd2, 2},
    {"gcd_ezgcd3", (lisp_proc *) my_gcd_ezgcd3, 2},
    {"gcd_ezgcd4", (lisp_proc *) my_gcd_ezgcd4, 2},
    {"gcd_sprs", (lisp_proc *) my_gcd_sprs, 2},
    {"gcd_pprs", (lisp_proc *) my_gcd_pprs, 2},
    {"gcd_rprs", (lisp_proc *) my_gcd_rprs, 2},
    {"gcd_euclid", (lisp_proc *) my_gcd_euclid, 2},
    {"gcd_heuristic", (lisp_proc *) my_gcd_heuristic, 2},
    {"gcd_huristic", (lisp_proc *) my_gcd_heuristic, 2},
    {"gcd_combine", (lisp_proc *) my_gcd_combine2, 2},
    {"cont_mgcd", (lisp_proc *) cont_mgcd, 1},
    {"remainder_by_int2", (lisp_proc *) remainder_by_int2, 3},
    {"quotient_by_int", (lisp_proc *) quotient_by_int, 2},
    {"remainder_by_int", (lisp_proc *) remainder_by_int, 2},
    {"wrem", (lisp_proc *) wrem, 2},
    {"wquo", (lisp_proc *) wquo, 2},
    {NULL, 0, 0}
};
