/*
 * groebner.c      (C) Copyright 2015 by Fujio Kako, Nara Women's University.
 * Groebner base Functions.
 *      $Header: /Users/kako/asuka/src/RCS/groebner.c,v 2.22 2015/06/26 08:33:54 kako Exp $
 */

static char const rcsid[] =
    "$Id: groebner.c,v 2.22 2015/06/26 08:33:54 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "entries.h"
#include "misc.h"
#include "fns.h"
#include "polynomial.h"
#include "tri.h"
#include "garith.h"

#define cddr(x)	cdr(cdr(x))
#define cdar(x)	cdr(car(x))
#define caar(x)	car(car(x))

Sex             Agroebner(Sex a);
Sex             Agroebner_algorithm(Sex a);
Sex             Agroebner_simple(Sex a);
Sex             Agroebner42(Sex a);
Sex             Agroebner44(Sex a);
Sex             Areducedgb(Sex a);
Sex             Aminimalgb(Sex a);
Sex             Ais_groebner(Sex a, Sex p);
Sex             Acyclic(Sex n);
Sex             Akatsura(Sex n);

extern Sex      Atdeg(Sex f);
extern Sex      Aspoly(Sex a, Sex b);
extern Sex      Afull_reduce(Sex s, Sex G);
extern Sex      Lnconc(Sex a, Sex b);
extern Sex      Laconc(Sex a, Sex b);
extern Sex      Lclone(Sex a);
extern Sex      Alt_reducible(Sex f, Sex g);
extern Sex      Asort_pairset(Sex Q);
extern Sex      Amerge_pairset(Sex P, Sex Q);
// extern Sex Aexpvec(Sex g);
extern Sex      Aordervec(Sex g);
extern Sex      Aexpvec_lcm(Sex a, Sex b);
extern Sex      Aexpvec_div(Sex a, Sex b);
extern Sex      Ldelete(Sex f, Sex P);
extern Sex      Asetdiff(Sex a, Sex b);
extern Sex      Ahave_nontrivial_gcd(Sex a, Sex b);
// extern Sex Alt_lcm_expvec(Sex a, Sex b);
extern Sex      Alt_lcm_ordervec(Sex a, Sex b);
extern Sex      Lmkarray1(Sex n);
extern Sex      Alist2array(Sex a, Sex b);
extern Sex      Lmake_id(Sex a);
extern Sex      symbol_to_polynomial(Sex a);
extern Sex      Areduce(Sex a, Sex b);
extern long     compare(Sex a, Sex b);
extern Sex      Lgetdipolyorder(void);

Sex             Agroebner_algorithm44(Sex G);
Sex             Agroebner_algorithm42(Sex G);
Sex             upatePairs(Sex D, Sex G, Sex s);
Sex             selectPair(Sex D);
Sex             selectPoly(Sex D);
Sex             reduceSet(Sex G);
Sex             createPair(Sex G);
Sex             makePair(Sex G, Sex f);
Sex             updatePairs(Sex D, Sex G, Sex gk);

bool            criterion1(Sex a);
bool            criterion2(Sex s, Sex p, Sex g);
bool            cdr_member(Sex s, Sex p);

Sex             criterionM(Sex G);
Sex             criterionF(Sex G);
Sex             criterionB(Sex G);

int             check_gb(Sex G, Sex P);
Sex             reduced_gb(Sex G);
Sex             minimal_gb(Sex G);

static Sex      mku(long l, long n);
static Sex      mkc(long l, long n);
static Sex      makeid(long c, long n);

/**************************************************************/
/**************************************************************/

extern Sex      Anewvector(Sex a);
extern Sex      make_matrix(long n, long m);
extern Sex      Mlcm(Sex a, Sex b);
static Sex      F4(Sex F);
static Sex      F4_HT(Sex f);
static Sex      F4_HT2(Sex F);
static Sex      F4_HM(Sex f);
static Sex      F4_HC(Sex f);
static Sex      makeP(Sex g1, Sex g2);
static Sex      makePair1(Sex G);
static Sex      makePair2(Sex h, Sex G);
static Sex      Left(Sex Pd);
static Sex      Right(Sex Pd);
static Sex      Left2(Sex Pd);
static Sex      Right2(Sex Pd);
static Sex      make_monolist(Sex f);
static Sex      make_coefflist(Sex f);
static Sex      make_monolist_from_list(Sex F);
static Sex      make_coeffmatrix(Sex F);
static Sex      symbolic_preprocessing(Sex L, Sex G);
static Sex      F4_Reduction(Sex L, Sex G);
static Sex      Sel(Sex P);
static Sex      row_echelon_form(Sex a, long n, long m);
static Sex      wari(Sex f, Sex g);
static Sex      LT(Sex f);
/**************************************************************/
/**************************************************************/

#define F5a

static Sex      poly(Sex f);
static Sex      poly_list(Sex R);
static Sex      check_dipoly_list(Sex F);
static Sex      F5(Sex F);
#ifdef F5a
static Sex      AlgorithmF5(long i, Sex f, Sex G);
#else
static Sex      F5_inc(long i, Sex f, Sex G);
#endif


static Sex      compare_signature2(Sex u1, Sex r1, Sex u2, Sex r2);
static Sex      top_reducible(Sex t, Sex G);
static Sex      CritPair(Sex r1, Sex r2, long k, Sex G);
static Sex      CritPair2(Sex r1, Sex r2, long k, Sex G);

static Sex      Spol(Sex P);
static Sex      NF(Sex f, Sex F);
static Sex      NF2(Sex f, Sex F);
static Sex      ht_div(Sex f, Sex g);
static Sex      F4_HM(Sex f);

static Sex      HT(Sex f);
static Sex      HC(Sex f);
static Sex      ResetRules(long m);
static Sex      AddRules(Sex r, long n);
#ifdef F5a
static Sex      Rewrite(Sex u, Sex rk);
static Sex      IsRewritable(Sex u, Sex rk);
static Sex      Reduction(Sex Todo, Sex Gp, Sex G);
static Sex      FindReductor(Sex rk, Sex Gp, Sex G);
static Sex      TopReduction(Sex rk, Sex Gp, Sex G);
#else
static Sex      Rewritten(Sex u, Sex rk);
static Sex      IsRewritten(Sex u, Sex rk);
static Sex      Reduction(Sex Todo, Sex Gp, long k, Sex G);
static Sex      IsReducible(Sex rk, Sex Gp, long k, Sex G);
static Sex      TopReduction(Sex rk, Sex Gp, long k, Sex G);
#endif
static Sex      term_div(Sex f, Sex g);
static Sex      merge_by_degree(Sex u, Sex v);

static Sex      sort_by_signature(Sex u);
static Sex      quick_sort_by_signature(Sex a, long l, long r);
static Sex      sort_by_degree(Sex u);
static Sex      quick_sort_by_degree(Sex a, long l, long r);
static bool     compare_by_degree(Sex u, Sex v);
static bool     compare_by_signature(Sex u, Sex v);
static long     compare_signature(Sex u, Sex v);
static Sex      minimal_by_signature(Sex u);
static Sex      NF_with_signature(Sex f, Sex F);

#define GROEBNER_SIMPLE	0
#define GROEBNER_42	1
#define GROEBNER_44	2
#define GROEBNER_F4	3
#define GROEBNER_F5	4
#define GROEBNER_F5C	5

static int      groebner_algorithm = 0;

/**************************************************************/
/**************************************************************/

Sex
Agroebner(Sex g)
{
    switch (groebner_algorithm) {
    case GROEBNER_SIMPLE:
	return Agroebner_simple(g);
    case GROEBNER_42:
	return Agroebner_algorithm42(g);
    case GROEBNER_44:
	return Agroebner_algorithm44(g);
    case GROEBNER_F4:
	return F4(g);
    case GROEBNER_F5:
	return F5(g);
    }
    return Agroebner_algorithm42(g);
}

Sex
Agroebner_algorithm(Sex a)
{
    Sex             c;
    long            t;
    t = mkint(a);
    c = mkfxat(groebner_algorithm);
    t &= 0xf;
    groebner_algorithm = t;
    return c;
}

Sex
Agroebner44(Sex g)
{
    return Agroebner_algorithm44(g);
}

Sex
Agroebner42(Sex g)
{
    return Agroebner_algorithm42(g);
}

Sex
Areducedgb(Sex g)
{
    return reduced_gb(g);
}

Sex
Aminimalgb(Sex g)
{
    return minimal_gb(g);
}

Sex
Ais_groebner(Sex g, Sex p)
{
    int             r;
    r = check_gb(g, p);
    return r > 0 ? true : nil;
}

/**************************************************************/
/**************************************************************/
Sex
Acyclic(Sex a)
{
    long            n,
                    i,
                    j,
                    l;
    Sex             c;

    if (!is_fixnum(a))
	return nil;
    n = mkint(a);
    if (n <= 0 || n >= 20)
	return nil;
    stackcheck();
    push3(nil, nil, nil);
#define a stack[-2]
#define b stack[-1]
#define r stack[-0]

    for (l = 1; l <= n; l++) {
	a = mkfxat(0);
	for (i = 0; i < n; i++) {
	    b = mkfxat(1);
	    for (j = i; j <= l + i - 1; j++) {
		c = mkc(j, n);
		b = gtimes2(b, c);
	    }
	    a = gplus2(a, b);
	}
	if (l == n)
	    a = gdiff2(a, mkfxat(n));
	r = cons(a, r);
    }
    c = r;
#undef a
#undef b
#undef r
    popv(3);
    return c;
}

Sex
Akatsura(Sex a)
{
    long            l,
                    i;
    long            n;
    Sex             c;

    if (!is_fixnum(a))
	return nil;
    n = mkint(a);
    if (n <= 0 || n >= 20)
	return nil;
    stackcheck();
    push3(nil, nil, nil);
#define a stack[-2]
#define b stack[-1]
#define r stack[0]
    for (l = 0; l < n; l++) {
	c = mku(l, n);
	a = gnegate(c);
	for (i = -n; i <= n; i++) {
	    b = mku(l, n);
	    c = mku(l - i, n);
	    c = gtimes2(c, b);
	    a = gplus2(a, c);
	}
	r = cons(a, r);
    }
    a = mkfxat(0);
    for (l = -n; l <= n; l++) {
	c = mku(l, n);
	a = gplus2(a, c);
    }
    a = gdiff2(a, mkfxat(1));
    c = cons(a, r);
    popv(3);
#undef a
#undef b
#undef r
    return c;
}

static          Sex
mkc(long l, long n)
{
    if (l >= n)
	l = l - n;
    if (l >= 0)
	return makeid('c', l + 1);
    return mkfxat(0);
}

static          Sex
mku(long l, long n)
{
    if (l < 0)
	l = -l;
    if (l <= n)
	return makeid('u', l + 1);
    return mkfxat(0);
}

static          Sex
makeid(long c, long n)
{
    Sex             a;
    a = Lexplode(mkfxat(n));
    a = cons(ASUKA_CHAR_OF(c), a);
    a = Lmake_id(a);
    a = symbol_to_polynomial(a);
    return a;
}
/**************************************************************/
/**************************************************************/

Sex
Agroebner_simple(Sex G)
{
    Sex             r;
    stackcheck();
    push4(G, nil, nil, nil);
    push3(nil, nil, nil);

#define G stack[-6]
#define S stack[-5]
#define gi stack[-4]
#define gj stack[-3]
#define s stack[-2]
#define i stack[-1]
#define j stack[-0]
    if (!is_cons(G)) {
	r = G;
	popv(7);
	return aerror1("not list", r);
    }
    G = Lclone(G);
    while (1) {
	S = nil;
	gi = G;
	while (gi != nil) {
	    i = car(gi);
	    gi = cdr(gi);
	    gj = gi;
	    while (gj != nil) {
		j = car(gj);
		gj = cdr(gj);
		/*
		 * lisp_printf("spoly: i=%p, j=%p\n", i, j);
		 */
		s = Aspoly(i, j);
		s = Afull_reduce(s, G);
		if (!gzerop(s))
		    S = Laconc(S, s);
	    }
	}
	if (S == nil)
	    break;
	G = Lnconc(G, S);
    }
    r = G;
#undef G
#undef S
#undef gi
#undef gj
#undef s
#undef i
#undef j
    popv(7);
    return r;
}

/**************************************************************/
/**************************************************************/

Sex
Agroebner_algorithm44(Sex G)
{
    Sex             r;
    stackcheck();
    push4(G, nil, nil, nil);
    push3(nil, nil, nil);

#define G stack[-6]
#define D stack[-5]
#define gi stack[-4]
#define gj stack[-3]
#define s stack[-2]
#define p stack[-1]
#define F stack[-0]
    if (!is_cons(G)) {
	r = G;
	popv(7);
	return aerror1("not list", r);
    }
    F = reduceSet(G);
    G = nil;
    D = nil;
    while (F != nil) {
	s = car(F);
	F = cdr(F);
	D = updatePairs(D, G, s);
	G = Laconc(G, s);
    }
    while (D != nil) {
	s = selectPair(D);
	D = Ldelete(s, D);
	gj = cadr(s);
	gi = cddr(s);
	p = Aspoly(gi, gj);
	p = Afull_reduce(p, G);
	if (!gzerop(p)) {
	    D = updatePairs(D, G, p);
	    G = Laconc(G, p);
	}
    }
    r = reduceSet(G);
#undef G
#undef D
#undef gi
#undef gj
#undef s
#undef p
#undef F
    popv(7);
    return r;
}

Sex
updatePairs(Sex D, Sex G, Sex gk)
{
    Sex             r;
    int             flag;

    stackcheck();
    push3(D, G, gk);
    push5(nil, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push4(nil, nil, nil, nil);

#define D stack[-16]
#define G stack[-15]
#define gk stack[-14]
#define D1 stack[-13]
#define Dp stack[-12]
#define H stack[-11]
#define d stack[-10]
#define d1 stack[-9]
#define tij stack[-8]
#define tijk stack[-7]
#define tik stack[-6]
#define tjk stack[-5]
#define tk stack[-4]
#define gj stack[-3]
#define gi stack[-2]
#define g stack[-1]
#define P stack[-0]
    tk = Aordervec(gk);
    P = D;
    while (P != nil) {
	d = car(P);
	P = cdr(P);
	gi = cadr(d);
	gj = cddr(d);
	tij = Alt_lcm_ordervec(gi, gj);
	tik = Alt_lcm_ordervec(gi, gk);
	tjk = Alt_lcm_ordervec(gj, gk);
	tijk = Aexpvec_lcm(tij, tk);
	if (Lequal(tijk, tij) != nil && Lequal(tik, tij) == nil &&
	    Lequal(tjk, tij) == nil)
	    D = Ldelete(d, D);
    }
    D1 = makePair(G, gk);
    H = D1;
    flag = NO;
    while (H != nil) {
	d = car(H);
	H = cdr(H);
	tik = car(d);
	P = H;
	while (P != nil) {
	    d1 = car(P);
	    P = cdr(P);
	    tjk = car(d1);
	    if (Aexpvec_div(tik, tjk) != nil) {
		D1 = Ldelete(d, D1);
		break;
	    }
	}
    }

    H = D1;
    while (H != nil) {
	d = car(H);
	gi = cadr(d);
	gj = cddr(d);
	flag = Ahave_nontrivial_gcd(gi, gj) == nil;
	H = cdr(H);
	tik = car(d);
	while (H != nil) {
	    d1 = car(H);
	    gi = car(d1);
	    if (Lequal(gi, tik) == nil)
		break;
	    if (!flag) {
		gi = cadr(d1);
		gj = cddr(d1);
		flag = Ahave_nontrivial_gcd(gi, gj) == nil;
	    }
	    D1 = Ldelete(d1, D1);
	    H = cdr(H);
	}
	if (flag)
	    D1 = Ldelete(d, D1);
    }
    r = Amerge_pairset(D, D1);
#undef D
#undef G
#undef gk
#undef D1
#undef Dp
#undef H
#undef d
#undef d1
#undef tij
#undef tijk
#undef tik
#undef tjk
#undef tk
#undef gj
#undef gi
#undef g
#undef P
    popv(17);
    return r;
}

Sex
selectPair(Sex D)
{
    return car(D);
}

Sex
selectPoly(Sex R)
{
    return car(R);
}

Sex
reduceSet(Sex G)
{
    Sex             r;
    stackcheck();
    push3(nil, nil, nil);
    push3(nil, nil, nil);
#define H stack[-5]
#define R stack[-4]
#define P stack[-3]
#define Q stack[-2]
#define h stack[-1]
#define q stack[-0]

    R = Lclone(G);
    P = nil;
    while (R != nil) {
	h = selectPoly(R);
	R = Ldelete(h, R);
	h = Afull_reduce(h, P);
	if (!gzerop(h)) {
	    Q = nil;
	    H = P;
	    while (H != nil) {
		q = car(H);
		H = cdr(H);
		if (Alt_reducible(q, h) != nil)
		    Q = cons(q, Q);
	    }
	    R = Lnconc(R, Q);
	    P = Asetdiff(P, Q);
	    P = Laconc(P, h);
	}
    }
    Q = nil;
    while (P != nil) {
	h = car(P);
	P = cdr(P);
	R = Ldelete(h, P);
	h = Afull_reduce(h, R);
	Q = Laconc(Q, h);
    }
    r = Q;
#undef H
#undef R
#undef P
#undef Q
#undef h
#undef q
    popv(6);
    return r;
}


Sex
createPair(Sex G)
{
    stackcheck();

    push5(G, nil, nil, nil, nil);
#define G stack[-4]
#define fi stack[-3]
#define fj stack[-2]
#define P stack[-1]
#define H stack[-0]

    P = nil;
    while (G != nil) {
	fi = car(G);
	G = cdr(G);
	H = G;
	while (H != nil) {
	    fj = car(H);
	    H = cdr(H);
	    fj = cons(fi, fj);
	    P = cons(fj, P);
	}
    }
#undef G
    G = Asort_pairset(P);
#undef f
#undef fi
#undef P
#undef H
    popv(5);
    return G;
}

Sex
makePair(Sex G, Sex f)
{
    stackcheck();
    push4(G, f, nil, nil);
#define G stack[-3]
#define f stack[-2]
#define fi stack[-1]
#define P stack[-0]

    P = nil;
    while (G != nil) {
	fi = car(G);
	G = cdr(G);
	fi = cons(fi, f);
	P = cons(fi, P);
    }
#undef G
    G = Asort_pairset(P);
#undef f
#undef fi
#undef P
    popv(4);
    return G;
}
/**************************************************************/
/**************************************************************/

Sex
Agroebner_algorithm42(Sex F)
{
    Sex             r;
    if (!is_cons(F))
	return aerror1("not list", F);

    stackcheck();
    push5(F, nil, nil, nil, nil);
    push2(nil, nil);
#define F stack[-6]
#define G stack[-5]
#define P stack[-4]
#define gi stack[-3]
#define gj stack[-2]
#define s stack[-1]
#define p stack[-0]

    G = reduceSet(F);
    P = createPair(G);
    while (P != nil) {
	s = selectPair(P);
	P = Ldelete(s, P);
	if (criterion1(cdr(s)) && !criterion2(s, P, G)) {
	    s = cdr(s);
	    gj = car(s);
	    gi = cdr(s);
	    s = Aspoly(gi, gj);
	    s = Afull_reduce(s, G);
	    if (!gzerop(s)) {
		p = makePair(G, s);
		P = Amerge_pairset(P, p);
		p = Lncons(s);
		G = Lappend(G, p);
	    }
	}
    }
    r = reduceSet(G);
#undef F
#undef G
#undef P
#undef gi
#undef gj
#undef s
#undef p
    popv(7);
    return r;
}


bool
criterion1(Sex s)
{
    Sex             a,
                    b;
    a = car(s);
    b = cdr(s);
    a = Ahave_nontrivial_gcd(a, b);
    return a != nil;
}

bool
criterion2(Sex s, Sex B, Sex G)
{
    stackcheck();
    push3(s, B, G);
    push4(nil, nil, nil, nil);
    push4(nil, nil, nil, nil);

#define s stack[-10]
#define B stack[-9]
#define G stack[-8]
#define g stack[-7]
#define gi stack[-6]
#define gj stack[-5]
#define gk stack[-4]
#define tij stack[-3]
#define tijk stack[-2]
#define tk stack[-1]
#define ss stack[-0]
    gi = cadr(s);
    gj = cddr(s);
    tij = Alt_lcm_ordervec(gi, gj);
    g = G;
    while (g != nil) {
	gk = car(g);
	g = cdr(g);

	if (Lequal(gi, gi) != nil || Lequal(gj, gk) != nil)
	    continue;
	tk = Aordervec(gk);
	tijk = Aexpvec_lcm(tij, tk);
	if (Lequal(tij, tijk) == nil)
	    continue;
	if (Lmember(gi, g) != nil)
	    ss = Llist2(gk, gi);
	else
	    ss = Llist2(gi, gk);
	if (cdr_member(ss, B))
	    continue;
	if (Lmember(gj, g) != nil)
	    ss = Llist2(gk, gj);
	else
	    ss = Llist2(gj, gk);
	if (cdr_member(ss, B))
	    continue;
	popv(11);
	return YES;
    }
    popv(11);
    return NO;
#undef s
#undef B
#undef G
#undef g
#undef gi
#undef gj
#undef gk
#undef tij
#undef tijk
#undef tk
#undef ss
}

bool
cdr_member(Sex a, Sex s)
{
    stackcheck();
    push2(a, s);
#define a stack[-1]
#define s stack[0]
    while (s != nil) {
	if (Lequal(a, cdar(s)) != nil) {
	    popv(2);
	    return YES;
	}
	s = cdr(s);
    }
    popv(2);
    return NO;
#undef a
#undef s
}

Sex
criterionM(Sex G)
{				/* M[i,j] == k < j and Tkj | Tij and Tkj # 
				 * Tij */
    Sex             r;
    stackcheck();
    push5(G, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push3(nil, nil, nil);
#define G stack[-12]
#define h stack[-11]
#define k stack[-10]
#define gi stack[-9]
#define gj stack[-8]
#define gk stack[-7]
#define g stack[-6]
#define tkj stack[-5]
#define tij stack[-4]
#define ti stack[-3]
#define tj stack[-2]
#define tk stack[-1]
#define M stack[-0]
    g = G;
    M = nil;
    while (g != nil) {
	gi = car(g);
	g = cdr(g);
	h = g;
	while (h != nil) {
	    gj = car(h);
	    h = cdr(h);
	    tij = Alt_lcm_ordervec(gi, gj);
	    k = G;
	    while (k != nil) {
		gk = car(k);
		k = cdr(k);
		if (Lequal(gk, gj) != nil)
		    break;
		tk = Aordervec(gk);
		tkj = Aexpvec_lcm(tk, tj);
		if (Aexpvec_div(tij, tkj) != nil
		    && Lequal(tij, tkj) == nil) {
		    gk = Llist2(gi, gj);
		    M = cons(gk, M);
		    break;
		}
	    }
	}
    }
    r = M;
    popv(13);
#undef G
#undef h
#undef k
#undef gi
#undef gj
#undef gk
#undef g
#undef tkj
#undef tij
#undef ti
#undef tj
#undef tk
#undef M
    return r;
}

Sex
criterionF(Sex G)
{
    Sex             r;
    stackcheck();
    push5(G, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push(nil);
#define G stack[-10]
#define h stack[-9]
#define k stack[-8]
#define gi stack[-7]
#define gj stack[-6]
#define gk stack[-5]
#define g stack[-4]
#define tij stack[-3]
#define tkj stack[-2]
#define tik stack[-1]
#define F stack[-0]
    g = G;
    F = nil;
    while (g != nil) {
	gi = car(g);
	g = cdr(g);
	h = g;
	while (h != nil) {
	    gj = car(h);
	    h = cdr(h);
	    tij = Alt_lcm_ordervec(gi, gj);
	    k = G;
	    while (k != nil) {
		gk = car(k);
		k = cdr(k);
		if (Lequal(gk, gi) != nil)
		    break;
		tkj = Alt_lcm_ordervec(gk, gj);
		if (Lequal(tik, tij) != nil) {
		    gk = Llist2(gk, gj);
		    F = cons(gk, F);
		    break;
		}
	    }
	}
    }
    r = F;
#undef G
#undef h
#undef k
#undef gi
#undef gj
#undef gk
#undef g
#undef tij
#undef tkj
#undef tik
#undef F
    popv(11);
    return r;
}


Sex
criterionB(Sex G)
{
    Sex             r;
    stackcheck();
    push5(G, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push2(nil, nil);
#define G stack[-11]
#define h stack[-10]
#define k stack[-9]
#define gi stack[-8]
#define gj stack[-7]
#define gk stack[-6]
#define g stack[-5]
#define tij stack[-4]
#define tjk stack[-3]
#define tik stack[-2]
#define tk stack[-1]
#define B stack[-0]
    g = G;
    B = nil;
    while (g != nil) {
	gi = car(g);
	g = cdr(g);
	h = g;
	while (h != nil) {
	    gj = car(h);
	    h = cdr(h);
	    tij = Alt_lcm_ordervec(gi, gj);
	    k = h;
	    while (k != nil) {
		gk = car(k);
		k = cdr(k);
		tk = Aordervec(gk);
		tij = Alt_lcm_ordervec(gi, gk);
		tjk = Alt_lcm_ordervec(gj, gk);
		if (Aexpvec_div(tij, tk) == nil && Lequal(tjk, tij) == nil
		    && Lequal(tik, tij) == nil) {
		    gk = Llist2(gi, gj);
		    B = cons(gk, B);
		    break;
		}
	    }
	}
    }
    r = B;
#undef G
#undef h
#undef k
#undef gi
#undef gj
#undef gk
#undef g
#undef tij
#undef tkj
#undef tik
#undef tk
#undef B
    popv(12);
    return r;
}

/**************************************************************/
/**************************************************************/
int
check_gb(Sex G, Sex P)
{
    stackcheck();
    push2(G, P);
    push4(nil, nil, nil, nil);
#define G stack[-5]
#define P stack[-4]
#define i stack[-3]
#define j stack[-2]
#define s stack[-1]
#define H stack[-0]
    while (P != nil) {
	i = car(P);
	P = cdr(P);
	s = Areduce(i, G);
	if (!gzerop(s)) {
	    popv(6);
	    return -1;
	}
    }
    H = G;
    while (H != nil) {
	i = car(H);
	H = cdr(H);
	P = H;
	while (P != nil) {
	    j = car(P);
	    P = cdr(P);
	    s = Aspoly(i, j);
	    s = Areduce(s, G);
	    if (!gzerop(s)) {
		popv(6);
		return 0;
	    }
	}
    }
    popv(6);
    return 1;
#undef G
#undef P
#undef i
#undef j
#undef s
#undef H
}

/**************************************************************/
/**************************************************************/
#if 0
Sex
minimal_gb(Sex F)
{
    Sex             r;

    stackcheck();
    push5(F, nil, nil, nil, nil);
#define F stack[-4]
#define G stack[-3]
#define H stack[-2]
#define f stack[-1]
#define g stack[0]
    G = nil;
    while (F != nil) {
	f = car(F);
	F = cdr(F);
	r = Minit(f);
	f = gquot2(f, r);
	G = Laconc(G, f);
    }
    F = G;
    while (F != nil) {
	f = car(F);
	F = cdr(F);
	H = G;
	while (H != nil) {
	    g = car(H);
	    H = cdr(H);
	    if (Alt_reducible(f, g) != nil) {
		G = Ldelete(f, G);
		break;
	    }
	}
    }
    r = G;
    popv(5);
#undef F
#undef G
#undef H
#undef f
#undef g
    return r;
}

#else

Sex
minimal_gb(Sex G)
{
    int             i,
                    j,
                    n;
    Sex             r;

    stackcheck();
    push5(G, nil, nil, nil, nil);
#define G stack[-4]
#define a stack[-3]
#define g stack[-2]
#define p stack[-1]
#define b stack[-0]
    n = list_length(G);
    a = Lmkarray1(mkfxat(n));
    for (i = 0; i < n; i++) {
	g = car(G);
	G = cdr(G);
	r = Minit(g);
	g = gquot2(g, r);
	array_value(a, i) = g;
    }

    for (i = 0; i < n; i++) {
	g = array_value(a, i);
	if (g == mkfxat(0))
	    continue;
	for (j = 0; j < n; j++) {
	    if (i == j)
		continue;
	    b = array_value(a, j);
	    if (b == mkfxat(0))
		continue;
	    if (Alt_reducible(g, b) != nil) {
		array_value(a, i) = mkfxat(0);
		break;
	    }
	}
    }
    p = nil;
    for (i = 0; i < n; i++) {
	g = array_value(a, i);
	if (g != mkfxat(0))
	    p = Laconc(p, g);
    }
    r = p;
    popv(5);
#undef G
#undef a
#undef g
#undef p
#undef b
    return r;
}

#endif

Sex
reduced_gb(Sex G)
{
    Sex             r;

    G = minimal_gb(G);
    stackcheck();
    push3(G, nil, nil);
#define G stack[-2]
#define H stack[-1]
#define g stack[-0]
    H = Lclone(G);
    while (G != nil) {
	g = car(G);
	G = cdr(G);
	H = Ldelete(g, H);
	g = Afull_reduce(g, H);
	if (!gzerop(g)) {
	    r = Minit(g);
	    g = gquot2(g, r);
	    H = Laconc(H, g);
	}
    }
    r = H;
    popv(3);
#undef G
#undef H
#undef g
    return r;
}

/**************************************************************/
/**************************************************************/

/**************************************************************/
/*
 * F4 algorithm 
 */
/**************************************************************/
/*
 * L:={{d,a*b*c+a*b*d+a*c*d+b*c*d},{1,a*b*c*d-1}}; G:={a*b*c*d-1}; 
 */

#include "f4a.c"
/**************************************************************/
/*
 * Fauger F5 algorithm 
 */
/**************************************************************/
#ifdef F5a
#include "f5a.c"
#else
#include "f5.c"
#endif
/**************************************************************/
/*
 * call fgb library 
 */
/**************************************************************/
#if 0
extern Sex      call_fgb_int(Sex F, long n, long m, long d);
extern Sex      call_fgb_mod(Sex F, long n, long m, long d, long q);
#else
static          Sex
call_fgb_int(Sex F, long n, long m, long d)
{
    return nil;
}
static          Sex
call_fgb_mod(Sex F, long n, long m, long d, long q)
{
    return nil;
}
#endif

static          Sex
FGb_int(Sex F)
{
    long            n,
                    m,
                    d;
    F = Ldipoly(F);
    n = list_length(F);
    m = list_length(dipoly_kernel_order);
    d = mkint(car(dipoly_kernel_order));
    F = call_fgb_int(F, n, m, d);
    return F;
}

static          Sex
FGb_mod(Sex F, Sex p)
{
    long            n,
                    m,
                    d;
    long            q;

    if (is_fixnum(p)) {
	q = mkint(p);
	if (q < 2 || q >= 4294967296)
	    aerror2("fgb_mod", F, p);
    } else {
	aerror2("fgb_mod", F, p);
    }

    F = Amkmodular(F, p);
    F = Ldipoly(F);
    n = list_length(F);
    m = list_length(dipoly_kernel_order);
    d = mkint(car(dipoly_kernel_order));
    F = call_fgb_mod(F, n, m, d, q);
    return nil;
}

/**************************************************************/
/**************************************************************/

setup_type const groebner_setup[] = {

    {"groebner", (lisp_proc *) Agroebner, 1},
    {"groebner_algorithm", (lisp_proc *) Agroebner_algorithm, 1},
    {"gb", (lisp_proc *) Agroebner, 1},
    {"gb42", (lisp_proc *) Agroebner42, 1},
    {"gb44", (lisp_proc *) Agroebner44, 1},
    {"mgb", (lisp_proc *) Aminimalgb, 1},
    {"rgb", (lisp_proc *) Areducedgb, 1},
    {"isgb", (lisp_proc *) Ais_groebner, 2},
    {"katsura", (lisp_proc *) Akatsura, 1},
    {"cyclic", (lisp_proc *) Acyclic, 1},
    {"F4", (lisp_proc *) F4, 1},
    {"F5", (lisp_proc *) F5, 1},
    {"fgb", (lisp_proc *) FGb_int, 1},
    {"fgb_p", (lisp_proc *) FGb_mod, 2},
    {NULL, 0, 0}
};
