/*
 * init.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/init.c,v 2.40 2017/11/27 00:45:12 kako Exp $
 */

static char const rcsid[] =
    "$Id: init.c,v 2.40 2017/11/27 00:45:12 kako Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "entries.h"
#include "code.h"

// #define DEBUG

#define LAST_PRIME	3571

Sex            *stackbase;
Sex            *stack;
Sex            *bindbase;
Sex            *bstack;

static Sex     *stacksegment;
static Sex     *bindsegment;

Sex            *stacklimit;
Sex            *bindlimit;

Cons_Cell      *cellsegment;
Sex            *fwdsegment;
void           *coresegment;

Cons_Cell      *cellpointer;
Cons_Cell      *celllimit;
Cons_Cell      *cellbase;

Sex            *fwdpointer;
Sex            *fwdlimit;
Sex            *fwdbase;

long            exit_reason;
int             gensym_cnt = 0,
    gensym1_cnt = 0,
    print_precision,
    print_bfloat_precision,
    read_precision;
unsigned long   miscflags;
long            current_w_precision;
long            current_b_precision;
long            current_d_precision;
double          current_precision;
long            current_modulus;
Sex             true,
                lambda,
                funarg;
Sex             undefined_var;
Sex             evalhook,
                applyhook;
Sex             catch_tags;
Sex             exit_tag,
                exit_val;
Sex             current_package;
Sex             current_charvec;

Sex             sugar_charvec;
Sex             asuka_charvec;
Sex             current_sugar_package;
Sex             current_asuka_package;


Sex             uniq_cons_table;
Sex             uniq_string_table;
long            uniq_cons_size,
                uniq_cons_count;
Sex             fake_cons_cell;
Sex             gensym_base;
char           *boffo,
               *read_buffer;
Sex             io_workspace,
                io_workspace2;
#ifdef SAVE_WORK_SPACE
Sex             work_space;
#endif
Sex             supervisor;
Sex             standard_input;
Sex             lisp_input,
                standard_output,
                lisp_output;
Sex             standard_errout,
                lisp_errout;
Sex             file_input,
                file_output;
Sex             fsl_input,
                fsl_output;
Sex             prompt_message,
                emsg_star;
Sex             lisp_prompt_message;
Sex             current_big_modulus;
Sex             star_modulus;
Sex             star_bmodulus;
Sex             star_var_order;
Sex             pi_value,
                e_value,
                log2_value,
                log10_value,
                sqrt2_value,
                sqrt5_value;
Sex             prime_list;
long            last_prime = LAST_PRIME;
long            last_prime_squared = LAST_PRIME * LAST_PRIME;
Sex             kernel_order;
Sex             dipoly_kernel_order;
Sex             asuka_mode;
Sex             plot_device;
Sex             plot_devices;

Sex             true_symbol;
Sex             false_symbol;
Sex             time_out_symbol;

#ifdef ASUKA
Sex             asuka_true_symbol;
Sex             asuka_false_symbol;
Sex             asuka_nil_symbol;
Sex             catch_symbol;
Sex             throw_symbol;
Sex             any_symbol;
Sex             switch_symbol;
Sex             case_symbol;
#endif

Sex             append_symbol;
Sex             input_symbol;
Sex             output_symbol;
Sex             aref_symbol;
Sex             bappend_symbol;
Sex             binput_symbol;
Sex             boutput_symbol;
Sex             car_symbol;
Sex             cdr_symbol;
Sex             caar_symbol;
Sex             cadr_symbol;
Sex             cdar_symbol;
Sex             cddr_symbol;
Sex             comma_at_symbol;
Sex             comma_symbol;
Sex             echo_symbol;
Sex             eecho_symbol;
Sex             alist_symbol;
Sex             eval_symbol;
Sex             subr_symbol;
Sex             fsubr_symbol;
Sex             expr_symbol;
Sex             fexpr_symbol;
Sex             macro_symbol;
Sex             cmacro_symbol;
Sex             smacro_symbol;
Sex             csmacro_symbol;

Sex             code_symbol;
Sex             fcode_symbol;
Sex             mcode_symbol;
Sex             scode_symbol;

Sex             nomess_symbol;

Sex             function_symbol;
Sex             progn_symbol;
Sex             prog2_symbol;
Sex             quote_symbol;
Sex             cons_symbol;
Sex             nconc_symbol;
Sex             bye_symbol;

Sex             raise_symbol;
Sex             lower_symbol;
Sex             hankaku_symbol;

Sex             bytecode_regs[16];
long            bytecode_nreg = 16;

Code           *bytecode;
long            bpsorg = 0;
long            bpsend = BPS;
Sex            *atoms;
long            atomorg = 0;
long            atomend = NATOM;

/*
 */


#ifdef ASUKA
Sex             expreope_symbol;
Sex             name_symbol;
Sex             for_symbol;
Sex             foreach_symbol;
Sex             if_symbol;
Sex             setq_symbol;

Sex             arrowope_symbol;
Sex             onbeepre_symbol;

Sex             lambda_symbol;

Sex             itrsyth_symbol;
Sex             noncomu_symbol;
Sex             omit_symbol;
Sex             lesser_equal_symbol;
Sex             greater_equal_symbol;
Sex             not_equal_symbol;
Sex             colon_equal_symbol;
Sex             semicolon_semicolon_symbol;
Sex             twbeepre_symbol;
Sex             thbeepre_symbol;

Sex             list_symbol;
Sex             set_symbol;
Sex             prog_symbol;
Sex             go_symbol;
Sex             return_symbol;
Sex             reversip_symbol;
Sex             defproc_symbol;

Sex             array_symbol;
Sex             mkarray1_symbol;
Sex             mkarray2_symbol;
Sex             mkarray3_symbol;
Sex             getarray1_symbol;
Sex             getarray2_symbol;
Sex             getarray3_symbol;
Sex             setarray1_symbol;
Sex             setarray2_symbol;
Sex             setarray3_symbol;
Sex             matrix_symbol;
Sex             mkmatrix_symbol;
Sex             vector_symbol;
Sex             mkvector_symbol;

Sex             ge_symbol;
Sex             le_symbol;
Sex             gt_symbol;
Sex             lt_symbol;
Sex             add_symbol;
Sex             sub_symbol;
Sex             times_symbol;

Sex             set_diff_symbol;
Sex             not_symbol;
Sex             equal_symbol;
Sex             minus_symbol;

Sex             error_symbol;
Sex             debug_symbol;
Sex             debugn_symbol;
Sex             write_symbol;
Sex             writen_symbol;
Sex             printf_symbol;
Sex             read_symbol;
Sex             import_symbol;
Sex             save_symbol;
Sex             load_symbol;
Sex             clear_symbol;
Sex             algebraic_symbol;

Sex             sin_symbol;
Sex             cos_symbol;
Sex             tan_symbol;
Sex             asin_symbol;
Sex             acos_symbol;
Sex             atan_symbol;
Sex             sinh_symbol;
Sex             cosh_symbol;
Sex             tanh_symbol;
Sex             asinh_symbol;
Sex             acosh_symbol;
Sex             atanh_symbol;
Sex             exp_symbol;
Sex             log_symbol;
Sex             log2_symbol;
Sex             log10_symbol;
Sex             sqrt_symbol;
Sex             cbrt_symbol;

extern Sex      F5_Rules;
extern Sex      star_evaluation_points;
#endif



/*
 */
Sex             proc_frame;
Sex             proc_vars;
Sex             proc_gvars;
Sex             loop_frame;
/*
 */
Sex             module_binding;
Sex            *parse_value_ptr;

/***************************************************************/

void            create_asuka_symbols(setup_type const s[]);
static void     cold_setup(void);

/***************************************************************/
/***************************************************************/


#ifdef _LP64

void
init_heap_segments()
{
    void           *p;
    long            core;
    long            check;

    stacksegment = (Sex *) malloc(stack_size);
    if (stacksegment == NULL)
	fatal_error("no more stack space");
    stackbase = (Sex *) alignit((long) stacksegment);
    bindsegment = (Sex *) malloc(stack_size);
    if (bindsegment == NULL)
	fatal_error("no more bind stack space");
    bindbase = (Sex *) alignit((long) bindsegment);

    bytecode = (Code *) malloc(sizeof(Code) * BPS);
    if (bytecode == NULL)
	fatal_error("no more bytecode space");
    bpsorg = 0;
    bpsend = BPS;

    atoms = (Sex *) malloc(sizeof(Sex) * NATOM);
    if (atoms == NULL)
	fatal_error("no more atoms space");
    atomorg = 0;
    atomend = NATOM;

    core = core_size;
    while ((p = malloc(core)) == NULL && core > 1024000) {
	core = core - (core >> 2);
	core = core & 0xffffffffffff0000L;
    }
    if (p == NULL)
	fatal_error("no more cell");

    coresegment = p;
    fwd_size = core / 2;

    cell_size = core - fwd_size;
    fwd_size = alignit(fwd_size);
    cell_size = alignit(cell_size);
    core_size = fwd_size + cell_size;

    fwdbase = (Sex *) alignit((long) p);
    fwdlimit = (Sex *) ((long) fwdbase + (fwd_size - 16));
    nil = (Sex) ((long) fwdbase + TAG_SYMBOL);
    fwdbase = (Sex *) ((long) fwdbase + symhdr_length);
    fwdpointer = fwdbase;
    p = (void *) ((char *) p + fwd_size);
    cellbase = (Cons_Cell *) ((char *) alignit((long) p + cell_size) - 16);
    celllimit =
	(Cons_Cell *) ((char *) cellbase - (cell_size - 16 - SPARE));
    cellpointer = cellbase;
#ifdef DEBUG
    printf("core segment =0x%08lx -- 0x%08lx\n", (long) coresegment,
	   (long) coresegment + core);
    printf("fwd segment =0x%08lx -- 0x%08lx\n", (long) fwdbase,
	   (long) fwdlimit);
    printf("cell segment =0x%08lx -- 0x%08lx\n", (long) cellbase,
	   (long) celllimit);
    printf("bytecode segment =0x%08lx -- 0x%08lx\n", (long) bytecode,
	   (long) (bytecode + BPS));
    printf("atoms segment =0x%08lx -- 0x%08lx\n", (long) atoms,
	   (long) (atoms + NATOM));
    printf("stack segment =0x%08lx -- 0x%08lx\n", (long) stackbase,
	   (long) stackbase + stack_size);
    printf("bind segment =0x%08lx -- 0x%08lx\n", (long) bindbase,
	   (long) bindbase + stack_size);
    printf("proc_frame =0x%08lx\n", (long) &proc_frame);
#endif
}

#else


#if MEMORY_MODEL == MEMORY_MIXED
int             memory_low;
#endif
#if MEMORY_MODEL == MEMORY_VARIABLE
int             memory_low;
int             memory_model = MEMORY_LOW;
#endif


void
init_heap_segments()
{
    void           *p;
    long            core;
    long            check;
#if MEMORY_MODEL == MEMORY_MIXED || MEMORY_MODEL == MEMORY_VARIABLE
    long            high;
    long            low;
#endif

    stacksegment = (Sex *) malloc(stack_size);
    if (stacksegment == NULL)
	fatal_error("no more stack space");
    stackbase = (Sex *) alignit((long) stacksegment);
    bindsegment = (Sex *) malloc(stack_size);
    if (bindsegment == NULL)
	fatal_error("no more bind stack space");
    bindbase = (Sex *) alignit((long) bindsegment);

    bytecode = (Code *) malloc(sizeof(Code) * BPS);
    if (bytecode == NULL)
	fatal_error("no more bytecode space");
    bpsorg = 0;
    bpsend = BPS;

    atoms = (Sex *) malloc(sizeof(Sex) * NATOM);
    if (atoms == NULL)
	fatal_error("no more atoms space");
    atomorg = 0;
    atomend = NATOM;

    core = core_size;
    while ((p = malloc(core)) == NULL && core > 1024000) {
	core = core - (core >> 2);
	core = core & 0xffff0000L;
    }
    if (p == NULL)
	fatal_error("no more cell");

    /*
     * check allocated memory address 
     */
#if MEMORY_MODEL == MEMORY_LOW
    check =
	(unsigned long) stacksegment | (unsigned long) bindsegment |
	(unsigned long) p;
    check |= (unsigned long) bytecode;
    check |= (unsigned long) atoms;
    check |= (unsigned long) stackbase + stack_size;
    check |= (unsigned long) bindbase + stack_size;;
    check |= (unsigned long) p + core;
    check |= (unsigned long) (bytecode + BPS);
    check |= (unsigned long) (atoms + NATOM);
    if (check < 0) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + (unsigned long) core);
	printf("stacksegment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stacksegment,
	       (unsigned long) stacksegment + (unsigned long) stack_size);
	printf("bindsegment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindsegment,
	       (unsigned long) bindsegment + (unsigned long) stack_size);
	printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
	printf("bytecode =0x%08lx -- 0x%08lx\n", (unsigned long) bytecode,
	       (unsigned long) (bytecode + BPS));
	printf("atoms =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	       (unsigned long) (atoms + NATOM));
	fatal_error("allocated memory exceed 0x80000000");
    }
    if ((int) (&proc_frame) < 0) {
	fatal_error("you must try mixed memory model");
    }
    if ((int) p > (int) p + core || (int) bytecode > (int) (bytecode + BPS)
	|| (int) atoms > (int) (atoms + NATOM)
	|| (int) stackbase > (int) stackbase + stack_size
	|| (int) bindbase > (int) bindbase + stack_size) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + (unsigned long) core);
	printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
	printf("atoms segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) atoms, (unsigned long) (atoms + NATOM));
	printf("stack segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stackbase,
	       (unsigned long) stackbase + (unsigned long) stack_size);
	printf("bind segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindbase,
	       (unsigned long) bindbase + (unsigned long) stack_size);
	fatal_error("sorry I can't handle this case");
    }
#ifdef DEBUG
    printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	   (unsigned long) p + core);
    printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
    printf("atoms segment =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	   (unsigned long) (atoms + NATOM));
    printf("stack segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) stackbase,
	   (unsigned long) stackbase + stack_size);
    printf("bind segment =0x%08lx -- 0x%08lx\n", (unsigned long) bindbase,
	   (unsigned long) bindbase + stack_size);
    printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
    printf("memory model = %d\n", memory_model);
#endif
#else
#if MEMORY_MODEL == MEMORY_HIGH
    check = (int) stacksegment & (int) bindsegment & (int) p;
    check &= (int) bytecode;
    check &= (int) atoms;
    check &= (int) stackbase + stack_size;
    check &= (int) bindbase + stack_size;;
    check &= (int) p + core;
    check &= (int) (bytecode + BPS);
    check &= (int) (atoms + NATOM);
    if (check > 0) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + core);
	printf("stacksegment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stacksegment,
	       (unsigned long) stacksegment + stack_size);
	printf("bindsegment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindsegment,
	       (unsigned long) bindsegment + stack_size);
	printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
	printf("bytecode =0x%08lx -- 0x%08lx\n", (unsigned long) bytecode,
	       (unsigned long) (bytecode + BPS));
	printf("atoms =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	       (unsigned long) (atoms + NATOM));
	fatal_error("allocated memory address is below 0x80000000");
    }
    if ((int) (&proc_frame) > 0) {
	fatal_error("you must try mixed memory model");
    }
    if ((int) p > (int) p + core || (int) bytecode > (int) (bytecode + BPS)
	|| (int) atoms > (int) (atoms + NATOM)
	|| (int) stackbase > (int) stackbase + stack_size
	|| (int) bindbase > (int) bindbase + stack_size) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + core);
	printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
	printf("atoms segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) atoms, (unsigned long) (atoms + NATOM));
	printf("stack segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stackbase,
	       (unsigned long) stackbase + stack_size);
	printf("bind segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindbase,
	       (unsigned long) bindbase + stack_size);
	fatal_error("sorry I can't handle this case");
    }
#ifdef DEBUG
    printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	   (unsigned long) p + core);
    printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
    printf("atoms segment =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	   (unsigned long) (atoms + NATOM));
    printf("stack segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) stackbase,
	   (unsigned long) stackbase + stack_size);
    printf("bind segment =0x%08lx -- 0x%08lx\n", (unsigned long) bindbase,
	   (unsigned long) bindbase + stack_size);
    printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
    printf("memory model = %d\n", memory_model);
#endif
#else
#if MEMORY_MODEL == MEMORY_MIXED

    high = 0;
    low = 0;
    if ((int) stacksegment > 0)
	high = (int) stacksegment;
    else
	low = (int) stacksegment;

    if ((int) bindsegment > 0) {
	if (high < (int) bindsegment)
	    high = (int) bindsegment;
    } else if (low > (int) bindsegment)
	low = (int) bindsegment;
    if ((int) p > 0) {
	if (high < (int) p)
	    high = (int) p;
    } else if (low > (int) p)
	low = (int) p;
    if ((int) bytecode > 0) {
	if (high < (int) bytecode)
	    high = (int) bytecode;
    } else if (low > (int) bytecode)
	low = (int) bytecode;
    if ((int) atoms > 0) {
	if (high < (int) atoms)
	    high = (int) atoms;
    } else if (low > (int) atoms)
	low = (int) atoms;
    if ((int) &proc_frame < 0) {
	if (high < (int) &proc_frame)
	    high = (int) &proc_frame;
    } else if (low > (int) &proc_frame)
	low = (int) &proc_frame;
    low = low & 0x7fffffff;
    memory_low = low;
    if (low < high) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + core);
	printf("stacksegment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stacksegment,
	       (unsigned long) stacksegment + stack_size);
	printf("bindsegment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindsegment,
	       (unsigned long) bindsegment + stack_size);
	printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
	printf("bytecode =0x%08lx -- 0x%08lx\n", (unsigned long) bytecode,
	       (unsigned long) (bytecode + BPS));
	printf("atoms =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	       (unsigned long) (atoms + NATOM));
	fatal_error("sorry I can't handle this case");
    }
    if ((int) p > (int) p + core || (int) bytecode > (int) (bytecode + BPS)
	|| (int) atoms > (int) (atoms + NATOM)
	|| (int) stackbase > (int) stackbase + stack_size
	|| (int) bindbase > (int) bindbase + stack_size) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + core);
	printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
	printf("atoms segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) atoms, (unsigned long) (atoms + NATOM));
	printf("stack segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stackbase,
	       (unsigned long) stackbase + stack_size);
	printf("bind segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindbase,
	       (unsigned long) bindbase + stack_size);
	fatal_error("sorry I can't handle this case");
    }
#ifdef DEBUG
    printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	   (unsigned long) p + core);
    printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
    printf("atoms segment =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	   (unsigned long) (atoms + NATOM));
    printf("stack segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) stackbase,
	   (unsigned long) stackbase + stack_size);
    printf("bind segment =0x%08lx -- 0x%08lx\n", (unsigned long) bindbase,
	   (unsigned long) bindbase + stack_size);
    printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
    printf("memory model = %d\n", memory_model);
#endif
#else
#if MEMORY_MODEL == MEMORY_VARIABLE
    high = 0;
    low = 0;
    if ((int) stacksegment > 0)
	high = (int) stacksegment;
    else
	low = (int) stacksegment;

    if ((int) bindsegment > 0) {
	if (high < (int) bindsegment)
	    high = (int) bindsegment;
    } else if (low > (int) bindsegment)
	low = (int) bindsegment;
    if ((int) p > 0) {
	if (high < (int) p)
	    high = (int) p;
    } else if (low > (int) p)
	low = (int) p;
    if ((int) bytecode > 0) {
	if (high < (int) bytecode)
	    high = (int) bytecode;
    } else if (low > (int) bytecode)
	low = (int) bytecode;
    if ((int) atoms > 0) {
	if (high < (int) atoms)
	    high = (int) atoms;
    } else if (low > (int) atoms)
	low = (int) atoms;
    if ((int) &proc_frame < 0) {
	if (high < (int) &proc_frame)
	    high = (int) &proc_frame;
    } else if (low > (int) &proc_frame)
	low = (int) &proc_frame;

#ifdef DEBUG
    printf("high =0x%08lx, low = 0x%08lx\n", high, low);
#endif
    if (low >= 0)
	memory_model = MEMORY_LOW;
    else if (high < 0)
	memory_model = MEMORY_HIGH;
    else {
	low = low & 0x7fffffff;
	if (low > high) {
	    memory_low = low;
	    memory_model = MEMORY_MIXED;
	} else {
	    printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
		   (unsigned long) p + core);
	    printf("stacksegment =0x%08lx -- 0x%08lx\n",
		   (unsigned long) stacksegment,
		   (unsigned long) stacksegment + stack_size);
	    printf("bindsegment =0x%08lx -- 0x%08lx\n",
		   (unsigned long) bindsegment,
		   (unsigned long) bindsegment + stack_size);
	    printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
	    printf("bytecode =0x%08lx -- 0x%08lx\n",
		   (unsigned long) bytecode,
		   (unsigned long) (bytecode + BPS));
	    printf("atoms =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
		   (unsigned long) (atoms + NATOM));
	    fatal_error("sorry I can't handle this case");
	}
    }
    if ((int) p > (int) p + core || (int) bytecode > (int) (bytecode + BPS)
	|| (int) atoms > (int) (atoms + NATOM)
	|| (int) stackbase > (int) stackbase + stack_size
	|| (int) bindbase > (int) bindbase + stack_size) {
	printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	       (unsigned long) p + core);
	printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
	printf("atoms segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) atoms, (unsigned long) (atoms + NATOM));
	printf("stack segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) stackbase,
	       (unsigned long) stackbase + stack_size);
	printf("bind segment =0x%08lx -- 0x%08lx\n",
	       (unsigned long) bindbase,
	       (unsigned long) bindbase + stack_size);
	fatal_error("sorry I can't handle this case");
    }
#ifdef DEBUG
    printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	   (unsigned long) p + core);
    printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
    printf("atoms segment =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	   (unsigned long) (atoms + NATOM));
    printf("stack segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) stackbase,
	   (unsigned long) stackbase + stack_size);
    printf("bind segment =0x%08lx -- 0x%08lx\n", (unsigned long) bindbase,
	   (unsigned long) bindbase + stack_size);
    printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
    printf("memory model = %d\n", memory_model);
#endif
#endif
#endif
#endif
#endif


    coresegment = p;
    fwd_size = core / 2;

    cell_size = core - fwd_size;
    fwd_size = alignit(fwd_size);
    cell_size = alignit(cell_size);
    core_size = fwd_size + cell_size;

    fwdbase = (Sex *) alignit((long) p);
    fwdlimit = (Sex *) ((long) fwdbase + (fwd_size - 16));
    nil = (Sex) ((long) fwdbase + TAG_SYMBOL);
    fwdbase = (Sex *) ((long) fwdbase + symhdr_length);
    fwdpointer = fwdbase;
    p = (void *) ((char *) p + fwd_size);
    cellbase = (Cons_Cell *) ((char *) alignit((long) p + cell_size) - 16);
    celllimit =
	(Cons_Cell *) ((char *) cellbase - (cell_size - 16 - SPARE));
    cellpointer = cellbase;
#ifdef DEBUG
    printf("core segment =0x%08lx -- 0x%08lx\n", (unsigned long) p,
	   (unsigned long) p + core);
    printf("bytecode segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) bytecode, (unsigned long) (bytecode + BPS));
    printf("atoms segment =0x%08lx -- 0x%08lx\n", (unsigned long) atoms,
	   (unsigned long) (atoms + NATOM));
    printf("stack segment =0x%08lx -- 0x%08lx\n",
	   (unsigned long) stackbase,
	   (unsigned long) stackbase + stack_size);
    printf("bind segment =0x%08lx -- 0x%08lx\n", (unsigned long) bindbase,
	   (unsigned long) bindbase + stack_size);
    printf("proc_frame =0x%08lx\n", (unsigned long) &proc_frame);
    printf("memory model = %d\n", memory_model);
#endif
}

#endif

/***************************************************************/
/***************************************************************/


void
create_symbols(setup_type const s[])
{
    long            i;
    for (i = 0; s[i].name != NULL; i++) {
	make_proc_symbol((char *) (s[i].name), s[i].fn, s[i].nargs);
    }
}

void
create_asuka_symbols(setup_type const s[])
{
    long            i;
    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    for (i = 0; s[i].name != NULL; i++) {
	make_proc_symbol((char *) (s[i].name), s[i].fn, s[i].nargs);
    }
    pop(current_package);
    current_charvec = package_charvec(current_package);
}

static void
cold_setup(void)
{
    long            i;
    Sex             cp,
                    a;

    miscflags = 0;


    plist(nil) = nil;
    symlen(nil) = sizeof(Symbol_Head) + 3 - CELL;
    proc(nil) = (lisp_proc *) undefined_proc;
    qnargs(nil) = 0;
    qtag(nil) = TAG_HDRS + T_SYMBOL + SYM_FLUID;
    value(nil) = nil;
#ifdef ALGEBRAIC_PROPERTY
    gtag(nil) = 0;
    gproc(nil) = undefined_proc;
    gnargs(nil) = 0;
    gvalue(nil) = nil;
#endif
    *(long *) (pname(nil)) = 0;
    memcpy(pname(nil), "nil", 3);

    exit_tag = exit_val = nil;
    exit_reason = 0;

    current_sugar_package = nil;
    current_asuka_package = nil;

    current_package = getvector_init(sizeof(Package), nil);
    a = getvector_init(GOBVEC_SIZE + HDR, mkfxat(0));
    cp = current_package;
    package_vec(cp) = a;
    package_charvec(cp) = nil;
    package_hdr(cp) = T_PACKAGE | TAG_HDRS;
    package_cnt(cp) = mkfxat(1);	/* 
					 * Allow for nil  
					 */
    vector_value(package_vec(cp),
		 hash_string("nil", 3) & (GOBVEC_SIZE / SEX - 1)) = nil;
    uniq_cons_table =
	getvector_init(INIT_HASH_SIZE * sizeof(HashEntry) + HDR, nil);
    uniq_cons_size = INIT_HASH_SIZE;
    uniq_cons_count = 0;
    //uniq_string_table = Lmake_hash_table();
    uniq_string_table = nil;
    //printf("uniq_string table = %016x, %ld\n", uniq_string_table, uniq_string_table);

    /*
     * gensym_cnt = 0;
     */
    print_precision = 13;
    print_bfloat_precision = 13;
    read_precision = 14;
    current_d_precision = 16;
#if 0
    current_w_precision = (current_d_precision * 7 + 63) >> 6;
    current_b_precision = current_w_precision * 32;
    current_precision = pow(0.1, (double) current_d_precision);
#endif
    Lprecision(mkfxat(current_d_precision));

    current_modulus = 1;
    undefined_var = nil;

    undefined_var = make_symbol("~undefined-value~");
    qtag(undefined_var) |= SYM_FLUID;
    value(undefined_var) = undefined_var;
    Lremob(undefined_var);

    true = make_symbol("t");
    qtag(true) |= SYM_FLUID;
    value(true) = true;

    for (i = 0; i < bytecode_nreg; i++)
	bytecode_regs[i] = nil;


    boffo = malloc(BUFFER_SIZE + 8);
    read_buffer = malloc(BUFFER_SIZE + 8);

    if (current_sugar_package == nil) {
	current_sugar_package = make_symbol("*sugar-package*");
	qtag(current_sugar_package) |= SYM_FLUID;
	value(current_sugar_package) = current_package;

	sugar_charvec = getvector_init(258 * SEX, nil);
	current_charvec = sugar_charvec;
	boffo[1] = 0;
	for (i = 0; i < 256; i++) {
	    boffo[0] = i;
	    vector_value(current_charvec, i) = intern(boffo, 1);
	}
	package_charvec(value(current_sugar_package)) = current_charvec;
    }
    if (current_asuka_package == nil) {
	current_asuka_package = make_symbol("*gal-package*");
	qtag(current_asuka_package) |= SYM_FLUID;
	value(current_asuka_package) =
	    getvector_init(sizeof(Package), nil);
	current_package = value(current_asuka_package);
	a = getvector_init(GOBVEC_SIZE + HDR, mkfxat(0));
	cp = value(current_asuka_package);
	package_vec(cp) = a;
	package_hdr(cp) = T_PACKAGE | TAG_HDRS;
	package_cnt(cp) = mkfxat(0);

	current_charvec = getvector_init(258 * SEX, nil);
	boffo[1] = 0;
	for (i = 0; i < 256; i++) {
	    boffo[0] = i;
	    vector_value(current_charvec, i) = intern(boffo, 1);
	}
	package_charvec(value(current_asuka_package)) = current_charvec;
	asuka_charvec = current_charvec;
    }
    current_package = value(current_sugar_package);
    current_charvec = package_charvec(current_package);


    env = nil;
    supervisor = nil;
    catch_tags = nil;
    io_workspace = nil;
    io_workspace2 = nil;
#ifdef SAVE_WORK_SPACE
    work_space = nil;
#endif
    comma_symbol = make_symbol("~comma");
    comma_at_symbol = make_symbol("~comma-at");
    lambda = make_symbol("lambda");
    funarg = make_symbol("funarg");

    package_id(value(current_sugar_package)) = make_symbol("sugar");
    package_id(value(current_asuka_package)) = make_symbol("asuka");


    evalhook = make_fluid_symbol("*evalhook*", nil);
    applyhook = make_fluid_symbol("*applyhook*", nil);
    gensym_base = make_string("G");

    true_symbol = make_fluid_symbol("true", true);
    false_symbol = make_fluid_symbol("false", nil);
#ifdef ASUKA
    time_out_symbol = make_asuka_symbol("@time_out@");
#else
    time_out_symbol = make_symbol("@time_out@");
#endif

#ifdef ASUKA
    asuka_true_symbol = make_asuka_symbol("true");
    asuka_false_symbol = make_asuka_symbol("false");
    asuka_nil_symbol = make_asuka_symbol("nil");
    catch_symbol = make_symbol("catch");
    throw_symbol = make_symbol("throw");
    any_symbol = make_asuka_symbol("any");
    switch_symbol = make_symbol("switch");
    case_symbol = make_symbol("case");
#endif


    asuka_mode = make_symbol("asuka_mode");
    input_symbol = make_symbol("input");
    output_symbol = make_symbol("output");
    append_symbol = make_symbol("append");
    binput_symbol = make_symbol("binput");
    boutput_symbol = make_symbol("boutput");
    bappend_symbol = make_symbol("bappend");
    raise_symbol = make_fluid_symbol("*raise", nil);
    lower_symbol = make_fluid_symbol("*lower", nil);
    nomess_symbol = make_fluid_symbol(".nomessage", nil);
    echo_symbol = make_fluid_symbol("*echo", nil);
    eecho_symbol = make_asuka_fluid_symbol("*echo", nil);
    alist_symbol = make_fluid_symbol("alist", nil);
#ifdef Kanji
    hankaku_symbol = make_fluid_symbol("*hankaku", nil);
#endif
    plot_device = make_fluid_symbol(".Device", nil);
    plot_devices = make_fluid_symbol(".Devices", nil);
    emsg_star = make_fluid_symbol("emsg*", nil);
    current_big_modulus = mkfxat(1);
    star_modulus = make_fluid_symbol("*modulus", nil);
    star_bmodulus = make_fluid_symbol("*bmodulus", nil);
    star_var_order = make_fluid_symbol("*var_order", nil);
    expr_symbol = make_symbol("expr");
    fexpr_symbol = make_symbol("fexpr");
    subr_symbol = make_symbol("subr");
    fsubr_symbol = make_symbol("fsubr");
    macro_symbol = make_symbol("macro");
    smacro_symbol = make_symbol("smacro");
    cmacro_symbol = make_symbol("cmacro");
    csmacro_symbol = make_symbol("csmacro");

    code_symbol = make_symbol("code");
    fcode_symbol = make_symbol("fcode");
    mcode_symbol = make_symbol("mcode");
    scode_symbol = make_symbol("scode");


    bye_symbol = make_symbol("bye");
    car_symbol = make_symbol("car");
    cdr_symbol = make_symbol("cdr");
    caar_symbol = make_symbol("caar");
    cadr_symbol = make_symbol("cadr");
    cdar_symbol = make_symbol("cdar");
    cddr_symbol = make_symbol("cddr");
    aref_symbol = make_symbol("aref");

    pi_value = nil;
    e_value = nil;
    log2_value = nil;
    log10_value = nil;
    sqrt2_value = nil;
    sqrt5_value = nil;
    prime_list = Lgen_prime(mkfxat(2), mkfxat(LAST_PRIME + 1));
    kernel_order = nil;
    dipoly_kernel_order = nil;
    /*
     * init_reserved_word_symbols();
     */

    fake_cons_cell = cons(car_symbol, cdr_symbol);
    prompt_message = nil;
    lisp_prompt_message = make_string("> ");



#ifdef ASUKA
    expreope_symbol = make_symbol(",");
    name_symbol = make_symbol(".");
    for_symbol = make_symbol("for");
    foreach_symbol = make_symbol("foreach");
    if_symbol = make_symbol("if");
    setq_symbol = make_symbol("setq");

    itrsyth_symbol = make_symbol("@@");
    noncomu_symbol = make_symbol("&*");
    omit_symbol = make_symbol("..");
    lesser_equal_symbol = make_symbol("<=");
    greater_equal_symbol = make_symbol(">=");
    not_equal_symbol = make_symbol("<>");
    colon_equal_symbol = make_asuka_symbol(":=");
    semicolon_semicolon_symbol = make_symbol(";;");
    arrowope_symbol = make_symbol("->");
    onbeepre_symbol = make_symbol("\"");

    twbeepre_symbol = make_symbol("\"\"");
    thbeepre_symbol = make_symbol("\"\"\"");
    lambda_symbol = make_symbol("lambda");
    equal_symbol = make_symbol("=");
    minus_symbol = make_symbol("minus");
    list_symbol = make_symbol("list");
    set_symbol = make_asuka_symbol("setm");
    prog_symbol = make_symbol("prog");
    go_symbol = make_symbol("go");
    return_symbol = make_symbol("return");
    reversip_symbol = make_symbol("reversip");
    defproc_symbol = make_symbol("de");
    not_symbol = make_symbol("not");
    set_diff_symbol = make_symbol("set_diff");
    ge_symbol = make_symbol(">=");
    le_symbol = make_symbol("<=");
    gt_symbol = make_symbol(">");
    lt_symbol = make_symbol("<");
    add_symbol = make_symbol("+");
    sub_symbol = make_symbol("-");
    times_symbol = make_symbol("*");

    array_symbol = make_symbol("array");
    mkarray1_symbol = make_symbol("mkarray1");
    mkarray2_symbol = make_symbol("mkarray2");
    mkarray3_symbol = make_symbol("mkarray3");
    getarray1_symbol = make_symbol("getarray1");
    getarray2_symbol = make_symbol("getarray2");
    getarray3_symbol = make_symbol("getarray3");
    setarray1_symbol = make_symbol("setarray1");
    setarray2_symbol = make_symbol("setarray2");
    setarray3_symbol = make_symbol("setarray3");
    matrix_symbol = make_symbol("matrix");
    mkmatrix_symbol = make_symbol("mkmatrix");
    vector_symbol = make_symbol("vector");
    mkvector_symbol = make_symbol("mkvector");

    error_symbol = make_asuka_symbol("error");
    debug_symbol = make_asuka_symbol("debug");
    debugn_symbol = make_asuka_symbol("debugn");
    write_symbol = make_asuka_symbol("write");
    writen_symbol = make_asuka_symbol("writen");
    printf_symbol = make_asuka_symbol("printf");
    read_symbol = make_asuka_symbol("read");
    import_symbol = make_asuka_symbol("import");
    save_symbol = make_asuka_symbol("save");
    load_symbol = make_asuka_symbol("load");
    clear_symbol = make_asuka_symbol("clear");
    algebraic_symbol = make_asuka_symbol("algebraic");
    sin_symbol = make_asuka_symbol("sin");
    cos_symbol = make_asuka_symbol("cos");
    tan_symbol = make_asuka_symbol("tan");
    asin_symbol = make_asuka_symbol("asin");
    acos_symbol = make_asuka_symbol("acos");
    atan_symbol = make_asuka_symbol("atan");
    sinh_symbol = make_asuka_symbol("sinh");
    cosh_symbol = make_asuka_symbol("cosh");
    tanh_symbol = make_asuka_symbol("tanh");
    asinh_symbol = make_asuka_symbol("asinh");
    acosh_symbol = make_asuka_symbol("acosh");
    atanh_symbol = make_asuka_symbol("atanh");
    exp_symbol = make_asuka_symbol("exp");
    log_symbol = make_asuka_symbol("log");
    log2_symbol = make_asuka_symbol("log2");
    log10_symbol = make_asuka_symbol("log10");
    sqrt_symbol = make_asuka_symbol("sqrt");
    cbrt_symbol = make_asuka_symbol("cbrt");
    {
	Sex             tmp;
	tmp = make_asuka_symbol("nil");
	value(tmp) = nil;
	tmp = make_asuka_symbol("true");
	value(tmp) = true;
	tmp = make_asuka_symbol("false");
	value(tmp) = nil;
    }

#endif

#ifdef ASUKA
    F5_Rules = nil;
    star_evaluation_points = nil;
#endif

    a = make_symbol("lispsystem*");
    value(a) = nil;

    proc_frame = nil;
    proc_vars = nil;
    proc_gvars = nil;
    loop_frame = nil;
}

void
setup()
{
    /*
     * int i; 
     */
    Sex             w,
                    n;

    stack = stackbase;
    bstack = bindbase;

    stacklimit = &stack[stack_size / SEX - 100];	/* 
							 * allow some slop at end 
							 */
    bindlimit = &bstack[bind_size / SEX - 100];	/* 
						 * allow some slop at end 
						 */
    cold_setup();


    function_symbol =
	make_proc_symbol("function", (lisp_proc *) function_fn, -1);
    quote_symbol = make_proc_symbol("quote", (lisp_proc *) quote_fn, -1);
    progn_symbol = make_proc_symbol("progn", (lisp_proc *) progn_fn, -1);
    prog2_symbol = make_proc_symbol("prog2", (lisp_proc *) prog2_fn, -1);
    cons_symbol = make_proc_symbol("cons", (lisp_proc *) cons, 0);
    nconc_symbol = make_proc_symbol("nconc", (lisp_proc *) cons, 0);
    eval_symbol = make_proc_symbol("eval", (lisp_proc *) Leval, 1);
    {
	push2(nil, nil);
#define w stack[-1]
#define s stack[0]
	s = make_symbol(OPSYS);
	n = make_symbol("opsys");
	w = Lacons(n, s, nil);
#ifdef VisualC
	if (win32s) {
	    n = make_symbol("win32s");
	    w = cons(n, w);
	}
#endif
	s = make_string(IMPNAME);
	n = make_symbol("name");
	w = Lacons(n, s, w);
	s = make_string(VERSION);
	n = make_symbol("version");
	w = Lacons(n, s, w);
	n = make_symbol("sugar");
	w = cons(n, w);
	n = make_symbol("lispsystem*");
	qtag(n) |= SYM_FLUID;
	value(n) = w;
#if defined(REDUCE) && defined(Android)
	w = make_symbol("algebraic");
	n = make_symbol("*mode");
	qtag(n) |= SYM_FLUID | SYM_GLOBAL;
	value(n) = w;
#endif
#undef w
#undef s
	popv(2);
    }

    create_symbols(arith_setup);
#ifdef ASUKA
    create_asuka_symbols(trig_setup);
#else
    create_symbols(trig_setup);
#endif
    create_symbols(modular_setup);
    create_symbols(char_setup);
    create_symbols(eval_setup);
    create_symbols(fns_setup);
    create_symbols(print_setup);
    create_symbols(read_setup);
    create_symbols(hash_setup);
#ifndef GAL
    create_symbols(fap_setup);
#endif

#ifdef ASUKA
    create_asuka_symbols(hash_setup);
    create_asuka_symbols(algnum_setup);
    create_asuka_symbols(polynomial_setup);


    create_asuka_symbols(mfunc_setup);
    create_symbols(gfunc_setup);
    create_asuka_symbols(triangular_setup);
    create_asuka_symbols(list_setup);
    create_asuka_symbols(misc_setup);
    create_asuka_symbols(groebner_setup);
    create_asuka_symbols(factor_setup);
    create_asuka_symbols(gcd_setup);
    create_asuka_symbols(matrix_setup);
#endif
#ifdef GRAPHICS
    create_symbols(plot_setup);
#endif

    w = Lspecial_char(mkfxat(1));	/* 
					 * newline 
					 */
    make_fluid_symbol("$eol$", w);
    w = Lspecial_char(mkfxat(3));	/* 
					 * tab 
					 */
    make_fluid_symbol("$tab$", w);
#ifdef ASUKA
    w = Lspecial_char(mkfxat(8));	/* 
					 * end of file 
					 */
    /*
     * make_fluid_symbol("$eof$", w); make_asuka_fluid_symbol("$eof$", w);
     * make_asuka_fluid_symbol("char_eof", w); 
     */
    make_fluid_symbol("$eof$", w);
    w = vector_value(asuka_charvec, 0x1a);
    make_asuka_fluid_symbol("$asuka_eof$", w);
#else
    w = Lspecial_char(mkfxat(8));	/* 
					 * end of file 
					 */
    make_fluid_symbol("$eof$", w);
#endif

#ifdef Android
    standard_input = lisp_input = make_file_handle(androidin, NOT_CHAR, 0);
    file_line(lisp_input) = 0;
    standard_output = lisp_output =
	make_file_handle(androidout, 0, char_to_androidout);
    standard_errout = lisp_errout =
	make_file_handle(androiderr, 0, char_to_androiderr);
#else
    standard_input = lisp_input = make_file_handle(stdin, NOT_CHAR, 0);
    file_line(lisp_input) = 0;
    standard_output = lisp_output =
	make_file_handle(stdout, 0, char_to_stdout);
    standard_errout = lisp_errout =
	make_file_handle(stderr, 0, char_to_stderr);
#endif
    file_input = file_output = nil;
    fsl_input = fsl_output = nil;
}
