
/*
 * interpreter.c      (C) Copyright 2010 by Fujio Kako, Nara Women's Univers
 ity.
 * bytecode interpreter 
 *
 *      $Header: /Users/kako/asuka/src/RCS/interpreter.c,v 2.6 2018/03/02 08:39:42 kako Exp $
 */

static char const rcsid[] =
    "$Id: interpreter.c,v 2.6 2018/03/02 08:39:42 kako Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"

/*
 */


#define T_INTERPRETER

#include "code.h"



/*
 * #define DEBUG
 */

#ifdef DEBUG
#define CAR_CHECK(a, r, s) \
	if(!is_cons(a)) { \
		printf("r = %d, s=%d, S=0x%08x\n", r, s, S); \
		lisp_printf("R = %p\n", bytecode_regs[r]); \
		lisp_printf("S = %p\n", bytecode_regs[s]); \
		carcdr_error(a); \
	}
#else
#define CAR_CHECK(a, r, s)
#endif

#define NILR	bytecode_regs[0]
#define A	bytecode_regs[1]
#define Q	bytecode_regs[2]
#define M	bytecode_regs[3]

#define caar(p) car(car(p))
#define cddr(p) cdr(cdr(p))
#define cdar(p) cdr(car(p))


#define next_code *ppc++

#define BYTECODE_STACK_SIZE 1000

Code           *bytecode_stack[BYTECODE_STACK_SIZE];
Code          **bytecode_sp = bytecode_stack;



Sex
carcdr_error(Sex a)
{
    return error1("taking car of atom", a);
}

Sex
Linterp0(Sex fn)
{
    stackcheck();
    return interpreter(fn, 0, nil);
}

Sex
Linterp1(Sex fn, Sex a1)
{
    A = a1;
    return interpreter(fn, 1, nil);
}

Sex
Linterp2(Sex fn, Sex a1, Sex a2)
{
    A = a1;
    Q = a2;
    return interpreter(fn, 2, nil);
}

Sex
Linterp3(Sex fn, Sex a1, Sex a2, Sex a3)
{
    A = a1;
    Q = a2;
    M = a3;
    return interpreter(fn, 3, nil);
}


Sex
Linterpreter(Sex fn, Sex args)
{
    int             nargs;
    int             i;
    Sex             def;

    nargs = 0;
    stackcheck();
    while (is_cons(args)) {
	Sex             w;
	push3(fn, args, env);
	w = car(args);
	w = eval(w, env);
	pop3(env, args, fn);
	if (nargs < MAX_NARGS) {
	    push(w);		/* 
				 * args build up on the Lisp stack
				 */
	    nargs++;
	}
	args = cdr(args);
    }
    if (nargs > MAX_NARGS) {
	error1("interpreter: too many arguments", fn);
    }
    for (i = 1; i <= nargs; i++)
	bytecode_regs[i] = stack[i - nargs];

    def = interpreter(fn, nargs, nil);
    popv(nargs);
    return def;
}


Sex
interpreter(long pc, long nargs, Sex env)
{
    Code           *ppc;
    Code          **sp = bytecode_sp;
    int             code;

    int             i;
    union {
	int             n;
	short           s[2];
    } n;

    int             r,
                    s,
                    t;
    int             c;

    Sex             R,
                    S;

    unsigned char   d;


    ppc = bytecode + pc;

    for (;;) {
	// printf("pc = %6d, code = 0x%04x\n", 2*(ppc - bytecode), *ppc);

	code = *ppc++;

	d = code >> 8;
	// printf("pc = 0x%08x, %6d, code = 0x%04x\n", ppc,
	// ppc-bytecode-1, code);

	switch (d) {
	case 0x00:		/* nop */
	    continue;
	case 0x01:		/* entry */
	    continue;
	case 0x02:		/* exit */
	    if (bytecode_sp > sp) {
		ppc = *--bytecode_sp;
		continue;
	    }
	    return bytecode_regs[1];
	case 0x03:		/* push */
	    *bytecode_sp++ = ppc;
	    if (bytecode_sp >= &bytecode_stack[BYTECODE_STACK_SIZE])
		error("bytecode stack space full");
	    continue;
	case 0x04:		/* pop */
	    if (bytecode_sp <= sp)
		error("bytecode stack over popup");
	    ppc = *--bytecode_sp;
	    continue;
	case 0x10:		/* alloc */
	    r = code & 0xff;
	    stack += r;
	    stackcheck();
	    for (i = 0; i < r; i++)
		stack[-i] = nil;
	    break;
	case 0x11:		/* dealloc */
	    r = code & 0xff;
	    stack -= r;
	    break;
	case 0x12:		/* alloc */
	    r = *ppc++;
	    stack += r;
	    stackcheck();
	    for (i = 0; i < r; i++)
		stack[-i] = nil;
	    break;
	case 0x13:		/* dealloc */
	    r = *ppc++;
	    stack -= r;
	    break;

	case 0x17:		/* callsr adr */
	    *bytecode_sp++ = ppc + 1;
	    if (bytecode_sp >= &bytecode_stack[BYTECODE_STACK_SIZE])
		error("bytecode stack space full");

	    r = *(signed short *) ppc;
	    ppc = ppc + r - 1;
	    break;
	case 0x19:		/* jmsr */
	    r = *(signed short *) ppc;
	    ppc = ppc + r - 1;
	    break;

	case 0x1b:		/* callr adr */
	    *bytecode_sp++ = ppc + 2;
	    if (bytecode_sp >= &bytecode_stack[BYTECODE_STACK_SIZE])
		error("bytecode stack space full");

	    /*
	     * n = *(int *)ppc;
	     */
	    n.s[0] = *ppc++;
	    n.s[1] = *ppc++;

	    ppc = ppc + n.n - 3;
	    break;
	case 0x1a:		/* call */
	    *bytecode_sp++ = ppc + 2;
	    if (bytecode_sp >= &bytecode_stack[BYTECODE_STACK_SIZE])
		error("bytecode stack space full");
	    /*
	     * n = *(int *)ppc;
	     */
	    n.s[0] = *ppc++;
	    n.s[1] = *ppc++;
	    ppc = bytecode + n.n;
	    break;
	case 0x1c:		/* jmp */
	    /*
	     * n = *(int *)ppc;
	     */
	    n.s[0] = *ppc++;
	    n.s[1] = *ppc++;
	    ppc = bytecode + n.n;
	    break;
	case 0x1d:		/* jmpr */
	    /*
	     * n = *(int *)ppc;
	     */
	    n.s[0] = *ppc++;
	    n.s[1] = *ppc++;
	    ppc = ppc + n.n - 3;
	    break;

	case 0x20:
	    c = (code & 0xf0) >> 4;
	    r = code & 0xf;
	    R = bytecode_regs[r];
	    r = *(signed short *) ppc - 2;
	    ppc += 1;
	  jmpc:
	    switch (c) {
	    case 0:
		if (R == nil)
		    ppc = ppc + r;
		break;
	    case 1:
		if (!is_cons(R))
		    ppc = ppc + r;
		break;
	    case 2:
		if (is_cons(R))
		    ppc = ppc + r;
		break;
	    case 3:
		if (is_symbol(R))
		    ppc = ppc + r;
		break;
	    case 4:
		if (is_string(R))
		    ppc = ppc + r;
		break;
	    case 5:
		if (is_simple_vector(R))
		    ppc = ppc + r;
		break;
	    case 6:
		if (is_logical(R))
		    ppc = ppc + r;
		break;
	    case 7:
		if (is_float(R))
		    ppc = ppc + r;
		break;
	    case 8:
		if (is_modular(R))
		    ppc = ppc + r;
		break;
	    case 9:
		if (is_integer(R))
		    ppc = ppc + r;
		break;
	    case 10:
		if (Lfixp(R) != nil)
		    ppc = ppc + r;
		break;
	    case 11:
		if (is_rational(R))
		    ppc = ppc + r;
		break;
	    case 12:
		if (is_complex(R))
		    ppc = ppc + r;
		break;
	    }
	    break;

	case 0x21:
	    c = (code & 0xf0) >> 4;
	    r = code & 0xf;
	    R = bytecode_regs[r];
	    r = *(signed short *) ppc - 1 - 1;
	    ppc += 1;
	  jmpnc:
	    switch (c) {
	    case 0:
		if (R != nil)
		    ppc = ppc + r;
		break;
	    case 1:
		if (is_cons(R))
		    ppc = ppc + r;
		break;
	    case 2:
		if (!is_cons(R))
		    ppc = ppc + r;
		break;
	    case 3:
		if (!is_symbol(R))
		    ppc = ppc + r;
		break;
	    case 4:
		if (!is_string(R))
		    ppc = ppc + r;
		break;
	    case 5:
		if (!is_simple_vector(R))
		    ppc = ppc + r;
		break;
	    case 6:
		if (!is_logical(R))
		    ppc = ppc + r;
		break;
	    case 7:
		if (!is_float(R))
		    ppc = ppc + r;
		break;
	    case 8:
		if (!is_modular(R))
		    ppc = ppc + r;
		break;
	    case 9:
		if (!is_integer(R))
		    ppc = ppc + r;
		break;
	    case 10:
		if (Lfixp(R) == nil)
		    ppc = ppc + r;
		break;
	    case 11:
		if (!is_rational(R))
		    ppc = ppc + r;
		break;
	    case 12:
		if (!is_complex(R))
		    ppc = ppc + r;
		break;
	    }
	    break;

	case 0x22:		/* jmpn */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (is_numhdr(R) && (numtype_of(R) == c)) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x23:		/* jmpn */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (!is_numhdr(R) || (numtype_of(R) != c)) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;

	case 0x24:		/* jmpg */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (is_vechdr(R) && (gnumtype_of(R) == c)) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x25:		/* jmpg */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (!is_vechdr(R) || (gnumtype_of(R) != c)) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;


	case 0x26:		/* jmpeq */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (bytecode_regs[r] == bytecode_regs[s]) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x27:		/* jmpne */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (bytecode_regs[r] != bytecode_regs[s]) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x28:		/* jmpequal */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (equal(bytecode_regs[r], bytecode_regs[s])) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x29:		/* jmpnequal */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (!equal(bytecode_regs[r], bytecode_regs[s])) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x2a:		/* jmpgt */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Lgreaterp(bytecode_regs[r], bytecode_regs[s]) != nil) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x2b:		/* jmpge */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Lgeq(bytecode_regs[r], bytecode_regs[s]) != nil) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x2c:		/* jmplt */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Llessp(bytecode_regs[r], bytecode_regs[s]) != nil) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;
	case 0x2d:		/* jmple */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Lleq(bytecode_regs[r], bytecode_regs[s]) != nil) {
		r = *(signed short *) ppc;
		ppc = ppc + r - 1;
	    } else
		ppc += 1;
	    break;

	case 0x30:
	    c = (code & 0xf0) >> 4;
	    r = code & 0xf;
	    R = bytecode_regs[r];
	    /*
	     * n = *(int *)ppc - 1 - 2;
	     */
	    n.s[0] = *ppc++;
	    n.s[1] = *ppc++;
	    n.n -= 3;
	    goto jmpc;
	case 0x31:
	    c = (code & 0xf0) >> 4;
	    r = code & 0xf;
	    R = bytecode_regs[r];
	    /*
	     * n = *(int *)ppc - 1 - 2;
	     */
	    n.s[0] = *ppc++;
	    n.s[1] = *ppc++;
	    n.n -= 3;
	    goto jmpnc;

	case 0x32:		/* jmpn */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (is_numhdr(R) && (numtype_of(R) == c)) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x33:		/* jmpn */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (!is_numhdr(R) || (numtype_of(R) != c)) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;

	case 0x34:		/* jmpg */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (is_vechdr(R) && (gnumtype_of(R) == c)) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x35:		/* jmpg */
	    c = (code & 0xf0) >> 4;
	    r = (code & 0xf);
	    R = bytecode_regs[r];
	    if (!is_vechdr(R) || (gnumtype_of(R) != c)) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;


	case 0x36:		/* jmpeq */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (bytecode_regs[r] == bytecode_regs[s]) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x37:		/* jmpne */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (bytecode_regs[r] != bytecode_regs[s]) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x38:		/* jmpequal */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (equal(bytecode_regs[r], bytecode_regs[s])) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x39:		/* jmpnequal */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (!equal(bytecode_regs[r], bytecode_regs[s])) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x3a:		/* jmpgt */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Lgreaterp(bytecode_regs[r], bytecode_regs[s]) != nil) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x3b:		/* jmpge */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Lgeq(bytecode_regs[r], bytecode_regs[s]) != nil) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x3c:		/* jmplt */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Llessp(bytecode_regs[r], bytecode_regs[s]) != nil) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;
	case 0x3d:		/* jmple */
	    r = (code & 0xf0) >> 4;
	    s = (code & 0xf);
	    if (Lleq(bytecode_regs[r], bytecode_regs[s]) != nil) {
		/*
		 * n = *(int *)ppc;
		 */
		n.s[0] = *ppc++;
		n.s[1] = *ppc++;
		ppc = ppc + n.n - 3;
	    } else
		ppc += 2;
	    break;

	case 0x40:
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    if (s != 0)
		bytecode_regs[s] = bytecode_regs[r];
	    break;


	case 0x41:
	    s = code & 0xf;
	    bytecode_regs[1] = bytecode_regs[s];
	    break;
	case 0x42:
	    s = code & 0xf;
	    bytecode_regs[2] = bytecode_regs[s];
	    break;
	case 0x43:
	    s = code & 0xf;
	    bytecode_regs[3] = bytecode_regs[s];
	    break;
	case 0x44:
	    s = code & 0xf;
	    bytecode_regs[4] = bytecode_regs[s];
	    break;
	case 0x45:
	    s = code & 0xf;
	    bytecode_regs[5] = bytecode_regs[s];
	    break;
	case 0x46:
	    s = code & 0xf;
	    bytecode_regs[6] = bytecode_regs[s];
	    break;
	case 0x47:
	    s = code & 0xf;
	    bytecode_regs[7] = bytecode_regs[s];
	    break;
	case 0x48:
	    s = code & 0xf;
	    bytecode_regs[8] = bytecode_regs[s];
	    break;
	case 0x49:
	    s = code & 0xf;
	    bytecode_regs[9] = bytecode_regs[s];
	    break;
	case 0x4a:
	    s = code & 0xf;
	    bytecode_regs[10] = bytecode_regs[s];
	    break;
	case 0x4b:
	    s = code & 0xf;
	    bytecode_regs[11] = bytecode_regs[s];
	    break;
	case 0x4c:
	    s = code & 0xf;
	    bytecode_regs[12] = bytecode_regs[s];
	    break;
	case 0x4d:
	    s = code & 0xf;
	    bytecode_regs[13] = bytecode_regs[s];
	    break;
	case 0x4e:
	    s = code & 0xf;
	    bytecode_regs[14] = bytecode_regs[s];
	    break;
	case 0x4f:
	    s = code & 0xf;
	    bytecode_regs[15] = bytecode_regs[s];
	    break;
	case 0x51:
	    s = code & 0xff;
	    bytecode_regs[1] = stack[-s];
	    break;
	case 0x52:
	    s = code & 0xff;
	    bytecode_regs[2] = stack[-s];
	    break;
	case 0x53:
	    s = code & 0xff;
	    bytecode_regs[3] = stack[-s];
	    break;
	case 0x54:
	    s = code & 0xff;
	    bytecode_regs[4] = stack[-s];
	    break;
	case 0x55:
	    s = code & 0xff;
	    bytecode_regs[5] = stack[-s];
	    break;
	case 0x56:
	    s = code & 0xff;
	    bytecode_regs[6] = stack[-s];
	    break;
	case 0x57:
	    s = code & 0xff;
	    bytecode_regs[7] = stack[-s];
	    break;
	case 0x58:
	    s = code & 0xff;
	    bytecode_regs[8] = stack[-s];
	    break;
	case 0x59:
	    s = code & 0xff;
	    bytecode_regs[9] = stack[-s];
	    break;
	case 0x5a:
	    s = code & 0xff;
	    bytecode_regs[10] = stack[-s];
	    break;
	case 0x5b:
	    s = code & 0xff;
	    bytecode_regs[11] = stack[-s];
	    break;
	case 0x5c:
	    s = code & 0xff;
	    bytecode_regs[12] = stack[-s];
	    break;
	case 0x5d:
	    s = code & 0xff;
	    bytecode_regs[13] = stack[-s];
	    break;
	case 0x5e:
	    s = code & 0xff;
	    bytecode_regs[14] = stack[-s];
	    break;
	case 0x5f:
	    s = code & 0xff;
	    bytecode_regs[15] = stack[-s];
	    break;
	case 0x60:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[0];
	    break;
	case 0x61:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[1];
	    break;
	case 0x62:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[2];
	    break;
	case 0x63:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[3];
	    break;
	case 0x64:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[4];
	    break;
	case 0x65:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[5];
	    break;
	case 0x66:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[6];
	    break;
	case 0x67:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[7];
	    break;
	case 0x68:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[8];
	    break;
	case 0x69:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[9];
	    break;
	case 0x6a:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[10];
	    break;
	case 0x6b:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[11];
	    break;
	case 0x6c:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[12];
	    break;
	case 0x6d:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[13];
	    break;
	case 0x6e:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[14];
	    break;
	case 0x6f:
	    s = code & 0xff;
	    stack[-s] = bytecode_regs[15];
	    break;
	case 0x70:		/* ldcar */
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    // printf("ldcar %d %d\n", r, s);
	    S = bytecode_regs[s];
	    CAR_CHECK(S, r, s)
		bytecode_regs[r] = car(S);
	    break;
	case 0x71:		/* ldcdr */
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    S = bytecode_regs[s];
	    CAR_CHECK(S, r, s)
		bytecode_regs[r] = cdr(S);
	    break;
	case 0x72:		/* stcar */
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    S = bytecode_regs[s];
	    CAR_CHECK(S, r, s)
		car(S) = bytecode_regs[r];
	    break;
	case 0x73:		/* stcdr */
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    S = bytecode_regs[s];
	    CAR_CHECK(S, r, s)
		cdr(S) = bytecode_regs[r];
	    break;

	case 0x80:		/* lda */
	    r = (code & 0xf);
	    S = atoms[*ppc++];
	    bytecode_regs[r] = S;
	    break;
	case 0x81:		/* ldv */
	    r = (code & 0xf);
	    S = atoms[*ppc++];
	    bytecode_regs[r] = value(S);
	    break;
	case 0x82:		/* stv */
	    r = (code & 0xf);
	    S = atoms[*ppc++];
	    value(S) = bytecode_regs[r];
	    break;

	case 0x90:		/* link */
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    S = atoms[*ppc++];
	    S = Llink(S, s);
	    if (r != 0)
		bytecode_regs[r] = S;
	    break;

	case 0x91:		/* link */
	    s = code & 0xf;
	    S = atoms[*ppc++];
	    Llink(S, s);
	    break;

	case 0x92:		/* link */
	    s = code & 0xf;
	    S = atoms[*ppc++];
	    A = Llink(S, s);
	    break;

	case 0x98:		/* linke */
	    r = (code & 0xf0) >> 4;
	    s = code & 0xf;
	    S = atoms[*ppc];
	    S = Llink(S, s);
	    if (r != 0)
		bytecode_regs[r] = S;
	    if (bytecode_sp <= sp)
		return S;
	    ppc = *--bytecode_sp;
	    break;
	case 0x99:		/* linke */
	    s = code & 0xf;
	    S = atoms[*ppc];
	    S = Llink(S, s);
	    if (bytecode_sp <= sp)
		return S;
	    ppc = *--bytecode_sp;
	    break;
	case 0x9a:		/* linke */
	    s = code & 0xf;
	    S = atoms[*ppc];
	    S = Llink(S, s);
	    A = S;
	    if (bytecode_sp <= sp)
		return S;
	    ppc = *--bytecode_sp;
	    break;

	case 0xa0:		/* progbind */
	    r = (code & 0xff);
	    bstackcheck();
	    for (i = 0; i < r; i++) {
		S = atoms[*ppc++];
		bstack[i * 2 + 2] = S;
		bstack[i * 2 + 1] = value(S);
		// printf("progbind: 0x%08x\n", n);
		value(S) = nil;
	    }
	    bstack = bstack + i * 2;

	    break;
	case 0xa1:		/* progbind 1 */
	    bstackcheck();
	    S = atoms[*ppc++];
	    bstack[0 * 2 + 2] = S;
	    bstack[0 * 2 + 1] = value(S);
	    value(S) = nil;
	    bstack = bstack + 1 * 2;

	    break;
	case 0xa2:		/* progbind 2 */
	    bstackcheck();
	    S = atoms[*ppc++];
	    bstack[0 * 2 + 2] = S;
	    bstack[0 * 2 + 1] = value(S);
	    value(S) = nil;
	    S = atoms[*ppc++];
	    bstack[1 * 2 + 2] = S;
	    bstack[1 * 2 + 1] = value(S);
	    value(S) = nil;
	    bstack = bstack + 2 * 2;

	    break;
	case 0xa3:		/* progbind 3 */
	    bstackcheck();
	    S = atoms[*ppc++];
	    bstack[0 * 2 + 2] = S;
	    bstack[0 * 2 + 1] = value(S);
	    value(S) = nil;
	    S = atoms[*ppc++];
	    bstack[1 * 2 + 2] = S;
	    bstack[1 * 2 + 1] = value(S);
	    value(S) = nil;
	    S = atoms[*ppc++];
	    bstack[2 * 2 + 2] = S;
	    bstack[2 * 2 + 1] = value(S);
	    value(S) = nil;
	    bstack = bstack + 3 * 2;

	    break;
	case 0xa4:		/* lamdbind */
	    r = (code & 0xff);
	    bstackcheck();
	    for (i = 0; i < r; i++) {
		s = *ppc++;
		S = atoms[*ppc++];

		bstack[i * 2 + 2] = S;
		// printf("lamdbind: 0x%08x\n", n);
		bstack[i * 2 + 1] = value(S);
		value(S) = bytecode_regs[s & 0xf];
	    }
	    bstack = bstack + i * 2;
	    break;
	case 0xa5:		/* lamdbind 1 */
	    bstackcheck();
	    s = *ppc++;
	    S = atoms[*ppc++];
	    bstack[0 * 2 + 2] = S;
	    bstack[0 * 2 + 1] = value(S);
	    value(S) = bytecode_regs[s & 0xf];
	    bstack = bstack + 1 * 2;
	    break;
	case 0xa6:		/* lamdbind 2 */
	    bstackcheck();
	    s = *ppc++;
	    S = atoms[*ppc++];

	    bstack[0 * 2 + 2] = S;
	    bstack[0 * 2 + 1] = value(S);
	    value(S) = bytecode_regs[s & 0xf];

	    s = *ppc++;
	    S = atoms[*ppc++];

	    bstack[1 * 2 + 2] = S;
	    bstack[1 * 2 + 1] = value(S);
	    value(S) = bytecode_regs[s & 0xf];
	    bstack = bstack + 2 * 2;
	    break;
	case 0xa7:		/* lamdbind 3 */
	    bstackcheck();
	    s = *ppc++;
	    S = atoms[*ppc++];

	    bstack[0 * 2 + 2] = S;
	    bstack[0 * 2 + 1] = value(S);
	    value(S) = bytecode_regs[s & 0xf];
	    s = *ppc++;
	    S = atoms[*ppc++];
	    bstack[1 * 2 + 2] = S;
	    bstack[1 * 2 + 1] = value(S);
	    value(S) = bytecode_regs[s & 0xf];
	    s = *ppc++;
	    S = atoms[*ppc++];
	    bstack[2 * 2 + 2] = S;
	    bstack[2 * 2 + 1] = value(S);
	    value(S) = bytecode_regs[s & 0xf];
	    bstack = bstack + 3 * 2;
	    break;
	case 0xa8:		/* freerstr */
	    r = (code & 0xff);
	    for (i = 0; i < r; i++) {
		value(bstack[-i * 2]) = bstack[-i * 2 - 1];
	    }
	    bstack = bstack - i * 2;
	    break;
	case 0xa9:		/* freerstr 1 */
	    value(bstack[-0 * 2]) = bstack[-0 * 2 - 1];
	    bstack = bstack - 1 * 2;
	    break;
	case 0xaa:		/* freerstr 2 */
	    value(bstack[-0 * 2]) = bstack[-0 * 2 - 1];
	    value(bstack[-1 * 2]) = bstack[-1 * 2 - 1];
	    bstack = bstack - 2 * 2;
	    break;
	case 0xab:		/* freerstr 3 */
	    value(bstack[-0 * 2]) = bstack[-0 * 2 - 1];
	    value(bstack[-1 * 2]) = bstack[-1 * 2 - 1];
	    value(bstack[-2 * 2]) = bstack[-2 * 2 - 1];
	    bstack = bstack - 3 * 2;
	    break;
	case 0xb0:		/* builtin 0 */
	    r = code & 0xff;
	    A = (f0[r]) ();
	    break;
	case 0xb1:		/* builtin 1 */
	    r = code & 0xff;
	    A = (f1[r]) (A);
	    break;
	case 0xb2:		/* builtin 2 */
	    r = code & 0xff;
	    A = (f2[r]) (A, Q);
	    break;
	case 0xb3:		/* builtin 3 */
	    r = code & 0xff;
	    A = (f3[r]) (A, Q, M);
	    break;
	case 0xc0:		/* builtin 0 */
	    r = code & 0xff;
	    A = (f0[r]) ();
	    if (bytecode_sp <= sp)
		return A;
	    ppc = *--bytecode_sp;
	    break;
	case 0xc1:		/* builtin 1 */
	    r = code & 0xff;
	    A = (f1[r]) (A);
	    if (bytecode_sp <= sp)
		return A;
	    ppc = *--bytecode_sp;
	    break;
	case 0xc2:		/* builtin 2 */
	    r = code & 0xff;
	    A = (f2[r]) (A, Q);
	    if (bytecode_sp <= sp)
		return A;
	    ppc = *--bytecode_sp;
	    break;
	case 0xc3:		/* builtin 3 */
	    r = code & 0xff;
	    A = (f3[r]) (A, Q, M);
	    if (bytecode_sp <= sp)
		return A;
	    ppc = *--bytecode_sp;
	    break;
	default:
	    // printf("invalid code 0x%04x at %d\n", code, ppc -
	    // bytecode);
	    error1("invalid opcode", mkfxat(code));
	}
    }
}



Sex
Llink(Sex fn, long n)
{
    Sex             def,
                    v;
    int             typ;
    int             traced;
    int             i;
    Sex             args[20];

    if (!is_symbol(fn))
	return error1("Llink", fn);
    if (qtag(fn) & PROC_BCODE)
	return interpreter(bcode(fn), n, nil);

    if (!is_symbol(fn))
	error1("bad function", fn);

    traced = trace_flag & qtag(fn);
    // traced = 1;
    stackcheck();
    if (traced) {
	push(fn);
	for (i = 1; i <= n; i++)
	    push(bytecode_regs[i]);
	trace_arg_print(n);
	for (i = n - 1; i >= 0; i--)
	    pop(args[i]);
	fn = stack[0];
    } else {
	for (i = 0; i < n; i++)
	    args[i] = bytecode_regs[i + 1];
    }
    typ = qtag(fn) & PROC_MASK;
    if (typ == PROC_EXPR) {
	Sex             bvl;
	def = (Sex) proc(fn);
	if (!is_cons(def)) {
	    error1("bad arg for ", fn);
	}
	bvl = car(def);
	for (i = 0; i < n; i++) {
	    if (bvl == nil)
		error1("too many args", fn);
	    v = car(bvl);
	    pushb(v);
	    value(v) = args[i];
	}

	if (cdr(bvl) != nil) {
#if 0
	    lisp_printf("interpreter: %p\n", args[0]);
		if(n > 0)
	    lisp_printf("interpreter: %p\n", args[1]);
#endif
	    error1("too few args", fn);
	}
	def = progn_fn(cdr(def), nil);
	popb();
    } else if (typ & PROC_CCODE) {
#if MAX_NARGS < 15
	if (n > MAX_NARGS)
	    return error1("arguments exceed MAX", fn);
#endif
	switch (n) {
	case 0:
	    def = (*proc0(fn)) ();
	    break;
	case 1:
	    def = (*proc1(fn)) (args[0]);
	    break;
	case 2:
	    def = (*proc2(fn)) (args[0], args[1]);
	    break;
	case 3:
	    def = (*proc3(fn)) (args[0], args[1], args[2]);
	    break;
	case 4:
	    def = (*proc4(fn)) (args[0], args[1], args[2], args[3]);
	    break;
	case 5:
	    def =
		(*proc5(fn)) (args[0], args[1], args[2], args[3], args[4]);
	    break;
	case 6:
	    def =
		(*proc6(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5]);
	    break;
	case 7:
	    def =
		(*proc7(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5], args[6]);
	    break;
	case 8:
	    def =
		(*proc8(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5], args[6], args[7]);
	    break;
	case 9:
	    def =
		(*proc9(fn)) (args[0], args[1], args[2], args[3], args[4],
			      args[5], args[6], args[7], args[8]);
	    break;
	case 10:
	    def =
		(*proc10(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8],
			       args[9]);
	    break;
	case 11:
	    def =
		(*proc11(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10]);
	    break;
	case 12:
	    def =
		(*proc12(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11]);
	    break;
	case 13:
	    def =
		(*proc13(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11], args[12]);
	    break;
	case 14:
	    def =
		(*proc14(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11], args[12], args[13]);
	    break;
	case 15:
	    def =
		(*proc15(fn)) (args[0], args[1], args[2], args[3], args[4],
			       args[5], args[6], args[7], args[8], args[9],
			       args[10], args[11], args[12], args[13],
			       args[14]);
	    break;
	default:
	    return error1("too many argument", fn);
	}
    } else {
	return error1("undefined function", fn);
    }
    if (traced) {
	push(def);
	trace_val_print();
	pop(def);
	popv(1);
    }
    return def;
}
