/*
 * kctype.h             (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * Handlings of characters in Japanese code
 * $Header: /Users/kako/asuka/src/RCS/kctype.h,v 2.2 2014/05/29 06:07:14 kako Exp $
 */

#define iskchar(c) (c&0xff00)

#ifdef Kanji
extern long     _iskalpha(long);
extern long     _iskalnum(long);
extern long     _iskspace(long);
extern long     _iskupper(long);
extern long     _isklower(long);
extern long     _toklower(long);
extern long     _tokupper(long);
#endif

typedef unsigned short kchar_t;

#define G_DELIM	0x10
#define G_SPACE	0x20
#define W_ESCAP 0x08

#define W_ALPHA 0x01
#define W_DIGIT 0x02
#define W_DELIM 0x04

#ifdef Kanji

#ifdef Shift_JIS
#define KT_SFT1 0x40		/* 
				 * * * * * Shift_JIS first character     
				 */
#define KT_SFT2 0x80		/* 
				 * * * * * Shift_JIS second character     
				 */
#endif
#ifdef EUC
#define KT_EUC1 0x40		/* 
				 * * * * * EUC first character     
				 */
#define KT_EUC2 0x80		/* 
				 * * * * * EUC second  character     
				 */
#endif
#endif

extern unsigned char _kctype_[];

#ifdef Kanji
#ifdef EUC
#define	is1byte(c) ((_kctype_[c+1] & KT_EUC1) == 0)
#define	is2byte(c) (_kctype_[c+1] & KT_EUC1)
#else
#ifdef Shift_JIS
#define	is1byte(c) ((_kctype_[c+1] & KT_SFT1) == 0)
#define	is2byte(c) (_kctype_[c+1] & KT_SFT1)
#else
#define	is1byte(c) ((c & 0x80) == 0)
#define is2byte(c) (c & 0x80)
#endif
#endif


#define isklower(c) ((c < 127) ? islower(c) : _isklower(c))
#define iskupper(c) ((c < 127) ? isupper(c) : _iskupper(c))
#define isedigit(c) ((c < 127) && isdigit(c) )
#define isexdigit(c) ((c < 127) && isxdigit(c) )
#define iskspace(c) ((c < 127) ? isspace(c) : _iskspace(c))
#define iskgraph(c) ((c < 127) && isgraph(c) )
/*
 * #define iskcntrl(c) ((c < 127) && iscntrl(c) ) 
 */
#define iskcntrl(c) (c < 32)
#define iskascii(c) ((c < 127) && isascii(c) )

#define toklower(c) ((c < 127) ? tolower(c) : _toklower(c))
#define tokupper(c) ((c < 127) ? toupper(c) : _tokupper(c))

#define iskalpha(c) ((c < 127) ?  isalpha(c) : _iskalpha(c))
#define iskdigit(c) ((c < 127) ?  isdigit(c) : _iskdigit(c))
#define iskalnum(c) ((c < 127) ?  isalnum(c) : _iskalnum(c))
#define ishankaku(c) (iskchar(c) == 0)
#define iszenkaku(c) iskchar(c)
#define	tozenkaku(c) ((c < 128) ? c : _tozenkaku(c))
#define	tohankaku(c) ((c < 128) ? c : _tohankaku(c))

#else

#define isklower(c) islower(c)
#define iskupper(c) isupper(c)
#define isedigit(c) isdigit(c)
#define isexdigit(c) isxdigit(c)
#define iskspace(c) isspace(c)
#define iskgraph(c) isgraph(c)
#define iskcntrl(c) iscntrl(c)
#define iskascii(c) isascii(c)
#define toklower(c) tolower(c)
#define tokupper(c) toupper(c)

#define iskalpha(c) isalpha(c)
#define iskdigit(c) isdigit(c)
#define iskalnum(c) isalnum(c)

#define ishankaku(c) (1)
#define iszenkaku(c) (0)
#define	tozenkaku(c) (c)
#define	tohankaku(c) (c)

#endif

#define is_first(c)   ((c < 127) && (_kctype_[c+1] & W_ALPHA) != 0)
#define is_second(c)  ((c < 127) && (_kctype_[c+1] & (W_ALPHA|W_DIGIT)) != 0)
#define is_delimit(c) ((c < 127) && (_kctype_[c+1] & W_DELIM) != 0)

#define is_gdelimit(c) ((c < 127) && (_kctype_[c+1] & G_DELIM) != 0)
#define is_gspace(c)   ((c < 127) && (_kctype_[c+1] & G_SPACE) != 0)


#ifdef Kanji
long            my_getwc(FILE * fp);
long            my_putwc(long c, FILE * fp);
#else
#define my_getwc(fp)	getc(fp)
#define my_putwc(c, fp)	putc(c, fp)
#endif


#ifdef Kanji
#ifdef EUC
long            _isklower(long c);
long            _iskupper(long c);
long            _iskblank(long c);
long            _iskxdigit(long c);
long            _iskspace(long c);
long            _iskpunct(long c);
long            _iskgraph(long c);
long            _iskcntrl(long c);
long            _iskascii(long c);
long            _toklower(long c);
long            _tokupper(long c);
long            _iskalpha(long c);
long            _iskdigit(long c);
long            _iskalnum(long c);
long            _tohankaku(long c);
#else
#ifdef Shift_JIS
long            _isklower(long c);
long            _iskupper(long c);
long            _iskblank(long c);
long            _iskxdigit(long c);
long            _iskspace(long c);
long            _iskpunct(long c);
long            _iskgraph(long c);
long            _iskcntrl(long c);
long            _iskascii(long c);
long            _toklower(long c);
long            _tokupper(long c);
long            _iskalpha(long c);
long            _iskdigit(long c);
long            _iskalnum(long c);
long            _tohankaku(long c);
#endif
#endif

long            euctosjis(long n);
long            sjistoeuc(long n);

#if defined(VisualC) && !defined(_CONSOLE)
long            my_putwc(long c, FILE * fp);
void            my_filbuf(void);
long            my_getwc(FILE * fp);
#else
long            my_putwc(long c, FILE * fp);
long            my_getwc(FILE * fp);
#endif

#endif
