/*
 * map.c                (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 *      mapping functions
 * $Header: /Users/kako/asuka/src/RCS/map.c,v 2.2 2013/01/11 04:15:30 kako Exp $
 */

static char const rcsid[] =
    "$Id: map.c,v 2.2 2013/01/11 04:15:30 kako Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "tags.h"
#include "kctype.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"


/*****************************************************************************/
/*
 * map composite functins
 */
/*****************************************************************************/


Sex
Lmap(Sex x, Sex fn)
{
    stackcheck();
    push2(fn, x);
#define fn stack[-1]
#define x stack[0]
    while (x != nil) {
	Lfuncall1(fn, x);
	x = cdr(x);
    }
    popv(2);
#undef fn
#undef x
    return nil;
}


Sex
Lmapc(Sex x, Sex fn)
{
    stackcheck();
    push2(fn, x);
#define fn stack[-1]
#define x stack[0]
    while (x != nil) {
	Lfuncall1(fn, car(x));
	x = cdr(x);
    }
    popv(2);
#undef fn
#undef x
    return nil;
}

Sex
Lmapcan(Sex x, Sex fn)
{
    Sex             r;

    stackcheck();
    push3(fn, x, nil);
#define fn stack[-2]
#define x stack[-1]
#define l stack[0]
    l = nil;
    while (x != nil) {
	r = Lfuncall1(fn, car(x));
	l = Lnconc(l, r);
	x = cdr(x);
    }
    r = l;
    popv(3);
#undef fn
#undef x
#undef l
    return r;
}

Sex
Lmapcon(Sex x, Sex fn)
{
    Sex             r;

    stackcheck();
    push3(fn, x, nil);
#define fn stack[-2]
#define x stack[-1]
#define l stack[0]
    l = nil;
    while (x != nil) {
	r = Lfuncall1(fn, x);
	l = Lnconc(l, r);
	x = cdr(x);
    }
    r = l;
    popv(3);
#undef fn
#undef x
#undef l
    return r;
}

Sex
Lmapcar(Sex x, Sex fn)
{
    Sex             r;

    stackcheck();
    push3(fn, x, nil);
#define fn stack[-2]
#define x stack[-1]
#define l stack[0]
    l = nil;
    while (is_cons(x)) {
	r = Lfuncall1(fn, car(x));
	l = Laconc(l, r);
	x = cdr(x);
    }
    r = l;
    popv(3);
#undef fn
#undef x
#undef l
    return r;
}

Sex
Lmaplist(Sex x, Sex fn)
{
    Sex             r;

    stackcheck();
    push3(fn, x, nil);
#define fn stack[-2]
#define x stack[-1]
#define l stack[0]
    l = nil;
    while (x != nil) {
	r = Lfuncall1(fn, x);
	l = Laconc(l, r);
	x = cdr(x);
    }
    r = l;
    popv(3);
#undef fn
#undef x
#undef l
    return r;
}

/*****************************************************************************/
/*
 * map composite functins with two arguments
 */
/*****************************************************************************/


Sex
Lmap2(Sex x, Sex y, Sex fn)
{
    stackcheck();
    push3(fn, x, y);
#define fn stack[-2]
#define x stack[-1]
#define y stack[0]
    while (x != nil) {
	Lfuncall2(fn, x, y);
	x = cdr(x);
	y = cdr(y);
    }
    popv(3);
#undef fn
#undef x
#undef y
    return nil;
}


Sex
Lmapc2(Sex x, Sex y, Sex fn)
{
    stackcheck();
    push3(fn, x, y);
#define fn stack[-2]
#define x stack[-1]
#define y stack[0]
    while (x != nil) {
	Lfuncall2(fn, car(x), car(y));
	y = cdr(y);
	x = cdr(x);
    }
    popv(3);
#undef fn
#undef x
#undef y
    return nil;
}

Sex
Lmapcan2(Sex x, Sex y, Sex fn)
{
    Sex             r;

    stackcheck();
    push4(fn, x, y, nil);
#define fn stack[-3]
#define x stack[-2]
#define y stack[-1]
#define l stack[0]
    l = nil;
    while (x != nil) {
	r = Lfuncall2(fn, car(x), car(y));
	l = Lnconc(l, r);
	y = cdr(y);
	x = cdr(x);
    }
    r = l;
    popv(4);
#undef fn
#undef x
#undef y
#undef l
    return r;
}

Sex
Lmapcon2(Sex x, Sex y, Sex fn)
{
    Sex             r;

    stackcheck();
    push4(fn, x, y, nil);
#define fn stack[-3]
#define x stack[-2]
#define y stack[-1]
#define l stack[0]
    l = nil;
    while (x != nil) {
	r = Lfuncall2(fn, x, y);
	l = Lnconc(l, r);
	y = cdr(y);
	x = cdr(x);
    }
    r = l;
    popv(4);
#undef fn
#undef x
#undef y
#undef l
    return r;
}

Sex
Lmapcar2(Sex x, Sex y, Sex fn)
{
    Sex             r;

    stackcheck();
    push4(fn, x, y, nil);
#define fn stack[-3]
#define x stack[-2]
#define y stack[-1]
#define l stack[0]
    l = nil;
    while (is_cons(x)) {
	r = Lfuncall2(fn, car(x), car(y));
	l = Laconc(l, r);
	y = cdr(y);
	x = cdr(x);
    }
    r = l;
    popv(4);
#undef fn
#undef x
#undef y
#undef l
    return r;
}

Sex
Lmaplist2(Sex x, Sex y, Sex fn)
{
    Sex             r;

    stackcheck();
    push4(fn, x, y, nil);
#define fn stack[-3]
#define x stack[-2]
#define y stack[-1]
#define l stack[0]
    l = nil;
    while (x != nil) {
	r = Lfuncall2(fn, x, y);
	l = Laconc(l, r);
	y = cdr(y);
	x = cdr(x);
    }
    r = l;
    popv(4);
#undef fn
#undef x
#undef y
#undef l
    return r;
}

/*****************************************************************************/
/*
 * map composite functins with vectors
 */
/*****************************************************************************/

Sex
Lmapv(Sex x, Sex fn)
{
    Sex             r;
    long            i,
                    siz;

    if (!is_simple_vector(x))
	error1("mapv: not vector", x);
    stackcheck();
    push3(fn, x, nil);
    siz = veclen(x);
#define fn stack[-2]
#define x stack[-1]
#define l stack[0]
    l = getvector(TAG_VECTOR, T_SIMPLE, siz);
    siz = bnum_n(siz);
    for (i = 0; i < siz; i++) {
	vector_value(l, i) = Lfuncall1(fn, vector_value(x, i));
    }
    r = l;
    popv(3);
#undef fn
#undef x
#undef l
    return r;
}

Sex
Lmapv2(Sex x, Sex y, Sex fn)
{
    Sex             r;
    long            i,
                    siz;

    if (!is_simple_vector(x) || !is_simple_vector(y))
	error("mapv2: not vector");
    stackcheck();
    siz = veclen(x);
    if (siz != veclen(y))
	error("mapv2: vector size mismatch");
    push4(fn, x, y, nil);
#define fn stack[-3]
#define x stack[-2]
#define y stack[-1]
#define l stack[0]
    l = getvector(TAG_VECTOR, T_SIMPLE, siz);
    siz = bnum_n(siz);
    for (i = 0; i < siz; i++) {
	vector_value(l, i) =
	    Lfuncall2(fn, vector_value(x, i), vector_value(y, i));
    }
    r = l;
    popv(4);
#undef fn
#undef x
#undef y
#undef l
    return r;
}
