/*
 * mfunc.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/mfunc.c,v 2.6 2016/04/08 07:19:55 kako Exp $
 */

static char const rcsid[] =
    "$Id: mfunc.c,v 2.6 2016/04/08 07:19:55 kako Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"

extern Sex      aref_setf_fn(Sex var, Sex ind, Sex val);

/***************************************************************/
/***************************************************************/
/***************************************************************/

static Sex      Msetf_fn(Sex args, Sex env);
Sex             Msetffn(Sex args, Sex env);
Sex             Mread(Sex a);
Sex             Mimport(Sex a);
Sex             Msave(Sex a);
Sex             Mfor0(Sex body, Sex env);
Sex             Mfor1(Sex wh1, Sex body, Sex env);
Sex             Mfor2(Sex to, Sex body, Sex env);
Sex             Mfor3x(Sex from, Sex by, Sex to, Sex wh1, Sex body,
		       Sex env);
Sex             Mfor4(Sex name, Sex from, Sex by, Sex wh1, Sex body,
		      Sex env);
Sex             Mfor5(Sex name, Sex from, Sex by, Sex to, Sex wh1,
		      Sex body, Sex env);
Sex             Mfor6(Sex name, Sex in, Sex wh1, Sex body, Sex env);
Sex             Mfor(Sex a, Sex env);
Sex             Mfactorial(Sex a);
Sex             Mintersect(Sex a, Sex b);
Sex             Mset_diff(Sex a, Sex b);
Sex             Mrange(Sex args);
Sex             Momit(Sex args);
Sex             Mitrsyth(Sex args, Sex env);
Sex             Mset(Sex a);
Sex             Mlist(Sex args);
Sex             Mconect(Sex args, Sex env);
Sex             Mlabel(Sex args);
Sex             Marray(Sex args);

/***************************************************************/
/***************************************************************/
/***************************************************************/


static          Sex
Msetf_fn(Sex args, Sex env)
{
    Sex             var,
                    val = nil;

    if (is_cons(cadr(args)) && car(cadr(args)) == lambda_symbol)
	return Msetffn(args, env);

    // lisp_printf("setf: %p\n", args);

    stackcheck();
    if (is_cons(args)) {
	var = car(args);
	push(var);
	args = cdr(args);
	if (is_cons(args)) {
	    push3(args, env, var);
	    val = car(args);
	    val = eval(val, env);
	    pop3(var, env, args);
	    args = cdr(args);
	} else
	    val = nil;
	pop(var);

	// lisp_printf("var = %p\n", var);

	if (is_symbol(var)) {
	    if (var == nil || var == true)
		aerror("setf (bad variable)");
	    value(var) = val;
	} else if (car(var) == aref_symbol || car(var) == array_symbol) {
	    return aref_setf_fn(cadr(var), cadr(cdr(var)), val);
	} else
	    return aerror1("setf", var);
    }
    return val;
}




Sex
Msetffn(Sex args, Sex env)
{
    Sex             var;
    Sex             val = nil;

    var = car(args);
    push(var);
    val = cadr(args);

    Lputd(var, expr_symbol, val);
    pop(var);
    return var;

}


Sex
Mread(Sex a)
{
    return nil;
}

Sex
Mimport(Sex a)
{
    return nil;
}

Sex
Msave(Sex a)
{
    return nil;
}

Sex
Mfor0(Sex body, Sex env)
{
    return eval(body, env);
}

/*
 * Sex Mfor1(Sex wh1, Sex body, Sex env) { Sex c, a = nil; while(1) { c =
 * eval(wh, env); if(c == nil) break; a = eval(body, env); } return a; } 
 */

Sex
Mfor1(Sex wh1, Sex body, Sex env)
{
    Sex             a = nil;

    push3(wh1, body, env);
    push2(a, nil);
#define wh1 stack[-4]
#define body stack[-3]
#define env stack[-2]
#define a stack[-1]
#define c stack[0]

    while (1) {
	c = eval(wh1, env);
	if (c == nil)
	    break;
	a = eval(body, env);
    }
#undef wh1
#undef body
#undef env
#undef a
    a = stack[-1];
#undef c
    popv(5);
    return a;
}


Sex
Mfor2(Sex to, Sex body, Sex env)
{
    Sex             from = mkfxat(0);
    Sex             by = mkfxat(1);
    Sex             a = nil;

    push3(to, body, env);
    push2(from, a);

#define to stack[-4]
#define body stack[-3]
#define env stack[-2]
#define from stack[-1]
#define a stack[0]


    while (1) {
	if (Lgeq(from, to) == nil)
	    break;
	a = eval(body, env);
	from = plus2(from, by);
    }
#undef to
#undef body
#undef env
#undef from
#undef a
    a = stack[0];
    popv(5);
    return a;
}

Sex
Mfor3x(Sex from, Sex by, Sex to, Sex wh1, Sex body, Sex env)
{
    Sex             a = nil;

    push4(wh1, to, body, env);
    push2(from, a);

#define wh1 stack[-5]
#define to stack[-4]
#define body stack[-3]
#define env stack[-2]
#define from stack[-1]
#define a stack[0]


    while (1) {
	if (Lgeq(from, to) != nil)
	    break;
	if (wh1 != nil && eval(wh1, env) == nil)
	    break;
	a = eval(body, env);
	from = plus2(from, by);
    }
#undef wh1
#undef to
#undef body
#undef env
#undef from
#undef a
    a = stack[0];
    popv(5);
    return a;
}


Sex
Mfor4(Sex name, Sex from, Sex by, Sex wh1, Sex body, Sex env)
{
    Sex             a = nil;

    push3(name, from, by)
	push4(wh1, body, env, a);


#define name stack[-6]
#define from stack[-5]
#define by stack[-4]
#define wh1 stack[-3]
#define body stack[-2]
#define env stack[-1]
#define a stack[0]


    while (1) {
	if (eval(wh1, env) == nil)
	    break;
	a = eval(body, env);
	value(name) = plus2(value(name), by);

    }
#undef wh1
#undef body
#undef env
#undef name
#undef from
#undef by
#undef a
    a = stack[0];
    popv(7);
    return a;
}


Sex
Mfor5(Sex name, Sex from, Sex by, Sex to, Sex wh1, Sex body, Sex env)
{
    Sex             a = nil;

    push4(to, wh1, body, env);
    push4(name, from, by, a);


#define to stack[-7]
#define wh1 stack[-6]
#define body stack[-5]
#define env stack[-4]
#define name stack[-3]
#define from stack[-2]
#define by stack[-1]
#define a stack[0]

    value(name) = from;

    while (1) {
	if (Lgeq(value(name), to) != nil)
	    break;
	if (wh1 != nil && eval(wh1, env) == nil)
	    break;
	a = eval(body, env);
	value(name) = plus2(value(name), by);
    }
#undef to
#undef wh1
#undef body
#undef env
#undef name
#undef from
#undef by
#undef a
    a = stack[0];
    popv(8);
    return a;
}

Sex
Mfor6(Sex name, Sex in, Sex wh1, Sex body, Sex env)
{
    Sex             a = nil;

    push3(wh1, body, env);
    push3(in, name, a);

#define wh1 stack[-5]
#define body stack[-4]
#define env stack[-3]
#define in stack[-2]
#define name stack[-1]
#define a stack[0]

    if (!is_cons(in))
	in = Lncons(in);
    while (is_cons(in)) {
	if (wh1 != nil && eval(wh1, env) == nil)
	    break;
	value(name) = car(in);
	in = cdr(in);
	a = eval(body, env);
    }

#undef wh1
#undef body
#undef env
#undef in
#undef name
#undef a

    a = stack[0];
    popv(6);
    return a;
}


Sex
Mfor(Sex a, Sex env)
{
    Sex             name,
                    from,
                    by,
                    to,
                    wh1,
                    body;
    Sex             in = nil;

    if ((Llength(a)) == mkfxat(6)) {
	name = car(a);
	a = cdr(a);
	from = car(a);
	a = cdr(a);
	by = car(a);
	a = cdr(a);
	to = car(a);
	a = cdr(a);
	wh1 = car(a);
	a = cdr(a);
	body = car(a);
    } else {
	name = car(a);
	a = cdr(a);
	in = car(a);
	a = cdr(a);
	wh1 = car(a);
	a = cdr(a);
	body = car(a);
    }

    if (in == nil) {
	if (name == nil) {
	    if (to == nil) {
		if (wh1 == nil)
		    return Mfor0(body, env);
		else
		    return Mfor1(wh1, body, env);
	    } else {
		if (from == nil)
		    from = mkfxat(1);
		if (by == nil)
		    by = mkfxat(1);
		return Mfor3x(from, by, to, wh1, body, env);
	    }
	} else {
	    if (from == nil)
		from = mkfxat(1);
	    if (by == nil)
		by = mkfxat(1);
	    if (to == nil) {
		if (wh1 == nil)
		    return Mfor0(body, env);
		else
		    return Mfor1(wh1, body, env);
		/*
		 * return Mfor4(name, from, by, wh1, body, env); 
		 */
	    } else {
		return Mfor5(name, from, by, to, wh1, body, env);
	    }
	}
    } else
	return Mfor6(name, in, wh1, body, env);
}



Sex
Mfactorial(Sex a)
{
    return Lfactorial(a);
}

Sex
Mintersect(Sex a, Sex b)
{
    Sex             c = nil;

    stackcheck();
    push3(a, b, c);

    while (is_cons(a)) {
	if ((Lmember(car(a), b)) != nil) {
	    c = Lxcons(c, car(a));
	}
	a = cdr(a);
    }
    popv(3);
    return Lreverse(c);
}


Sex
Mset_diff(Sex a, Sex b)
{
    Sex             c = nil;
    Sex             d = Mintersect(a, b);

    stackcheck();
    push3(a, b, c);

    while (is_cons(a)) {
	if ((Lmember(car(a), d)) == nil) {
	    c = Lxcons(c, car(a));
	}
	a = cdr(a);
    }
    popv(3);
    return Lreverse(c);
}

Sex
Mrange(Sex args)
{
    Sex             c = nil;

    if (!is_cons(car(args))) {
	Sex             a = car(args);
	Sex             b = cadr(args);
	while (b > mkfxat(0)) {
	    c = cons(a, c);
	    b = diff2(b, mkfxat(1));
	}
	return c;
    } else {
	c = cdr(car(args));
	return Momit(c);
    }
}

Sex
Momit(Sex args)
{
    Sex             a = car(args);
    Sex             b = cadr(args);
    Sex             c = nil;

    push4(args, a, b, c);

    while (b > a) {
	c = cons(b, c);
	b = diff2(b, mkfxat(1));
    }
    c = cons(a, c);
    popv(4);
    return c;
}

Sex
Mitrsyth(Sex args, Sex env)
{
    Sex             a = car(args);
    Sex             b = cadr(args);
    Sex             c = nil;

    push4(args, a, b, c);

    if (is_cons(b)) {
	c = Llist2(a, cadr(b));
	b = car(b);
    } else
	c = Lncons(a);
    while (b > mkfxat(1)) {
	c = Llist2(a, c);
	b = diff2(b, mkfxat(1));
    }
    popv(4);
    return eval(c, env);

}


Sex
Mset(Sex a)
{

    push3(a, nil, nil);
#define a stack[-2]
#define b stack[-1]
#define c stack[0]

    while (is_cons(a)) {
	b = car(a);
	b = eval(b, nil);
	if ((Lmember(b, c)) == nil)
	    c = cons(b, c);
	a = cdr(a);
    }
#undef a
    a = c;
#undef b
#undef c
    popv(3);
    return Lreverse(a);
}


Sex
Mlist(Sex args)
{
    return Levlis(args);
}


Sex
Mconect(Sex args, Sex env)
{
    Sex             a = car(args);
    Sex             b = cdr(args);

    while (is_cons(b)) {
	Sex             c = car(b);
	if (!is_number(c))
	    if (value(c) != undefined_var)
		c = eval(c, env);
	a = Lstrcat(a, c);
	b = cdr(b);
    }
    return a;
}


Sex
Mlabel(Sex args)
{
    Sex             a = car(args);

    char            buf[20];
    sprintf(buf, "%%%lld", a);
    return make_symbol(buf);

}

Sex
Marray(Sex args)
{
    Sex             name = car(args);
    Sex             box = cdr(args);

    while (is_cons(box)) {
	if (is_cons(car(box)))
	    name = Lstrcat(name, car(box));
	else
	    name = Lstrcat(name, Lncons(car(box)));
	box = cdr(box);
    }
    return name;
}


// extern Sex Lsetkorder(Sex a);
// extern Sex Lreorder(Sex a);

setup_type const mfunc_setup[] = {
    // {"setkorder", (lisp_proc *)Lsetkorder, 1},
    // {"reorder", (lisp_proc *)Lreorder, 1},

    {":=", (lisp_proc *) Msetf_fn, -1},
    {"setq", (lisp_proc *) Msetf_fn, -1},
    {"read", (lisp_proc *) Mread, -1},
    {"import", (lisp_proc *) Mimport, -1},
    {"save", (lisp_proc *) Msave, 1},
    {"for", (lisp_proc *) Mfor, -1},
    {"!", (lisp_proc *) Mfactorial, 1},
    {"intersect", (lisp_proc *) Mintersect, 2},
    {"set_diff", (lisp_proc *) Mset_diff, 2},
    {"$", (lisp_proc *) Mrange, -1},
    {"..", (lisp_proc *) Momit, -1},
    {"@@", (lisp_proc *) Mitrsyth, -1},
    {"setm", (lisp_proc *) Mset, -1},
    {"listq", (lisp_proc *) Mlist, -1},
    {".", (lisp_proc *) Mconect, -1},
    {"%", (lisp_proc *) Mlabel, -1},
    {"array", (lisp_proc *) Marray, -1},
    {NULL, 0, 0}
};

setup_type const gfunc_setup[] = {
    // {"setkorder", (lisp_proc *)Lsetkorder, 1},
    // {"reorder", (lisp_proc *)Lreorder, 1},
    {"intersect", (lisp_proc *) Mintersect, 2},
    {"set_diff", (lisp_proc *) Mset_diff, 2},

    {":=", (lisp_proc *) Msetf_fn, -1},
    {"setq", (lisp_proc *) Msetf_fn, -1},
    {"!", (lisp_proc *) Mfactorial, 1},
    {"$", (lisp_proc *) Mrange, -1},
    {"..", (lisp_proc *) Momit, -1},
    {"@@", (lisp_proc *) Mitrsyth, -1},
    {".", (lisp_proc *) Mconect, -1},
    {"%", (lisp_proc *) Mlabel, -1},
    {NULL, 0, 0}
};
