/*
 * misc.c      (C) Copyright 2006-2007 by Fujio Kako, Nara Women's University.
 * Transcendental Functions.
 *      $Header: /Users/kako/asuka/src/RCS/misc.c,v 2.63 2016/09/06 05:50:02 kako Exp $
 */

static char const rcsid[] =
    "$Id: misc.c,v 2.63 2016/09/06 05:50:02 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
// #include "print.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "entries.h"
#include "misc.h"
#include "fns.h"
#include "polynomial.h"
#include "tri.h"
#include "garith.h"

#define caar(x)	(car(car(x)))
#define cdar(x)	(cdr(car(x)))

#define lcoeff(a) caar(a)
#define ldeg(a) cdar(a)
#define coeff(a) car(a)
#define power(a) cdr(a)
#define mkterm(c,p) cons(c, p)
#define reduct(a) cdr(a)
#define lterm(a) car(a)


extern Sex      with_timeout(Sex n, Sex u);

extern Sex      recipiM(Sex a, Sex b);

extern Sex      Lfrexp(Sex a);

extern Sex      Lceiling(Sex a);
extern Sex      Lfloor(Sex a);
extern Sex      Llcm(Sex a, Sex b);
extern Sex      Lround(Sex a);
extern Sex      Ltruncate(Sex a);
extern Sex      gcd(Sex a, Sex b);
extern Sex      gcd(Sex a, Sex b);
extern Sex      Ldenominator(Sex a);
extern Sex      Lnumerator(Sex a);
extern Sex      Lash(Sex a, Sex b);
extern Sex      Lash1(Sex a, Sex b);
extern Sex      Ldivide(Sex a, Sex b);
extern Sex      Levenp(Sex a);
extern Sex      logand2(Sex a, Sex b);
extern Sex      logeqv2(Sex a, Sex b);
extern Sex      lognot(Sex a);
extern Sex      logxor2(Sex a, Sex b);
extern Sex      logor2(Sex a, Sex b);
extern Sex      Llsd(Sex a);
extern Sex      Lsrandom(Sex a);
extern Sex      Lrandom(Sex a);
extern Sex      Lmax2(Sex a, Sex b);
extern Sex      Lmin2(Sex a, Sex b);
#ifdef ASUKA
extern Sex      gnegate(Sex a);
#else
extern Sex      negate(Sex a);
#endif
extern Sex      Lminusp(Sex a);
extern Sex      modulus(Sex a, Sex b);
extern Sex      Lmsd(Sex a);
extern Sex      Loddp(Sex a);
extern Sex      Lonep(Sex a);
extern Sex      plus2(Sex a, Sex b);
extern Sex      Lplusp(Sex a);
extern Sex      times2(Sex a, Sex b);
extern Sex      Lzerop(Sex a);
extern Sex      Ladd1(Sex a);
extern Sex      diff2(Sex a, Sex b);
extern Sex      Leqn(Sex a, Sex b);
extern Sex      Lappeq(Sex a, Sex b);
extern Sex      Lgeq(Sex a, Sex b);
extern Sex      Lgreaterp(Sex a, Sex b);
extern Sex      Lleq(Sex a, Sex b);
extern Sex      Lleq(Sex a, Sex b);
extern Sex      Lgeq(Sex a, Sex b);
extern Sex      Llessp(Sex a, Sex b);
extern Sex      Lcgeq(Sex a, Sex b);
extern Sex      Lcgreaterp(Sex a, Sex b);
extern Sex      Lcleq(Sex a, Sex b);
extern Sex      Lcleq(Sex a, Sex b);
extern Sex      Lcgeq(Sex a, Sex b);
extern Sex      Lclessp(Sex a, Sex b);
extern Sex      ratio2(Sex a, Sex b);
extern Sex      quot2(Sex a, Sex b);
extern Sex      remaind2(Sex a, Sex b);
extern Sex      reciprocal(Sex a);
extern Sex      Lsub1(Sex a);
extern Sex      Lprecision(Sex a);
/*
 * type conversion routines 
 */
extern Sex      Lfix(Sex a);
extern Sex      Llogical(Sex a);
extern Sex      Llogical(Sex a);
extern Sex      Lrational(Sex a);
extern Sex      Lmkrational(Sex a, Sex b);
extern Sex      Lfloat(Sex a);
extern Sex      Lfloat(Sex a);
extern Sex      Lefloat(Sex a);
extern Sex      Lmkefloat(Sex a, Sex b);
extern Sex      Lmkefloat1(Sex a);
extern Sex      Linterval(Sex a);
extern Sex      Linterval_by_ratio(Sex a, Sex f);
extern Sex      Lmkinterval(Sex a, Sex b);
extern Sex      Lcomplex(Sex a);
extern Sex      Lmkcomplex(Sex a, Sex b);
extern Sex      Lecomplex(Sex a);
extern Sex      Lmkecomplex(Sex a, Sex b, Sex c);
extern Sex      Lbfloat(Sex a);
extern Sex      Lbfloat(Sex a);
extern Sex      Lbefloat(Sex a);
extern Sex      Lmkbefloat(Sex a, Sex b);
extern Sex      Lbinterval(Sex a);
extern Sex      Lmkbinterval(Sex a, Sex b);
extern Sex      Lbecomplex(Sex a);
extern Sex      Lmkbecomplex(Sex a, Sex b, Sex c);
extern Sex      Lmkcomplex(Sex a, Sex b);
extern Sex      Lmkecomplex(Sex a, Sex b, Sex c);
extern Sex      Lcomplex_conjugate(Sex a);
extern Sex      Lmkmodular(Sex a);
/*
 * interval 
 */
extern Sex      Linterval_lower(Sex a);
extern Sex      Linterval_upper(Sex a);
extern Sex      Linterval_mid(Sex a);
extern Sex      Linterval_width(Sex a);
extern Sex      Lmkinterval(Sex a, Sex b);
/*
 * efloat
 */
extern Sex      Lefloat_error(Sex a);
extern Sex      Lefloat_value(Sex a);
extern Sex      Lecomplex_error(Sex a);
extern Sex      Lecomplex_real(Sex a);
extern Sex      Lecomplex_imag(Sex a);
extern Sex      Lcomplex_real(Sex a);
extern Sex      Lcomplex_imag(Sex a);
/*
 */
#ifdef ASUKA
extern Sex      gplus2(Sex a, Sex b);
extern Sex      gdiff2(Sex a, Sex b);
extern Sex      gtimes2(Sex a, Sex b);
extern Sex      gquot2(Sex a, Sex b);
extern Sex      gremaind2(Sex a, Sex b);
extern Sex      gexpt(Sex a, Sex b);
extern Sex      Lgeqn(Sex a, Sex b);
extern Sex      gcross(Sex a, Sex b);
#else
extern Sex      plus2(Sex a, Sex b);
extern Sex      diff2(Sex a, Sex b);
extern Sex      times2(Sex a, Sex b);
extern Sex      quot2(Sex a, Sex b);
extern Sex      remaind2(Sex a, Sex b);
extern Sex      Lexpt(Sex a, Sex b);
extern Sex      Leqn(Sex a, Sex b);
#endif



extern Sex      Lappeq(Sex a, Sex b);
extern Sex      Llessp(Sex a, Sex b);
extern Sex      Lgreaterp(Sex a, Sex b);
extern Sex      Lgeq(Sex a, Sex b);
extern Sex      Lleq(Sex a, Sex b);
/*
 */
extern Sex      Lbfloat_man(Sex a);
extern Sex      Lbfloat_exp(Sex a);
extern Sex      Lleftshift(Sex a, Sex b);
extern Sex      Lgen_prime(Sex a, Sex b);
extern Sex      Lnext_prime(Sex a);
extern Sex      Lprime_before(Sex a);

extern Sex      Lprimep(Sex a);

extern Sex      return_fn(Sex a);

extern Sex      Lprint_precision(Sex a);

extern Sex      Lclone(Sex u);

extern Sex      Mgcd(Sex a, Sex b);
extern Sex      Mlcm(Sex a, Sex b);

extern Sex      Lvprintf(Sex f, Sex u);
extern Sex      Lprintf0(Sex f);
extern Sex      Lprintf1(Sex f, Sex u);
extern Sex      Lprintf2(Sex f, Sex u, Sex v);

extern Sex      Lsetorder_lex(Sex a);
extern Sex      Lsetorder_tlex(Sex a);
extern Sex      Lsetorder_revlex(Sex a);
extern Sex      Lsetorder_trevlex(Sex a);
extern Sex      Lsetorder_gradings(Sex a, Sex b);

extern Sex      all_variables(Sex a);


static int      compare_pair(Sex a, Sex b);
static Sex      append_lcm(Sex a);
Sex             Asort_pairset(Sex a);
Sex             Amerge_pairset(Sex a, Sex b);
static void     quick_sort_pair(Sex * a, long left, long right);

static Sex      Adipoly_order(Sex a);
static Sex      Adipoly_pol(Sex a);
static Sex      Ais_dipoly_list(Sex a);
static Sex      Asugar_strategy(Sex a);
static Sex      Ais_sugar_strategy(void);
Sex             Aordervec(Sex a);
static Sex      Apause(void);
static Sex      Adump(Sex a);
static Sex      Amake_id(Sex a, Sex b);
static Sex      Amake_uid(Sex a, Sex b);

extern Sex      Lmake_id(Sex a);
extern Sex      Lmake_uid(Sex a);

static int      sugar_strategy = 0;

Sex             Acompare(Sex u, Sex v);
long            compare(Sex u, Sex v);


static Sex      Adebug_on(void);
static Sex      Adebug_off(void);
static Sex      Adebug_level(Sex);
static Sex      Adebug_depth(Sex);
Sex             Adebug(Sex a);
Sex             Adebugn(Sex a);
static int      star_debug = NO;
static int      star_debug_level = 0;
static int      star_debug_depth = 0;
static int      star_debug_newline = 0;

Sex             Aenable(Sex);
Sex             Adisable(Sex);
/**************************************************************/
/**************************************************************/

Sex
Aprecision(Sex a)
{
    Sex             p;
    push(a)
	p = Lprecision(a);
    a = stack[0];
    stack[0] = p;
    Lprint_precision(a);
    p = stack[0];
    popv(1);
    return p;
}

/**************************************************************/
/**************************************************************/

static          Sex
Adebug_level(Sex a)
{
    if (is_fixnum(a))
	star_debug_level += mkint(a);
    else if (a == true)
	star_debug_level++;
    else if (a == nil)
	star_debug_level--;
    return mkfxat(star_debug_level);
}


static          Sex
Adebug_depth(Sex a)
{
    if (is_fixnum(a))
	star_debug_depth = mkint(a);
    return mkfxat(star_debug_depth);
}

static          Sex
Adebug_on(void)
{
    star_debug = YES;
    star_debug_level = 0;
    star_debug_newline = 0;
    return nil;
}


static          Sex
Adebug_off(void)
{
    star_debug = NO;
    return nil;
}

#if 0
Sex
Adebug(Sex args)
{
    star_debug_newline = 0;
    if (star_debug)
	lisp_printf("%a\n", args);
    return nil;
}
#endif

Sex
Adebug(Sex args)
{
    Sex             p;
    int             n;

    star_debug_newline = 0;
    if (!star_debug)
	return nil;
    push2(args, nil);
    p = args;
    n = star_debug_level;
    while (n-- > 0)
	lisp_printf(" ");
#ifdef DEBUG_DEPTH
    print_flag = 0;
#endif
    while (p != nil) {
	stack[0] = cdr(p);
#ifdef DEBUG_DEPTH
	if (star_debug_depth)
	    hprint_sex(car(p), char_to_stdout, star_debug_depth);
	else
	    print_sex(car(p), char_to_stdout);
#else
	lisp_printf("%a", car(p));
#endif

	p = stack[0];
    }
    lisp_printf("\n");
    p = stack[-1];
    popv(2);
    // return p;
    return nil;
}

Sex
Adebugn(Sex args)
{
    Sex             p;
    int             n;
    if (!star_debug)
	return nil;

    push2(args, nil);
    p = args;
    if (star_debug_newline == 0) {
	n = star_debug_level;
	while (n-- > 0)
	    lisp_printf(" ");
    }
    star_debug_newline = 1;
#ifdef DEBUG_DEPTH
    print_flag = 0;
#endif
    while (p != nil) {
	stack[0] = cdr(p);
#ifdef DEBUG_DEPTH
	if (star_debug_depth)
	    hprint_sex(car(p), char_to_stdout, star_debug_depth);
	else
	    print_sex(car(p), char_to_stdout);
#else
	lisp_printf("%a", car(p));
#endif
	p = stack[0];
    }
    p = stack[-1];
    popv(2);
    // return p;
    return nil;
}


Sex
Awrite(Sex args)
{
    Sex             p;
    push2(args, nil);
    p = args;
    while (p != nil) {
	stack[0] = cdr(p);
	lisp_printf("%a", car(p));
	p = stack[0];
    }
    lisp_printf("\n");
    p = stack[-1];
    popv(2);
    // return p;
    return nil;
}

Sex
Awriten(Sex args)
{
    Sex             p;
    long            pl = line_length;
    push2(args, nil);
    p = args;
    line_length = 0;
    while (p != nil) {
	stack[0] = cdr(p);
	lisp_printf("%a", car(p));
	p = stack[0];
    }
    line_length = pl;
    p = stack[-1];
    popv(2);
    // return p;
    return nil;
}

Sex
Aclear(Sex a)
{
    Sex             p;
    Sex             b;
    p = a;
    while (p != nil) {
	b = car(p);
	if (is_symbol(b)) {
	    qtag(b) &= ~(SYM_FLUID | SYM_GLOBAL);
	    value(b) = undefined_var;
	}
	p = cdr(p);
    }
    // return a;
    return nil;
}

Sex
Aload(Sex a)
{
    Sex             fd,
                    old;
    Sex            *init_stack;
    Sex            *curr_stack;
    extern int      value_print_flag;
    int             f = value_print_flag;

    curchar = NOT_CHAR;
    fd = Lopen(a, input_symbol);
    push(fd);
    old = Lrds(fd);
    push(old);
    asuka_read_eval_print(f ? true : nil);
    pop(old);
    Lrds(old);
    pop(fd);
    if (is_file(fd)) {
	Lclose(fd);
    } else {
	current_package = value(current_asuka_package);
	current_charvec = package_charvec(current_package);
    }

    return true;
}



Sex
Aimport(Sex a)
{
    Sex             fd,
                    old;
    Sex            *init_stack;
    Sex            *curr_stack;
    char            name[1024];
    char           *p,
                   *q;
    long            len;

    strcpy(name, "~/lib/asuka/");
    p = name + strlen(name);
    stackcheck();
    push(a);
    if (is_symbol(a)) {
	len = symbol_length(a);
	q = pname(a);
	strncpy(p, q, len);
    } else if (!is_string(a)) {
	popv(1);
	return aerror("import");
    } else {
	len = string_length(a);
	q = &string_value(a, 0);
	strncpy(p, q, len);
    }
    p[len] = 0;
    a = make_string(name);

    curchar = NOT_CHAR;
    fd = Lopen_file(a, input_symbol);
    a = stack[0];
    if (fd == nil)
	fd = Lopen_file(a, input_symbol);
    if (fd == nil) {
	pop(a);
	return aerror1("import file not found", a);
    }
    stack[0] = fd;
    old = Lrds(fd);
    push(old);
    asuka_read_eval_print(nil);
    pop(old);
    Lrds(old);
    pop(fd);
    if (is_file(fd)) {
	Lclose(fd);
    } else {
	current_package = value(current_asuka_package);
	current_charvec = package_charvec(current_package);
    }

    return true;
}


Sex
Asize(Sex a)
{
    Sex             c;

    if (is_array(a))
	return array_dim(a);
    else if (is_matrix(a)) {
	push(a);
	c = matrix_dimy(a);
	c = Lncons(c);
	pop(a);
	c = cons(matrix_dimx(a), c);
	return c;
    } else if (is_vect(a)) {
	return vect_dimx(a);
    }



    return aerror1("not an array", a);
}

Sex
Amcoeff(p)
     Sex             p;
{
    long            n;
    long            i;
    Sex            *m;
    if (!is_poly(p))
	return nil;
    n = mkint(ldeg(poly_pol(p)));
    push4(p, nil, nil, nil);
#define p stack[-3]
#define q stack[-2]
#define b stack[-1]
#define a stack[-0]
    q = mkarray(n + 1);
    /*
     * m = &array_value(q,0); for(i = 0; i <= n; i++) *m++ = mkfxat(0); 
     */
    a = poly_pol(p);
    m = &array_value(q, 0);
    while (a != nil) {
	b = lterm(a);
	i = mkint(power(b));
	b = coeff(b);
	if (i < 0 || i > n) {
	    popv(4);
	    return aerror("System error, polynomial degree invalid");
	}
	m[i] = b;
	a = reduct(a);
    }
#undef p
    p = stack[-2];
#undef q
#undef b
#undef a
    popv(4);
    return p;
}


Sex
Acoeff(p, x)
     Sex             p,
                     x;
{
    long            n;
    long            i;
    Sex            *m;
    if (is_poly(x))
	x = poly_var(x);
    stackcheck();
    push5(p, x, nil, nil, nil);
#define p stack[-4]
#define x stack[-3]
#define q stack[-2]
#define b stack[-1]
#define a stack[-0]
    kernel_order = cons(x, kernel_order);
    p = Lreorder(p);
    if (!is_poly(p) || poly_var(p) != x) {
	n = 0;
	q = mkarray(1);
	m = &array_value(q, 0);
	m[0] = p;
    } else {
	a = poly_pol(p);
	n = mkint(ldeg(a));
	q = mkarray(n + 1);
	m = &array_value(q, 0);
	while (a != nil) {
	    b = lterm(a);
	    i = mkint(power(b));
	    b = coeff(b);
	    if (i < 0 || i > n) {
		popv(5);
		kernel_order = cdr(kernel_order);
		return aerror("System error, polynomial degree invalid");
	    }
	    m[i] = b;
	    a = reduct(a);
	}
    }
    kernel_order = cdr(kernel_order);
    for (i = 0; i < n; i++) {
	m = &array_value(q, 0);
	a = Lreorder(m[i]);
	m = &array_value(q, 0);
	m[i] = a;
    }
#undef x
#undef p
    p = stack[-2];
#undef q
#undef b
#undef a
    popv(5);
    return p;
}


Sex
Acoeffn(p, x, k)
     Sex             p,
                     x,
                     k;
{
    Sex             n;

    ASSERT(is_poly(p));
    ASSERT(is_fixnum(k));

    if (is_poly(x))
	x = poly_var(x);
    stackcheck();
    push5(p, x, nil, nil, nil);
#define p stack[-4]
#define x stack[-3]
#define q stack[-2]
#define b stack[-1]
#define a stack[-0]
    kernel_order = cons(x, kernel_order);
    p = Lreorder(p);
    q = mkfxat(0);
    if (!is_poly(p) || poly_var(p) != x) {
	if (k == mkfxat(0))
	    q = p;
    } else {
	a = poly_pol(p);
	while (a != nil) {
	    b = lterm(a);
	    n = power(b);
	    b = coeff(b);
	    if (k == n) {
		q = b;
		break;
	    } else if (k > n) {
		break;
	    }
	    a = reduct(a);
	}
    }
    kernel_order = cdr(kernel_order);
    q = Lreorder(q);
#undef x
#undef p
    p = stack[-2];
#undef q
#undef b
#undef a
    popv(5);
    return p;
}


Sex
Acoeffx(p, x)
     Sex             p,
                     x;
{
    long            o;
    Sex             q,
                    y;
    Sex             c,
                    r;
    Sex             n;

    for (;;) {

	if (zerop(x))
	    return mkfxat(0);

	if (!is_poly(p)) {
	    if (!is_poly(x))
		return p;
	    return mkfxat(0);
	}
	if (!is_poly(x)) {
	    while (is_poly(p)) {
		q = poly_pol(p);
		while (q != nil && ldeg(q) != mkfxat(0))
		    q = reduct(q);
		if (q == nil)
		    return mkfxat(0);
		p = lcoeff(q);
	    }
	    return p;
	}

	while (is_poly(p)) {
	    o = compare_var(poly_var(p), poly_var(x));
	    if (o > 0)
		return mkfxat(0);
	    if (o == 0)
		break;
	    q = poly_pol(p);
	    while (q != nil && ldeg(q) != mkfxat(0))
		q = reduct(q);
	    if (q == nil)
		return mkfxat(0);
	    p = lcoeff(q);
	}
	if (!is_poly(p))
	    break;

	q = poly_pol(p);
	y = poly_pol(x);
	x = lcoeff(y);
	n = ldeg(y);
	while (q != nil && n != ldeg(q)) {
	    q = reduct(q);
	}
	if (q == nil)
	    break;
	p = lcoeff(q);
    }
    return mkfxat(0);
}


Sex
Acoeffy(Sex p, Sex x)
{
    Sex             b;

    b = nil;
    push3(p, x, b);
#define p stack[-2]
#define x stack[-1]
#define b stack[0]
    while (is_cons(x)) {
	b = cons(Acoeffx(p, car(x)), b);
	x = cdr(x);
    }
#undef p
    p = b;
#undef x
#undef b
    popv(3);
    return Lnreverse(p);
}


Sex
Asupport(Sex p)
{
    Sex             x,
                    q;
    Sex             s;

    q = nil;
    if (zerop(p))
	return q;
    if (!is_poly(p))
	return cons(mkfxat(1), q);

    x = poly_var(p);
    p = poly_pol(p);
    push2(x, p);
    push4(nil, nil, nil, nil);
#define x stack[-5]
#define p stack[-4]
#define a stack[-3]
#define b stack[-2]
#define q stack[-1]
#define r stack[-0]
    while (p != nil) {
	a = power(lterm(p));
	b = Asupport(lcoeff(p));
	r = nil;
	while (is_cons(b)) {
	    s = make_polynomial(x, car(b), a);
	    r = cons(s, r);
	    b = cdr(b);
	}
	r = Lnreverse(r);
	q = Amerge(q, r);
	p = reduct(p);
    }
#undef x
#undef p
    p = q;
#undef a
#undef b
#undef q
#undef r
    popv(6);
    return p;
}

Sex
Asupport_ap(Sex p, Sex e)
{
    Sex             x,
                    q;
    Sex             s;

    q = nil;
    if (!is_poly(p)) {
	push(e);
	p = Labs(p);
	pop(e);
/**
lisp_printf("support: %p", p);
lisp_printf("  %p\n", e);
*/
	if (lessp2(p, e))
	    return q;
	return cons(mkfxat(1), q);
    }

    stackcheck();

    x = poly_var(p);
    p = poly_pol(p);
    push3(x, p, e);
    push4(nil, nil, nil, nil);
#define x stack[-6]
#define p stack[-5]
#define e stack[-4]
#define a stack[-3]
#define b stack[-2]
#define q stack[-1]
#define r stack[-0]
    while (p != nil) {
	a = power(lterm(p));
	b = Asupport_ap(lcoeff(p), e);
	r = nil;
	while (is_cons(b)) {
	    s = make_polynomial(x, car(b), a);
	    r = cons(s, r);
	    b = cdr(b);
	}
	r = Lnreverse(r);
	q = Amerge(q, r);
	p = reduct(p);
    }
#undef x
#undef p
#undef e
    p = q;
#undef a
#undef b
#undef q
#undef r
    popv(7);
    return p;
}



Sex
Anew_symbol1(x, n1)
     Sex             x,
                     n1;
{
    char            name[64];
    long            n;

    if (is_poly(x))
	x = Mvar(x);
    if (is_string(x)) {
	n = string_length(x);
	if (n > 30)
	    n = 30;
	memcpy(name, &string_value(x, 0), n);
    } else if (is_symbol(x)) {
	n = symbol_length(x);
	if (n > 30)
	    n = 30;
	memcpy(name, pname(x), n);
    } else
	return aerror("new_symbol1");
    sprintf(name + n, "%ld", mkint(n1));
    return intern(name, (long) strlen(name));
}

Sex
Anew_symbol2(x, n1, n2)
     Sex             x,
                     n1,
                     n2;
{
    char            name[64];
    long            n;

    if (is_poly(x))
	x = Mvar(x);
    if (is_string(x)) {
	n = string_length(x);
	if (n > 30)
	    n = 30;
	memcpy(name, &string_value(x, 0), n);
    } else if (is_symbol(x)) {
	n = symbol_length(x);
	if (n > 30)
	    n = 30;
	memcpy(name, pname(x), n);
    } else
	return aerror("new_symbol2");
    sprintf(name + n, "%ld_%ld", mkint(n1), mkint(n2));
    return intern(name, (long) strlen(name));
}

/**************************************************************/
/********* spolynomial and mreduce         ********************/
/**************************************************************/
Sex
Aspoly(Sex a, Sex b)
{
    if (is_number(b)) {
	if (zerop(b))
	    return a;
	return mkfxat(1);
    }
    if (is_number(a)) {
	if (zerop(a))
	    return b;
	return mkfxat(1);
    }
    if (is_dipoly(b)) {
	if (is_dipoly(a))
	    return Adipoly_spoly(a, b);
	if (is_poly(a)) {
	    push(b);
	    a = Ldipoly(a);
	    pop(b);
	    return Adipoly_spoly(a, b);
	}

	return aerror2("spoly", a, b);
    }
    if (is_poly(b)) {
	if (is_dipoly(a)) {
	    push(a);
	    b = Ldipoly(b);
	    pop(a);
	    return Adipoly_spoly(a, b);
	}
	if (is_poly(a))
	    return Apoly_spoly(a, b);
    }
    return aerror2("spoly", a, b);
}

Sex
Areduce(Sex a, Sex g)
{
    if (g == nil)
	return a;
    if (!is_cons(g))
	return Areduce1(a, g);
    if (is_number(g)) {
	return mkfxat(0);
    }
    if (is_number(a)) {
	return a;
    }
    if (is_poly(a)) {
	if (is_poly(car(g)))
	    return Apoly_reduce(a, g);
	if (is_dipoly(car(g))) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_reduce(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(car(g))) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_reduce(a, g);
	}
	if (is_dipoly(car(g)))
	    return Adipoly_reduce(a, g);
    }
    return aerror2("reduce", a, g);
}

Sex
Areduce1(Sex a, Sex g)
{
    if (is_number(g)) {
	return mkfxat(0);
    }
    if (is_number(a)) {
	return a;
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_reduce1(a, g);
	if (is_dipoly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_reduce1(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_reduce1(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_reduce1(a, g);
    }
    return aerror2("reduce", a, g);
}


Sex
Afull_reduce(Sex a, Sex g)
{
    if (g == nil)
	return a;
    if (!is_cons(g))
	return Afull_reduce1(a, g);
    if (is_number(g)) {
	return mkfxat(0);
    }
    if (is_number(a)) {
	return a;
    }
    if (is_poly(a)) {
	if (is_poly(car(g)))
	    return Apoly_full_reduce(a, g);
	if (is_dipoly(car(g))) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_full_reduce(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(car(g))) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_full_reduce(a, g);
	}
	if (is_dipoly(car(g)))
	    return Adipoly_full_reduce(a, g);
    }
    return aerror2("reduce", a, g);
}

Sex
Afull_reduce1(Sex a, Sex g)
{
    if (is_number(g)) {
	return mkfxat(0);
    }
    if (is_number(a)) {
	return a;
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_full_reduce1(a, g);
	if (is_dipoly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_full_reduce1(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_full_reduce1(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_remainder(a, g);
    }
    return aerror2("reduce", a, g);
}

Sex
Areduce_set(Sex a)
{
    if (!is_cons(a))
	return a;
    if (is_dipoly(car(a)))
	return Adipoly_reduce_set(a);
    if (is_poly(car(a)))
	return Apoly_reduce_set(a);
    return aerror1("reduce_set", a);
}

Sex
Ahave_nontrivial_gcd(Sex a, Sex g)
{
    if (is_number(g)) {
	return nil;
    }
    if (is_number(a)) {
	return nil;
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_have_nontrivial_gcd(a, g);
	if (is_dipoly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_have_nontrivial_gcd(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_have_nontrivial_gcd(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_have_nontrivial_gcd(a, g);
    }
    return aerror2("have_nontrivial_gcd", a, g);
}

#if 0
Sex
Alt_lcm_ordervec(Sex a, Sex b)
{
    if (!is_dipoly(a) || !is_dipoly(b))
	aerror2("lt_lcm_ordervec: not dipoly", a, b);
    return Adipoly_lt_lcm_ordervec(a, b);
}
#endif

Sex
Alt_lcm_expvec(Sex a, Sex b)
{
    if (!is_dipoly(a) || !is_dipoly(b))
	aerror2("lt_lcm_expvec: not dipoly", a, b);
    return Adipoly_lt_lcm_expvec(a, b);
}

Sex
Alt_reducible(Sex a, Sex g)
{
    if (is_number(g)) {
	return nil;
    }
    if (is_number(a)) {
	return nil;
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_lt_reducible(a, g);
	if (is_dipoly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_lt_reducible(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_lt_reducible(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_lt_reducible(a, g);
    }
    return aerror2("lt_reducible", a, g);
}

Sex
Alt_divisible(Sex a, Sex g)
{
    if (is_number(g) || (is_dipoly(g) && dipoly_is_number(g))) {
	return true;
    }
    if (is_number(a) || (is_dipoly(a) && dipoly_is_number(a))) {
	return nil;
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_lt_divisible(a, g);
	if (is_dipoly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_lt_divisible(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_lt_divisible(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_lt_divisible(a, g);
    }
    return aerror2("lt_divisible", a, g);
}


Sex
Aunknowns(Sex a)
{
    if (!is_poly(a))
	return nil;
    return all_variables(a);
}

Sex
Amax_coeff(Sex f)
{
    if (!is_poly(f))
	return f;
    stackcheck();
    push3(f, nil, nil);
#define f stack[-2]
#define c stack[-1]
#define a stack[-0]
    a = mkfxat(0);
    f = poly_pol(f);
    while (f != nil) {
	c = lcoeff(f);
	f = reduct(f);
	if (is_poly(c))
	    c = Amax_coeff(c);
	else if (minusp(c))
	    c = negate(c);
	if (lessp2(a, c))
	    a = c;
    }
#undef f
#undef c
#undef a
    f = stack[0];
    popv(3);
    return f;
}

Sex
Anterms(Sex f)
{
    if (!is_poly(f))
	return mkfxat(1);
    stackcheck();
    push3(f, nil, nil);
#define f stack[-2]
#define c stack[-1]
#define n stack[-0]
    n = mkfxat(0);
    f = poly_pol(f);
    while (f != nil) {
	c = lcoeff(f);
	f = reduct(f);
	if (is_poly(c))
	    c = Anterms(c);
	else
	    c = mkfxat(1);
	n = plus2(n, c);
    }
#undef f
#undef c
#undef n
    f = stack[0];
    popv(3);
    return f;
}

/**************************************************************/
/**************************************************************/

Sex
Amonic(Sex g)
{
    Sex             a;
    if (g == nil)
	return g;
    if (is_cons(g))
	return Amonic1(g);
    if (is_number(g)) {
	return mkfxat(1);
    }
    if (is_poly(g)) {
	a = g;
	while (is_poly(a))
	    a = coeff(lterm(poly_pol(a)));
	if (a == mkfxat(1))
	    return g;
	return gratio2(g, a);
    }
    if (is_dipoly(g)) {
	a = dipoly_pol(g);
	if (a == nil)
	    return mkfxat(0);
	a = car(a);
#if DIPOLY_COEFF == 0
	a = car(a);
#else
	a = vect_value(a, 0);
#endif
	if (a == mkfxat(1))
	    return g;
	return gratio2(g, a);
    }
    return aerror1("monic", g);
}

Sex
Amonic1(Sex g)
{
    Sex             r;
    stackcheck();
    push3(g, nil, nil);
#define g stack[-2]
#define f stack[-1]
#define p stack[0]
    while (is_cons(g)) {
	f = car(g);
	g = cdr(g);
	f = Amonic(f);
	p = Laconc(p, f);
    }
#undef g
    g = stack[-2];
    r = p;
#undef p
#undef f
    popv(3);
    if (g != nil) {
	aerror1("monic", g);
    }
    return r;
}

/**************************************************************/
/**************************************************************/

Sex
ALT_gcd(Sex a, Sex g)
{
    if (is_number(g)) {
	return mkfxat(1);
    }
    if (is_number(a)) {
	return mkfxat(1);
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_LT_gcd(a, g);
	if (is_dipoly(g)) {
	    push(a);
	    g = Lpolynomial(g);
	    pop(a);
	    return Apoly_LT_gcd(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_LT_gcd(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_LT_gcd(a, g);
    }
    return aerror2("LTgcd", a, g);
}

Sex
ALT_lcm(Sex a, Sex g)
{
    if (is_number(g)) {
	return a;
    }
    if (is_number(a)) {
	return g;
    }
    if (is_poly(a)) {
	if (is_poly(g))
	    return Apoly_LT_lcm(a, g);
	if (is_dipoly(g)) {
	    push(a);
	    g = Lpolynomial(g);
	    pop(a);
	    return Apoly_LT_lcm(a, g);
	}
    }
    if (is_dipoly(a)) {
	if (is_poly(g)) {
	    push(g);
	    a = Lpolynomial(a);
	    pop(g);
	    return Apoly_LT_lcm(a, g);
	}
	if (is_dipoly(g))
	    return Adipoly_LT_lcm(a, g);
    }
    return aerror2("LTlcm", a, g);
}

/**************************************************************/


Sex
Alt_lcm_ordervec(Sex f, Sex g)	// lcm of exponent vectors 
{
    long            i,
                    n,
                    m;
    Sex             q;
    Sex             p;
    Sex            *pa,
                   *pb,
                   *pc;
    Sex             ax,
                    bx;
    long            order;
    long            sugar;

    if (!is_dipoly(f) || !is_dipoly(g))
	aerror("lt_lcm_ordervec: not distributed polynomial");


    q = dipoly_order(f);
    if (q != dipoly_order(g))
	aerror("lt_lcm_ordervec: order mismath");

    order = mkint(car(q));
    sugar = (sugar_strategy || (order & 2) != 0) ? 1 : 0;

    q = cdr(q);
    n = list_length(q) + sugar;
    push2(f, g);
    q = make_vector(n);
    pop2(g, f);
    f = dipoly_pol(f);
    g = dipoly_pol(g);
    if (f != nil && g != nil) {
	f = car(f);
	g = car(g);
#if DIPOLY_COEFF == 0
	f = cdr(f);
	g = cdr(g);
	pa = &vect_value(f, order / 2);
	pb = &vect_value(g, order / 2);
#else
	pa = &vect_value(f, order / 2 + 1);
	pb = &vect_value(g, order / 2 + 1);
#endif
	pc = &vect_value(q, sugar);
	for (i = sugar; i < n; i++) {
	    ax = *pa++;
	    bx = *pb++;
	    *pc++ = ax > bx ? ax : bx;
	}
	if (sugar) {
	    m = 0;
	    for (i = 1; i < n; i++) {
		m += *(--pc) - TAG_FIXNUM;
	    }
	    vect_value(q, 0) = m + TAG_FIXNUM;
	}
    } else {
	if (f == nil)
	    f = g;
	if (f == nil)
	    return q;
	f = car(f);
#if DIPOLY_COEFF == 0
	f = cdr(f);
	pa = &vect_value(f, order / 2);
#else
	pa = &vect_value(f, order / 2 + 1);
#endif
	pb = &vect_value(q, sugar);
	for (i = sugar; i < n; i++)
	    *pb++ = *pa++;
	if (sugar) {
	    m = 0;
	    for (i = 1; i < n; i++) {
		m += *(--pa) - TAG_FIXNUM;
	    }
	    vect_value(q, 0) = m + TAG_FIXNUM;
	}
    }
    return q;
}

/**************************************************************/

Sex
Aordervec(Sex f)		// exponent vector of monomial
{
    long            i,
                    n,
                    m;
    Sex             q;
    Sex             p;
    Sex            *pa,
                   *pb;
    long            order;
    long            sugar;

    if (!is_dipoly(f))
	aerror("ordervec: not distributed polynomial");


    q = dipoly_order(f);
    order = mkint(car(q));
    sugar = (sugar_strategy || (order & 2) != 0) ? 1 : 0;

    q = cdr(q);
    n = list_length(q) + sugar;
    push(f);
    q = make_vector(n);
    pop(f);
    f = dipoly_pol(f);
    if (f != nil) {
	f = car(f);
#if DIPOLY_COEFF == 0
	f = cdr(f);
	pa = &vect_value(f, order / 2);
#else
	pa = &vect_value(f, order / 2 + 1);
#endif
	pb = &vect_value(q, sugar);
	for (i = sugar; i < n; i++)
	    *pb++ = *pa++;
	if (sugar) {
	    m = 0;
	    for (i = 1; i < n; i++) {
		m += *(--pa) - TAG_FIXNUM;
	    }
	    vect_value(q, 0) = m + TAG_FIXNUM;
	}
    }
    return q;
}

/**************************************************************/

int
compare_pair(Sex a, Sex b)
{
    long            n;
    long            r;
    long            i;
    Sex            *pa,
                   *pb;

    /*
     * lisp_printf("a=%p, ", a); lisp_printf(", b=%p\n", b); 
     */
    a = car(a);
    b = car(b);
    n = mkint(vect_dimx(a));
#if DIPOLY_COEFF == 0
    pa = &vect_value(a, 0);
    pb = &vect_value(b, 0);
    for (i = 0; i < n; i++) {
	r = *pa++ - *pb++;
	if (r != 0)
	    return -r;
    }
#else
    pa = &vect_value(a, 1);
    pb = &vect_value(b, 1);
    for (i = 1; i < n; i++) {
	r = *pa++ - *pb++;
	if (r != 0)
	    return -r;
    }
#endif
    return 0;
}

static void
quick_sort_pair(Sex * a, long left, long right)
{
    long            l_save,
                    r_save;
    long            ll;
    Sex             pivot;

    l_save = left;
    r_save = right;
    pivot = a[left];

    while (left < right) {
	while (compare_pair(a[right], pivot) <= 0 && left < right)
	    right--;
	if (left != right) {
	    a[left] = a[right];
	    left++;
	}
	while (compare_pair(a[left], pivot) >= 0 && left < right)
	    left++;
	if (left != right) {
	    a[right] = a[left];
	    right--;
	}
    }
    a[left] = pivot;
    ll = left;
    left = l_save;
    right = r_save;
    if (left < ll)
	quick_sort_pair(a, left, ll - 1);
    if (right > ll)
	quick_sort_pair(a, ll + 1, right);
}



Sex
append_lcm(Sex a)
{
    Sex             p,
                    q,
                    r;
    long            i,
                    n,
                    m;
    Sex             c,
                    d;
    Sex            *pa,
                   *pb,
                   *pc;
    long            order;
    long            sugar;

    p = car(a);
    q = cdr(a);
    if (!is_dipoly(p) || !is_dipoly(q))
	aerror2("not distributed poly", p, q);
    c = dipoly_order(p);
    order = mkint(car(c));
    sugar = (sugar_strategy || (order & 2) != 0) ? 1 : 0;
    p = (dipoly_pol(p));
    q = (dipoly_pol(q));
    if (p == nil || q == nil)
	return nil;
    // lisp_printf("q=%p\n", q);
    p = car(p);
    q = car(q);

#if DIPOLY_COEFF == 0
    p = cdr(p);
    q = cdr(q);
#endif
    n = mkint(vect_dimx(p));
    if (sugar_strategy && (order & 2) == 0)
	n++;
    push3(a, p, q);
#if DIPOLY_COEFF == 0
    r = make_vector(n);
#else
    n--;
    r = make_vector(n);
#endif
    pop3(q, p, a);
#if DIPOLY_COEFF == 0
    pa = &vect_value(p, order / 2);
    pb = &vect_value(q, order / 2);
#else
    pa = &vect_value(p, order / 2 + 1);
    pb = &vect_value(q, order / 2 + 1);
#endif
    pc = &vect_value(r, sugar);
    for (i = sugar; i < n; i++) {
	c = *pa++;
	d = *pb++;
	*pc++ = c > d ? c : d;
    }
    if (sugar) {
	m = 0;
	for (i = 1; i < n; i++)
	    m += *(--pc) - TAG_FIXNUM;
	vect_value(r, 0) = m + TAG_FIXNUM;
    }

    a = cons(r, a);
    return a;
}


Sex
Asort_pairset(Sex a)
{
    Sex             p;
    long            i,
                    j,
                    n;

    p = a;
    n = 0;
    stackcheck();
    while (p != nil) {
	p = cdr(p);
	n++;
    }
    push3(a, p, nil);
#define a stack[-2]
#define p stack[-1]
#define q stack[0]
    p = make_vector(n);
    for (i = 0; i < n; i++) {
	q = car(a);
	q = append_lcm(q);
	if (q == nil) {
	    n--;
	    i++;
	} else {
	    vect_value(p, i) = q;
	}
	a = cdr(a);
    }
    // qsort(&vect_value(p, 0), n, sizeof(Sex), (compare_fn
    // *)compare_pair);
    quick_sort_pair(&vect_value(p, 0), 0, n - 1);
    a = nil;
    for (j = n - 1; j >= 0; j--)
	a = cons(vect_value(p, j), a);
#undef a
    a = stack[-2];
#undef p
#undef q
    popv(3);
    return a;
}

Sex
Amerge_pairset(Sex a, Sex b)
{
    stackcheck();
    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define p stack[0]
    c = cons(nil, nil);
    p = c;
    while (a != nil && b != nil) {
	if (compare_pair(car(a), car(b)) >= 0) {
	    c = rcons(c, car(a));
	    a = cdr(a);
	} else {
	    c = rcons(c, car(b));
	    b = cdr(b);
	}
    }
    if (a != nil)
	cdr(c) = a;
    else if (b != nil)
	cdr(c) = b;
#undef a
#undef b
#undef c
#undef p
    a = stack[0];
    popv(4);
    return cdr(a);
}


/**************************************************************/
/********* Conversion between various typs ********************/
/**************************************************************/

Sex
converter(Sex a, one_args * cf)
{

    if (is_cons(a))
	return list_converter(a, cf);
    if (is_number(a))
	return cf(a);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_ALGNUM:
	case N_TRNNUM:
	    break;
	case N_POLY:
	    return polynomial_converter(a, cf);
	case N_RPOLY:
	    return rpoly_converter(a, cf);
	case N_DIPOLY:
	    return dipoly_converter(a, cf);
	case N_VECT:
	    return vector_converter(a, cf);
	case N_MATRIX:
	    return matrix_converter(a, cf);
	}
    }
    return aerror1("converter", a);
}

Sex
list_converter(Sex a, one_args * cf)
{
    Sex             b;

    stackcheck();
    push2(a, nil);
#define a stack[-1]
#define c stack[0]
    while (a != nil) {
	b = car(a);
	b = converter(b, cf);
	c = cons(b, c);
	a = cdr(a);
    }
#undef a
#undef c
    a = stack[0];
    popv(2);
    return Lnreverse(a);
}

Sex
polynomial_converter(Sex a, one_args * cf)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = converter(b, cf);
	c = cdar(a);		/* leading degree */
	if (!zerop(b)) {
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

Sex
rpoly_converter(Sex a, one_args * cf)
{
    Sex             b,
                    c;

    b = rpoly_num(a);
    c = rpoly_den(a);
    push2(b, c);
    b = converter(b, cf);
    stack[-1] = b;
    c = stack[0];
    c = converter(c, cf);
    b = stack[-1];
    popv(2);
    return make_rpoly(a, b);
}

Sex
dipoly_converter(Sex a, one_args * cf)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = converter(c, cf);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = converter(c, cf);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

/**************************************************************/
/**************************************************************/



Sex
converter2(Sex a, Sex b, two_args * cf)
{

    if (is_cons(a))
	return list_converter2(a, b, cf);
    if (is_number(a))
	return cf(a, b);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_ALGNUM:
	case N_TRNNUM:
	    break;
	case N_POLY:
	    return polynomial_converter2(a, b, cf);
	case N_RPOLY:
	    return rpoly_converter2(a, b, cf);
	case N_DIPOLY:
	    return dipoly_converter2(a, b, cf);
	case N_VECT:
	    return vector_converter2(a, b, cf);
	case N_MATRIX:
	    return matrix_converter2(a, b, cf);
	}
    }
    return aerror2("converter2", a, b);
}

Sex
list_converter2(Sex a, Sex bb, two_args * cf)
{
    Sex             b;

    stackcheck();
    push3(bb, a, nil);
#define bb stack[-2]
#define a stack[-1]
#define c stack[0]
    while (a != nil) {
	b = car(a);
	b = converter2(b, bb, cf);
	c = cons(b, c);
	a = cdr(a);
    }
#undef bb
#undef a
#undef c
    a = stack[0];
    popv(3);
    return Lnreverse(a);
}

Sex
polynomial_converter2(Sex a, Sex b, two_args * cf)
{
    push(b);
    push5(a, nil, nil, nil, nil);

#define bb stack[-5]
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = converter2(b, bb, cf);
	c = cdar(a);		/* leading degree */
	if (!zerop(b)) {
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef bb
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(6);
    return a;
}

Sex
rpoly_converter2(Sex a, Sex d, two_args * cf)
{
    Sex             b,
                    c;

    b = rpoly_num(a);
    c = rpoly_den(a);
    push3(b, c, d);
    b = converter2(b, d, cf);
    stack[-2] = b;
    c = stack[-1];
    d = stack[0];
    c = converter2(c, d, cf);
    b = stack[-2];
    popv(3);
    return make_rpoly(a, b);
}

Sex
dipoly_converter2(Sex a, Sex b, two_args * cf)
{
    push(b);
    push5(a, nil, nil, nil, nil);

#define bb stack[-5]
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = converter2(c, bb, cf);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = converter2(c, bb, cf);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef bb
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(6);
    return a;
}

/**************************************************************/
/**************************************************************/


Sex
Aunmodular(Sex a)
{

    if (is_number(a))
	return Lunmodular(a);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_ALGNUM:
	case N_TRNNUM:
	    break;
	case N_POLY:
	    return Aunmodular_pol(a);
	case N_RPOLY:
	    break;
	case N_DIPOLY:
	    return Aunmodular_dipoly(a);
	case N_VECT:
	    return Aunmodular_vect(a);
	case N_MATRIX:
	    return Aunmodular_mat(a);
	}
    }
    return aerror1("unmodular", a);
}

Sex
Amodular(Sex a)
{
    return converter(a, Lmkmodular);
}

Sex
Amkmodular(Sex a, Sex b)
{
    if (is_fixnum(b)) {
	Lset_modulus(b);
	return converter2(a, b, Lmodular);
    } else {
	Lset_big_modulus(b);
	return converter2(a, b, Lbmodular);
    }
}




Sex
Afmodular(Sex a)
{
    return converter(a, Lfmodular);
}


Sex
Afloat(Sex a)
{
    return converter(a, Lfloat);
}

Sex
Abfloat(Sex a)
{
    return converter(a, Lbfloat);
}

Sex
Aefloat(Sex a)
{
    return converter(a, Lefloat);
}

Sex
Abefloat(Sex a)
{
    return converter(a, Lbefloat);
}


Sex
Ainterval(Sex a, Sex f)
{
    return converter2(a, f, Linterval_by_ratio);
}

Sex
Amkefloat(Sex a, Sex f)
{
    return converter2(a, f, Lmkefloat);
}

Sex
Amkefloat_by_ratio(Sex a, Sex f)
{
    return converter2(a, f, Lmkefloat_by_ratio);
}

Sex
Aefloat_value(Sex a)
{
    return converter(a, Lefloat_value);
}
Sex
Aefloat_error(Sex a)
{
    return converter(a, Lefloat_error);
}

Sex
Amkbefloat(Sex a, Sex f)
{
    return converter2(a, f, Lmkbefloat);
}

Sex
Amkbefloat_by_ratio(Sex a, Sex f)
{
    return converter2(a, f, Lmkbefloat_by_ratio);
}


/**************************************************************/

Sex
Aunmodular_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Aunmodular(b);
	c = cdar(a);		/* leading degree */
	c = make_polynomial(e, b, c);
	d = gplus2(d, c);
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}



Sex
Amkmodular_pol(Sex a, Sex b)
{
    push(b);
    push5(a, nil, nil, nil, nil);

#define bb stack[-5]
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Amkmodular(b, bb);
	if (!zerop(b)) {
	    c = cdar(a);	/* leading degree */
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef bb
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(6);
    return a;
}

Sex
Amodular_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Amodular(b);
	if (!zerop(b)) {
	    c = cdar(a);	/* leading degree */
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

Sex
Afmodular_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Afmodular(b);
	if (!zerop(b)) {
	    c = cdar(a);	/* leading degree */
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}



Sex
Afloat_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Afloat(b);
	c = cdar(a);		/* leading degree */
	c = make_polynomial(e, b, c);
	d = gplus2(d, c);
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}


Sex
Abfloat_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Abfloat(b);
	c = cdar(a);		/* leading degree */
	c = make_polynomial(e, b, c);
	d = gplus2(d, c);
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

Sex
Amkefloat_pol(Sex a, Sex f)
{
    push5(a, nil, nil, nil, nil);
    push(f);

#define a stack[-5]
#define b stack[-4]
#define c stack[-3]
#define d stack[-2]
#define e stack[-1]
#define f stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	// if(plusp(b))
	// b = Amkefloat(b, times2(b, f));
	// else
	// b = Amkefloat(b, negate(times2(b, f)));
	b = Amkefloat(b, f);
	c = cdar(a);		/* leading degree */
	c = make_polynomial(e, b, c);
	d = gplus2(d, c);
	a = cdr(a);
    }
#undef a
    a = stack[-2];
#undef b
#undef c
#undef d
#undef e
#undef f
    popv(6);
    return a;
}


Sex
Amkbefloat_pol(Sex a, Sex f)
{
    push5(a, nil, nil, nil, nil);
    push(f);

#define a stack[-5]
#define b stack[-4]
#define c stack[-3]
#define d stack[-2]
#define e stack[-1]
#define f stack[0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Amkbefloat(b, f);
	c = cdar(a);		/* leading degree */
	c = make_polynomial(e, b, c);
	d = gplus2(d, c);
	a = cdr(a);
    }
#undef a
    a = stack[-2];
#undef b
#undef c
#undef d
#undef e
#undef f
    popv(6);
    return a;
}


Sex
Aefloat_value_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Aefloat_value(b);
	if (float_val(b) != 0.0) {
	    c = cdar(a);	/* leading degree */
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

Sex
Aefloat_error_pol(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = mkfxat(0);
    e = poly_var(a);		/* main variable */
    a = poly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = caar(a);		/* leading coefficient */
	b = Aefloat_error(b);
	if (float_val(b) != 0.0) {
	    c = cdar(a);	/* leading degree */
	    c = make_polynomial(e, b, c);
	    d = gplus2(d, c);
	}
	a = cdr(a);
    }
#undef a
    a = stack[-1];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

/**************************************************************/


Sex
Aunmodular_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Aunmodular(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Aunmodular(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}






Sex
Amkmodular_dipoly(Sex a, Sex b)
{
    push(b);
    push5(a, nil, nil, nil, nil);

#define bb stack[-5]
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Amkmodular(c, bb);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Amkmodular(c, bb);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef bb
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(6);
    return a;
}




Sex
Amodular_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Amodular(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Amodular(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}





Sex
Afmodular_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Afmodular(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Afmodular(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}





Sex
Afloat_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Afloat(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Afloat(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}




Sex
Abfloat_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Abfloat(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Abfloat(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}




Sex
Amkefloat_dipoly(Sex a, Sex f)
{
    push(f);
    push5(a, nil, nil, nil, nil);

#define f stack[-5]
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Amkefloat(c, f);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Amkefloat(c, f);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef f
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(6);
    return a;
}



Sex
Amkbefloat_dipoly(Sex a, Sex f)
{
    push(f);
    push5(a, nil, nil, nil, nil);

#define f stack[-5]
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Amkbefloat(c, f);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Amkbefloat(c, f);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef f
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(6);
    return a;
}



Sex
Aefloat_value_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Aefloat_value(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Aefloat_value(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}


Sex
Aefloat_error_dipoly(Sex a)
{
    push5(a, nil, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[ 0]

    d = nil;
    e = dipoly_order(a);	/* main variable */
    a = dipoly_pol(a);		/* polynomial list */
    while (a != nil) {
	b = car(a);		/* leading coefficient */
#if DIPOLY_COEFF == 0
	c = car(b);
	b = cdr(b);
	c = Aefloat_error(c);
	b = cons(c, b);
#else
	c = vect_value(b, 0);
	b = Lclone(b);
	c = Aefloat_error(c);
	vect_value(b, 0) = c;
#endif
	if (!zerop(c)) {
	    d = cons(b, d);
	}
	a = cdr(a);
    }
    a = Lnreverse(d);
    a = make_dipoly(a, e);
#undef a
    a = stack[-4];
#undef b
#undef c
#undef d
#undef e
    popv(5);
    return a;
}

/**************************************************************/
Sex
Adipoly_order(Sex a)
{
    if (!is_dipoly(a))
	return aerror("not dipoly");
    return dipoly_order(a);
}

Sex
Adipoly_pol(Sex a)
{
    if (!is_dipoly(a))
	return aerror("not dipoly");
    return dipoly_pol(a);
}

Sex
Ais_dipoly_list(Sex a)
{
    Sex             o,
                    f,
                    l;
    if (Llistp(a) == nil)
	return nil;
    f = car(a);
    l = cdr(a);
    if (!is_dipoly(f))
	return nil;
    o = dipoly_order(f);
    if (o != dipoly_kernel_order)
	return nil;
    while (l != nil) {
	f = car(f);
	l = cdr(l);
	if (!is_dipoly(f))
	    return nil;
	if (dipoly_order(f) != o)
	    return nil;
    }
    return true;
}

static          Sex
Asugar_strategy(Sex a)
{
    Sex             b;

    b = sugar_strategy ? true : nil;
    sugar_strategy = a == nil ? 0 : 1;
    return b;
}

static          Sex
Ais_sugar_strategy(void)
{
    Sex             b;

    b = sugar_strategy ? true : nil;
    return b;
}

/**************************************************************/
/**************************************************************/

Sex
Adump_array(Sex a)
{
    long            i,
                    n,
                    m;
    if (!is_array(a))
	return nil;
    m = veclen(a);
    n = mkint(car(array_dim(a)));
    printf("array %ld, header length = %ld\n", n, m);
    for (i = 0; i < n; i++)
	printf("dump[%ld]: 0x%08llx\n", i, array_value(a, i));
    return nil;
}

/**************************************************************/
/**************************************************************/

Sex
Ais_cons(Sex a)
{
    return is_cons(a) ? true : nil;
}


Sex
Ais_fixnum(Sex a)
{
    return is_fixnum(a) ? true : nil;
}

Sex
Ais_hdrs(Sex a)
{
    return is_hdrs(a) ? true : nil;
}

Sex
Ais_reloc(Sex a)
{
    return is_relhdr(a) ? true : nil;
}

Sex
Ais_symbol(Sex a)
{
    return is_symbol(a) ? true : nil;
}

Sex
Ais_numhdr(Sex a)
{
    return is_numhdr(a) ? true : nil;
}

Sex
Ais_lisp_vector(Sex a)
{
    return is_vechdr(a) ? true : nil;
}

Sex
Ais_binhdr(Sex a)
{
    return is_binhdr(a) ? true : nil;
}

Sex
Ais_number(Sex a)
{
    return is_number(a) ? true : nil;
}

Sex
Ais_immed(Sex a)
{
    return is_immed(a) ? true : nil;
}

Sex
Ais_logical(Sex a)
{
    return is_logical(a) ? true : nil;
}

Sex
Ais_bigint(Sex a)
{
    return is_bigint(a) ? true : nil;
}

Sex
Ais_modular(Sex a)
{
    return is_modular(a) ? true : nil;
}

Sex
Ais_float(Sex a)
{
    return is_float(a) ? true : nil;
}

Sex
Ais_rational(Sex a)
{
    return is_rational(a) ? true : nil;
}

Sex
Ais_ratio(Sex a)
{
    return is_ratio(a) ? true : nil;
}

Sex
Ais_complex(Sex a)
{
    return is_complex(a) ? true : nil;
}

Sex
Ais_bfloat(Sex a)
{
    return is_bfloat(a) ? true : nil;
}

Sex
Ais_efloat(Sex a)
{
    return is_efloat(a) ? true : nil;
}

Sex
Ais_interval(Sex a)
{
    return is_interval(a) ? true : nil;
}

Sex
Ais_ecomplex(Sex a)
{
    return is_ecomplex(a) ? true : nil;
}

Sex
Ais_befloat(Sex a)
{
    return is_befloat(a) ? true : nil;
}

Sex
Ais_binterval(Sex a)
{
    return is_binterval(a) ? true : nil;
}

Sex
Ais_becomplex(Sex a)
{
    return is_becomplex(a) ? true : nil;
}

Sex
Ais_bmodular(Sex a)
{
    return is_bmodular(a) ? true : nil;
}

Sex
Ais_qmodular(Sex a)
{
    return is_qmodular(a) ? true : nil;
}

Sex
Ais_fmodular(Sex a)
{
    return is_fmodular(a) ? true : nil;
}

Sex
Ais_poly(Sex a)
{
    return is_poly(a) ? true : nil;
}

Sex
Ais_tropical(Sex a)
{
    return is_poly(a) ? true : nil;
}

Sex
Ais_dipoly(Sex a)
{
    return is_dipoly(a) ? true : nil;
}

Sex
Ais_rpoly(Sex a)
{
    return is_rpoly(a) ? true : nil;
}

Sex
Ais_bigneg(Sex a)
{
    return is_bigneg(a) ? true : nil;
}

Sex
Ais_integer(Sex a)
{
    return is_integer(a) ? true : nil;
}

Sex
Ais_fix(Sex a)
{
    return is_fix(a) ? true : nil;
}

Sex
Ais_neg(Sex a)
{
    return is_neg(a) ? true : nil;
}

Sex
Ais_app(Sex a)
{
    return is_app(a) ? true : nil;
}

Sex
Ais_log(Sex a)
{
    return is_log(a) ? true : nil;
}

Sex
Ais_eff(Sex a)
{
    return is_eff(a) ? true : nil;
}

Sex
Ais_cmp(Sex a)
{
    return is_cmp(a) ? true : nil;
}

Sex
Ais_big(Sex a)
{
    return is_big(a) ? true : nil;
}

Sex
Ais_vector(Sex a)
{
    return is_vect(a) ? true : nil;
}

Sex
Ais_matrix(Sex a)
{
    return is_matrix(a) ? true : nil;
}

Sex
Ais_array(Sex a)
{
    return is_array(a) ? true : nil;
}

Sex
Ais_string(Sex a)
{
    return is_string(a) ? true : nil;
}

Sex
Ais_iarray(Sex a)
{
    return is_iarray(a) ? true : nil;
}

Sex
Ais_sarray(Sex a)
{
    return is_sarray(a) ? true : nil;
}

Sex
Ais_darray(Sex a)
{
    return is_darray(a) ? true : nil;
}

Sex
Ais_iarray2(Sex a)
{
    return is_iarray2(a) ? true : nil;
}

Sex
Ais_sarray2(Sex a)
{
    return is_sarray2(a) ? true : nil;
}

Sex
Ais_darray2(Sex a)
{
    return is_darray2(a) ? true : nil;
}

Sex
Ais_iarray3(Sex a)
{
    return is_iarray3(a) ? true : nil;
}

Sex
Ais_sarray3(Sex a)
{
    return is_sarray3(a) ? true : nil;
}

Sex
Ais_darray3(Sex a)
{
    return is_darray3(a) ? true : nil;
}

Sex
Ais_package(Sex a)
{
    return is_package(a) ? true : nil;
}

/**************************************************************/
/**************************************************************/
Sex
Aws(Sex a)
{
#ifdef SAVE_WORK_SPACE
    a = Lassoc(a, work_space);
    if (a != nil)
	a = cdr(a);
    return a;
#else
    if (a == mkfxat(2))
	return value(twbeepre_symbol);
    if (a == mkfxat(3))
	return value(thbeepre_symbol);
    return value(onbeepre_symbol);

#endif
}
/**************************************************************/
/**************************************************************/
Sex
Apause(void)
{
    extern Sex      erdn(void);

    switch (ttygetchar("\n Pause called. Continue ?(Y/N) = ")) {
    case 'Y':
    case 'y':
	return true;
    case 'N':
    case 'n':
	erdn();
    }
    return nil;
}


Sex
Adump(Sex a)
{
    Cons_Cell      *b;
    int             i;

    if (a == nil)
	b = (Cons_Cell *) cellbase;
    else {
	b = (Cons_Cell *) celllimit;
	b += 4;
    }
    for (i = 0; i < 10; i++) {
	b--;
	printf("0x%08lx(%4d): 0x%08lx, 0x%08lx\n", (long) b, i,
	       (long) (b->kar), (long) (b->kdr));
    }
    printf("cell pointer = 0x%08lx\n", (long) cellpointer);
    return nil;
}

static          Sex
Amake_id(Sex a, Sex b)
{
    a = Lexplodec(a);
    push(a);
    b = Lexplodec(b);
    pop(a);
    a = Lappend(a, b);
    a = Lmake_id(a);
    return a;
}

static          Sex
Amake_uid(Sex a, Sex b)
{
    a = Lexplodec(a);
    push(a);
    b = Lexplodec(b);
    pop(a);
    a = Lappend(a, b);
    a = Lmake_uid(a);
    return a;
}


#ifdef UNIX
static          Sex
Asystem(Sex a)
{
    char           *command;
    a = Lexplodec(a);
    a = Lmake_uid(a);
    command = pname(a);
    system(command);
    return nil;
}
#endif
/**************************************************************/
/**************************************************************/
static          Sex
Alist2bnum(Sex a)
{
    Sex             b,
                    c;
    long            i,
                    n;
    n = 0;
    b = a;
    while (is_cons(b)) {
	b = cdr(b);
	n++;
    }
    // printf("n=%d\n", n);
    push(a);
    b = new_bignum(bnum_len(n));
    pop(a);
    for (i = 0; i < n; i++) {
	c = car(a);
	if (is_fixnum(c)) {
	    bignum_value(b)[i] = mkint(c);
	} else if (is_bigint(c)) {
	    bignum_value(b)[i] = bignum_value(c)[0];
	} else if (is_logical(c)) {
	    bignum_value(b)[i] = bignum_value(c)[0];
	}
	a = cdr(a);
    }
    return b;
}

static          Sex
Abnum2list(Sex a)
{
    long            i,
                    m;
    unsigned long   n;
    Sex             c;

    if (!is_bigint(a))
	return aerror1("not big", a);
    push2(a, nil);
#define a stack[-1]
#define b stack[-0]
    b = nil;
    m = bignum_length(a);
    m = bnum_n(m);
    for (i = 0; i < m; i++) {
	n = bignum_value(a)[i];
	c = mklogical(n);
	b = cons(c, b);
    }
    b = Lnreverse(b);
#undef a
#undef b
    a = stack[0];
    popv(2);
    return a;
}
/**************************************************************/
/**************************************************************/
Sex
Acompare(Sex u, Sex v)
{
    long            r = compare(u, v);
    if (r > 0)
	return mkfxat(1);
    if (r < 0)
	return mkfxat(-1);
    return mkfxat(0);
}

long
compare(Sex u, Sex v)
{
    long            r;

    if (is_number(u)) {
	if (is_number(v))
	    return (0);
	if (is_poly(v))
	    return (-1);
	if (is_dipoly(v)) {
	    return (-1);
	}
    }
    if (is_poly(u)) {
	if (is_number(v))
	    return (1);
	if (is_poly(v)) {
	    return compare_term(u, v);
	}
	u = Ldipoly(u);
	return compare_dipoly(u, v);
    }
    if (is_dipoly(u)) {
	if (is_dipoly(v))
	    return compare_dipoly(u, v);
	v = Ldipoly(v);
	return compare_dipoly(u, v);
    }
    return aerror2("compare:", u, v);
}

/**************************************************************/
/**************************************************************/
extern bool     modular_print;

Sex
Aenable(Sex a)
{
    Sex             b;
    stackcheck();
    // lisp_printf("enable : %p\n", a);
    while (is_cons(a)) {
	b = car(a);
	a = cdr(a);
	if (is_symbol(b)) {
	    if (strncmp(pname(b), "debug", symbol_length(b)) == 0) {
		push(a);
		Adebug_on();
		pop(a);
	    } else if (strncmp(pname(b), "modular", symbol_length(b)) == 0) {
		push(a);
		modular_print = YES;
		pop(a);
	    } else if (strncmp(pname(b), "traceall", symbol_length(b)) ==
		       0) {
		push(a);
		Ltraceall(true);
		pop(a);
	    } else if (strncmp(pname(b), "echo", symbol_length(b)) == 0) {
		value(eecho_symbol) = true;
	    }
	}
    }
    return true;
}

Sex
Adisable(Sex a)
{
    Sex             b;
    stackcheck();
    // lisp_printf("disable : %p\n", a);
    while (is_cons(a)) {
	b = car(a);
	a = cdr(a);
	if (is_symbol(b)) {
	    if (strncmp(pname(b), "debug", symbol_length(b)) == 0) {
		push(a);
		Adebug_off();
		pop(a);
	    } else if (strncmp(pname(b), "modular", symbol_length(b)) == 0) {
		push(a);
		modular_print = NO;
		pop(a);
	    } else if (strncmp(pname(b), "traceall", symbol_length(b)) ==
		       0) {
		push(a);
		Ltraceall(nil);
		pop(a);
	    } else if (strncmp(pname(b), "echo", symbol_length(b)) == 0) {
		value(eecho_symbol) = nil;
	    }
	}
    }
    return true;
}


static          Sex
Aecho(Sex a)
{
    if (a == nil)
	value(eecho_symbol) = nil;
    else
	value(eecho_symbol) = true;
    return a;
}
/**************************************************************/
/**************************************************************/

setup_type const misc_setup[] = {
    {"with_timeout", (lisp_proc *) with_timeout, -1},

    {"dump_array", (lisp_proc *) Adump_array, 1},
    {"clone", (lisp_proc *) Lclone, 1},

    {"frexp", (lisp_proc *) Lfrexp, 1},
    {"ceiling", (lisp_proc *) Lceiling, 1},
    {"floor", (lisp_proc *) Lfloor, 1},
    // {"lcm", (lisp_proc *) Llcm, 2},
    {"round", (lisp_proc *) Lround, 1},
    {"truncate", (lisp_proc *) Ltruncate, 1},
    {"gcdn", (lisp_proc *) gcd, 2},
    {"gcd", (lisp_proc *) Mgcd, 2},
    {"lcm", (lisp_proc *) Mlcm, 2},
    {"denominator", (lisp_proc *) Ldenominator, 1},
    {"numerator", (lisp_proc *) Lnumerator, 1},
    {"ash", (lisp_proc *) Lash, 2},
    {"ash1", (lisp_proc *) Lash1, 2},
    {"divide", (lisp_proc *) Ldivide, 2},
    {"evenp", (lisp_proc *) Levenp, 1},
    {"logand2", (lisp_proc *) logand2, 2},
    {"logeqv2", (lisp_proc *) logeqv2, 2},
    {"lognot", (lisp_proc *) lognot, 1},
    {"logxor2", (lisp_proc *) logxor2, 2},
    {"logor2", (lisp_proc *) logor2, 2},
    {"lsd", (lisp_proc *) Llsd, 1},
    {"srandom", (lisp_proc *) Lsrandom, 1},
    {"random", (lisp_proc *) Lrandom, 1},
    {"random_int1", (lisp_proc *) Lrandom_int1, 0},
    {"random_int2", (lisp_proc *) Lrandom_int2, 0},
    {"random_real1", (lisp_proc *) Lrandom_real1, 0},
    {"random_real2", (lisp_proc *) Lrandom_real2, 0},
    {"random_real3", (lisp_proc *) Lrandom_real3, 0},
    {"max2", (lisp_proc *) Lmax2, 2},
    {"min2", (lisp_proc *) Lmin2, 2},
#ifdef ASUKA
    {"minus", (lisp_proc *) gnegate, 1},
#else
    {"minus", (lisp_proc *) negate, 1},
#endif
    {"minusp", (lisp_proc *) Lminusp, 1},
    {"mod", (lisp_proc *) modulus, 2},
    {"msd", (lisp_proc *) Lmsd, 1},
    {"oddp", (lisp_proc *) Loddp, 1},
    {"onep", (lisp_proc *) Lonep, 1},
    {"plus2", (lisp_proc *) plus2, 2},
    {"plusp", (lisp_proc *) Lplusp, 1},
    {"times2", (lisp_proc *) times2, 2},
    {"zerop", (lisp_proc *) Lgzerop, 1},
    {"add1", (lisp_proc *) Ladd1, 1},
    {"difference", (lisp_proc *) diff2, 2},
    {"eqn", (lisp_proc *) Leqn, 2},
    {"appeqn", (lisp_proc *) Lappeq, 2},
    {"geq", (lisp_proc *) Lgeq, 2},
    {"greaterp", (lisp_proc *) Lgreaterp, 2},
    {"leq", (lisp_proc *) Lleq, 2},
    {"lep", (lisp_proc *) Lleq, 2},
    {"gep", (lisp_proc *) Lgeq, 2},
    {"lessp", (lisp_proc *) Llessp, 2},
    {"cgeq", (lisp_proc *) Lcgeq, 2},
    {"cgreaterp", (lisp_proc *) Lcgreaterp, 2},
    {"cleq", (lisp_proc *) Lcleq, 2},
    {"clep", (lisp_proc *) Lcleq, 2},
    {"cgep", (lisp_proc *) Lcgeq, 2},
    {"clessp", (lisp_proc *) Lclessp, 2},
    {"ratio", (lisp_proc *) ratio2, 2},
    {"quotient", (lisp_proc *) quot2, 2},
    {"remainder", (lisp_proc *) remaind2, 2},
    {"reciprocal", (lisp_proc *) reciprocal2, 2},
    {"sub1", (lisp_proc *) Lsub1, 1},
    {"precision", (lisp_proc *) Lprecision, 1},
    {"number_log2", (lisp_proc *) Lnumber_log2, 1},
    /*
     * type conversion routines 
     */
    {"fix", (lisp_proc *) Lfix, 1},
    {"logical", (lisp_proc *) Llogical, 1},
    {"rational", (lisp_proc *) Lrational, 1},
    {"float", (lisp_proc *) Afloat, 1},
    {"efloat", (lisp_proc *) Aefloat, 1},
    {"mkefloat", (lisp_proc *) Amkefloat, 2},
    {"mkefloat_by_ratio", (lisp_proc *) Amkefloat_by_ratio, 2},
    {"interval", (lisp_proc *) Linterval, 1},
    {"interval_by_ratio", (lisp_proc *) Ainterval, 2},
    {"complex", (lisp_proc *) Lcomplex, 1},
    {"mkcomplex", (lisp_proc *) Lmkcomplex, 2},
    {"ecomplex", (lisp_proc *) Lecomplex, 1},
    {"mkecomplex", (lisp_proc *) Lmkecomplex, 3},
    {"bfloat", (lisp_proc *) Abfloat, 1},
    {"befloat", (lisp_proc *) Abefloat, 1},
    {"mkbefloat", (lisp_proc *) Amkbefloat, 2},
    {"mkbefloat_by_ratio", (lisp_proc *) Amkbefloat_by_ratio, 2},
    {"binterval", (lisp_proc *) Lbinterval, 1},
    {"mkbinterval", (lisp_proc *) Lmkbinterval, 2},
    {"becomplex", (lisp_proc *) Lbecomplex, 1},
    {"mkbecomplex", (lisp_proc *) Lmkbecomplex, 3},
    {"modular", (lisp_proc *) Amodular, 1},
    {"mkmodular", (lisp_proc *) Amkmodular, 2},
    {"unmodular", (lisp_proc *) Aunmodular, 1},
    {"setmodulus", (lisp_proc *) Lset_modulus, 1},
    {"fmodular", (lisp_proc *) Afmodular, 1},
    {"set_fmodulus", (lisp_proc *) Lset_fmodulus, 1},
    /*
     * interval 
     */
    {"interval_lower", (lisp_proc *) Linterval_lower, 1},
    {"interval_upper", (lisp_proc *) Linterval_upper, 1},
    {"interval_mid", (lisp_proc *) Linterval_mid, 1},
    {"interval_width", (lisp_proc *) Linterval_width, 1},
    {"mkinterval", (lisp_proc *) Lmkinterval, 2},
    /*
     * efloat
     */
    {"efloat_error", (lisp_proc *) Aefloat_error, 1},
    {"efloat_value", (lisp_proc *) Aefloat_value, 1},
    {"ecomplex_error", (lisp_proc *) Lecomplex_error, 1},
    {"ecomplex_real", (lisp_proc *) Lecomplex_real, 1},
    {"ecomplex_imag", (lisp_proc *) Lecomplex_imag, 1},
    {"complex_real", (lisp_proc *) Lcomplex_real, 1},
    {"complex_imag", (lisp_proc *) Lcomplex_imag, 1},

    {"polynomial", (lisp_proc *) Lpolynomial, 1},
    {"tropical", (lisp_proc *) Ltropical, 1},
    {"dipoly", (lisp_proc *) Ldipoly, 1},
    {"setorder_lex", (lisp_proc *) Lsetorder_lex, 1},
    {"setorder_tlex", (lisp_proc *) Lsetorder_tlex, 1},
    {"setorder_revlex", (lisp_proc *) Lsetorder_revlex, 1},
    {"setorder_trevlex", (lisp_proc *) Lsetorder_trevlex, 1},
    {"setorder_gradings", (lisp_proc *) Lsetorder_gradings, 2},
    {"unknowns", (lisp_proc *) Aunknowns, 1},
    {"max_coeff", (lisp_proc *) Amax_coeff, 1},
    {"nterms", (lisp_proc *) Anterms, 1},
    /*
     */
#ifdef ASUKA
    {"+", (lisp_proc *) gplus2, 2},
    {"-", (lisp_proc *) gdiff2, 2},
    {"*", (lisp_proc *) gtimes2, 2},
    {"/", (lisp_proc *) gratio2, 2},
    {"//", (lisp_proc *) gremaind2, 2},
    {"^", (lisp_proc *) gexpt, 2},
    {"cross", (lisp_proc *) gcross, 2},
#else
    {"+", (lisp_proc *) plus2, 2},
    {"-", (lisp_proc *) diff2, 2},
    {"*", (lisp_proc *) times2, 2},
    {"/", (lisp_proc *) quot2, 2},
    {"//", (lisp_proc *) remaind2, 2},
    {"^", (lisp_proc *) Lexpt, 2},
#endif
#ifdef ASUKA
    {"=", (lisp_proc *) Lgeqn, 2},
#else
    {"=", (lisp_proc *) Leqn, 2},
#endif
    {"=:=", (lisp_proc *) Lappeq, 2},
    {"<", (lisp_proc *) Llessp, 2},
    {">", (lisp_proc *) Lgreaterp, 2},
    {">=", (lisp_proc *) Lgeq, 2},
    {"<=", (lisp_proc *) Lleq, 2},
    /*
     */
    {"bfloat_man", (lisp_proc *) Lbfloat_man, 1},
    {"bfloat_exp", (lisp_proc *) Lbfloat_exp, 1},
    {"leftshift", (lisp_proc *) Lleftshift, 2},
    {"gen_prime", (lisp_proc *) Lgen_prime, 2},
    {"next_prime", (lisp_proc *) Lnext_prime, 1},
    {"prime_before", (lisp_proc *) Lprime_before, 1},

    {"primep", (lisp_proc *) Lprimep, 1},

    {"DEBUG", (lisp_proc *) Adebug, -1},
    {"debug_on", (lisp_proc *) Adebug_on, 0},
    {"debug_off", (lisp_proc *) Adebug_off, 0},
    {"debug_level", (lisp_proc *) Adebug_level, 1},
    {"debug_depth", (lisp_proc *) Adebug_depth, 1},
    {"debug", (lisp_proc *) Adebug, 1},
    {"debugn", (lisp_proc *) Adebugn, 1},
    {"write", (lisp_proc *) Awrite, 1},
    {"writen", (lisp_proc *) Awriten, 1},
    {"clear", (lisp_proc *) Aclear, -1},
    {"load", (lisp_proc *) Aload, 1},
    {"echo", (lisp_proc *) Aecho, 1},
    {"import", (lisp_proc *) Aimport, 1},
    {"size", (lisp_proc *) Asize, 1},
    {"mcoeff", (lisp_proc *) Amcoeff, 1},
    {"coeff", (lisp_proc *) Acoeff, 2},
    {"coeffx", (lisp_proc *) Acoeffx, 2},
    {"coeffy", (lisp_proc *) Acoeffy, 2},
    {"coeffn", (lisp_proc *) Acoeffn, 3},
    {"support", (lisp_proc *) Asupport, 1},
    {"support_ap", (lisp_proc *) Asupport_ap, 2},
    {"new_symbol1", (lisp_proc *) Anew_symbol1, 2},
    {"new_symbol2", (lisp_proc *) Anew_symbol2, 3},
    {"print_precision", (lisp_proc *) Lprint_precision, 1},
    {"Precision", (lisp_proc *) Aprecision, 1},

    {"setorder", (lisp_proc *) Lsetorder, 1},
    {"getorder", (lisp_proc *) Lgetorder, 0},
    {"getdipolyorder", (lisp_proc *) Lgetdipolyorder, 0},

    {"printf", (lisp_proc *) Lvprintf, 2},
    {"printf0", (lisp_proc *) Lprintf0, 1},
    {"printf1", (lisp_proc *) Lprintf1, 2},
    {"printf2", (lisp_proc *) Lprintf2, 3},

    {"is_cons", (lisp_proc *) Ais_cons, 1},
    {"is_fixnum", (lisp_proc *) Ais_fixnum, 1},
    {"is_hdrs", (lisp_proc *) Ais_hdrs, 1},
    {"is_reloc", (lisp_proc *) Ais_reloc, 1},
    {"is_symbol", (lisp_proc *) Ais_symbol, 1},
    {"is_numhdr", (lisp_proc *) Ais_numhdr, 1},
    {"is_vechdr", (lisp_proc *) Ais_lisp_vector, 1},
    {"is_binhdr", (lisp_proc *) Ais_binhdr, 1},
    {"is_number", (lisp_proc *) Ais_number, 1},
    {"is_immed", (lisp_proc *) Ais_immed, 1},
    {"is_logical", (lisp_proc *) Ais_logical, 1},
    {"is_bigint", (lisp_proc *) Ais_bigint, 1},
    {"is_modular", (lisp_proc *) Ais_modular, 1},
    {"is_float", (lisp_proc *) Ais_float, 1},
    {"is_rational", (lisp_proc *) Ais_rational, 1},
    {"is_ratio", (lisp_proc *) Ais_ratio, 1},
    {"is_complex", (lisp_proc *) Ais_complex, 1},
    {"is_bfloat", (lisp_proc *) Ais_bfloat, 1},
    {"is_efloat", (lisp_proc *) Ais_efloat, 1},
    {"is_interval", (lisp_proc *) Ais_interval, 1},
    {"is_ecomplex", (lisp_proc *) Ais_ecomplex, 1},
    {"is_befloat", (lisp_proc *) Ais_befloat, 1},
    {"is_binterval", (lisp_proc *) Ais_binterval, 1},
    {"is_becomplex", (lisp_proc *) Ais_becomplex, 1},
    {"is_bmodular", (lisp_proc *) Ais_bmodular, 1},
    {"is_qmodular", (lisp_proc *) Ais_qmodular, 1},
    {"is_fmodular", (lisp_proc *) Ais_fmodular, 1},
    {"is_poly", (lisp_proc *) Ais_poly, 1},
    {"is_tropical", (lisp_proc *) Ais_tropical, 1},
    {"is_dipoly", (lisp_proc *) Ais_dipoly, 1},
    {"is_rpoly", (lisp_proc *) Ais_rpoly, 1},
    {"is_bigneg", (lisp_proc *) Ais_bigneg, 1},
    {"is_integer", (lisp_proc *) Ais_integer, 1},
    {"is_fix", (lisp_proc *) Ais_fix, 1},
    {"is_neg", (lisp_proc *) Ais_neg, 1},
    {"is_app", (lisp_proc *) Ais_app, 1},
    {"is_log", (lisp_proc *) Ais_log, 1},
    {"is_eff", (lisp_proc *) Ais_eff, 1},
    {"is_cmp", (lisp_proc *) Ais_cmp, 1},
    {"is_big", (lisp_proc *) Ais_big, 1},
    {"is_vector", (lisp_proc *) Ais_vector, 1},
    {"is_matrix", (lisp_proc *) Ais_matrix, 1},
    {"is_array", (lisp_proc *) Ais_array, 1},
    {"is_string", (lisp_proc *) Ais_string, 1},
    {"is_iarray", (lisp_proc *) Ais_iarray, 1},
    {"is_sarray", (lisp_proc *) Ais_sarray, 1},
    {"is_darray", (lisp_proc *) Ais_darray, 1},
    {"is_iarray2", (lisp_proc *) Ais_iarray2, 1},
    {"is_sarray2", (lisp_proc *) Ais_sarray2, 1},
    {"is_darray2", (lisp_proc *) Ais_darray2, 1},
    {"is_iarray3", (lisp_proc *) Ais_iarray3, 1},
    {"is_sarray3", (lisp_proc *) Ais_sarray3, 1},
    {"is_darray3", (lisp_proc *) Ais_darray3, 1},
    {"is_package", (lisp_proc *) Ais_package, 1},
    {"recipiM", (lisp_proc *) recipiM, 2},
    {"spoly", (lisp_proc *) Aspoly, 2},
    {"reduce", (lisp_proc *) Areduce, 2},
    {"monic", (lisp_proc *) Amonic, 1},
    {"reduce_set", (lisp_proc *) Areduce_set, 1},
    {"full_reduce", (lisp_proc *) Afull_reduce, 2},
    {"dipoly_remainder", (lisp_proc *) Adipoly_remainder, 2},
    {"dipoly_quotient", (lisp_proc *) Adipoly_quotient, 2},
    {"dipoly_order", (lisp_proc *) Adipoly_order, 1},
    {"dipoly_pol", (lisp_proc *) Adipoly_pol, 1},
    {"is_dipoly_list", (lisp_proc *) Ais_dipoly_list, 1},
    {"lt_reducible", (lisp_proc *) Alt_reducible, 2},
    {"lt_divisible", (lisp_proc *) Alt_divisible, 2},
    {"have_nontrivial_gcd", (lisp_proc *) Ahave_nontrivial_gcd, 2},
    {"has_nontrivial_gcd", (lisp_proc *) Ahave_nontrivial_gcd, 2},
    {"lt_lcm_expvec", (lisp_proc *) Alt_lcm_expvec, 2},
    {"lt_lcm_ordervec", (lisp_proc *) Alt_lcm_ordervec, 2},
    {"sortPairset", (lisp_proc *) Asort_pairset, 1},
    {"mergePairset", (lisp_proc *) Amerge_pairset, 2},
    {"ws", (lisp_proc *) Aws, 1},

    {"expvec_lcm", (lisp_proc *) Aexpvec_lcm, 2},
    {"expvec_gcd", (lisp_proc *) Aexpvec_gcd, 2},
    {"expvec_div", (lisp_proc *) Aexpvec_div, 2},

    {"lt_gcd", (lisp_proc *) ALT_gcd, 2},
    {"lt_lcm", (lisp_proc *) ALT_lcm, 2},

    {"homogenize", (lisp_proc *) Ahomogenize, 2},
    {"ordervec", (lisp_proc *) Aordervec, 1},
    {"sugar_strategy", (lisp_proc *) Asugar_strategy, 1},
    {"is_sugar_strategy", (lisp_proc *) Ais_sugar_strategy, 0},
    {"append_lcm", (lisp_proc *) append_lcm, 1},

    {"pause", (lisp_proc *) Apause, 0},
    {"dump", (lisp_proc *) Adump, 1},

    {"makeid", (lisp_proc *) Amake_id, 2},
    {"makeuid", (lisp_proc *) Amake_uid, 2},
#ifdef UNIX
    {"system", (lisp_proc *) Asystem, 1},
#endif

    {"list2bnum", (lisp_proc *) Alist2bnum, 1},
    {"bnum2list", (lisp_proc *) Abnum2list, 1},
    {"compare", (lisp_proc *) Acompare, 2},
    {NULL, 0, 0}
};
