/*
 * vector.c             (C) Copyright 2015 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/prefix.c,v 2.9 2016/03/28 06:02:37 kako Exp $
 */




static char const rcsid[] =
    "$Id: prefix.c,v 2.9 2016/03/28 06:02:37 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "read.h"
#include "entries.h"
#include "misc.h"

/***********************************************************************/
/***********************************************************************/

Sex             new_prefix(void);
void            print_prefix_sub(Sex a, int l, putchar_fn * fn);
void            print_plus(Sex a, int l, putchar_fn * fn);
void            print_uplus(Sex a, int l, putchar_fn * fn);
void            print_minus(Sex a, int l, putchar_fn * fn);
void            print_uminus(Sex a, int l, putchar_fn * fn);
void            print_times(Sex a, int l, putchar_fn * fn);
void            print_quot(Sex a, int l, putchar_fn * fn);
void            print_remd(Sex a, int l, putchar_fn * fn);
void            print_expt(Sex a, int l, putchar_fn * fn);
void            print_args(Sex a, int l, putchar_fn * fn);

#if 0
Sex             prefix_diff(Sex a, Sex x);
Sex             prefix_diff_plus(Sex a, Sex x);
Sex             prefix_diff_uplus(Sex a, Sex x);
Sex             prefix_diff_minus(Sex a, Sex x);
Sex             prefix_diff_uminus(Sex a, Sex x);
Sex             prefix_diff_times(Sex a, Sex x);
Sex             prefix_diff_quot(Sex a, Sex x);
Sex             prefix_diff_remd(Sex a, Sex x);
Sex             prefix_diff_expt(Sex a, Sex x);
Sex             prefix_diff_args(Sex a, Sex x);
#else
Sex             prefix_diff(Sex a, Sex x);

Sex
prefix_diff(Sex a, Sex x)
{
    return aerror2("diff", a, x);
}

#endif

extern Sex      Adiff(Sex a, Sex x);
extern Sex      Lsimplify(Sex a);
/***********************************************************************/
/***********************************************************************/


Sex
new_prefix()
{
    Sex             p;

    p = getvector(TAG_VECTOR, T_PREFIX, sizeof(Prefix_Form));
    prefix_form(p) = nil;
    return p;
}

void
print_prefix(Sex a, putchar_fn * fn)
{
    a = prefix_form(a);
//lisp_printf("prefix: %p\n", a);
    print_prefix_sub(a, 0, fn);
}

void
print_prefix_sub(Sex a, int l, putchar_fn * fn)
{
    long            n,
                    k;

    stackcheck();
    push2(a, nil);
#define a stack[-1]
#define b stack[0]
    if (!is_cons(a)) {
	if (is_number(a)) {
	    print_sex(a, fn);
	} else {
	    if (l)
		fn('(');
	    print_sex(a, fn);
	    if (l)
		fn(')');
	}
	popv(2);
	return;
    }
    b = a;
    a = car(a);
    if (a == ASUKA_CHAR_OF('+')) {
	if (l)
	    fn('(');
	b = cdr(b);
	if (cdr(b) == nil) {
	    print_uplus(car(b), l + 1, fn);
	} else {
	    print_plus(b, l + 1, fn);
	}
	if (l)
	    fn(')');
    } else if (a == ASUKA_CHAR_OF('-')) {
	if (l)
	    fn('(');
	b = cdr(b);
	if (cdr(b) == nil) {
	    print_uminus(car(b), l + 1, fn);
	} else {
	    print_minus(b, l + 1, fn);
	}
	if (l)
	    fn(')');
    } else if (a == ASUKA_CHAR_OF('*')) {
	if (l)
	    fn('(');
	b = cdr(b);
	print_times(b, l + 1, fn);
	if (l)
	    fn(')');
    } else if (a == ASUKA_CHAR_OF('/')) {
	if (l)
	    fn('(');
	b = cdr(b);
	print_quot(b, l + 1, fn);
	if (l)
	    fn(')');
    } else if (a == ASUKA_CHAR_OF('%')) {
	if (l)
	    fn('(');
	b = cdr(b);
	print_remd(b, l + 1, fn);
	if (l)
	    fn(')');
    } else if (a == ASUKA_CHAR_OF('^')) {
	if (l)
	    fn('(');
	b = cdr(b);
	print_expt(b, l + 1, fn);
	if (l)
	    fn(')');
    } else {
	print_sex(a, fn);
	fn('(');
	a = cdr(b);
	print_args(a, 0, fn);
	fn(')');
    }
    popv(2);
#undef a
#undef b
}

void
print_plus(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first)
	    fn('+');
	stack[0] = cdr(a);
	print_prefix_sub(car(a), l + 1, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}

void
print_minus(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first)
	    fn('-');
	stack[0] = cdr(a);
	print_prefix_sub(car(a), l + 1, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}


void
print_uplus(Sex a, int l, putchar_fn * fn)
{
    stackcheck();
    push(a);
    fn('+');
    pop(a);
    print_prefix_sub(a, l + 1, fn);
}

void
print_uminus(Sex a, int l, putchar_fn * fn)
{
    stackcheck();
    push(a);
    fn('-');
    pop(a);
    print_prefix_sub(a, l + 1, fn);
}

void
print_times(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first)
	    fn('*');
	stack[0] = cdr(a);
	print_prefix_sub(car(a), l + 1, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}

void
print_quot(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first)
	    fn('/');
	stack[0] = cdr(a);
	print_prefix_sub(car(a), l + 1, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}


void
print_remd(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first)
	    fn('%');
	stack[0] = cdr(a);
	print_prefix_sub(car(a), l + 1, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}


void
print_expt(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first) {
	    fn('^');
	}
	stack[0] = cdr(a);
	print_prefix_sub(car(a), l + 1, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}


void
print_args(Sex a, int l, putchar_fn * fn)
{
    int             first;
    stackcheck();
    push(a);

    first = YES;
    while (a != nil) {
	if (!first)
	    fn(',');
	stack[0] = cdr(a);
	print_prefix_sub(car(a), 0, fn);
	a = stack[0];
	first = NO;
    }
    popv(1);
}
/***********************************************************************/
/***********************************************************************/
Sex
simplify_prefix(Sex a)
{
    Sex             b,
                    g;
    a = prefix_form(a);
    if (!is_cons(a))
	return a;
    b = car(a);
    if (!is_symbol(b))
	return aerror1("simplify", a);
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define p stack[-0]
    switch (pname(b)[0]) {
    case '+':
	if (b == ASUKA_CHAR_OF('+')) {
	    a = cdr(a);
	    if (a == nil)
		break;
	    p = Lsimplify(car(a));
	    a = cdr(a);
	    while (a != nil) {
		b = car(a);
		a = cdr(a);
		g = Lsimplify(b);
		p = gplus2(g, p);
	    }
	    g = p;
	    popv(3);
	    return g;
	}
	break;
    case '-':
	if (b == ASUKA_CHAR_OF('-')) {
	    a = cdr(a);
	    if (a == nil)
		break;
	    p = Lsimplify(car(a));
	    if (cdr(a) == nil) {
		g = gnegate(p);
	    } else {
		a = cdr(a);
		while (a != nil) {
		    b = car(a);
		    a = cdr(a);
		    g = Lsimplify(b);
		    p = gdiff2(p, g);
		}
		g = p;
	    }
	    popv(3);
	    return g;
	}
	break;
    case '*':
	if (b == ASUKA_CHAR_OF('*')) {
	    a = cdr(a);
	    if (a == nil)
		break;
	    p = Lsimplify(car(a));
	    a = cdr(a);
	    while (a != nil) {
		b = car(a);
		a = cdr(a);
		g = Lsimplify(b);
		p = gtimes2(p, g);
	    }
	    g = p;
	    popv(3);
	    return g;
	}
	break;
    case '/':
	if (b == ASUKA_CHAR_OF('/')) {
	    a = cdr(a);
	    if (a == nil)
		break;
	    p = Lsimplify(car(a));
	    a = cdr(a);
	    while (a != nil) {
		b = car(a);
		a = cdr(a);
		g = Lsimplify(b);
		p = gquot2(p, g);
	    }
	    g = p;
	    popv(3);
	    return g;
	}
	break;
    case '%':
	if (b == ASUKA_CHAR_OF('%')) {
	    a = cdr(a);
	    if (a == nil)
		break;
	    p = Lsimplify(car(a));
	    a = cdr(a);
	    while (a != nil) {
		b = car(a);
		a = cdr(a);
		g = Lsimplify(b);
		p = gremaind2(p, g);
	    }
	    g = p;
	    popv(3);
	    return g;
	}
	break;
    case '^':
	if (b == ASUKA_CHAR_OF('^')) {
	    a = cdr(a);
	    if (a == nil)
		break;
	    p = Lsimplify(car(a));
	    a = cdr(a);
	    while (a != nil) {
		b = car(a);
		a = cdr(a);
		g = Lsimplify(b);
		p = gexpt(p, g);
	    }
	    g = p;
	    popv(3);
	    return g;
	}
	break;
    case 's':
    case 'c':
    case 't':
    case 'a':
    case 'e':
    case 'l':
	break;
    default:
	break;
    }
    popv(3);
#undef a
#undef b
#undef p
    return aerror1("simplify", a);
}

/***********************************************************************/
/***********************************************************************/
Sex
gplus2PP(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('+'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
gdiff2PP(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('-'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
gtimes2PP(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('*'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
gquot2PP(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('/'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
gratio2PP(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('/'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
grem2PP(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('/'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
gexpt2Pi(Sex a, Sex b)
{
    Sex             p;
    a = Llist3(ASUKA_CHAR_OF('^'), a, b);
    push(a);
    p = new_prefix();
    pop(a);
    prefix_form(p) = a;
    return p;
}

Sex
ggcd2PP(Sex a, Sex b)
{
	return aerror2("gcd2", a, b);
}

/***********************************************************************/
/***********************************************************************/
#if 0
Sex
prefix_diff(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push3(a, nil, x);
#define a stack[-2]
#define b stack[-1]
#define x stack[0]
    if (!is_cons(a)) {
	r = Adiff(a, x);
	popv(2);
	return r;
    }
    b = a;
    a = car(a);
    if (a == ASUKA_CHAR_OF('+')) {
	b = cdr(b);
	if (cdr(b) == nil) {
	    prefix_diff_uplus(car(b), x);
	} else {
	    prefix_diff_plus(b, x);
	}
    } else if (a == ASUKA_CHAR_OF('-')) {
	b = cdr(b);
	if (cdr(b) == nil) {
	    prefix_diff_uminus(car(b), x);
	} else {
	    prefix_diff_minus(b, x);
	}
    } else if (a == ASUKA_CHAR_OF('*')) {
	b = cdr(b);
	prefix_diff_times(b, x);
    } else if (a == ASUKA_CHAR_OF('/')) {
	b = cdr(b);
	prefix_diff_quot(b, x);
    } else if (a == ASUKA_CHAR_OF('%')) {
	b = cdr(b);
	prefix_diff_remd(b, x);
    } else if (a == ASUKA_CHAR_OF('^')) {
	b = cdr(b);
	prefix_diff_expt(b, x);
    } else {
	r = prefix_diff_fn(a, x);
    }
    popv(3);
#undef a
#undef b
#undef x
    return r;
}

Sex
prefix_diff_plus(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	prefix_diff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}

Sex
prefix_diff_minus(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	Adiff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}


Sex
prefix_diff_uplus(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);
    pop(a);
    Adiff(a, x);
    return r;
}

Sex
prefix_diff_uminus(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);
    pop(a);
    Adiff(a, x);
    return r;
}

Sex
prefix_diff_times(Sex a, Sex x)
{
    Sex             r;
    int             first;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	Adiff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}

Sex
prefix_diff_quot(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	Adiff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}


Sex
prefix_diff_remd(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	Adiff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}


Sex
prefix_diff_expt(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	Adiff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}


Sex
prefix_diff_fn(Sex a, Sex x)
{
    Sex             r;
    Sex             fn;


    fn = car(a);
    if (fn == sin_symbol) {
    } else if (fn == cos_symbol) {
    } else if (fn == tan_symbol) {
    } else if (fn == exp_symbol) {
    } else if (fn == log_symbol) {
    } else if (fn == log2_symbol) {
    } else if (fn == log10_symbol) {
    } else if (fn == sqrt_symbol) {
    } else if (fn == cbrt_symbol) {
    } else if (fn == sinh_symbol) {
    } else if (fn == cosh_symbol) {
    } else if (fn == tanh_symbol) {
    } else if (fn == asin_symbol) {
    } else if (fn == acos_symbol) {
    } else if (fn == atan_symbol) {
    } else if (fn == asinh_symbol) {
    } else if (fn == acosh_symbol) {
    } else if (fn == atanh_symbol) {
    } else
	return aerror2("diff", a, x);

}

Sex
prefix_diff_args(Sex a, Sex x)
{
    Sex             r;
    stackcheck();
    push(a);

    while (a != nil) {
	stack[0] = cdr(a);
	Adiff(car(a), x);
	a = stack[0];
    }
    popv(1);
    return r;
}

#endif
/***********************************************************************/
/***********************************************************************/
