/*
 * print.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/print.c,v 2.35 2016/09/21 05:47:21 kako Exp $
 */

static char const rcsid[] =
    "$Id: print.c,v 2.35 2016/09/21 05:47:21 kako Exp $";


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "config.h"
#include "kctype.h"

#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "entries.h"
#include "print.h"

/*
 * #define PRINT_SHARP_ARRAY 
 */

long            print_flag;


FILE           *spool_file = NULL;
FILE           *debug_file = NULL;

long            column = 0,
    line_length = 80;
long            fcolumn = 0;

static void     lisp_vprintf(putchar_fn *, char *, Sex, va_list);
static Sex      scan_format(char *, va_list);

Sex             Lpdebug(Sex a);

void
debug_printf(char *fmt0, ...)
{
    Sex             s;
    va_list         ap,
                    ap2;
    va_start(ap, fmt0);
    va_copy(ap2, ap);
    s = scan_format(fmt0, ap2);
    va_end(ap2);
    lisp_vprintf(char_to_debug, fmt0, s, ap);
    va_end(ap);
}

void
lisp_printf(char *fmt0, ...)
{
    Sex             s;
    va_list         ap,
                    ap2;
    va_start(ap, fmt0);
    va_copy(ap2, ap);
    s = scan_format(fmt0, ap2);
    va_end(ap2);
    lisp_vprintf(char_to_stdout, fmt0, s, ap);
    va_end(ap);
}

static          Sex
scan_format(char *fmt, va_list a)
{
    unsigned long   _ulong;
    double          _double;
    Sex             _sex;
    Sex             s;
    long            ch;
    char           *t;

    push(nil);

    for (;; ++fmt) {
	while ((ch = *fmt) && ch != '%') {
	    fmt++;
	}
	if (!ch)
	    break;
      rflag:
	switch (*++fmt) {
	case ' ':
	case '-':
	case '+':
	    goto rflag;
	case '.':
	    fmt++;
	    while (isdigit(*fmt++));
	    --fmt;
	    goto rflag;
	case '0':
	    goto rflag;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    while (isdigit(*++fmt));
	    --fmt;
	    goto rflag;
	case 'c':
	    ch = va_arg(a, long);
	    continue;
	case 'u':
	case 'x':
	case 'd':
	case 'i':
	    _ulong = va_arg(a, long);
	    continue;
	case 'f':
	case 'g':
	    _double = va_arg(a, double);
	    continue;
	case 'a':
	case 'p':
	case 'U':
	case 'L':
	case 'A':
	case 'P':
	case 'T':
	    _sex = va_arg(a, Sex);
	    s = stack[0];
	    s = cons(_sex, s);
	    stack[0] = s;
	    continue;

	case 's':
	    t = va_arg(a, char *);
	    continue;;
	case '\0':
	    break;
	default:
	    continue;
	}
	break;
    }
    pop(s);
    s = Lnreverse(s);
    return s;
}



static void
lisp_vprintf(putchar_fn * fn, char *fmt0, Sex s, va_list a)
{
    char            buf[20];
    char           *fmt;
    long            n;
    long            ch;
    long            flags;
    long            size;
    long            sign;
    long            base;
    long            prec;
    long            width;
    unsigned long   _ulong;
    double          _double;
    Sex             _sex;
    char           *t;
    static char     digs[] = "0123456789abcdef";

    push(s);

    fmt = fmt0;
    for (;; ++fmt) {
	while ((ch = *fmt) && ch != '%') {
	    (*fn) (ch);
	    fmt++;
	}
	if (!ch)
	    break;
	flags = 0;
	sign = 0;
	prec = -1;
	width = 0;
      rflag:
	switch (*++fmt) {
	case ' ':
	    if (!sign)
		sign = ' ';
	    goto rflag;
	case '-':
	    flags |= P_LADJUST;
	    goto rflag;
	case '+':
	    sign = '+';
	    goto rflag;
	case '.':
	    n = 0;
	    fmt++;
	    while (isdigit(*fmt))
		n = 10 * n + *fmt++ - '0';
	    --fmt;
	    prec = n < 0 ? -1 : n;
	    goto rflag;
	case '0':
	    flags |= P_ZEROPAD;
	    goto rflag;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    n = 0;
	    do {
		n = 10 * n + *fmt - '0';
	    } while (isdigit(*++fmt));
	    width = n;
	    --fmt;
	    goto rflag;
	case 'c':
	    ch = va_arg(a, long);
	    (*fn) (ch);
	    continue;
	case 'u':
	    _ulong = va_arg(a, long);
	    base = 10;
	    goto nosign;
	case 'x':
	    _ulong = va_arg(a, long);
	    base = 16;
	    goto nosign;
	case 'd':
	    _ulong = va_arg(a, long);
	    base = 10;
	    goto number;
	case 'i':
	    _ulong = va_arg(a, long);
	    base = 10;
	  number:
	    if ((long) _ulong < 0) {
		sign = '-';
		_ulong = -(long) _ulong;
	    }
	  nosign:
	    if (prec >= 0)
		flags &= ~P_ZEROPAD;
	    t = buf;
	    if (_ulong != 0 || prec != 0) {
		do {
		    *t++ = digs[_ulong % base];
		    _ulong /= base;
		} while (_ulong);
	    }
	    size = t - buf;
	    if (sign)
		size++;
	    if ((flags & (P_LADJUST | P_ZEROPAD)) == 0 && width)
		for (n = size; n < width; n++)
		    (*fn) (' ');
	    if (sign)
		(*fn) (sign);
	    if ((flags & (P_LADJUST | P_ZEROPAD)) == P_ZEROPAD)
		for (n = size; n < width; n++)
		    (*fn) ('0');
	    while (--t >= buf)
		(*fn) (*t);
	    if (flags & P_LADJUST)
		for (n = size; n < width; n++)
		    (*fn) (' ');
	    continue;
	case 'f':
	    _double = va_arg(a, double);
	    sprintf(buf, "%.2f", _double);
	    t = buf;
	    while (*t)
		(*fn) (*t++);
	    continue;
	case 'g':
	    _double = va_arg(a, double);
	    sprintf(buf, "%.2g", _double);
	    t = buf;
	    while (*t)
		(*fn) (*t++);
	    continue;
	case 'a':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    stack[0] = cdr(s);
	    print_flag = 0;
	    print_sex(_sex, fn);
	    continue;
	case 'p':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    stack[0] = cdr(s);
	    print_flag = P_ESCAPED;
	    print_sex(_sex, fn);
	    continue;
	case 'U':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    stack[0] = cdr(s);
	    print_flag = P_ESCAPED | P_UPPERCASE;
	    print_sex(_sex, fn);
	    continue;
	case 'L':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    stack[0] = cdr(s);
	    print_flag = P_ESCAPED | P_DOWNCASE;
	    print_sex(_sex, fn);
	    continue;
	case 'A':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    stack[0] = cdr(s);
	    print_flag = 0;
	    hprint_sex(_sex, fn, trace_level);
	    continue;
	case 'P':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    stack[0] = cdr(s);
	    print_flag = P_ESCAPED;
	    hprint_sex(_sex, fn, trace_level);
	    continue;
	case 'T':
	    (void) va_arg(a, Sex);
	    s = stack[0];
	    _sex = car(s);
	    n = mkint(_sex);
	    stack[0] = cdr(s);
	    while (n-- > 0)
		(*fn) (' ');
	    continue;
	case 's':
	    t = va_arg(a, char *);
	    while (*t)
		(*fn) (*t++);
	    continue;;
	case '\0':
	    break;
	default:
	    (*fn) (*fmt);
	    continue;
	}
	break;
    }
    popv(1);
    return;
}


Sex
Lputch(Sex a)
{
    /*
     * send a single byte to stdout.
     */
    if (!is_fixnum(a))
	return aerror("putch");
    if (a != CHAR_EOF) {
	long            c = (int) mkint(a);
	char_to_stdout(c);
    }
    return a;
}

Sex
make_file_handle(FILE * f, long a, putchar_fn * fn)
{
    Sex             w;
    SFILE          *sf;
    w = getvector(TAG_BINARY, T_FILE, sizeof(File_Head));
    file_FILE(w) = f;
    sf = &file_SFILE(w);
    sf->data = NULL;
    sf->type = 2;
    sf->size = 0;
    sf->pos = 0;
    file_char(w) = a;
    file_putc(w) = fn;
    file_line(w) = 0;
    file_pos(w) = 0;
    return w;
}

Sex
make_string_handle(unsigned char *p, int size, int flag)
{
    SFILE          *sf;
    Sex             w;
    w = getvector(TAG_BINARY, T_FILE, sizeof(File_Head));
    file_FILE(w) = NULL;
    sf = &file_SFILE(w);
    sf->data = p;
    if (flag)
	sf->type = 3;
    else
	sf->type = 1;
    sf->size = size;
    sf->pos = 0;
    file_char(w) = NOT_CHAR;
    file_putc(w) = NULL;
    file_line(w) = 0;
    file_pos(w) = 0;
    return w;
}


static long     char_count;

void
char_to_count(long c)
{
    if (c == BLANK_CHAR)
	char_count++;
    else if ((c & NOT_CHAR) == 0)
	char_count++;
}

void
wchar_to_count(long c)
{
    if (c == BLANK_CHAR)
	char_count++;
    else if ((c & NOT_CHAR) == 0) {
	if (iskchar(c))
	    char_count += 2;
	else
	    char_count++;
    }
}


void
char_to_stdout(long c)
{
#ifdef Android
    char_to_androidout(c);
#else
    switch (c) {
    case BREAK_CHAR:
	if ((print_flag & P_NOLINEBREAK) == 0 && line_length
	    && column > line_length) {
	    c = '\n';
	} else
	    return;
	break;
    case BLANK_CHAR:
	if ((print_flag & P_NOLINEBREAK) == 0 && line_length
	    && column > line_length) {
	    c = '\n';
	} else {
	    c = ' ';
	}
	break;
    }
    if (c == '\n')
	column = 0;
    else
	column++;

    my_putwc(c, stdout);
    if (spool_file != NULL)
	my_putwc(c, spool_file);
#endif
}

void
char_to_stderr(long c)
{
    switch (c) {
    case BREAK_CHAR:
	return;
    case BLANK_CHAR:
	c = ' ';
	break;
    }
    my_putwc(c, stderr);
}

void
char_to_debug(long c)
{
    switch (c) {
    case BREAK_CHAR:
	return;
    case BLANK_CHAR:
	c = ' ';
	break;
    }
    my_putwc(c, debug_file == NULL ? stderr : debug_file);
}

static void
char_to_file(long c)
{
    switch (c) {
    case BREAK_CHAR:
	if ((print_flag & P_NOLINEBREAK) == 0 && line_length
	    && fcolumn > line_length) {
	    c = '\n';
	} else
	    return;
	break;
    case BLANK_CHAR:
	if ((print_flag & P_NOLINEBREAK) == 0 && line_length
	    && fcolumn > line_length) {
	    c = '\n';
	} else {
	    c = ' ';
	}
	break;
    }
    if (c == '\n')
	fcolumn = 0;
#ifdef Kanji
    else if (iskchar(c))
	fcolumn += 2;
#endif
    else
	fcolumn++;
    my_putwc(c, file_FILE(file_output));
}

#ifdef Android
static char     android_buffer[512];
static char    *android_pointer = &android_buffer[0];

void
char_to_androidout(long c)
{
    if (android_pointer < &android_buffer[510])
	*android_pointer++ = c;
    putchar(c);
}
void
init_androidout()
{
    android_pointer = android_buffer;
}
char           *
get_androidout()
{
    *android_pointer = 0;
#ifdef _LP64
    // printf("p = 0x%016lx\n, %016lx", (unsigned long)android_pointer,
    // (unsigned long)&android_buffer[0]);
#else
    // printf("p = 0x%08lx\n, %08lx", (unsigned long)android_pointer,
    // (unsigned long)&android_buffer[0]);
#endif
    return android_buffer;
}

void
char_to_androiderr(long c)
{
}
#endif

static void
char_to_lispout(long c)
{
    switch (c) {
    case '\0':
	error("print out null character");
	break;
    case BREAK_CHAR:
	if ((print_flag & P_NOLINEBREAK) == 0 && line_length
	    && column > line_length) {
	    c = '\n';
	} else
	    return;
	break;
    case BLANK_CHAR:
	if ((print_flag & P_NOLINEBREAK) == 0 && line_length
	    && column > line_length) {
	    c = '\n';
	} else {
	    c = ' ';
	}
	break;
    }
    if (c == '\n')
	column = 0;
#ifdef Kanji
    else if (iskchar(c))
	column += 2;
#endif
    else
	column++;
    my_putwc(c, file_FILE(lisp_output));
}

static          Sex
Lfiledate(Sex name)
{
    char            filename[FILENAME_MAX],
                    tt[32];
    long            len;
    char           *p;

    if (is_symbol(name)) {
	p = pname(name);
	len = symbol_length(name);
    } else if (is_string(name)) {
	p = &string_value(name, 0);
	len = string_length(name);
    } else
	return aerror("filep");
    if (len >= sizeof(filename))
	len = sizeof(filename);
    if (!file_exists(filename, p, (size_t) len, tt))
	return nil;
    name = make_string2(tt, 24);
    return name;
}

Sex
Lfilep(Sex name)
{
    name = Lfiledate(name);
    if (name != nil)
	name = true;
    return name;
}

Sex
Ltmpnam()
{

    return error("tmpnam: not implemented");
#if 0
    char           *s;
    Sex             r;
    s = tmpnam(NULL);
    if (s == NULL)
	return nil;
    r = make_string(s);
    return r;
#endif
}

Sex
Lmkstemp()
{
    int             s;
    Sex             r;
    FILE           *file;

#ifdef VisualC
    return error("mkstemp: not implemented");
#else
    s = mkstemp("/tmp/asuka_tmpXXXXXX");
    if (s == -1)
	return nil;
    file = fdopen(s, "w");
    if (file == NULL)
	return nil;
    r = make_file_handle(file, 0, char_to_lispout);
    return r;
#endif
}

Sex
Lopen(Sex name, Sex dir)
{
    FILE           *file;
    Sex             r;
    char           *fname;
    char           *dstr;
    char            filename[FILENAME_MAX];
    long            len;
    long            c;

    // lisp_printf("open %p", name);
    // lisp_printf(" with %p\n", dir);
    // printf("input = 0x%08x, dir = 0x%08x\n", input_symbol, dir);
    // printf("name = 0x%08x\n", name);

    if (dir == input_symbol) {
	dstr = "r";
	c = 1;
    } else if (dir == output_symbol) {
	dstr = "w";
	c = 0;
    } else if (dir == append_symbol) {
	dstr = "a";
	c = 0;
    } else if (dir == binput_symbol) {
	dstr = "rb";
	c = 1;
    } else if (dir == boutput_symbol) {
	dstr = "wb";
	c = 0;
    } else if (dir == bappend_symbol) {
	dstr = "ab";
	c = 0;
    } else {
	return aerror("open");
    }

    if (is_symbol(name)) {
	fname = pname(name);
	len = symbol_length(name);
    } else if (!is_string(name)) {
	return aerror("open");
    } else {
	fname = &string_value(name, 0);
	len = string_length(name);
    }
    if (len >= sizeof(filename))
	len = sizeof(filename);
    file = open_file(filename, fname, (size_t) len, dstr, NULL);
    if (file == NULL)
	return error1("open failed", name);
    if (c)
	r = make_file_handle(file, NOT_CHAR, 0);
    else
	r = make_file_handle(file, 0, char_to_lispout);
    return r;
}

Sex
Lopen_file(Sex name, Sex dir)
{
    FILE           *file;
    Sex             r;
    char           *fname;
    char           *dstr;
    char            filename[FILENAME_MAX];
    long            len;
    long            c;


    if (dir == input_symbol) {
	dstr = "r";
	c = 1;
    } else if (dir == output_symbol) {
	dstr = "w";
	c = 0;
    } else if (dir == append_symbol) {
	dstr = "a";
	c = 0;
    } else if (dir == binput_symbol) {
	dstr = "rb";
	c = 1;
    } else if (dir == boutput_symbol) {
	dstr = "wb";
	c = 0;
    } else if (dir == bappend_symbol) {
	dstr = "ab";
	c = 0;
    } else {
	return nil;
    }
    if (is_symbol(name)) {
	fname = pname(name);
	len = symbol_length(name);
    } else if (!is_string(name)) {
	return nil;
    } else {
	fname = &string_value(name, 0);
	len = string_length(name);
    }
    if (len >= sizeof(filename))
	len = sizeof(filename);
    file = open_file(filename, fname, (size_t) len, dstr, NULL);
    if (file == NULL)
	return nil;
    if (c)
	r = make_file_handle(file, NOT_CHAR, 0);
    else
	r = make_file_handle(file, 0, char_to_lispout);
    return r;
}

Sex
Lwrs(Sex a)
{
    Sex             old = lisp_output;
    if (a == nil)
	a = standard_output;
    if (a == old)
	return old;
    else if (!is_file(a))
	return aerror("wrs");
    else if (file_FILE(a) == NULL || file_putc(a) == 0)
	return aerror("wrs");
    file_pos(old) = column;
    lisp_output = a;
    column = file_pos(a);
    return old;
}

Sex
Lclose(Sex a)
{
    FILE           *f;
    static SFILE    snull = { NULL, NULL, 0, 0, 0 };
#ifdef USE_ZLIB
    SFILE          *sf;
#endif

    if (a == standard_input || a == standard_output
	|| a == standard_errout)
	return nil;
    else if (!is_file(a))
	return aerror("close");
    else if (file_type(a) == 0 || file_type(a) == 1 || file_type(a) == 3) {
#ifdef USE_ZLIB
	if (file_type(a) == 3) {
	    sf = &file_SFILE(a);
	    free(sf->data);
	}
#endif
	file_SFILE(a) = snull;
	return nil;
    } else if (file_FILE(a) == NULL)
	return nil;

    if (a == lisp_input) {
	lisp_input = standard_input;
	curchar = file_char(lisp_input);
    } else if (a == lisp_output) {
	lisp_output = standard_output;
	column = file_pos(lisp_output);
    }
    f = file_FILE(a);
    fclose(f);
    file_FILE(a) = NULL;
    return nil;
}

/*****************************************************************************/
/*
 * Printing.                                                            
 */
/*****************************************************************************/

void
print_string(char *p, putchar_fn * fn)
{
    char            c;
    while ((c = *p++) != 0)
	fn(c);
}

Sex
Lprint_precision(Sex a)
{
    long            old = print_bfloat_precision;
    if (a == nil)
	return mkfxat(old);
    if (!is_fixnum(a))
	return aerror("print-precision");
    print_precision = mkint(a);
    print_bfloat_precision = print_precision;
    if (print_precision > 16)
	print_precision = 15;
    return mkfxat(old);
}

static char     my_buff[128];
static char    *hexdig = "0123456789abcdef";

void
print_barray_int(long n, long f, putchar_fn * fn)
{
    char           *p;
    char            ch;
    sprintf(my_buff, "%ld", n);
    p = my_buff;
    while ((ch = *p++) != 0)
	fn(ch);
    if (!f)
	fn(BLANK_CHAR);
}
void
print_barray_long(long n, long f, putchar_fn * fn)
{
    char           *p;
    char            ch;
    sprintf(my_buff, "%ld", n);
    p = my_buff;
    while ((ch = *p++) != 0)
	fn(ch);
    if (!f)
	fn(BLANK_CHAR);
}
void
print_barray_float(float d, long f, putchar_fn * fn)
{
    char           *p;
    char            ch;
    sprintf(my_buff, "%f", d);
    p = my_buff;
    while ((ch = *p++) != 0)
	fn(ch);
    if (!f)
	fn(BLANK_CHAR);
}
void
print_barray_double(double d, long f, putchar_fn * fn)
{
    char           *p;
    char            ch;
    sprintf(my_buff, "%lf", d);
    p = my_buff;
    while ((ch = *p++) != 0)
	fn(ch);
    if (!f)
	fn(BLANK_CHAR);
}

void
print_barray(Sex u, long type, putchar_fn * fn)
{
    long            i,
                    n;
    int            *mi;
    long           *ml;
    float          *mf;
    double         *md;

    pop(u);
    switch (type) {
    case 0x50:
	mi = &iarray_value(u, 0);
	n = (binlen(u) - HDR) / sizeof(int);
	break;
    case 0x60:
	ml = &larray_value(u, 0);
	n = (binlen(u) - HDR) / sizeof(long);
	break;
    case 0x70:
	mf = &sarray_value(u, 0);
	n = (binlen(u) - HDR) / sizeof(float);
	break;
    case 0x80:
	md = &darray_value(u, 0);
	n = (binlen(u) - HDR) / sizeof(double);
	break;
    default:
	error("print binary array");
    }
    fn('[');
    for (i = 1; i <= n; i++) {
	switch (type) {
	case 0x50:
	    print_barray_int(*mi++, i == n, fn);
	    break;
	case 0x60:
	    print_barray_long(*ml++, i == n, fn);
	    break;
	case 0x70:
	    print_barray_float(*mf++, i == n, fn);
	    break;
	case 0x80:
	    print_barray_double(*md++, i == n, fn);
	    break;
	}
    }
    fn(']');
}

void
print_barray2(Sex u, long type, putchar_fn * fn)
{
    long            i,
                    j,
                    n,
                    m;
    int            *mi;
    long           *ml;
    float          *mf;
    double         *md;

    pop(u);
    n = barray_dimy(u);
    switch (type) {
    case 0x50:
	mi = &iarray2_value(u, 0);
	m = (binlen(u) - HDR - SEX) / sizeof(int);
	break;
    case 0x60:
	ml = &larray2_value(u, 0);
	m = (binlen(u) - HDR - SEX) / sizeof(long);
	break;
    case 0x70:
	mf = &sarray2_value(u, 0);
	m = (binlen(u) - HDR - SEX) / sizeof(float);
	break;
    case 0x80:
	md = &darray2_value(u, 0);
	m = (binlen(u) - HDR - SEX) / sizeof(double);
	break;
    default:
	error("print binary array");
    }
    m = m / n;
    fn('[');
    for (i = 1; i <= m; i++) {
	fn('[');
	for (j = 1; j <= n; j++) {
	    switch (type) {
	    case 0x50:
		print_barray_int(*mi++, j == n, fn);
		break;
	    case 0x60:
		print_barray_long(*ml++, j == n, fn);
		break;
	    case 0x70:
		print_barray_float(*mf++, j == n, fn);
		break;
	    case 0x80:
		print_barray_double(*md++, j == n, fn);
		break;
	    }
	}
	fn(']');
    }
    fn(']');
}

void
print_barray3(Sex u, long type, putchar_fn * fn)
{
    long            i,
                    j,
                    k,
                    n,
                    m,
                    l;
    int            *mi;
    long           *ml;
    float          *mf;
    double         *md;

    pop(u);
    n = barray_dimy(u);
    m = barray_dimz(u);
    switch (type) {
    case 0x50:
	mi = &iarray3_value(u, 0);
	l = (binlen(u) - HDR - 2 * SEX) / sizeof(int);
	break;
    case 0x60:
	ml = &larray3_value(u, 0);
	l = (binlen(u) - HDR - 2 * SEX) / sizeof(long);
	break;
    case 0x70:
	mf = &sarray3_value(u, 0);
	l = (binlen(u) - HDR - 2 * SEX) / sizeof(float);
	break;
    case 0x80:
	md = &darray3_value(u, 0);
	l = (binlen(u) - HDR - 2 * SEX) / sizeof(double);
	break;
    default:
	error("print binary array");
    }
    l = l / (n * m);
    fn('[');
    for (i = 1; i <= l; i++) {
	fn('[');
	for (j = 1; j <= n; j++) {
	    fn('[');
	    for (k = 1; k <= m; k++) {
		switch (type) {
		case 0x50:
		    print_barray_int(*mi++, k == m, fn);
		    break;
		case 0x60:
		    print_barray_long(*ml++, k == m, fn);
		    break;
		case 0x70:
		    print_barray_float(*mf++, k == m, fn);
		    break;
		case 0x80:
		    print_barray_double(*md++, k == m, fn);
		    break;
		}
	    }
	    fn(']');
	}
	fn(']');
    }
    fn(']');
}


void
print_sex(Sex u, putchar_fn * fn)
{
    Sex             w;
    extern long     run_mode;

    stackcheck();
    push(u);
    fn(BREAK_CHAR);
    pop(u);
    switch ((long) u & TAG_BITS) {
    case TAG_CONS:
	if (u == 0)
	    fatal_error("u should not 0");
	else if (car(u) == quote_symbol) {
	    push(u);
	    fn('\'');
	    print_sex(car(cdr(stack[0])), fn);
	    popv(1);
	    return;
	}
	push(u);
	/*
	 * ??? 2006.7.19 by f.kako 
	 */
	w = u;
	while (is_cons(w))
	    w = cdr(w);
	if (w != nil)
	    goto no_list;
#ifdef ASUKA
	if ((run_mode & RUN_MASK) == RUN_ASUKA) {
	    fn('{');
	    print_sex(car(stack[0]), fn);
	    w = stack[0];
	    while ((w = cdr(w)) != nil) {
		stack[0] = w;
		fn(',');
		fn(BLANK_CHAR);
		print_sex(car(stack[0]), fn);
		w = stack[0];
	    }
	    fn('}');
	} else {
	    fn('(');
	    print_sex(car(stack[0]), fn);
	    w = stack[0];
	    while ((w = cdr(w)) != nil) {
		stack[0] = w;
		fn(BLANK_CHAR);
		print_sex(car(stack[0]), fn);
		w = stack[0];
	    }
	    fn(')');
	}
#else
	fn('(');
	print_sex(car(stack[0]), fn);
	w = stack[0];
	while ((w = cdr(w)) != nil) {
	    stack[0] = w;
	    /*
	     * fn(' ');
	     */
	    fn(BLANK_CHAR);
	    print_sex(car(stack[0]), fn);
	    w = stack[0];
	}
	fn(')');
#endif
	popv(1);
	return;

      no_list:
	/*
	 * ??? 2006.7.19 by f.kako 
	 */

	fn('(');
	print_sex(car(stack[0]), fn);
	w = stack[0];
	while (is_cons(w = cdr(w))) {
	    stack[0] = w;
	    fn(BLANK_CHAR);
	    print_sex(car(stack[0]), fn);
	    w = stack[0];
	}
	if (w != nil) {
	    stack[0] = w;
	    fn(' ');
	    fn('.');
	    fn(' ');
	    print_sex(stack[0], fn);
	}
	popv(1);
	fn(')');
	return;


    case TAG_FIXNUM:
	sprintf(my_buff, "%ld", mkint(u));
	break;

    case TAG_HDRS:
	goto error_case;

    case TAG_BINARY:
	{
	    Tag             h = binhdr(u);
	    long            k,
	                    len = binlen(u);	/* 
						 * counts in bytes 
						 */
	    push(u);
	    switch (type_of_header(h)) {
	    case T_STRING:
		len -= CELL;
		if (print_flag & P_ESCAPED) {
		    fn('"');
		    for (k = 0; k < len; k++) {
			long            ch = string_value(stack[0], k);
#ifdef Kanji
			if (is2byte(ch)) {
			    k++;
			    ch = (ch << 8) + string_value(stack[0], k);
			}
#endif
			if (ch == '"' || ch == '\\') {
			    fn(ch);
			    fn(ch);
			} else if (iskcntrl(ch)) {
			    fn('\\');
			    fn(hexdig[(ch >> 4) & 0xf]);
			    fn(hexdig[ch & 0xf]);
			} else {
			    fn(ch);
			}
		    }
		    fn('"');
		} else if (print_flag & P_CESCAPED) {
		    fn('\\');
		    fn('"');
		    for (k = 0; k < len; k++) {
			long            ch = string_value(stack[0], k);
#ifdef Kanji
			if (is2byte(ch)) {
			    k++;
			    ch = (ch << 8) + string_value(stack[0], k);
			}
#endif
			if (ch == '"') {
			    fn('\\');
			    fn('"');
			    fn('\\');
			    fn('"');
			} else if (ch == '\\') {
			    fn('\\');
			    fn('\\');
			} else if (ch == '\n') {
			    fn('\\');
			    fn('\n');
			} else if (iskcntrl(ch)) {
			    fn('\\');
			    fn(hexdig[(ch >> 4) & 0xf]);
			    fn(hexdig[ch & 0xf]);
			} else {
			    fn(ch);
			}
		    }
		    fn('\\');
		    fn('"');
		} else {
		    for (k = 0; k < len; k++) {
			long            ch = string_value(stack[0], k);
#ifdef Kanji
			if (is2byte(ch)) {
			    k++;
			    ch = (ch << 8) + string_value(stack[0], k);
			}
			if ((print_flag & P_DOWNCASE) && iskupper(ch))
			    ch = toklower(ch);
			else if ((print_flag & P_UPPERCASE)
				 && isklower(ch))
			    ch = tokupper(ch);
#else
			if ((print_flag & P_DOWNCASE) && isupper(ch))
			    ch = tolower(ch);
			else if ((print_flag & P_UPPERCASE) && islower(ch))
			    ch = toupper(ch);
#endif
			fn(ch);
		    }
		}
		popv(1);
		return;

	    case T_IARRAY:
	    case T_SARRAY:
	    case T_DARRAY:
		pop(u);
		print_barray(u, barray_type(u, h), fn);
		return;
	    case T_IARRAY2:
	    case T_SARRAY2:
	    case T_DARRAY2:
		pop(u);
		print_barray2(u, barray_type(u, h), fn);
		return;
	    case T_IARRAY3:
	    case T_SARRAY3:
	    case T_DARRAY3:
		pop(u);
		print_barray3(u, barray_type(u, h), fn);
		return;
	    case T_FILE:
		pop(u);
		sprintf(my_buff, "#<file: %.8lx>",
			(unsigned long) file_FILE(u));
		goto print_my_buff;


	    default:
		pop(u);
		goto error_case;
	    }
	}
	popv(1);
	return;
    case TAG_VECTOR:
	{
	    Tag             h = vechdr(u);
	    long            k,
	                    len = veclen(u);	/* 
						 * counts in bytes 
						 */
	    long            bl;
	    switch (type_of_header(h)) {
#ifdef ASUKA
	    case T_VAR:
		print_variable(u, fn);
		return;
	    case T_ALGNUM:
		print_algnum(u, fn);
		return;
	    case T_TRNNUM:
		print_const(u, fn);
		return;
	    case T_POLY:
		print_polynomial(u, fn);
		return;
	    case T_DIPOLY:
		print_dipoly(u, fn);
		return;
	    case T_RPOLY:
		print_rpoly(u, fn);
		return;
	    case T_VECT:
		print_vector(u, fn);
		return;
	    case T_MATRIX:
		print_matrix(u, fn);
		return;
	    case T_PREFIX:
		print_prefix(u, fn);
		return;
#endif



	    case T_SIMPLE:
	    case T_ARRAY:
		if (type_of_header(h) == T_SIMPLE)
		    bl = '[';
		else {
		    bl = '(';
		    fn('#');
		}
#ifdef PRINT_SHARP_ARRAY
		fn('a');
		fn('r');
		fn('r');
		fn('a');
		fn('y');
#else
		fn(bl);
		len = bnum_n(len);
		push(u);
		for (k = 0; k < len; k++) {
		    fn(BLANK_CHAR);
		    print_sex(vector_value(stack[0], k), fn);
		}
		popv(1);
		if (type_of_header(h) == T_SIMPLE)
		    fn(']');
		else
		    fn(')');
#endif
		return;
#if 0
	    case T_MATRIX:
		push(u);
		fn('#');
		fn('m');
		fn('a');
		fn('t');
		fn('r');
		fn('i');
		fn('x');
		fn('(');
		print_sex(matrix_dimx(stack[0]), fn);
		fn(',');
		print_sex(matrix_dimy(stack[0]), fn);
		fn(')');
		popv(1);
		return;
#endif

	    case T_V_ARRAY:
		{
		    long            size = v_array_size(u);
		    long            i;
		    push(u);
#ifdef ASUKA
		    if ((run_mode & RUN_MASK) == RUN_ASUKA) {
			fn('{');
		    } else {
			fn('(');
		    }
#else
		    fn('(');
#endif
		    u = v_array_value(stack[0], 0);
		    print_sex(u, fn);
		    for (i = 1; i < size; i++) {
#ifdef ASUKA
			if ((run_mode & RUN_MASK) == RUN_ASUKA) {
			    fn(',');
			}
#endif
			fn(BLANK_CHAR);
			u = v_array_value(stack[0], i);
			print_sex(u, fn);
		    }
#ifdef ASUKA
		    if ((run_mode & RUN_MASK) == RUN_ASUKA) {
			fn('}');
		    } else {
			fn(')');
		    }
#else
		    fn(')');
#endif
		    pop(u);
		}
		return;

	    case T_PACKAGE:
		sprintf(my_buff, "#<package: %.16lx>", (unsigned long) u);
		goto print_my_buff;
	    case T_HASH:
		sprintf(my_buff, "#<hash-table: %.16lx>",
			(unsigned long) u);
		goto print_my_buff;

	    default:
		/*
		 * push(u); 
		 */
		goto error_case;
	    }
	}
	/*
	 * popv(1); 
	 */
	return;

    case TAG_SYMBOL:
#ifdef ASUKA
	push(u);
	{
	    long            k,
	                    len = symbol_length(u);

	    for (k = 0; k < len; k++) {
		long            ch;
		u = stack[0];
		ch = pname(u)[k] & 0xff;
#ifdef Kanji
		if (is2byte(ch)) {
		    k++;
		    ch = (ch << 8) + (pname(u)[k] & 0xff);
		}
		if ((print_flag & P_DOWNCASE) && iskupper(ch))
		    ch = toklower(ch);
		else if ((print_flag & P_UPPERCASE) && isklower(ch))
		    ch = tokupper(ch);
#else
		if ((print_flag & P_DOWNCASE) && isupper(ch))
		    ch = tolower(ch);
		else if ((print_flag & P_UPPERCASE) && islower(ch))
		    ch = toupper(ch);
#endif
		fn(ch);
	    }
	}
#else
	push(u);
	{
	    long            k,
	                    len = symbol_length(u);
	    // printf("print_sex: %s, len = %d\n", pname(u), len);
	    if (print_flag & (P_ESCAPED | P_CESCAPED)) {
		k = 0;
		if (len > 0) {
		    long            ch;
		    ch = pname(u)[0];
#ifdef Kanji
		    if (is2byte(ch)) {
			ch = (ch << 8) + pname(u)[1];
			k = 2;
		    } else
#endif
			k = 1;
		    if ((print_flag & P_CESCAPED)
			&& (ch == '"' || ch == '\\')) {
			fn(ESCAPE_CHAR);
			fn(CESCAPE_CHAR);
			fn(ch);
		    } else if (!is_first(ch)) {
			fn(ESCAPE_CHAR);
			fn(ch);
#ifdef Kanji
		    } else if (iskcntrl(ch)) {
			fn('#');
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[(ch) & 0xf]);
#else
		    } else if (iscntrl(ch)) {
			fn('#');
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[(ch) & 0xf]);
#endif
		    } else
			fn(ch);
		}
		for (; k < len; k++) {
		    long            ch;
		    u = stack[0];
		    ch = pname(u)[k] & 0xff;
#ifdef Kanji
		    if (is2byte(ch)) {
			k++;
			ch = (ch << 8) + (pname(u)[k] & 0xff);
		    }
#endif
		    if ((print_flag & P_CESCAPED)
			&& (ch == '"' || ch == '\\')) {
			fn(ESCAPE_CHAR);
			fn(CESCAPE_CHAR);
			fn(ch);
		    } else if (!is_second(ch)) {
			fn(ESCAPE_CHAR);
			fn(ch);
#ifdef Kanji
		    } else if (iskcntrl(ch)) {
			fn('#');
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[(ch) & 0xf]);
#else
		    } else if (iscntrl(ch)) {
			fn('#');
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[(ch) & 0xf]);
#endif
		    } else {
			fn(ch);
		    }
		}
	    } else {
		for (k = 0; k < len; k++) {
		    long            ch;
		    u = stack[0];
		    ch = pname(u)[k] & 0xff;
#ifdef Kanji
		    if (is2byte(ch)) {
			k++;
			ch = (ch << 8) + (pname(u)[k] & 0xff);
		    }
		    if ((print_flag & P_DOWNCASE) && iskupper(ch))
			ch = toklower(ch);
		    else if ((print_flag & P_UPPERCASE) && isklower(ch))
			ch = tokupper(ch);
#else
		    if ((print_flag & P_DOWNCASE) && isupper(ch))
			ch = tolower(ch);
		    else if ((print_flag & P_UPPERCASE) && islower(ch))
			ch = toupper(ch);
#endif
		    fn(ch);
		}
	    }
	}
#endif
	popv(1);
	return;

    case TAG_NUMBERS:
	switch (numtype_of(u)) {
	case N_LOGICAL:
	    print_bighex(u, fn);
	    return;
	case N_BIGNUM:
	    print_bignum(u, fn);
	    return;
	case N_MODULAR:
	    print_modular(u, fn);
	    return;
	case N_FLOAT:
	    sprintf(my_buff, "%#.*g", print_precision, float_val(u));
#ifdef REDUCE
	    /*
	     * skipped by f.kako 20006.7.19 
	     */
	    {
		long            i = 0,
		    j,
		    k,
		    c;
		while ((c = my_buff[i]) != 0 && c != '.')
		    i++;
		if (c == 0)
		    break;
		j = i + 1;
		while ((c = my_buff[j]) != 'e' && c != 0)
		    j++;
		k = j - 1;
		if (k == i) {
		    long            l = j;
		    while (my_buff[l] != 0)
			l++;
		    while (l >= j) {
			my_buff[l + 1] = my_buff[l];
			l--;
		    }
		    my_buff[j++] = '0';
		} else {
		    i++;
		    while (k > i && my_buff[k] == '0')
			k--;
		    if (k != j - 1)
			strcpy(&my_buff[k + 1], &my_buff[j]);
		}
	    }
#endif
	    goto print_my_buff;
	case N_RATIONAL:
	    push(u);
	    /*
	     * fn('('); 
	     */
	    print_sex(rational_num(stack[0]), fn);
	    pop(u);
	    fn('/');
	    print_sex(rational_den(u), fn);
	    /*
	     * fn(')'); 
	     */
	    return;
	case N_COMPLEX:
	    push(u);
	    fn('#'), fn('c');
	    fn('(');
	    print_sex(complex_real(stack[0]), fn);
	    pop(u);
	    fn(' ');
	    print_sex(complex_imag(u), fn);
	    fn(')');
	    return;
	case N_EFLOAT:
	    sprintf(my_buff, "#e(%#.*g,%#.*g)",
		    print_precision, efloat_value(u),
		    print_precision, efloat_error(u));
	    goto print_my_buff;
	case N_INTERVAL:
	    sprintf(my_buff, "#v(%#.*g,%#.*g)",
		    print_precision, interval_lower(u),
		    print_precision, interval_upper(u));
	    goto print_my_buff;
	case N_ECOMPLEX:
	    sprintf(my_buff, "#ec(%#.*g,%#.*g,%#.*g)",
		    print_precision, ecomplex_real(u),
		    print_precision, ecomplex_imag(u),
		    print_precision, ecomplex_error(u));
	    goto print_my_buff;
	case N_BFLOAT:
	    print_bfloat(u, fn);
	    return;
	case N_BEFLOAT:
	    print_befloat(u, fn);
	    return;
	case N_BINTERVAL:
	    print_binterval(u, fn);
	    return;
	case N_BECOMPLEX:
	    print_becomplex(u, fn);
	    return;
	case N_BMODULAR:
	    print_bmodular(u, fn);
	    return;
	case N_FMODULAR:
	    print_fmodular(u, fn);
	    return;
	default:
	    break;
	}
	/*
	 * Else drop through to treat as an error 
	 */
    default:
      error_case:
	sprintf(my_buff, "?%.16lx?", (unsigned long) u);
	goto print_my_buff;
    }
  print_my_buff:
    {
	char           *p = my_buff;
	long            ch;
	while ((ch = *p++) != 0)
	    fn(ch);
    }
    return;
}

void
hprint_sex(Sex u, putchar_fn * fn, long lev)
{
    Sex             w;
    char           *p;

    lev--;
    if (lev < 0) {
	fn('.');
	fn('.');
	return;
    }
    stackcheck();
    if (line_length && column >= line_length)
	fn('\n');
    switch ((long) u & TAG_BITS) {
    case TAG_CONS:
	if (u == 0)
	    u = nil;
	else if (car(u) == quote_symbol) {
	    push(u);
	    fn('\'');
	    print_sex(car(cdr(stack[0])), fn);
	    popv(1);
	    return;
	}
	push(u);
	fn('(');
	hprint_sex(car(stack[0]), fn, lev);
	w = stack[0];
	while (is_cons(w = cdr(w))) {
	    stack[0] = w;
	    if (line_length && column >= line_length)
		fn('\n');
	    else
		fn(' ');
	    hprint_sex(car(stack[0]), fn, lev--);
	    w = stack[0];
	    if (lev < 0)
		break;
	}
	if (w != nil && lev >= 0) {
	    stack[0] = w;
	    fn(' ');
	    fn('.');
	    fn(' ');
	    hprint_sex(stack[0], fn, lev--);
	}
	if (lev < 0) {
	    fn('.');
	    fn('.');
	}
	popv(1);
	fn(')');
	return;


    case TAG_FIXNUM:
	sprintf(my_buff, "%ld", mkint(u));
	break;

    case TAG_HDRS:
	goto error_case;

    case TAG_BINARY:
	{
	    Tag             h = binhdr(u);
	    long            k,
	                    len = binlen(u);	/* 
						 * counts in bytes 
						 */
	    push(u);
	    switch (type_of_header(h)) {
	    case T_STRING:
		len -= CELL;
		/*
		 * I will write out the fast, easy, common case here
		 */
		fn('"');
		for (k = 0; k < len; k++) {
		    long            ch = string_value(stack[0], k);
#ifdef Kanji
		    if (is2byte(ch)) {
			k++;
			ch = (ch << 8) + string_value(stack[0], k);
		    }
#endif
		    if (ch == '"') {
			fn('"');
			fn('"');
		    } else if (iskcntrl(ch)) {
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[ch & 0xf]);
		    } else {
			fn(ch);
		    }
		}
		popv(1);
		fn('"');
		return;

	    case T_IARRAY:
	    case T_IARRAY2:
	    case T_IARRAY3:
		pop(u);
		sprintf(my_buff, "#<intvec%ld: %.8llx>", h >> 8, u);
		goto print_my_buff;
	    case T_SARRAY:
	    case T_SARRAY2:
	    case T_SARRAY3:
		pop(u);
		sprintf(my_buff, "#<sfloatvec%ld: %.16llx>", h >> 8, u);
		goto print_my_buff;
	    case T_DARRAY:
	    case T_DARRAY2:
	    case T_DARRAY3:
		pop(u);
		sprintf(my_buff, "#<dfloatvec%ld: %.16llx>", h >> 8, u);
		goto print_my_buff;
	    case T_FILE:
		pop(u);
		sprintf(my_buff, "#<file: %.16lx>",
			(unsigned long) file_FILE(u));
		goto print_my_buff;


	    default:
		pop(u);
		goto error_case;
	    }
	}
	popv(1);
	return;
    case TAG_VECTOR:
	{
	    Tag             h = vechdr(u);
	    long            k,
	                    len = veclen(u);	/* 
						 * counts in bytes 
						 */
	    long            bl;
	    switch (type_of_header(h)) {
#ifdef ASUKA
	    case T_VAR:
		print_variable(u, fn);
		return;
	    case T_ALGNUM:
		print_algnum(u, fn);
		return;
	    case T_TRNNUM:
		print_const(u, fn);
		return;
	    case T_POLY:
		print_polynomial(u, fn);
		return;
	    case T_DIPOLY:
		print_dipoly(u, fn);
		return;
	    case T_RPOLY:
		print_rpoly(u, fn);
		return;
	    case T_VECT:
		// print_vector(u, fn);
		// return;
		sprintf(my_buff, "#<vector: %.8lx>", (unsigned long) u);
		goto print_my_buff;
	    case T_MATRIX:
		// print_matrix(u, fn);
		// return;
		sprintf(my_buff, "#<matrix: %.8lx>", (unsigned long) u);
		goto print_my_buff;
	    case T_PREFIX:
		print_prefix(u, fn);
		return;
#endif

	    case T_SIMPLE:
	    case T_ARRAY:
		if (type_of_header(h) == T_SIMPLE)
		    bl = '[';
		else {
		    bl = '(';
		    fn('#');
		}
		fn(bl);
		len = bnum_n(len);
		push(u);
		for (k = 0; k < len; k++) {
		    if (line_length && column >= line_length)
			fn('\n');
		    else
			fn(' ');

		    hprint_sex(vector_value(stack[0], k), fn, lev--);
		    if (lev < 0)
			break;
		}
		if (lev < 0) {
		    fn('.');
		    fn('.');
		}
		popv(1);
		if (type_of_header(h) == T_SIMPLE)
		    fn(']');
		else
		    fn(')');
		return;
	    case T_V_ARRAY:
		sprintf(my_buff, "#<variable array: %.8lx>",
			(unsigned long) u);
		goto print_my_buff;
	    case T_PACKAGE:
		sprintf(my_buff, "#<package: %.8lx>", (unsigned long) u);
		goto print_my_buff;

	    case T_HASH:
		sprintf(my_buff, "#<hashtable: %.8lx>", (unsigned long) u);
		goto print_my_buff;
	    default:
		goto error_case;
	    }
	}
	return;

    case TAG_SYMBOL:
	push(u);
	p = pname(u);
	{
	    long            k,
	                    len = symbol_length(u);
	    if (len > 0) {
		long            ch = p[0];
#ifdef Kanji
		if (is2byte(ch)) {
		    ch = (ch << 8) + p[1];
		    k = 2;
		} else
#endif
		    k = 1;
#ifdef Kanji
		if (!is_first(ch)) {
		    fn(ESCAPE_CHAR);
		    fn(ch);
		} else if (iskcntrl(ch)) {
		    fn('#');
		    fn('\\');
		    fn(hexdig[(ch >> 4) & 0xf]);
		    fn(hexdig[(ch) & 0xf]);
#else
		if (iscntrl(ch)) {
		    fn('#');
		    fn('\\');
		    fn(hexdig[(ch >> 4) & 0xf]);
		    fn(hexdig[(ch) & 0xf]);
#endif
		} else
		    fn(ch);
		for (; k < len; k++) {
		    long            ch;
		    u = stack[0];
		    p = pname(u);
		    ch = p[k];
#ifdef Kanji
		    if (is2byte(ch)) {
			k++;
			ch = (ch << 8) + p[k];
		    }
#endif
#ifdef Kanji
		    if (!is_second(ch)) {
			fn(ESCAPE_CHAR);
			fn(ch);
		    } else if (iskcntrl(ch)) {
			fn('#');
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[(ch) & 0xf]);
#else
		    if (iscntrl(ch)) {
			fn('#');
			fn('\\');
			fn(hexdig[(ch >> 4) & 0xf]);
			fn(hexdig[(ch) & 0xf]);
#endif
		    } else {
			fn(ch);
		    }
		}
	    }
	}
	popv(1);
	return;

    case TAG_NUMBERS:
	switch (numtype_of(u)) {
	case N_LOGICAL:
	    sprintf(my_buff, "0x%lx", logical_value(u)[0]);
	    goto print_my_buff;
	case N_BIGNUM:
	    print_bignum(u, fn);
	    return;
	case N_MODULAR:
	    print_modular(u, fn);
	    return;
	case N_FLOAT:
	    sprintf(my_buff, "%#.*g", print_precision, float_val(u));
	    goto print_my_buff;
	case N_RATIONAL:
	    push(u);
	    fn('(');
	    hprint_sex(rational_num(stack[0]), fn, lev);
	    pop(u);
	    fn('/');
	    hprint_sex(rational_den(u), fn, lev);
	    fn(')');
	    return;
	case N_COMPLEX:
	    push(u);
	    fn('#'), fn('c');
	    fn('(');
	    hprint_sex(complex_real(stack[0]), fn, lev);
	    pop(u);
	    fn(' ');
	    hprint_sex(complex_imag(u), fn, lev);
	    fn(')');
	    return;
	case N_EFLOAT:
	    sprintf(my_buff, "#e(%g,%g)", efloat_value(u),
		    efloat_error(u));
	    goto print_my_buff;
	case N_INTERVAL:
	    sprintf(my_buff, "#v(%g,%g)", interval_lower(u),
		    interval_upper(u));
	    goto print_my_buff;
	case N_ECOMPLEX:
	    sprintf(my_buff, "#ec(%#.*g,%#.*g,%#.*g)",
		    print_precision, ecomplex_real(u),
		    print_precision, ecomplex_imag(u),
		    print_precision, ecomplex_error(u));
	    goto print_my_buff;
	case N_BFLOAT:
	    print_bfloat(u, fn);
	    return;
	case N_BEFLOAT:
	    print_befloat(u, fn);
	    return;
	case N_BINTERVAL:
	    print_binterval(u, fn);
	    return;
	case N_BECOMPLEX:
	    print_becomplex(u, fn);
	    return;
	case N_FMODULAR:
	    print_fmodular(u, fn);
	    return;
	case N_BMODULAR:
	    print_bmodular(u, fn);
	    return;
	default:
	    break;
	}
    default:
      error_case:
	sprintf(my_buff, "?%.8lx?", (unsigned long) u);
	goto print_my_buff;
    }
  print_my_buff:
    {
	char           *p = my_buff;
	long            ch;
	while ((ch = *p++) != 0)
	    fn(ch);
    }
    return;
}

Sex
prin(Sex u)
{
    print_flag = P_ESCAPED;
    stackcheck();
    push(u);
    print_sex(u, file_putc(lisp_output));
    pop(u);
    return u;
}

void
prin_to_stdout(Sex u)
{
    print_flag = P_ESCAPED;
    print_sex(u, char_to_stdout);
}

void
prin_to_stderr(Sex u)
{
    print_flag = P_ESCAPED;
    print_sex(u, char_to_stderr);
}

void
prin_to_debug(Sex u)
{
    print_flag = P_ESCAPED;
    print_sex(u, char_to_debug);
}

Sex
princ(Sex u)
{
    print_flag = 0;
    stackcheck();
    push(u);
    print_sex(u, file_putc(lisp_output));
    pop(u);
    return u;
}

Sex
print(Sex u)
{
    stackcheck();
    push(u);
    print_flag = P_ESCAPED;
    print_sex(u, file_putc(lisp_output));
    file_putc(lisp_output) ('\n');
    pop(u);
    return u;
}

Sex
printc(Sex u)
{
    stackcheck();
    push(u);
    print_flag = 0;
    print_sex(u, file_putc(lisp_output));
    file_putc(lisp_output) ('\n');
    pop(u);
    return u;
}

Sex
terpri(void)
{
    file_putc(lisp_output) ('\n');
    return nil;
}

static void
char_to_list(long c)
{
    Sex             k;
    long            ch;

    if (c == BLANK_CHAR)
	c = ' ';
    else if (NOT_CHAR & c)
	return;
#ifdef Kanji
    ch = (c >> 8) & 0xff;
    if (ch) {
	k = CHAR_OF(ch);
	if (k == nil) {
	    lisp_printf("character code(%d) not installed\n", ch);
	}
	k = cons(k, io_workspace);
	io_workspace = k;
    }
    k = CHAR_OF(c & 0xff);
    if (k == nil) {
	lisp_printf("character code(%d) not installed\n", c);
    }
#else
    k = CHAR_OF(c & 0xff);
    if (k == nil) {
	lisp_printf("character code(%d) not installed\n", c);
    }
#endif
    k = cons(k, io_workspace);
    io_workspace = k;
}


static void
wchar_to_list(long c)
{
    Sex             k;

    if (c == BLANK_CHAR)
	c = ' ';
    if (NOT_CHAR & c)
	return;
#ifdef Kanji
    if (iskchar(c)) {
	boffo[0] = c >> 8;
	boffo[1] = c;
	boffo[2] = 0;
	k = intern(boffo, 2);
    } else {
	k = CHAR_OF(c & 0xff);
	if (k == nil)
	    lisp_printf("character code(%d) not installed\n", c);
    }
#else
    k = CHAR_OF(c & 0xff);
    if (k == nil) {
	lisp_printf("character code(%d) not installed\n", c);
    }
#endif
    k = cons(k, io_workspace);
    io_workspace = k;
}

static void
int_to_list(long c)
{
    if (c == BLANK_CHAR)
	c = ' ';
    if (NOT_CHAR & c)
	return;
#ifdef Kanji
    if (iskchar(c)) {
	c &= 0xffff;
    } else {
	c &= 0xff;
    }
#else
    c &= 0xff;
#endif
    io_workspace = cons(mkfxat(c), io_workspace);
}

Sex
Lstrlen(Sex u)
{
    if (is_symbol(u))
	return mkfxat(symbol_length(u));
    else if (is_string(u))
	return mkfxat(string_length(u));

    char_count = 0;
    print_flag = P_NOLINEBREAK;
    print_sex(u, char_to_count);
    return mkfxat(char_count);
}

Sex
string2list(Sex u)
{
    io_workspace = nil;
    // lisp_printf("string2list : u=%p\n", u);
    // printf("string2list: u = %016llx\n", u);
    print_flag = 0;
    print_sex(u, int_to_list);
    u = io_workspace;
    io_workspace = nil;
    return Lnreverse(u);
}

static          Sex
explode(Sex u)
{
    io_workspace = nil;
    print_sex(u, wchar_to_list);
    u = io_workspace;
    io_workspace = nil;
    return Lnreverse(u);
}


static          Sex
explodew(Sex u)
{
    io_workspace = nil;
    print_sex(u, wchar_to_list);
    u = io_workspace;
    io_workspace = nil;
    return Lnreverse(u);
}

static void
code_to_list(long c)
{
    Sex             k;
    k = mkfxat((long) c);
    k = cons(k, io_workspace);
    io_workspace = k;
}

static          Sex
exploden(Sex u)
{
    io_workspace = nil;
    print_sex(u, code_to_list);
    u = io_workspace;
    io_workspace = nil;
    return Lnreverse(u);
}

Sex
Llinelength(Sex a)
{
    long            oll = line_length;
    long            n;
    if (a == nil)
	return mkfxat(oll);
    else if (!is_fixnum(a))
	return aerror("linelength");
    n = mkint(a);
    if (n >= 0) {
//	if (n < 10)
//	    n = 10;
	line_length = n;
    }
    return mkfxat(oll);
}

Sex
Lfprin(Sex f, Sex a)
{
    if (is_file(f) && file_putc(f) != NULL) {
	stackcheck();
	push(a);
	print_flag = P_ESCAPED;
	file_output = f;
	print_sex(a, file_putc(f));
	pop(a);
    }
    return a;
}

Sex
Lprin(Sex a)
{
    stackcheck();
    push(a);
    print_flag = P_ESCAPED;
    print_sex(a, file_putc(lisp_output));
    pop(a);
    return a;
}



Sex
Lposn()
{
    return mkfxat((long) column);
}

Sex
Llposn()
{
    return mkfxat(0);
}

Sex
Lpagelength(Sex a)
{
    return a;
}

Sex
Lfprinc(Sex f, Sex a)
{
    if (is_file(f) && file_putc(f) != NULL) {
	stackcheck();
	push(a);
	print_flag = 0;
	print_sex(a, file_putc(f));
	pop(a);
    }
    return a;
}

Sex
Lprinc(Sex a)
{
    stackcheck();
    push(a);
    print_flag = 0;
    print_sex(a, file_putc(lisp_output));
    pop(a);
    return a;
}

Sex
CC_debug_prin2(Sex a)
{
    return Ldebug_prin2(a);
}
Sex
CC_debug_prin2t(Sex a)
{
    return Ldebug_prin2t(a);
}

Sex
Ldebug_prin2(Sex a)
{
    stackcheck();
    push(a);
    print_flag = 0;
    print_sex(a, char_to_debug);
    pop(a);
    return a;
}

Sex
Ldebug_prin2t(Sex a)
{
    stackcheck();
    push(a);
    print_flag = 0;
    print_sex(a, char_to_debug);
    char_to_debug('\n');
    pop(a);
    return a;
}

Sex
Lprin2t(Sex a)
{
    stackcheck();
    push(a);
    print_flag = 0;
    print_sex(a, file_putc(lisp_output));
    file_putc(lisp_output) ('\n');
    pop(a);
    return a;
}

Sex
Lfprin2a(Sex f, Sex a)
{
    if (is_file(f) && file_putc(f) != NULL) {
	stackcheck();
	push(a);
	print_flag = P_NOLINEBREAK;
	print_sex(a, file_putc(f));
	pop(a);
    }
    return a;
}


Sex
Lprin2a(Sex a)
{
    stackcheck();
    push(a);
    print_flag = P_NOLINEBREAK;
    print_sex(a, file_putc(lisp_output));
    pop(a);
    return a;
}

Sex
Llprim(Sex a)
{
    stackcheck();
    push(a);
    print_flag = P_NOLINEBREAK;
    lisp_printf("Warnings :");
    a = stack[0];
    print_sex(a, file_putc(standard_output));
    file_putc(standard_output) ('\n');
    pop(a);
    return a;
}

Sex
Llprie(Sex a)
{
    stackcheck();
    push(a);
    print_flag = P_NOLINEBREAK;
    lisp_printf("Error :");
    a = stack[0];
    print_sex(a, file_putc(standard_output));
    file_putc(standard_output) ('\n');
    pop(a);
    return a;
}


Sex
Llengthc(Sex a)
{
    print_flag = P_NOLINEBREAK;
    char_count = 0;
    print_sex(a, wchar_to_count);
    return mkfxat(char_count);
}


void
bldmsg1(Sex f, Sex u, putchar_fn * fn)
{
    long            i,
                    len;
    long            c;
    long            n;

    stackcheck();
    push(u);
    if (is_symbol(f))
	f = make_string2(pname(f), symbol_length(f));
    if (!is_string(f)) {
	error1("invalid format string", f);
	popv(1);
	return;
    }
    push(f);
    len = string_length(f);
#define f stack[0]
#define u stack[-1]
    for (i = 0; i < len; i++) {
	for (; i < len; i++) {
	    c = string_value(f, i);
	    if (c == '%')
		break;
	    fn(c);
	}
	if (i >= len)
	    break;
	i++;
	c = string_value(f, i);
	switch (c) {
	case 'b':		/* 
				 * print spaces 
				 */
	    n = mkint(car(u));
	    u = cdr(u);
	    while (n > 0)
		fn(' ');
	    break;
	case 'o':
	case 'x':
	case 'w':
	case 'W':
	    print_flag = P_CESCAPED | P_NOLINEBREAK;
	    goto p_print;
	case 'p':
	    print_flag = 0 | P_NOLINEBREAK;
	  p_print:
	    print_sex(car(u), fn);
	    u = cdr(u);
	    break;
	case 'n':
	    fn('\n');
	    break;
	case 't':
	    fn('\t');
	    break;
	default:
	    fn(c);
	    break;
	}
    }
#undef f
#undef u
    popv(2);
}


void
lisp_vfprintf(Sex f, Sex u, putchar_fn * fn)
{
    long            i,
                    len;
    long            c;
    long            n;

    stackcheck();
    push(u);
    if (is_symbol(f))
	f = make_string2(pname(f), symbol_length(f));
    if (!is_string(f)) {
	error1("invalid format string", f);
	popv(1);
	return;
    }
    push(f);
    len = string_length(f);
#define f stack[0]
#define u stack[-1]
    for (i = 0; i < len; i++) {
	for (; i < len; i++) {
	    c = string_value(f, i);
	    if (c == '%' || c == '\\')
		break;
	    fn(c);
	}
	if (i >= len)
	    break;
	i++;
	if (c == '%') {
	    c = string_value(f, i);
	    switch (c) {
	    case 'U':
		print_flag = P_NOLINEBREAK | P_UPPERCASE;
		goto p_print;
	    case 'L':
		print_flag = P_NOLINEBREAK | P_DOWNCASE;
		goto p_print;
	    case 'a':
		print_flag = P_CESCAPED | P_NOLINEBREAK;
		goto p_print;
	    case 'p':
		print_flag = 0 | P_NOLINEBREAK;
	      p_print:
		print_sex(car(u), fn);
		u = cdr(u);
		break;
	    case 'T':
		n = mkint(car(u));
		u = cdr(u);
		while (n-- > 0)
		    fn(' ');
		break;
	    case 'n':
		fn('\n');
		break;
	    case 't':
		fn('\t');
		break;
	    default:
		fn(c);
		break;
	    }
	} else {
	    c = string_value(f, i);
	    switch (c) {
	    case 'n':
		fn('\n');
		break;
	    case 't':
		fn('\t');
		break;
	    default:
		fn(c);
		break;
	    }
	}
    }
#undef f
#undef u
    popv(2);
}

void
char_to_boffo(long c)
{
    if (c == BLANK_CHAR)
	c = ' ';
    else if (c == BREAK_CHAR)
	return;
    if (c == '\n')
	column = 0;
    else
	column++;

    if (char_count > BUFFER_SIZE)
	return;
    boffo[char_count++] = c;
}

Sex
Lbldmsg1(Sex f, Sex u)
{
    char_count = 0;
    bldmsg1(f, u, char_to_boffo);
    return make_string2(boffo, char_count);
}

Sex
printf_fn(Sex args, Sex env)
{
    Sex             f,
                    u;
    if (!is_cons(args))
	return nil;
    stackcheck();
    push4(args, env, nil, nil);
#define args stack[-3]
#define env stack[-2]
#define f stack[-1]
#define u stack[0]
    f = car(args);
    f = eval(f, env);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	w = car(args);
	w = eval(w, env);
	u = cons(w, u);
	args = cdr(args);
    }
    u = Lnreverse(u);
#undef u
#undef f
#undef env
#undef args
    f = stack[-1];
    u = stack[0];
    lisp_vfprintf(f, u, file_putc(lisp_output));
    return nil;
}

Sex
Lvprintf(Sex f, Sex u)
{
    lisp_vfprintf(f, u, file_putc(lisp_output));
    return nil;
}

Sex
Lvsprintf(Sex f, Sex u)
{
    char_count = 0;
    lisp_vfprintf(f, u, char_to_boffo);
    return make_string2(boffo, char_count);
}


Sex
Lvfprintf(Sex f, Sex fmt, Sex u)
{
    if (is_file(f) && file_putc(f) != NULL) {
	lisp_vfprintf(fmt, u, file_putc(f));
    }
    return nil;
}

Sex
Lprintf0(Sex f)
{
    lisp_vfprintf(f, nil, file_putc(lisp_output));
    return nil;
}

Sex
Lsprintf0(Sex f)
{
    char_count = 0;
    lisp_vfprintf(f, nil, char_to_boffo);
    return make_string2(boffo, char_count);
}


Sex
Lfprintf0(Sex f, Sex fmt)
{
    if (is_file(f) && file_putc(f) != NULL) {
	lisp_vfprintf(fmt, nil, file_putc(f));
    }
    return nil;
}

Sex
Lprintf1(Sex f, Sex u)
{
    push(f);
    u = Lncons(u);
    pop(f);
    lisp_vfprintf(f, u, file_putc(lisp_output));
    return nil;
}

Sex
Lsprintf1(Sex f, Sex u)
{
    char_count = 0;
    push(f);
    u = Lncons(u);
    pop(f);
    lisp_vfprintf(f, u, char_to_boffo);
    return make_string2(boffo, char_count);
}


Sex
Lfprintf1(Sex f, Sex fmt, Sex u)
{
    if (is_file(f) && file_putc(f) != NULL) {
	push2(f, fmt);
	u = Lncons(u);
	pop2(fmt, u);
	lisp_vfprintf(fmt, u, file_putc(f));
    }
    return nil;
}

Sex
Lprintf2(Sex f, Sex u, Sex v)
{
    push2(f, u);
    v = Lncons(v);
    pop(u);
    u = cons(u, v);
    pop(f);
    lisp_vfprintf(f, u, file_putc(lisp_output));
    return nil;
}



Sex
Lfprint(Sex f, Sex a)
{
    if (is_file(f) && file_putc(f) != NULL) {
	stackcheck();
	push(a);
	print_flag = P_ESCAPED;
	print_sex(a, file_putc(f));
	file_putc(f) ('\n');
	pop(a);
    }
    return a;
}

Sex
Lprint(Sex a)
{
    stackcheck();
    push(a);
    print_flag = P_ESCAPED;
    print_sex(a, file_putc(lisp_output));
    file_putc(lisp_output) ('\n');
    pop(a);
    return a;
}

Sex
Aprint(Sex a)
{
    stackcheck();
    push(a);
    print_flag = P_ESCAPED;
    if (a == true)
	a = true_symbol;
    else if (a == nil)
	a = false_symbol;
    print_sex(a, file_putc(lisp_output));
    file_putc(lisp_output) ('\n');
    pop(a);
    return a;
}

Sex
Lfprintc(Sex f, Sex a)
{
    if (is_file(f) && file_putc(f) != NULL) {
	stackcheck();
	push(a);
	print_flag = 0;
	print_sex(a, file_putc(f));
	file_putc(f) ('\n');
	pop(a);
    }
    return a;
}

Sex
Lprintc(Sex a)
{
    stackcheck();
    push(a);
    print_flag = 0;
    print_sex(a, file_putc(lisp_output));
    file_putc(lisp_output) ('\n');
    pop(a);
    return a;
}

Sex
Lphex(Sex a, Sex b)
{
    int             n = 0;
    if (is_fixnum(b))
	n = mkint(b);
    if (n < 0)
	n = 0;
    stackcheck();
    push(a);
    print_flag = 0;
    print_phex(a, n, file_putc(lisp_output));
    pop(a);
    return a;
}

Sex
Lpdebug(Sex a)
{
    if (is_fixnum(a))
	lisp_printf("fix number %p\n", a);
    else if (is_symbol(a))
	lisp_printf("symbol %p\n", a);
    else if (is_number(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    lisp_printf("logical %p\n", a);
	    break;
	case N_BIGNUM:
	    lisp_printf("bignum %p\n", a);
	    break;
	case N_MODULAR:
	    lisp_printf("modular %p\n", a);
	    break;
	case N_RATIONAL:
	    lisp_printf("rational %p\n", a);
	    break;
	case N_FLOAT:
	    lisp_printf("float %p\n", a);
	    break;
	case N_EFLOAT:
	    lisp_printf("efloat %p\n", a);
	    break;
	case N_COMPLEX:
	    lisp_printf("complex %p\n", a);
	    break;
	case N_INTERVAL:
	    lisp_printf("interval %p\n", a);
	    break;
	case N_BFLOAT:
	    lisp_printf("bfloat %p\n", a);
	    break;
	case N_ECOMPLEX:
	    lisp_printf("ecomplex %p\n", a);
	    break;
	case N_BEFLOAT:
	    lisp_printf("befloat %p\n", a);
	    break;
	case N_BECOMPLEX:
	    lisp_printf("becomplex %p\n", a);
	    break;
	case N_BINTERVAL:
	    lisp_printf("binterval %p\n", a);
	    break;
	case N_BMODULAR:
	    lisp_printf("bmodular %p\n", a);
	    break;
	case N_QMODULAR:
	    lisp_printf("qmodular %p\n", a);
	    break;
	case N_FMODULAR:
	    lisp_printf("fmodular %p\n", a);
	    break;
	default:
	    lisp_printf("unknown number %p\n", a);
	    break;
	}
    } else if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_NON:
	    lisp_printf("non %p\n", a);
	    break;
	case N_VAR:
	    lisp_printf("variable %p\n", a);
	    break;
	case N_ALGNUM:
	    lisp_printf("algnum %p\n", a);
	    break;
	case N_TRNNUM:
	    lisp_printf("trignum  %p\n", a);
	    break;
	case N_POLY:
	    lisp_printf("polynomial  %p\n", a);
	    break;
	case N_DIPOLY:
	    lisp_printf("distributed polynomial %p\n", a);
	    break;
	case N_ALGFUN:
	    lisp_printf("algebraic function %p\n", a);
	    break;
	case N_RPOLY:
	    lisp_printf("rational polynomial  %p\n", a);
	    break;
	case N_VECT:
	    lisp_printf("vector %p\n", a);
	    break;
	case N_MATRIX:
	    lisp_printf("matrix %p\n", a);
	    break;
	case N_TENSOR:
	    lisp_printf("tensor %p\n", a);
	    break;
	case N_PREF:
	    lisp_printf("prefix form %p\n", a);
	    break;
	case N_RELEXP:
	    lisp_printf("relatinal expression %p\n", a);
	    break;
	case N_SETEXP:
	    lisp_printf("set expression %p\n", a);
	    break;
	case N_LOGEXP:
	    lisp_printf("logical expression %p\n", a);
	    break;
	case N_TROPICAL:
	    lisp_printf("tropical number  %p\n", a);
	    break;
	default:
	    lisp_printf("unknown gtype %p\n", a);
	    break;

	}
    } else if (Llistp(a)) {
	lisp_printf("list %p\n", a);
    } else if (is_cons(a)) {
	lisp_printf("cons %p\n", a);
    } else
	lisp_printf("unknown type %p\n", a);
    return nil;
}

Sex
Lfterpri(Sex f)
{
    if (is_file(f) && file_putc(f) != NULL)
	file_putc(f) ('\n');
    return nil;
}

Sex
Lterpri()
{
    file_putc(lisp_output) ('\n');
    return nil;
}

Sex
Lfflush(Sex f)
{
    if (is_file(f) && file_putc(f) != NULL)
	fflush(file_FILE(f));
    return nil;
}

Sex
Lflush()
{
    fflush(file_FILE(lisp_output));
    return nil;
}

Sex
Lttab(Sex a)
{
    int             n;
    putchar_fn     *f = file_putc(lisp_output);
    if (!is_fixnum(a))
	return aerror("ttab");
    n = mkint(a);
    while (column < n)
	f(' ');
    return nil;
}

Sex
Lxtab(Sex a)
{
    int             n;
    putchar_fn     *f = file_putc(lisp_output);
    if (!is_fixnum(a))
	return aerror("xtab");
    n = mkint(a);
    while (n-- > 0)
	f(' ');
    return nil;
}

Sex
Lfeject(Sex f)
{
    if (is_file(f) && file_putc(f) != NULL) {
	file_putc(f) ('\f');
	file_pos(f) = 0;
    }
    return nil;
}

Sex
Leject()
{
    file_putc(lisp_output) ('\f');
    column = 0;
    return nil;
}

Sex
Lsprin2(Sex a)
{
    stackcheck();
    print_flag = 0;
    char_count = 0;
    print_sex(a, char_to_boffo);
    return intern(boffo, char_count);
}

Sex
Lexplodew(Sex a)
{
    print_flag = P_ESCAPED + P_NOLINEBREAK;
    a = explodew(a);
    return a;
}

Sex
Lexplode(Sex a)
{
    print_flag = P_ESCAPED + P_NOLINEBREAK;
    a = explode(a);
    return a;
}



Sex
Lexplodec(Sex a)
{
    print_flag = P_NOLINEBREAK;
    a = explode(a);
    return a;
}

Sex
Lexplodelc(Sex a)
{
    print_flag = P_DOWNCASE | P_NOLINEBREAK;
    a = explode(a);
    return a;
}

Sex
Lexplodeuc(Sex a)
{
    print_flag = P_UPPERCASE | P_NOLINEBREAK;
    a = explode(a);
    return a;
}

Sex
Lexploden(Sex a)
{
    print_flag = P_ESCAPED + P_NOLINEBREAK;
    a = exploden(a);
    return a;
}

Sex
Lexplodecn(Sex a)
{
    print_flag = P_NOLINEBREAK;
    a = exploden(a);
    return a;
}


setup_type const print_setup[] = {
    {"eject", (lisp_proc *) Leject, 0},
    {"filep", (lisp_proc *) Lfilep, 1},
    {"filedate", (lisp_proc *) Lfiledate, 1},
    {"flush", (lisp_proc *) Lflush, 0},
    {"lengthc", (lisp_proc *) Llengthc, 1},
    {"linelength", (lisp_proc *) Llinelength, 1},
    {"lposn", (lisp_proc *) Llposn, 0},
    {"open", (lisp_proc *) Lopen, 2},
    {"pagelength", (lisp_proc *) Lpagelength, 1},
    {"posn", (lisp_proc *) Lposn, 0},
    {"spaces", (lisp_proc *) Lxtab, 1},
    {"terpri", (lisp_proc *) Lterpri, 0},
    {"tmpnam", (lisp_proc *) Ltmpnam, 0},
    {"mkstemp", (lisp_proc *) Lmkstemp, 0},
    {"ttab", (lisp_proc *) Lttab, 1},
    {"wrs", (lisp_proc *) Lwrs, 1},
    {"xtab", (lisp_proc *) Lxtab, 1},
    {"close", (lisp_proc *) Lclose, 1},
    {"strlen", (lisp_proc *) Lstrlen, 1},
    {"string2list", (lisp_proc *) string2list, 1},
    {"explode", (lisp_proc *) Lexplode, 1},
    {"explodew", (lisp_proc *) Lexplodew, 1},
    {"explodec", (lisp_proc *) Lexplodec, 1},
    {"explode2", (lisp_proc *) Lexplodec, 1},
    {"exploden", (lisp_proc *) Lexploden, 1},
    {"explodecn", (lisp_proc *) Lexplodecn, 1},
    {"explode2n", (lisp_proc *) Lexplodecn, 1},
    {"explode2lc", (lisp_proc *) Lexplodelc, 1},
    {"explode2uc", (lisp_proc *) Lexplodeuc, 1},
    {"prin", (lisp_proc *) Lprin, 1},
    {"prin1", (lisp_proc *) Lprin, 1},
    {"princ", (lisp_proc *) Lprinc, 1},
    {"prin2", (lisp_proc *) Lprinc, 1},
    {"prin2t", (lisp_proc *) Lprin2t, 1},
    {"prin2a", (lisp_proc *) Lprin2a, 1},
    {"print", (lisp_proc *) Lprint, 1},
    {"printc", (lisp_proc *) Lprintc, 1},
    {"debug_prin2", (lisp_proc *) Ldebug_prin2, 1},
    {"debug_prin2t", (lisp_proc *) Ldebug_prin2t, 1},
    {"fflush", (lisp_proc *) Lfflush, 1},
    {"feject", (lisp_proc *) Lfeject, 1},
    {"fterpri", (lisp_proc *) Lfterpri, 1},
    {"fprin", (lisp_proc *) Lfprin, 2},
    {"fprin1", (lisp_proc *) Lfprin, 2},
    {"fprinc", (lisp_proc *) Lfprinc, 2},
    {"fprin2", (lisp_proc *) Lfprinc, 2},
    {"fprin2a", (lisp_proc *) Lfprin2a, 2},
    {"fprint", (lisp_proc *) Lfprint, 2},
    {"fprintc", (lisp_proc *) Lfprintc, 2},
    {"set-print-precision", (lisp_proc *) Lprint_precision, 1},
    {"putch", (lisp_proc *) Lputch, 1},
    {"lprim", (lisp_proc *) Llprim, 1},
    {"lprie", (lisp_proc *) Llprie, 1},
    {"maprin", (lisp_proc *) Lprin, 1},	/* 
					 * for initialization problem 
					 */
    {"terpri*", (lisp_proc *) Lterpri, 0},	/* 
						 * for initialization problem 
						 */
    {"vprintf", (lisp_proc *) Lvprintf, 2},
    {"vsprintf", (lisp_proc *) Lvsprintf, 2},
    {"vfprintf", (lisp_proc *) Lvfprintf, 3},
#ifndef GAL
    {"printf", (lisp_proc *) Lvprintf, 2},
#endif
    {"printf0", (lisp_proc *) Lprintf0, 1},
    {"sprintf0", (lisp_proc *) Lsprintf0, 1},
    {"fprintf0", (lisp_proc *) Lfprintf0, 2},
    {"printf1", (lisp_proc *) Lprintf1, 2},
    {"sprintf1", (lisp_proc *) Lsprintf1, 2},
    {"fprintf1", (lisp_proc *) Lfprintf1, 3},
    {"bldmsg1", (lisp_proc *) Lbldmsg1, 2},
    {"phex", (lisp_proc *) Lphex, 2},
    {"pdebug", (lisp_proc *) Lpdebug, 1},
    {NULL, 0, 0}
};
