/*
 * read.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/read.c,v 2.71 2018/06/11 09:08:26 kako Exp $
 */

static char const rcsid[] =
    "$Id: read.c,v 2.71 2018/06/11 09:08:26 kako Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#define READ_LOGICAL_AS_INTEGER

#if defined(FreeBSD) || defined(Darwin)
#include <fcntl.h>
#endif

#ifndef VisualC
#if GCCTYPE == GCC_DJGPP || GCCTYPE == GNU_WIN32
#include <unistd.h>
#endif
#endif

#include "config.h"
#include "kctype.h"

#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "read.h"
#include "arith.h"
#include "entries.h"
#ifdef ASUKA
#include "asuka.tab.h"
#endif

// #define DEBUG

/***************************************************************/
/***************************************************************/





static long     showtime = 0;

extern long     run_mode;
extern Sex      CC_prettyprint(Sex u);

#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#ifdef Android
extern double   my_atof(const char *p);
#endif

#define CTRL_C    3
#define CTRL_D    4



#ifndef FAPLIB_PATH
#define FAPLIB_PATH "/usr/home/kako/reduce3.7/"
#endif

#ifdef REDUCE
typedef int     pkg_init_fn(void);

#include "stubs.h"

extern Sex      CC_bloadKpackage(Sex);	/* 
					 * this function is defined
					 * in fisl.red 
					 */


typedef struct packages {
    char           *pkg_name;
    pkg_init_fn    *init_fn;
    int             already_loaded;
} packages;


packages        preloaded[] = {
    {"acfsf", init_acfsf, 0},
    {"alg", init_alg, 0},
    {"algint", init_algint, 0},
    {"applysym", init_applysym, 0},
    {"arith", init_arith, 0},
    {"arnum", init_arnum, 0},
    {"assert", init_assert, 0},
    {"assist", init_assist, 0},
    {"atensor", init_atensor, 0},
    {"avector", init_avector, 0},	/* */
    {"bfap", init_bfap, 0},
    {"bibasis", init_bibasis, 0},
    {"boolean", init_boolean, 0},
    {"breduce", init_breduce, 0},
    {"cali", init_cali, 0},
    {"camal", init_camal, 0},
    {"cantens", init_cantens, 0},
    {"ccomp", init_ccomp, 0},
    {"cdiff", init_cdiff, 0},
    {"cedit", init_cedit, 0},
    {"cfap", init_cfap, 0},
    {"cgb", init_cgb, 0},
    {"changevr", init_changevr, 0},
    {"cl", init_cl, 0},
    {"clprl", init_clprl, 0},
    {"cmacro", init_cmacro, 0},
    {"compact", init_compact, 0},
    {"compiler", init_compiler, -1},	/* */
    {"complr", init_complr, 0},
    {"conlaw", init_conlaw, -1},	/* */
    {"crack", init_crack, -1},	/* */
    {"cuba", init_cuba, -1},	/* */
    {"cvit", init_cvit, 0},
    {"dcfsf", init_dcfsf, 0},
    {"defint", init_defint, 0},
    {"desir", init_desir, 0},
    {"dfpart", init_dfpart, 0},
    {"dipoly", init_dipoly, 0},
    {"dummy", init_dummy, 0},
    {"dvfsf", init_dvfsf, 0},
    {"eds", init_eds, 0},
    {"entry", init_entry, 0},
    {"excalc", init_excalc, 0},
    {"ezgcd", init_ezgcd, 0},
    {"factor", init_factor, 0},
    {"fap", init_fap, 0},
    {"fide", init_fide, -1},	/* */
    {"fide1", init_fide1, 0},
    {"fisl", init_fisl, 0},
    {"fmprint", init_fmprint, 0},
    {"fps", init_fps, 0},
    {"ftr", init_ftr, 0},
    {"gcref", init_gcref, -1},	/* */
    {"gentran", init_gentran, 0},
    {"geometry", init_geometry, 0},
    {"geoprover", init_geoprover, 0},
    {"gnuplot", init_gnuplot, -1},	/* */
    {"groebfac", init_groebfac, 0},
    {"groebner", init_groebner, 0},
    {"groebnr2", init_groebnr2, 0},
    {"groext", init_groext, 0},
    {"guardian", init_guardian, 0},
    {"gurobi", init_gurobi, -1},	/* */
    {"hephys", init_hephys, 0},
    {"ibalp", init_ibalp, 0},
    {"ideals", init_ideals, 0},
    {"ineq", init_ineq, 0},
    {"int", init_int, 0},
    {"invbase", init_invbase, 0},
    {"lalr", init_lalr, 0},
    {"lap", init_lap, 0},
    {"laplace", init_laplace, 0},
    {"libreduce", init_libreduce, 0},
    {"lie", init_lie, 0},
    {"liepde", init_liepde, 0},
    {"limits", init_limits, 0},
    {"linalg", init_linalg, 0},
    {"linineq", init_linineq, 0},
    {"listvecops", init_listvecops, 0},
    {"lpdo", init_lpdo, 0},
    {"mathml", init_mathml, 0},
    {"mathmlom", init_mathmlom, 0},
    {"mathpr", init_mathpr, 0},
    {"matrix", init_matrix, 0},
    {"misc", init_misc, 0},
    {"mma", init_mma, -1},	/* */
    {"modsr", init_modsr, 0},
    {"module", init_module, -1},	/* */
    {"mri", init_mri, 0},
    {"mri_ofsf", init_mri_ofsf, 0},
    {"mri_pasf", init_mri_pasf, 0},
    {"mrvlimit", init_mrvlimit, 0},
    {"ncpoly", init_ncpoly, 0},
    {"nlopt", init_nlopt, -1},	/* */
    {"noncom2", init_noncom2, 0},
    {"normform", init_normform, 0},
    {"numeric", init_numeric, 0},
    {"odesolve", init_odesolve, 0},
    {"ofsf", init_ofsf, 0},
    {"orthovec", init_orthovec, 0},
    {"otrgsimp", init_otrgsimp, -1},	/* */
    {"pasf", init_pasf, 0},
    {"pf", init_pf, 0},
    {"pgauss", init_pgauss, 0},
    {"physop", init_physop, 0},
    {"plot", init_plot, 0},
    {"pm", init_pm, 0},
    {"poly", init_poly, 0},
    //{"polydiv", init_polydiv, 0},
    {"pretty", init_pretty, 0},
    {"profile", init_profile, 0},
    {"prolog", init_prolog, 0},
    {"qepcad", init_qepcad, -1},	/* */
    {"qerrc", init_qerrc, 0},
    {"qhull", init_qhull, 0},
    {"qqe", init_qqe, 0},
    {"qqe_ofsf", init_qqe_ofsf, 0},
    {"qsum", init_qsum, -0},	/* */
    {"randpoly", init_randpoly, 0},
    {"ranum", init_ranum, 0},
    // {"rarith", init_rarith, -1}, /* */
    {"rataprx", init_rataprx, 0},
    {"ratint", init_ratint, 0},
    {"rcref", init_rcref, -1},	/* */
    {"reacteqn", init_reacteqn, 0},
    {"redfront", init_redfront, 0},
    {"redlog", init_redlog, 0},
    {"reduce4", init_reduce4, 0},
    {"rend", init_rend, 0},
    {"reset", init_reset, 0},
    {"residue", init_residue, 0},
    {"rl", init_rl, 0},
    {"rlfi", init_rlfi, 0},
    {"rlisp", init_rlisp, 0},
    {"rlisp88", init_rlisp88, 0},
    {"rlsupport", init_rlsupport, 0},
    {"rltools", init_rltools, 0},
    {"roots", init_roots, 0},
    {"roots2", init_roots2, 0},
    {"rprint", init_rprint, 0},
    {"rsolve", init_rsolve, 0},
    {"rtrace", init_rtrace, 0},
    {"rubi_red", init_rubi_red, -1},	/* */
    {"scope", init_scope, 0},
    {"sets", init_sets, 0},
    {"sfgamma", init_sfgamma, 0},
    {"smt", init_smt, 0},
    {"solve", init_solve, 0},
    {"sparse", init_sparse, 0},
    {"spde", init_spde, 0},
    {"specbess", init_specbess, 0},
    {"specfaux", init_specfaux, 0},
    {"specfn", init_specfn, 0},
    {"specfn2", init_specfn2, 0},
    {"sstools", init_sstools, 0},
    {"sum", init_sum, 0},
    {"support", init_support, -1},	/* */
    {"susy2", init_susy2, 0},
    {"symaux", init_symaux, 0},
    {"symmetry", init_symmetry, 0},	/* */
    {"tables", init_tables, 0},
    {"talp", init_talp, 0},
    {"taylor", init_taylor, 0},
    {"tmprint", init_tmprint, -1},	/* */
    {"tplp", init_tplp, 0},
    {"tps", init_tps, 0},
    {"tri", init_tri, 0},
    {"trigd", init_trigd, 0},
    {"trigint", init_trigint, 0},
    {"trigsimp", init_trigsimp, 0},
    {"turtle", init_turtle, -1},	/* */
    {"utf8", init_utf8, 0},
    {"v3tools", init_v3tools, -1},	/* */
    {"wu", init_wu, 0},
    {"xcolor", init_xcolor, 0},
    {"xideal", init_xideal, 0},
    {"z3", init_z3, -1},	/* */
    {"zeilberg", init_zeilberg, 0},
    {"ztrans", init_ztrans, 0},
    {NULL, NULL, 0}
};
#endif

/***************************************************************/
/***************************************************************/
/***************************************************************/


long            curchar = NOT_CHAR;

static long     boffop;



Sex
make_literal(const char *p)
{

    return read_from_string(p);
}
#ifdef ASUKA
Sex
make_asuka_literal(const char *p)
{
    Sex             r;
    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);

    r = read_from_string(p);

    pop(current_package);
    current_charvec = package_charvec(current_package);
    return r;
}
#endif

Sex
make_string(char *str)
{
    return make_string2(str, strlen(str));
}

Sex
make_string2(char *str, long n)
/*
 * Given a Lisp string, plus its length, create a Lisp (simple-) string.
 */
{
    Sex             r;
    char           *s;
    long            k;

    r = getvector(TAG_BINARY, T_STRING, HDR + n);
    k = (n + (SEX - 1)) & -(2 * SEX);
    s = (char *) r - TAG_BINARY;

    *(long *) (s + k + 2 * SEX) = 0;
    if (k != 0)
	*(long *) (s + k + SEX) = 0;
    memcpy(s + HDR, str, (size_t) n);
    return r;
}

Sex
Lbatchp()
{
    return Bool_to_Sex(batchp());
}

Sex
Lgetenv(Sex a)
{
    char            parmname[FILENAME_MAX];
    Sex             r;
    long            len;
    char           *w;

    if (is_symbol(a)) {
	w = pname(a);
	len = symbol_length(a);
    } else if (!is_string(a)) {
	return aerror("getenv");
    } else {
	w = &string_value(a, 0);
	len = string_length(a);
    }
    if (len >= FILENAME_MAX)
	len = FILENAME_MAX - 1;
    memcpy(parmname, w, (size_t) len);
    parmname[len] = 0;
    w = getenv(parmname);
    if (w == NULL)
	return nil;
    r = make_string(w);
    return r;
}

#ifdef UNIX
Sex
Lsystem(Sex a)
{
    char            parmname[FILENAME_MAX];
    long            len;
    long            w;
    char           *p;

    if (a == nil) {
	w = system(NULL);
	return Bool_to_Sex(w != 0);
    }
    if (is_symbol(a)) {
	p = pname(a);
	len = symbol_length(a);
    } else if (!is_string(a)) {
	return aerror("system");
    } else {
	p = &string_value(a, 0);
	len = string_length(a);
    }
    if (len >= FILENAME_MAX)
	len = FILENAME_MAX - 1;
    memcpy(parmname, p, (size_t) len);
    parmname[len] = 0;
    w = system(parmname);
    return mkfxat((long) w);
}
#else
Sex
Lsystem(Sex a)
{
    return error1("system not implemented", a);
}
#endif

/***************************************************************/
/***************************************************************/
/***************************************************************/
void
close_all_files()
{
}

void
warm_start()
{
}

Sex
Lrestart(Sex a)
{
    long            core = core_size;
    if (is_fixnum(a))
	core = mkint(a);
    if (core >= 1024 * 1024)
	core_size = core;
    close_all_files();
    warm_start();
    return nil;
}
/***************************************************************/
/***************************************************************/
/***************************************************************/

unsigned long
hash_string(char *s, long n)
{
    unsigned long   hh = 0;
    unsigned char  *b = (unsigned char *) s;
    union {
	unsigned long   h;
	unsigned char   b[4];
    } h;

    h.h = 0;
    while (n >= 4) {
	h.b[0] = *b++;
	h.b[1] = *b++;
	h.b[2] = *b++;
	h.b[3] = *b++;
	hh = (hh << 7) ^ (hh << 1) ^ h.h;
	n -= 4;
    }
    h.h = 0;
    while (n > 0) {
	h.b[n] = *b++;
	n--;
    }
    hh = (hh << 7) ^ (hh << 1) ^ h.h;
    return ((0x7a3c * hh) & int_max);
}

Sex
make_fluid_symbol(char *s, Sex value)
{
    Sex             v;
    push(value);
    v = intern(s, (long) strlen(s));
    pop(value);
    qtag(v) |= SYM_FLUID;
    value(v) = value;
    return v;
}

Sex
make_symbol(char *s)
{
    Sex             v;
    v = intern(s, (long) strlen(s));
    return v;
}

#ifdef ASUKA
Sex
Lasuka_symbol(Sex s)
{
    Sex             v;
    if (!is_symbol(s))
	return aerror1("asuka_symbol", s);
    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    v = intern(pname(s), symbol_length(s));
    pop(current_package);
    current_charvec = package_charvec(current_package);
    return v;
}

Sex
Lsugar_symbol(Sex s)
{
    Sex             v;
    if (!is_symbol(s))
	return aerror1("sugar_symbol", s);
    push(current_package);
    current_package = value(current_sugar_package);
    current_charvec = package_charvec(current_package);
    v = intern(pname(s), symbol_length(s));
    pop(current_package);
    current_charvec = package_charvec(current_package);
    return v;
}
#endif


Sex
make_asuka_symbol(char *s)
{
    Sex             v;
    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    v = intern(s, (long) strlen(s));
    pop(current_package);
    current_charvec = package_charvec(current_package);
    return v;
}

Sex
make_asuka_fluid_symbol(char *s, Sex value)
{
    Sex             v;
    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    v = intern(s, (long) strlen(s));
    qtag(v) |= SYM_FLUID;
    value(v) = value;
    pop(current_package);
    current_charvec = package_charvec(current_package);
    return v;
}

Sex
make_proc_symbol(char *s, lisp_proc * f1, long nargs)
{
    Sex             v;
    Tag             ta;
    v = intern(s, (long) strlen(s));
    proc(v) = f1;
    ta = qtag(v) & ~PROC_MASK;
    if (nargs < -2) {
	qnargs(v) = -3 - nargs;
	ta = ta | PROC_CSMACRO;
    } else if (nargs == -2) {
	qnargs(v) = 1;
	ta = ta | PROC_CMACRO;
    } else if (nargs == -1) {
	qnargs(v) = 1;
	ta = ta | PROC_FSUBR;
    } else {
	qnargs(v) = nargs;
	ta = ta | PROC_SUBR;
    }
    qtag(v) = ta;
    return v;
}

static          bool
add_to_hash(Sex s, Sex vector, unsigned long hash)
{
    Tag             h;
    long            size,
                    size1,
                    i,
                    step,
                    probes;

    h = vechdr(vector);
    size = bnum_n(veclen(vector));
    size1 = size - 1;
    i = (long) (hash & size1);
    step = 1 | ((hash >> 10) & size1);
    probes = 0;

    while (++probes < size1) {
	if (is_fixnum(vector_value(vector, i))) {
	    vector_value(vector, i) = s;
	    return YES;		/* 
				 * Success 
				 */
	}
	i = (i + step) & size1;
    }
    return NO;			/* 
				 * Table is totally full 
				 */
}


static          Sex
rehash(Sex v)
{
    long            h,
                    i;
    Sex             new_obvec;
    Sex             s;
    unsigned long   hash;

    h = veclen(v) - HDR;
    stackcheck();
    push(v);
    new_obvec = getvector_init(h * 2 + HDR, mkfxat(0));
    v = stack[0];
    h = h / SEX;
    for (i = 0; i < h; i++) {
	s = vector_value(v, i);
	if (is_fixnum(s))
	    continue;
	hash = hash_string(pname(s), symbol_length(s));
	if (!add_to_hash(s, new_obvec, hash)) {
	    fatal_error(" object table over flow (this can't happen)");
	}
    }
    popv(1);
    return new_obvec;
}


static          Sex
lookup(char *str, long strsize, Sex v, unsigned long hash)
/*
 * Return Lisp number 0 if not found.
 */
{
    Tag             h;
    long            size,
                    size1;
    long            i,
                    step;

    // printf("lookup: string = %s\n", str);

    h = vechdr(v);
    size = bnum_n(veclen(v));
    size1 = size - 1;
    i = (long) (hash & size1);
    step = 1 | ((hash >> 10) & size1);
    while (YES) {
	Sex             w = vector_value(v, i);
	// printf("lookup: w=0x%16lx, i=%ld\n", (long)w, i);
	// printf("step = 0x%x, size = 0x%x\n", step, size1);
	if (w == mkfxat(0))
	    return w;
	if (w != mkfxat(1)) {
	    if (memcmp(str,
		       pname(w),
		       (size_t) strsize) == 0 &&
		symbol_length(w) == (unsigned long) strsize)
		return w;
	}
	i = (i + step) & size1;
    }
}

static long
ordersymbol(Sex v1, Sex v2)
{
    char           *pn1 = pname(v1),
	*pn2 = pname(v2);
    long            c;
    long            l1 = symbol_length(v1),
	l2 = symbol_length(v2);
    c = memcmp(pn1, pn2, (size_t) (l1 < l2 ? l1 : l2));
    if (c == 0)
	c = l1 - l2;
    return -c;
}


static long
ordpv(Sex u, Sex v)
{
    Tag             hu = vechdr(u),
	hv = vechdr(v);
    long            lu = veclen(u),
	lv = veclen(v),
	n = HDR;
    if (type_of_header(hu) != type_of_header(hv))
	return (type_of_header(hu) < type_of_header(hv) ? -1 : 1);
    stackcheck();
    while (n < lu && n < lv) {
	Sex             eu = *(Sex *) (u - TAG_VECTOR + n),
	    ev = *(Sex *) (v - TAG_VECTOR + n);
	long            w;
	push2(u, v);
	w = orderp(eu, ev);
	pop2(v, u);
	if (w != 0)
	    return w;
	n += SEX;
    }
    return (lu == lv ? 0 : lu < lv ? -1 : 1);
}


#if defined(REDUCE) || defined(GAL)

long
orderp(Sex u, Sex v)
{
    long            w;

    while (YES) {
	if (u == v)
	    return 0;
	if (u == nil)
	    return -1;
	else if (v == nil)
	    return 1;
	else if (!is_cons(u)) {
	    if (!is_cons(v)) {
		if (is_vechdr(u)) {
		    if (is_vechdr(v))
			return ordpv(u, v);
		    else
			return 1;
		} else if (is_number(u)) {
		    if (is_number(v)) {
			push2(u, v);
			w = lessp2(u, v);
			pop2(v, u);
			if (w)
			    return -1;
			w = eql(u, v);
			return w == YES ? 0 : 1;
		    } else
			return -1;
		} else if (is_number(v))
		    return 1;
		else if (is_symbol(u)) {
		    if (is_symbol(v))
			return ordersymbol(u, v);
		    else
			return -1;
		} else if (is_symbol(v))
		    return 1;
		else
		    return (long) u - (long) v;
	    } else
		return -1;
	} else if (!is_cons(v))
	    return 1;
	else {
	    Sex             cu = car(u),
		cv = car(v);
	    long            w;
	    push2(u, v);
	    stackcheck();
	    w = orderp(cu, cv);
	    pop2(v, u);
	    if (w != 0) {
		return w;
	    }
	    u = cdr(u);
	    v = cdr(v);
	}
    }
}

#else

long
orderp(Sex u, Sex v)
{
    Sex             cu,
                    cv;
    long            t1,
                    t2;
    long            w;
    static int      ord[] = { 6, 0, 0, 1, 3, 4, 5, 2 };
    /*
     * fixnum, numbers, symbol, binary, vector, cons 
     */

    while (YES) {
	if (u == v)
	    return 0;
	else if (u == nil)
	    return -1;
	else if (v == nil)
	    return 1;
	t1 = ((long) u) & TAG_BITS;
	t2 = ((long) v) & TAG_BITS;
	if (t1 != t2)
	    return (ord[t1] - ord[t2]);
	switch (t1) {
	case TAG_CONS:
	    cu = car(u);
	    cv = car(v);
	    stackcheck();
	    push2(u, v);
	    w = orderp(cu, cv);
	    pop2(v, u);
	    if (w != 0)
		return w;
	    u = cdr(u);
	    v = cdr(v);
	    break;
	case TAG_FIXNUM:
	    return (mkfxat(u) - mkfxat(v));
	case TAG_SYMBOL:
	    return compare_var(u, v);
	case TAG_BINARY:
	    w = binhdr(u) - binhdr(v);
	    if (w != 0)
		return w;
	    w = binlen(u) - binlen(v);
	    return w;
	case TAG_VECTOR:
	    w = vechdr(u) - vechdr(v);
	    if (w != 0)
		return w;
	    w = veclen(u) - veclen(v);
	    if (w != 0)
		return w;
	    t1 = type_of_header(vechdr(u));
	    if (t1 == T_POLY)
		return compare_term(u, v);
	    else if (t1 == T_DIPOLY) {
#if 0
		long            o;
		o = compare_dipoly(u, v);
		lisp_printf("compare terms: %p, %p ==> %d\n", u, v, o);
		return o;
#else
		return compare_dipoly(u, v);
#endif
	    }
	    t1 = veclen(u);
	    t2 = HDR;
	    while (t2 < t1) {
		long            w;
		cu = *(Sex *) (u - TAG_VECTOR + t2);
		cv = *(Sex *) (v - TAG_VECTOR + t2);
		push2(u, v);
		w = orderp(cu, cv);
		pop2(v, u);
		if (w != 0)
		    return w;
		t2 += SEX;
	    }
	    return 0;
	case TAG_NUMBERS:
	    if (eql(u, v) != NO)
		return 0;
	    return lessp2(u, v) ? -1 : 1;
	default:
	    return ((long) u - (long) v);
	}
    }
}

#endif

Sex
Lorderp(Sex a, Sex b)
{
    long            w;
    w = orderp(a, b);
    return Bool_to_Sex(w >= 0);
}

Sex
Lfluid_list()
{
    long            i,
                    size;
    Sex             p,
                    v;
    p = current_package;
    v = package_vec(p);
    size = bnum_n(veclen(v));
    for (i = 0; i < size; i++) {
	p = vector_value(v, i);
	if (!is_symbol(p))
	    continue;
	if (qtag(p) & SYM_FLUID)
	    lisp_printf("%p\n", p);
    }
    return nil;
}


Sex
Lglobal_list()
{
    long            i,
                    size;
    Sex             p,
                    v;
    p = current_package;
    v = package_vec(p);
    size = bnum_n(veclen(v));
    for (i = 0; i < size; i++) {
	p = vector_value(v, i);
	if (!is_symbol(p))
	    continue;
	if (qtag(p) & SYM_GLOBAL)
	    lisp_printf("%p\n", p);
    }
    return nil;
}


static          bool
remob(Sex sym, Sex v)
{
    Tag             h;
    unsigned long   hash;
    long            i,
                    size,
                    step;

    hash = hash_string(pname(sym), symbol_length(sym));
    h = vechdr(v);
    size = bnum_n(veclen(v));
    i = (long) (hash & (size - 1));
    step = 1 | ((hash >> 10) & (size - 1));
    while (YES) {
	Sex             w = vector_value(v, i);
	if (w == mkfxat(0))
	    return NO;
	if (w == sym) {
	    vector_value(v, i) = mkfxat(1);
	    return YES;
	}
	i = i + step;
	if (i >= size)
	    i -= size;
    }
}


Sex
gensym2(char *name)
{
    Sex             id;
    long            n;

    n = strlen(name);
    id = getvector(TAG_SYMBOL, T_SYMBOL, sizeof(Symbol_Head) - CELL + n);
    value(id) = undefined_var;
    memcpy(pname(id), name, n);
    plist(id) = nil;
    proc(id) = (lisp_proc *) undefined_proc;
    qnargs(id) = 0;
    return id;
}


Sex
Lgensym()
{
    char            name[16];

    sprintf(name, "   G%04d", gensym_cnt++);
    if (gensym_cnt >= 10000)
	gensym_cnt = 0;
    return gensym2(name);
}

Sex
Lreset_gensym(Sex a)
{
    int             old = gensym_cnt;
    int             new;
    if (is_fixnum(a)) {
	new = mkint(a);
	gensym_cnt = new;
    }
    return mkfxat(old);
}

Sex
Lgensym1(Sex a)
{
    char            name[64];
    long            n;

    if (is_string(a)) {
	n = string_length(a);
	if (n > 56)
	    n = 56;
	memcpy(name, &string_value(a, 0), n);
    } else if (is_symbol(a)) {
	n = symbol_length(a);
	if (n > 56)
	    n = 56;
	memcpy(name, pname(a), n);
    } else
	return aerror("gensym1");
    sprintf(name + n, "%04d", gensym1_cnt++);
    if (gensym1_cnt >= 10000)
	gensym1_cnt = 0;
    return gensym2(name);
}

Sex
Lgensym2(Sex a, Sex b)
{
    char            name[128];
    long            n,
                    m;

    if (is_string(a)) {
	n = string_length(a);
	if (n > 56)
	    n = 56;
	memcpy(name, &string_value(a, 0), n);
    } else if (is_symbol(a)) {
	n = symbol_length(a);
	if (n > 56)
	    n = 56;
	memcpy(name, pname(a), n);
    } else
	return aerror("gensym2");
    if (is_string(b)) {
	m = string_length(a);
	if (m > 56)
	    m = 56;
	memcpy(name + n, &string_value(b, 0), m);
    } else if (is_symbol(b)) {
	m = symbol_length(b);
	if (m > 56)
	    m = 56;
	memcpy(name + n, pname(b), m);
    } else if (is_fixnum(b))
	m = sprintf(name + n, "%ld", mkint(b));
    else
	return aerror("gensym2");
    name[n + m] = 0;
    return gensym2(name);
}

Sex
Lgen_sharp_symbol(Sex a)
{
    char            name[64];
    long            n;

    if (is_string(a)) {
	n = string_length(a);
	if (n > 56)
	    n = 56;
	memcpy(name + 1, &string_value(a, 0), n);
    } else if (is_symbol(a)) {
	n = symbol_length(a);
	if (n > 56)
	    n = 56;
	memcpy(name + 1, pname(a), n);
    } else
	return aerror("gen_sharp_symbol");
    name[0] = '#';
    name[n + 1] = 0;
    return gensym2(name);
}


Sex
intern(char *str, long h)
{
    Sex             r;
    Sex             p = current_package;
    Sex             s;
    unsigned long   hash;

    hash = hash_string(str, h);
    r = lookup(str, h, package_vec(p), hash);
    if (r != mkfxat(0)) {
	return r;
    }
    s = (Sex) getvector(TAG_SYMBOL, T_SYMBOL, symhdr_length + h - CELL);
    value(s) = undefined_var;
    ((long *) pname(s))[h / SEX] = 0;
    memcpy(pname(s), str, h);
    plist(s) = nil;
    proc(s) = (lisp_proc *) undefined_proc;
    qnargs(s) = 0;
#ifdef ALGEBRAIC_PROPERTY
    gvalue(s) = undefined_var;
    gproc(s) = undefined_proc;
    gtag(s) = 0;
    gnargs(s) = 0;
#endif

    p = current_package;
    r = package_vec(p);
    if (package_cnt(p) >= 3 * veclen(r) / (LPSIZE / 32)) {	/* 
								 * ratio ~~ 3/4 
								 */
	stackcheck();
	push2(s, p);
	r = rehash(r);
	pop2(p, s);
	package_vec(p) = r;
    }
    package_cnt(p) = package_cnt(p) + 0x10;
    add_to_hash(s, r, hash);
    return s;
}

Sex
Lintern(Sex str)
{
    long            len;
    if (is_symbol(str)) {
	len = symbol_length(str);
	if (len >= BUFFER_SIZE)
	    len = BUFFER_SIZE - 1;
	memcpy(boffo, pname(str), len);
	boffo[len] = 0;
	return intern(boffo, len);
    }
    if (!is_string(str)) {
	return str;
    }
    len = string_length(str);
    if (len >= BUFFER_SIZE)
	len = BUFFER_SIZE - 1;
    memcpy(boffo, &string_value(str, 0), len);
    boffo[len] = 0;
    return intern(boffo, len);
}


Sex
Linternp(Sex str)
{
    long            len;
    Sex             r;
    Sex             p = current_package;
    Sex             s;
    unsigned long   hash;

    if (is_symbol(str)) {
	len = symbol_length(str);
	if (len >= BUFFER_SIZE)
	    len = BUFFER_SIZE - 1;
	memcpy(boffo, pname(str), len);
	boffo[len] = 0;
    	hash = hash_string(boffo, len);
    	r = lookup(boffo, len, package_vec(p), hash);
    	if(r != mkfxat(0))
	    return true;
    }
    return nil;
}


Sex
ndelete(Sex a, Sex l)
{
    if (!is_cons(l))
	return l;
    if (a == car(l))
	return cdr(l);
    {
	Sex             z1 = l,
	    z2 = cdr(l);
	while (is_cons(z2)) {
	    if (a == car(z2)) {
		cdr(z1) = cdr(z2);
		return l;
	    } else {
		z1 = z2;
		z2 = cdr(z2);
	    }
	}
    }
    return l;
}

Sex
Lremob(Sex sym)
{
    Sex             p = current_package;
    if (!is_symbol(sym))
	return nil;
    if (remob(sym, package_vec(p))) {
	package_cnt(p) = package_cnt(p) - 0x10;
	return true;
    }
    return nil;
}



long            tty_count;
#define TTYBUF_SIZE 256
#ifdef USE_READLINE
static char    *tty_buffer = NULL;
static unsigned char *tty_pointer = NULL;
#else
#ifdef Kanji
static kchar_t  tty_buffer[TTYBUF_SIZE];
static kchar_t *tty_pointer;
#else
static char     tty_buffer[TTYBUF_SIZE];
static unsigned char *tty_pointer;
#endif
#endif

static long     prevchar = '\n';

void
flush_tty()
{
    prevchar = '\n';
    tty_count = 0;
}

static char     prompt_string[256];
#ifdef SAVE_WORK_SPACE
static int      work_space_count = 1;
#endif

long
ttyget()
{
    double          tv0;
    extern long     batch_mode;

    if (batch_mode) {
	long            c;
	c = my_getwc(stdin);
	if (c == EOF) {
	    clearerr(stdin);
	    c = CTRL_D;
	}
	return c;
    }
    if (tty_count == 0) {
#ifdef USE_READLINE
	tv0 = get_time();
#else
	tv0 = get_time();
	if (prevchar == '\n') {
	    print_flag = 0;
#ifdef SAVE_WORK_SPACE
	    print_sex(mkfxat(work_space_count), char_to_stdout);
#endif
	    print_sex(prompt_message, char_to_stdout);
	    // lisp_printf(prompt_message, work_space_count);
	}
#endif
#ifdef NO_SIGNAL_HANDLER
#ifdef USE_READLINE
	if (tty_buffer) {
	    free(tty_buffer);
	    tty_buffer = NULL;
	    io_time += get_time() - tv0;
	    return '\n';
	}
	tty_buffer = readline(prompt_string);
	if (tty_buffer == NULL) {
	    tty_count = 0;
	    io_time += get_time() - tv0;
	    return CTRL_D;
	}
	if (tty_buffer && *tty_buffer)
	    add_history(tty_buffer);
	tty_count = strlen(tty_buffer);
#else
	while (tty_count < TTYBUF_SIZE && !interrupt_pending) {
	    long            c;
	    c = my_getwc(stdin);
	    if (c == EOF) {
		clearerr(stdin);	/* 
					 * Believed to be what is
					 * * wanted 
					 */
		c = CTRL_D;	/* 
				 * Use ASCII ^D as EOF marker 
				 */
	    }
	    tty_buffer[tty_count++] = c;
	    if (c == '\n' || c == '\v' || c == CTRL_D)
		break;
	}
#endif
#else
	while (YES) {
	    fflush(stdout);
	    if (setjmp(sigint_buf)) {
		sigint_must_longjmp = NO;
		interrupt_pending = YES;
		tty_count = 0;
	    } else {
#ifdef USE_READLINE
		if (tty_buffer) {
		    free(tty_buffer);
		    tty_buffer = NULL;
		    io_time += get_time() - tv0;
		    return '\n';
		}
		tty_buffer = readline(prompt_string);
		if (tty_buffer == NULL) {
		    tty_count = 0;
		    io_time += get_time() - tv0;
		    return CTRL_D;
		}
		if (tty_buffer && *tty_buffer)
		    add_history(tty_buffer);
		tty_count = strlen(tty_buffer);
#else
		while (tty_count < TTYBUF_SIZE && !interrupt_pending) {
		    long            c;
		    sigint_must_longjmp = YES;
		    c = my_getwc(stdin);
		    sigint_must_longjmp = NO;
		    if (c == EOF) {
			clearerr(stdin);	/* 
						 * Believed to be what is
						 * * wanted 
						 */
			c = CTRL_D;	/* 
					 * Use ASCII ^D as EOF marker 
					 */
		    }
		    tty_buffer[tty_count++] = c;
		    if (c == '\n' || c == '\v' || c == CTRL_D)
			break;
		}
#endif
	    }
	    if (interrupt_pending) {
		interrupt_pending = NO;
		interrupted(nil);
		tty_count = 0;
		continue;
	    }
	    break;
	}
#endif
	io_time += get_time() - tv0;
#ifdef Kanji
	tty_pointer = tty_buffer;
#else
	tty_pointer = (unsigned char *) tty_buffer;
#endif
    }
    if (tty_count == 0)
	prevchar = '\n';	/* 
				 * ^C odd case 
				 */
    else {
	tty_count--;
	prevchar = *tty_pointer++;
    }
    return prevchar;
}

#if defined(FreeBSD)
long
ttygetchar(char *s)
{
    long            c = 0;
    long            d = 0;
    long            fd;

    fd = open("/dev/tty", O_RDWR, 0);
    if (fd < 0) {
#if 0
	fprintf(stderr,
		"\n *** Lisp attention and can't open /dev/tty.\n");
	exit(1);
#else
	fprintf(stderr, "%s", s);
	fprintf(stderr,
		"\n *** Lisp attention and can't open /dev/tty.\n");
	return 'q';
#endif
    }
#if 0
    write(fd, "\n *** Lisp attention !  ?(q/c/t) = ",
	  sizeof("\n *** Lisp attention !  ?(q/c/t) = "));
#else
    write(fd, s, strlen(s));
#endif
    read(fd, &c, 1);
    if (c == '\n')
	c = 'c';
    else {
	d = 0;
	while (d != '\n')
	    read(fd, &d, 1);

    }
    close(fd);
    flush_tty();
    return c;
}

#else
long
ttygetchar(char *s)
{
    double          tv0;
    long            c;
    long            d;
    extern long     batch_mode;

    if (batch_mode) {
	return 'q';
    }
    tv0 = get_time();
#if 0
    fprintf(stderr, "\n *** Lisp attention !  ?(q/c/t) = ");
#else
    fprintf(stderr, "%s", s);
#endif
    c = '\n';
#ifdef NO_SIGNAL_HANDLER
    c = my_getwc(stdin);
    d = c;
    while (d != '\n' && d != '\v' && d != CTRL_D && !interrupt_pending) {
	d = my_getwc(stdin);
	if (d == EOF) {
	    clearerr(stdin);	/* 
				 * Believed to be what is
				 * * wanted 
				 */
	    d = CTRL_D;		/* 
				 * Use ASCII ^D as EOF marker 
				 */
	}
    }
#else
    while (YES) {
	fflush(stdout);
	if (setjmp(sigint_buf)) {
	    sigint_must_longjmp = NO;
	    interrupt_pending = YES;
	    tty_count = 0;
	} else {
	    sigint_must_longjmp = YES;
	    c = my_getwc(stdin);
	    sigint_must_longjmp = NO;
	    d = c;
	    while (d != '\n' && d != '\v' && d != CTRL_D
		   && !interrupt_pending) {
		sigint_must_longjmp = YES;
		d = my_getwc(stdin);
		sigint_must_longjmp = NO;
		if (d == EOF) {
		    clearerr(stdin);	/* 
					 * Believed to be what is
					 * * wanted 
					 */
		    d = CTRL_D;	/* 
				 * Use ASCII ^D as EOF marker 
				 */
		}
	    }
	}
	if (interrupt_pending) {
	    interrupt_pending = NO;
	    interrupted(nil);
	    tty_count = 0;
	    continue;
	}
	break;
    }
#endif
    io_time += get_time() - tv0;
    flush_tty();
    return c;
}
#endif

Sex
Lfile_line()
{
    return mkfxat(file_line(lisp_input));
}

Sex
Lrds(Sex a)
{
    Sex             old = lisp_input;
    if (a == nil)
	a = standard_input;
    if (a == old)
	return old;
    else if (!is_file(a))
	return aerror("rds");
    else if (file_FILE(a) == NULL || file_putc(a) != 0)
	return aerror("rds");
    file_char(old) = curchar;
    lisp_input = a;
    curchar = file_char(a);
    return old;
}

Sex
Lrtell()
{
    long            n;
    n = ftell(file_FILE(lisp_input));
    if (n == -1L)
	return nil;
#ifdef Kanji
    if (iskchar(curchar))
	n -= 2;
    else if (curchar != NOT_CHAR)
	n--;
#else
    if (curchar != NOT_CHAR)
	n--;
#endif
    return mkfxat(n);
}

Sex
Lrseek(Sex a)
{
    long            n;
    if (is_fixnum(a))
	n = mkint(a);
    else
	return aerror("rseek");
    curchar = NOT_CHAR;
    if (fseek(file_FILE(lisp_input), n, SEEK_SET) != 0)
	return nil;
    else
	return true;
}


static void
skip_space(getchar_fn * fn)
{
    while (YES) {
	switch (curchar) {
	case NOT_CHAR:
	case 0:
	case '\v':
	case '\f':
	case ' ':
	case '\t':
	case '\n':
	case '\r':
	case CTRL_C:
	    fn();
	    continue;

	case '%':
	    while (curchar != '\n' && curchar != EOF && curchar != CTRL_D) {
		fn();
	    }
	    continue;

	default:
#ifdef Kanji
	    if (iskspace(curchar)) {
		fn();
		continue;
	    }
#endif
	    return;
	}
    }
}

static Sex      read_sex(getchar_fn * fn);

static          Sex
read_list(getchar_fn * fn)
{
    Sex             l,
                    w;
    stackcheck();
    if (curchar == ')') {
	curchar = NOT_CHAR;
	return nil;
    }
    l = read_sex(fn);
    l = Lncons(l);
    push(l);			/* 
				 * this hold the final result 
				 */
    while (YES) {
	skip_space(fn);
	switch (curchar) {

	case ',':
	    curchar = NOT_CHAR;
	    continue;
	case ']':
	case ')':
	    curchar = NOT_CHAR;
	    pop(l);
	    return l;

	case EOF:
	case CTRL_D:
	    pop(l);
	    return l;

	case '.':
	    curchar = NOT_CHAR;
	    push(l);
	    w = read_sex(fn);
	    pop(l);
	    cdr(l) = w;
	    skip_space(fn);
	    if (curchar == ')')
		curchar = NOT_CHAR;
	    pop(l);
	    return l;

	default:
	    push(l);
	    w = read_sex(fn);
	    w = Lncons(w);
	    pop(l);
	    cdr(l) = w;
	    l = w;
	    continue;
	}
    }
}

Sex
list_to_vector(Sex l)
{
    int             len = 0, n;
    Sex             p = l;
    while (is_cons(p))
	len++, p = cdr(p);
    stackcheck();
    push(l);
    n = bnum_len(len);
    p = getvector(TAG_VECTOR, T_SIMPLE, n);
    pop(l);
    len = 0;
    while (is_cons(l)) {
	vector_value(p, len) = car(l);
	len++;
	l = cdr(l);
    }
    return p;
}


static          Sex
backquote_expander(Sex a)
{
    Sex             w1,
                    f;
    if (a == nil)
	return a;
    if (!is_cons(a))
	return Llist2(quote_symbol, a);
    stackcheck();
    f = car(a);
    if (f == comma_symbol)
	return car(cdr(a));
    if (is_cons(f) && car(f) == comma_at_symbol) {
	w1 = car(cdr(f));
	push(w1);
	a = backquote_expander(cdr(a));
	pop(w1);
	w1 = Llist2(w1, a);
	return cons(append_symbol, w1);
    }
    push(a);
    f = backquote_expander(f);
    pop(a);
    push(f);
    a = backquote_expander(cdr(a));
    pop(f);
    a = Llist2(f, a);
    return cons(cons_symbol, a);
}

static void
packbyte(long c)
{
    if (boffop >= BUFFER_SIZE) {
	buffer_over();
	return;
    }
#ifdef Kanji
    if (iskchar(c))
	boffo[boffop++] = c >> 8;
#endif
    boffo[boffop++] = c;
}

static          Sex
read_fixnum()
{
    Sex             v = mkfxat(0);
    bool            sign = NO;
    char           *p = boffo;
    long            d = 0,
	d1 = 1;
    if (*p == '+') {
	p++;
    }
    if (*p == '-') {
	p++;
	sign = YES;
    }
    boffo[boffop] = 0;
    while (*p) {
	d = 10 * d + (long) *p++ - '0';
	d1 = 10 * d1;
	if (d1 == M10_READ_EXP) {
	    v = times2(v, mkfxat(d1));
	    v = plus2(v, mkfxat(d));
	    d = 0;
	    d1 = 1;
	}
    }
    v = times2(v, mkfxat(d1));
    v = plus2(v, mkfxat(d));
    if (sign)
	v = negate(v);
    return v;
}

long
hex_to_num(long a)
{
    if (a >= 'a')
	return a - 'a' + 10;
    else if (a >= 'A')
	return a - 'A' + 10;
    else
	return a - '0';
}

static          Sex
read_hexnum()
{
    Sex             v;
    bool            sign = NO;
    long            j,
                    len;
    long            d;
    char           *p = boffo;

    boffo[boffop] = 0;
    if (*p == '+') {
	p++;
    }
    if (*p == '-') {
	p++;
	sign = YES;
    }
    p += 2;			/* 
				 * skip 0x 
				 */

#ifdef READ_LOGICAL_AS_INTEGER
    while (*p == '0')
	p++;
    len = boffo - p + boffop;
    d = 0;
    if (len < (CELL)) {
	for (j = len - 1; j >= 0; j--) {
	    d = (d << 4) + hex_to_num(*p++);
	}
	if (sign) {
	    d = -d;
	    if (d >= fix_min && d <= fix_max)
		return mkfxat(d);
	    return mkbigat1s(-1, -d);
	} else {
	    if (d >= fix_min && d <= fix_max)
		return mkfxat(d);
	    return mkbigat1s(1, d);
	}
    }
    v = new_bignum(alignit(len) / 2 + HDR);
    if (sign)
	((Big_Number *) v)->h = T_BIGNEG | TAG_HDRS;
    j = len;
    while (*p) {
	d = (d << 4) + hex_to_num(*p++);
	j--;
	if (j % (CELL) == 0) {
	    bignum_value(v)[j / (CELL)] = d;
	    d = 0;
	}
    }
    return v;
#else
    len = boffo - p + boffop;
    v = new_logical(alignit(len) / 2 + HDR);
    d = 0;
    j = len;
    while (*p) {
	d = (d << 4) + hex_to_num(*p++);
	j--;
	if (j % (CELL) == 0) {
	    logical_value(v)[j / (CELL)] = d;
	    d = 0;
	}
    }
    if (sign) {
	len = len / (CELL);
	for (j = 0; j < len; j++)
	    logical_value(v)[j] = ~logical_value(v)[j];
    }
    return v;
#endif
}


static          Sex
read_rational()
{
    Sex             p = mkfxat(0);
    Sex             q = mkfxat(0);
    bool            sign = NO;
    char           *s = boffo;
    long            d = 0,
	d1 = 1;

    boffo[boffop] = 0;
    if (*s == '+') {
	s++;
    }
    if (*s == '-') {
	s++;
	sign = YES;
    }
    while (*s && *s != '/') {
	d = 10 * d + (long) *s++ - '0';
	d1 = 10 * d1;
	if (d1 == M10_READ_EXP) {
	    p = times2(p, mkfxat(d1));
	    p = plus2(p, mkfxat(d));
	    d = 0;
	    d1 = 1;
	}
    }
    p = times2(p, mkfxat(d1));
    p = plus2(p, mkfxat(d));
    if (sign) {
	p = negate(p);
    }
    if (*s++ == 0)
	return p;
    d = 0;
    d1 = 1;
    stackcheck();
    push(p);
    while (*s) {
	d = 10 * d + (long) *s++ - '0';
	d1 = 10 * d1;
	if (d1 == M10_READ_EXP) {
	    q = times2(q, mkfxat(d1));
	    q = plus2(q, mkfxat(d));
	    d = 0;
	    d1 = 1;
	}
    }
    q = times2(q, mkfxat(d1));
    q = plus2(q, mkfxat(d));
    pop(p);
    return Lmkrational(p, q);
}

static          Sex
read_bfloat()
{
    Sex             p = mkfxat(0);
    Sex             q;
    bool            sign = NO;
    char           *s = boffo;
    long            exp = 0;
    long            d = 0,
	d1 = 1;

    boffo[boffop] = 0;
    if (*s == '+') {
	s++;
    }
    if (*s == '-') {
	s++;
	sign = YES;
    }
    for (; *s; s++) {
	if (*s == '.')
	    exp = 1;
	else if (*s != '0')
	    break;
	else if (exp)
	    exp++;
    }
    while (1) {
	switch (*s) {
	case 'e':
	    s++;
	case '\0':
	    p = times2(p, mkfxat(d1));
	    p = plus2(p, mkfxat(d));
	    break;
	case '.':
	    s++;
	    exp = 1;
	    continue;
	default:
	    if (exp)
		exp++;
	    d = 10 * d + (long) *s++ - '0';
	    d1 = 10 * d1;
	    if (d1 == M10_READ_EXP) {
		p = times2(p, mkfxat(d1));
		p = plus2(p, mkfxat(d));
		d = 0;
		d1 = 1;
	    }
	    continue;
	}
	break;
    }
    if (*s)
	exp = exp - atoi(s);
    if (sign) {
	p = negate(p);
    }
    stackcheck();
    push(p);
    exp--;
    if (exp > 0) {
	q = expti(mkfxat(10), exp);
	q = is_fixnum(q) ? bfloatiti(q) : bfloatitb(q);
	pop(p);
	p = quot2(p, q);
    } else if (exp < 0) {
	q = expti(mkfxat(10), -exp);
	q = is_fixnum(q) ? bfloatiti(q) : bfloatitb(q);
	pop(p);
	p = times2(p, q);
    } else
	p = bfloatit(p);
    return p;
}

static          Sex
read_float()
{
    double          d;
    long            dig;
    Sex             r;
    char           *p = boffo;

    boffo[boffop] = 0;
    if (*p == '+' || *p == '-')
	p++;
    while (*p == '0' || *p == '.')
	p++;
    dig = 0;
    while (*p) {
	if (isdigit(*p))
	    dig++;
	else if (*p != '.')
	    break;
	p++;
    }
#ifdef REDUCE
#else
    if (dig > read_precision)
	return read_bfloat();
#endif
#ifdef Android
    d = my_atof((const char *) boffo);
#else
    d = atof(boffo);
#endif
    r = new_float();
    float_val(r) = d;
    return r;
}

static          Sex
read_number(getchar_fn * fn)
{
    if (curchar == '0') {
	packbyte(curchar);
	fn();
	if (curchar == 'x') {
	    packbyte(curchar);
	    fn();
	    while (isexdigit(curchar)) {
		packbyte(curchar);
		fn();
	    }
	    if (!is_delimit(curchar))
		return nil;
	    return read_hexnum();
	}
    }
    while (isedigit(curchar)) {
	packbyte(curchar);
	fn();
    }
    if (curchar == '.') {
	packbyte(curchar);
	fn();
	while (isedigit(curchar)) {
	    packbyte(curchar);
	    fn();
	}
	if (curchar == 'e' || curchar == 'E') {
	    packbyte('e');
	    fn();
	    if (curchar == '+' || curchar == '-') {
		packbyte(curchar);
		fn();
	    }
	    while (isedigit(curchar)) {
		packbyte(curchar);
		fn();
	    }
	}
	if (!is_delimit(curchar))
	    return nil;
	return read_float();
    } else if (curchar == '/') {
	packbyte(curchar);
	fn();
	while (isedigit(curchar)) {
	    packbyte(curchar);
	    fn();
	}
	if (!is_delimit(curchar))
	    return nil;
	return read_rational();
    }
    if (!is_delimit(curchar))
	return nil;
    return read_fixnum();
}

static          Sex
read_sharp(getchar_fn * fn)
{
    Sex             x;

    boffop = 0;
    while (YES) {
	fn();
	if (curchar == ESCAPE_CHAR)
	    fn();
	else if (is_delimit(curchar))
	    break;
	packbyte(curchar);
    }
    if (boffop == 0)
	return CHAR_OF('#' & 0xff);
    boffo[boffop] = 0;
    x = intern(boffo, boffop);
    return x;
}

static          Sex
read_sex(getchar_fn * fn)
{
    Sex             w;

    while (YES) {
	skip_space(fn);
	switch (curchar) {
	case EOF:
	case CTRL_D:
	    return CHAR_EOF;

	case '(':
	    curchar = NOT_CHAR;
	    skip_space(fn);
	    return read_list(fn);

	case '[':
	    curchar = NOT_CHAR;
	    skip_space(fn);
	    w = read_list(fn);
	    return list_to_vector(w);

	case ']':
	case ')':
	    curchar = NOT_CHAR;
	    continue;		/* 
				 * ignore it 
				 */

	case '\'':
	    curchar = NOT_CHAR;
	    w = read_sex(fn);
	    return Llist2(quote_symbol, w);

	case '`':
	    curchar = NOT_CHAR;
	    w = read_sex(fn);
	    return backquote_expander(w);

	case ',':
	    fn();
	    if (curchar == '@') {
		curchar = NOT_CHAR;
		w = read_sex(fn);
		return Llist2(comma_at_symbol, w);
	    } else {
		w = read_sex(fn);
		return Llist2(comma_symbol, w);
	    }

	case '.':
	    curchar = NOT_CHAR;
	    continue;

	case '#':
	    return read_sharp(fn);

	case '"':
	    boffop = 0;
	    while (YES) {
		fn();
		if (curchar == EOF || curchar == CTRL_D) {
		    return CHAR_EOF;
		}
		while (curchar != '"' &&
		       curchar != EOF && curchar != CTRL_D) {
		    packbyte(curchar);
		    fn();
		}
		fn();
		if (curchar == '"') {	/* 
					 * one more '"' 
					 */
		    packbyte(curchar);
		    continue;
		}
		return make_string2(boffo, boffop);
	    }

	case '+':
	case '-':
	    boffop = 0;
	    packbyte(curchar);
	    fn();
	    if (!isedigit(curchar))
		goto atom;
	    goto number;
	default:
	    boffop = 0;
	    if (isedigit(curchar)) {
	      number:
		w = read_number(fn);
		if (w != nil)
		    return w;
	    }
	  atom:
	    while (YES) {
		if (curchar == ESCAPE_CHAR)
		    fn();
		else if (is_delimit(curchar))
		    break;
		packbyte(curchar);
		fn();
	    }
	    boffo[boffop] = 0;
	    return intern(boffo, boffop);
	}
    }
}


static const char *reading_string = NULL;

static void
char_from_string(void)
{
    long            ch;
    if (reading_string == NULL) {
	curchar = EOF;
	return;
    }
#ifdef Kanji
    if ((ch = *reading_string++) == 0) {
	reading_string--;
	curchar = EOF;
	return;
    }
    if (!is1byte(ch)) {
	long            c = *reading_string++;
	if (c == 0) {
	    reading_string--;
	} else {
	    ch = (ch << 8) | (c & 0xff);
	}
    }
#else
    if ((ch = *reading_string++) == 0) {
	reading_string--;
	curchar = EOF;
	return;
    }
#endif
    curchar = ch;
}


static const char *command,
               *curr_cmd,
               *end_cmd;

#ifdef Android
int
getc_from_androidinput()
{
    if (curr_cmd > end_cmd) {
	return EOF;
    }
    if (*curr_cmd == '\0') {
	curr_cmd++;
	return ';';
    }
    return *curr_cmd++;
}

void
set_androidinput(char *p)
{
    command = p;
    curr_cmd = p;
    end_cmd = p + strlen(p);
}
#endif

void
read_from_command(void)
{
    if (curr_cmd > end_cmd) {
	curchar = EOF;
	return;
    }
    if (*curr_cmd == '\0') {
	curr_cmd++;
	// curchar = EOF;
	curchar = ';';
	return;
    }
    curchar = *curr_cmd++;
}

void
init_reading_command(const char *p)
{
    command = p;
    curr_cmd = p;
    end_cmd = p + strlen(p);
}


void
read_from_lispin(void)
{
    long            ch;
#ifdef Android
    ch = file_FILE(lisp_input) == androidin ?
	getc_from_androidinput() : my_getwc(file_FILE(lisp_input));
#else
    ch = file_FILE(lisp_input) ==
	stdin ? ttyget() : my_getwc(file_FILE(lisp_input));
#endif
    if (ch == '\n')
	file_line(lisp_input)++;
    if (ch == EOF || ch == CTRL_D) {
	curchar = EOF;
	return;
    }
    if (value(echo_symbol) != nil)
	file_putc(lisp_output) (ch);
#ifdef ASUKA
    else if (spool_file != NULL && value(eecho_symbol) != nil)
	my_putwc(ch, spool_file);
#else
    else if (spool_file != NULL)
	my_putwc(ch, spool_file);
#endif
    /*
     * fprintf(stderr, "read_from_lispin:%04x(%c)\n", ch, ch);
     */
    curchar = ch;
}

Sex
Lread()
{
    Sex             w;
    w = read_sex((getchar_fn *) read_from_lispin);
    return w;
}

Sex
Lsugar_read()
{
    Sex             w;
    push(current_package);
    current_package = value(current_sugar_package);
    current_charvec = package_charvec(current_package);
    w = read_sex((getchar_fn *) read_from_lispin);
    pop(current_package);
    current_charvec = package_charvec(current_package);
    return w;
}



Sex
Lread_from_string(Sex a)
{
    char           *p;
    Sex             q;
    long            len;
    extern char    *read_buffer;

    if (is_symbol(a)) {
	p = pname(a);
	len = symbol_length(a);
    } else if (is_string(a)) {
	p = &string_value(a, 0);
	len = string_length(a);
    } else
	return nil;
    if (len > BUFFER_SIZE)
	len = BUFFER_SIZE;
    memcpy(read_buffer, p, len);
    read_buffer[len] = 0;
    q = read_from_string((const char *) read_buffer);
    return q;
}

Sex
read_from_string(const char *p)
{
    Sex             w;
    long            ch = curchar;

    reading_string = p;
    w = read_sex((getchar_fn *) char_from_string);
    reading_string = NULL;
    curchar = ch;
    return w;
}


static void
char_from_list(void)
{
    Sex             ch;
    if (!is_cons(io_workspace2)) {
	curchar = EOF;
	return;
    }
    ch = car(io_workspace2);
    io_workspace2 = cdr(io_workspace2);
#ifdef Kanji
    if (is_symbol(ch)) {
	curchar = pname(ch)[0];
	if ((is2byte(curchar)) && pname(ch)[1])
	    curchar = (curchar << 8) + pname(ch)[1];
    } else if (is_fixnum(ch))
	curchar = mkint(ch) & 0xffff;
    else
	curchar = EOF;		/* 
				 * Bad item in the list 
				 */
#else
    if (is_symbol(ch))
	curchar = pname(ch)[0];
    else if (is_fixnum(ch))
	curchar = (char) mkint(ch);
    else
	curchar = EOF;		/* 
				 * Bad item in the list 
				 */
#endif
}

static void
int_from_list(void)
{
    Sex             ch;
    if (!is_cons(io_workspace2)) {
	curchar = EOF;
	return;
    }
    ch = car(io_workspace2);
    io_workspace2 = cdr(io_workspace2);
#ifdef Kanji
    if (is_fixnum(ch))
	curchar = mkint(ch);
    else
	curchar = EOF;		/* 
				 * Bad item in the list 
				 */
#else
    if (is_fixnum(ch))
	curchar = (char) mkint(ch);
    else
	curchar = EOF;		/* 
				 * Bad item in the list 
				 */
#endif
}

Sex
Lstrcat(Sex a, Sex b)
{
    long            savecur = curchar;

    stackcheck();
    push2(a, b);
    stack[-1] = Lexplodec(a);
    b = stack[0];
    stack[0] = Lexplodec(b);
    a = Lnconc(stack[-1], stack[0]);
    popv(2);

    io_workspace2 = a;
    boffop = 0;
    char_from_list();
    while (curchar != EOF) {
	packbyte(curchar);
	char_from_list();
    }
    a = make_string2(boffo, boffop);
    io_workspace2 = nil;
    curchar = savecur;
    return a;
}

Sex
Lcompress(Sex stream)
{
    long            savecur = curchar;
    Sex             w;
    io_workspace2 = stream;
    boffop = 0;
    char_from_list();
    if (curchar == '"') {
      again:
	char_from_list();
	while (curchar != '"' && curchar != EOF) {
	    packbyte(curchar);
	    char_from_list();
	}
	char_from_list();
	if (curchar == '"') {
	    packbyte(curchar);
	    goto again;
	}
	w = make_string2(boffo, boffop);
    } else if (isedigit(curchar)) {
	w = read_sex((getchar_fn *) char_from_list);
    } else {
	while (curchar != EOF) {
	    if (curchar == ESCAPE_CHAR) {
		char_from_list();
		if (curchar == EOF)
		    break;
	    }
	    packbyte(curchar);
	    char_from_list();
	}
	w = getvector(TAG_SYMBOL, T_SYMBOL,
		      sizeof(Symbol_Head) - CELL + boffop);
	value(w) = undefined_var;
	memcpy(pname(w), boffo, boffop);
	plist(w) = nil;
	proc(w) = (lisp_proc *) undefined_proc;
	qnargs(w) = 0;
#ifdef ALGEBRAIC_PROPERTY
	gvalue(w) = undefined_var;
	gtag(w) = 0;
	gproc(w) = undefined_proc;
	gnargs(w) = 0;
#endif
    }
    io_workspace2 = nil;
    curchar = savecur;
    return w;
}

#if 1
Sex
Llist2string(Sex stream)
{
    Sex             w;
    Sex             ch;
    long            n,
                    k;
    char           *s;

    n = 0;
    w = stream;
    while (is_cons(w)) {
	n++;
	w = cdr(w);
    }
    stackcheck();
    push(stream);
    w = getvector(TAG_BINARY, T_STRING, HDR + n);
    k = (n + (SEX - 1)) & -(2 * SEX);
    s = (char *) w - TAG_BINARY;
    *(long *) (s + k + 2 * SEX) = 0;
    if (k)
	*(long *) (s + k + SEX) = 0;
    pop(stream);
    s += HDR;
    while (is_cons(stream)) {
	ch = car(stream);
	stream = cdr(stream);
#ifdef Kanji
	if (is_symbol(ch)) {
	    char            c;
	    c = pname(ch)[0];
	    *s++ = c;
	    if (is2byte(c) && pname(ch)[1])
		*s++ = pname(ch)[1];
	}
#else
	if (is_symbol(ch))
	    *s++ = pname(ch)[0];
#endif
	else if (is_fixnum(ch))
	    *s++ = mkint(ch) & 0xff;
    }
    return w;
}

#else

Sex
Llist2string(Sex stream)
{
    long            savecur = curchar;
    Sex             w;
    io_workspace2 = stream;
    boffop = 0;
    int_from_list();
    if (curchar == '"') {
      again:
	int_from_list();
	while (curchar != '"' && curchar != EOF) {
	    packbyte(curchar);
	    int_from_list();
	}
	int_from_list();
	if (curchar == '"') {
	    packbyte(curchar);
	    goto again;
	}
	w = make_string2(boffo, boffop);
    } else if (isedigit(curchar)) {
	w = read_sex((getchar_fn *) int_from_list);
    } else {
	while (curchar != EOF) {
	    if (curchar == ESCAPE_CHAR) {
		int_from_list();
		if (curchar == EOF)
		    break;
	    }
	    packbyte(curchar);
	    int_from_list();
	}
	w = getvector(TAG_SYMBOL, T_SYMBOL,
		      sizeof(Symbol_Head) - CELL + boffop);
	value(w) = undefined_var;
	memcpy(pname(w), boffo, boffop);
	plist(w) = nil;
	proc(w) = (lisp_proc *) undefined_proc;
	qnargs(w) = 0;
#ifdef ALGEBRAIC_PROPERTY
	gvalue(w) = undefined_var;
	gtag(w) = 0;
	gproc(w) = undefined_proc;
	gnargs(w) = 0;
#endif
    }
    io_workspace2 = nil;
    curchar = savecur;
    return w;
}
#endif

Sex
Lmake_id(Sex stream)
{
    long            savecur = curchar;
    Sex             w;
    io_workspace2 = stream;
    boffop = 0;
    char_from_list();
    if (curchar == '"') {
      again:
	char_from_list();
	while (curchar != '"' && curchar != EOF) {
	    packbyte(curchar);
	    char_from_list();
	}
	char_from_list();
	if (curchar == '"') {
	    packbyte(curchar);
	    goto again;
	}
    } else {
	while (curchar != EOF) {
	    if (curchar == ESCAPE_CHAR) {
		char_from_list();
		if (curchar == EOF)
		    break;
	    }
	    packbyte(curchar);
	    char_from_list();
	}
    }
    boffo[boffop] = 0;
    w = intern(boffo, boffop);
    io_workspace2 = nil;
    curchar = savecur;
    return w;
}

Sex
Lmake_uid(Sex stream)
{
    long            savecur = curchar;
    Sex             w;
    io_workspace2 = stream;
    boffop = 0;
    char_from_list();
    if (curchar == '"') {
      again:
	char_from_list();
	while (curchar != '"' && curchar != EOF) {
	    packbyte(curchar);
	    char_from_list();
	}
	char_from_list();
	if (curchar == '"') {
	    packbyte(curchar);
	    goto again;
	}
    } else {
	while (curchar != EOF) {
	    if (curchar == ESCAPE_CHAR) {
		char_from_list();
		if (curchar == EOF)
		    break;
	    }
	    packbyte(curchar);
	    char_from_list();
	}
    }
    boffo[boffop] = 0;
    w = gensym2(boffo);
    io_workspace2 = nil;
    curchar = savecur;
    return w;
}

void
read_eval_print(void)
{
    Sex            *save,
                   *bsave;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    save = stack;
    bsave = bstack;
    while (YES) {		/* 
				 * Loop for each s-expression found 
				 */
	Sex             u;
	miscflags |= MESSAGES_FLAG;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    continue;
	} else {
	    errorset_buffer = &this_level;
	    u = Lread();
	    if (u != CHAR_EOF)
		u = Lprint(u);
	}

	if (u == CHAR_EOF) {
	    errorset_buffer = saved_buffer;
	    return;
	}
	miscflags |= MESSAGES_FLAG;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    if (spool_file != NULL)
		fflush(spool_file);
	    continue;
	}
	u = eval(u, nil);
	Lprint(u);
    }
}



Sex
Levload(Sex a)
{
    Sex             c;
    stackcheck();
    while (is_cons(a)) {
	c = car(a);
	a = cdr(a);
	push(a);
	if (!is_symbol(c)) {
	    warnings("not a package name", a, nil);
	} else {
	    load_module(c);
	}
	pop(a);
    }
    return nil;
}

#ifdef REDUCE
long
preloaded_module(char *name)
{
    packages       *p = preloaded;
    long            r;
    char            c;
    char           *q;

    c = *name;
    while ((q = (p)->pkg_name) != NULL) {
	if (*q != c) {
	    p++;
	    continue;
	}
	if (strcmp(p->pkg_name, name) == 0) {
	    if (p->already_loaded)
		return p->already_loaded;	/* already loaded package */
	    /*
	     *  printf("Initialize preloaded package %s\n", name); 
	     */
	    r = (p->init_fn) ();
	    if (r == 0)
		return 0;
	    p->already_loaded = 1;
	    return 1;
	}
	p++;
    }
    return 0;
}

#endif

#if defined(UNIX)
#include <dlfcn.h>
Sex
load_module(Sex file)
{
    long            len = symbol_length(file);
    char           *p = pname(file);
    char            buf[64] = "init_";
    void            (*fn) (void);
    void           *fd;
    long            r;
    /*
     * printf("Dlopen %s\n", p);                
     */
    strncpy(buf, p, len);
    buf[len] = 0;
#ifdef REDUCE
    stackcheck();
    push(file);
    r = preloaded_module(buf);
    pop(file);
    if (r != 0)
	return nil;
    return CC_bloadKpackage(file);
#endif
#ifdef RTLD_NOW
    strcpy(buf + len, ".so");
    if ((fd = dlopen(buf, RTLD_NOW)) == NULL) {
	strcpy(buf, FAPLIB_PATH);
	strncpy(buf + strlen(buf), p, len);
	strcpy(buf + strlen(buf), ".so");
	if ((fd = dlopen(buf, RTLD_NOW)) == NULL) {
	    p = (char *) dlerror();
	    if (p == NULL)
		p = "dlopen failed";
	    error(p);
	    return nil;
	}
    }
#endif
    strcpy(buf, "init_");
    strncpy(buf + 5, p, len);
    buf[5 + len] = 0;
    fn = dlsym(fd, buf);
    (*fn) ();
    return nil;
}

#else
Sex
load_module(Sex file)
{
    long            len = symbol_length(file);
    char           *p = pname(file);
    char            buf[64] = "init_";
    long            r;
    /*
     * void (*fn) (); 
     */
    /*
     * void *fd; 
     */
    /*
     * printf("Dlopen %s\n", p);                
     */
    strncpy(buf, p, len);
    buf[len] = 0;
#ifdef REDUCE
    stackcheck();
    push(file);
    r = preloaded_module(buf);
    pop(file);
    if (r != 0)
	return nil;
    return CC_bloadKpackage(file);
#endif
#ifdef RTLD_NOW
    strcpy(buf + len, ".so");
    if ((fd = dlopen(buf, RTLD_NOW)) == NULL) {
	strcpy(buf, FAPLIB_PATH);
	strncpy(buf + strlen(buf), p, len);
	strcpy(buf + strlen(buf), ".so");
	if ((fd = dlopen(buf, RTLD_NOW)) == NULL) {
	    p = (char *) dlerror();
	    if (p == NULL)
		p = "dlopen failed";
	    error(p);
	    return nil;
	}
    }
    strcpy(buf, "init_");
    strncpy(buf + 5, p, len);
    buf[5 + len] = 0;
    fn = dlsym(fd, buf);
    (*fn) ();
    return nil;
#else
    p = "dlopen failed";
    error(p);
    return nil;
#endif
}
#endif

Sex
Lload(Sex file)
{
    Sex            *save,
                   *bsave;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    Sex             r = nil;

#ifdef  UNIX
#ifdef RTLD_NOW
    if (is_string(file)) {
	long            len = string_length(file);
	char           *p = &string_value(file, 0);
	char            buf[32] = "init_";
	void            (*fn) (void);
	void           *fd;
	if (strcmp(&p[len - 3], ".so") == 0) {
	    printf("Dlopen %s\n", p);
	    if ((fd = dlopen(p, RTLD_NOW)) == NULL) {
		p = (char *) dlerror();
		if (p != NULL)
		    printf("Dlopen failed by '%s'\n", p);
	    } else {
		strncpy(buf + 5, p, len - 3);
		buf[5 + len - 3] = 0;
		fn = dlsym(fd, buf);
		(*fn) ();
	    }
	    return nil;
	}
    }
#endif
#endif
    stackcheck();
    if (file != nil) {
	push(file);
	r = Lopen(file, input_symbol);
	r = Lrds(r);
	pop(file);
	if (miscflags & MESSAGES_FLAG) {
	    push(r);
	    lisp_printf("\nLoading ");
	    prin_to_stdout(file);
	    lisp_printf("\n");
	    pop(r);
	}
    }
    push(r);
    save = stack;
    bsave = bstack;
    while (YES) {
	Sex             u;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    continue;
	} else {
	    errorset_buffer = &this_level;
	    u = Lread();
	}

	if (u == CHAR_EOF) {
	    errorset_buffer = saved_buffer;
	    break;
	}
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    if (spool_file != NULL)
		fflush(spool_file);
	    continue;
	}
	if (miscflags & LOAD_ECHO_FLAG)
	    u = Lprint(u);
	u = eval(u, nil);
	if (miscflags & LOAD_ECHO_FLAG) {
	    Lprint(u);
	}
    }
    pop(r);
    if (r != nil) {
	push(r);
	Lclose(lisp_input);
	pop(r);
	Lrds(r);
    }
    return nil;
}

Sex
Lfload(Sex file)
{
    Sex            *save,
                   *bsave;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    Sex             r = nil;
    Sex             u;

    stackcheck();
    if (file != nil) {
	push(file);
	r = Lopen(file, input_symbol);
	r = Lrds(r);
	pop(file);
	if (miscflags & MESSAGES_FLAG) {
	    push(r);
	    lisp_printf("\nLoading ");
	    prin_to_stdout(file);
	    lisp_printf("\n");
	    pop(r);
	}
    }
    push(r);
    save = stack;
    bsave = bstack;
    while (YES) {
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    continue;
	} else {
	    errorset_buffer = &this_level;
	    u = Lread();
	}

	if (u == CHAR_EOF) {
	    errorset_buffer = saved_buffer;
	    break;
	}
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    if (spool_file != NULL)
		fflush(spool_file);
	    continue;
	}
	u = Lprint(u);
	u = eval(u, nil);
	Lprint(u);
    }
    pop(r);
    if (r != nil) {
	push(r);
	Lclose(lisp_input);
	pop(r);
	Lrds(r);
    }
    return nil;
}

Sex
Lspool(Sex file)
{
    char            filename[FILENAME_MAX];
    char           *p;
    long            len;
    if (spool_file != NULL) {
	fclose(spool_file);
	spool_file = NULL;
    }
    if (file == nil)
	return true;
    if (is_symbol(file)) {
	p = pname(file);
	len = symbol_length(file);
    } else if (!is_string(file)) {
	return aerror("spool");
    } else {
	p = &string_value(file, 0);
	len = string_length(file);
    }
    spool_file = open_file(filename, p, (size_t) len, "w", NULL);
    return Bool_to_Sex(spool_file != NULL);
}



Sex
Lsetpchar(Sex a)
{
    long            l;
    Sex             old = prompt_message;
    prompt_message = a;
    if (is_string(a)) {
	l = string_length(a);
	if (l > 250)
	    l = 250;
	strncpy(prompt_string, &string_value(a, 0), l);
	prompt_string[l] = 0;
    } else if (is_symbol(a)) {
	l = symbol_length(a);
	if (l > 250)
	    l = 250;
	strncpy(prompt_string, pname(a), l);
	prompt_string[l] = 0;
    }
    return old;
}


Sex
Lreadch()
{
    Sex             w;
    long            ch;
    if (curchar == NOT_CHAR) {
#ifdef Android
	ch = file_FILE(lisp_input) == androidin ?
	    getc_from_androidinput() : my_getwc(file_FILE(lisp_input));
#else
	ch = file_FILE(lisp_input) == stdin ?
	    ttyget() : my_getwc(file_FILE(lisp_input));
#endif
    } else {
	ch = curchar;
	curchar = NOT_CHAR;
    }
    /*
     * fprintf(stderr, "read:%c\n", ch);
     */
    if (ch == '\n')
	file_line(lisp_input)++;
    if (ch == EOF || ch == CTRL_D)
	w = CHAR_EOF;
    else {
	if (value(echo_symbol) != nil)
	    file_putc(lisp_output) (ch);
#ifdef ASUKA
	else if (spool_file != NULL && value(eecho_symbol) != nil)
	    my_putwc(ch, spool_file);
#else
	else if (spool_file != NULL)
	    my_putwc(ch, spool_file);
#endif
#ifdef Kanji
	if (iskchar(ch)) {
	    boffo[0] = ch >> 8;
	    boffo[1] = ch;
	    boffo[2] = 0;
	    w = intern(boffo, 2);
	} else if ((w = CHAR_OF(ch & 0xff)) == nil) {
	    lisp_printf("char code(%d) not installed\n", ch);
	}
#else
	w = CHAR_OF(ch & 0xff);
	if (w == nil) {
	    lisp_printf("char code(%d) not installed\n", ch);
	}
#endif
    }
    return w;
}


static          Sex
Lread_precision(Sex a)
{
    long            old = read_precision;
    if (a == nil)
	return mkfxat(old);
    if (!is_fixnum(a))
	return aerror("read-precision");
    read_precision = mkint(a);
    return mkfxat(old);
}

Sex
Lmodule(Sex a)
{
    return a;
}

Sex
Lendmodule()
{
    return nil;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
#ifdef ASUKA

/*
 * Sex             yylval;
 */

static          Sex
read_asuka_number_token(getchar_fn * fn)
{
    if (curchar == '0') {
	packbyte(curchar);
	fn();
	if (curchar == 'x') {
	    packbyte(curchar);
	    fn();
	    while (isexdigit(curchar)) {
		packbyte(curchar);
		fn();
	    }
	    yylval.sex = read_hexnum();
	    return NUMBER;
	}
    }
    while (isedigit(curchar)) {
	packbyte(curchar);
	fn();
    }
    if (curchar == '.') {
	packbyte(curchar);
	fn();
	while (isedigit(curchar)) {
	    packbyte(curchar);
	    fn();
	}
	if (curchar == 'e' || curchar == 'E') {
	    packbyte('e');
	    fn();
	    if (curchar == '+' || curchar == '-') {
		packbyte(curchar);
		fn();
	    }
	    while (isedigit(curchar)) {
		packbyte(curchar);
		fn();
	    }
	}
	// yylval.sex = read_bfloat();
	yylval.sex = read_bfloat();
	return NUMBER;
    }
#if 0
    else if (curchar == '/') {
	packbyte(curchar);
	fn();
	while (isedigit(curchar)) {
	    packbyte(curchar);
	    fn();
	}
	yylval.sex = read_rational();
	return NUMBER;
    }
#endif
    yylval.sex = read_fixnum();
    return NUMBER;
}


#if 0
static          Sex
read_asuka_sharp_token(getchar_fn * fn)
{
    Sex             x;

    boffop = 0;
    while (YES) {
	fn();
	if (curchar == ESCAPE_CHAR)
	    fn();
	else if (is_gdelimit(curchar))
	    break;
	packbyte(curchar);
    }
    if (boffop == 0) {
	switch (curchar) {
	case '[':
	    yylval.sex = hash_bracket_symbol;
	    curchar = NOT_CHAR;
	    return HASH_BRACKET;
	case '(':
	    yylval.sex = hash_paren_symbol;
	    curchar = NOT_CHAR;
	    return HASH_PAREN;
	default:
	    yylval.sex = LISP_CHAR_OF('#' & 0xff);
	    return LITERAL;
	}
    }
    boffo[boffop] = 0;
    x = intern(boffo, boffop);
    yylval.sex = make_polynomial(x, mkfxat(1), mkfxat(1));
    return LITERAL;
}
#endif


struct resword {
    char           *word;
    int             token;
};


struct resword  resword_table[] = {
    {"by", BY},
    {"bye", QUIT},
    {"do", DO},
    {"done", DONE},
    {"elif", ELIF},
    {"else", ELSE},
    {"end", END},
    {"fi", FI},
    {"for", FOR},
    {"foreach", FOREACH},
    {"from", FROM},
    {"if", IF},
    {"in", IN},
    {"on", ON},
    {"disable", DISABLE},
    {"enable", ENABLE},
    {"local", LOCAL},
    {"global", GLOBAL},
    {"od", OD},
    {"option", OPTION},
    {"options", OPTIONS},
    {"proc", PROC},
    {"quit", QUIT},
    {"write", WRITE},
    {"writen", WRITEN},
    {"debug", XDEBUG},
    {"debugn", XDEBUGN},
    {"printf", PRINTF},
    {"read", READ},
    {"save", SAVE},
    {"load", LOAD},
    {"stop", STOP},
    {"then", THEN},
    {"to", TO},
    {"downto", DOWNTO},
    {"while", WHILE},
    {"mod", MOD},
    {"smod", SMOD},
    {"union", UNION},
    {"minus", SET_DIFF},
    {"intersect", INTERSECT},
    {"and", AND},
    {"or", OR},
    {"not", NOT},
    {"repeat", REPEAT},
    {"until", UNTIL},
    {"switch", SWITCH},
    {"hctiws", HCTIWS},
    {"begin", BEGIN},
    {"return", RETURN},
    {"break", BREAK},
    {"continue", CONTINUE},
    {"array", ARRAY},
    {"vector", VECTOR},
    {"matrix", MATRIX},
    {"algebraic", ALGEBRAIC},
    {"with", WITH},
    {"clear", CLEAR},
    {"setorder", SETORDER},
    {"lex", ORDER_LEX},
    {"tlex", ORDER_TLEX},
    {"revlex", ORDER_RLEX},
    {"gradings", ORDER_GRADINGS},
    {"import", IMPORT},
    {"sum", SUM},
    {"product", PRODUCT},
    {"collect", COLLECT},
    {"join", JOIN},
    {"conc", CONC},
    {"try", TRY},
    {"throw", THROW},
    {"catch", CATCH},
    {"case", CASE},
    {"esac", ESAC},
    {"of", OF},
    {"default", DEFAULT},
    /*
     * {"error", ERROR}, 
     */
    {NULL, 0}
};

int
symbol_or_resword(char *b)
{
    struct resword *p = resword_table;
    /*
     * yylval.sex = w; 
     */
    while (p->word) {
	if (strcmp(b, p->word) == 0)
	    return p->token;
	p++;
    }
    return SYMBOL;
}


int
read_asuka_token(getchar_fn * fn)
{
    Sex             w;
    int             t;
    int             saved_char = NOT_CHAR;

  again:
    while (YES) {
	while (curchar == NOT_CHAR || is_gspace(curchar))
	    fn();

	switch (curchar) {
	case EOF:
	case CTRL_D:
	    curchar = NOT_CHAR;
	    yylval.sex = CHAR_EOF;
	    return EOF_TOKEN;

	case ';':
	    saved_char = curchar;
	    fn();
	    if (curchar == ';') {
		curchar = NOT_CHAR;
		yylval.sex = semicolon_semicolon_symbol;
		return SEMICOLON_SEMICOLON;
	    }
	    yylval.sex = LISP_CHAR_OF(curchar & 0xff);
	    return saved_char;
	case '!':
	case '=':
	case '(':
	case ')':
	case '[':
	case ']':
	case '{':
	case '}':
	case '`':
	case '|':
	case '\\':
	case '#':
	case '?':
	case '+':
	case '^':
	case '$':
	case ',':
	case '%':
	    yylval.sex = LISP_CHAR_OF(curchar & 0xff);
	    saved_char = curchar;
	    curchar = NOT_CHAR;
	    return saved_char;
	case '/':
	    saved_char = curchar;
	    fn();
	    if (curchar == '*') {
		while (YES) {
		    fn();
		    while (curchar != '*' && curchar != EOF
			   && curchar != CTRL_D)
			fn();
		    while (curchar == '*')
			fn();

		    if (curchar == '/') {	/* end of comment */
			curchar = NOT_CHAR;
			break;
		    }
		    if (curchar == EOF || curchar == CTRL_D) {
			curchar = NOT_CHAR;
			yylval.sex = CHAR_EOF;
			return EOF_TOKEN;
		    }
		}
		continue;
	    } else if (curchar == '/') {
		fn();
		while (curchar != EOF && curchar != CTRL_D
		       && curchar != '\n')
		    fn();
		if (curchar == EOF || curchar == CTRL_D) {
		    curchar = NOT_CHAR;
		    yylval.sex = CHAR_EOF;
		    return EOF_TOKEN;
		}
		curchar = NOT_CHAR;
		continue;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
#if 0
	case '\'':
	    curchar = NOT_CHAR;
	    yylval.sex = quote_symbol;
	    return QUOTE;
#else
	case '\'':
	    boffop = 0;
	    while (YES) {
		fn();
		if (curchar == EOF || curchar == CTRL_D) {
		    return CHAR_EOF;
		}
		while (curchar != '\'' && curchar != EOF
		       && curchar != CTRL_D) {
		    packbyte(curchar);
		    fn();
		}
		fn();
		if (curchar == '\'') {
		    packbyte(curchar);
		    continue;
		}
		break;
	    }
	    boffo[boffop] = 0;
	    yylval.sex = make_symbol(boffo);
	    return QUOTE_SYMBOL;
#endif
	case '*':
	    saved_char = curchar;
	    fn();
	    if (curchar == '*') {
		curchar = NOT_CHAR;
		yylval.sex = LISP_CHAR_OF('^' & 0xff);
		return '^';	/* 
				 * POWER 
				 */
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '@':
	    saved_char = curchar;
	    fn();
	    if (curchar == '@') {
		curchar = NOT_CHAR;
		yylval.sex = itrsyth_symbol;
		return ITRSYTH;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '&':
	    saved_char = curchar;
	    fn();
	    if (curchar == '*') {
		curchar = NOT_CHAR;
		yylval.sex = noncomu_symbol;
		return NONCOMU;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '.':
	    saved_char = curchar;
	    fn();
	    if (curchar == '.') {
		curchar = NOT_CHAR;
		yylval.sex = omit_symbol;
		return OMIT;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '<':
	    saved_char = curchar;
	    fn();
	    if (curchar == '=') {
		curchar = NOT_CHAR;
		yylval.sex = lesser_equal_symbol;
		return LESSER_EQUAL;
	    } else if (curchar == '>') {
		curchar = NOT_CHAR;
		yylval.sex = not_equal_symbol;
		return NOT_EQUAL;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '>':
	    saved_char = curchar;
	    fn();
	    if (curchar == '=') {
		curchar = NOT_CHAR;
		yylval.sex = greater_equal_symbol;
		return GREATER_EQUAL;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case ':':
	    saved_char = curchar;
	    fn();
	    if (curchar == '=') {
		curchar = NOT_CHAR;
		yylval.sex = colon_equal_symbol;
		return COLON_EQUAL;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '-':
	    saved_char = curchar;
	    fn();
	    if (curchar == '>') {
		curchar = NOT_CHAR;
		yylval.sex = arrowope_symbol;
		return ARROWOPE;
	    }
	    yylval.sex = LISP_CHAR_OF(saved_char & 0xff);
	    return saved_char;
	case '"':
	    boffop = 0;
	    while (YES) {
		fn();
		if (curchar == EOF || curchar == CTRL_D) {
		    return CHAR_EOF;
		}
		while (curchar != '"' && curchar != EOF
		       && curchar != CTRL_D) {
		    packbyte(curchar);
		    fn();
		}
		fn();
		if (curchar == '"') {
		    packbyte(curchar);
		    continue;
		}
		break;
	    }
	    yylval.sex = make_string2(boffo, boffop);
	    return STRING;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    boffop = 0;
	    return read_asuka_number_token(fn);
	case '_':
	    fn();
	    if (!iskalpha(curchar)) {
		yylval.sex = LISP_CHAR_OF('_' & 0xff);
		return '_';
	    }
	    boffop = 0;
	    packbyte('_');
	    goto name;
	default:
	    if (!iskalpha(curchar)) {
		yylval.sex = LISP_CHAR_OF(curchar & 0xff);
		while (curchar != ';')
		    fn();
		printf("\n Illegal character. Skipping until ';'");
		yylval.sex = LISP_CHAR_OF(';' & 0xff);
		curchar = NOT_CHAR;
		return UNKNOWN;
	    }
	    boffop = 0;
	  name:
	    while (iskalnum(curchar) || curchar == '_') {
		packbyte(curchar);
		fn();
	    }
	    if (curchar == ':' && boffop == 4
		&& strncmp(boffo, "lisp", 4) == 0) {
		boffop = 0;
		fn();
		while (iskalnum(curchar) || curchar == '_') {
		    packbyte(curchar);
		    fn();
		}
		boffo[boffop] = 0;
		push(current_package);
		current_package = value(current_sugar_package);
		current_charvec = package_charvec(current_package);
		w = intern(boffo, boffop);
		pop(current_package);
		current_charvec = package_charvec(current_package);
		t = symbol_or_resword(boffo);
		yylval.sex = w;
		return t;
	    }
	    boffo[boffop] = 0;
	    t = symbol_or_resword(boffo);
	    if (t == SYMBOL) {
		w = intern(boffo, boffop);
	    } else {
		push(current_package);
		current_package = value(current_sugar_package);
		current_charvec = package_charvec(current_package);
		w = intern(boffo, boffop);
		pop(current_package);
		current_charvec = package_charvec(current_package);
	    }
	    yylval.sex = w;
	    return t;
	}
    }
}


int
yylex()
{
    int             w;
#ifdef Android
    w = read_asuka_token((getchar_fn *) read_from_command);
#else
    w = read_asuka_token((getchar_fn *) read_from_lispin);
#endif
    // printf("yylex: token is 0x%08x\n", w);
    return w;
}

Sex
Lread_asuka_token()
{
    read_asuka_token((getchar_fn *) read_from_lispin);

    return yylval.sex;
}


Sex
Lasuka_parse()
{
    Sex             parse_value = nil;
    extern Sex     *parse_value_ptr;

    parse_value_ptr = &parse_value;
    yyparse();
    return parse_value;
}

int             value_print_flag = 1;


void
asuka_read_eval_print(Sex flg)
{
    Sex             u;
    Sex            *save,
                   *bsave;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    int             f = flg == nil ? 0 : 1;
    double          clock1,
                    clock2;
    extern long     batch_mode;


    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    save = stack;
    bsave = bstack;

    while (YES) {		/* 
				 * Loop for each s-expression found 
				 */
	miscflags |= MESSAGES_FLAG;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    continue;
	} else {
	    errorset_buffer = &this_level;
	    u = Lasuka_parse();
	    if (run_mode & RUN_LBP)
		CC_prettyprint(u);
#ifdef DEBUG
	    Lprint(u);
#endif
	    if (u == bye_symbol)
		Lbye();
	}

	if (u == CHAR_EOF) {
	    // printf("end of file\n");
	    break;
	}
	miscflags |= MESSAGES_FLAG;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    if (spool_file != NULL)
		fflush(spool_file);
	    continue;
	}
	if (f == 0)
	    miscflags &= ~MESSAGES_FLAG;
	if (showtime) {
	    clock1 = get_time() - base_time - io_time - gc_time;
	    u = eval(u, nil);
	    clock2 = get_time() - base_time - io_time - gc_time;
	} else {
	    u = eval(u, nil);
	    clock1 = clock2 = 0;
	}
	miscflags |= MESSAGES_FLAG;
	value(thbeepre_symbol) = value(twbeepre_symbol);
	value(twbeepre_symbol) = value(onbeepre_symbol);
	value(onbeepre_symbol) = u;
#ifdef SAVE_WORK_SPACE
	if (!batch_mode) {
	    push(u);
	    work_space =
		cons(cons(mkfxat(work_space_count), u), work_space);
	    work_space_count++;
	    pop(u);
	}
#endif
	if (f && value_print_flag) {
	    // if (u != nil)
	    Aprint(u);
	}
	if (showtime)
	    lisp_printf("\n\tTime %fms\n", (clock2 - clock1) * 1000);
    }
    errorset_buffer = saved_buffer;
    pop(current_package);
    current_charvec = package_charvec(current_package);
}


void
asuka_read_compile(Sex flg)
{				/* incomplete 20007.10.23 */
    Sex             u;
    Sex            *save,
                   *bsave;
    jmp_buf         this_level,
                   *saved_buffer = errorset_buffer;
    int             f = flg == nil ? 0 : 1;


    push(current_package);
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    save = stack;
    bsave = bstack;

    while (YES) {		/* 
				 * Loop for each s-expression found 
				 */
	miscflags |= MESSAGES_FLAG;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    continue;
	} else {
	    errorset_buffer = &this_level;
	    u = Lasuka_parse();
	    Lprint(u);
	    if (u == bye_symbol)
		Lbye();
	}

	if (u == CHAR_EOF) {
	    break;
	}
	miscflags |= MESSAGES_FLAG;
	if (setjmp(this_level)) {
	    unwind_bind(bsave);
	    stack = save;
	    if (spool_file != NULL)
		fflush(spool_file);
	    continue;
	}
	u = eval(u, nil);
	value(thbeepre_symbol) = value(twbeepre_symbol);
	value(twbeepre_symbol) = value(onbeepre_symbol);
	value(onbeepre_symbol) = u;
	if (f && value_print_flag) {
	    if (u != nil)
		Lprint(u);
	}
    }
    errorset_buffer = saved_buffer;
    pop(current_package);
    current_charvec = package_charvec(current_package);
}

#endif
/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
Lshowtime(Sex a)
{
    if (a == nil)
	showtime = 0;
    else
	showtime = 1;
    return a;
}

Sex
Lecho(Sex a)
{
    if (a == nil)
	value(echo_symbol) = nil;
    else
	value(echo_symbol) = true;
    return nil;
}

Sex
Lgetpid()
{
#ifdef VisualC
    int             id;
    extern int      getpid();
#else
    pid_t           id;
#endif
    id = getpid();
    return mkfxat(id);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


setup_type const read_setup[] = {
    {"restart", (lisp_proc *) Lrestart, 1},
    {"set-read-precision", (lisp_proc *) Lread_precision, 1},
    {"batchp", (lisp_proc *) Lbatchp, 0},
    {"strcat", (lisp_proc *) Lstrcat, 2},
    {"compress", (lisp_proc *) Lcompress, 1},
    {"list2string", (lisp_proc *) Llist2string, 1},
    {"list_to_vector", (lisp_proc *) list_to_vector, 1},
    {"list2widestring", (lisp_proc *) Llist2string, 1},
    {"mk.id", (lisp_proc *) Lmake_id, 1},
    {"rseek", (lisp_proc *) Lrseek, 1},
    {"rtell", (lisp_proc *) Lrtell, 0},
    {"gensym", (lisp_proc *) Lgensym, 0},
    {"reset_gensym", (lisp_proc *) Lreset_gensym, 1},
    {"gensym1", (lisp_proc *) Lgensym1, 1},
    {"gensym2", (lisp_proc *) Lgensym2, 2},
    {"getenv", (lisp_proc *) Lgetenv, 1},
    {"intern", (lisp_proc *) Lintern, 1},
    {"internp", (lisp_proc *) Linternp, 1},
    {"gensymp", (lisp_proc *) Linternp, 1},
    {"orderp", (lisp_proc *) Lorderp, 2},
    {"rds", (lisp_proc *) Lrds, 1},
    {"read", (lisp_proc *) Lread, 0},
    {"sugar_read", (lisp_proc *) Lsugar_read, 0},
    {"read_from_string", (lisp_proc *) Lread_from_string, 1},
    {"setpchar", (lisp_proc *) Lsetpchar, 1},
    {"spool", (lisp_proc *) Lspool, 1},

    {"system", (lisp_proc *) Lsystem, 1},

  //  {"ordp", (lisp_proc *) Lorderp, 2},
    {"remob", (lisp_proc *) Lremob, 1},

    {"evload", (lisp_proc *) Levload, 1},
    {"load_module", (lisp_proc *) load_module, 1},

    {"load", (lisp_proc *) Lload, 1},
    {"fload", (lisp_proc *) Lfload, 1},
    {"readch", (lisp_proc *) Lreadch, 0},
#ifdef ASUKA
    {"read_token", (lisp_proc *) Lread_asuka_token, 0},
    {"asuka_parse", (lisp_proc *) Lasuka_parse, 0},
    {"asuka_symbol", (lisp_proc *) Lasuka_symbol, 1},
    {"sugar_symbol", (lisp_proc *) Lsugar_symbol, 1},
    {"lisp_symbol", (lisp_proc *) Lsugar_symbol, 1},
#endif
    {"file-line", (lisp_proc *) Lfile_line, 0},
    {"fluid-list", (lisp_proc *) Lfluid_list, 0},
    {"global-list", (lisp_proc *) Lglobal_list, 0},

    {"module", (lisp_proc *) Lmodule, 1},
    {"endmodule", (lisp_proc *) Lendmodule, 0},
    {"create-package", (lisp_proc *) Lmodule, 1},
    {"showtime", (lisp_proc *) Lshowtime, 1},
    {"echo", (lisp_proc *) Lecho, 1},
    {"getpid", (lisp_proc *) Lgetpid, 0},

    {NULL, 0, 0}
};
