/*
 * read.h               (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/read.h,v 2.23 2018/06/11 09:08:26 kako Exp $
 */

/*
 * FILE structure          
 */
#define FILE_HEADER (TAG_HDRS + T_FILE)


typedef struct SFILE {
    FILE           *fp;
    unsigned char  *data;
    int             type;	/* 0 for null, 1 for read from string 2
				 * for read/write file */
    int             size;
    int             pos;
} SFILE;

typedef struct File_Head {
    Tag             h;
    long            len;
    SFILE           sf;

    long            line;
    long            pos;	/* 
				 * file position 
				 */
    long            Char;	/* 
				 * look-ahead char 
				 */
    putchar_fn     *Fn;
} File_Head;

#define	_SFFILE	0x0001		/* read/write file */
#define	_SFPIPE	0x0002		/* read/write pipe */
#define	_SFSTR	0x0004		/* read/write string */
#define _SFRD	0x0008		/* read */
#define	_SFWR	0x0010		/* write */
#define	_SFRW	0x0020		/* read and write */
#define	_SFEOF	0x0040		/* eof */
#define	_SFERR	0x0080		/* error */
#define	_SFALL	0x0100		/* allocate space by malloc */

#ifdef Android
FILE            AndroidFile[3];
#define androidin	&AndroidFile[0]
#define androidout	&AndroidFile[1]
#define androiderr	&AndroidFile[2]
#endif

typedef void    getchar_fn(void);

extern void     char_to_stdout(long c);
extern void     char_to_stderr(long c);
extern void     char_to_debug(long c);

extern void     char_to_androidout(long c);
extern void     char_to_androiderr(long c);
extern void     init_androidout(void);
extern char    *get_androidout(void);

extern Sex      make_file_handle(FILE * f, long data, putchar_fn * fn);
extern Sex      make_string_handle(unsigned char *f, int size, int flag);

#define file_FILE(v)    (((File_Head *) ((char *)(v) - TAG_BINARY))->sf.fp)
#define file_char(v)    (((File_Head *) ((char *)(v) - TAG_BINARY))->Char)
#define file_putc(v)    (((File_Head *) ((char *)(v) - TAG_BINARY))->Fn)
#define file_pos(v)     (((File_Head *) ((char *)(v) - TAG_BINARY))->pos)
#define file_line(v)    (((File_Head *) ((char *)(v) - TAG_BINARY))->line)

#define file_type(v)    (((File_Head *) ((char *)(v) - TAG_BINARY))->sf.type)
#define file_SFILE(v)    (((File_Head *) ((char *)(v) - TAG_BINARY))->sf)

#define is_file(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_FILE)




#define  ESCAPE_CHAR '!'
#define CESCAPE_CHAR '\\'

#define NO_PREFIX           'x'

extern Sex      intern(char *str, long h);
extern Sex      Lopen(Sex name, Sex dir);
extern Sex      Lopen_file(Sex name, Sex dir);
extern Sex      read_from_string(const char *);

#ifdef Kanji
#define NOT_CHAR        0x40000
#define BREAK_CHAR	0x40001
#define BLANK_CHAR	0x40002
#else
#define NOT_CHAR        0x400
#define BREAK_CHAR	0x401
#define BLANK_CHAR	0x402
#endif

extern long     ttyget(void);
extern long     ttygetchar(char *);
extern bool     sigint_must_longjmp;
extern jmp_buf  sigint_buf;



#ifdef YYDEBUG
extern int      yydebug;
#endif

extern int      yylex(void);
extern Sex      yacc_warnings(char *, Sex, Sex);
extern int      yyparse(void);
extern int      value_print_flag;

/***************************************************************/
/***************************************************************/

Sex             make_literal(const char *p);
#ifdef ASUKA
Sex             make_asuka_literal(const char *p);
#endif
Sex             make_string(char *str);
Sex             make_string2(char *str, long n);
Sex             Lbatchp(void);
Sex             Lgetenv(Sex a);
#ifdef UNIX
Sex             Lsystem(Sex a);
#endif
void            close_all_files(void);
void            warm_start(void);
Sex             Lrestart(Sex a);
unsigned long   hash_string(char *s, long n);
Sex             make_fluid_symbol(char *s, Sex value);
Sex             make_symbol(char *s);
Sex             make_asuka_symbol(char *s);
Sex             make_asuka_fluid_symbol(char *s, Sex v);
Sex             make_proc_symbol(char *s, lisp_proc * f1, long nargs);
static bool     add_to_hash(Sex s, Sex vector, unsigned long hash);
static Sex      rehash(Sex v);
static Sex      lookup(char *str, long strsize, Sex v, unsigned long hash);
static long     ordersymbol(Sex v1, Sex v2);
static long     ordpv(Sex u, Sex v);
Sex             Lorderp(Sex a, Sex b);
Sex             Lfluid_list(void);
Sex             Lglobal_list(void);
static bool     remob(Sex sym, Sex v);
Sex             gensym2(char *name);
Sex             Lgensym(void);
Sex             Lreset_gensym(Sex);
Sex             Lgensym1(Sex a);
Sex             Lgensym2(Sex a, Sex b);
Sex             Lgen_sharp_symbol(Sex a);
Sex             intern(char *str, long h);
Sex             Lintern(Sex str);
Sex             Linternp(Sex str);
Sex             ndelete(Sex a, Sex l);
Sex             Lremob(Sex sym);
void            flush_tty(void);
long            ttyget(void);
Sex             Lfile_line(void);
Sex             Lrds(Sex a);
Sex             Lrtell(void);
Sex             Lrseek(Sex a);
static void     skip_space(getchar_fn * fn);
static Sex      read_list(getchar_fn * fn);
Sex      list_to_vector(Sex l);
static Sex      backquote_expander(Sex a);
static void     packbyte(long c);
static Sex      read_fixnum(void);
long            hex_to_num(long a);
static Sex      read_hexnum(void);
static Sex      read_rational(void);
static Sex      read_bfloat(void);
static Sex      read_float(void);
void            read_from_command(void);
void            init_reading_command(const char *p);
static Sex      read_number(getchar_fn * fn);
static Sex      read_sharp(getchar_fn * fn);
static Sex      read_sex(getchar_fn * fn);
static void     char_from_string(void);
void            read_from_lispin(void);
Sex             Lread(void);
Sex             Lsugar_read(void);
Sex             Lread_from_string(Sex a);
Sex             read_from_string(const char *p);
static void     char_from_list(void);
Sex             Lstrcat(Sex a, Sex b);
Sex             Lcompress(Sex stream);
Sex             Llist2string(Sex stream);
Sex             Lmake_id(Sex stream);
Sex             Lmake_uid(Sex stream);
void            read_eval_print(void);
Sex             Levload(Sex a);
#ifdef REDUCE
long            preloaded_module(char *name);
#endif
Sex             load_module(Sex file);
Sex             Lload(Sex file);
Sex             Lfload(Sex file);
Sex             Lspool(Sex file);
Sex             Lsetpchar(Sex a);
Sex             Lreadch(void);
static Sex      Lread_precision(Sex a);
Sex             Lmodule(Sex a);
Sex             Lendmodule(void);
#ifdef ASUKA
static Sex      read_asuka_number_token(getchar_fn * fn);
#if 0
static Sex      read_asuka_sharp_token(getchar_fn * fn);
#endif
int             symbol_or_resword(char *b);
int             read_asuka_token(getchar_fn * fn);
int             yylex(void);
Sex             Lread_asuka_token(void);
Sex             Lasuka_parse(void);
void            asuka_read_eval_print(Sex flg);
void            asuka_read_compile(Sex flg);
#endif
Sex             Lshowtime(Sex a);
Sex             Lecho(Sex a);
Sex             Lgetpid(void);

#ifdef ASUKA
Sex             Lsugar_symbol(Sex a);
Sex             Lasuka_symbol(Sex a);
#endif
/***************************************************************/
/***************************************************************/
