/*
 * sugar.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      sugar main  routine written by C lang.
 *      $Header: /Users/kako/asuka/src/RCS/sugar.c,v 2.33 2018/01/05 02:35:21 kako Exp kako $
 */

static char const rcsid[] =
    "$Id: sugar.c,v 2.33 2018/01/05 02:35:21 kako Exp kako $";


/*
 * #define GAL
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <math.h>

#include "config.h"

#ifdef USE_GMP
#include <gmp.h>
#ifdef USE_MPFR
#include <mpfr.h>
#endif
#endif

#include "tags.h"
#include "externs.h"
#include "arith.h"
#include "read.h"
#include "entries.h"
#include "errors.h"
#include "init.h"
#include "mt19937.h"


Sex             env;
Sex             nil;

long            core_size = CORE_SIZE;
long            cell_size;
long            fwd_size;
long            stack_size = STACK_SIZE;
long            bind_size = BIND_SIZE;

long            showlap = 0;

jmp_buf        *main_buffer;


void            erdn(void);
Sex             aerror2_a(char *s, char *m, Sex a, Sex b);
Sex             aerror2(char *s, Sex a, Sex b);
Sex             aerror1(char *s, Sex a);
Sex             aerror0(char *s);
Sex             warning2(char *s, va_list a);
Sex             warnings(char *s, Sex a, Sex b);
void            STACKCHECK(void);
void            stack_error(void);
#ifdef ASUKA
void            init_asuka(void);
#endif
#ifdef Android
char           *android_eval(const char *cmd);
void            android_init2(long ac, char *av[]);
extern void     android_error(char *fmt, char *msg);
extern void     asuka_warning(char *msg);
extern void     asuka_warning1(char *msg, char *t);
int             android_initialized = 0;
#endif
long            my_main(long argc, char *argv[]);
long            lisp_atoi(char *w);


#ifdef REDUCE
static char    *compile_file = NULL;
static char    *cobject_path = NULL;
#endif
#ifdef GAL
static char    *compile_file = NULL;
static char    *cobject_path = NULL;
#endif
#ifdef SLISP
static char    *compile_file = NULL;
static char    *cobject_path = NULL;
#endif
#ifdef ASUKA
static char    *compile_file = NULL;
static char    *cobject_path = NULL;
#endif
long            compile_flag = NO;

#ifdef _LP64
#define	LP_MODE	"(64)"
#else
#define LP_MODE	"(32)"
#endif

/*****************************************************************************/
/*
 * Error reporting and recovery                                         
 */
/*****************************************************************************/


void
erdn(void)
{
    /*
     * print backtrace 
     */
    longjmp(*errorset_buffer, 2);
}

void
buffer_over(void)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	asuka_warning("\n*** Input buffer overflow.\n");
#else
	lisp_printf("\n*** Input buffer overflow.\n");
#endif
    }
    erdn();
}
void
stack_over(void)
{
    if (miscflags & MESSAGES_FLAG) {
	if (already_in_gc == NO)
#ifdef Android
	    asuka_warning("\n*** Stack overflow\n");
#else
	    lisp_printf("\n*** Stack overflow\n");
#endif
	else
#ifdef Android
	    asuka_warning
		("\n*** Stack overflow while garbage collection\n");
#else
	    lisp_printf("\n*** Stack overflow while garbage collection\n");
#endif
    }
    erdn();
}

void
bind_over(void)
{
    if (miscflags & MESSAGES_FLAG) {
	if (already_in_gc == NO)
#ifdef Android
	    asuka_warning("\n*** Bind Stack overflow\n");
#else
	    lisp_printf("\n*** Bind Stack overflow\n");
#endif
	else
#ifdef Android
	    asuka_warning
		("\n*** Bind Stack overflow while garbage collection\n");
#else
	    lisp_printf
		("\n*** Bind Stack overflow while garbage collection\n");
#endif
    }
    erdn();
}


Sex
error(char *s)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s. \n", s);
#else
	lisp_printf("\n*** Error %s. \n", s);
#endif
    }
    erdn();
    return nil;
}

Sex
error1(char *s, Sex a)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s: \n", s);
#else
	lisp_printf("\n*** Error %s: %P\n", s, a);
#endif
    }
    erdn();
    return nil;
}

Sex
error2(char *s, Sex a, Sex b)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s: \n", s);
#else
	lisp_printf("\n*** Error %s: \n", s);
	push(b);
	lisp_printf("%P\n", a);
	pop(b);
	lisp_printf("         %P\n", b);
#endif
    }
    erdn();
    return nil;
}

void
stack_error()
{
}

void
STACKCHECK()
{
    Sex            *q;
    for (q = stackbase; q < stack; q++) {
	if (*q == 6) {
	    printf("found 6 at stack 0x%llx in (0x%llx, 0x%llx)\n",
		   (Sex) q, (Sex) stackbase, (Sex) stack);
	    stack_error();
	}
    }
}


Sex
interrupted(Sex p)
{
    char           *s = "\n *** Lisp attention !  ?(q/c/t) = ";

    switch (ttygetchar(s)) {
    case 'a':
    case 'A':
	break;
    default:
    case 'c':
    case 'C':			/* 
				 * proceed as if no interrupt 
				 */
	return p;
    case 't':
    case 'T':			/* 
				 * raise an exception 
				 */
	Lsetpchar(lisp_prompt_message);
	longjmp(*main_buffer, 1);
    case 'q':
    case 'Q':
	exit(EXIT_FAILURE);	/* 
				 * * Rather abrupt 
				 */
    }
    erdn();
    return nil;
}

Sex
aerror(char *s)
{
    if (miscflags & MESSAGES_FLAG)
#ifdef Android
	android_error("\n*** Error bad args for %s\n", s);
#else
	lisp_printf("\n*** Error bad args for %s\n", s);
#endif
    erdn();
    return nil;
}

Sex
aerror0(char *s)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s:\n", s);
#else
	lisp_printf("\n*** Error %s\n", s);
#endif
    }
    erdn();
    return nil;
}

Sex
aerror1(char *s, Sex a)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s:\n", s);
#else
	lisp_printf("\n*** Error %s:%p\n", s, a);
#endif
    }
    erdn();
    return nil;
}

Sex
aerror2(char *s, Sex a, Sex b)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s:\n", s);
#else
	push(b);
	lisp_printf("\n*** Error %s: %p ", s, a);
	pop(b);
	lisp_printf(" %p\n", b);
#endif
    }
    erdn();
    return nil;
}


Sex
aerror3(char *s, Sex a, Sex b, Sex c)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error("\n*** Error %s: \n", s);
#else
	push2(c, b);
	lisp_printf("\n*** Error %s: %p ", s, a);
	pop(b);
	lisp_printf("%p ", b);
	pop(c);
	lisp_printf("%p\n", c);
#endif
    }
    erdn();
    return nil;
}

Sex
aerror2_a(char *s, char *m, Sex a, Sex b)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	android_error2("\n*** Error %s at %s: \n", s, m);
#else
	push(b);
	lisp_printf("\n*** Error %s at %s: %p ", s, m, a);
	pop(b);
	lisp_printf(" %p\n", b);
#endif
    }
    erdn();
    return nil;
}


Sex
warning2(char *s, va_list a)
{

    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	asuka_warning(s);
#else
	lisp_printf(s, a);
#endif
    }
    return nil;
}

Sex
warnings(char *s, Sex a, Sex b)
{
    if (miscflags & MESSAGES_FLAG) {
#ifdef Android
	asuka_warning1("\n*** Warning %s: \n", s);
#else
	push(b);
	lisp_printf("\n*** Warning %s: %p", s, a);
	pop(b);
	lisp_printf(" %p\n", b);
#endif
    }
    return nil;
}

Sex
yacc_warnings(char *s, Sex a, Sex b)
{
#ifdef Android
    asuka_warning1("\n*** Parsing error %s: \n", s);
#else
    push(b);
    lisp_printf("\n*** Parsing error %s: %p", s, a);
    pop(b);
    lisp_printf(" %p", b);
    lisp_printf(" at line %d\n", file_line(lisp_input) + 1);
#endif
    erdn();
    return nil;
}

Sex
messages(char *s, Sex a, Sex b)
{
    if (value(nomess_symbol) == nil && (miscflags & MESSAGES_FLAG)) {
#ifdef Android
	asuka_message("\n*** Message %s: \n", s);
#else
	push(b);
	lisp_printf("\n*** Message %s: %p ", s, a);
	pop(b);
	lisp_printf(" %p\n", b);
#endif
    }
    return nil;
}


void
fatal_error(char *msg)
{
#ifdef Android
    android_error("*** Fatal error %s\n", msg);
#else
    lisp_printf("*** Fatal error %s\n", msg);
    if (spool_file != NULL)
	fclose(spool_file);
    if (debug_file != NULL)
	fclose(debug_file);
    fflush(stdout);
#if defined(VisualC) && !defined(_CONSOLE)
    fail("Fatal Error");
#endif
#endif
    exit(EXIT_FAILURE);
}


static long     return_code;
long            batch_mode;
#ifdef ASUKA
long            run_mode = RUN_ASUKA;
#else


#ifdef GAL
long            run_mode = RUN_GAL;
#else
#ifdef REDUCE
long            run_mode = RUN_REDUCE;
#else
#ifdef SLISP
long            run_mode = RUN_SLISP;
#else
long            run_mode = RUN_SUGAR;
#endif

#endif
#endif
#endif

static bool     stayslow = NO;

void
lisp_main(void)
{
    jmp_buf         this_level;
    extern long     trace_count;

    tty_count = 0;
    curchar = '\n';
    Lsetpchar(lisp_prompt_message);
    batch_mode = batchp();
    while (YES) {
	Sex            *save = stack;
	Sex            *bind_save = bstack;
	errorset_buffer = (jmp_buf *) (&this_level);
	trace_count = 1;
	main_buffer = errorset_buffer;
#ifdef REDUCE
	if (run_mode & RUN_COMPILER) {
	    push3(nil, nil, nil);

	    stack[0] = make_symbol("*stayslow");
	    value(stack[0]) = stayslow ? true : nil;
	    if (showlap) {
		stack[0] = make_symbol("*showlap");
		value(stack[0]) = showlap ? true : nil;
	    }
	    stack[0] = make_symbol("*mode");
	    stack[-1] = make_symbol("symbolic");
	    value(stack[0]) = stack[-1];

	    stack[0] = make_symbol(compile_file);
	    if (cobject_path)
		stack[-1] = make_symbol(cobject_path);
	    else
		stack[-1] = nil;
	    stack[-2] = make_symbol(" xfaslout");
	    if (cobject_path)
		lisp_printf
		    ("  *** Compiling file %s.red and generate %s/r_%s.c\n",
		     compile_file, cobject_path, compile_file);
	    else
		lisp_printf
		    ("  *** Compiling file %s.red and generate r_%s.c\n",
		     compile_file, compile_file);
	    if (setjmp(this_level))
		break;
	    apply2(stack[-2], stack[-1], stack[0]);
	    popv(3);
	    break;
	}
#endif
#ifdef GAL
	if (run_mode & RUN_COMPILER) {
	    char           *pp,
	                   *p;
	    push3(nil, nil, nil);
	    stack[0] = make_symbol(".stayslow");
	    value(stack[0]) = stayslow ? true : nil;
	    stack[0] = make_symbol(".procmode");
	    stack[-1] = make_symbol("lisp");
	    value(stack[0]) = stack[-1];
	    stack[0] = make_symbol(compile_file);
	    if (cobject_path)
		stack[-1] = make_symbol(cobject_path);
	    else
		stack[-1] = nil;
	    stack[-2] = make_symbol(" xfaslout");
	    pp = compile_file;
	    for (p = pp; *p; p++) {
		if (*p == '/')
		    pp = p + 1;
	    }
	    if (cobject_path)
		lisp_printf
		    ("  *** Compiling file %s.gal and generate %s/g_%s.c\n",
		     compile_file, cobject_path, pp);
	    else
		lisp_printf
		    ("  *** Compiling file %s.gal and generate g_%s.c\n",
		     compile_file, pp);
	    if (setjmp(this_level))
		break;
#if 0
	    lisp_printf("call apply2 with %p %p %p \n", stack[-2],
			stack[-1], stack[0]);
	    lisp_printf("call apply2 with %p %p %p \n", stack[-2],
			stack[-1], stack[0]);
#endif
	    apply2(stack[-2], stack[-1], stack[0]);
	    popv(3);
	    break;
	} else if (run_mode & RUN_BP) {
	    push2(nil, nil);
	    if (run_mode & RUN_GBP)
		stack[0] = make_symbol("rprint");
	    else
		stack[0] = make_symbol("prettyprint");
	    stack[-1] = make_symbol("prettyprintf*");
	    value(stack[-1]) = stack[0];
	    stack[0] = make_symbol(".startupmode");
	    stack[-1] = make_symbol("lisp");
	    value(stack[0]) = stack[-1];
	    stack[0] = make_symbol(".priprog");
	    value(stack[0]) = true;
	    stack[0] = make_symbol(".nomessage");
	    value(stack[0]) = true;
	    popv(2);
	}
#endif
#ifdef SLISP
	if (run_mode & RUN_COMPILER) {
	    push2(nil, nil);
	    stack[0] = make_symbol(compile_file);
	    stack[-1] = make_symbol(" acompilefile");
	    lisp_printf("  *** Compiling file %s.sl and generate %s.c\n",
			compile_file, compile_file);
	    if (setjmp(this_level))
		break;
	    apply1(stack[-1], stack[0]);
	    popv(2);
	    break;
	}
#endif
#ifdef ASUKA
	if (run_mode & RUN_COMPILER) {
	    /*
	     * incomplete 2007.10.23 
	     */
	    char           *p,
	                   *q;
	    push2(nil, nil);
	    stack[0] = make_symbol(compile_file);
	    stack[-1] = make_symbol("ccompilefile");
	    p = compile_file;
	    for (q = p; *q; q++)
		if (*q == '/')
		    p = q + 1;
	    lisp_printf("  *** Compiling file %s.m and generate a_%s.c\n",
			compile_file, p);
	    if (setjmp(this_level))
		break;
	    apply1(stack[-1], stack[0]);
	    popv(2);
	    break;
	}
#endif
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    supervisor = nil;
	    continue;
	} else {
	    switch (run_mode & RUN_MASK) {
	    case RUN_GAL:
	    case RUN_REDUCE:
		if (supervisor != nil) {
		    miscflags |= MESSAGES_FLAG;
		    apply_n(supervisor, 0, nil);
		} else
		    read_eval_print();
		break;
#ifdef ASUKA
	    case RUN_ASUKA:
		asuka_read_eval_print(true);
		break;
#endif
	    case RUN_SUGAR:
		read_eval_print();
		break;
	    }
	}
	break;
    }
}

#ifdef Android
char           *
android_eval(const char *cmd)
{
    jmp_buf         this_level;
    extern long     trace_count;
    char           *r;

    Sex            *save = stack;
    Sex            *bind_save = bstack;

    if (!android_initialized) {
	return "Asuka is not initialized!";
    }
    init_reading_command(cmd);
    init_androidout();

    tty_count = 0;
    curchar = '\n';
    batch_mode = true;
    errorset_buffer = (jmp_buf *) (&this_level);
    trace_count = 1;
    main_buffer = errorset_buffer;
    if (setjmp(this_level)) {
	stack = save;
	unwind_bind(bind_save);
	supervisor = nil;
    } else {
#ifdef GAL
	CC_gal_read_eval_print(true);
#endif
#ifdef REDUCE
	CC_reduce_read_eval_print(true);
#endif
#ifdef ASUKA
	asuka_read_eval_print(true);
#endif
    }

    r = get_androidout();
    return r;
}
#endif

#ifndef NO_SIGNAL_HANDLER

bool            sigint_must_longjmp = NO;
jmp_buf         sigint_buf;

void
oninti(int code)
{
    char           *s = "\n *** Lisp attention !  ?(q/c/t) = ";

    signal(code, oninti);
    if (sigint_must_longjmp) {
	interrupt_pending = 1;
	sigint_must_longjmp = NO;
	longjmp(sigint_buf, 1);
    }
#ifdef UNIX			/* 
				 * * * signal works   
				 */
    if (already_in_gc) {	/* 
				 * we are now collecting  garbages * wait
				 * the end of processsing 
				 */
	interrupt_pending = 1;
	return;
    } else if (batch_mode) {
	fprintf(stderr, "Interrupted!!. \nprocess terminated\n");
	exit(1);
    }
    switch (ttygetchar(s)) {
    case 'a':
    case 'A':			/* 
				 * abort processing 
				 */
	erdn();
	break;
    default:
    case 'c':
    case 'C':			/* 
				 * continue processing 
				 */
	return;
    case 't':
    case 'T':			/* 
				 * raise an exception 
				 */
	Lsetpchar(lisp_prompt_message);
	longjmp(*main_buffer, 1);
    case 'q':
    case 'Q':
	exit(EXIT_FAILURE);	/* 
				 * * Rather abrupt 
				 */
    }
#else
    interrupt_pending = 1;
#endif
    return;
}

void
onintr(int code)
{
    if (miscflags & MESSAGES_FLAG) {
	switch (code) {
	default:
	    lisp_printf("\nSignal (code = %d) detected...\n", code);
	    break;
#ifdef SIGFPE
	case SIGFPE:
	    lisp_printf("\nFloating point exception detected...\n");
	    exit(80 + code);
	    break;
#endif
#ifdef SIGSEGV
	case SIGSEGV:
	    lisp_printf("\nMemory access violation detected...\n");
	    exit(80 + code);
	    break;
#endif
#ifdef SIGBUS
	case SIGBUS:
	    lisp_printf("\nBus error detected...\n");
	    exit(80 + code);
	    break;
#endif
	}
    }
    if (!is_file(lisp_input))
	lisp_input = standard_input;
    if (already_in_gc)
	exit(80 + code);
    signal(code, onintr);
    longjmp(*errorset_buffer, 1);
}
#endif



#ifdef REDUCE
extern long     number_of_literal_vectors;

void
init_reduce(void)
{
    jmp_buf         this_level;
    Sex            *save = stack;
    Sex            *bind_save = bstack;
    Sex             t;

    tty_count = 0;
    curchar = '\n';
    Lsetpchar(lisp_prompt_message);
    t = make_symbol("*compiler-target*");
    qtag(t) |= SYM_GLOBAL | SYM_FLUID;
    value(t) = CHAR_OF('r' & 0xff);

    errorset_buffer = &this_level;
    if (setjmp(this_level)) {
	stack = save;
	unwind_bind(bind_save);
    } else {
	miscflags |= MESSAGES_FLAG;
	value(nomess_symbol) = nil;
	number_of_literal_vectors = 0;
#if REDUCE_MINOR_VERSION <= 6
	preloaded_module("module");
#endif

	/*
	 */

	preloaded_module("prolog");
	/*
	 * init_patches(); 
	 */
	preloaded_module("rlisp");
	preloaded_module("rend");

	preloaded_module("complr");
	preloaded_module("cfap");
	preloaded_module("entry");
	preloaded_module("cmacro");
	preloaded_module("lap");
	preloaded_module("fap");
	preloaded_module("fisl");

	preloaded_module("pretty");
	preloaded_module("rprint");


	/*
	 * init_cedit(); 
	 */
	preloaded_module("poly");
#if 1
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization ezgcd failed\n");
	} else {
	    preloaded_module("ezgcd");
	}
#endif
	preloaded_module("alg");
	preloaded_module("arith");
	preloaded_module("mathpr");


#if 1

	preloaded_module("matrix");
	preloaded_module("hephys");
#if REDUCE_MINOR_VERSION >= 6
#if 0
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization ezgcd failed\n");
	} else {
	    preloaded_module("ezgcd");
	}
#endif
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization factor failed\n");
	} else {
	    preloaded_module("factor");
	}
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization int failed\n");
	} else {
	    preloaded_module("int");
	}

	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization solve failed\n");
	} else {
	    preloaded_module("solve");
	}

	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization roots failed\n");
	} else {
	    preloaded_module("roots");
	}

	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization roots2 failed\n");
	} else {
	    preloaded_module("roots2");
	}

#endif

#if 0
#ifndef VisualC
	preloaded_module("taylor");
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization rlisp88 failed\n");
	} else {
	    preloaded_module("rlisp88");
	}
#endif
#endif
#if 1
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization arnum failed\n");
	} else {
	    preloaded_module("arnum");
	}
#endif
#if 0
	if (setjmp(this_level)) {
	    stack = save;
	    unwind_bind(bind_save);
	    lisp_printf("initialization solve failed\n");
	} else {
	    /*
	     * miscflags |= MESSAGES_FLAG;
	     */
	    /*
	     * value(nomess_symbol) = nil;
	     */
	    preloaded_module("solve");
	}
#endif

#if REDUCE_MINOR_VERSION == 6
	preloaded_module("dipoly");
	preloaded_module("groebner");
	preloaded_module("groebnr2");
#else
#if REDUCE_MINOR_VERSION >= 7
	preloaded_module("dipoly");
	preloaded_module("groebner");
	preloaded_module("groebnr2");
	/*
	 * preloaded_module("ideals");
	 */
	preloaded_module("sum");
	//preloaded_module("polydiv");
#endif
#endif

#endif

	value(nomess_symbol) = nil;
	supervisor = make_symbol("begin");
    }
}

#endif

#ifdef ASUKA
void
init_asuka(void)
{
    jmp_buf         this_level;
    Sex            *save = stack;
    Sex            *bind_save = bstack;
    // Sex t;

    tty_count = 0;
    curchar = '\n';
    Lsetpchar(lisp_prompt_message);

    errorset_buffer = &this_level;
    if (setjmp(this_level)) {
	stack = save;
	unwind_bind(bind_save);
    } else {
	miscflags |= MESSAGES_FLAG;
#if 0

	push(current_package);
	current_package = value(current_asuka_package);
	current_charvec = package_charvec(current_package);

#endif
	init_pretty();
	init_complr();
	init_sfap();
#ifdef XASUKA
	init_z();
#endif
#if 0

	pop(current_package);
	current_charvec = package_charvec(current_package);

#endif
    }
}
#endif

#ifdef SUGAR
void
init_sugar(void)
{
    jmp_buf         this_level;
    Sex            *save = stack;
    Sex            *bind_save = bstack;
    // Sex t;

    tty_count = 0;
    curchar = '\n';
    Lsetpchar(lisp_prompt_message);

    errorset_buffer = &this_level;
    if (setjmp(this_level)) {
	stack = save;
	unwind_bind(bind_save);
    } else {
    }
}
#endif

#ifdef SLISP
void
init_slisp(void)
{
    jmp_buf         this_level;
    Sex            *save = stack;
    Sex            *bind_save = bstack;
    Sex             t;

    tty_count = 0;
    curchar = '\n';
    Lsetpchar(lisp_prompt_message);
    t = make_symbol("*compiler-target*");
    qtag(t) |= SYM_GLOBAL | SYM_FLUID;
    value(t) = CHAR_OF('s' & 0xff);

    errorset_buffer = &this_level;
    if (setjmp(this_level)) {
	stack = save;
	unwind_bind(bind_save);
    } else {
	miscflags |= MESSAGES_FLAG;
	init_complr();
	init_cfap();
	init_reduce2();
    }
}
#endif

#ifdef GAL

extern long     number_of_literal_vectors;

void
init_gal(void)
{
    jmp_buf         this_level;
    Sex            *save = stack;
    Sex            *bind_save = bstack;
    Sex             t;

    tty_count = 0;
    curchar = '\n';
    Lsetpchar(lisp_prompt_message);
    t = make_symbol("*compiler-target*");
    qtag(t) |= SYM_GLOBAL | SYM_FLUID;
    value(t) = CHAR_OF('g' & 0xff);

    errorset_buffer = &this_level;
    if (setjmp(this_level)) {
	stack = save;
	unwind_bind(bind_save);
    } else {
	miscflags |= MESSAGES_FLAG;
	value(nomess_symbol) = true;
	number_of_literal_vectors = 0;
	/*
	 * init_fend();
	 */
	/*
	 * init_gend();
	 */
	/*
	 * init_pretty();
	 */
	init_parser();
	init_bsatom();
	init_bsdefn();
	init_bseval();
	init_bsmapf();
	init_bspoly();
	init_bspref();
	init_bsques();
	init_bssimp();
	init_bswrit();
	init_comcomp();
	init_comconv();
	init_comform();
	init_compart();
	init_comproc();
	init_comrset();
	init_libalg();
	init_libapp();
	/*
	 * init_libnag(); 
	 */
	init_gpretty();
	init_complr();
	init_cfap();
	init_bfap();
	init_bi2r();
#ifdef OLD
	init_pkggb6();
#else
	init_pkggb();
	init_add();
	init_linalg();
	init_pkgfac();
	init_missing();
	init_pkgtest();
	init_pkgnc();
	init_pkgtest();
	init_pkgleq();
#endif
	value(nomess_symbol) = nil;
	supervisor = make_symbol(" gal");
    }
}

#endif



char           *files_to_read[MAX_INPUT_FILES];


long            number_of_input_files = 0;

long
lisp_atoi(char *w)
{
    long            n = 0;
    long            c;

    while ((c = *w++) != 0) {
	if (isdigit(c))
	    n = n * 10 + c - '0';
	else if (c == 'k' || c == 'K')
	    return n * 1024;
	else if (c == 'm' || c == 'M')
	    return n * 1024 * 1024;
	else if (c == 'g' || c == 'G')
	    return n * 1024 * 1024 * 1024;
	else
	    break;
    }
    return n;
}

#if !defined(VisualC) || defined(_CONSOLE)

long
my_main(long argc, char *argv[])
{
    long            i,
                    t = 0;

    spool_file = NULL;
    debug_file = NULL;
    number_of_input_files = 0;
    column = 0;
    line_length = 78;

    return_code = EXIT_SUCCESS;

    argc--;
    for (i = 1; i <= argc; i++) {
	char           *opt = argv[i];
	if (opt[0] == '-') {
	    char           *w;
	    long            c1 = opt[1],
		c2 = opt[2];
	    switch (c1) {

	    case 'c':
	    case 'm':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */

		core_size = lisp_atoi(w);
		if (core_size <= 0)
		    core_size = CORE_SIZE;
		continue;
	    case 's':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		stack_size = lisp_atoi(w) * SEX;
		if (stack_size <= 0)
		    stack_size = STACK_SIZE;
		continue;
	    case 'b':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		bind_size = lisp_atoi(w) * SEX;
		if (bind_size <= 0)
		    bind_size = BIND_SIZE;
		continue;

	    case 'l':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		{
		    char            filename[FILENAME_MAX];
		    spool_file = open_file(filename, w,
					   strlen(w), "w", NULL);
		    if (spool_file == NULL)
			lisp_printf("Unable to write to \"%s\"\n",
				    filename);
		}
		continue;
	    case 'd':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		{
		    char            filename[FILENAME_MAX];
		    debug_file = open_file(filename, w,
					   strlen(w), "w", NULL);
		    if (debug_file == NULL)
			lisp_printf("Unable to write to \"%s\"\n",
				    filename);
		}
		continue;
	    case 't':
		t |= 1;
		continue;
	    case 'n':
		run_mode = RUN_SUGAR;
		continue;
	    case 'P':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		run_mode |= RUN_LBP;
		continue;
#ifdef ASUKA
	    case 'p':
		run_mode |= RUN_LBP;
		continue;
#else
	    case 'p':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		run_mode |= RUN_GBP;
		continue;
#endif
	    case 'x':
		stayslow = YES;
		continue;
	    case 'o':
		if (i != argc) {
		    cobject_path = argv[++i];
		    continue;
		}
		break;
#ifdef YYDEBUG
	    case 'y':
		yydebug = 1;
		continue;
#endif
#ifdef REDUCE
	    case 'Z':
		showlap = 1;
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		t |= 2;
		continue;
#endif
#ifdef GAL
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		continue;
#endif
#ifdef ASUKA
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		continue;
#endif
#ifdef SLISP
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		continue;
#endif

	    case 'X':
		trace_flag |= (SYM_TRACED | SYM_PROC);
		continue;

	    default:
	      usage:
		lisp_printf("Unrecognized option \"-%c\"\n", c1);
#ifdef REDUCE
		lisp_printf
		    ("Usage: reduce [-c memory size] [-s stacksize] [-b bind_stacksize] [input files ..]\n");
		lisp_printf("       reduce [-z module_name] \n");
#else
#ifdef GAL
		lisp_printf
		    ("Usage: gal [-c memory size] [-s stacksize] [-b bind_stacksize] [input files ..]\n");
		lisp_printf("       gal [-z module_name] \n");
#else
		lisp_printf
		    ("Usage: sugar [-c memory size] [-s stacksize] [-b bind_stacksize] [input files ..]\n");
#endif
#endif
		exit(1);
	    }
	    lisp_printf("Option \"-%c\" needs an argument: ignored\n", c1);
	    break;
	} else
	    files_to_read[number_of_input_files++] = opt;
    }



    base_time = get_time();
    gc_time = 0.0;
    io_time = 0.0;
    trap_time = 0.0;

    init_heap_segments();

    setup();

    {
	long            core = core_size / 1024;
	long            typ = 'K';
	char            current_date[sizeof(__DATE__) + 1] = __DATE__;
	current_date[sizeof(__DATE__) - 6] = ',';
	if (core > 16000) {
	    core = core / 1000;
	    typ = 'M';
	}
#ifdef REDUCE
	lisp_printf("  Open REDUCE %s/%s %s%s (%s) : %d %cbytes\n",
		    REDUCE_VERSION,
		    IMPNAME, OPVER, LP_MODE, current_date, core, typ);
#endif
#ifdef ASUKA
	lisp_printf("  Asuka/%s %s%s (ver %s, %s) : %d %cbytes\n",
		    IMPNAME, OPVER, LP_MODE, ASUKA_VERSION, current_date,
		    core, typ);
#endif
#ifdef SUGAR
	lisp_printf("  CSugar/%s %s%s (ver %s, %s) : %d %cbytes\n",
		    IMPNAME, OPVER, LP_MODE, VERSION, current_date, core,
		    typ);
#endif
#ifdef GAL
	if (run_mode & RUN_BP)
	    lisp_printf
		("||  CGAL on %s %s%s (ver %s, %s with  %d %cbytes)\n",
		 IMPNAME, OPVER, LP_MODE, GAL_VERSION, current_date, core,
		 typ);
	else
	    lisp_printf
		("  CGAL on %s %s%s (ver %s, %s with  %d %cbytes)\n",
		 IMPNAME, OPVER, LP_MODE, GAL_VERSION, current_date, core,
		 typ);
#endif
#ifdef SLISP
	lisp_printf("  CSlisp/%s %s%s (ver %s, %s) : %d %cbytes\n",
		    IMPNAME, OPVER, LP_MODE, VERSION, current_date, core,
		    typ);
#endif
#ifdef USE_GMP
#ifdef USE_MPFR
	lisp_printf
	    ("      with GMP %d.%d.%d and MPFR %d.%d.%d libraries\n",
	     __GNU_MP_VERSION, __GNU_MP_VERSION_MINOR,
	     __GNU_MP_VERSION_PATCHLEVEL, MPFR_VERSION_MAJOR,
	     MPFR_VERSION_MINOR, MPFR_VERSION_PATCHLEVEL);
#else
	lisp_printf("      with GMP %d.%d.%d library\n", __GNU_MP_VERSION,
		    __GNU_MP_VERSION_MINOR, __GNU_MP_VERSION_PATCHLEVEL);
#endif
#endif
	lisp_printf("\n");
    }


    interrupt_pending = already_in_gc = NO;

#ifndef NO_SIGNAL_HANDLER

    sigint_must_longjmp = NO;

#ifdef SIGINT
    signal(SIGINT, oninti);
#endif
#ifdef SIGILL
    signal(SIGILL, onintr);
#endif
#ifdef SIGFPE
    signal(SIGFPE, onintr);
#endif
#ifdef SIGBUS
    signal(SIGBUS, onintr);
#endif
#ifdef SIGEMT
    signal(SIGEMT, onintr);
#endif
#ifdef SIGSEGV
    signal(SIGSEGV, onintr);
#endif
#endif

#ifdef ASUKA
    init_asuka();
#endif

#ifdef SUGAR
    init_sugar();
#endif
#ifdef REDUCE
    init_reduce();
#endif
#ifdef SLISP
    init_slisp();
#endif
#ifdef GAL
    init_gal();
#endif
    /*
     * supervisor = nil;
     */

#ifdef _LP64
    unsigned long   init_rand64[4] = {
	0x12345UL, 0x23456UL, 0x34567UL, 0x45678UL
    };
    init_by_array64(init_rand64, 4);
#else
    unsigned long   init_rand32[4] = {
	0x123, 0x234, 0x345, 0x456
    };
    init_by_array(init_rand32, 4);
#endif

    if (t & 1)
	trace_flag = SYM_TRACED | SYM_PROC;
    if (t & 2)
	compile_flag = YES;
    if (number_of_input_files == 0)
	lisp_main();
    else {
	long            i;
	for (i = 0; i < number_of_input_files; i++) {
	    char            filename[FILENAME_MAX];
	    FILE           *f;

	    f = open_file(filename, files_to_read[i],
			  strlen(files_to_read[i]), "r", NULL);
	    if (f == NULL)
		lisp_printf("\n   *** Could not read file %s\n",
			    files_to_read[i]);
	    else {
		lisp_printf("\n   *** About to read file %s\n",
			    files_to_read[i]);
		// vector_value(standard_input, 0) = (Sex) f;
		lisp_input = make_file_handle(f, NOT_CHAR, 0);
		lisp_main();
		fclose(f);
		lisp_input = standard_input;
	    }
	}
    }

    if (spool_file != NULL)
	fclose(spool_file);
    if (debug_file != NULL)
	fclose(debug_file);
    exit(return_code);
    return 0;
}
#endif


#ifdef Android
void
android_init2(long argc, char *argv[])
{
    long            i,
                    t = 0;

    spool_file = NULL;
    debug_file = NULL;
    number_of_input_files = 0;
    column = 0;
    line_length = 78;

    return_code = EXIT_SUCCESS;

    argc--;
    for (i = 1; i <= argc; i++) {
	char           *opt = argv[i];
	if (opt[0] == '-') {
	    char           *w;
	    long            c1 = opt[1],
		c2 = opt[2];
	    switch (c1) {
	    case 'c':
	    case 'm':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */

		core_size = lisp_atoi(w);
		if (core_size <= 0)
		    core_size = CORE_SIZE;
		continue;
	    case 's':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		stack_size = lisp_atoi(w) * SEX;
		if (stack_size <= 0)
		    stack_size = STACK_SIZE;
		continue;
	    case 'b':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		bind_size = lisp_atoi(w) * SEX;
		if (bind_size <= 0)
		    bind_size = BIND_SIZE;
		continue;

	    case 'l':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		{
		    char            filename[FILENAME_MAX];
		    spool_file = open_file(filename, w,
					   strlen(w), "w", NULL);
		    if (spool_file == NULL)
			lisp_printf("Unable to write to \"%s\"\n",
				    filename);
		}
		continue;
	    case 'd':
		if (c2 != 0)
		    w = &opt[2];
		else if (i != argc)
		    w = argv[++i];
		else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		{
		    char            filename[FILENAME_MAX];
		    debug_file = open_file(filename, w,
					   strlen(w), "w", NULL);
		    if (debug_file == NULL)
			lisp_printf("Unable to write to \"%s\"\n",
				    filename);
		}
		continue;
	    case 't':
		t |= 1;
		continue;
	    case 'n':
		run_mode = RUN_SUGAR;
		continue;
	    case 'P':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		run_mode |= RUN_LBP;
		continue;
#ifdef ASUKA
	    case 'p':
		run_mode |= RUN_LBP;
		continue;
#else
	    case 'p':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		run_mode |= RUN_GBP;
		continue;
#endif
	    case 'x':
		stayslow = YES;
		continue;
	    case 'o':
		if (i != argc) {
		    cobject_path = argv[++i];
		    continue;
		}
		break;
#ifdef YYDEBUG
	    case 'y':
		yydebug = 1;
		continue;
#endif
#ifdef REDUCE
	    case 'Z':
		showlap = 1;
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		t |= 2;
		continue;
#endif
#ifdef GAL
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		continue;
#endif
#ifdef ASUKA
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		continue;
#endif
#ifdef SLISP
	    case 'z':
		if (run_mode & (RUN_COMPILER | RUN_BP))
		    goto usage;
		if (c2 != 0) {
		    compile_file = &opt[2];
		    run_mode |= RUN_COMPILER;
		} else if (i != argc) {
		    compile_file = argv[++i];
		    run_mode |= RUN_COMPILER;
		} else
		    break;	/* 
				 * Illegal at end of command-line 
				 */
		continue;
#endif


	    default:
	      usage:
		lisp_printf("Unrecognized option \"-%c\"\n", c1);
#ifdef REDUCE
		lisp_printf
		    ("Usage: reduce [-c memory size] [-s stacksize] [-b bind_stacksize] [input files ..]\n");
		lisp_printf("       reduce [-z module_name] \n");
#else
#ifdef GAL
		lisp_printf
		    ("Usage: gal [-c memory size] [-s stacksize] [-b bind_stacksize] [input files ..]\n");
		lisp_printf("       gal [-z module_name] \n");
#else
		lisp_printf
		    ("Usage: sugar [-c memory size] [-s stacksize] [-b bind_stacksize] [input files ..]\n");
#endif
#endif
		exit(1);
	    }
	    lisp_printf("Option \"-%c\" needs an argument: ignored\n", c1);
	    break;
	} else
	    files_to_read[number_of_input_files++] = opt;
    }



    base_time = get_time();
    gc_time = 0.0;
    io_time = 0.0;
    trap_time = 0.0;

    init_heap_segments();

    setup();

    {
	long            core = core_size / 1024;
	long            typ = 'K';
	char            current_date[sizeof(__DATE__) + 1] = __DATE__;
	current_date[sizeof(__DATE__) - 6] = ',';
	if (core > 16000) {
	    core = core / 1000;
	    typ = 'M';
	}
#ifdef REDUCE
	lisp_printf("  Open REDUCE %s/%s %s%s (%s) : %d %cbytes\n",
		    REDUCE_VERSION,
		    IMPNAME, OPVER, LP_MODE, current_date, core, typ);
#endif
#ifdef ASUKA
	lisp_printf("  Asuka/%s %s%s (ver %s, %s) : %d %cbytes\n",
		    IMPNAME, OPVER, LP_MODE, ASUKA_VERSION, current_date,
		    core, typ);
#endif
#ifdef SUGAR
	lisp_printf("  CSugar/%s %s%s (ver %s, %s) : %d %cbytes\n",
		    IMPNAME, OPVER, LP_MODE, VERSION, current_date, core,
		    typ);
#endif
#ifdef GAL
	if (run_mode & RUN_BP)
	    lisp_printf
		("||  CGAL on %s %s%s (ver %s, %s with  %d %cbytes)\n",
		 IMPNAME, OPVER, LP_MODE, GAL_VERSION, current_date, core,
		 typ);
	else
	    lisp_printf
		("  CGAL on %s %s%s (ver %s, %s with  %d %cbytes)\n",
		 IMPNAME, OPVER, LP_MODE, GAL_VERSION, current_date, core,
		 typ);
#endif
#ifdef SLISP
	lisp_printf("  CSlisp/%s %s%s (ver %s, %s) : %d %cbytes\n",
		    IMPNAME, OPVER, LP_MODE, VERSION, current_date, core,
		    typ);
#endif
#ifdef USE_GMP
#ifdef USE_MPFR
	lisp_printf
	    ("      with GMP %d.%d.%d and MPFR %d.%d.%d libraries\n",
	     __GNU_MP_VERSION, __GNU_MP_VERSION_MINOR,
	     __GNU_MP_VERSION_PATCHLEVEL, MPFR_VERSION_MAJOR,
	     MPFR_VERSION_MINOR, MPFR_VERSION_PATCHLEVEL);
#else
	lisp_printf("      with GMP %d.%d.%d library\n", __GNU_MP_VERSION,
		    __GNU_MP_VERSION_MINOR, __GNU_MP_VERSION_PATCHLEVEL);
#endif
#endif
	lisp_printf("\n");
    }


    interrupt_pending = already_in_gc = NO;

#ifndef NO_SIGNAL_HANDLER

    sigint_must_longjmp = NO;

#ifdef SIGINT
    signal(SIGINT, oninti);
#endif
#ifdef SIGILL
    signal(SIGILL, onintr);
#endif
#ifdef SIGFPE
    signal(SIGFPE, onintr);
#endif
#ifdef SIGBUS
    signal(SIGBUS, onintr);
#endif
#ifdef SIGEMT
    signal(SIGEMT, onintr);
#endif
#ifdef SIGSEGV
    signal(SIGSEGV, onintr);
#endif
#endif

#ifdef ASUKA
    init_asuka();
#endif

#ifdef SUGAR
    init_sugar();
#endif
#ifdef REDUCE
    init_reduce();
    {
	push2(nil, nil);
#define w       stack[0]
#define s       stack[-1]
	w = make_symbol("algebraic");
	s = make_symbol("*mode");
	qtag(s) |= SYM_FLUID | SYM_GLOBAL;
	value(s) = w;
#undef s
#undef w
	popv(2);
    }
#endif
#ifdef SLISP
    init_slisp();
#endif
#ifdef GAL
    init_gal();
#endif
    /*
     * supervisor = nil;
     */

#ifdef _LP64
    unsigned long   init_rand64[4] = {
	0x12345UL, 0x23456UL, 0x34567UL, 0x45678UL
    };
    init_by_array64(init_rand64, 4);
#else
    unsigned long   init_rand32[4] = {
	0x123, 0x234, 0x345, 0x456
    };
    init_by_array(init_rand32, 4);
#endif

    if (t & 1)
	trace_flag = SYM_TRACED | SYM_PROC;
    if (t & 2)
	compile_flag = YES;

    if (spool_file != NULL)
	fclose(spool_file);
    if (debug_file != NULL)
	fclose(debug_file);

    android_initialized = 1;
}
#endif
