/*
 * sysdosv.c		(C) Copyright 1998-2000 by Fujio Kako, Nara Women's University.
 * General Dynabook MS-DOS 6.2/V system-specific code
 * $Header: /Users/kako/asuka/src/RCS/sysdosv.c,v 1.1 2015/11/08 23:22:39 kako Exp $
 */

static char const rcsid[] =
"$Id: sysdosv.c,v 1.1 2015/11/08 23:22:39 kako Exp $";


#include "config.h"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WINDOWS95
#include <sys/times.h>
#include <sys/time.h>
#endif
#if GCCTYPE == GCC_DJGPP
#include <dos.h>
#endif

#include "tags.h"
#include "externs.h"

static void     get_home_directory(char *b);
static void     get_users_home_directory(char *b);

static void
expand_file_name(char *filename, char *old, size_t n)
{
    int             j;
    int             c;
    char           *o,
                   *tail;
    if (n == 0) {
	*filename = 0;
	return;
    }
    o = filename;
    c = *old;
    if (c == '~') {
	tail = old + 1;
	for (j = (int) n - 1; j != 0; j--) {
	    c = *tail;
	    if (c == '.' || c == '/' || c == '\\')
		break;
	    tail++;
	    *o++ = c;
	}
	*o = 0;
	if (o == filename) {	/*
				 * '~' on its own 
				 */
	    get_home_directory(filename);
	    o = filename + strlen(filename);
	} else {
	    get_users_home_directory(filename);
	    o = filename + strlen(filename);
	}
	for (; j != 0; j--)
	    *o++ = *tail++;
	*o = 0;
    } else if (c == '$' && n > 1 &&
	       (c = old[1]) != '.' && c != '/' && c != '\\') {
	char           *w;
	if (c == '{') {
	    tail = old + 2;
	    for (j = (int) n - 2; j != 0; j--) {
		c = *tail++;
		if (c == '}')
		    break;
		*o++ = c;
	    }
	} else {
	    tail = old + 1;
	    for (j = (int) n - 1; j != 0; j--) {
		c = *tail;
		if (c == '.' || c == '/' || c == '\\')
		    break;
		tail++;
		*o++ = c;
	    }
	}
	*o = 0;
	w = getenv(filename);
	if (w != NULL) {
	    strcpy(filename, w);
	    o = filename + strlen(filename);
	} else
	    o = filename;
	for (; j != 0; j--)
	    *o++ = *tail++;
	*o = 0;
    } else {
	memcpy(filename, old, n);
	filename[n] = 0;
    }
}

FILE           *
open_file(char *filename, char *old, size_t n,
	  char *mode, FILE * old_file)
{
    expand_file_name(filename, old, n);
    if (*filename == 0)
	return NULL;
#ifdef NO_BINARY_FOPEN
    if (mode[0] == 'w') {
	if (mode[1] == '+')
	    mode = "w+";
	else
	    mode = "w";
    } else if (mode[1] == '+')
	mode = "r+";
    else
	mode = "r";
#endif

    if (old_file == NULL)
	return fopen(filename, mode);
    else
	return freopen(filename, mode, old_file);
}

bool
file_exists(char *filename, char *old, size_t n, char *tt)
{
    struct stat     statbuff;
    expand_file_name(filename, old, n);
    if (*filename == 0)
	return NO;
    if (stat(filename, &statbuff) != 0)
	return NO;
    strcpy(tt, ctime(&(statbuff.st_ctime)));
    return YES;
}

#ifdef USE_GETUID

#include <pwd.h>
extern int      getuid(void);

static void
get_home_directory(char *b)
{
    struct passwd  *pw = getpwuid(getuid());
    strcpy(b, pw->pw_dir);
}

static void
get_users_home_directory(char *b)
{
    struct passwd  *pw = getpwnam(b);
    if (pw != NULL)
	strcpy(b, pw->pw_dir);
    else
	strcpy(b, ".");
}

#else				/*
				 * * * USE_GETUID   
				 */

static void
get_home_directory(char *b)
{
    strcpy(b, getenv("HOME"));
}

static void
get_users_home_directory(char *b)
{
    strcpy(b, ".");
}


#endif


#if GCCTYPE == GNU_WIN32
double
get_time()
{
    double          tv;
    struct timeval  t;
    int             r;

    r = gettimeofday(&t, NULL);
    tv = (double) (t.tv_sec + t.tv_usec / 1000000.0);
    return tv;
}

#else
#if GCCTYPE == GCC_DJGPP

double
get_time()
{
    union REGS      r;
    double          tv;
    int             days,
                    i;
    static int      month[] =
    {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

    r.h.ah = 0x2c;
    int86(0x21, &r, &r);
    tv = (r.h.ch * 60 + r.h.cl) * 60 + r.h.dh;
    tv = tv + r.h.dl * 0.01;
    r.h.ah = 0x2a;
    int86(0x21, &r, &r);

    days = r.h.cl * 365 + (r.h.cl / 4) - (r.h.cl / 100) + (r.h.cl / 400);
    for (i = 0; i < r.h.dh; i++)
	days += month[i];
    days += r.h.dl - 1;
    if ((r.h.cl % 4 == 0 && r.h.cl % 100 != 0) || r.h.cl % 400 == 0) {
	if (r.h.dh > 1)
	    days++;
    }
    tv = tv + days * 24.0 * 60 * 60;
    return tv;
}
#else
double
get_time()
{
    double          tv;
    struct my_tms {
	clock_t         tms_utime;
	clock_t         tms_stime;
	clock_t         tms_cutime;
	clock_t         tms_cstime;
    } tmsbuf;
    extern void     times(	/*
				 * struct my_tms * 
				 */ );
    times(&tmsbuf);
    return (double) (tmsbuf.tms_utime + tmsbuf.tms_stime) / (double) HZ;
}
#endif
#endif


int
batchp()
{
    return !isatty(fileno(stdin));
}


#ifdef GNU_WIN32
Sex
Lint10h(ax, bx, cx, dx)
    Sex             ax,
                    bx,
                    cx,
                    dx;
{
    return nil;
}

Sex
Lint10hr(ax, bx, cx, dx)
    Sex             ax,
                    bx,
                    cx,
                    dx;
{
    return nil;
}
#else
extern          Sex
Lint10h(Sex ax, Sex bx, Sex cx, Sex dx)
{
    register Sex    nilr = nil;
    int             a,
                    b,
                    c,
                    d;
    union REGS      r;
    if (ax == nilr)
	ax = mkfxat(0);
    if (bx == nilr)
	bx = mkfxat(0);
    if (cx == nilr)
	cx = mkfxat(0);
    if (dx == nilr)
	dx = mkfxat(0);
    if (!is_fixnum(ax) ||
	!is_fixnum(bx) ||
	!is_fixnum(cx) ||
	!is_fixnum(dx))
	return aerror("int10h");
    r.x.ax = mkint(ax);
    r.x.bx = mkint(bx);
    r.x.cx = mkint(cx);
    r.x.dx = mkint(dx);
    int86(0x10, &r, &r);
    return nilr;
}

extern          Sex
Lint10hr(Sex ax, Sex bx, Sex cx, Sex dx)
{
    int             a,
                    b,
                    c,
                    d;
    register Sex    nilr = nil;
    union REGS      r;
    if (ax == nilr)
	ax = mkfxat(0);
    if (bx == nilr)
	bx = mkfxat(0);
    if (cx == nilr)
	cx = mkfxat(0);
    if (dx == nilr)
	dx = mkfxat(0);
    if (!is_fixnum(ax) ||
	!is_fixnum(bx) ||
	!is_fixnum(cx) ||
	!is_fixnum(dx))
	return aerror("int10hr");
    r.x.ax = mkint(ax);
    r.x.bx = mkint(bx);
    r.x.cx = mkint(cx);
    r.x.dx = mkint(dx);
    int86(0x10, &r, &r);
    cx = Llist2(mkfxat(r.x.cx), mkfxat(r.x.dx));
    ax = Llist2star(mkfxat(r.x.ax), mkfxat(r.x.bx), cx);
    return ax;
}
#endif


#if GCCTYPE == GCC_DJGPP
#ifdef DynaBook
#if 1
void
times(struct tms *buffer)
{
    buffer->tms_utime = 0;
    buffer->tms_stime = 0;
}

#else

void
times(struct tms *buffer)
{
    union REGS      r;
    long            usec,
                    sec;

    r.h.ah = 0x2c;
    int86(0x21, &r, &r);

    sec = ((r.h.ch * 60 + r.h.cl) * 60 + r.h.dh) * 100 + r.h.dl;
    buffer->tms_utime = sec;
    buffer->tms_stime = 0;
}
#endif
#else

struct _system_date {
    int             dayofweek;
    int             year;
    int             month;
    int             day;
};

struct _system_time {
    int             hours;
    int             minutes;
    int             seconds;
    int             subsecs;
};

#ifdef dummy
static struct _system_date
_dos_get_current_date()
{
    struct _system_date date;

    asm volatile    ("
		     movb % 4, %%ah
		     int $0x21
		     movzbl %% al, %0
		     movzwl %% cx, %1
		     movzbl %% dh, %2
		     movzbl %% dl, %3 "
		     :"=r"           (date.dayofweek), "=r"(date.year),
		     "=r"            (date.month), "=r"(date.day)
		     :"i"            (0x2a)
		     :"ax", "cx", "dx");

    return date;
}


static struct _system_time
_dos_get_current_time()
{
    struct _system_time time;

    asm volatile    ("
		     movb % 4, %%ah
		     int $0x21
		     movzbl %% ch, %0
		     movzbl %% cl, %1
		     movzbl %% dh, %2
		     movzbl %% dl, %3 "
		     :"=r"           (time.hours), "=r"(time.minutes),
		     "=r"            (time.seconds), "=r"(time.subsecs)
		     :"i"            (0x2c)
		     :"ax", "cx", "dx");

    return time;
}

#else
extern struct _system_date _dos_get_current_date();
extern struct _system_time _dos_get_current_time();
#endif

#define JST	(-9)		/*
				 * * * Japan standard time   
				 */

time_t          _subsec_time(time_t *);

time_t
time(time_t * tloc)
{
    unsigned long int ctime;
    unsigned long int subsec;

    ctime = _subsec_time(&subsec);
    ctime = ctime * 100 + subsec;

    if (tloc)
	*tloc = ctime;

    return ctime;
}

time_t
_subsec_time(time_t * subsec)
{
    unsigned long int today,
                    ctime;
    struct _system_date date;
    struct _system_time time;

    date = _dos_get_current_date();
    time = _dos_get_current_time();

    ctime = ((today * 24 + time.hours + JST) * 60 + time.minutes) * 60
	+ time.seconds;

    if (subsec)
	*subsec = time.subsecs;

    return ctime;
}


char           *
ctime(t)
    const time_t   *t;
{
    struct _system_date date;
    struct _system_time time;
    char            buf[29];
    static char    *month[] =
    {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
    static char    *week[] =
    {
	"Sun", "Mon", "Tur", "Wed", "The", "Fri", "Sat"};

    date = _dos_get_current_date();
    time = _dos_get_current_time();

    sprintf(buf, "%s %s %2d %02d:%02d:%02d.%02d %4d",
	    week[date.dayofweek], month[date.month],
	    date.day, time.hours, time.minutes, time.seconds,
	    date.subsecs,
	    date.year);
    return buf;
}

#endif
#endif
