/*
 * sysunix.c            (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * General Unix system-specific code 
 * $Header: /Users/kako/asuka/src/RCS/sysunix.c,v 2.3 2015/01/21 05:30:39 kako Exp $
 */

static char const rcsid[] =
    "$Id: sysunix.c,v 2.3 2015/01/21 05:30:39 kako Exp $";


#include "config.h"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include "tags.h"
#include "externs.h"

static void     get_home_directory(char *b);
static void     get_users_home_directory(char *b);

static void
expand_file_name(char *filename, char *old, size_t n)
{
    int             j;
    int             c;
    char           *o,
                   *tail;
    if (n == 0) {
	*filename = 0;
	return;
    }
    o = filename;
    c = *old;
    if (c == '~') {
	tail = old + 1;
	for (j = (int) n - 1; j != 0; j--) {
	    c = *tail;
	    if (c == '.' || c == '/' || c == '\\')
		break;
	    tail++;
	    *o++ = c;
	}
	*o = 0;
	if (o == filename) {	/* 
				 * '~' on its own 
				 */
	    get_home_directory(filename);
	    o = filename + strlen(filename);
	} else {
	    get_users_home_directory(filename);
	    o = filename + strlen(filename);
	}
	for (; j != 0; j--)
	    *o++ = *tail++;
	*o = 0;
    } else if (c == '$' && n > 1 &&
	       /*
	        * $ on its own, or $/... etc stay unaltered 
	        */
	       (c = old[1]) != '.' && c != '/' && c != '\\') {
	char           *w;
	if (c == '{') {
	    tail = old + 2;
	    for (j = (int) n - 2; j != 0; j--) {
		c = *tail++;
		if (c == '}')
		    break;
		*o++ = c;
	    }
	} else {
	    tail = old + 1;
	    for (j = (int) n - 1; j != 0; j--) {
		c = *tail;
		if (c == '.' || c == '/' || c == '\\')
		    break;
		tail++;
		*o++ = c;
	    }
	}
	*o = 0;
	w = getenv(filename);
	if (w != NULL) {
	    strcpy(filename, w);
	    o = filename + strlen(filename);
	} else
	    o = filename;
	for (; j != 0; j--)
	    *o++ = *tail++;
	*o = 0;
    } else {
	memcpy(filename, old, n);
	filename[n] = 0;
    }
}

FILE           *
open_file(char *filename, char *old, size_t n, char *mode, FILE * old_file)
{
    expand_file_name(filename, old, n);

    if (*filename == 0)
	return NULL;
#ifdef NO_BINARY_FOPEN
    if (mode[0] == 'w') {
	if (mode[1] == '+')
	    mode = "w+";
	else
	    mode = "w";
    } else if (mode[1] == '+')
	mode = "r+";
    else
	mode = "r";
#endif


    if (old_file == NULL)
	return fopen(filename, mode);
    else
	return freopen(filename, mode, old_file);
}

bool
file_exists(char *filename, char *old, size_t n, char *tt)
{
    struct stat     statbuff;
    expand_file_name(filename, old, n);
    if (*filename == 0)
	return NO;
    if (stat(filename, &statbuff) != 0)
	return NO;
#ifdef Android
    strcpy(tt, ctime((const time_t *) &(statbuff.st_ctime)));
#else
    strcpy(tt, ctime(&(statbuff.st_ctime)));
#endif
    return YES;
}





#ifdef USE_GETUID

#include <pwd.h>
extern int      getuid(void);

static void
get_home_directory(char *b)
{
    struct passwd  *pw = getpwuid(getuid());
    strcpy(b, pw->pw_dir);
}

static void
get_users_home_directory(char *b)
{
    struct passwd  *pw = getpwnam(b);
    if (pw != NULL)
	strcpy(b, pw->pw_dir);
    else
	strcpy(b, ".");
}

#else

static void
get_home_directory(char *b)
{
    strcpy(b, getenv("HOME"));
}

static void
get_users_home_directory(char *b)
{
    strcpy(b, ".");
}


#endif


#include <sys/time.h>
#include <sys/times.h>
#include <sys/resource.h>

double
get_time()
{
    double          tv;
    /*
     * struct my_tms { clock_t tms_utime; clock_t tms_stime; clock_t
     * tms_cutime; clock_t tms_cstime; } tmsbuf; 
     */
    struct tms      tmsbuf;
    struct rusage   ru;

    if (getrusage(RUSAGE_SELF, &ru) == 0) {
	tv = (double) (ru.ru_utime.tv_sec + ru.ru_stime.tv_sec)
	    + (double) (ru.ru_utime.tv_usec + ru.ru_stime.tv_usec)
	    / (double) 1000000;
    } else {
	clock_t         w1;
	// extern void times(tmsbuf *);
	times(&tmsbuf);
	w1 = tmsbuf.tms_utime;	/* 
				 * * User time in HZ ticks 
				 */
	tv = (double) (tmsbuf.tms_utime + tmsbuf.tms_stime) / (double) HZ;
    }
    return tv;
}


int
batchp()
{
    return !isatty(fileno(stdin));
}
