/*
 * tri.c               (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/tri.c,v 2.23 2016/09/06 05:48:29 kako Exp $
 */

static char const rcsid[] =
    "$Id: tri.c,v 2.23 2016/09/06 05:48:29 kako Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "kctype.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "entries.h"
#include "compare.h"
#include "misc.h"
#include "tri.h"

#define lcoeff(a) caar(a)
#define ldeg(a) cdar(a)
#define coeff(a) car(a)
#define power(a) cdr(a)
#define mkterm(c,p) cons(c, p)
#define reduct(a) cdr(a)
#define lterm(a) car(a)

long            compare_var(Sex a, Sex b);
Sex             Mgcd(Sex a, Sex b);
Sex             Mlcm(Sex a, Sex b);
extern Sex      wuremainder(Sex a, Sex b);

extern Sex      gpquotient2(Sex a, Sex b, Sex x);
extern Sex      gpremaind2(Sex a, Sex b, Sex x);

/*
 * #define putv(v,i,a) vector_value(v,i) = a
 * #define getv(v,i)   vector_value(v,i)
 */
#define putv(v,i,a)        Lputv(v,mkfxat(i),a)
#define getv(v,i)        Lgetv(v,mkfxat(i))

#define ZEROP(f)	zerop(f)


extern Sex      Lreorder(Sex a);

int             kernel_order_of_var(Sex x, Sex p);
extern Sex      Apoly2dist(Sex f, Sex p);
extern Sex      Lclone(Sex a);
extern Sex      make_dipoly(Sex a, Sex v);

Sex             homogenize(Sex f, Sex x, long n);
/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
Mcompare_var(Sex a, Sex b)
{
    long            r = 0;
    if (is_number(a)) {
	if (!is_number(b))
	    r = 1;
    } else if (is_number(b))
	r = -1;
    else {
	if (is_poly(a))
	    a = poly_var(a);
	if (is_poly(b))
	    b = poly_var(b);
	r = compare_var(a, b);
    }
    return mkfxat(r);
}

Sex
Mvar(Sex a)			// a ̎ϐo
{
    if (is_dipoly(a)) {
	a = dipoly_order(a);
	return cdr(a);
    }
    if (!is_poly(a))
	return nil;
    return poly_var(a);
}

Sex
Mpoly_pol(Sex a)
{
    if (!is_poly(a))
	return nil;
    return poly_pol(a);
}

Sex
Mmkterm(Sex c, Sex a)
{
    return cons(c, a);
}

#if 1
Sex
Adeg(Sex f, Sex x)
{
    if (!is_poly(f))
	return mkfxat(0);
    if (is_poly(x))
	x = poly_var(x);
    return Adeg_aux(f, x);
}

Sex
Adeg_aux(Sex f, Sex x)
{
    Sex             n = mkfxat(0);
    Sex             b,
                    m;

    if (!is_poly(f))
	return n;
    if (poly_var(f) == x) {
	return power(lterm(poly_pol(f)));
    }
    f = poly_pol(f);
    while (f != nil) {
	b = lterm(f);
	f = reduct(f);
	m = Adeg_aux(coeff(b), x);
	if (m > n)
	    n = m;
    }
    return n;
}
#else
Sex
Adeg(Sex f, Sex x)
{
    Sex             n = mkfxat(0);

    if (!is_poly(f))
	return n;
    if (is_poly(x))
	x = poly_var(x);
    stackcheck();
    push2(x, f);
    kernel_order = cons(x, kernel_order);
    f = Lreorder(stack[0]);
    x = stack[-1];
    popv(2);
    if (poly_var(f) == x) {
	n = power(lterm(poly_pol(f)));
    }
    kernel_order = cdr(kernel_order);
    return n;
}
#endif

int
kernel_order_of_var(Sex x, Sex p)
{
    int             n = 0;
    Sex             v;

    while (p != nil) {
	v = car(p);
	if (is_poly(v))
	    v = poly_var(v);
	if (v == x)
	    return n;
	n++;
	p = cdr(p);
    }
    return -1;
}


Sex
Atdeg(Sex f)			// total degree
{
    long            n;
    n = total_deg(f);
    return mkfxat(n);
}

Sex
Atdeg_coeff(Sex f, Sex x)	// total degree of coefficients
{
    Sex             g;
    long            n,
                    m;
    n = 0;

    if (!is_poly(f))
	return mkfxat(0);
    if (is_poly(x))
	x = poly_var(x);
    stackcheck();
    push2(x, f);
    kernel_order = cons(x, kernel_order);
    f = Lreorder(stack[0]);
    x = stack[-1];
    popv(2);
    if (poly_var(f) == x) {
	f = poly_pol(f);
	while (f != nil) {
	    g = lterm(f);
	    f = reduct(f);
	    m = total_deg(coeff(g));
	    if (n < m)
		n = m;
	}
    } else
	n = total_deg(f);

    kernel_order = cdr(kernel_order);
    return mkfxat(n);
}

Sex
Amtdeg_coeff(Sex f)		// total degree of coefficients
{
    Sex             g;
    long            n,
                    m;
    n = 0;

    if (!is_poly(f))
	return mkfxat(0);
    f = poly_pol(f);
    while (f != nil) {
	g = lterm(f);
	f = reduct(f);
	m = total_deg(coeff(g));
	if (n < m)
	    n = m;
    }
    return mkfxat(n);
}


long
dipoly_total_deg(Sex f)
{
    Sex             q,
                    a;
    long            order;
    long            n,
                    m,
                    k,
                    i;

    q = dipoly_order(f);
    order = mkint(car(q));
    q = dipoly_pol(f);
    if (!is_cons(q)) {
	return 0;
    }
    switch (order) {
    case TERM_ORDER_LEX:
	k = 0;
	while (is_cons(q)) {
	    a = car(q);
	    q = cdr(q);
#if DIPOLY_COEFF==0
	    a = cdr(a);
	    m = mkint(vect_dimx(a));
	    n = 0;
	    for (i = 0; i < m; i++)
		n = n + (long) vect_value(a, i) - TAG_FIXNUM;
	    n = mkint(n);
	    if (k < n)
		k = n;
#else
	    m = mkint(vect_dimx(a));
	    n = 0;
	    for (i = 1; i < m; i++)
		n = n + (long) vect_value(a, i) - TAG_FIXNUM;
	    n = mkint(n);
	    if (k < n)
		k = n;
#endif
	}
	return k;
    case TERM_ORDER_REVLEX:
    case TERM_ORDER_TLEX:
    case TERM_ORDER_TREVLEX:
	a = car(q);
#if DIPOLY_COEFF==0
	a = cdr(a);
	n = mkint(vect_value(a, 0));
#else
	n = mkint(vect_value(a, 1));
#endif
	return n;
    default:
	return 0;
    }
}

long
total_deg(Sex f)		// total degree
{
    Sex             g;
    long            n,
                    m;
    n = 0;

    if (is_dipoly(f))
	return dipoly_total_deg(f);
    if (!is_poly(f))
	return 0;
    f = poly_pol(f);
    while (f != nil) {
	g = lterm(f);
	f = reduct(f);
	m = mkint(power(g)) + total_deg(coeff(g));
	if (n < m)
	    n = m;
    }
    return n;
}

Sex
Ahomogenize(Sex f, Sex x)
{
    long            n;
    if (!is_poly(f))
	aerror1("homogenize: not poly", f);
    n = total_deg(f);
    f = homogenize(f, x, n);
    return f;
}

Sex
homogenize(Sex f, Sex x, long n)
{
    Sex             p;
    Sex             g;
    long            m;

    if (n == 0)
	return f;
    if (!is_poly(f)) {
	p = gexpt(x, mkfxat(n));
	return gtimes2(f, p);
    }
    push2(f, x);
    push3(nil, nil, nil);
#define f stack[-4]
#define x stack[-3]
#define p stack[-2]
#define a stack[-1]
#define y stack[0]
    y = poly_var(f);
    f = poly_pol(f);
    p = mkfxat(0);
    while (f != nil) {
	g = lterm(f);
	f = reduct(f);
	a = homogenize(coeff(g), x, n - mkint(power(g)));
	a = cons(a, power(g));
	p = cons(a, p);
    }
    p = Lnreverse(p);
    p = make_poly(y, p);
#undef f
#undef x
#undef p
#undef a
#undef y
    f = stack[-2];
    popv(5);
    return f;
}

Sex
Atdeg_monomial(Sex f)		// total degree of monomial
{
    long            n = 0;

    while (is_poly(f)) {
	f = lterm(poly_pol(f));
	n = n + mkint(power(f));
	f = coeff(f);
    }
    return mkfxat(n);
}

Sex
Ainit(Sex f, Sex x)
{
    return Alcoeff(f, x);
}

Sex
Alcoeff(Sex f, Sex x)
{
    Sex             c = nil;

    if (!is_poly(f))
	return f;
    if (is_poly(x))
	x = poly_var(x);
    ASSERT(is_symbol(x));
    stackcheck();
    push2(x, f);
    kernel_order = cons(x, kernel_order);
    f = Lreorder(stack[0]);
    x = stack[-1];
    popv(2);
    if (poly_var(f) == x) {
	c = coeff(lterm(poly_pol(f)));
    } else
	c = f;
    kernel_order = cdr(kernel_order);
    c = Lreorder(c);
    return c;
}

Sex
Areduct(Sex f, Sex x)
{
    Sex             c = nil;

    if (!is_poly(f))
	return mkfxat(0);
    if (is_poly(x))
	x = poly_var(x);
    ASSERT(is_symbol(x));
    stackcheck();
    push2(x, f);
    kernel_order = cons(x, kernel_order);
    f = Lreorder(stack[0]);
    x = stack[-1];
    popv(2);

    if (poly_var(f) == x) {
	c = reduct(poly_pol(f));
	if (c == nil)
	    c = mkfxat(0);
	else if (ldeg(c) == mkfxat(0))
	    c = lcoeff(c);
	else
	    c = make_poly(x, c);
    } else
	c = mkfxat(0);

    kernel_order = cdr(kernel_order);
    c = Lreorder(c);
    return c;
}


Sex
Alpower(Sex f, Sex x)
{
    Sex             c = nil;

    if (!is_poly(f))
	return mkfxat(1);
    if (is_poly(x))
	x = poly_var(x);
    ASSERT(is_symbol(x));
    stackcheck();
    push2(x, f);
    kernel_order = cons(x, kernel_order);
    f = Lreorder(stack[0]);
    x = stack[-1];
    popv(2);
    if (poly_var(f) == x) {
	c = make_polynomial(x, mkfxat(1), ldeg(poly_pol(f)));
    } else
	c = mkfxat(1);
    kernel_order = cdr(kernel_order);
    c = Lreorder(c);
    return c;
}



Sex
Alterm(Sex f, Sex x)
{
    Sex             c = nil;

    if (!is_poly(f))
	return f;
    if (is_poly(x))
	x = poly_var(x);
    ASSERT(is_symbol(x));
    stackcheck();
    push2(x, f);
    kernel_order = cons(x, kernel_order);
    f = Lreorder(stack[0]);
    x = stack[-1];
    popv(2);
    if (poly_var(f) == x) {
	c = make_polynomial(x, lcoeff(poly_pol(f)), ldeg(poly_pol(f)));
    } else
	c = f;
    kernel_order = cdr(kernel_order);
    c = Lreorder(c);
    return c;
}




Sex
Mdeg(Sex a)			// a ̎oĂ
{
    if (is_dipoly(a)) {
	a = dipoly_pol(a);
#if DIPOLY_COEFF == 0
	if (a != nil)
	    a = cdar(a);
#else
	if (a != nil)
	    a = car(a);
#endif
	return a;
    }


    if (is_poly(a))
	return power(lterm(poly_pol(a)));
    if (!is_number(a))
	return aerror1("mdeg: not poly", a);
    return (mkfxat(0));
    // error1("mdeg: not a vector", a);
}

Sex
Mlpower(Sex a)			// xL[[ŏ̍̃xL
{
    Sex             x;
    if (is_dipoly(a)) {
	x = dipoly_order(a);
	a = dipoly_pol(a);
	if (a == nil)
	    return mkfxat(0);
	a = car(a);
	push(x);
#if DIPOLY_COEFF == 0
	a = cons(mkfxat(1), cdr(a));
#else
	a = Lclone(a);
	vect_value(a, 0) = mkfxat(1);
#endif
	a = cons(a, nil);
	pop(x);
	return make_dipoly(a, x);
    }
    if (is_poly(a)) {
	x = poly_var(a);
	a = power(lterm(poly_pol(a)));
	return make_polynomial(x, mkfxat(1), a);
    }
    if (!is_number(a))
	return aerror1("mpower: not poly", a);
    return mkfxat(1);
}

Sex
Minit(Sex a)
{
    return Mlcoeff(a);
}

Sex
Mlcoeff(Sex a)			// W[[ŏ̍̌W
{

    if (is_dipoly(a)) {
	a = dipoly_pol(a);
	if (a == nil)
	    return mkfxat(0);
	a = car(a);
#if DIPOLY_COEFF == 0
	a = car(a);
#else
	a = vect_value(a, 0);
#endif
	return a;
    }
    if (is_poly(a))
	return coeff(lterm(poly_pol(a)));
    if (!is_number(a))
	return aerror1("minit: not poly", a);
    return a;
}

Sex
Mtail(Sex a)			// a區菜c
{
    Sex             x;
    if (is_dipoly(a)) {
	x = dipoly_order(a);
	a = dipoly_pol(a);
	if (a != nil)
	    a = cdr(a);
	return make_dipoly(a, x);
    } else if (is_poly(a))
	return reduct(poly_pol(a));
    else if (!is_number(a))
	return aerror1("mtail: not poly", a);
    return mkfxat(0);
}

Sex
Mreduct(Sex a)
{
    Sex             b;
    if (is_dipoly(a)) {
	b = dipoly_order(a);
	a = dipoly_pol(a);
	if (a != nil)
	    a = cdr(a);
	return make_dipoly(a, b);
    }
    if (is_poly(a)) {
	b = reduct(poly_pol(a));
	if (b == nil)
	    return mkfxat(0);
	else if (ldeg(b) == mkfxat(0))
	    return lcoeff(b);
	return make_poly(Mvar(a), Mtail(a));
    }
    if (!is_number(a))
	return aerror1("mreduct: not poly", a);
    return mkfxat(0);
}

Sex
Mlterm(Sex a)			// a̎區
{
    Sex             x;
    if (is_dipoly(a)) {
	x = dipoly_order(a);
	a = dipoly_pol(a);
	if (a == nil)
	    return mkfxat(0);
	a = car(a);
	push(x);
	a = cons(a, nil);
	pop(x);
	return make_dipoly(a, x);
    } else if (is_poly(a))
	return make_polynomial(Mvar(a), Minit(a), Mdeg(a));
    else if (!is_number(a))
	return aerror1("mlterm: not poly", a);
    else
	return a;
}

Sex
Mdiff(Sex a, Sex b)
{				/* 
				 * Z
				 */
    return gdiff2(a, b);
}

Sex
Mtimes(Sex a, Sex b)
{				/* 
				 * Z
				 */
    return gtimes2(a, b);
}

Sex
Mquot(Sex a, Sex b)
{				/* 
				 * Z
				 */
    return gquot2(a, b);
}


Sex
Mcont(Sex a)
{				/* 
				 * WGCD
				 */
    if (is_dipoly(a))
	return Mcont_dipoly(a);
    if (!is_poly(a))
	return a;
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = poly_pol(a);
    p = lcoeff(r);
    r = reduct(r);

    while (r != nil) {
	q = lcoeff(r);
	p = Mgcd(p, q);
	r = reduct(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}


Sex
Mcont_dipoly(Sex a)
{
    if (dipoly_pol(a) == nil)
	return mkfxat(0);
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define p stack[-2]
#define q stack[-1]
#define r stack[0]

    r = dipoly_pol(a);
    p = caar(r);
    r = cdr(r);
    while (r != nil) {
#if DIPOLY_COEFF == 0
	q = caar(r);
#else
	q = vect_value(car(r), 0);
#endif
	p = Mgcd(p, q);
	r = cdr(r);
    }
#undef a
#undef p
#undef q
#undef r
    a = stack[-2];
    popv(4);
    return a;
}


Sex
Mpp(Sex a)
{				/* 
				 * primitive part
				 */
    Sex             p;
    if (is_cons(a))
	return Mlist_pp(a);
    if (is_dipoly(a)) {
	return Mpp_dipoly(a);
    }
    if (!is_poly(a)) {
	if (a == mkfxat(0))
	    return a;
	return mkfxat(1);
    }
    push(a);
    p = Mcont(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}

Sex
Mlist_pp(Sex a)
{
    Sex             b;

    stackcheck();
    push2(a, nil);
#define a stack[-1]
#define c stack[0]
    while (a != nil) {
	b = car(a);
	b = Mpp(b);
	c = cons(b, c);
	a = cdr(a);
    }
#undef a
#undef c
    a = stack[0];
    popv(2);
    return Lnreverse(a);
}


Sex
Mpp_dipoly(Sex a)
{
    Sex             p;
    push(a);
    p = Mcont_dipoly(a);
    pop(a);
    p = Mquot(a, p);
    return p;
}


Sex
Mpquot(Sex a, Sex b)
{				/* 
				 * [
				 */
    long            o;
    Sex             c;

    stackcheck();
    push3(a, b, nil);

#define a stack[-2]
#define b stack[-1]
#define d stack[ 0]

    o = compare_var(Mvar(a), Mvar(b));

    if (o > 0) {
	return mkfxat(0);
    } else if (o < 0) {
	d = gplus2(Mdeg(a), mkfxat(1));

	while (d != mkfxat(0)) {
	    a = gtimes2(a, b);
	    d = gdiff2(d, mkfxat(1));
	}

    } else if (o == 0) {
	d = gdiff2(Mdeg(a), Mdeg(b));
	d = gplus2(d, mkfxat(1));

	while (d != mkfxat(0)) {
	    a = gtimes2(a, Minit(b));
	    d = gdiff2(d, mkfxat(1));
	}
    }
    c = Mquot(a, b);
#undef a
#undef b
#undef d
    popv(3);
    return c;
}

#if 0
Sex
fpol_rem(Sex a, Sex b)
{
    Sex             r;
    push5(a, b, nil, nil, nil);

#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define q stack[-1]
#define rem stack[0]


    if (is_number(a) && is_number(b)) {
	q = gratio2(a, b);
	c = gtimes2(b, q);
	r = gdiff2(a, c);
	popv(5);
	return r;
    } else if (is_number(a) && is_poly(b)) {
	r = a;
	popv(5);
	return r;
    } else if (is_poly(a) && is_number(b)) {
	popv(5);
	return mkfxat(0);
    } else if (is_poly(a) && is_poly(b)) {
	o = compare_var(Mvar(a), Mvar(b));
	if (o < 0) {
	    popv(5);
	    return mkfxat(0);
	} else if (o > 0) {
	    r = a;
	    popv(5);
	    return r;
	} else if (o == 0) {
	    d = mkint(gdiff2(Mdeg(a), Mdeg(b)));
	    if (d >= 0) {
		c = gquot2(a, b);
		c = gtimes2(b, c);
		rem = gdiff2(a, c);
		rem = f_normalize(rem);
		// lisp_printf("Mrem: rem = %p\n", rem); 
		if (rem == mkfxat(0)) {
		    r = rem;
		    popv(5);
		    return r;
		} else if (rem != mkfxat(0)) {
		    r = Mrem(rem, b);
		    popv(5);
		    return r;
		}
	    } else if (d < 0) {
		r = a;
		popv(5);
		return r;
	    }
	}
    }
    return error("rem: no such args");

#undef a
#undef b
#undef c
#undef q
#undef rem

}
#endif


Sex
Mrem(Sex a, Sex b)
{				/* 
				 * ]
				 */
    long            o,
                    d;
    long            db;
    Sex             r;

    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define q stack[-1]
#define rem stack[0]


    // lisp_printf("Mrem: a=%p", a); lisp_printf(", b=%p\n", b);
    if (is_number(a) && is_number(b)) {
	q = gratio2(a, b);
	c = gtimes2(b, q);
	r = gdiff2(a, c);
	popv(5);
	return r;
    } else if (is_number(a) && is_poly(b)) {
	r = a;
	popv(5);
	return r;
    } else if (is_poly(a) && is_number(b)) {
	popv(5);
	return mkfxat(0);
    } else if (is_poly(a) && is_poly(b)) {
	o = compare_var(Mvar(a), Mvar(b));
	if (o < 0) {
	    popv(5);
	    return mkfxat(0);
	} else if (o > 0) {
	    r = a;
	    popv(5);
	    return r;
	} else if (o == 0) {
	    d = mkint(gdiff2(Mdeg(a), Mdeg(b)));
	    if (d >= 0) {
		db = mkint(Mdeg(b));
		c = gquot2(a, b);
		c = gtimes2(b, c);
		rem = gdiff2(a, c);
		while (is_poly(rem) && compare_var(Mvar(rem), Mvar(b)) == 0
		       && mkint(Mdeg(rem)) >= db)
		    rem = Mreduct(rem);
		// rem = f_normalize(rem);
		// lisp_printf("Mrem: rem = %p\n", rem);
		if (rem == mkfxat(0)) {
		    r = rem;
		    popv(5);
		    return r;
		} else if (rem != mkfxat(0)) {
		    r = Mrem(rem, b);
		    popv(5);
		    return r;
		}
	    } else if (d < 0) {
		r = a;
		popv(5);
		return r;
	    }
	}
    }
    return error("rem: no such args");

#undef a
#undef b
#undef c
#undef q
#undef rem
}

Sex
Mprem(Sex a, Sex b)
{				/* 
				 * []
				 */
    long            o;

    if (is_number(a) && is_number(b)) {
	return Mrem(a, b);
    } else if (is_number(a) && is_poly(b))
	return a;
    else if (is_poly(a) && is_number(b)) {
	return mkfxat(0);
/**
	push(b);
	a = gtimes2(a, b);
	pop(b);
	return Mrem(a, b);
*/
    } else if (is_poly(a) && is_poly(b)) {
	o = compare_var(Mvar(a), Mvar(b));
	if (o < 0) {
	    return mkfxat(0);
	} else if (o > 0) {
	    return a;
	} else if (o == 0) {
	    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define d stack[0]
	    d = gdiff2(Mdeg(a), Mdeg(b));
	    if (d < mkfxat(0)) {
		a = b;
		b = a;
		d = gdiff2(Mdeg(a), Mdeg(b));
	    }
	    d = gplus2(d, mkfxat(1));

	    while (d != mkfxat(0)) {
		a = gtimes2(a, Minit(b));
		d = gdiff2(d, mkfxat(1));
	    }
	    b = Mrem(a, b);
#undef a
	    a = b;
#undef b
#undef d
	    popv(3);
	    return a;
	}
    }
    return error(" prem : invalid arguement");
}

Sex
Mpgcd(Sex a, Sex b)
{				/* 
				 * [GCD
				 */
    Sex             c;


    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define c stack[0]

    // c = Mprem(a, b);
    c = wuremainder(a, b);

    while (c != mkfxat(0)) {
	a = b;
	b = c;
	// c = Mprem(a, b);
	c = wuremainder(a, b);
    }
#undef c
    c = Mpp(b);
#undef a
#undef b
    popv(3);
    return c;

}

Sex
Mlcm(Sex a, Sex b)
{
    Sex             g;
    push2(a, b);
    g = Mgcd(a, b);
    pop2(b, a);
    push(g);
    a = gtimes2(a, b);
    pop(g);
    a = gquot2(a, g);
    return a;
}

Sex
Mgcd(Sex a, Sex b)
{
    Sex             gcd_sprs(Sex, Sex);
    int             dirty = NO;
    if (is_dipoly(a)) {
	push(b);
	a = Lpolynomial(a);
	pop(b);
	dirty = YES;
    }
    if (is_dipoly(b)) {
	push(a);
	b = Lpolynomial(b);
	pop(a);
	dirty = YES;
    }
    a = gcd_sprs(a, b);
    if (dirty)
	a = Ldipoly(a);
    return a;
}

Sex
gcd_pprs(Sex a, Sex b)
{				/* 
				 * GCD primitive prs
				 */
    Sex             g;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (a == mkfxat(0))
	    return b;
	push(a);
	b = Mcont(b);
	pop(a);
	return Mgcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (b == mkfxat(0))
	    return a;
	push(b);
	a = Mcont(a);
	pop(b);
	return Mgcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(Mvar(a), Mvar(b));

	if (o > 0) {
	    push(a);
	    b = Mcont(b);	/* 
				 * b̌Wgcd
				 */
	    pop(a);
	    return Mgcd(a, b);
	} else if (o < 0) {
	    push(b);
	    a = Mcont(a);
	    pop(b);
	    return Mgcd(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push(nil);
#define a stack[-5]
#define b stack[-4]
#define p stack[-3]
#define q stack[-2]
#define g1 stack[-1]
#define g2 stack[0]

	    g1 = Mcont(a);
	    g2 = Mcont(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = Mgcd(g1, g2);

	    g1 = wuremainder(p, q);
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		p = q;
		q = g1;
		g1 = wuremainder(p, q);
		g1 = Mpp(g1);
	    }
	    if (g1 == mkfxat(0)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = Mpp(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
	    popv(6);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}

Sex
gcd_rprs(Sex a, Sex b)
{				/* 
				 * GCD reduced prs
				 */
    Sex             g;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (a == mkfxat(0))
	    return b;
	push(a);
	b = Mcont(b);
	pop(a);
	return Mgcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (b == mkfxat(0))
	    return a;
	push(b);
	a = Mcont(a);
	pop(b);
	return Mgcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(Mvar(a), Mvar(b));

	if (o > 0) {
	    push(a);
	    b = Mcont(b);	/* 
				 * b̌Wgcd
				 */
	    pop(a);
	    return Mgcd(a, b);
	} else if (o < 0) {
	    push(b);
	    a = Mcont(a);
	    pop(b);
	    return Mgcd(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push2(nil, nil);
#define a stack[-6]
#define b stack[-5]
#define p stack[-4]
#define q stack[-3]
#define g1 stack[-2]
#define g2 stack[-1]
#define e1 stack[0]

	    if (power(lterm(poly_pol(a))) < power(lterm(poly_pol(b)))) {
		p = a;
		a = b;
		b = p;
	    }
	    g1 = Mcont(a);
	    g2 = Mcont(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = Mgcd(g1, g2);

	    g1 = wuremainder(p, q);
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		d1 = mkint(power(lterm(poly_pol(p)))) -
		    mkint(power(lterm(poly_pol(q)))) + 1;
		e1 = gexpt(coeff(lterm(poly_pol(q))), mkfxat(d1));
		p = q;
		q = g1;
		g1 = wuremainder(p, q);
		g1 = gquot2(g1, e1);
	    }
	    if (g1 == mkfxat(0)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = Mpp(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
#undef e1
	    popv(7);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}

Sex
gcd_sprs(Sex a, Sex b)
{				/* 
				 * GCD subresultant prs
				 */
    Sex             g;
    Sex             s;
    long            o,
                    d;
    long            d1,
                    d2;

    if (is_number(a) && is_number(b))
	return gcd(a, b);

    stackcheck();
    if (is_number(a) && is_poly(b)) {
	if (ZEROP(a))
	    return b;
	push(a);
	b = Mcont(b);
	pop(a);
	return Mgcd(a, b);

    }
    if (is_poly(a) && is_number(b)) {
	if (ZEROP(b))
	    return a;
	push(b);
	a = Mcont(a);
	pop(b);
	return Mgcd(a, b);

    }
    if (is_poly(a) && is_poly(b)) {

	o = compare_var(Mvar(a), Mvar(b));

	if (o > 0) {
	    push(a);
	    b = Mcont(b);	/* 
				 * b̌Wgcd
				 */
	    pop(a);
	    return Mgcd(a, b);
	} else if (o < 0) {
	    push(b);
	    a = Mcont(a);
	    pop(b);
	    return Mgcd(a, b);
	} else if (o == 0) {
	    stackcheck();
	    push5(a, b, nil, nil, nil);
	    push4(nil, nil, nil, nil);
#define a stack[-8]
#define b stack[-7]
#define p stack[-6]
#define q stack[-5]
#define g1 stack[-4]
#define g2 stack[-3]
#define e1 stack[-2]
#define s1 stack[-1]
#define t1 stack[0]

	    if (power(lterm(poly_pol(a))) < power(lterm(poly_pol(b)))) {
		p = a;
		a = b;
		b = p;
	    }
	    g1 = Mcont(a);
	    g2 = Mcont(b);
	    p = gquot2(a, g1);
	    q = gquot2(b, g2);
	    g2 = Mgcd(g1, g2);

	    g1 = wuremainder(p, q);
	    d1 = mkint(power(lterm(poly_pol(p)))) -
		mkint(power(lterm(poly_pol(q))));
	    t1 = mkfxat(1);
	    while (is_poly(g1) && poly_var(a) == poly_var(g1)) {
		e1 = gexpt(t1, mkfxat(d1 - 1));
		t1 = gexpt(lcoeff(poly_pol(q)), mkfxat(d1));
		t1 = gquot2(t1, e1);
		d1 = mkint(power(lterm(poly_pol(q)))) -
		    mkint(power(lterm(poly_pol(g1))));
		s1 = gexpt(t1, mkfxat(d1));
		s1 = gtimes2(s1, lcoeff(poly_pol(q)));
		p = q;
		q = g1;
		g1 = wuremainder(p, q);
		g1 = gquot2(g1, s1);
	    }
	    if (ZEROP(g1)) {
		if (is_poly(q) && poly_var(q) == poly_var(a))
		    g1 = Mpp(q);
		else
		    g1 = mkfxat(1);
	    } else
		g1 = mkfxat(1);

	    g = gtimes2(g1, g2);
#undef a
#undef b
#undef p
#undef q
#undef g1
#undef g2
#undef e1
#undef s1
#undef t1
	    popv(9);
	    return g;
	}
    }
    return error(" gcd : invalid arguement ");
}


#if 0
Sex
wuremainder(Sex a, Sex b)
{
    Sex             s;
    long            i,
                    j,
                    k;
    long            n,
                    m;
    long            o;

    if (!is_poly(b))
	return mkfxat(0);

    if (!is_poly(a))
	return a;

    stackcheck();
    push3(nil, a, b);
    push4(nil, nil, nil, nil);
    push2(nil, nil);

#define cu         stack[-8]
#define a stack[-7]
#define b stack[-6]
#define x stack[-5]
#define u stack[-4]
#define v stack[-3]
#define cv stack[-2]
#define tu stack[-1]
#define tv stack[0]


    x = poly_var(b);
    u = poly_var(a);
    o = compare_var(x, u);
    if (o < 0) {
	b = a;
	goto owari1;
    } else if (o > 0) {
	b = mkfxat(0);
	goto owari1;
    } else {
	a = poly_pol(a);
	b = poly_pol(b);
	m = mkint(power(lterm(a)));
	n = mkint(power(lterm(b)));
	k = m - n;
	if (k < 0) {
	    b = a;
	    goto owari2;
	}
	u = Lmkvect(mkfxat(m + 1));
	for (i = m; i >= 0; i--) {
	    if (a != nil && power(lterm(a)) == mkfxat(i)) {
		putv(u, i, coeff(lterm(a)));
		a = reduct(a);
	    } else
		putv(u, i, mkfxat(0));
	}
	v = Lmkvect(mkfxat(n + 1));
	for (i = n; i >= 0; i--) {
	    if (b != nil && power(lterm(b)) == mkfxat(i)) {
		putv(v, i, coeff(lterm(b)));
		b = reduct(b);
	    } else
		putv(v, i, mkfxat(0));
	}
	cv = getv(v, n);
	for (; k >= 0; k--) {
	    cu = getv(u, n + k);
	    for (j = n + k - 1; j >= k; j--) {
		// printf("k, j = %d, %d\n", k, j);
		tu = gtimes2(cv, getv(u, j));
		tv = gtimes2(cu, getv(v, j - k));
		tu = gdiff2(tu, tv);
		putv(u, j, tu);
	    }
	    for (j = k - 1; j >= 0; j--) {
		tu = gtimes2(cv, getv(u, j));
		putv(u, j, tu);
	    }
	}
	b = nil;
	for (k = 0; k < n; k++) {
	    a = getv(u, k);
	    if (!zerop(a)) {
		s = cons(a, mkfxat(k));
		b = cons(s, b);
	    }
	}
    }
  owari2:
    b = make_poly(x, b);
  owari1:
#undef tu
#undef tv
#undef cu
#undef x
#undef u
#undef v
#undef cv
#undef a
    a = b;
#undef b
    popv(9);
    return a;
}
#endif

Sex
Msqfree(Sex a, Sex x)
{				/* 
				 * square-free decomposition
				 */
    Sex             s;
    Sex             r;
    long            i;
    Sex             AAdiff(Sex, Sex);

    if (!is_poly(a))
	return a;
    if (is_poly(x))
	x = poly_var(x);
    else if (!is_symbol(x))
	x = poly_var(a);

    push5(a, x, nil, nil, nil);
    push2(nil, nil);
#define v stack[-6]
#define x stack[-5]
#define q stack[-4]
#define w stack[-3]
#define r stack[-2]
#define g stack[-1]
#define c stack[0]

    w = AAdiff(v, x);
    g = Mgcd(v, w);
    c = g;
    i = 0;
    while (is_poly(v)) {
	v = gquot2(v, g);
	w = gquot2(w, g);
	g = AAdiff(v, x);
	w = gdiff2(w, g);
	i++;
	g = Mgcd(v, w);
	if (is_poly(g)) {
	    s = Lncons(mkfxat(i));
	    s = cons(g, s);
	    s = Lncons(s);
	    r = Lappend(r, s);
	    s = gexpt(g, mkfxat(i - 1));
	    c = gquot2(c, s);
	}
    }
    a = Lappend(r, Lncons(c));
#undef r
#undef x
#undef q
#undef v
#undef w
#undef g
#undef c
    popv(7);
    return a;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
MinitRed(Sex a, Sex b)
{				/* 
				 * initialy reduced ǂ̔
     *//*
     * * ^ȂtrueCUȂnil Ԃ 
     */
    long            d;
    long            o;

    if (is_number(a) && !is_number(b)) {
	return true;
    } else if (!is_number(a) && !is_number(b)) {
	o = compare_var(Mvar(a), Mvar(b));
	if (o > 0) {
	    return true;
	} else if (o == 0) {
	    d = mkint(Mdeg(a)) - mkint(Mdeg(b));
	    if (d < 0)
		return true;
	    else if (d >= 0) {	/* 
				 * * iredłȂƂ 
				 */
		return nil;
	    }
	} else if (o < 0) {
	    push(b);
	    a = Minit(a);
	    pop(b);
	    return MinitRed(a, b);
	}
    }
    return error(" initred : invalid arguement ");
}

Sex
MiRed(Sex a, Sex b)
{				/* 
				 * initially reduced ֐
				 */
    stackcheck();
    push5(a, b, nil, nil, nil);
    push3(nil, nil, nil);
#define a stack[-7]
#define b stack[-6]
#define c stack[-5]
#define la stack[-4]
#define lb stack[-3]
#define ka stack[-2]
#define kb stack[-1]
#define g stack[0]


    while (MinitRed(a, b) == nil) {
	la = Mlterm(a);
	lb = Mlterm(b);
	g = Mgcd(la, lb);

	ka = gquot2(la, g);
	kb = gquot2(lb, g);

	a = gtimes2(a, kb);
	c = gtimes2(b, ka);

	a = gdiff2(a, c);
    }
#undef a
#undef b
#undef c
#undef la
#undef lb
#undef ka
#undef kb
#undef g
    a = stack[-7];
    popv(8);
    return a;
}

/*
 * Sex McharSet(Sex F) { Sex R=nil,Q=nil,C=nil,P=nil,p;
 * 
 * R=Mexcept_zero(F);
 * 
 * if(R != nil && !Mnumber_in(R)){ Q=Lunion(R,Q); Q=Lunion(C,Q);
 * C=MbasicSet(Q); Q=Mrset(Q); R=MiRed(Q,C); return Q; } p=mkfxat(1);
 * P=cons(p,nil); return P; }
 */

extern Sex      Mintersect(Sex a, Sex b);

Sex
Minsec(Sex a, Sex b)
{
    return Mintersect(a, b);
}


#if 0
Sex
Mcset(Sex a)
{				/* 
				 * characteristic set
				 */

    long            flag1 = 0,
	flag2 = 0,
	o = 0;
    Sex             x,
                    rr;

    stackcheck();
    push5(a, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push(nil);
#define a stack[-10]
#define ps stack[-9]
#define cs stack[-8]
#define rs stack[-7]
#define c stack[-6]
#define r stack[-5]
#define rem stack[-4]
#define rem1 stack[-3]
#define tailp stack[-2]
#define tailc stack[-1]
#define tail stack[0]

    /*
     * ps =Msort2(a); 
     */
    ps = a;
    for (;;) {
	cs = MbasicSet2(ps);
	// lisp_printf("ps=%p\n", ps);
	// lisp_printf("cs=%p\n", cs);
	/*
	 * ps  descending set
	 * 傫
	 */

	tailp = ps;
	rs = nil;
	while (tailp != nil) {
	    r = car(tailp);
	    tailp = cdr(tailp);
	    tailc = cs;
	    while (tailc != nil) {
		c = car(tailc);
		tailc = cdr(tailc);
		rem = wuremainder(r, c);
		// lisp_printf("prem1=%p\n", rem);
	    }
	    if (!zerop(rem))
		rs = cons(rem, rs);
	}
	if (rs == nil) {
	    rr = Msort2(cs);
	    popv(11);
	    return rr;		/* 
				 * *   
				 */
	} else if (Mnumber_in(rs) == mkfxat(1)) {
	    rr = cons(mkfxat(1), nil);
	    popv(11);
	    return rr;
	}
	ps = Mjoin(ps, rs);
    }
#undef a
#undef ps
#undef cs
#undef rs
#undef c
#undef r
#undef rem
#undef rem1
#undef tailp
#undef tailc
#undef tail
}
#endif

Sex
Mtriangular(Sex a)
{				/* 
				 * triangular set ǂ
				 */
    long            o,
                    flag = 0;

    push4(a, nil, nil, nil);
#define a stack[-3]
#define p1 stack[-2]
#define p2 stack[-1]
#define tail stack[0]
    a = Msort1(a);
    p1 = car(a);
    tail = cdr(a);
    p2 = car(tail);

    while (tail != nil) {
	if (is_number(p1) || is_number(p2)) {
	    flag = 1;
	    break;
	}
	o = compare_var(poly_var(p1), poly_var(p2));
	if (o == 0) {
	    flag = 2;
	    break;
	}
	p1 = p2;
	tail = cdr(tail);
	p2 = car(tail);
    }

    if (flag != 0)
	lisp_printf("This set is NOT triangular:");
    else
	lisp_printf("This set is triangular:");

#undef a
    a = stack[-3];
#undef p1
#undef p2
#undef tail
    popv(4);
    return a;
}
/*
 */


int
Mcomp(const Sex * a, const Sex * b)
{
    long            o,
                    d;
    // k = 1;

    if (is_number(*a) && is_number(*b)) {
	if (lessp2(*a, *b))
	    return (1);
	else if (*a > *b)
	    return (-1);
	return (0);
    }
    if (is_number(*a) && is_poly(*b)) {
	return (1);
    }
    if (is_poly(*a) && is_number(*b)) {
	return (-1);
    }
    if (is_poly(*a) && is_poly(*b)) {
	o = compare_var(Mvar(*a), Mvar(*b));
	if (o < 0)
	    return (1);
	else if (o > 0)
	    return (-1);
	else if (o == 0) {
	    d = Mdeg(*a) - Mdeg(*b);
	    if (d < 0)
		return (-1);
	    else if (d > 0)
		return (1);
	    else if (d == 0)
		return 0;
	}
    }
    return 0;
}

Sex
Msort1(Sex a)
{				/* 
				 * ascending set
				 */
    Sex             a1,
                    p,
                    q,
                    tail,
                    data,
                    listp = nil;
    long            i = 1,
	SIZE = 1,
	j;

    a1 = car(a);
    tail = cdr(a);

    while (tail != nil) {
	a1 = car(tail);
	tail = cdr(tail);
	SIZE++;
    }

    p = Lmkvect(mkfxat(SIZE - 1));
    q = car(a);
    tail = cdr(a);
    Lputv(p, mkfxat(0), q);

    while (tail != nil) {
	q = car(tail);
	Lputv(p, mkfxat(i), q);
	tail = cdr(tail);
	i++;
    }
    qsort(&vector_value(p, 0), SIZE, SEX, (compare_fn *) Mcomp);

    j = SIZE - 1;
    while (j >= 0) {
	data = Lgetv(p, mkfxat(j));
	listp = cons(data, listp);
	j--;
    }
    return listp;
}

Sex
Msort2(Sex a)
{				/* 
				 * descending set~
				 */
    Sex             a1,
                    p,
                    q,
                    tail,
                    data,
                    listp = nil;
    long            i = 1,
	SIZE = 1,
	j;

    a1 = car(a);
    tail = cdr(a);

    while (tail != nil) {
	a1 = car(tail);
	tail = cdr(tail);
	SIZE++;
    }

    p = Lmkvect(mkfxat(SIZE - 1));
    q = car(a);
    tail = cdr(a);
    Lputv(p, mkfxat(0), q);

    while (tail != nil) {
	q = car(tail);
	Lputv(p, mkfxat(i), q);
	tail = cdr(tail);
	i++;
    }
    qsort(&vector_value(p, 0), SIZE, 4, (compare_fn *) Mcomp);

    j = 0;
    while (j <= SIZE - 1) {
	data = Lgetv(p, mkfxat(j));
	listp = cons(data, listp);
	j++;
    }
    return listp;
}

Sex
MbasicSet2(Sex a)
{				/* 
				 * descending set偨
				 */
    Sex             p0,
                    p1,
                    p2,
                    tail,
                    bset = nil,
	rset = nil;
    long            o;

    a = Msort2(a);		/* 
				 * descending set
				 */

    p1 = car(a);
    tail = cdr(a);
    p2 = car(tail);
    while (tail != nil) {
	if (is_number(p1) && is_number(p2))
	    return nil;
	else if (!is_number(p1) && is_number(p2))
	    break;

	o = compare_var(poly_var(p1), poly_var(p2));
	if (o == 0) {
	    rset = cons(p1, rset);	/* 
					 * abset菜
					 */
	    p0 = p1;
	    p1 = p2;
	    tail = cdr(tail);
	    p2 = car(tail);
	} else {
	    bset = cons(p1, bset);
	    p0 = p1;
	    p1 = p2;
	    tail = cdr(tail);
	    p2 = car(tail);
	}
    }
    bset = cons(p1, bset);
    bset = Msort2(bset);
    return bset;
}

Sex
MbasicSet1(Sex a)
{				/* 
				 * ascending set
				 */
    Sex             p0,
                    p1,
                    p2,
                    tail,
                    bset = nil,
	rset = nil;
    long            o,
                    flag = 0;

    a = Msort1(a);		/* 
				 * ascending set
				 */
    // lisp_printf("a=%p\n", a);

    p1 = car(a);
    tail = cdr(a);
    p2 = car(tail);
    while (tail != nil) {
	if (is_number(p1) && is_number(p2)) {
	    return nil;
	} else if (is_number(p1) && is_poly(p2)) {
	    p1 = p2;
	    tail = cdr(tail);
	    p2 = car(tail);
	} else if (is_poly(p1) && is_poly(p2)) {
	    o = compare_var(poly_var(p1), poly_var(p2));
	    if (o == 0) {
		if (flag == 0) {
		    bset = cons(p1, bset);
		    flag++;
		}
		rset = cons(p2, rset);	/* 
					 * abset菜
					 */
		p0 = p1;
		p1 = p2;
		tail = cdr(tail);
		p2 = car(tail);
	    } else {
		if (flag == 0) {
		    bset = cons(p1, bset);
		    flag++;
		}
		bset = cons(p2, bset);
		p0 = p1;
		p1 = p2;
		tail = cdr(tail);
		p2 = car(tail);
	    }
	}
    }
    /*
     * bset=cons(p1,bset);
     */
    bset = Msort1(bset);
    return bset;
}

bool
Mnumber_in(Sex a)
{
    Sex             a1,
                    tail;
    long            flag = 0;

    a1 = car(a);
    tail = cdr(a);

    while (tail != nil) {
	if (is_number(a1)) {
	    flag++;
	    a1 = car(tail);
	    tail = cdr(tail);
	} else {
	    a1 = car(tail);
	    tail = cdr(tail);
	}
    }
    if (flag > 0)
	return YES;
    return NO;
}

Sex             Lunion(Sex a, Sex b);

Sex
Mexcept_zero(Sex a)
{				/* 
				 * 0菜֐
				 */
    Sex             a1,
                    tail,
                    p = nil;

    a1 = car(a);
    tail = cdr(a);

    while (tail != nil) {
	if (zerop(a1)) {
	    a1 = car(tail);
	    tail = cdr(tail);
	} else if (!zerop(a1)) {
	    p = cons(a1, p);
	    a1 = car(tail);
	    tail = cdr(tail);
	}
    }
    if (!zerop(a1))
	p = cons(a1, p);
    return p;
}


Sex
Mrset(Sex a)
{				/* 
				 * Waabasicset֐
				 */
    Sex             p0,
                    p1,
                    p2,
                    tail,
                    rset = nil;
    long            o;

    a = Msort1(a);		/* 
				 * ascending set
				 */

    p1 = car(a);
    tail = cdr(a);
    p2 = car(tail);

    while (tail != nil) {
	o = compare_var(poly_var(p1), poly_var(p2));
	if (o == 0) {
	    rset = cons(p1, rset);	/* 
					 * abset菜
					 */
	    p0 = p1;
	    p1 = p2;
	    tail = cdr(tail);
	    p2 = car(tail);
	} else {
	    p0 = p1;
	    p1 = p2;
	    tail = cdr(tail);
	    p2 = car(tail);
	}
    }
    rset = Msort1(rset);
    return rset;
}

/*
 */

Sex
Norm(Sex a)
{
    Sex             norm;

    push4(nil, a, nil, nil);

#define onorm stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[0]

    c = mkfxat(0);
    while (is_poly(a)) {
	b = gtimes2(Minit(a), Minit(a));
	c = gplus2(c, b);
	a = Mreduct(a);
    }
    b = gtimes2(a, a);
    c = gplus2(c, b);

    c = Lsqrt(c);
    norm = c;
#undef onorm
#undef a
#undef b
#undef c
    popv(4);
    return norm;
}
/*
 * Sex Norm(Sex a) { Sex norm;
 * 
 * push3(a,nil,nil);
 * 
 * #define a stack[-2] #define b stack[-1] #define norm stack[0]
 * 
 * norm = mkfxat(0); while (is_poly(a)) { //Mdeg(a) != mkfxat(0)){ b =
 * gtimes2(Minit(a), Minit(a)); norm = gplus2(norm, b); //norm += Mvar(a)
 * * Mvar(a); a = Mreduct(a); // lisp_printf("a = %p\n", a); } b =
 * gtimes2(a,a); norm = gplus2(norm, b);
 * 
 * norm=Lsqrt(norm);
 * 
 * #undef norm norm = stack[0]; lisp_printf("norm = %p\n", norm); #undef a
 * #undef b popv(3); return norm; }
 */


Sex
AAdiff(Sex a, Sex x)
{
    Sex             res;

    if (!is_poly(a))
	return mkfxat(0);

    push5(a, x, nil, nil, nil);

#define a stack[-4]
#define x stack[-3]
#define b stack[-2]
#define c stack[-1]
#define res stack[0]

    res = mkfxat(0);
    if (is_poly(x))
	x = Mvar(x);

    while (is_poly(a)) {
	if (Mvar(a) == x) {
	    b = Minit(a);
	    c = Mdeg(a);
	    b = gtimes2(b, c);
	    c = gdiff2(c, mkfxat(1));
	    c = make_polynomial(Mvar(a), b, c);
	    res = gplus2(res, c);
	} else {
	    b = AAdiff(Minit(a), x);
	    if (!gzerop(b)) {
		c = make_polynomial(Mvar(a), b, Mdeg(a));
		res = gplus2(res, c);
	    }
	}
	a = Mreduct(a);
    }

#undef res
    res = stack[0];
#undef a
#undef x
#undef b
#undef c
    popv(5);
    return res;
}

Sex
Lbibun(Sex a)
{
    Sex             norm;
    push4(a, nil, nil, nil);

#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define norm stack[0]

    norm = mkfxat(0);
    while (is_poly(a)) {
	// Mdeg(a) != mkfxat(0)){
	b = Minit(a);
	c = Mdeg(a);
	b = gtimes2(b, c);
	c = gdiff2(c, mkfxat(1));
	c = make_polynomial(Mvar(a), b, c);
	norm = gplus2(norm, c);

	a = Mreduct(a);
	// lisp_printf("a = %p\n", a);
    }

#undef norm
    norm = stack[0];
#undef a
#undef b
#undef c
    popv(4);
    return norm;
}

Sex
Float(Sex a)
{
    push4(a, nil, nil, nil);

#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[0]

    d = mkfxat(0);
    while (is_poly(a)) {
	// Mdeg(a) != mkfxat(0)){
	b = Minit(a);
	b = Float(b);
	c = Mdeg(a);
	c = make_polynomial(Mvar(a), b, c);
	d = gplus2(d, c);

	a = Mreduct(a);
	// lisp_printf("a = %p\n", a);
    }
    a = Lfloat(a);
    d = gplus2(d, a);

#undef a
    a = stack[0];
#undef b
#undef c
#undef d
    popv(4);
    return a;


}

Sex
w_define(Sex a, Sex b)
{
    Sex             w;
    // Sex s;
    // Sex t;
    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define s stack[-1]
#define t stack[0]

    s = Minit(a);
    t = Minit(b);
    if (s * s > t * t)
	w = s;
    else {
	if (s * s > t * t)
	    w = t;
	else {
	    if (s > 0)
		w = s;
	    else
		w = t;
	}
    }
#undef a
#undef b
#undef s
#undef t
    popv(4);
    return w;
}

Sex
STrem(Sex a)
{
    Sex             A;
    push5(a, nil, nil, nil, nil);
    push5(nil, nil, nil, nil, nil);
    push3(nil, nil, nil);

#define a stack[-12]
#define b stack[-11]
#define q stack[-10]
#define r stack[-9]
#define s0 stack[-8]
#define s1 stack[-7]
#define s2 stack[-6]
#define t0 stack[-5]
#define t1 stack[-4]
#define t2 stack[-3]
#define m stack[-2]
#define w stack[-1]
#define norm stack[0]
    m = Mdeg(a);
    b = Lbibun(a);
    b = Mquot(b, m);
    s0 = make_float(1.0);
    s1 = make_float(0.0);
    t0 = make_float(0.0);
    t1 = make_float(1.0);
    norm = make_float(1.0);
    norm = Norm(a);
    norm = Norm(b);

    // while(is_poly(a)){
    while (Mdeg(a) >= mkfxat(2)) {
	// lisp_printf("\na = %p\n", a);
	// lisp_printf("b = %p\n",b);

	q = Mquot(a, b);
	// lisp_printf("q = %p\n", q);

	r = Mrem(a, b);
	// lisp_printf("r = %p\n", r);

	s2 = gtimes2(q, s1);
	// lisp_printf("q*s1 = %p\n", s2);
	s2 = gdiff2(s0, s2);
	// lisp_printf("s2 = %p\n", s2);
	t2 = gtimes2(q, t1);
	// lisp_printf("q*t1 = %p\n", t2);
	t2 = gdiff2(t0, t2);
	// lisp_printf("t2 = %p\n", t2);

	w = w_define(s2, t2);
	// lisp_printf("w = %p\n", w);
	s2 = Mquot(s2, w);
	t2 = Mquot(t2, w);
	s0 = s1;
	s1 = s2;
	t0 = t1;
	t1 = t2;
	// lisp_printf("s0 = %p, ",s0);
	// lisp_printf("s1 = %p\n",s1);
	// lisp_printf("t0 = %p, ",t0);
	// lisp_printf("t1 = %p\n",t1);
	r = Mquot(r, w);
	// lisp_printf("r = %p\n", r);
	norm = Norm(r);
	a = b;
	b = r;
    }
    A = a;
#undef a
#undef b
#undef q
#undef r
#undef s0
#undef s1
#undef s2
#undef t0
#undef t1
#undef t2
#undef m
#undef w
#undef norm

    popv(13);
    return A;
}

Sex
Lis_number(Sex a)
{
    return is_number(a) ? true : nil;
}
Sex
Lis_poly(Sex a)
{
    return is_poly(a) ? true : nil;
}

setup_type const triangular_setup[] = {
    {"poly_pol", (lisp_proc *) Mpoly_pol, 1},
    {"poly_var", (lisp_proc *) Mvar, 1},
    {"mkterm", (lisp_proc *) Mmkterm, 1},

    {"mvar", (lisp_proc *) Mvar, 1},
    {"mdeg", (lisp_proc *) Mdeg, 1},
    {"mpower", (lisp_proc *) Mlpower, 1},
    {"mlpower", (lisp_proc *) Mlpower, 1},
    {"minit", (lisp_proc *) Minit, 1},
    {"mtail", (lisp_proc *) Mreduct, 1},
    {"mreduct", (lisp_proc *) Mreduct, 1},
    {"mlterm", (lisp_proc *) Mlterm, 1},
    {"mlcoeff", (lisp_proc *) Minit, 1},
    // {"mcoeff", (lisp_proc *)Mcoeff, 2},
    // {"mquot", (lisp_proc *)Mquot, 2},
    // {"mrem", (lisp_proc *)Mrem, 2},
    {"mquot", (lisp_proc *) gquot2, 2},
    {"mrem", (lisp_proc *) gremaind2, 2},
    {"deg", (lisp_proc *) Adeg, 2},
    {"tdeg", (lisp_proc *) Atdeg, 1},
    {"mcdeg", (lisp_proc *) Amtdeg_coeff, 1},
    {"tcdeg", (lisp_proc *) Atdeg_coeff, 2},
    {"poly2dist", (lisp_proc *) Apoly2dist, 2},
    {"expvec", (lisp_proc *) Aexpvec, 1},
    {"expvec2", (lisp_proc *) Aexpvec2, 2},
    {"init", (lisp_proc *) Ainit, 2},
    {"tail", (lisp_proc *) Areduct, 2},
    {"reduct", (lisp_proc *) Areduct, 2},
    {"lterm", (lisp_proc *) Alterm, 2},
    {"lpower", (lisp_proc *) Alpower, 2},
    {"lcoeff", (lisp_proc *) Ainit, 2},
    // {"coeff", (lisp_proc *)Acoeff, 3},
    // {"pquot", (lisp_proc *)Aquot, 3},
    // {"prem", (lisp_proc *)Arem, 3},
    {"pquot", (lisp_proc *) gpquotient2, 3},
    {"prem", (lisp_proc *) gpremaind2, 3},
    // {"diff", (lisp_proc *)Mdiff, 2},
    {"times", (lisp_proc *) Mtimes, 2},
    // {"pquot", (lisp_proc *)Mpquot, 2},
    // {"prem", (lisp_proc *)Mprem, 2},
    {"pgcd", (lisp_proc *) Mpgcd, 2},
    {"Mgcd", (lisp_proc *) Mgcd, 2},
    {"sqfree", (lisp_proc *) Msqfree, 2},
    {"cont", (lisp_proc *) Mcont, 1},
    {"pp", (lisp_proc *) Mpp, 1},
    {"norm", (lisp_proc *) Norm, 1},	// insert
    {"lbibun", (lisp_proc *) Lbibun, 1},
    {"diff", (lisp_proc *) Adiff, 2},
    // {"strem", (lisp_proc *)STrem, 1},
    // {"Float", (lisp_proc *)Float, 1},
    {"is_poly", (lisp_proc *) Lis_poly, 1},
    {"is_number", (lisp_proc *) Lis_number, 1},
    {"compare_var", (lisp_proc *) Mcompare_var, 2},
    {"orderp", (lisp_proc *) Lorderp, 2},
    {"orderpp", (lisp_proc *) Mcompare_var, 2},

    {NULL, 0, 0}
};
// a:=(x+1)*(x+2)*(x+3);
