/*
 * tropical.c             (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/tropical.c,v 2.5 2015/04/23 04:04:47 kako Exp $
 */




static char const rcsid[] =
    "$Id: tropical.c,v 2.5 2015/04/23 04:04:47 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "tri.h"
#include "polynomial.h"
#include "tropical.h"
#include "print.h"



#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)




extern long     compare_term(Sex a, Sex b);
extern long     compare_term_lex(Sex a, Sex b);
extern long     compare_term_tlex(Sex a, Sex b);
extern long     compare_term_trevlex(Sex a, Sex b);
extern long     compare_term_gradings(Sex a, Sex b);

extern long     gradings_matrix_m;
extern long     gradings_matrix_n;
extern long     gradings_matrix[];

/*
 * #define ONEP(x) onep(x) #define ZEROP(x) zerop(x) 
 */
#define ONEP(x) (x == mkfxat(1))
// #define ZEROP(x) (x == mkfxat(0))
#define ZEROP(x)        zerop(x)

/***************************************************************/
/***************************************************************/
/***************************************************************/





void
print_tropical_aux(Sex a, var_list * v, putchar_fn * fn, long first)
{
    Sex             x,
                    p;
    struct var_list vv,
                   *vp;

    x = tropical_var(a);
    p = tropical_pol(a);
    stackcheck();
    push4(x, p, nil, nil);
#define a stack[-1]
#define c stack[0]
#define p stack[-2]
#define x stack[-3]

    vv.var = &x;
    vv.next = v;
    while (p != nil) {		/* 
				 * ( (coef exp) ...)  
				 */
	a = lterm(p);
	p = reduct(p);
	c = power(a);
	vv.pow = &c;
	a = coeff(a);
	if (is_tropical(a)) {
	    print_tropical_aux(a, &vv, fn, first);
	} else {
	    if (!first && !minusp(a))
		fn('+');
	    if (a == mkfxat(1)) {
		for (vp = &vv; vp != NULL; vp = vp->next) {
		    if (*(vp->pow) != mkfxat(0))
			break;
		}
		if (vp == NULL)
		    fn('1');
	    } else
		print_sex(a, fn);
	    for (vp = &vv; vp != NULL; vp = vp->next) {
		if (*(vp->pow) != mkfxat(0)) {
		    print_sex(*(vp->var), fn);
		    if (*(vp->pow) != mkfxat(1)) {
			fn('^');
			print_sex(*(vp->pow), fn);
		    }
		}
	    }
	}
	first = NO;
    }
#undef a
#undef c
#undef p
#undef x
    popv(4);
}

void
print_tropical(Sex a, putchar_fn * fn)
{
    Sex             x,
                    p;
    long            first = YES;

    x = tropical_var(a);
    p = tropical_pol(a);
    stackcheck();
    push4(x, p, nil, nil);
#define a stack[-1]
#define c stack[0]
#define p stack[-2]
#define x stack[-3]

    while (p != nil) {		/* 
				 * ( (coef exp) ...)  
				 */
      again:
	a = lterm(p);
	p = reduct(p);
	c = power(a);
	a = coeff(a);
	if (is_tropical(a)) {
	    if (p == nil && c == mkfxat(0)) {
		x = tropical_var(a);
		p = tropical_pol(a);
		goto again;
	    }
	    if (!first && !minuspP(a))
		fn('+');
	    if (c == mkfxat(0)) {
		fn('(');
		print_sex(a, fn);
		fn(')');
	    } else {
		if (reduct(tropical_pol(a)) == nil) {
		    print_sex(a, fn);
		} else {
		    fn('(');
		    print_sex(a, fn);
		    fn(')');
		}
		fn('*');
		print_sex(x, fn);
		if (c != mkfxat(1)) {
		    fn('^');
		    print_sex(c, fn);
		}
	    }
	} else {
	    if (!first && !minuspP(a))
		fn('+');
	    if (a == mkfxat(-1) && c != mkfxat(0)) {
		fn('-');
	    } else if (a != mkfxat(1) && c != mkfxat(0)) {
		print_sex(a, fn);
		fn('*');
	    } else if (c == mkfxat(0)) {
		print_sex(a, fn);
	    }
	    if (c != mkfxat(0)) {
		print_sex(x, fn);
		if (c != mkfxat(1)) {
		    fn('^');
		    print_sex(c, fn);
		}
	    }
	}
	first = NO;
    }
#undef a
#undef c
#undef p
#undef x
    popv(4);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
Ltropical(Sex a)
{
    return nil;
}

Sex
make_tropical(Sex x, Sex p)
{
    Sex             d;

    if (p == nil || zerop(p))
	return mkfxat(0);
    d = lterm(p);
    if (reduct(p) == nil && power(d) == mkfxat(0))
	return coeff(d);

    stackcheck();
    push2(x, p);
    d = getvector(TAG_VECTOR, T_TROPICAL, sizeof(Tropical));
    pop2(p, x);
    tropical_var(d) = x;
    tropical_pol(d) = p;
    return d;
}

Sex
make_tropical3(Sex a, Sex b, Sex c)
{
    Sex             d;

    if (c == mkfxat(0))
	return b;
    stackcheck();
    push(a);
    b = mkterm(b, c);
    b = cons(b, nil);
    push(b);
    d = getvector(TAG_VECTOR, T_TROPICAL, sizeof(Tropical));
    pop2(b, a);
    tropical_var(d) = a;
    tropical_pol(d) = b;
    return d;
}

Sex
symbol_to_tropical(Sex a)
{
    Sex             b,
                    d;

    stackcheck();
    push(a);
    b = mkterm(mkfxat(1), mkfxat(1));
    b = cons(b, nil);
    push(b);
    d = getvector(TAG_VECTOR, T_TROPICAL, sizeof(Tropical));
    pop2(b, a);
    tropical_var(d) = a;
    tropical_pol(d) = b;
    return d;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2xX(Sex t, Sex b)
{
    Sex             p;

    stackcheck();
    push2(t, b);
    p = cons(nil, nil);
    push4(p, p, nil, nil);
#define t stack[-5]
#define b stack[-4]
#define p stack[-2]
#define bc stack[-1]
#define c stack[0]
    while (b != nil) {
	bc = lterm(b);
	if (power(bc) == mkfxat(0)) {
	    c = gplus2(t, coeff(bc));
	    if (!ZEROP(c)) {
		c = mkterm(c, mkfxat(0));
		p = rcons(p, c);
	    }
	    cdr(p) = reduct(b);
	    goto ax;
	} else if (plusp(power(bc))) {
	    p = rcons(p, bc);
	    b = reduct(b);
	} else {
	    c = mkterm(t, mkfxat(0));
	    p = rcons(p, c);
	    cdr(p) = b;
	    goto ax;
	}
    }
    c = mkterm(t, mkfxat(0));
    p = rcons(p, c);
  ax:
#undef t
#undef b
#undef p
#undef bc
#undef c
    p = stack[-3];
    popv(6);
    return cdr(p);
}




Sex
gplus2mX(Sex a, Sex b)
{
    Sex             x,
                    p;

    if (ZEROP(a))
	return b;

    x = tropical_var(b);
    p = tropical_pol(b);
    stackcheck();
    push(x);
    p = plus2xX(a, p);
    pop(x);
    return make_tropical(x, p);
}

Sex
plus2XX(Sex a, Sex b)
{
    Sex             p;

    stackcheck();
    push2(a, b);
    p = cons(nil, nil);
    push5(p, p, nil, nil, nil);
#define a stack[-6]
#define b stack[-5]
#define p stack[-3]
#define ac stack[-2]
#define bc stack[-1]
#define c stack[0]
    while (a != nil && b != nil) {
	ac = lterm(a);
	bc = lterm(b);
	if (numeq2(power(ac), power(bc))) {
	    c = gplus2(coeff(ac), coeff(bc));
	    if (!ZEROP(c)) {
		c = mkterm(c, power(ac));
		p = rcons(p, c);
	    }
	    a = reduct(a);
	    b = reduct(b);
	} else if (lessp2(power(ac), power(bc))) {
	    p = rcons(p, bc);
	    b = reduct(b);
	} else {
	    p = rcons(p, ac);
	    a = reduct(a);
	}
    }
    if (a != nil)
	cdr(p) = a;
    else if (b != nil)
	cdr(p) = b;
#undef a
#undef b
#undef p
#undef ac
#undef bc
#undef c
    p = cdr(stack[-4]);
    popv(7);
    return p;
}



Sex
gplus2XX(Sex a, Sex b)
{
    long            o;

    o = compare_var(tropical_var(a), tropical_var(b));
    if (o > 0)
	return gplus2mX(a, b);
    else if (o < 0)
	return gplus2mX(b, a);
    stackcheck();
    push(a);
    b = plus2XX(tropical_pol(a), tropical_pol(b));
    pop(a);
    return make_tropical(tropical_var(a), b);
}





/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
times2xX(Sex t, Sex a)
{
    Sex             d,
                    p;

    stackcheck();
    push2(a, t);
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-4]
#define t stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = gtimes2(t, coeff(c));
	if (!ZEROP(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef t
#undef p
#undef c
    p = stack[-2];
    popv(5);
    return cdr(p);
}



Sex
gtimes2mX(Sex t, Sex p)
{
    if (ZEROP(t))
	return t;
    if (ONEP(t))
	return p;
    stackcheck();
    push(p);
    t = times2xX(t, tropical_pol(p));
    pop(p);
    if (t == nil)
	return mkfxat(0);
    return make_tropical(tropical_var(p), t);
}



Sex
times2tX(Sex t, Sex p)
{
    Sex             a;

    if (p == nil)
	return nil;
    stackcheck();
    push5(t, p, nil, nil, nil);
#define t stack[-4]
#define p stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[0]
    c = lterm(p);
    a = gtimes2(coeff(t), coeff(c));
    b = gplus2(power(t), power(c));
    a = mkterm(a, b);
    p = times2tX(t, reduct(p));
    a = cons(a, p);
#undef a
#undef b
#undef c
#undef p
#undef t
    a = stack[-2];
    popv(5);
    return a;
}
#ifdef NEW_POLYMUL
#else

Sex
times2XX(Sex a, Sex b)
{
    Sex             e;

    if (a == nil || b == nil)
	return nil;
    stackcheck();
    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[ 0]
    c = times2pp(reduct(a), b);
    d = times2tp(lterm(a), reduct(b));
    c = plus2pp(c, d);
    a = lterm(a);
    b = lterm(b);
    d = gtimes2(coeff(a), coeff(b));
    e = gplus2(power(a), power(b));
    d = mkterm(d, e);
    e = cons(d, c);
#undef a
#undef b
#undef c
#undef d
    popv(4);
    return e;
}

#endif

Sex
gtimes2XX(Sex a, Sex b)
{
    long            o;

    o = compare_var(tropical_var(a), tropical_var(b));
    if (o > 0)
	return gtimes2mX(a, b);
    else if (o < 0)
	return gtimes2mX(b, a);
    stackcheck();
    push(a);
    b = times2XX(tropical_pol(a), tropical_pol(b));
    pop(a);
    return make_tropical(tropical_var(a), b);
}


Sex
exptiX(Sex a, long n)
{
    Sex             p;
    long            i;

    if (n == 1)
	return a;
    p = lterm(a);
    a = reduct(a);

    stackcheck();
    push2(a, p);
    push5(nil, nil, nil, nil, nil);
#define a stack[-6]
#define p stack[-5]
#define c stack[-4]
#define f stack[-3]
#define q stack[-2]
#define r stack[-1]
#define s stack[ 0]

    c = expti(coeff(p), n);
    if (!ZEROP(c)) {
	q = times2(power(p), mkfxat(n));
	s = mkterm(c, q);
	s = cons(s, nil);
    } else {
	s = nil;
    }
    r = cons(mkterm(mkfxat(1), mkfxat(0)), nil);
    if (a != nil) {
	f = mkfxat(n);
	for (i = 1; i < n; i++) {
	    c = expti(coeff(p), n - i);
	    q = gtimes2(power(p), mkfxat(n - i));
	    c = gtimes2(f, c);
	    f = gtimes2(f, mkfxat(n - i));
	    f = gquot2(f, mkfxat(i + 1));
	    r = times2pp(r, a);
	    if (!ZEROP(c))
		s = plus2pp(s, times2tp(mkterm(c, q), r));
	}
	r = times2pp(r, a);
	s = plus2pp(s, r);
    }
#undef a
#undef p
#undef c
#undef f
#undef q
#undef r
#undef s
    p = stack[0];
    popv(7);
    return p;
}

Sex
gexpt2Xi(Sex a, Sex b)
{
    Sex             x;
    long            n = mkint(b);

    if (n == 0)
	return mkfxat(1);
    else if (n == 1)
	return a;
    x = tropical_var(a);
    a = tropical_pol(a);
    stackcheck();
    push(x);
    a = exptiX(a, n);
    pop(x);
    return make_tropical(x, a);
}




/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
ggcd2XX(Sex a, Sex b)
{
    return aerror2("gcd", a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
