/*
 * variable.c           (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/variable.c,v 2.4 2015/06/26 08:33:54 kako Exp $
 */

static char const rcsid[] =
    "$Id: variable.c,v 2.4 2015/06/26 08:33:54 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"

#define coeff(a)	car(a)
#define power(a)	cdr(a)
#define lterm(a)	car(a)
#define mkterm(c,p)	cons(c, p)
#define reduct(a)	cdr(a)


Sex             intervalitp(Sex);
Sex             negatep(Sex);

#define ONEP(x)		onep(x)
#define ZEROP(x)	zerop(x)

Sex             make_variable(Sex x);
Sex             gnegatev(Sex a);

/***************************************************************/
/***************************************************************/
/***************************************************************/

void
print_variable(Sex a, putchar_fn * fn)
{
    Sex             x;

    x = var_var(a);
    stackcheck();
    push4(x, nil, nil, nil);
#define a stack[-1]
#define c stack[0]
#define p stack[-2]
#define x stack[-3]

    if (is_symhdr(x))
	print_sex(x, fn);
    else {
	p = cdr(x);
	x = car(x);
	print_sex(x, fn);
	fn('(');
	while (p != nil) {
	    print_sex(car(p), fn);
	    p = cdr(p);
	    if (p != nil)
		fn(',');
	}
	fn(')');
    }
#undef a
#undef c
#undef p
#undef x
    popv(4);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
make_variable(Sex x)
{
    Sex             d;

    if (!is_symhdr(x))
	x = getuniq(x);
    stackcheck();
    push(x);
    d = getvector(TAG_VECTOR, T_VAR, sizeof(Variable));
    pop(x);
    var_var(d) = x;
    var_def(d) = nil;
    return d;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
gplus2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gplus2pn(a, b);
}

Sex
gplus2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gplus2np(a, b);
}

Sex
gplus2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gplus2pp(a, b);
}

Sex
gdiff2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gdiff2pn(a, b);
}

Sex
gdiff2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gdiff2np(a, b);
}

Sex
gdiff2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gdiff2pp(a, b);
}

Sex
gnegatev(Sex a)
{
    return make_polynomial(a, mkfxat(-1), mkfxat(1));
}
/***************************************************************/
/***************************************************************/
/***************************************************************/
Sex
gplus2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gplus2pp(a, b);
}

Sex
gplus2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gplus2pp(a, b);
}

Sex
gdiff2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gdiff2pp(a, b);
}

Sex
gdiff2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gdiff2pp(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
Sex
gtimes2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gtimes2pn(a, b);
}

Sex
gtimes2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gtimes2np(a, b);
}

Sex
gtimes2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gtimes2pp(a, b);
}

Sex
gquot2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gquot2pn(a, b);
}

Sex
gquot2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gquot2np(a, b);
}

Sex
gquot2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gquot2pp(a, b);
}

Sex
gratio2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gratio2pn(a, b);
}

Sex
gratio2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gratio2np(a, b);
}

Sex
gratio2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gratio2pp(a, b);
}

Sex
grem2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return grem2pn(a, b);
}

Sex
grem2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return grem2np(a, b);
}

Sex
grem2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return grem2pp(a, b);
}
/***************************************************************/
/***************************************************************/
/***************************************************************/
Sex
gtimes2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gtimes2pp(a, b);
}

Sex
gtimes2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gtimes2pp(a, b);
}


Sex
gquot2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gquot2pp(a, b);
}

Sex
gquot2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gquot2pp(a, b);
}


Sex
gratio2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gratio2pp(a, b);
}

Sex
gratio2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return gratio2pp(a, b);
}


Sex
grem2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return grem2pp(a, b);
}

Sex
grem2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return grem2pp(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
gexpt2vi(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return gexpt2pi(a, b);
}


/***************************************************************/
/***************************************************************/
Sex
ggcd2vn(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return ggcd2pn(a, b);
}

Sex
ggcd2nv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return ggcd2np(a, b);
}

Sex
ggcd2vv(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return ggcd2pp(a, b);
}

Sex
ggcd2vp(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = var_var(a);
    a = symbol_to_polynomial(a);
    pop(b);
    return ggcd2pp(a, b);
}

Sex
ggcd2pv(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = var_var(b);
    b = symbol_to_polynomial(b);
    pop(a);
    return ggcd2pp(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
