/*
 * vector.c             (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/vector.c,v 2.5 2014/07/02 02:58:14 kako Exp $
 */




static char const rcsid[] =
    "$Id: vector.c,v 2.5 2014/07/02 02:58:14 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "read.h"
#include "entries.h"
#include "misc.h"

/***********************************************************************/
/***********************************************************************/

extern Sex      make_vector(long n);
extern Sex      new_vector(long n);
extern Sex      addVector(Sex a, Sex b);
extern Sex      subVector(Sex a, Sex b);
extern Sex      scaleVector(Sex a, Sex b);
extern Sex      scaleaddVector(Sex a, Sex x, Sex b, Sex y);
Sex             vect_subst(Sex a, Sex x, Sex b);
Sex             vect_reval(Sex a);
Sex             vect_diff(Sex a, Sex x);
Sex             gcross(Sex a, Sex b);

/***********************************************************************/
/***********************************************************************/

Sex
make_vector(long n)
{
    long            i;
    Sex             p;
    Sex            *m;

    if (n < 0 || n > 0x0fffffff)
	return aerror1("mkvector", mkfxat(n));

    p = getvector(TAG_VECTOR, T_VECT,
		  sizeof(Vector) + (n - 1) * sizeof(Sex));
    vect_dimx(p) = mkfxat(n);
    m = &vect_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKVECTOR_INIT_VALUE;
    }
    return p;
}

Sex
new_vector(long n)
{
    long            i;
    Sex             p;
    Sex            *m;

    if (n < 0 || n > 0x0fffffff)
	return aerror1("mkvector", mkfxat(n));

    p = getvector(TAG_VECTOR, T_VECT,
		  sizeof(Vector) + (n - 1) * sizeof(Sex));
    vect_dimx(p) = mkfxat(n);
    return p;
}

void
print_vector(Sex a, putchar_fn * fn)
{
    long            n,
                    k;

    push(a);
    n = mkint(vect_dimx(a));
    fn('[');
    for (k = 0; k < n; k++) {
	if (k != 0)
	    fn(',');
	print_sex(vect_value(stack[0], k), fn);
    }
    fn(']');
    popv(1);
}

/***********************************************************************/
/***********************************************************************/


Sex
gplus2VV(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             c,
                    p;

    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("gplus2:", a, b);
    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	p = gplus2(vect_value(a, i), vect_value(b, i));
	vect_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}


Sex
gdiff2VV(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             c,
                    p;

    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("gplus2:", a, b);
    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	p = gdiff2(vect_value(a, i), vect_value(b, i));
	vect_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}


Sex
gtimes2Vx(Sex a, Sex b)
{
    return gtimes2xV(b, a);
}

Sex
gtimes2xV(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             c,
                    p;

    dx = mkint(vect_dimx(b));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	p = gtimes2(a, vect_value(b, i));
	vect_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
gtimes2VV(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             c,
                    p;

    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("gtimes2:", a, b);
    dx = mkint(vect_dimx(a));
    push2(a, b);
    push(nil);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    c = mkfxat(0);
    for (i = 0; i < dx; i++) {
	p = gtimes2(vect_value(a, i), vect_value(b, i));
	c = gplus2(c, p);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}



/***********************************************************************/
/***********************************************************************/


Sex
gratio2Vx(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             c,
                    p;

    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	p = gratio2(vect_value(a, i), b);
	vect_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

/***********************************************************************/
/***********************************************************************/
Sex
addVector(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             p;

    if (!is_vect(a) || !is_vect(b))
	return aerror("addVector");
    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("addVector:", a, b);
    dx = mkint(vect_dimx(a));
    push2(a, b);
#define a stack[-1]
#define b stack[-0]

    for (i = 0; i < dx; i++) {
	p = gplus2(vect_value(a, i), vect_value(b, i));
	vect_value(a, i) = p;
    }
#undef a
#undef b
    a = stack[-1];
    popv(2);
    return a;
}


Sex
subVector(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             p;

    if (!is_vect(a) || !is_vect(b))
	return aerror("addVector");
    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("addVector:", a, b);
    dx = mkint(vect_dimx(a));
    push2(a, b);
#define a stack[-1]
#define b stack[-0]

    for (i = 0; i < dx; i++) {
	p = gdiff2(vect_value(a, i), vect_value(b, i));
	vect_value(a, i) = p;
    }
#undef a
#undef b
    a = stack[-1];
    popv(2);
    return a;
}


Sex
scaleVector(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             p;

    if (!is_vect(a))
	aerror("scaleVector");
    dx = mkint(vect_dimx(a));
    push2(a, b);
#define a stack[-1]
#define b stack[-0]

    for (i = 0; i < dx; i++) {
	p = gtimes2(b, vect_value(a, i));
	vect_value(a, i) = p;
    }
#undef a
#undef b
    a = stack[-1];
    popv(2);
    return a;
}

Sex
scaleaddVector(Sex a, Sex x, Sex b, Sex y)
{
    Sex             dx;
    int             i;
    Sex             p,
                    q;

    if (!is_vect(a) || !is_vect(b))
	return aerror("addVector");
    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("addVector:", a, b);
    dx = mkint(vect_dimx(a));
    push4(a, b, x, y);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define x stack[-2]
#define y stack[-1]
#define p stack[0]

    for (i = 0; i < dx; i++) {
	p = gtimes2(x, vect_value(a, i));
	q = gtimes2(y, vect_value(b, i));
	q = gplus2(p, q);
	vect_value(a, i) = q;
    }
#undef a
#undef b
#undef x
#undef y
#undef p
    a = stack[-4];
    popv(5);
    return a;
}


/***********************************************************************/
/***********************************************************************/

Sex
gexpt2Vi(Sex a, Sex b)
{
    long            n = mkint(b);
    Sex             c;

    if (n == 0)
	return mkfxat(1);
    if (n < 0) {
	push(a);
	b = gtimes2VV(a, a);
	pop(a);
	a = gratio2(a, b);
	n = -n;
    }
    if (n == 1)
	return a;
    push(a);
    c = mkfxat(1);
    push(c);
#define a stack[-1]
#define c stack[0]
    while (n > 1) {
	if (n & 1)
	    c = gtimes2(c, a);
	a = gtimes2(a, a);
	n = n >> 1;
    }
    c = gtimes2(a, c);
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}
/***********************************************************************/
/***********************************************************************/

Sex
vector_converter(Sex a, one_args * cf)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = converter(vect_value(a, i), cf);
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
vector_converter2(Sex a, Sex b, two_args * cf)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = converter2(vect_value(a, i), b, cf);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}


Sex
Aunmodular_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Aunmodular(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Amkmodular_vect(Sex a, Sex b)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Amkmodular(vect_value(a, i), b);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
Amodular_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Amodular(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Afmodular_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Afmodular(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Afloat_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Afloat(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Abfloat_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Abfloat(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Amkefloat_vect(Sex a, Sex f)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define f stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Amkefloat(vect_value(a, i), f);
    }
#undef a
#undef f
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
Amkbefloat_vect(Sex a, Sex f)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push2(a, f);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define f stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Amkbefloat(vect_value(a, i), f);
    }
#undef a
#undef f
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
Aefloat_value_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Aefloat_value(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Aefloat_error_vect(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Aefloat_error(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
gnegateV(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = gnegate(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
gmodulusV(Sex a, Sex b)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = gmodulus(vect_value(a, i), b);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
gsmodulusV(Sex a, Sex b)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    push2(a, b);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = gsmodulus(vect_value(a, i), b);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

/***********************************************************************/
/***********************************************************************/
Sex
vect_diff(Sex a, Sex v)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    stackcheck();
    push2(a, v);
    c = make_vector(dx);
    push(c);
#define a stack[-2]
#define x stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Adiff(vect_value(a, i), x);
    }
#undef a
#undef c
#undef x
    c = stack[-0];
    popv(3);
    return c;
}

/***********************************************************************/
/***********************************************************************/
Sex
vect_subst(Sex a, Sex v, Sex b)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    stackcheck();
    push3(a, v, b);
    c = make_vector(dx);
    push(c);
#define a stack[-3]
#define x stack[-2]
#define b stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Asub(vect_value(a, i), x, b);
    }
#undef a
#undef c
#undef x
#undef b
    c = stack[-0];
    popv(4);
    return c;
}

Sex
vect_reval(Sex a)
{
    long            dx;
    long            i;
    Sex             c;

    dx = mkint(vect_dimx(a));
    stackcheck();
    push(a);
    c = make_vector(dx);
    push(c);
#define a stack[-1]
#define c stack[-0]

    for (i = 0; i < dx; i++) {
	vect_value(c, i) = Areval(vect_value(a, i));
    }
#undef a
#undef c
    c = stack[-0];
    popv(2);
    return c;
}

/***********************************************************************/
/***********************************************************************/

Sex
gcross(Sex a, Sex b)
{
    Sex             dx;
    int             i,
                    n;
    Sex             c,
                    p;
    int             j,
                    k;

    if (vect_dimx(a) != vect_dimx(b))
	return aerror2("gcross:", a, b);
    dx = mkint(vect_dimx(a));
    if (dx != 3)
	return aerror2("gcross:", a, b);
    push2(a, b);
    c = make_vector(dx);
    push2(c, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define q stack[-0]

    for (i = 0; i < 3; i++) {
	j = (i + 1) % 3;
	k = (i + 2) % 3;
	q = gtimes2(vect_value(a, j), vect_value(b, k));
	p = gtimes2(vect_value(a, k), vect_value(b, j));
	p = gplus2(p, q);
	vect_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[-1];
    popv(4);
    return c;
}


/***********************************************************************/
/***********************************************************************/
